OOPS
====

Project home: https://confluence.ecmwf.int/display/OOPS/OOPS+Home

Contact: Marcin Chrust (marcin.chrust@ecmwf.int)

Object Oriented Prediction System (OOPS) is an ECMWF library for generic variational
data assimilation algorithms. The library was designed with the aim of abstracting
the implementation of variational algorithms from implementation of particular models.
This separation of concerns allows for easier development, prototyping and testing of
new algorithms.

OOPS is written in C++ and comes along with a unit testing environment and two toy
models: Lorenz95 (written in C++) and Quasi-geostrophic (QG) model (written in Fortran).

Requirements
------------

Required dependencies:

- CMake --- For use and installation see http://www.cmake.org/
- ecbuild --- ECMWF library of CMake macros, see https://git.ecmwf.int/projects/ECSDK/repos/ecbuild/browse
- eckit --- ECMWF C++ support library, see https://git.ecmwf.int/projects/ECSDK/repos/eckit/browse
- fckit --- ECMWF Fortran support library, see https://git.ecmwf.int/projects/ECSDK/repos/fckit/browse
- Eigen3 -- For FFT and linear algebra operations


Installation
------------

OOPS employs an out-of-source build/install based on CMake.

Make sure ecbuild, eckit and fckit are installed and the ecbuild
executable script is found ( `which ecbuild` ). Following environment variables
help the build system to detect the right dependencies:

```bash
# For finding eckit
ECKIT_PATH               # Path to eckit prefix

# For finding fckit
FCKIT_PATH               # Path to fckit prefix

# For finding Eigen3
Eigen3_DIR               # Path to directory containing Eigen3Config.cmake
```

Proceed with installation as follows

```bash
# Environment --- Edit as needed
OOPS_SRC=$(pwd)
OOPS_BUILD=build
OOPS_INSTALL=$HOME/local

# 1. Create the build directory:
mkdir -p $OOPS_BUILD
cd $OOPS_BUILD

# 2. Run CMake
ecbuild --prefix=$OOPS_INSTALL -- $OOPS_SRC

# 3. Compile / Install
make -j4
make install

# 4. Run CI tests
cd $OOPS_BUILD
ctest
```

Extra flags maybe added to step 2 to fine-tune configuration.

- `--build=DEBUG|RELEASE|BIT` --- Optimisation level
  * DEBUG:   No optimisation (`-O0 -g`)
  * BIT:     Maximum optimisation while remaning bit-reproducible (`-O2`)
  * RELEASE: Maximum optimisation (`-O3`)

Bundle Installation (recommended at ECMWF)
---------------------

It is advised to compile in OOPS in bundle with eckit and fckit:

```bash
# Environment --- Edit as needed
OOPS_BUNDLE_SRC=$(pwd)
OOPS_BUNDLE_BUILD=build
OOPS_BUNDLE_INSTALL=$HOME/local

# 1. Create a CMakeLitst.txt file for the OOPS bundle
cd $OOPS_BUNDLE_SRC
cat > ./CMakeLists.txt <<EOF

cmake_minimum_required( VERSION 3.3.2 FATAL_ERROR )
project( oops-bundle VERSION 1.0 LANGUAGES C CXX Fortran )
set( CMAKE_MODULE_PATH ${CMAKE_MODULE_PATH} )
set( ENABLE_MPI OFF CACHE BOOL "Do not compile with MPI" )

ecbuild_bundle_initialize()
ecbuild_requires_macro_version( 2.7 )

ecbuild_bundle( PROJECT eckit    GIT "ssh://git@git.ecmwf.int:7999/ecsdk/eckit.git"    BRANCH develop )
ecbuild_bundle( PROJECT fckit    GIT "ssh://git@git.ecmwf.int:7999/ecsdk/fckit.git"    BRANCH develop )
ecbuild_bundle( PROJECT atlas    GIT "ssh://git@git.ecmwf.int:7999/ecsdk/atlas.git"    BRANCH develop )
ecbuild_bundle( PROJECT oops     GIT "ssh://git@git.ecmwf.int:7999/oops/oops.git"      BRANCH feature/develop-IFS )

ecbuild_bundle_finalize()

EOF

# 2. Create the build directory:
mkdir -p $OOPS_BUNDLE_BUILD
cd $OOPS_BUNDLE_BUILD

# 3. Run CMake
ecbuild --prefix=$OOPS_BUNDLE_INSTALL -- $OOPS_BUNDLE_SRC

# 4. Compile / Install
make -j4
make install

# 5. Run CI tests
cd $OOPS_BUNDLE_BUILD
ctest
```

Using SABER
-----------

To use SABER with OOPS, the bundle should be as follows:

```bash
# Environment --- Edit as needed
OOPS_BUNDLE_SRC=$(pwd)
OOPS_BUNDLE_BUILD=build
OOPS_BUNDLE_INSTALL=$HOME/local

# 1. Create a CMakeLitst.txt file for the OOPS bundle
cd $OOPS_BUNDLE_SRC
cat > ./CMakeLists.txt <<EOF

cmake_minimum_required( VERSION 3.3.2 FATAL_ERROR )
find_package( ecbuild 3.6 REQUIRED )
project( oops-bundle VERSION 1.0 LANGUAGES C CXX Fortran )
set( CMAKE_MODULE_PATH ${CMAKE_MODULE_PATH} )

# Enable OpenMP and MPI
set( ENABLE_MPI ON CACHE BOOL "Compile with MPI" )
set( ENABLE_OMP ON CACHE BOOL "Compile with OpenMP" )

ecbuild_bundle_initialize()
ecbuild_requires_macro_version( 2.7 )

ecbuild_bundle( PROJECT eckit    GIT "ssh://git@git.ecmwf.int:7999/ecsdk/eckit.git"    BRANCH develop )
ecbuild_bundle( PROJECT fckit    GIT "ssh://git@git.ecmwf.int:7999/ecsdk/fckit.git"    BRANCH develop )
ecbuild_bundle( PROJECT fiat     GIT "ssh://git@git.ecmwf.int:7999/ecsdk/fiat.git"     BRANCH main )
ecbuild_bundle( PROJECT ectrans  GIT "ssh://git@git.ecmwf.int:7999/ecsdk/ectrans.git"  BRANCH main )
ecbuild_bundle( PROJECT atlas    GIT "ssh://git@git.ecmwf.int:7999/ecsdk/atlas.git"    BRANCH develop )
ecbuild_bundle( PROJECT oops     GIT "ssh://git@git.ecmwf.int:7999/oops/oops.git"      BRANCH feature/develop-IFS )
ecbuild_bundle( PROJECT ecsaber  GIT "ssh://git@git.ecmwf.int:7999/sab/ecsaber.git"    BRANCH develop )

ecbuild_bundle_finalize()

EOF

# 2. Create the build directory:
mkdir -p $OOPS_BUNDLE_BUILD
cd $OOPS_BUNDLE_BUILD

# 3. Run CMake
ecbuild --prefix=$OOPS_BUNDLE_INSTALL -DECSABER=ON $OOPS_BUNDLE_SRC

# 4. Compile / Install
make -j4
make install

# 5. Run CI tests
cd $OOPS_BUNDLE_BUILD
ctest
```

Runtime Configuration
---------------------

OOPS behaviour can be configured through some environment variables with defaults marked in square brackets

- `OOPS_DEBUG=<[0]|1>` --- Control printing of OOPS debug information
- `OOPS_TRACE=<[0]|1>` --- Control printing of OOPS traces (includes timings)

Documentation
-------------
OOPS is documented using "annotations" in the source files. It is strongly recommended
to document all the new classes and update documentation where consdidered necessary.

The documentation can be generated using Doxygen and compiled into a convenient web
format following these steps:

```bash
cd $OOPS_SRC/Documents
gmake
firefox html/index.html
```

Contributing
------------

Contributions to OOPS are welcome. In order to do so, please open an issue
in the OOPS JIRA space https://jira.ecmwf.int/projects/OOPS/summary; all developments
should be done in a branch which name includes the JIRA ticket tag. Include the JIRA ticket
tag in commit messages. Issue a pull request on the `feature/develop-IFS` branch when you are
ready with your contribution.
