/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include <memory> // for std::unique_ptr

#include "./TestConfig.h"

#include "eckit/config/LocalConfiguration.h"
#include "eckit/testing/Test.h"

#include "oops/runs/Test.h"

#include "util/DateTime.h"

#include "lorenz95/GomL95.h"
#include "lorenz95/NoVariables.h"
#include "lorenz95/ObservationL95.h"
#include "lorenz95/ObsTable.h"
#include "lorenz95/Resolution.h"

#include "test/TestFixture.h"

namespace test {

// -----------------------------------------------------------------------------
class GomTestFixture : TestFixture {
 public:
  GomTestFixture() {
    const eckit::LocalConfiguration conf(TestConfig::config(), "Observations");
    const util::DateTime bgn(conf.getString("window_begin"));
    const util::DateTime end(conf.getString("window_end"));
    const eckit::LocalConfiguration otconf(conf, "Observation");
    eckit::LocalConfiguration res(TestConfig::config(), "resolution");
    resol_.reset(new lorenz95::Resolution(res));
    ot_.reset(new lorenz95::ObsTable(otconf, *resol_, bgn, end));
    t1_.reset(new util::DateTime("2010-01-01T03:00:00Z"));
    t2_.reset(new util::DateTime("2010-01-02T06:00:00Z"));
    novar_.reset(new lorenz95::NoVariables());
  }
  ~GomTestFixture() {}
  lorenz95::IncrementL95 * incr_;  // Not actually used
  std::unique_ptr<lorenz95::Resolution> resol_;
  std::unique_ptr<lorenz95::ObsTable> ot_;
  std::unique_ptr<util::DateTime> t1_;
  std::unique_ptr<util::DateTime> t2_;
  std::unique_ptr<lorenz95::NoVariables> novar_;
};
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
CASE("test_GomL95") {
  GomTestFixture f;
// -----------------------------------------------------------------------------
  SECTION("test_GomL95_constructor") {
    std::unique_ptr<lorenz95::GomL95> gom(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    EXPECT(gom.get() != NULL);
  }
// -----------------------------------------------------------------------------
  SECTION("test_GomL95_nobs") {
    std::unique_ptr<lorenz95::GomL95> gom(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    EXPECT(gom->nobs() == 160);
  }
// -----------------------------------------------------------------------------
  SECTION("test_gomL95_classname") {
    std::unique_ptr<lorenz95::GomL95> gom(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    EXPECT(gom->classname() == "lorenz95::GomL95");
  }
// -----------------------------------------------------------------------------
  SECTION("test_gomL95_zero") {
    std::unique_ptr<lorenz95::GomL95> gom(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    gom->zero();
    for (int i = 0; i < gom->nobs(); ++i) {
      EXPECT((*gom)[i] == 0.0);
    }
  }
// -----------------------------------------------------------------------------
  SECTION("test_gomL95_dot_product_with") {
    std::unique_ptr<lorenz95::GomL95> gom1(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    gom1->zero();
    std::unique_ptr<lorenz95::GomL95> gom2(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    gom2->zero();

    double zz = gom1->dot_product_with(*gom2);
    EXPECT(zz == 0.0);
  }
// -----------------------------------------------------------------------------
  SECTION("test_gomL95_operator") {
    std::unique_ptr<lorenz95::GomL95> gom1(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    gom1->zero();
    std::unique_ptr<lorenz95::GomL95> gom2(new lorenz95::GomL95(*f.ot_, *f.novar_, *f.incr_, *f.t1_, *f.t2_));
    gom2->zero();

    (*gom1)[1] = 1.0;
    (*gom2)[3] = 1.0;
    double zz = gom1->dot_product_with(*gom2);
    EXPECT(zz == 0.0);

    (*gom1)[3] = 1.0;
    zz = gom1->dot_product_with(*gom2);
    EXPECT(zz == 1.0);
  }
// -----------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

}  // namespace test
int main(int argc, char **argv)
{
    return eckit::testing::run_tests ( argc, argv );
}
