/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/GeometryQG.h"

#include "atlas/field.h"
#include "atlas/functionspace.h"

#include "eckit/config/Configuration.h"

#include "util/Logger.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
GeometryQG::GeometryQG(const eckit::Configuration & conf) {
  const eckit::Configuration * configc = &conf;
  qg_geo_setup_f90(keyGeom_, &configc);

  // Get grid size
  int nx;
  int ny;
  qg_geo_info_f90(keyGeom_, nx, ny);

  // Set ATLAS grid
  eckit::LocalConfiguration confGrid;
  confGrid.set("type", "regional");
  confGrid.set("nx", nx);
  confGrid.set("ny", ny);
  confGrid.set("north", 50);
  confGrid.set("south", 30);
  confGrid.set("east", 10);
  confGrid.set("west", -10);
  atlas::Grid grid(confGrid);

  // Create ATLAS function space
  functionSpace_.reset(new atlas::functionspace::StructuredColumns(grid));

  // Fill ATLAS fieldset
  fields_.reset(new atlas::FieldSet());
  atlas::Field vert_coord = functionSpace_->createField<double>(atlas::option::name("vert_coord")
    | atlas::option::levels(2));
  auto view = atlas::array::make_view<double, 2>(vert_coord);
  for (int jj = 0; jj < functionSpace_->size(); ++jj) {
    view(jj, 0) = 1000.0;
    view(jj, 1) = 3500.0;
  }
  fields_->add(vert_coord);
}
// -----------------------------------------------------------------------------
GeometryQG::GeometryQG(const GeometryQG & other) {
  const int key_geo = other.keyGeom_;
  qg_geo_clone_f90(key_geo, keyGeom_);

  // Copy ATLAS function space
  functionSpace_.reset(new atlas::functionspace::StructuredColumns(other.functionSpace_->grid()));

  // Copy ATLAS fieldset
  fields_.reset(new atlas::FieldSet());
  for (int jfield = 0; jfield < other.fields_->size(); ++jfield) {
    atlas::Field field = other.fields_->field(jfield);
    fields_->add(field);
  }
}
// -----------------------------------------------------------------------------
GeometryQG::~GeometryQG() {
  qg_geo_delete_f90(keyGeom_);
}
// -------------------------------------------------------------------------------------------------
std::vector<size_t> GeometryQG::variableSizes(const VariablesQG & vars) const {
  std::vector<size_t> sizes(vars.variablesList().size(), 2);
  return sizes;
}
// -----------------------------------------------------------------------------
void GeometryQG::print(std::ostream & os) const {
  int nx;
  int ny;
  qg_geo_info_f90(keyGeom_, nx, ny);
  os << "nx = " << nx << ", ny = " << ny;
}
// -----------------------------------------------------------------------------
}  // namespace qg
