/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ObsBiasIncrement.h"

#include <cmath>
#include <iostream>
#include <sstream>
#include <string>
#include <fstream>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/filesystem/LocalPathName.h"
#include "eckit/filesystem/PathName.h"

#include "util/formats.h"
#include "util/Logger.h"

#include "model/ObsBias.h"
#include "model/ObsBiasCovariance.h"

using oops::Log;

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
static oops::ObsAuxIncrementMaker<QgTraits, ObsBiasIncrement> makesObsBiasIncrement_("ObsBiasQG");
// -----------------------------------------------------------------------------
ObsBiasIncrement::ObsBiasIncrement(const ObsSpaceQG &, const eckit::Configuration & conf)
  : active_(ObsBias::ntypes, false), bias_(ObsBias::ntypes, 0.0),
    dummyKey_(0)
{
  active_[0] = conf.has("stream");
  active_[1] = conf.has("uwind");
  active_[2] = conf.has("vwind");
  active_[3] = conf.has("wspeed");
  bool on = false;
  std::string strn = "";
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (jj > 0) strn += ", ";
    if (active_[jj]) {
      strn += "on";
      on = true;
    } else {
      strn += "off";
    }
  }
  if (on) {Log::trace() << "ObsBiasIncrement created : " << strn << std::endl;}
}
// -----------------------------------------------------------------------------
ObsBiasIncrement::ObsBiasIncrement(const ObsBiasIncrement & other,
                                   const bool copy)
  : active_(other.active_), bias_(ObsBias::ntypes, 0.0),
    dummyKey_(other.dummyKey_)
{
  if (copy) {
    for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] = other.bias_[jj];
  }
  this->makePassive();
}
// -----------------------------------------------------------------------------
ObsBiasIncrement::ObsBiasIncrement(const ObsBiasIncrement & other,
                                   const eckit::Configuration &)
  : active_(other.active_), bias_(ObsBias::ntypes, 0.0),
    dummyKey_(other.dummyKey_)
{
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] = other.bias_[jj];
  this->makePassive();
}
// -----------------------------------------------------------------------------
ObsBiasIncrement* ObsBiasIncrement::clone(const bool copy) const {
  return new ObsBiasIncrement(*this, copy);
}
// -----------------------------------------------------------------------------
ObsBiasIncrement* ObsBiasIncrement::clone(const eckit::Configuration & conf) const {
  return new ObsBiasIncrement(*this, conf);
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::makePassive() {
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (!active_[jj]) bias_[jj] = 0.0;
  }
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::diff(const oops::ObsAuxControlBase<QgTraits> & b1,
                            const oops::ObsAuxControlBase<QgTraits> & b2) {
  const ObsBias * pb1 = dynamic_cast<const ObsBias*>(&b1);
  const ObsBias * pb2 = dynamic_cast<const ObsBias*>(&b2);
  ASSERT(pb1 != nullptr && pb2 != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    bias_[jj] = (*pb1)[jj] - (*pb2)[jj];
  }
  this->makePassive();
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::zero() {
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] = 0.0;
}
// -----------------------------------------------------------------------------
ObsBiasIncrement &
ObsBiasIncrement::operator=(const oops::ObsAuxIncrementBase<QgTraits> & rhs) {
  const ObsBiasIncrement * prhs = dynamic_cast<const ObsBiasIncrement*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] = (*prhs)[jj];
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasIncrement &
ObsBiasIncrement::operator+=(const oops::ObsAuxIncrementBase<QgTraits> & rhs) {
  const ObsBiasIncrement * prhs = dynamic_cast<const ObsBiasIncrement*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] += (*prhs)[jj];
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasIncrement &
ObsBiasIncrement::operator-=(const oops::ObsAuxIncrementBase<QgTraits> & rhs) {
  const ObsBiasIncrement * prhs = dynamic_cast<const ObsBiasIncrement*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] -= (*prhs)[jj];
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasIncrement & ObsBiasIncrement::operator*=(const double & fact) {
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] *= fact;
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::axpy(const double & fact,
                            const oops::ObsAuxIncrementBase<QgTraits> & rhs) {
  const ObsBiasIncrement * prhs = dynamic_cast<const ObsBiasIncrement*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) bias_[jj] += fact * (*prhs)[jj];
  this->makePassive();
}
// -----------------------------------------------------------------------------
double ObsBiasIncrement::dot_product_with(const oops::ObsAuxIncrementBase<QgTraits> & rhs) const {
  double zz = 0.0;
  const ObsBiasIncrement * prhs = dynamic_cast<const ObsBiasIncrement*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (active_[jj]) zz += bias_[jj] * (*prhs)[jj];
  }
  return zz;
}
// -----------------------------------------------------------------------------
double ObsBiasIncrement::norm() const {
  double zz = 0.0;
  int ii = 0;
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (active_[jj]) {
      zz += bias_[jj] * bias_[jj];
      ++ii;
    }
  }
  if (ii>0) zz = std::sqrt(zz/ii);
  return zz;
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::print(std::ostream & os) const {
  bool on = false;
  std::string strn = "";
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (jj > 0) strn += ", ";
    if (active_[jj]) {
      on = true;
      std::ostringstream strs;
      strs << bias_[jj];
      strn += strs.str();
    } else {
      strn += "0.0";
    }
  }
  if (on) os << std::endl << "ObsBiasIncrement = " << strn;
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::read(const eckit::Configuration & conf) {
  std::string filename = conf.getString("filename");
  Log::info() << " ObsBiasIncrement::read opening" << filename << std::endl;
  eckit::PathName path = eckit::LocalPathName::cwd() + "/" + filename;
  std::ifstream ifs(path.localPath());
  if (!ifs)
    throw eckit::CantOpenFile(path.localPath(), Here());
  for (size_t ii = 0; ii < ntypes; ++ii) {
    if (!ifs.bad()) {
      std::string bias;
      ifs >> bias;
      if (ifs.good()) {
        bias_[ii] = std::stod(bias);
      } else if (ifs.eof()) {
        throw eckit::ReadError("Failed to read entire ObsBiasIncrement", Here());
      } else if (ifs.fail()) {
        throw eckit::ReadError("Failed to read ObsBiasIncrement", Here());
      }
    }
  }
  Log::info() << " ObsBiasIncrement::read " << util::full_precision(bias_) << std::endl;
}
// -----------------------------------------------------------------------------
void ObsBiasIncrement::write(const eckit::Configuration & conf) const {
  std::string filename = conf.getString("filename");
  Log::info() << " ObsBiasIncrement::write opening" << filename << std::endl;
  eckit::PathName path = eckit::LocalPathName::cwd() + "/" + filename;
  std::ofstream ofs(path.localPath());
  if (!ofs)
    throw eckit::CantOpenFile(path.localPath(), Here());
  for (size_t ii = 0; ii < ntypes; ++ii) {
    ofs << util::full_precision(bias_[ii]) << std::endl;
    if (ofs.fail())
      throw eckit::WriteError("Failed to write ObsBiasIncrement", Here());
  }
  ofs.close();
  Log::info() << " ObsBiasIncrement::write " << util::full_precision(bias_) << std::endl;
}
// -----------------------------------------------------------------------------
}  // namespace qg
