/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef QG_MODEL_STATEQG_H_
#define QG_MODEL_STATEQG_H_

#include <memory> // for std::shared_ptr, std::unique_ptr
#include <ostream>
#include <string>

#include "util/DateTime.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"

#include "model/FieldsQG.h"

namespace eckit {
  class Configuration;
}

namespace qg {
  class GomQG;
  class GeometryQG;
  class IncrementQG;
  class LocationsQG;
  class ModelQG;
  class TlmQG;
  class VariablesQG;

/// QG model state
/*!
 * A State contains everything that is needed to propagate the state
 * forward in time.
 */

// -----------------------------------------------------------------------------
class StateQG : public util::Printable,
                private util::ObjectCounter<StateQG> {
 public:
  static const std::string classname() {return "qg::StateQG";}

/// Constructor, destructor
  StateQG(const GeometryQG &, const ModelQG &,
          const eckit::Configuration &);
  StateQG(const GeometryQG &, const TlmQG &,
          const eckit::Configuration &) {}
  StateQG(const GeometryQG &, const ModelQG &,
          const StateQG &);
  StateQG(const GeometryQG &, const ModelQG &,
          const StateQG &, const eckit::Configuration &);
  StateQG(const StateQG &);
  virtual ~StateQG();
  StateQG & operator=(const StateQG &);

/// Interpolate to observation location
  void interpolate(const LocationsQG &, GomQG &) const;

/// Interpolate full fields
  void forceWith(const StateQG &, const VariablesQG &);

/// Interactions with Increment
  StateQG & operator+=(const IncrementQG &);

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const {return fields_->norm();}
  const util::DateTime & validTime() const {return fields_->time();}
  util::DateTime & validTime() {return fields_->time();}

/// Access to fields
  FieldsQG & fields() {return *fields_;}
  const FieldsQG & fields() const {return *fields_;}

  std::shared_ptr<const GeometryQG> geometry() const {
    return fields_->geometry();
  }

/// Other
  void activateModel();
  void deactivateModel();

  void zero();
  void accumul(const double &, const StateQG &);

/// ATLAS FieldSet
  void toFieldSet(atlas::FieldSet & fset) const {fields_->toFieldSet(fset);}
  void fromFieldSet(const atlas::FieldSet & fset) {fields_->fromFieldSet(fset);}
  const atlas::FieldSet & fieldSet() const {return fields_->fieldSet();}
  atlas::FieldSet & fieldSet() {return fields_->fieldSet();}
  void synchronizeFields() {fields_->synchronizeFields();}

 private:
  void print(std::ostream &) const;
  std::unique_ptr<FieldsQG> fields_;
  std::unique_ptr<FieldsQG> stash_;
};
// -----------------------------------------------------------------------------

}  // namespace qg

#endif  // QG_MODEL_STATEQG_H_
