/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_BMATRIX_H_
#define OOPS_ASSIMILATION_BMATRIX_H_

#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"

namespace oops {

/// The \f$ B \f$ matrix.
/*!
 *  The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply the \f$ B \f$ matrix.
 */

template <typename MODEL>
class BMatrix : private eckit::NonCopyable {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using ObsAuxIncrements_ = ObsAuxIncrements<MODEL>;
  using ObsAuxCtlVecs_ = ObsAuxCtlVecs<MODEL>;

 public:
  explicit BMatrix(const CostFct_& j)
      : j_(j),
        dxObsAux_(j.jb().getFirstGuessIncrement().obsVar(), false),
        dvObsAux_(j.jb().jbObsBias()) {}
  void multiply(const CtrlInc_& x, CtrlInc_& bx) const {
    j_.jb().multiplyB(x, bx);
  }
  void multiplySqrt(const CtrlVec_& dv, CtrlInc_& dx) const {
    j_.jb().multiplyBsqrt(dv, dx);
  }
  void multiplySqrtTrans(const CtrlInc_& dx, CtrlVec_& dv) const {
    j_.jb().multiplyBsqrtTrans(dx, dv);
  }
  void multiplyId(const CtrlInc_& dx, CtrlInc_& jbdx) const {
    /*!          = | I             |
     *             |   B^-1 U'U'^T |
     * where B = UU^T and U' does not have to be equal to U
     */

    // State part
    jbdx.state() = dx.state();

    // Mod Aux part
    jbdx.modVar() = dx.modVar();

    // Obs Aux part
    dvObsAux_.zero();
    dxObsAux_.zero();

    // U'^T
    j_.jb().jbObsBias().multiplySqrtTrans(dx.obsVar(), dvObsAux_);
    // U' U'^T
    j_.jb().jbObsBias().multiplySqrt(dvObsAux_, dxObsAux_);
    // B^-1 U' U'^T
    j_.jb().jbObsBias().inverseMultiply(dxObsAux_, jbdx.obsVar());
  }

  void multiplyB(const CtrlInc_& dx, CtrlInc_& jbdx) const {
    /*!          = | B          |
     *             |     U'U'^T |
     * where B = UU^T and U' does not have to be equal to U
     */

    // State part
    j_.jb().jbState().Bmult(dx.state(), jbdx.state());

    // Mod Aux part
    jbdx.modVar() = dx.modVar();

    // Obs Aux part
    dvObsAux_.zero();

    // U'^T
    j_.jb().jbObsBias().multiplySqrtTrans(dx.obsVar(), dvObsAux_);
    // U' U'^T
    j_.jb().jbObsBias().multiplySqrt(dvObsAux_, jbdx.obsVar());
  }

 private:
  CostFct_ const& j_;

  mutable ObsAuxIncrements_ dxObsAux_;
  mutable ObsAuxCtlVecs_ dvObsAux_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_BMATRIX_H_
