/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_CONTROLINCREMENT_H_
#define OOPS_ASSIMILATION_CONTROLINCREMENT_H_

#include <ostream>
#include <sstream>
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/base/Increment4D.h"
#include "oops/base/ObsAuxIncrements.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/dot_product.h"

namespace oops {

template <typename MODEL>
class CostJbTotal;

/*! @brief Control Increment

 *  The control variable acts as a container for the inputs of the variational
 * data assimilation cost functions in physical space. That includes the states
 * at the start the assimilation window or of each  sub-window but also
 * additional variables such as model bias, VarBC coefficients, or other
 * control variables for algorithms that use them. The control variable
 * increment contains variations of the control variable.

 *  The control increment \f$ dx \f$ has the following structure

 *   \f[ dx = \begin{bmatrix}
               dx_{state4D} \\
               dx_{modaux}  \\
               dx_{obsaux}
             \end{bmatrix}
     \f]

 *   where

     \f[  dx_{state4D} =  \begin{bmatrix} dx^0 \\ dx^1 \\ .. \\ dx^N
 \end{bmatrix} \f]

 *   and \f$ N \f$ is the number of sub-windows.

 *    ->  \f$ dx_{state4D} \f$ is implemented by Increment4D containing a vector
 *       of Increment objects

     \f[  dx_{modaux} = \begin{bmatrix} dx_{modaux}^0 \end{bmatrix} \f]

 *    ->  \f$ dx_{modaux} \f$ is implemented by ModelAuxIncrement

     \f[
          dx_{obsaux} = \begin{bmatrix} dx_{obsaux}^0 \\ dx_{obsaux}^1 \\
                                         ..           \\ dx_{obsaux}^K
                        \end{bmatrix}
     \f]

 *    ->  \f$ dx_{obsaux} \f$ is implemented by ObsAuxIncrements which contains
 *       a vector of ObsAuxIncrement objects (\f$ dx_{obsaux}^i \f$), which in
 *       turn contain a map of ObsAuxIncrementBase objects
 *       (e.g \f$ dx_{VarBC}, dx_{TOVSCV} \f$):

      \f[
        dx_{obsaux}^i = \begin{bmatrix} dx_{obsaux,i}^0 \\
                                 ..  \\ dx_{obsaux,i}^L
                       \end{bmatrix}
      \f]

 *    ->  IFS: K=0, L=1:

      \f[
        dx_{obsaux}= \begin{bmatrix}
                      \begin{bmatrix}
                       dx_{VarBC} \\ dx_{TOVSCV}
                      \end{bmatrix}
                     \end{bmatrix}
      \f]
 */

// -----------------------------------------------------------------------------
template <typename MODEL>
class ControlIncrement : public util::Printable,
                         private util::ObjectCounter<ControlIncrement<MODEL> > {
  using JbTotal_ = CostJbTotal<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using ModelAuxIncr_ = ModelAuxIncrement<MODEL>;
  using ObsAuxIncrs_ = ObsAuxIncrements<MODEL>;

 public:
  static const std::string classname() { return "oops::ControlIncrement"; }

  /// Constructor (using the B matrix to define shape/components)
  explicit ControlIncrement(const JbTotal_ &);
  /// Copy constructor
  ControlIncrement(const ControlIncrement &, const bool copy = true);
  /// Copy constructor with additional configuration
  ControlIncrement(const ControlIncrement &, const eckit::Configuration &);
  /// Copy constructor with interpolation
  ControlIncrement(const Geometry_ &, const ControlIncrement &);
  /// Destructor
  ~ControlIncrement();

  // Linear algebra operators
  /// \f$ dx_1 = 0 \f$
  void zero();
  /// \f$ dx_1 = dx_2 \f$
  ControlIncrement &operator=(const ControlIncrement &);
  /// \f$ dx_1 = dx_1 + dx_2 \f$
  ControlIncrement &operator+=(const ControlIncrement &);
  /// \f$ dx_1 = dx_1 - dx_2 \f$
  ControlIncrement &operator-=(const ControlIncrement &);
  /// \f$ dx_1 = dx_1 * c \f$
  ControlIncrement &operator*=(const double);
  /// \f$ dx_1 = dx_1 * c dx_2 \f$
  void axpy(const double, const ControlIncrement &);
  /// \f$ <dx_1,dx_1> = dx_1^T dx_1 \f$
  double dot_product_with(const ControlIncrement &) const;

  // I/O and diagnostics
  /// Read control increment
  void read(const eckit::Configuration &);
  /// Write control increment
  void write(const eckit::Configuration &) const;

  /// Get geometry
  Geometry_ geometry() const { return incrm4d_.geometry(); }

  /// Get state control increment
  Increment4D_ &state() { return incrm4d_; }
  /// Get constant state control increment
  const Increment4D_ &state() const { return incrm4d_; }

  /// Get augmented model control increment
  ModelAuxIncr_ &modVar() { return modbias_; }
  /// Get constant augmented model control increment
  const ModelAuxIncr_ &modVar() const { return modbias_; }

  /// Get augmented observation control increment
  ObsAuxIncrs_ &obsVar() { return obsbias_; }
  /// Get constant augmented observation control increment
  const ObsAuxIncrs_ &obsVar() const { return obsbias_; }

 private:
  void print(std::ostream &) const;

  Increment4D_ incrm4d_;   ///< 4D increment component
  ModelAuxIncr_ modbias_;  ///< Auxiliary model space component
  ObsAuxIncrs_ obsbias_;   ///< Auxiliary observation space component
};

// =============================================================================

template <typename MODEL>
ControlIncrement<MODEL>::ControlIncrement(const JbTotal_ &jb)
    : incrm4d_(jb.jbState()),
      modbias_(jb.resolution(), jb.jbModBias().config()),
      obsbias_(jb.jbObsBias().obspaces(), jb.jbObsBias().config()) {
  Log::trace() << "ControlIncrement:ControlIncrement created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL>::ControlIncrement(const ControlIncrement &other,
                                          const bool copy)
    : incrm4d_(other.incrm4d_, copy),
      modbias_(other.modbias_, copy),
      obsbias_(other.obsbias_, copy) {
  Log::trace() << "ControlIncrement:ControlIncrement copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL>::ControlIncrement(const ControlIncrement &other,
                                          const eckit::Configuration &tlConf)
    : incrm4d_(other.incrm4d_, tlConf),
      modbias_(other.modbias_, tlConf),
      obsbias_(other.obsbias_, tlConf) {
  Log::trace() << "ControlIncrement:ControlIncrement copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL>::ControlIncrement(const Geometry_ &geom,
                                          const ControlIncrement &other)
    : incrm4d_(geom, other.incrm4d_),
      modbias_(other.modbias_, true),
      obsbias_(other.obsbias_, true) {
  Log::trace() << "ControlIncrement:ControlIncrement copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL>::~ControlIncrement() {}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL> &ControlIncrement<MODEL>::operator=(
    const ControlIncrement &rhs) {
  incrm4d_ = rhs.incrm4d_;
  modbias_ = rhs.modbias_;
  obsbias_ = rhs.obsbias_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL> &ControlIncrement<MODEL>::operator+=(
    const ControlIncrement &rhs) {
  incrm4d_ += rhs.incrm4d_;
  modbias_ += rhs.modbias_;
  obsbias_ += rhs.obsbias_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL> &ControlIncrement<MODEL>::operator-=(
    const ControlIncrement &rhs) {
  incrm4d_ -= rhs.incrm4d_;
  modbias_ -= rhs.modbias_;
  obsbias_ -= rhs.obsbias_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlIncrement<MODEL> &ControlIncrement<MODEL>::operator*=(const double zz) {
  incrm4d_ *= zz;
  modbias_ *= zz;
  obsbias_ *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlIncrement<MODEL>::zero() {
  incrm4d_.zero();
  modbias_.zero();
  obsbias_.zero();
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlIncrement<MODEL>::axpy(const double zz,
                                   const ControlIncrement &rhs) {
  incrm4d_.axpy(zz, rhs.incrm4d_);
  modbias_.axpy(zz, rhs.modbias_);
  obsbias_.axpy(zz, rhs.obsbias_);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlIncrement<MODEL>::read(const eckit::Configuration &config) {
  if (config.has("increment")) incrm4d_.read(config);
  if (config.has("incrementModAux")) modbias_.read(config);
  if (config.has("incrementsObsAux")) obsbias_.read(config);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlIncrement<MODEL>::write(const eckit::Configuration &config) const {
  if (config.has("increment")) incrm4d_.write(config);
  if (config.has("incrementModAux")) modbias_.write(config);
  if (config.has("incrementsObsAux")) obsbias_.write(config);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlIncrement<MODEL>::print(std::ostream &outs) const {
  outs << incrm4d_;
  outs << modbias_;
  outs << obsbias_;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ControlIncrement<MODEL>::dot_product_with(
    const ControlIncrement &x2) const {
  double zz = 0.0;
  zz += dot_product(incrm4d_, x2.incrm4d_);
  zz += dot_product(modbias_, x2.modbias_);
  zz += dot_product(obsbias_, x2.obsbias_);
  return zz;
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_ASSIMILATION_CONTROLINCREMENT_H_
