/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_CONTROLVARIABLE_H_
#define OOPS_ASSIMILATION_CONTROLVARIABLE_H_

#include <cmath>
#include <ostream>
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/State4D.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxControlEstimator.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"

namespace oops {

/*! @brief Control Variable

 *  The control variable acts as a container for the inputs of the variational
 * data assimilation cost functions in physical space.
 * That includes the states at the start the assimilation window or of each
 * sub-window but also additional variables such as model bias, VarBC
 * coefficients, or other control variables for algorithms that use them.
 * This is mostly a convenience class that is used to keep things together
 * and reduce the number of arguments to be passed around.

 *  The control variable \f$ x \f$ has the following structure

 *   \f[ x = \begin{bmatrix}
               x_{state4D} \\
               x_{modaux}  \\
               x_{obsaux}
             \end{bmatrix}
     \f]

 *   where

     \f[  x_{state4D} =  \begin{bmatrix} x^0 \\ x^1 \\ .. \\ x^N \end{bmatrix}
     \f]

 *   and \f$ N \f$ is the number of sub-windows.

 *    ->  \f$ x_{state4D} \f$ is implemented by State4D containing a vector of
 *       State objects

     \f[  x_{modaux} = \begin{bmatrix} x_{modaux}^0 \end{bmatrix} \f]

 *    ->  \f$ x_{modaux} \f$ is implemented by ModelAuxControl

     \f[
          x_{obsaux} = \begin{bmatrix} x_{obsaux}^0 \\ x_{obsaux}^1 \\
                                       ..           \\ x_{obsaux}^K
                       \end{bmatrix}
     \f]

 *    ->  \f$ x_{obsaux} \f$ is implemented by ObsAuxControls which contains a
 *       vector of ObsAuxControl objects (\f$ x_{obsaux}^i \f$), which in
 turns
 *       contain a map of ObsAuxControlBase objects
 *       (e.g \f$ x_{VarBC}, x_{TOVSCV} \f$):

      \f[
        x_{obsaux}^i = \begin{bmatrix} x_{obsaux,i}^0 \\
                                ..  \\ x_{obsaux,i}^L
                       \end{bmatrix}
      \f]

 *    ->  IFS: K=0, L=1:

      \f[
        x_{obsaux}= \begin{bmatrix}
                     \begin{bmatrix}
                       x_{VarBC} \\ x_{TOVSCV}
                     \end{bmatrix}
                    \end{bmatrix}
      \f]

 */

template <typename MODEL>
class ControlVariable;

// -----------------------------------------------------------------------------
template <typename MODEL>
class ControlVariable : public util::Printable,
                        private util::ObjectCounter<ControlVariable<MODEL> > {
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using ModelAuxCtrl_ = ModelAuxControl<MODEL>;
  using ModelAuxCtrlEstimator_ = ModelAuxControlEstimator<MODEL>;
  using ObsAuxCtrls_ = ObsAuxControls<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;
  using State_ = State<MODEL>;
  using State4D_ = State4D<MODEL>;

 public:
  static const std::string classname() { return "oops::ControlVariable"; }

  /// The arguments define the number of sub-windows and the resolution
  ControlVariable(const eckit::Configuration &, const Geometry_ &,
                  const Model_ &, const ObsSpaces_ &);
  ControlVariable(const Geometry_ &, const ControlVariable &);
  ControlVariable(const Geometry_ &, const Model_ &, const ControlVariable &);
  explicit ControlVariable(const ControlVariable &);
  /// Destructor
  ~ControlVariable();

  // I/O and diagnostics
  /// Read a control variable
  void read(const eckit::Configuration &);
  /// Write a control variable
  void write(const eckit::Configuration &) const;
  /// Compute the L-2 norm of the control variable
  double norm() const;

  /// Relase state and model error component
  void releaseStateMaux();

  // Get state control variable
  /// Returns a 4D state component of the control variable
  State4D_ &state() { return *state4d_; }
  /// Returns a constant 4D state component of the control variable
  const State4D_ &state() const { return *state4d_; }

  // Get augmented model control variable
  /// Returns the auxiliary model space component of the control variable
  ModelAuxCtrl_ &modVar() { return *modbias_; }
  ///  Returns the constant auxiliary model space component of the control
  /// variable
  const ModelAuxCtrl_ &modVar() const { return *modbias_; }

  // Get augmented observation control variable
  /// Returns the auxiliary observation space component of the control
  /// variable
  ObsAuxCtrls_ &obsVar() { return obsbias_; }
  ///  Returns the constant auxiliary observation space component of the
  ///  control
  /// variable
  const ObsAuxCtrls_ &obsVar() const { return obsbias_; }

 private:
  ControlVariable &operator=(const ControlVariable &);  // No assignment
  void print(std::ostream &) const;

  std::unique_ptr<State4D_> state4d_;       ///< 4D state component
  std::unique_ptr<ModelAuxCtrl_> modbias_;  ///< Auxiliary model space component
  ObsAuxCtrls_ obsbias_;  ///< Auxiliary observation space component
};

// =============================================================================

template <typename MODEL>
ControlVariable<MODEL>::ControlVariable(const eckit::Configuration &conf,
                                        const Geometry_ &resol,
                                        const Model_ &model,
                                        const ObsSpaces_ &odb)
    : state4d_(new State4D_(eckit::LocalConfiguration(conf, "Jb.Background"),
                            resol, model)),
      modbias_(new ModelAuxCtrl_(resol, model,
                                 conf.getSubConfiguration("Jb.Background"))),
      obsbias_(odb, eckit::LocalConfiguration(conf, "Jo")) {
  if (conf.has("Jb.Background.stateModAuxEstimator")) {
    Log::info() << "ControlVariable<MODEL>::ControlVariable Estimate Model Bias"
                << std::endl;
    //  Setup NN geometry
    const eckit::LocalConfiguration resolConfig(
        conf, "Jb.Background.stateModAuxEstimator.resolution");
    const Geometry_ resolNN(resolConfig);

    //  Setup model error estimator
    const ModelAuxCtrlEstimator_ modauxEstimator(
        resolNN, model, conf.getSubConfiguration("Jb.Background"));
    //  Setup MODERR at lowres
    ModelAuxCtrl_ me(resolNN, *modbias_);

    //  Setup BG predictor at lowres
    State_ bg(resolNN, (*state4d_)[0]);

    if (conf.has("Jb.Background.stateModAuxEstimator.an")) {
      //  Setup AN state predictor
      const eckit::LocalConfiguration anConfig(
          conf, "Jb.Background.stateModAuxEstimator.an");
      State_ anhres(resol, model, anConfig);
      State_ an(resolNN, anhres);

      //  Estimate model error
      modauxEstimator.estimate(bg, an, me);
    } else {
      //  Estimate model error
      modauxEstimator.estimate(bg, me);
    }

    //  Interpolate model error
    modbias_.reset(new ModelAuxCtrl_(resol, me));
  }
  Log::trace() << "ControlVariable contructed" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlVariable<MODEL>::ControlVariable(const ControlVariable &other)
    : state4d_(new State4D_(*other.state4d_)),
      modbias_(new ModelAuxCtrl_(*other.modbias_)),
      obsbias_(other.obsbias_) {
  Log::trace() << "ControlVariable copied" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlVariable<MODEL>::ControlVariable(const Geometry_ &resol,
                                        const ControlVariable &other)
    : state4d_(new State4D_(resol, *other.state4d_)),
      modbias_(new ModelAuxCtrl_(*other.modbias_)),
      obsbias_(other.obsbias_) {
  Log::trace() << "ControlVariable copied with change of resolution"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlVariable<MODEL>::ControlVariable(const Geometry_ &resol,
                                        const Model_ &model,
                                        const ControlVariable &other)
    : state4d_(new State4D_(resol, model, *other.state4d_)),
      modbias_(new ModelAuxCtrl_(*other.modbias_)),
      obsbias_(other.obsbias_) {
  Log::trace() << "ControlVariable copied with change of resolution"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlVariable<MODEL>::~ControlVariable() {
  Log::trace() << "ControlVariable destructed" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ControlVariable<MODEL>::read(const eckit::Configuration &config) {
  ASSERT(state4d_);
  state4d_->read(config);
  if (config.has("stateModAux")) modbias_->read(config);
  if (config.has("statesObsAux")) obsbias_.read(config);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ControlVariable<MODEL>::write(const eckit::Configuration &config) const {
  ASSERT(state4d_);
  state4d_->write(config);
  if (config.has("stateModAux")) modbias_->write(config);
  if (config.has("statesObsAux")) obsbias_.write(config);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ControlVariable<MODEL>::releaseStateMaux() {
  state4d_.reset();
  modbias_.reset();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double ControlVariable<MODEL>::norm() const {
  ASSERT(state4d_);
  double zz = state4d_->norm();
  double zn = zz * zz;
  zz = modbias_->norm();
  zn += zz * zz;
  zz = obsbias_.norm();
  zn += zz * zz;
  return sqrt(zn);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVariable<MODEL>::print(std::ostream &outs) const {
  ASSERT(state4d_);
  outs << *state4d_;
  outs << *modbias_;
  outs << obsbias_;
}

// -----------------------------------------------------------------------------

}  // namespace oops
#endif  // OOPS_ASSIMILATION_CONTROLVARIABLE_H_
