/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_CONTROLVECTOR_H_
#define OOPS_ASSIMILATION_CONTROLVECTOR_H_

#include <ostream>
#include <sstream>
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/base/IncrCtlVec4D.h"
#include "oops/base/ObsAuxCtlVecs.h"
#include "oops/interface/ModelAuxCtlVec.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/dot_product.h"

namespace oops {

template <typename MODEL>
class CostJbTotal;

/*! @brief Control Vector

 *  The control vector acts as a container for the inputs of the variational
 * data assimilation cost functions in the control variable space. That includes
 * the control vectors at the start the assimilation window or of each
 * sub-window but also additional variables such as model bias, VarBC
 * coefficients, or other control variables for algorithms that use them.
 *  The control vector structure is determined by the B matrix.

 *  The control vector \f$ dv \f$ has the following structure

 *   \f[ dv = \begin{bmatrix}
               dv_{state4D} \\
               dv_{modaux}  \\
               dv_{obsaux}
              \end{bmatrix}
     \f]

 *   where

 *       \f[  dv_{state4D} =  \begin{bmatrix} dv^0 \\
                                              dv^1 \\
                                              ..   \\
                                              dv^N
                               \end{bmatrix}
         \f]

 *    and N is the number of sub-windows. For each sub-window:

 *    ->  \f$ dv_{state4D} \f$ is implemented by IncrCtlVec4D containing
 *       a vector of IncrCtlVec objects:

 *    \f[ dv^K = \begin{bmatrix}
                   dv^K_{mod}  \\
                   dv^K_{ens}
                 \end{bmatrix}
               = \begin{bmatrix}
                   dv^K_{mod}  \\
                   \begin{bmatrix}
                    dv^K_{ens_{1}}  \\
                    ..              \\
                    dv^K_{ens_{L}}
                   \end{bmatrix}
                 \end{bmatrix}
      \f]

 *    ->  \f$ dv^K_{mod} \f$ is implemented by IncrModCtlVec

 *    ->  \f$ dv^K_{ens_m} \f$ is implemented by IncrEnsCtlVec, L is the
 *       number of ensemble perturbations used to construct the localized
 *       sampled covariance matrix

      \f[  dv_{modaux} = \begin{bmatrix} dv_{modaux}^0 \end{bmatrix} \f]

 *    ->  \f$ dv_{modaux} \f$ is implemented by ModelAuxCtlVec

     \f[
          dv_{obsaux} = \begin{bmatrix} dv_{obsaux}^0 \\ dv_{obsaux}^1 \\
                                         ..           \\ dv_{obsaux}^K
                        \end{bmatrix}
     \f]

*    ->  \f$ dv_{obsaux} \f$ is implemented by ObsAuxCtlVecs which contains a
*       vector of ObsAuxCtlVec objects (\f$ dv_{obsaux}^i \f$), which in turn
*       contain a map of ObsAuxCtlVecBase objects
*       (e.g \f$ dv_{VarBC}, dv_{TOVSCV} \f$):

      \f[
        dv_{obsaux}^i = \begin{bmatrix} dv_{obsaux,i}^0 \\
                                 ..  \\ dv_{obsaux,i}^L
                        \end{bmatrix}
      \f]

 *    ->  IFS: K=0, L=1:

      \f[
        dv_{obsaux}= \begin{bmatrix}
                      \begin{bmatrix}
                       dv_{VarBC} \\ dv_{TOVSCV}
                      \end{bmatrix}
                     \end{bmatrix}
      \f]

 */

// -----------------------------------------------------------------------------
template <typename MODEL>
class ControlVector : public util::Printable,
                      private util::ObjectCounter<ControlVector<MODEL> > {
  using JbTotal_ = CostJbTotal<MODEL>;
  using IncrCtlVec4D_ = IncrCtlVec4D<MODEL>;
  using ModelAuxCtlVec_ = ModelAuxCtlVec<MODEL>;
  using ObsAuxCtlVecs_ = ObsAuxCtlVecs<MODEL>;

 public:
  static const std::string classname() { return "oops::ControlVector"; }

  /// Constructor
  explicit ControlVector(const JbTotal_ &);
  /// Constructor with copy initialization
  ControlVector(const JbTotal_ &, const ControlVector &);
  /// Copy constructor
  ControlVector(const ControlVector &, const bool copy = true);
  /// Copy constructor with additional configuration
  ControlVector(const ControlVector &, const eckit::Configuration &);
  /// Destructor
  ~ControlVector();

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  void zero();
  /// \f$ dv_1 = dv_2 \f$
  ControlVector &operator=(const ControlVector &);
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  ControlVector &operator+=(const ControlVector &);
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  ControlVector &operator-=(const ControlVector &);
  /// \f$ dv_1 = dv_1 * c \f$
  ControlVector &operator*=(const double);
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  void axpy(const double, const ControlVector &);
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  double dot_product_with(const ControlVector &) const;

  // I/O and diagnostics
  /// Read control vector
  void read(const eckit::Configuration &);
  /// Write control vector
  void write(const eckit::Configuration &) const;

  /// Get state control vector
  IncrCtlVec4D_ &state() { return incrCtlVec4D_; }
  /// Get constant state control vector
  const IncrCtlVec4D_ &state() const { return incrCtlVec4D_; }

  /// Get augmented model control vector
  ModelAuxCtlVec_ &modVar() { return modAuxCtlVec_; }
  /// Get constant augmented model control vector
  const ModelAuxCtlVec_ &modVar() const { return modAuxCtlVec_; }

  /// Get augmented observation control vector
  ObsAuxCtlVecs_ &obsVar() { return obsAuxCtlVecs_; }
  /// Get constant augmented observation control vector
  const ObsAuxCtlVecs_ &obsVar() const { return obsAuxCtlVecs_; }

 private:
  void print(std::ostream &) const;

  IncrCtlVec4D_ incrCtlVec4D_;    ///< 4D state component
  ModelAuxCtlVec_ modAuxCtlVec_;  ///< Auxiliary model space component
  ObsAuxCtlVecs_ obsAuxCtlVecs_;  ///< Auxiliary observation space component
};

// =============================================================================

template <typename MODEL>
ControlVector<MODEL>::ControlVector(const JbTotal_ &jb)
    : incrCtlVec4D_(jb.jbState()),
      modAuxCtlVec_(jb.jbModBias().errorCovariance()),
      obsAuxCtlVecs_(jb.jbObsBias()) {
  Log::trace() << "ControlVector:ControlVector created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL>::ControlVector(const JbTotal_ &jb,
                                    const ControlVector &other)
    : incrCtlVec4D_(jb.jbState()),
      modAuxCtlVec_(jb.jbModBias().errorCovariance()),
      obsAuxCtlVecs_(jb.jbObsBias()) {
  incrCtlVec4D_ = other.incrCtlVec4D_;
  modAuxCtlVec_ = other.modAuxCtlVec_;
  obsAuxCtlVecs_ = other.obsAuxCtlVecs_;
  Log::trace() << "ControlVector:ControlVector created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL>::ControlVector(const ControlVector &other, const bool copy)
    : incrCtlVec4D_(other.incrCtlVec4D_, copy),
      modAuxCtlVec_(other.modAuxCtlVec_, copy),
      obsAuxCtlVecs_(other.obsAuxCtlVecs_, copy) {
  Log::trace() << "ControlVector:ControlVector copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL>::ControlVector(const ControlVector &other,
                                    const eckit::Configuration &tlConf)
    : incrCtlVec4D_(other.incrCtlVec4D_, tlConf),
      modAuxCtlVec_(other.modAuxCtlVec_, tlConf),
      obsAuxCtlVecs_(other.obsAuxCtlVecs_) {
  Log::trace() << "ControlVector:ControlVector copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL>::~ControlVector() {}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL> &ControlVector<MODEL>::operator=(
    const ControlVector &rhs) {
  incrCtlVec4D_ = rhs.incrCtlVec4D_;
  modAuxCtlVec_ = rhs.modAuxCtlVec_;
  obsAuxCtlVecs_ = rhs.obsAuxCtlVecs_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL> &ControlVector<MODEL>::operator+=(
    const ControlVector &rhs) {
  incrCtlVec4D_ += rhs.incrCtlVec4D_;
  modAuxCtlVec_ += rhs.modAuxCtlVec_;
  obsAuxCtlVecs_ += rhs.obsAuxCtlVecs_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL> &ControlVector<MODEL>::operator-=(
    const ControlVector &rhs) {
  incrCtlVec4D_ -= rhs.incrCtlVec4D_;
  modAuxCtlVec_ -= rhs.modAuxCtlVec_;
  obsAuxCtlVecs_ -= rhs.obsAuxCtlVecs_;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlVector<MODEL> &ControlVector<MODEL>::operator*=(const double zz) {
  incrCtlVec4D_ *= zz;
  modAuxCtlVec_ *= zz;
  obsAuxCtlVecs_ *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVector<MODEL>::zero() {
  incrCtlVec4D_.zero();
  modAuxCtlVec_.zero();
  obsAuxCtlVecs_.zero();
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVector<MODEL>::axpy(const double zz, const ControlVector &rhs) {
  incrCtlVec4D_.axpy(zz, rhs.incrCtlVec4D_);
  modAuxCtlVec_.axpy(zz, rhs.modAuxCtlVec_);
  obsAuxCtlVecs_.axpy(zz, rhs.obsAuxCtlVecs_);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVector<MODEL>::read(const eckit::Configuration &config) {
  incrCtlVec4D_.read(config);
  if (config.has("ctlvecModAux")) modAuxCtlVec_.read(config);
  if (config.has("ctlvecsObsAux")) obsAuxCtlVecs_.read(config);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVector<MODEL>::write(const eckit::Configuration &config) const {
  incrCtlVec4D_.write(config);
  if (config.has("ctlvecModAux")) modAuxCtlVec_.write(config);
  if (config.has("ctlvecsObsAux")) obsAuxCtlVecs_.write(config);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVector<MODEL>::print(std::ostream &outs) const {
  outs << incrCtlVec4D_;
  outs << modAuxCtlVec_;
  outs << obsAuxCtlVecs_;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ControlVector<MODEL>::dot_product_with(const ControlVector &x2) const {
  double zz = 0.0;
  zz += dot_product(incrCtlVec4D_, x2.incrCtlVec4D_);
  zz += dot_product(modAuxCtlVec_, x2.modAuxCtlVec_);
  zz += dot_product(obsAuxCtlVecs_, x2.obsAuxCtlVecs_);
  return zz;
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_ASSIMILATION_CONTROLVECTOR_H_
