/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTFCT4DVARHYB4D_H_
#define OOPS_ASSIMILATION_COSTFCT4DVARHYB4D_H_

#include <map>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/CostFct4DVarMF.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

/// 4DVarHyb4D Cost Function in the ARPEGE
/*!
 * This class implements a 4D-Var cost function for ARPEGE with specific TL/AD
 * methods.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class CostFct4DVarHyb4D : public CostFct4DVarMF<MODEL> {
  using Increment_ = Increment<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;
  using Model_ = Model<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  CostFct4DVarHyb4D(const eckit::Configuration &, const Geometry_ &,
                    const Model_ &);
  ~CostFct4DVarHyb4D() {}

  /// Run the tangent linear model with 4D Increment
  void runTL(CtrlInc_ &, PostProcessorTL<Increment_> &,
             PostProcessor<Increment_>,
             const bool idModel = false) const override;

  /// Run the adjoint model with 4D Increment
  void runAD(CtrlInc_ &, PostProcessorAD<Increment_> &,
             PostProcessor<Increment_>,
             const bool idModel = false) const override;

  /// Nullify the 4D increment prior to running the adjoint model
  void zeroAD(CtrlInc_ &) const override;

 private:
  /// Create the \f$J_b\f$  4D term of the cost function
  CostJbState<MODEL> *newJb(const eckit::Configuration &, const Geometry_ &,
                            const CtrlVar_ &) const override;

  util::Duration windowLength_;
  util::DateTime windowBegin_;
  util::DateTime windowEnd_;
  util::Duration windowSub_;
  util::Duration zero_;
  unsigned int ncontrol_;
  const Variables_ ctlvars_;
  const Geometry_ geom_;
};

// =============================================================================

template <typename MODEL>
CostFct4DVarHyb4D<MODEL>::CostFct4DVarHyb4D(const eckit::Configuration &config,
                                            const Geometry_ &geom,
                                            const Model_ &model)
    : CostFct4DVarMF<MODEL>::CostFct4DVarMF(config, geom, model),
      zero_(0),
      ctlvars_(config),
      geom_(geom) {
  windowLength_ = util::Duration(config.getString("window_length"));
  windowBegin_ = util::DateTime(config.getString("window_begin"));
  windowEnd_ = windowBegin_ + windowLength_;
  windowSub_ = util::Duration(config.getString("window_sub"));

  ncontrol_ = windowLength_.toSeconds() / windowSub_.toSeconds();
  ASSERT(windowLength_.toSeconds() == windowSub_.toSeconds() * ncontrol_);
  Log::trace() << "CostFct4DVarHyb4D constructed" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
CostJbState<MODEL> *CostFct4DVarHyb4D<MODEL>::newJb(
    const eckit::Configuration &jbConf, const Geometry_ &geom,
    const CtrlVar_ &xb) const {
  return new CostJb4D<MODEL>(jbConf, geom, ctlvars_, zero_, xb.state());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarHyb4D<MODEL>::runTL(CtrlInc_ &dx,
                                     PostProcessorTL<Increment_> &cost,
                                     PostProcessor<Increment_> post,
                                     const bool idModel) const {
  PostProcessorTL<Increment_> no_cost;
  PostProcessor<Increment_> no_post;

  /// In the IFS, cost.finalizeTL has to be called once after all horizontal
  /// interpolations ( cost.processTL ). But a forecast is called at each
  /// timeslot in order to connect dx with the model object.

  CostFct_::getTLM().forecastTL(dx.state()[0], dx.modVar(), util::Duration(0),
                                no_post, no_cost);
  cost.initializeTL(dx.state()[0], dx.state()[ncontrol_].validTime(),
                    util::Duration(0));

  for (unsigned int jsub = 1; jsub <= ncontrol_; ++jsub) {
    util::DateTime now(windowBegin_ + jsub * windowSub_);

    ASSERT(dx.state()[jsub].validTime() == now);
    CostFct_::getTLM().forecastTL(dx.state()[jsub], dx.modVar(),
                                  util::Duration(0), no_post, no_cost);
    cost.processTL(dx.state()[jsub]);
  }
  cost.finalizeTL(dx.state()[0]);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarHyb4D<MODEL>::zeroAD(CtrlInc_ &dx) const {
  util::DateTime now(windowBegin_);
  for (unsigned int jsub = 0; jsub <= ncontrol_; ++jsub) {
    dx.state()[jsub].zero(now);
    now += windowSub_;
  }
  dx.modVar().zero();
  dx.obsVar().zero();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarHyb4D<MODEL>::runAD(CtrlInc_ &dx,
                                     PostProcessorAD<Increment_> &cost,
                                     PostProcessor<Increment_> post,
                                     const bool idModel) const {
  PostProcessorAD<Increment_> no_cost;
  PostProcessor<Increment_> no_post;

  /// In the IFS, cost.finalizeTL has to be called once after all horizontal
  /// interpolations ( cost.processTL ). But a forecast is called at each
  /// timeslot in order to connect dx with the model object.

  CostFct_::getTLM().forecastAD(dx.state()[ncontrol_], dx.modVar(),
                                util::Duration(0), no_post, no_cost);
  cost.initializeAD(dx.state()[ncontrol_], dx.state()[0].validTime(),
                    util::Duration(0));

  for (int jsub = ncontrol_; jsub >= 1; --jsub) {
    util::DateTime now(windowBegin_ + jsub * windowSub_);

    ASSERT(dx.state()[jsub].validTime() == now);
    CostFct_::getTLM().forecastAD(dx.state()[jsub], dx.modVar(),
                                  util::Duration(0), no_post, no_cost);
    cost.processAD(dx.state()[jsub]);
  }
  CostFct_::getTLM().forecastAD(dx.state()[0], dx.modVar(), util::Duration(0),
                                no_post, no_cost);
  cost.finalizeAD(dx.state()[0]);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTFCT4DVARHYB4D_H_
