/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTFCT4DVARMF_H_
#define OOPS_ASSIMILATION_COSTFCT4DVARMF_H_

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/CostJb3D.h"
#include "oops/assimilation/CostJcDFI.h"
#include "oops/assimilation/CostJo.h"
#include "oops/assimilation/CostTermBase.h"
#include "oops/base/AdvectorHandler.h"
#include "oops/base/DFIProcessor.h"
#include "oops/base/DolphChebyshev.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/StateInfo.h"
#include "oops/base/TrajectorySaverMF.h"
#include "oops/base/WeightedDiff.h"
#include "oops/base/WeightingFct.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/MallocTrim.h"

namespace oops {

/// 4D-Var Cost Function in the ARPEGE
/*!
 * This class implements a specific linearize method for ARPEGE. The model
 * trajectory is calculated at low resolution and the changes of resolution
 * are made with FullPos
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class CostFct4DVarMF : public CostFunction<MODEL> {
  using Increment_ = Increment<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;
  using Model_ = Model<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Advector_ = AdvectorHandler<MODEL>;
  using JqTerm_ = JqTerm<MODEL>;
  using TrajectorySaver_ = TrajectorySaverMF<MODEL>;
  using WeightedDiff_ = WeightedDiff<MODEL, Increment_, State_>;
  using WeightedMean_ = DFIProcessor<MODEL>;

 public:
  CostFct4DVarMF(const eckit::Configuration &, const Geometry_ &,
                 const Model_ &);
  ~CostFct4DVarMF() {}

  /// Evaluate the non-linear cost function
  double evaluate(
      const CtrlVar_ &,
      const eckit::Configuration &config = eckit::LocalConfiguration(),
      PostProcessor<State_> post = PostProcessor<State_>()) const override;

  /// Set up the quadratic cost function. Specific linearization for ARPEGE
  /// 4D-Var
  double linearize(
      const CtrlVar_ &, const eckit::Configuration &,
      PostProcessor<State_> post = PostProcessor<State_>()) override;

  /// Run the tangent linear model
  void runTL(CtrlInc_ &, PostProcessorTL<Increment_> &,
             PostProcessor<Increment_>,
             const bool idModel = false) const override;

  /// Run the adjoint model
  void runAD(CtrlInc_ &, PostProcessorAD<Increment_> &,
             PostProcessor<Increment_>,
             const bool idModel = false) const override;

  /// Nullify the increment prior to running the adjoint model
  void zeroAD(CtrlInc_ &) const override;

  /// Run the non-linear model (HR)
  void runNL(CtrlVar_ &xx, PostProcessor<State_> &pp) const override;

  void setupTrajTL(const CtrlVar_ &xx, const util::Duration &dt,
                   PostProcessor<State_> &pp) const override;

  /// Run the non-linear model (Trajectory)
  void runNL(CtrlVar_ &, const Model_ &model, PostProcessor<State_> &) override;

  /// Get the JoJc cost function value
  const double &getCostJoJc() const override { return costJoJc_; }

  /// Get the Jb cost function value
  const double &getCostJb() const override { return costJb_; }

 private:
  /// Add increment to the first guess ( with FullPos change of resolution )
  void addIncr(CtrlVar_ &, const CtrlInc_ &,
               PostProcessor<Increment_> &) const override;

  /// Create the \f$J_b, J_q\f$ term of the cost function
  CostJbState<MODEL> *newJb(const eckit::Configuration &, const Geometry_ &,
                            const CtrlVar_ &) const override;

  /// Create the \f$J_o\f$ term of the cost function
  CostJo<MODEL> *newJo(const eckit::Configuration &) const override;

  /// Create other terms of the cost function \f$J_c, ..\f$
  CostTermBase<MODEL> *newJc(const eckit::Configuration &,
                             const Geometry_ &) const override;

  double costJoJc_;
  double costJb_;

  util::Duration windowLength_;
  util::DateTime windowBegin_;
  util::DateTime windowEnd_;
  const Geometry_ &geom_;
  const Variables_ ctlvars_;

  /// Model that calculates the trajectory at low resolution.
  std::unique_ptr<Model_> trajectoryModel_;
  /// Guess at low resolution
  std::unique_ptr<State_> fglr_;

  /// biasDFI calculated in the first trajectory and used at the end
  std::unique_ptr<Increment_> biasDFI_;

  // Temporary hack, for writing not filtered analysis files
  mutable int fgnum_;
  const bool write_analyse_;
};

// =============================================================================

template <typename MODEL>
CostFct4DVarMF<MODEL>::CostFct4DVarMF(const eckit::Configuration &config,
                                      const Geometry_ &geom,
                                      const Model_ &model)
    : CostFunction<MODEL>::CostFunction(config, geom, model),
      geom_(geom),
      ctlvars_(config),
      fgnum_(0),
      write_analyse_(config.getBool("write_analyses", false)) {
  Log::trace() << "CostFct4DVarMF:CostFct4DVarMF" << std::endl;
  windowLength_ = util::Duration(config.getString("window_length"));
  windowBegin_ = util::DateTime(config.getString("window_begin"));
  windowEnd_ = windowBegin_ + windowLength_;
  /// Create J components \f$J_b\f$ , \f$J_o\f$, \f$J_c\f$ ...
  this->setupTerms(config);
  Log::trace() << "CostFct4DVarMF constructed" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFct4DVarMF<MODEL>::evaluate(const CtrlVar_ &fguess,
                                       const eckit::Configuration &config,
                                       PostProcessor<State_> post) const {
  Log::trace() << "CostFct4DVarMF::evaluate start" << std::endl;

  /// 1. Setup DFI
  PostProcessor<State_> pp(post);
  std::shared_ptr<WeightedMean_> pdfi;
  if (config.has("dfi")) {
    const eckit::LocalConfiguration dfiConf(config, "dfi");
    const util::Duration dfispan(dfiConf.getString("filter_span"));
    const util::DateTime dfitime(fguess.state()[0].validTime() + dfispan / 2);
    pdfi.reset(new WeightedMean_(dfitime, dfispan, geom_, dfiConf));
    pp.enrollProcessor(pdfi);
  }

  /// 2. Call standard J.evaluate()
  double zzz = this->CostFunction<MODEL>::evaluate(fguess, config, pp);

  /*! 3. Write analysis state in the middle of the window, filtered with
   incremental DFI

   *   \f$\widetilde{x}^{a}=DFI(x^{a})\f$
   *
   *   \f$x_{+3h}^{a}=x_{+3h}^{b}+(\widetilde{x}^{a}-\widetilde{x}^{b})
   *                 =x_{+3h}^{b}+\widetilde{\delta x}\f$
   *
   *   \f$x_{+3h}^{a}=\widetilde{x}^{a}+(x_{+3h}^{b}-\widetilde{x}^{b})
   *                 =x^{b}+biasDFI\f$
   */
  if (config.has("dfi")) {
    const eckit::LocalConfiguration dfiConf(config, "dfi");
    std::unique_ptr<State_> xdfi(pdfi->releaseMean());
    (*xdfi) += (*biasDFI_);
    Log::info()
        << "CostFunction: Write filtered analysis in the middle of the window.";
    xdfi->write(eckit::LocalConfiguration(dfiConf, "output"));
  }

  Log::trace() << "CostFct4DVarMF::evaluate done" << std::endl;
  return zzz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFct4DVarMF<MODEL>::linearize(const CtrlVar_ &fguess,
                                        const eckit::Configuration &innerConf,
                                        PostProcessor<State_> post) {
  /*!  This method sets up the quadratic cost function by linearizing
   *  the non-linear cost function in the specific case of ARPEGE

   *   Jo, Jc, .. (jterms_) terms of the cost function call their 'initialize'
   *  methods which return post-processors. These are subsequently attached to
   *  the non-linear trajectory and do their post-processing prior/during/after
   *  the trajectory evaluation; once the non-linear trajectory is completed,
   *  the terms of the cost function call their 'finalize' method which makes
   *  further use of the output acquired by the post-processors during the
   *  trajectory.
   *
   *    E.g. the 'initialize' method of the Jo term of the cost functions
   *  returns an Observer post-processor. Its task is to compute model
   *  equivalents at observation locations. Once the trajectory evaluation is
   *  completed, the 'finalize' method is called to compute departures, perform
   *  screening if requested, set up the observation error covariance and print
   *  diagnostics.

   *   If required, a DFI bias is calculated.
   *
   *   Prior to the computation of the linearisation trajectory by the
   *  TrajectorySaver, the Jb, Jo, Jc, .. terms of the cost function call their
   *  'initializeTraj' methods which return post-processors. These are
   *  subsequently enrolled into the TrajectorySaver and attached to the
   *  linearisation trajectory computation. Once the linearisation trajectory
   *  computation is completed, the terms of the cost function call the
   *  'finalizeTraj' method which makes further use of the output acquired by
   *  the post-processors during the computation of the linearisation
   *  trajectory.

   *    E.g. the 'initializeTraj' method of the Jo term of the cost function
   *  sets up the linear observation operator and returns an Observer
   *  post-processor. Its task is to compute the linearisation for the linear
   *  observation operator. Once the linearisation trajectory computation is
   *  completed, the 'finailzeTraj' method is called to compute the departures
   *  and print diagnostics.
   */

  /// 1. Read inner loop configuration
  const eckit::LocalConfiguration resConf(innerConf, "resolution");
  const eckit::LocalConfiguration tlmConf(innerConf, "linearmodel");
  std::unique_ptr<eckit::LocalConfiguration> dfiConf;
  const Geometry_ lowres(resConf);
  std::shared_ptr<WeightedDiff_> pdfi;
  std::unique_ptr<WeightingFct> wfct;

  PostProcessor<State_> pp(post);
  pp.enrollProcessor(this->getJterm(0).initialize(fguess, innerConf));

  /// 2. Prepare DFI if required
  ///   If needed, the background +3H is written
  if (innerConf.has("screening.dfi")) {
    dfiConf.reset(new eckit::LocalConfiguration(innerConf, "screening.dfi"));
    const util::Duration dfispan(dfiConf->getString("filter_span"));
    const util::DateTime dfitime(fguess.state()[0].validTime() + dfispan / 2);
    wfct.reset(new DolphChebyshev(*dfiConf));
    pdfi.reset(new WeightedDiff_(dfitime, dfispan, geom_, *dfiConf,
                                 util::Duration(0), *wfct));
    pp.enrollProcessor(pdfi);
    if (dfiConf->has("output")) {
      const eckit::LocalConfiguration outConfig(*dfiConf, "output");
      pp.enrollProcessor(
          new StateWriter<State_>(fguess.state()[0].validTime(), outConfig));
    }
  }

  /// 3. Run non linear model at high resolution with screening, if required
  ///   Calculate non linear Jo (without Jc)
  CtrlVar_ fghr(fguess);
  this->runNL(fghr, pp);
  costJoJc_ = this->getJterm(0).finalize(innerConf);

  /*! 4. Put filter aside for usage after the final trajectory (biasDFI)

   *   \f$\widetilde{x}^{b}=DFI(x^{b})\f$
   *
   *   \f$biasDFI=x_{+3h}^{b}-\widetilde{x}^{b}\f$
   */
  if (innerConf.has("screening.dfi")) {
    biasDFI_.reset(pdfi->releaseDiff());
    (*biasDFI_) *= -1.0;
  }

  /// 5. Calculate first guess at low resolution with FullPos
  std::unique_ptr<CtrlVar_> fglr;
  PostProcessor<State_> no_post;
  const eckit::LocalConfiguration modelConfig(innerConf,
                                              "low_resolution_model");
  trajectoryModel_.reset(new Model_(lowres, modelConfig));
  fglr.reset(new CtrlVar_(lowres, *trajectoryModel_, fguess));
  int status = util::MallocTrim(0);

  // 6. Setup trajectory calculation at low resolution
  PostProcessor<State_> pplr(post);
  std::shared_ptr<JqTerm_> jq(
      this->getJb().initializeTraj(*fglr, lowres, innerConf));
  pplr.enrollProcessor(jq);

  for (unsigned iterm = 0; iterm < this->nterms(); ++iterm) {
    pplr.enrollProcessor(
        this->getJterm(iterm).initializeTraj(*fglr, lowres, innerConf));
  }
  std::shared_ptr<TrajectorySaver_> trajectorySaver(new TrajectorySaver_(
      fglr->state()[0], fglr->modVar(), lowres, this->getTLMs(), tlmConf));
  pplr.enrollProcessor(trajectorySaver);

  std::shared_ptr<Advector_> advector = Advector_::get();
  if (advector) pplr.enrollProcessor(advector);

  /*! 7. Run NL model at low resolution. The PostProcessor object contains:

   *   Only horizontal interpolation of observation operators
   *   Model trajectory saver
   *   Advector builder in the case of Hybrid4D, the wind fields at each
   timeslot of
   *    the 4D ControlVariable is given to the Advector object.
   */
  if (innerConf.has("low_resolution_model")) {
    util::LogbookWriter<std::string> log("MF4DVarTrajectory",
                                         "onlyHorizInterpol");
    this->runNL(*fglr, *trajectoryModel_, pplr);
  }

  /// 8. Calculate Jb value
  costJb_ = this->getJb().finalizeTraj(jq.get(), innerConf);
  /* Do not change Jo.gradFG */
  double costJoJc = 0.0;
  for (unsigned iterm = 1; iterm < this->nterms(); ++iterm) {
    costJoJc += this->getJterm(iterm).finalizeTraj(innerConf);
  }

  /// 9. Print cost function value
  double costJ = costJb_ + costJoJc_;
  Log::test() << "CostFunction: Nonlinear J = " << costJ << std::endl;
  return costJ;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
CostJbState<MODEL> *CostFct4DVarMF<MODEL>::newJb(
    const eckit::Configuration &jbConf, const Geometry_ &geom,
    const CtrlVar_ &xb) const {
  ASSERT(xb.state().checkStatesNumber(1));
  return new CostJb3D<MODEL>(jbConf, geom, ctlvars_, windowLength_,
                             xb.state()[0]);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
CostJo<MODEL> *CostFct4DVarMF<MODEL>::newJo(
    const eckit::Configuration &joConf) const {
  const util::Duration tslot(joConf.getString("timeslot"));
  return new CostJo<MODEL>(joConf, geom_, windowBegin_, windowEnd_, tslot);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
CostTermBase<MODEL> *CostFct4DVarMF<MODEL>::newJc(
    const eckit::Configuration &jcConf, const Geometry_ &geom) const {
  const eckit::LocalConfiguration jcdfi(jcConf, "jcdfi");
  const util::DateTime vt(windowBegin_ + windowLength_ / 2);
  return new CostJcDFI<MODEL>(jcdfi, geom, vt, windowLength_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::runNL(CtrlVar_ &xx,
                                  PostProcessor<State_> &post) const {
  ASSERT(xx.state().checkStatesNumber(1));
  ASSERT(xx.state()[0].validTime() == windowBegin_);
  CostFct_::getModel().forecast(xx.state()[0], xx.modVar(), windowLength_,
                                post);
  ASSERT(xx.state()[0].validTime() == windowEnd_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::setupTrajTL(const CtrlVar_ &xx,
                                        const util::Duration &dt,
                                        PostProcessor<State_> &post) const {
  ASSERT(xx.state().checkStatesNumber(1));
  ASSERT(xx.state()[0].validTime() == windowBegin_);
  CostFct_::getModel().setupTrajTL(xx.state()[0], dt, windowLength_, post);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::runNL(CtrlVar_ &xx, const Model_ &model,
                                  PostProcessor<State_> &post) {
  ASSERT(xx.state().checkStatesNumber(1));
  ASSERT(xx.state()[0].validTime() == windowBegin_);
  fglr_.reset(new State_(xx.state()[0]));
  State_ x0(xx.state()[0]);
  model.forecast(x0, xx.modVar(), windowLength_, post);
  ASSERT(x0.validTime() == windowEnd_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::runTL(CtrlInc_ &dx,
                                  PostProcessorTL<Increment_> &cost,
                                  PostProcessor<Increment_> post,
                                  const bool idModel) const {
  ASSERT(dx.state()[0].validTime() == windowBegin_);
  CostFct_::getTLM().forecastTL(dx.state()[0], dx.modVar(), windowLength_, post,
                                cost, idModel);
  ASSERT(dx.state()[0].validTime() == windowEnd_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::zeroAD(CtrlInc_ &dx) const {
  dx.state()[0].zero(windowEnd_);
  dx.modVar().zero();
  dx.obsVar().zero();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::runAD(CtrlInc_ &dx,
                                  PostProcessorAD<Increment_> &cost,
                                  PostProcessor<Increment_> post,
                                  const bool idModel) const {
  ASSERT(dx.state()[0].validTime() == windowEnd_);
  CostFct_::getTLM().forecastAD(dx.state()[0], dx.modVar(), windowLength_, post,
                                cost, idModel);
  ASSERT(dx.state()[0].validTime() == windowBegin_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarMF<MODEL>::addIncr(CtrlVar_ &xfghr, const CtrlInc_ &dxlr,
                                    PostProcessor<Increment_> &) const {
  ASSERT(xfghr.state().checkStatesNumber(1));

  /// \f$x_{hr}^{g} = {fullpos}_{l2h}(x_{lr}^{g})\f$
  State_ fghr(geom_, *trajectoryModel_, *fglr_);

  /// \f$x_{hr}^{a} = {fullpos}_{l2h}(x_{lr}^{g} + {\delta x}_{lr})\f$
  (*fglr_) += dxlr.state()[0];
  State_ xahr(geom_, *trajectoryModel_, *fglr_);

  /// \f${\delta x}_{hr} = x_{hr}^{a} - x_{hr}^{g}\f$
  Increment_ dxhr(geom_, dxlr.state()[0]);
  dxhr.diff(xahr, fghr);

  /// \f$x_{hr}^{g} += {\delta x}_{hr}\f$
  xfghr.state()[0] += dxhr;

  // Quick fix: write low_resolution analysis for FSOi
  if (write_analyse_) {
    eckit::LocalConfiguration outputConfig;
    std::ostringstream naml;
    naml << "naml_write_anaminbr" << (++fgnum_);
    outputConfig.set("namelist", naml.str());
    fglr_->write(outputConfig);
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTFCT4DVARMF_H_
