/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTFCT4DVARTRAJHR_H_
#define OOPS_ASSIMILATION_COSTFCT4DVARTRAJHR_H_

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/CostFct4DVarMF.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/base/AdvectorHandler.h"
#include "oops/base/DFIProcessor.h"
#include "oops/base/DolphChebyshev.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/TrajectorySaverFP.h"
#include "oops/base/WeightedDiff.h"
#include "oops/base/WeightingFct.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/MallocTrim.h"

namespace oops {

/// 4D-Var Cost Function in the ARPEGE
/*!
 * This class implements a specific linearize method for ARPEGE. The model
 * trajectory is calculated at high resolution and the changes of resolution
 * are made with FullPOS
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class CostFct4DVarTrajHR : public CostFct4DVarMF<MODEL> {
  using Increment_ = Increment<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;
  using Model_ = Model<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Advector_ = AdvectorHandler<MODEL>;
  using JqTerm_ = JqTerm<MODEL>;
  using ObsAuxControls_ = ObsAuxControls<MODEL>;
  using TrajectorySaver_ = TrajectorySaverFP<MODEL>;
  using WeightedDiff_ = WeightedDiff<MODEL, Increment_, State_>;
  using WeightedMean_ = DFIProcessor<MODEL>;

 public:
  CostFct4DVarTrajHR(const eckit::Configuration &, const Geometry_ &,
                     const Model_ &);
  ~CostFct4DVarTrajHR() {}

  /// Evaluate the non-linear cost function
  double evaluate(
      const CtrlVar_ &,
      const eckit::Configuration &config = eckit::LocalConfiguration(),
      PostProcessor<State_> post = PostProcessor<State_>()) const override;

  /// Specific linearization for ARPEGE 4D-Var with trajectory from high
  /// resolution.
  double linearize(
      const CtrlVar_ &, const eckit::Configuration &,
      PostProcessor<State_> post = PostProcessor<State_>()) override;

  /// Get the JoJc cost function value
  const double &getCostJoJc() const override { return costJoJc_; }

  /// Get the Jb cost function value
  const double &getCostJb() const override { return costJb_; }

 private:
  /// Add increment to the first guess ( with FullPos change of resolution )
  void addIncr(CtrlVar_ &, const CtrlInc_ &,
               PostProcessor<Increment_> &) const override;

  double costJoJc_;
  double costJb_;
  const Geometry_ geom_;

  /// Model that calculates the trajectory at low resolution.
  std::unique_ptr<Model_> trajectoryModel_;
  /// Guess at low resolution
  std::unique_ptr<State_> fglr_;

  /// biasDFI calculated in the first trajectory and used at the end.
  std::unique_ptr<Increment_> biasDFI_;
};

// =============================================================================

template <typename MODEL>
CostFct4DVarTrajHR<MODEL>::CostFct4DVarTrajHR(
    const eckit::Configuration &config, const Geometry_ &resol,
    const Model_ &model)
    : CostFct4DVarMF<MODEL>::CostFct4DVarMF(config, resol, model),
      geom_(resol) {}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFct4DVarTrajHR<MODEL>::evaluate(const CtrlVar_ &fguess,
                                           const eckit::Configuration &config,
                                           PostProcessor<State_> post) const {
  Log::trace() << "CostFct4DVarTrajHR::evaluate start" << std::endl;

  /// 1. Setup DFI
  PostProcessor<State_> pp(post);
  std::shared_ptr<WeightedMean_> pdfi;
  if (config.has("dfi")) {
    const eckit::LocalConfiguration dfiConf(config, "dfi");
    const util::Duration dfispan(dfiConf.getString("filter_span"));
    const util::DateTime dfitime(fguess.state()[0].validTime() + dfispan / 2);
    pdfi.reset(new WeightedMean_(dfitime, dfispan, geom_, dfiConf));
    pp.enrollProcessor(pdfi);
  }

  /// 2. Call standard J.evaluate()
  double zzz = this->CostFunction<MODEL>::evaluate(fguess, config, pp);

  /*! 3. Write analysis state in the middle of the window, filtered with
   incremental DFI

   *   \f$\widetilde{x}^{a}=DFI(x^{a})\f$
   *
   *   \f$x_{+3h}^{a}=x_{+3h}^{b}+(\widetilde{x}^{a}-\widetilde{x}^{b})
   *                 =x_{+3h}^{b}+\widetilde{\delta x}\f$
   *
   *   \f$x_{+3h}^{a}=\widetilde{x}^{a}+(x_{+3h}^{b}-\widetilde{x}^{b})
   *                 =x^{b}+biasDFI\f$
   */
  if (config.has("dfi")) {
    const eckit::LocalConfiguration dfiConf(config, "dfi");
    std::unique_ptr<State_> xdfi(pdfi->releaseMean());
    (*xdfi) += (*biasDFI_);
    Log::info()
        << "CostFunction: Write filtered analysis in the middle of the window.";
    xdfi->write(eckit::LocalConfiguration(dfiConf, "output"));
  }

  Log::trace() << "CostFct4DVarTrajHR::evaluate done" << std::endl;
  return zzz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFct4DVarTrajHR<MODEL>::linearize(
    const CtrlVar_ &fguess, const eckit::Configuration &innerConf,
    PostProcessor<State_> post) {
  /*!  This method sets up the quadratic cost function by linearizing
   *  the non-linear cost function in the specific case of ARPEGE

   *   Jo, Jc, .. (jterms_) terms of the cost function call their 'initialize'
   *  methods which return post-processors. These are subsequently attached to
   *  the non-linear trajectory and do their post-processing prior/during/after
   *  the trajectory evaluation; once the non-linear trajectory is completed,
   *  the terms of the cost function call their 'finalize' method which makes
   *  further use of the output acquired by the post-processors during the
   *  trajectory.
   *
   *    E.g. the 'initialize' method of the Jo term of the cost functions
   *  returns an Observer post-processor. Its task is to compute model
   *  equivalents at observation locations. Once the trajectory evaluation is
   *  completed, the 'finalize' method is called to compute departures, perform
   *  screening if requested, set up the observation error covariance and print
   *  diagnostics.
   *
   *   Along with with the post-processors set up by the terms of the cost
   *  function, the TrajectorySaver post-processor is set up. Its role is to
   *  record the interpolated high resolution states at the linear model time
   *  steps during the non-linear trajectory run. These interpolated high
   *  resolution states are subsequently used to force the linearisation
   *  trajectory.
   *

   *   If required, a DFI bias is calculated.
   *
   *   Prior to the computation of the linearisation trajectory by the
   *  TrajectorySaver, the Jb, Jo, Jc, .. terms of the cost function call their
   *  'initializeTraj' methods which return post-processors. These are
   *  subsequently enrolled into the TrajectorySaver and attached to the
   *  linearisation trajectory computation. Once the linearisation trajectory
   *  computation is completed, the terms of the cost function call the
   *  'finalizeTraj' method which makes further use of the output acquired by
   *  the post-processors during the computation of the linearisation
   *  trajectory.

   *    E.g. the 'initializeTraj' method of the Jo term of the cost function
   *  sets up the linear observation operator and returns an Observer
   *  post-processor. Its task is to compute the linearisation for the linear
   *  observation operator. Once the linearisation trajectory computation is
   *  completed, the 'finailzeTraj' method is called to compute the departures
   *  and print diagnostics.
   */

  /// 1. Read inner loop configuration
  PostProcessor<State_> no_post;
  const eckit::LocalConfiguration resConf(innerConf, "resolution");
  const eckit::LocalConfiguration tlmConf(innerConf, "linearmodel");
  std::unique_ptr<eckit::LocalConfiguration> dfiConf;
  const Geometry_ lowres(resConf);
  std::shared_ptr<WeightedDiff_> pdfi;
  std::unique_ptr<WeightingFct> wfct;

  /// 2. Calculate Guess at low resolution for future change of resolution in
  /// addIncr()
  const eckit::LocalConfiguration lowresModelConfig(innerConf,
                                                    "low_resolution_model");
  trajectoryModel_.reset(new Model_(lowres, lowresModelConfig));
  CtrlVar_ fglr(lowres, *trajectoryModel_, fguess);
  fglr_.reset(new State_(fglr.state()[0]));
  int status = util::MallocTrim(0);

  /// 3. Initialize Observation operator and TrajectorySaver, possibly DFI.
  PostProcessor<State_> pp(post);
  pp.enrollProcessor(this->getJterm(0).initialize(fguess, innerConf));

  std::shared_ptr<TrajectorySaver_> trajectorySaver(new TrajectorySaver_(
      fglr.state()[0], fglr.modVar(), lowres, this->getTLMs(), tlmConf));
  pp.enrollProcessor(trajectorySaver);

  if (innerConf.has("screening.dfi")) {
    dfiConf.reset(new eckit::LocalConfiguration(innerConf, "screening.dfi"));
    const util::Duration dfispan(dfiConf->getString("filter_span"));
    const util::DateTime dfitime(fguess.state()[0].validTime() + dfispan / 2);
    wfct.reset(new DolphChebyshev(*dfiConf));
    pdfi.reset(new WeightedDiff_(dfitime, dfispan, geom_, *dfiConf,
                                 util::Duration(0), *wfct));
    pp.enrollProcessor(pdfi);
    if (dfiConf->has("output")) {
      const eckit::LocalConfiguration outConfig(*dfiConf, "output");
      pp.enrollProcessor(
          new StateWriter<State_>(fguess.state()[0].validTime(), outConfig));
    }
  }

  /*! 4. Run non linear model at high resolution with screening, if required

   *   Calculate non linear Jo (without Jc)
   */
  CtrlVar_ fghr(fguess);
  this->runNL(fghr, pp);

  costJoJc_ = this->getJterm(0).finalize(innerConf);

  /*! 5. Put filter aside for usage after the final trajectory (biasDFI)

   *   \f$\widetilde{x}^{b}=DFI(x^{b})\f$
   *
   *   \f$biasDFI=x_{+3h}^{b}-\widetilde{x}^{b}\f$
   */
  if (innerConf.has("screening.dfi")) {
    biasDFI_.reset(pdfi->releaseDiff());
    (*biasDFI_) *= -1.0;
  }

  // TODO:
  //  5. Reset the observation space component of the first guess
  // ObsAuxControls_ &obsvarfg = const_cast<ObsAuxControls_ &>(fglr.obsVar());
  // obsvarfg.reset(*obsSpaces_, "fg", innerConf);

  /*! 6. Record trajectory with:

   *   JbJq trajectory
   *   JoJc trajectory
   *   Advector winds fields along the assimilation window
   */
  std::shared_ptr<JqTerm_> jq(
      this->getJb().initializeTraj(fglr, lowres, innerConf));
  trajectorySaver->enrollProcessor(jq);
  for (unsigned iterm = 0; iterm < this->nterms(); ++iterm) {
    trajectorySaver->enrollProcessor(
        this->getJterm(iterm).initializeTraj(fglr, lowres, innerConf));
  }
  std::shared_ptr<Advector_> advector = Advector_::get();
  if (advector) trajectorySaver->enrollProcessor(advector);

  trajectorySaver->recordTrajectory();

  /// 8. Calculate trajectory cost function value
  costJb_ = this->getJb().finalizeTraj(jq.get(), innerConf);
  double costJoJcTraj = 0.0;
  for (unsigned iterm = 0; iterm < this->nterms(); ++iterm) {
    costJoJcTraj += this->getJterm(iterm).finalizeTraj(innerConf);
  }

  /// 9. Print cost function value
  double costJ = costJb_ + costJoJc_;
  Log::test() << "CostFunction: Nonlinear J = " << costJ << std::endl;
  return costJ;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFct4DVarTrajHR<MODEL>::addIncr(CtrlVar_ &xfghr, const CtrlInc_ &dxlr,
                                        PostProcessor<Increment_> &) const {
  ASSERT(xfghr.state().checkStatesNumber(1));

  /// \f$x_{hr}^{g} = {fullpos}_{l2h}(x_{lr}^{g})\f$
  State_ fghr(geom_, *trajectoryModel_, *fglr_);

  /// \f$x_{hr}^{a} = {fullpos}_{l2h}(x_{lr}^{g} + {\delta x}_{lr})\f$
  (*fglr_) += dxlr.state()[0];
  State_ xahr(geom_, *trajectoryModel_, *fglr_);

  /// \f${\delta x}_{hr} = x_{hr}^{a} - x_{hr}^{g}\f$
  Increment_ dxhr(geom_, dxlr.state()[0]);
  dxhr.diff(xahr, fghr);

  /// \f$x_{hr}^{g} += {\delta x}_{hr}\f$
  xfghr.state()[0] += dxhr;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTFCT4DVARTRAJHR_H_
