/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTFUNCTION_H_
#define OOPS_ASSIMILATION_COSTFUNCTION_H_

#include <map>
#include <memory>  // for std::shared_ptr, std::unique_ptr
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostJbState.h"
#include "oops/assimilation/CostJbTotal.h"
#include "oops/assimilation/CostJo.h"
#include "oops/assimilation/CostTermBase.h"
#include "oops/assimilation/DualVector.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/StateWriter.h"
#include "oops/base/TrajectorySaver.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/LinearModel.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {
template <typename MODEL>
class JqTerm;

template <typename MODEL>
class MMatrix;

// -----------------------------------------------------------------------------

/*! @brief Abstract base class for a cost function

 *  CostFunction defines and manages the computation of all the terms of the
 * variational data assimilation cost function. It is a base class for cost
 * functions specific to 3D-Var, 4D-Var, 4DEns-Var and weak constraint 4D-Var
 * implemented by oops::CostFct3DVar<MODEL>, oops::CostFct4DVar<MODEL>,
 * oops::CostFct4DEnsVar<MODEL> and oops::CostFctWeak<MODEL>, respectively.

 *  Concrete cost functions are produced by CostFactory.

 *  IFS: 4D-Var (CostFct4DVar) is set up by default.
 */

template <typename MODEL>
class CostFunction : private eckit::NonCopyable {
  using CostBase_ = CostTermBase<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using JbTotal_ = CostJbTotal<MODEL>;
  using JqTerm_ = JqTerm<MODEL>;
  using LinearModel_ = LinearModel<MODEL>;
  using Model_ = Model<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::CostFunction"; }

  /// Construct a cost function
  CostFunction(const eckit::Configuration &, const Geometry_ &, const Model_ &);

  /// Destroy a cost function
  virtual ~CostFunction() {}

  /// Evaluate the non-linear cost function
  virtual double evaluate(
      const CtrlVar_ &,
      const eckit::Configuration &config = eckit::LocalConfiguration(),
      PostProcessor<State_> post = PostProcessor<State_>()) const;

  /// Linearise the cost function: set up the quadratic cost function
  virtual double linearize(
      const CtrlVar_ &, const eckit::Configuration &,
      PostProcessor<State_> post = PostProcessor<State_>());

  /// Run the tangent linear model (virtual method)
  virtual void runTL(
      CtrlInc_ &, PostProcessorTL<Increment_> &,
      PostProcessor<Increment_> post = PostProcessor<Increment_>(),
      const bool idModel = false) const = 0;

  /// Run the adjoint model (virtual method)
  virtual void runAD(
      CtrlInc_ &, PostProcessorAD<Increment_> &,
      PostProcessor<Increment_> post = PostProcessor<Increment_>(),
      const bool idModel = false) const = 0;

  /// Nullify the increment prior to running the adjoint model (virtual method)
  virtual void zeroAD(CtrlInc_ &) const = 0;

  /// Run the non-linear model (virtual method)
  virtual void runNL(CtrlVar_ &, PostProcessor<State_> &) const = 0;

  /// Run the non-linear model (virtual method)
  virtual void runNL(CtrlVar_ &xx, const Model_ &, PostProcessor<State_> &pp) {
    ASSERT(false);
  }

  /// Run the non-linear model (virtual method)
  virtual void setupTrajTL(const CtrlVar_ &, const util::Duration &,
                           PostProcessor<State_> &) const = 0;

  /// Add increment to the first guess
  void addIncrement(
      CtrlVar_ &, const CtrlInc_ &,
      PostProcessor<Increment_> post = PostProcessor<Increment_>()) const;

  /// Rest the quadratic cost function
  void resetLinearization();

  /// Compute the cost function gradient at first guess (without Jb)
  void computeGradientFG(CtrlInc_ &) const;

  /// Compute the cost function gradient at first guess (without Jb)
  double computeGradientFG(const CtrlVec_ &, CtrlInc_ &) const;

  /// Checkpoint
  void checkpoint(eckit::LocalConfiguration &) const;

  /// Restart
  void restart(const eckit::Configuration &);

  /// Access \f$ J_b\f$
  const JbTotal_ &jb() const { return *jb_; }

  /// Access \f$ J_b\f$
  const JbTotal_ &jbRestart() const { return *jbRestart_; }

  /// Access terms of the cost function other than \f$J_b\f$
  const CostBase_ &jterm(const unsigned ii) const { return *jterms_[ii]; }

  /// Return the numer of other cost function terms \f$J_o, J_c, ..\f$
  unsigned nterms() const { return jterms_.size(); }

  /// Get the Jb cost function value
  virtual const double &getCostJb() const { return costJb_; }

  /// Get the JoJc cost function value
  virtual const double &getCostJoJc() const { return costJoJc_; }

 protected:
  /// Set up terms of the cost function \f$J_b, J_q, J_o, J_c, ..\f$
  void setupTerms(const eckit::Configuration &);
  /// Access the non-linear model
  const Model_ &getModel() const { return model_; }
  /// Access the linear model
  const LinearModel_ &getTLM(const unsigned isub = 0) const {
    return *tlm_[isub];
  }
  /// Re-centre the background
  void recentreBackground(const eckit::Configuration &);

  /// Non const access to JTerms for linearize() method in concrete
  /// cost-functions
  CostBase_ &getJterm(const unsigned ii) const { return *jterms_[ii]; }

  /// Non const access to Jb for linearize() method in concrete cost-functions
  JbTotal_ &getJb() const { return *jb_; }

  /// Non const access to TL Model for linearize() method in concrete
  /// cost-functions
  std::vector<std::unique_ptr<LinearModel_> > &getTLMs() { return tlm_; }

 private:
  /// Virtual method to add the increment to the state
  virtual void addIncr(CtrlVar_ &, const CtrlInc_ &,
                       PostProcessor<Increment_> &) const = 0;

  /// Create the \f$J_b, J_q\f$ term of the cost function (virtual method)
  virtual CostJbState<MODEL> *newJb(const eckit::Configuration &,
                                    const Geometry_ &,
                                    const CtrlVar_ &) const = 0;
  /// Create the \f$J_o\f$ term of the cost function (virtual method)
  virtual CostJo<MODEL> *newJo(const eckit::Configuration &) const = 0;
  /// Create other terms of the cost function \f$J_c, ..\f$ (virtual method)
  virtual CostTermBase<MODEL> *newJc(const eckit::Configuration &,
                                     const Geometry_ &) const = 0;

  /// This method is initializing JbRestart to allow reading in control vector
  /// at their original resolution
  void initializeTrajJbRestart(const CtrlVar_ &, const eckit::Configuration &);

  /// This method is finalizing JbRestart to allow reading in control vector at
  /// their original resolution
  void finalizeTrajJbRestart(const eckit::Configuration &);

  /// Reduce memory footprit (for multi-executable mode only)
  const bool &reduceMemoryFootprint() const { return lReduceMemoryFootprint_; }
  void setReduceMemoryFootprint(const bool value) {
    lReduceMemoryFootprint_ = value;
  }

  /// Is the CostFunction restarted (skip NL traj)
  void setIsRestarted(const bool &isRestarted) const {
    isRestarted_ = isRestarted;
  }

  /// Is the CostFunction restarted (skip NL traj)
  const bool &isRestarted() const { return isRestarted_; }

  /// Data members:

  /// Configuration
  eckit::LocalConfiguration conf_;

  // Data members
  const Geometry_ &resol_;  ///< Outer-loop geometry
  const Model_ &model_;     ///< Non-linear model

  /// Terms of the cost function
  std::unique_ptr<CtrlVar_> xb_;  ///< Background control variable
  std::unique_ptr<JbTotal_> jb_;  ///< \f$J_b, J_q\f$ term of the cost function
  std::unique_ptr<JbTotal_> jbRestart_;
  std::shared_ptr<JqTerm_> jqRestart_;
  std::vector<std::unique_ptr<CostBase_> >
      jterms_;  ///< \f$J_o, J_c, ..\f$ terms of the cost function

  /// Linear model (includes trajectory)
  std::vector<std::unique_ptr<LinearModel_> > tlm_;  ///< Tangent-linear model

  /// Reduce memory footprint (only in multi-executable mode)
  bool lReduceMemoryFootprint_;

  /// Skip non-linear trajectory (in case of restart)
  mutable bool isRestarted_;

  double costJb_;
  double costJoJc_;

  /// Friend classes requiring access to protected members
  friend class MMatrix<MODEL>;
};

// -----------------------------------------------------------------------------

/// Cost Function Factory
template <typename MODEL>
class CostFactory {
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;

 public:
  static CostFunction<MODEL> *create(const eckit::Configuration &,
                                     const Geometry_ &, const Model_ &);
  virtual ~CostFactory() { getMakers().clear(); }

 protected:
  explicit CostFactory(const std::string &);

 private:
  virtual CostFunction<MODEL> *make(const eckit::Configuration &,
                                    const Geometry_ &, const Model_ &) = 0;
  static std::map<std::string, CostFactory<MODEL> *> &getMakers() {
    static std::map<std::string, CostFactory<MODEL> *> makers_;
    return makers_;
  }
};

template <class MODEL, class FCT>
class CostMaker : public CostFactory<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;

 private:
  virtual CostFunction<MODEL> *make(const eckit::Configuration &config,
                                    const Geometry_ &resol,
                                    const Model_ &model) {
    return new FCT(config, resol, model);
  }

 public:
  explicit CostMaker(const std::string &name) : CostFactory<MODEL>(name) {}
};

// =============================================================================

//  Factory
// -----------------------------------------------------------------------------

template <typename MODEL>
CostFactory<MODEL>::CostFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in cost function factory."
                 << std::endl;
    ABORT("Element already registered in CostFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
CostFunction<MODEL> *CostFactory<MODEL>::create(
    const eckit::Configuration &config, const Geometry_ &resol,
    const Model_ &model) {
  std::string id = config.getString("cost_type");
  Log::trace() << "Variational Assimilation Type=" << id << std::endl;
  typename std::map<std::string, CostFactory<MODEL> *>::iterator j =
      getMakers().find(id);
  if (j == getMakers().end()) {
    Log::error() << id << " does not exist in cost function factory."
                 << std::endl;
    ABORT("Element does not exist in CostFactory.");
  }
  return (*j).second->make(config, resol, model);
}

// -----------------------------------------------------------------------------
//  Cost Function
// -----------------------------------------------------------------------------

template <typename MODEL>
CostFunction<MODEL>::CostFunction(const eckit::Configuration &config,
                                  const Geometry_ &resol, const Model_ &model)
    : conf_(config),
      resol_(resol),
      model_(model),
      xb_(),
      jb_(),
      jbRestart_(),
      jqRestart_(),
      jterms_(),
      tlm_(),
      lReduceMemoryFootprint_(false),
      isRestarted_(false),
      costJb_(0.0),
      costJoJc_(0.0) {
  /// The constructor sets the geomery and model objects
  Log::trace() << "CostFunction:created" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::setupTerms(const eckit::Configuration &config) {
  Log::trace() << "CostFunction::setupTerms starting" << std::endl;
  /// This method sets up the terms of the non-linear cost function
  /// 1. \f$J_o\f$
  eckit::LocalConfiguration joConf(config, "Jo");
  CostJo<MODEL> *jo = this->newJo(joConf);
  jterms_.emplace_back(jo);
  Log::trace() << "CostFunction: setupTerms Jo added" << std::endl;

  /// 2. \f$J_b\f$
  xb_.reset(new CtrlVar_(config, resol_, model_, jo->obspaces()));
  if (config.has("Jb")) {
    const eckit::LocalConfiguration jbConf(config, "Jb");
    if (jbConf.has("RecentreBackground"))
      this->recentreBackground(
          jbConf.getSubConfiguration("RecentreBackground"));
    jb_.reset(new JbTotal_(*xb_, this->newJb(jbConf, resol_, *xb_), config,
                           resol_, jo->obspaces()));
    Log::trace() << "CostFunction::setupTerms Jb added" << std::endl;
  }

  /// 3. Set the background in the \f$J_o\f$ term
  /// MC: \f$J_b\f$ requires the \f$J_o\f$ term to setup the observation space
  ///     first in order to initialize observation space components
  jo->setBackground(*xb_);

  /// 4. Other constraints: \f$J_c, ..\f$
  std::vector<eckit::LocalConfiguration> jcs;
  config.get("Jc", jcs);
  for (size_t jj = 0; jj < jcs.size(); ++jj) {
    CostTermBase<MODEL> *jc = this->newJc(jcs[jj], resol_);
    jterms_.emplace_back(jc);
  }
  Log::trace() << "CostFunction::setupTerms done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFunction<MODEL>::evaluate(const CtrlVar_ &fguess,
                                     const eckit::Configuration &config,
                                     PostProcessor<State_> post) const {
  /// This method evaluates the final non-linear cost function and returns the
  /// cost function value
  util::LogbookWriter<bool> log("NonlinearCostFunction", true);
  Log::trace() << "CostFunction::evaluate start" << std::endl;

  /// 1. Setup terms of cost function: \f$J_b, J_q, J_o, J_c\f$
  PostProcessor<State_> pp(post);
  for (auto &jterm : jterms_) {
    pp.enrollProcessor(jterm->initialize(fguess, config));
  }

  /// 2. Run the non-linear model
  std::unique_ptr<CtrlVar_> mfguess(new CtrlVar_(fguess));
  this->runNL(*mfguess, pp);
  mfguess.reset();

  /// 3. Compute the non-linear cost function value
  double zzz = 0.0;
  for (auto &jterm : jterms_) {
    zzz += jterm->finalize(config);
  }

  Log::trace() << "CostFunction::evaluate done" << std::endl;
  return zzz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFunction<MODEL>::linearize(const CtrlVar_ &fguess,
                                      const eckit::Configuration &innerConf,
                                      PostProcessor<State_> post) {
  /*!  This method sets up the quadratic cost function by linearizing
   *  the non-linear cost function.

   *   Jo, Jc, .. (jterms_) terms of the cost function call their 'initialize'
   *  methods which return post-processors. These are subsequently attached to
   *  the non-linear trajectory and do their post-processing prior/during/after
   *  the trajectory evaluation; once the non-linear trajectory is completed,
   *  the terms of the cost function call their 'finalize' method which makes
   *  further use of the output acquired by the post-processors during the
   *  trajectory.
   *
   *    E.g. the 'initialize' method of the Jo term of the cost functions
   *  returns an Observer post-processor. Its task is to compute model
   *  equivalents at observation locations. Once the trajectory evaluation is
   *  completed, the 'finalize' method is called to compute departures, perform
   *  screening if requested, set up the observation error covariance and print
   *  diagnostics.

   *   Along with with the post-processors set up by the terms of the cost
   *  function, the TrajectorySaver post-processor is set up. Its role is to
   *  record the interpolated high resolution states at the linear model time
   *  steps during the non-linear trajectory run. These interpolated high
   *  resolution states are subsequently used to force the linearisation
   *  trajectory.
   *
   *   Prior to the computation of the linearisation trajectory by the
   *  TrajecotrySaver, the Jb, Jo, Jc, .. terms of the cost function call their
   *  'initializeTraj' methods which return post-processors. These are
   *  subsequently enrolled into the TrajecotrySaver and attached to the
   *  linearisation trajectory computation. Once the linearisation trajectory
   *  computation is completed, the terms of the cost function call the
   *  'finalizeTraj' method which makes further use of the output acquired by
   *  the post-processors during the computation of the linearisation
   *  trajectory.

   *    E.g. the 'initializeTraj' method of the Jo term of the cost function
   *  sets up the linear observation operator and returns an Observer
   *  post-processor. Its task is to compute the linearisation for the linear
   *  observation operator. Once the linearisation trajectory computation is
   *  completed, the 'finailzeTraj' method is called to compute the departures
   *  and print diagnostics.
   */

  /// 0. Check if the minimization is skipped
  const bool lSkipMinimization = innerConf.getBool("skipMinimization", false);

  /// 1. Enable the option to reduce the memory footprint in case late 4d start
  ///   is enabled to make more memory available prior to writing out 4V/AN
  ///   states
  if (innerConf.has("late4d")) {
    this->setReduceMemoryFootprint(true);
    Log::info() << "CostFunction::linearize reduced memory footprint enabled"
                << std::endl;
  }

  /// 2. Read the inner loop configuration
  const eckit::LocalConfiguration resConf(innerConf, "resolution");
  const eckit::LocalConfiguration tlmConf(innerConf, "linearmodel");

  /// 3. Set up the inner loop geomerty
  const Geometry_ lowres(resConf);

  /// 4. Set up the TrajectorySaver post-processor
  PostProcessor<State_> pp(post);
  std::shared_ptr<TrajectorySaver<MODEL> > trajectorySaver(
      new TrajectorySaver<MODEL>(fguess.state()[0], fguess.modVar(), lowres,
                                 tlm_, tlmConf, isRestarted()));
  pp.enrollProcessor(trajectorySaver);

  if (isRestarted()) {
    util::Duration dt(tlmConf.getString("tstep"));
    /// Set up the linearisation trajectory
    this->setupTrajTL(fguess, dt, pp);
  } else {
    /// 5.A Optional diagnostic output
    if (innerConf.has("output")) {
      const eckit::LocalConfiguration outConfig(innerConf, "output");
      const util::DateTime bgndate(fguess.state()[0].validTime());
      pp.enrollProcessor(new StateWriter<State_>(bgndate, outConfig));
    }

    /// 5.B Release background if the minimization is skipped
    if (lSkipMinimization) xb_->releaseStateMaux();

    /// 6. Evaluate the non-linear cost function
    costJoJc_ = this->evaluate(fguess, innerConf, pp);

    ///  Do not linearise the cost function if the minimization is
    //  skipped
    if (lSkipMinimization) return 0;
  }

  /*! 7. Compute the linearisation trajectory

   *   The TrajecotrySaver was attached to the non-linear model trajecotry in
   *  order to record the interpolated high resolution states at each linear
   *  model time step. Terms of the cost function (Jb, Jq, Jo, Jc) set up post
   *  processors which perform operations required to set up the quadratic cost
   *  function. These post-processors are enrolled into the TrajectorySaver.
   *  They do their post processing during the forced (with store interpolated
   *  high resolution fields) linearisation trajectory computation.

   *   Such a design was required for IFS as the linearisation of the
   *  observation operator needs to be performed with the screened ODB (CCMA
   *  rather than ECMA).
   */

  /// Initialize JbRestart for reading in control vectors at their original
  /// resolution
  this->initializeTrajJbRestart(fguess, innerConf);

  // JbJq trajectory
  std::shared_ptr<JqTerm_> jq(jb_->initializeTraj(fguess, lowres, innerConf));
  trajectorySaver->enrollProcessor(jq);
  // JoJc trajectory
  for (auto &jterm : jterms_) {
    trajectorySaver->enrollProcessor(
        jterm->initializeTraj(fguess, lowres, innerConf));
  }

  trajectorySaver->recordTrajectory();

  /// 8. Finalize setting up the terms of the quadratic cost function
  costJb_ = jb_->finalizeTraj(jq.get(), innerConf);
  double costJoJcTraj = 0.0;
  for (auto &jterm : jterms_) {
    costJoJcTraj +=
        jterm->finalizeTraj(innerConf.getSubConfiguration("diagnostics"));
  }

  /// 9. Print the cost function value
  double costJ = costJb_ + costJoJc_;
  double costJtraj = costJb_ + costJoJcTraj;
  Log::test() << "CostFunction: Nonlinear J = " << costJ << std::endl;
  Log::info() << "CostFunction: Nonlinear J (traj) = " << costJtraj
              << std::endl;

  /// Finalize JbRestart for reading in control vectors at their original
  /// resolution
  this->finalizeTrajJbRestart(innerConf);

  /// 10. Reset xb
  if (reduceMemoryFootprint()) xb_.reset();

  return costJ;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void CostFunction<MODEL>::computeGradientFG(CtrlInc_ &grad) const {
  /// This method computes and returns the first guess gradient in model space:
  /// \f$H^T R^{-1} d\f$

  /// 1. Set up post-processors for computation of the first guess gradient for
  /// \f$J_o\f$ and \f$J_c\f$
  PostProcessor<Increment_> pp;
  PostProcessorAD<Increment_> costad;
  this->zeroAD(grad);

  for (auto &jterm : jterms_) {
    std::shared_ptr<const GeneralizedDepartures> tmp(jterm->newGradientFG());
    costad.enrollProcessor(jterm->setupAD(tmp, grad));
  }

  /// 2. Run the adjoint model with attached post-processors to compute the
  /// gradient
  this->runAD(grad, costad, pp);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostFunction<MODEL>::computeGradientFG(const CtrlVec_ &dv0,
                                              CtrlInc_ &grad) const {
  /// This method computes and returns the first guess gradient in control
  /// space:
  /// \f$H^T R^{-1} [d + H (dx_{fg} - U dv_0) ]\f$

  /// 1. Compute \f$dx_{fg} - U dv_0\f$
  CtrlInc_ dxfg(this->jb().getFirstGuessIncrement(), true);
  CtrlInc_ dx(this->jb().getFirstGuessIncrement(), false);

  this->jb().multiplyBsqrt(dv0, dx);
  dxfg -= dx;

  /// 2. Compute \f$H (dx_{fg} - U dv_0)\f$
  PostProcessorTL<Increment_> costtl;
  for (unsigned jj = 0; jj < this->nterms(); ++jj) {
    costtl.enrollProcessor(this->jterm(jj).setupTL(dxfg));
  }
  CtrlInc_ mdxfg(dxfg);
  this->runTL(mdxfg, costtl);

  /// 3. Compute \f$R^{-1} [d + H (dx_{fg} - U dv_0) ]\f$
  PostProcessorAD<Increment_> costad;
  this->zeroAD(grad);

  DualVector<MODEL> yy;
  DualVector<MODEL> ww;
  DualVector<MODEL> zz;

  // \f$ d + H (dx_{fg} - U dv_0) \f$
  for (unsigned jj = 0; jj < this->nterms(); ++jj) {
    yy.append(this->jterm(jj).newDepartures());
    ww.append(costtl.releaseOutputFromTL(jj));
  }
  yy += ww;

  // \f$ R^{-1} [d + H (dx_{fg} - U dv_0) ] \f$
  for (unsigned jj = 0; jj < this->nterms(); ++jj) {
    zz.append(this->jterm(jj).multiplyCoInv(*yy.getv(jj)));
  }

  /// 4. Run ADJ: \f$ H^T R^{-1} [d + H (dx_{fg} - U dv_0) ] \f$
  for (unsigned jj = 0; jj < this->nterms(); ++jj) {
    costad.enrollProcessor(this->jterm(jj).setupAD(zz.getv(jj), grad));
  }
  this->runAD(grad, costad);

  /// 5. Return:
  ///  \f$ 1/2 [d + H (dx_{fg} - U dv_0) ]^T R^{-1} [d + H (dx_{fg} - U dv_0) ]
  ///  \f$
  double costJ0JoJc = 0.5 * dot_product(yy, zz);

  return costJ0JoJc;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::addIncrement(CtrlVar_ &xx, const CtrlInc_ &dx,
                                       PostProcessor<Increment_> post) const {
  /// This methods adds an increment to the first guess
  Log::trace() << "CostFunction::addIncrement start" << std::endl;
  Log::info() << "CostFunction::addIncrement: First guess:" << xx << std::endl;
  Log::info() << "CostFunction::addIncrement: Increment:" << dx << std::endl;

  /// 1. Add the auxiliary observation space components (e.g. VarBC, TOVSCV, ..)
  xx.obsVar() += dx.obsVar();
  /// 2. Add the auxiliary model space components (e.g. model bias)
  xx.modVar() += dx.modVar();
  /// 3. Add the increment to the state
  this->addIncr(xx, dx, post);

  Log::test() << "CostFunction::addIncrement: Analysis norm: " << xx.norm()
              << std::endl;
  Log::trace() << "CostFunction::addIncrement done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::recentreBackground(const eckit::Configuration &conf) {
  /// This method is re-centres the background
  Log::trace() << "CostFunction::recentreBackground start" << std::endl;
  Log::info() << "CostFunction::recentreBackground enabled" << std::endl;

  /// Setup re-centring increment
  Geometry_ bkgRecResol(conf.getSubConfiguration("resolution"));
  Variables_ bkgRecVars(conf);
  util::DateTime bkgRecValidTime(xb_->state()[0].validTime());

  Increment_ bkgRecIncr(bkgRecResol, bkgRecVars, bkgRecValidTime);
  bkgRecIncr.read(conf.getSubConfiguration("increment"));

  /// Re-centre the background
  xb_->state()[0] += bkgRecIncr;

  /// Write out re-centred background (optional)
  if (conf.has("writeOutRecentredBackground")) {
    xb_->state().write(conf.getSubConfiguration("writeOutRecentredBackground"));
  }

  Log::trace() << "CostFunction::recentreBackground done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::resetLinearization() {
  /// This method resets the linearization of the non-linear cost function
  /// Reset Jb
  if (reduceMemoryFootprint()) {
    jb_.reset();
    jbRestart_.reset();
  }
  /// Reset Jo, Jc
  for (auto &jterm : jterms_) {
    jterm->resetLinearization();
    if (reduceMemoryFootprint()) jterm.reset();
  }
  /// Reset linearization trajectory
  tlm_.clear();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::checkpoint(eckit::LocalConfiguration &conf) const {
  /// This method is checkpointing the terms of the cost function required for
  /// restarting

  if (isRestarted()) {
    Log::info() << "CostFunction::checkpoint skipping checkpointing (restarted)"
                << std::endl;
  } else {
    util::LogbookWriter<size_t> log("CostFunctionCheckpoint", 999);

    /// 1. Call the checkpoint method of \f$J_o, J_c, ..\f$
    for (auto &jterm : jterms_) {
      jterm->checkpoint(conf);
    }

    /// 2. Checkpoint the non-linear cost fucntion value JoJc
    conf.set("costJoJc", util::full_precision(costJoJc_));

    this->setIsRestarted(false);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::restart(const eckit::Configuration &conf) {
  /// This method is restarting the terms of the cost function
  this->setIsRestarted(true);

  /// 1. Call the restart method of \f$J_o, J_c , .. \f$
  for (auto &jterm : jterms_) {
    jterm->restart(conf);
  }

  /// 2. Restart the non-linear cost function value JoJc
  costJoJc_ = conf.getDouble("costJoJc");
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::initializeTrajJbRestart(
    const CtrlVar_ &fguess, const eckit::Configuration &conf) {
  /// This method is initializing JbRestart to allow reading in control vector
  /// at their original resolution
  if (conf.has("resolution_restart")) {
    util::LogbookWriter<bool> log_restart("Restart_ctlvec", true);
    const eckit::LocalConfiguration jbConf(conf_, "Jb");
    jbRestart_.reset(new JbTotal_(*xb_, this->newJb(jbConf, resol_, *xb_),
                                  conf_, resol_, xb_->obsVar().obspaces()));
    const eckit::LocalConfiguration resRestartConf(conf, "resolution_restart");
    const Geometry_ lowresRestart(resRestartConf);
    jqRestart_.reset(jbRestart_->initializeTraj(fguess, lowresRestart, conf));
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostFunction<MODEL>::finalizeTrajJbRestart(
    const eckit::Configuration &conf) {
  /// This method is finalizing JbRestart to allow reading in control vector at
  /// their original resolution
  if (conf.has("resolution_restart")) {
    jbRestart_->finalizeTraj(jqRestart_.get(), conf);
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTFUNCTION_H_
