/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJB4D_H_
#define OOPS_ASSIMILATION_COSTJB4D_H_

#include <memory>  // for std::unique_ptr
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/CostJbState.h"
#include "oops/base/IncrCtlVec4D.h"
#include "oops/base/Increment4D.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/base/State4D.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/IncrEnsCtlVec.h"
#include "oops/interface/IncrModCtlVec.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/dot_product.h"

namespace oops {
template <typename MODEL>
class ControlIncrement;
template <typename MODEL>
class IncrCtlVec;
template <typename MODEL>
class JqTerm;
template <typename MODEL>
class JqTermTL;
template <typename MODEL>
class JqTermAD;

// -----------------------------------------------------------------------------

/// 4D Jb Cost Function
/*!
 * CostJb4D encapsulates the generalized four dimensional Jb term of
 * the 4D-Ens-Var cost function. NOTE: for the square root formulation
 * it is assumed that the localization matrix is constant
 * L_bar = [I I .. I]^T L [I I .. I]
 */

template <typename MODEL>
class CostJb4D : public CostJbState<MODEL> {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrCtlVec4D_ = IncrCtlVec4D<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  /// Construct \f$ J_b\f$.
  CostJb4D(const eckit::Configuration &, const Geometry_ &, const Variables_ &,
           const util::Duration &, const State4D_ &);

  /// Destructor
  virtual ~CostJb4D() {}

  /// Empty Jq observer.
  JqTerm<MODEL> *initializeJq() const override { return 0; }

  /// Get increment from state (usually first guess).
  void computeIncrement(const State4D_ &, const State4D_ &,
                        Increment4D_ &) const override;

  /// Linearize before the linear computations.
  void linearize(const State4D_ &, const Geometry_ &,
                 const eckit::Configuration &) override;

  /// Add Jb gradient.
  void addGradient(const Increment4D_ &, Increment4D_ &,
                   Increment4D_ &) const override;

  /// Empty TL Jq observer.
  JqTermTL<MODEL> *initializeJqTL() const override { return 0; }

  /// Empty AD Jq observer.
  JqTermAD<MODEL> *initializeJqAD(const Increment4D_ &) const override {
    return 0;
  }

  /// Multiply by \f$ B\f$ and \f$ B^{-1}\f$.
  void Bmult(const Increment4D_ &, Increment4D_ &) const override;
  void Bminv(const Increment4D_ &, Increment4D_ &) const override;

  /// Multiply by \f$ B^{1/2}\f$ and \f$ [B^{1/2}]^T\f$.
  void BmultSqrt(const IncrCtlVec4D_ &, Increment4D_ &) const override;
  void BmultSqrtTrans(const Increment4D_ &, IncrCtlVec4D_ &) const override;

  /// Access to B
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return B4D_->covar(); }

  /// Randomize
  void randomize(Increment4D_ &) const override;

  /// Create new increment (set to 0).
  unsigned int nstates() const override { return times_.size(); }
  Increment_ *newStateIncrement(const unsigned int) const override;

  /// Create new control vector
  unsigned int nchistates() const override { return 1; }
  IncrCtlVec_ *newIncrCtlVec(const unsigned int) const override;

  /// Create new control vector (mod)
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return B4D_->newIncrModCtlVec();
  }

  /// Create new control vector (ens)
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return B4D_->newIncrEnsCtlVec();
  }

 private:
  std::shared_ptr<ModelSpaceCovariance4DBase<MODEL> > B4D_;
  const Variables_ controlvars_;
  std::unique_ptr<const Geometry_> resol_;
  std::vector<util::DateTime> times_;
};

// =============================================================================

//  Generalized Jb Term of Cost Function
// -----------------------------------------------------------------------------

template <typename MODEL>
CostJb4D<MODEL>::CostJb4D(const eckit::Configuration &config,
                          const Geometry_ &resolouter,
                          const Variables_ &ctlvars, const util::Duration &,
                          const State4D_ &xb)
    : B4D_(), controlvars_(ctlvars), resol_(), times_() {
  // Create one row of blocks of the whole BMatrix, one object for each
  // subwindow. Each object stands for all blocks in the same column of the B
  // Matrix.
  // It can be from any concrete class of ModelSpaceCovarianceBase,
  // according to the configuration file.
  const eckit::LocalConfiguration covar(config, "Covariance");
  B4D_.reset(
      Covariance4DFactory<MODEL>::create(covar, resolouter, controlvars_, xb));
  times_.clear();

  if (covar.has("envar_times")) {
    std::vector<eckit::LocalConfiguration> envar_times;
    covar.get("envar_times", envar_times);
    for (size_t jsub = 0; jsub < envar_times.size(); ++jsub) {
      times_.push_back(
          util::DateTime(envar_times[jsub].getString("reference_time")));
    }
  } else {
    for (unsigned jsub = 0; jsub < xb.statesNumber(); ++jsub) {
      times_.push_back(xb[jsub].validTime());
    }
  }
  Log::trace() << "CostJb4D contructed." << std::endl;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void CostJb4D<MODEL>::linearize(const State4D_ &fg, const Geometry_ &resolinner,
                                const eckit::Configuration &innerConf) {
  resol_.reset(new Geometry_(resolinner));
  B4D_->linearize(fg, *resol_, innerConf);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::computeIncrement(const State4D_ &xb, const State4D_ &fg,
                                       Increment4D_ &dx) const {
  if (xb.statesNumber() == 1) {
    dx.zero();
    dx[0].diff(fg[0], xb[0]);
  } else {
    dx.diff(fg, xb);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::addGradient(const Increment4D_ &dxFG, Increment4D_ &grad,
                                  Increment4D_ &gradJb) const {
  grad -= gradJb;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::Bmult(const Increment4D_ &dxin,
                            Increment4D_ &dxout) const {
  B4D_->multiply(dxin, dxout);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::Bminv(const Increment4D_ &dxin,
                            Increment4D_ &dxout) const {
  B4D_->inverseMultiply(dxin, dxout);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::BmultSqrt(const IncrCtlVec4D_ &dvin,
                                Increment4D_ &dxout) const {
  B4D_->multiplySqrt(dvin[0], dxout);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::BmultSqrtTrans(const Increment4D_ &dxin,
                                     IncrCtlVec4D_ &dvout) const {
  B4D_->multiplySqrtTrans(dxin, dvout[0]);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJb4D<MODEL>::randomize(Increment4D_ &dx) const {
  /* TODO:
    B4D_[jsub].randomize(dx[jsub]);
  */
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> *CostJb4D<MODEL>::newStateIncrement(
    const unsigned int isub) const {
  Increment_ *incr =
      new Increment_(*resol_, controlvars_, times_[0], times_[isub]);
  return incr;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec<MODEL> *CostJb4D<MODEL>::newIncrCtlVec(
    const unsigned int isub) const {
  return new IncrCtlVec_(*this, times_[isub]);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJB4D_H_
