/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJBSTATE_H_
#define OOPS_ASSIMILATION_COSTJBSTATE_H_

#include <memory>

#include "eckit/memory/NonCopyable.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"

namespace util {
class Duration;
}

namespace oops {

// Forward declaration
template <typename MODEL>
class ControlIncrement;
template <typename MODEL>
class Increment4D;
template <typename MODEL>
class IncrCtlVec;
template <typename MODEL>
class IncrCtlVec4D;
template <typename MODEL>
class IncrModCtlVec;
template <typename MODEL>
class IncrEnsCtlVec;
template <typename MODEL>
class JqTerm;
template <typename MODEL>
class JqTermTL;
template <typename MODEL>
class JqTermAD;
template <typename MODEL>
class State4D;

// -----------------------------------------------------------------------------

/*! @brief Pure abstract base class for the Jb term state component

 *   The CostJbState is the base class for the Jb term corresponding to the
 *  state part (3D or 4D) of the control variable.

 *   Consider applying the matrix vector product:
 *    \f[ B_{state} dx_{state} =
          \begin{bmatrix} B_0 & 0   & .. & 0   \\
                          0   & B_1 & .. & 0   \\
                          ..  & ..  & .. & ..  \\
                          0   & 0   & .. & B_N

          \end{bmatrix}
          \begin{bmatrix} dx_0 \\
                          dx_1 \\
                          ..   \\
                          dx_N \\
          \end{bmatrix}
      \f]

 *   Both \f$ B_{state} \f$, as implemented by the JbState class, and
 *  \f$ dx_{state} \f$, as implemented by Increment4D class, have in the
 *  most general case a 4D structure. In case of 3/4D-Var, CostJb3D class
 *  contains only a single component \f$ B_0 \f$ and likewise  Increment4D,
 *  implemented as a vector of increments, contains only single component
 *  \f$ dx_0 \f$.
 */

template <typename MODEL>
class CostJbState : private eckit::NonCopyable {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrCtlVec4D_ = IncrCtlVec4D<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using State4D_ = State4D<MODEL>;

 public:
  /// Constructor
  CostJbState() {}

  /// Destructor
  virtual ~CostJbState() {}

  /// Initialize Jq computations (relevant for weak constraint).
  virtual JqTerm<MODEL> *initializeJq() const = 0;

  /// Get increment from state (usually first guess).
  virtual void computeIncrement(const State4D_ &, const State4D_ &,
                                Increment4D_ &) const = 0;

  /// Linearize before the linear computations.
  virtual void linearize(const State4D_ &xx, const Geometry_ &geom,
                         const eckit::Configuration &) = 0;

  /// Add Jb gradient.
  virtual void addGradient(const Increment4D_ &, Increment4D_ &,
                           Increment4D_ &) const = 0;

  /// Finalize \f$ J_b\f$ after the TL run (relevant for weak constraint).
  virtual JqTermTL<MODEL> *initializeJqTL() const = 0;

  /// Initialize \f$ J_b\f$ before the AD run (relevant for weak constraint).
  virtual JqTermAD<MODEL> *initializeJqAD(const Increment4D_ &) const = 0;

  /// Multiply by \f$ B\f$.
  virtual void Bmult(const Increment4D_ &, Increment4D_ &) const = 0;
  /// Multiply by \f$B^{-1}\f$.
  virtual void Bminv(const Increment4D_ &, Increment4D_ &) const = 0;

  /// Multiply by \f$ B^{1/2}\f$.
  virtual void BmultSqrt(const IncrCtlVec4D_ &, Increment4D_ &) const = 0;
  /// Multiply by \f$ [B^{1/2}]^T\f$.
  virtual void BmultSqrtTrans(const Increment4D_ &, IncrCtlVec4D_ &) const = 0;

  /// Randomize
  virtual void randomize(Increment4D_ &) const = 0;

  /// Access to B
  virtual const ModelSpaceCovarianceBase<MODEL> &covar() const = 0;

  /// Return size of the state component of the control variable
  virtual unsigned int nstates() const = 0;

  /// Create new increment (set to 0).
  virtual Increment_ *newStateIncrement(const unsigned int) const = 0;

  /// Return size of the state component of the control vector
  virtual unsigned int nchistates() const = 0;

  /// Create new control vector
  virtual IncrCtlVec_ *newIncrCtlVec(const unsigned int) const = 0;

  /// Create new control vector (mod)
  virtual IncrModCtlVec_ *newIncrModCtlVec() const = 0;

  /// Create new control vector (ens)
  virtual IncrEnsCtlVec_ *newIncrEnsCtlVec() const = 0;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJBSTATE_H_
