/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJBTOTAL_H_
#define OOPS_ASSIMILATION_COSTJBTOTAL_H_

#include <limits>
#include <memory>  // for std::unique_ptr

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostJbState.h"
#include "oops/base/ObsAuxCovariances.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxCovariance.h"
#include "util/Logger.h"

namespace oops {
template <typename MODEL>
class JqTerm;
template <typename MODEL>
class JqTermTL;
template <typename MODEL>
class JqTermAD;

// -----------------------------------------------------------------------------

/*! @brief Jb component of the cost function

 *   CostJbTotal encapsulates B matrices corresponding to various components
 *  of the control increment/control vector:

 *   Consider applying the matrix vector product:
 *    \f[ B dx =
          \begin{bmatrix} B_{state}  &  0           &  0          \\
                          0          &  B_{modaux}  &  0          \\
                          0          &  0           &  B_{obsaux}
          \end{bmatrix}
          \begin{bmatrix} dx_{state}  \\
                          dx_{modaux} \\
                          dx_{obsaux} \\
          \end{bmatrix}
      \f]

 *   - CostJbState        - \f$ B_{state} \f$: abstract base class encapsulating
 *                          the background covariance matrix corresponding to
 *                          the state component of the control increment.
 *                          Concrete classes deriving from it are CostJb3D
 *                          (3/4D-Var), CostJb4D (4D-Ens-Var), CostJbJq (weak
 *                          constraint)

 *   - ModelAuxCovariance - \f$ B_{modaux} \f$: class encapsulating the
 *                          background covariance matrix of an auxiliary model
 *                          space component of the control increment
 *                          (e.g.Q matrix - model bias covariance)

 *   - ObsAuxCovariances  - \f$ B_{obsaux} \f$: class encapsulating background
 *                          covariance matrices of auxiliary observation space
 *                          components of the control increment (e.g. VarBC,
 *                          TOVSCV)

 */

template <typename MODEL>
class CostJbTotal {
  using State4D_ = State4D<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using JbState_ = CostJbState<MODEL>;
  using JqTerm_ = JqTerm<MODEL>;
  using JqTermAD_ = JqTermAD<MODEL>;
  using JqTermTL_ = JqTermTL<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using ModelAuxCovar_ = ModelAuxCovariance<MODEL>;
  using ObsAuxCovars_ = ObsAuxCovariances<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  static const std::string classname() { return "oops::CostJbTotal"; }

  /// Construct \f$ J_b\f$.
  CostJbTotal(const CtrlVar_ &, JbState_ *, const eckit::Configuration &,
              const Geometry_ &, const ObsSpaces_ &odb);

  /// Destructor
  ~CostJbTotal() {}

  /// Set up Jq term for weak constraint
  JqTerm_ *initialize(const CtrlVar_ &) const;

  /// Various set ups prior to the linearisation trajectory computation
  JqTerm_ *initializeTraj(const CtrlVar_ &, const Geometry_ &,
                          const eckit::Configuration &);

  /// Compute model error for weak constraint
  void finalize(JqTerm_ *, CtrlInc_ &, const eckit::Configuration &) const;

  /// Various computations after the linearisation trajectory computation
  double finalizeTraj(JqTerm_ *, const eckit::Configuration &);

  /// Initialize JqTermTL (weak constraint only) before starting the TL run
  JqTermTL_ *initializeTL() const;

  /// Model error computation (weak constraint only) after the TL run
  void finalizeTL(JqTermTL_ *, const CtrlInc_ &, CtrlInc_ &) const;

  /// Initialize JqTermAD (weak constraint only) before starting the AD run
  JqTermAD_ *initializeAD(CtrlInc_ &, const CtrlInc_ &) const;

  /// Perform clean ups after the AD run
  void finalizeAD(JqTermAD_ *) const;

  /// Multiply by the covariance matrix
  void multiplyB(const CtrlInc_ &, CtrlInc_ &) const;

  /// Multiply by the inverse covariance matrix
  void multiplyBinv(const CtrlInc_ &, CtrlInc_ &) const;

  /// Multiply by the square root of the covariance matrix
  void multiplyBsqrt(const CtrlVec_ &, CtrlInc_ &) const;

  /// Multiply by the transpose of the square root of the covariance matrix
  void multiplyBsqrtTrans(const CtrlInc_ &, CtrlVec_ &) const;

  /// Randomize (using the square root of the covariance matrix)
  void randomize(CtrlInc_ &) const;

  /// Add Jb gradient at first guess.
  void addGradientFG(CtrlInc_ &) const;
  void addGradientFG(CtrlInc_ &, CtrlInc_ &) const;
  void addGradientFG(CtrlVec_ &, CtrlVec_ &) const;

  /// Return background.
  const CtrlVar_ &getBackground() const { return xb_; }

  /// Return first guess increment \f$ x_0-x_b\f$.
  const CtrlInc_ &getFirstGuessIncrement() const { return *dxFG_; }

  // Jb terms for ControlIncrement constructor.

  /// Access (inner loop) geometry
  const Geometry_ &resolution() const { return *geom_; }

  /// Access covariance matrix for the state component of ctrl vec
  const JbState_ &jbState() const { return *jb_; }

  /// Access  covariance matrix for the auxiliary model space component of ctrl
  /// vec
  const ModelAuxCovar_ &jbModBias() const { return jbModBias_; }

  /// Access  covariance matrix for the auxiliary obs space component of ctrl
  /// vec
  const ObsAuxCovars_ &jbObsBias() const { return jbObsBias_; }

 private:
  /// Compute Jb cost function value
  double evaluate(const CtrlInc_ &) const;

  /// Write the increment
  void writeIncrement(const CtrlInc_ &, const eckit::Configuration &) const;

  /// Reference to the background control variable
  const CtrlVar_ &xb_;

  // Covariance
  std::unique_ptr<JbState_>
      jb_;                    ///< Cov matrix for state component of ctrl vec
  ModelAuxCovar_ jbModBias_;  ///< Cov matrix for auxiliary model space
                              /// component of ctrl vec
  ObsAuxCovars_
      jbObsBias_;  ///< Cov matrix for auxiliary obs space component of ctrl vec

  /// Pointer to the first guess
  const CtrlVar_ mutable *fg_;

  /// First guess increment \f$x_0-x_b\f$ or more generally \f$
  /// x_i-M(x_{i-1})\f$.
  std::unique_ptr<CtrlInc_> dxFG_;

  /// Inner loop geometry
  std::unique_ptr<Geometry_> geom_;
};

// =============================================================================

template <typename MODEL>
CostJbTotal<MODEL>::CostJbTotal(const CtrlVar_ &xb, JbState_ *jb,
                                const eckit::Configuration &conf,
                                const Geometry_ &resol, const ObsSpaces_ &odb)
    : xb_(xb),
      jb_(jb),
      jbModBias_(conf.getSubConfiguration("Jb.ModelBiasCovariance"), resol),
      jbObsBias_(odb, eckit::LocalConfiguration(conf, "Jo")),
      fg_(),
      dxFG_(),
      geom_() {
  /*!  Set the pointer to the covariance matrix for the state component of
   *  the control vector. Construct the covariance matrices for the auxiliary
   *  observation and model space components of the control vector.

   *   The covariance matrix for the state component of the control vector
   *  is constructed outside by calling the newJb method of one of the cost
   *  function classes derived from the abstract base class CostFunction:
   *  CostFct3DVar (returns CostJb3D), CostFct4DVar (returns CostJb3D),
   *  CostFct4DEnsVar (returns CostJb4D), CostFctWeak (returns CostJbJq).
   */
  Log::trace() << "CostJbTotal contructed." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
JqTerm<MODEL> *CostJbTotal<MODEL>::initialize(const CtrlVar_ &fg) const {
  /*!  This method is called prior to the non-linear model integration.
   *  Relevant for weak constraint only. CostJbJq (inheriting from
   *  CostJbState) returns a post-processor that is subsequently attached
   *  to the model integration. Its purpose is to compute the model error
   *  \f$ x_i - {\cal M}(x_{i-1}\f$. Returns JqTerm post-processor.
   */
  fg_ = &fg;
  JqTerm_ *jqnl = jb_->initializeJq();
  return jqnl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::finalize(JqTerm_ *jqnl, CtrlInc_ &dx,
                                  const eckit::Configuration &conf) const {
  /*!  The purpose of this method is to compute the model error
   *  for the weak constraint formulation \f$ x_i - {\cal M}(x_{i-1}\f$.
   */

  /// 1. Compute model error \f$ x_i - {\cal M}(x_{i-1}\f$ (weak constraint
  /// only)
  if (jqnl) jqnl->computeModelError(fg_->state(), dx.state());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::writeIncrement(
    const CtrlInc_ &dx, const eckit::Configuration &conf) const {
  /// 1. Write the state component of the increment
  if (conf.has("increment")) {
    // Write increment
    dx.state().write(conf);
  }

  /// 2. Write the auxiliary model space component of the increment
  if (conf.has("incrementModAux")) {
    dx.modVar().write(conf);
  }

  /// 3. Write the auxiliary observation space components of the increment
  if (conf.has("incrementsObsAux")) {
    dx.obsVar().write(conf);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
JqTerm<MODEL> *CostJbTotal<MODEL>::initializeTraj(
    const CtrlVar_ &fg, const Geometry_ &resol,
    const eckit::Configuration &config) {
  /*!  This method is called prior to the forced model linearisation
   *  trajectory computation in the TrajecotrySaver. Its purpose is to
   *  reset the inner loop geometry, set up the background error covariances.
   *  Calls initialize method which returns JqTerm post-processor (relevant
   *  for weak constraint). Returns JqTerm post-processor.
   */

  /// 1. Set the pointer to the first guess
  fg_ = &fg;

  /// 2. Set up the inner loop geometry
  geom_.reset(new Geometry_(resol));

  /// 3. Set up covariance matrices
  jb_->linearize(fg_->state(), *geom_, config);
  jbModBias_.linearize(fg_->modVar(), *geom_);
  jbObsBias_.linearize(fg_->obsVar(), fg_->obsVar().obspaces());

  /// 4. Call the initialize method, linearize JqTerm post-processor and return
  /// it.
  JqTerm_ *jqnl = this->initialize(fg);
  if (jqnl) jqnl->linearize();
  return jqnl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJbTotal<MODEL>::finalizeTraj(JqTerm_ *jqnl,
                                        const eckit::Configuration &conf) {
  /*!  This method is called once the model linearisation trajectory
   *  is computed. Its purpose is to compute the first guess increment,
   *  update background covariances corresponding to the auxiliary observation
   *  space components of the control vector (relevant for IFS). In case of weak
   *  constraint JqTerm computes the model error. Returns the value of the Jb
   *  component of the cost function.
   */

  /// 1. Compute and save the first guess increment.
  dxFG_.reset(new CtrlInc_(*this));

  // Compute x_0 - x_b for Jb
  if (conf.has("explicitChangeResolution")) {
    State4D_ xblr(resolution(), xb_.state());
    jb_->computeIncrement(xblr, fg_->state(), dxFG_->state());
  } else {
    jb_->computeIncrement(xb_.state(), fg_->state(), dxFG_->state());
  }

  // Model and Obs biases
  dxFG_->modVar().diff(fg_->modVar(), xb_.modVar());
  dxFG_->obsVar().diff(fg_->obsVar(), xb_.obsVar());

  /// 2. Compute the model error (relevant for weak constraint)
  if (jqnl) this->finalize(jqnl, *dxFG_, conf);
  Log::info() << "CostJb: FG-BG" << *dxFG_ << std::endl;

  /// 3.  Update covariances corresponding to the auxiliary observation space
  ///    components of the control vector
  if (conf.has("covariancesObsAux")) {
    // Update ObsAuxCovariance
    jbObsBias_.update(fg_->obsVar(), conf);
  }

  // Reset first guess pointer
  fg_ = nullptr;

  /// 4. Compute and return the Jb value
  double zjb = this->evaluate(*dxFG_);
  return zjb;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJbTotal<MODEL>::evaluate(const CtrlInc_ &dx) const {
  CtrlInc_ gg(dx, false);
  this->multiplyBinv(dx, gg);

  double zjb = 0.0;
  double zz = 0.5 * dot_product(dx.state(), gg.state());
  Log::info() << "CostJb   : Nonlinear Jb State = " << zz << std::endl;
  zjb += zz;
  zz = 0.5 * dot_product(dx.modVar(), gg.modVar());
  Log::info() << "CostJb   : Nonlinear Jb Model Aux = " << zz << std::endl;
  zjb += zz;
  zz = 0.5 * dot_product(dx.obsVar(), gg.obsVar());
  Log::info() << "CostJb   : Nonlinear Jb Obs Aux = " << zz << std::endl;
  zjb += zz;

  Log::info() << "CostJb   : Nonlinear Jb = " << zjb << std::endl;

  // Get rid of very small values for test
  double ztest = zjb;
  if (zjb >= 0.0 && zjb <= std::numeric_limits<double>::epsilon()) ztest = 0.0;
  Log::test() << "CostJb   : Nonlinear Jb = " << ztest << std::endl;

  return zjb;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::addGradientFG(CtrlInc_ &grad) const {
  /*!  Add Jb component of the first gradient. Used for primal formulation.
   */
  CtrlInc_ gg(grad, false);
  this->multiplyBinv(*dxFG_, gg);
  grad -= gg;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::addGradientFG(CtrlInc_ &grad, CtrlInc_ &gradJb) const {
  /*!  Add Jb component of the first gradient. Used for B preconditioned
   *  formulation.
   */
  jb_->addGradient(dxFG_->state(), grad.state(), gradJb.state());
  grad.modVar() -= gradJb.modVar();
  grad.obsVar() -= gradJb.obsVar();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::addGradientFG(CtrlVec_ &grad, CtrlVec_ &gradJb) const {
  /*!  Add Jb component of the first gradient. Used for square root B
   *  preconditioned formulation.
   */
  grad.state() -= gradJb.state();
  grad.modVar() -= gradJb.modVar();
  grad.obsVar() -= gradJb.obsVar();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
JqTermTL<MODEL> *CostJbTotal<MODEL>::initializeTL() const {
  JqTermTL_ *jqtl = jb_->initializeJqTL();
  return jqtl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::finalizeTL(JqTermTL_ *jqtl, const CtrlInc_ &bgns,
                                    CtrlInc_ &dx) const {
  dx = bgns;
  if (jqtl) jqtl->computeModelErrorTL(dx.state());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
JqTermAD<MODEL> *CostJbTotal<MODEL>::initializeAD(CtrlInc_ &bgns,
                                                  const CtrlInc_ &dx) const {
  JqTermAD_ *jqad = jb_->initializeJqAD(dx.state());
  bgns += dx;
  return jqad;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::finalizeAD(JqTermAD_ *jqad) const {
  if (jqad) jqad->clear();
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::multiplyB(const CtrlInc_ &dxin,
                                   CtrlInc_ &dxout) const {
  jb_->Bmult(dxin.state(), dxout.state());
  jbModBias_.multiply(dxin.modVar(), dxout.modVar());
  jbObsBias_.multiply(dxin.obsVar(), dxout.obsVar());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::multiplyBinv(const CtrlInc_ &dxin,
                                      CtrlInc_ &dxout) const {
  jb_->Bminv(dxin.state(), dxout.state());
  jbModBias_.inverseMultiply(dxin.modVar(), dxout.modVar());
  jbObsBias_.inverseMultiply(dxin.obsVar(), dxout.obsVar());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::multiplyBsqrt(const CtrlVec_ &dvin,
                                       CtrlInc_ &dxout) const {
  jb_->BmultSqrt(dvin.state(), dxout.state());
  jbModBias_.multiplySqrt(dvin.modVar(), dxout.modVar());
  jbObsBias_.multiplySqrt(dvin.obsVar(), dxout.obsVar());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::multiplyBsqrtTrans(const CtrlInc_ &dxin,
                                            CtrlVec_ &dvout) const {
  jb_->BmultSqrtTrans(dxin.state(), dvout.state());
  jbModBias_.multiplySqrtTrans(dxin.modVar(), dvout.modVar());
  jbObsBias_.multiplySqrtTrans(dxin.obsVar(), dvout.obsVar());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJbTotal<MODEL>::randomize(CtrlInc_ &dx) const {
  jb_->randomize(dx.state());
  jbModBias_.randomize(dx.modVar());
  jbObsBias_.randomize(dx.obsVar());
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJBTOTAL_H_
