/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJCDFI_H_
#define OOPS_ASSIMILATION_COSTJCDFI_H_

#include <memory>  // for std::dynamic_pointer_cast, std::shared_ptr, std::unique_ptr

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostTermBase.h"
#include "oops/base/DolphChebyshev.h"
#include "oops/base/PostBase.h"
#include "oops/base/PostBaseAD.h"
#include "oops/base/PostBaseTL.h"
#include "oops/base/WeightedDiff.h"
#include "oops/base/WeightedDiffAD.h"
#include "oops/base/WeightedDiffTL.h"
#include "oops/base/WeightingFct.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/JcCovar.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Jc DFI Cost Function
/*!
 * Digital filter based constraint term for the cost function.
 */

template <typename MODEL>
class CostJcDFI : public CostTermBase<MODEL> {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using JcCovar_ = JcCovar<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  const std::string classname() const override { return "JcDFI"; }

  /// Construct \f$ J_c\f$.
  CostJcDFI(const eckit::Configuration &, const Geometry_ &,
            const util::DateTime &, const util::Duration &,
            const util::Duration &tstep = util::Duration(0));

  /// Destructor
  virtual ~CostJcDFI() {}

  /// Initialize before nonlinear model integration.
  std::shared_ptr<PostBase<State_> > initialize(
      const CtrlVar_ &, const eckit::Configuration &) const override;
  std::shared_ptr<PostBase<State_> > initializeTraj(
      const CtrlVar_ &, const Geometry_ &,
      const eckit::Configuration &) override;

  /// Finalize computation after nonlinear model integration.
  double finalize(const eckit::Configuration &) const override;
  double finalizeTraj(const eckit::Configuration &) override;

  /// Initialize \f$ J_c\f$ before starting the TL run.
  std::shared_ptr<PostBaseTL<Increment_> > setupTL(
      const CtrlInc_ &) const override;

  /// Initialize \f$ J_c\f$ before starting the AD run.
  std::shared_ptr<PostBaseAD<Increment_> > setupAD(
      std::shared_ptr<const GeneralizedDepartures>, CtrlInc_ &) const override;

  /// Multiply by \f$ C\f$ and \f$ C^{-1}\f$.
  Increment_ *multiplyCovar(const GeneralizedDepartures &) const override;
  Increment_ *multiplyCoInv(const GeneralizedDepartures &) const override;
  Increment_ *randomizeCovar() const override {
    ABORT("randomizeCovar not implemented for JcDFI");
    return this->newGradientFG();
  }

  /// Provide new increment.
  Increment_ *newDualVector() const override;

  /// Provide new departures (not relevant to JcDFI).
  Increment_ *newDepartures() const override { return new Increment_(*xdep_); }

  /// Gradient of \f$ J_c\f$ at first guess.
  Increment_ *newGradientFG() const override {
    return new Increment_(*gradFG_);
  }

  /// Reset trajectory.
  void resetLinearization() override;

  /// Checkpoint
  void checkpoint(eckit::LocalConfiguration &) const {}

  /// Restart
  void restart(const eckit::Configuration &) {}

 private:
  const eckit::LocalConfiguration conf_;
  const Geometry_ resol_;
  std::unique_ptr<Geometry_> tlres_;

  bool ltraj_;

  util::DateTime vt_;
  util::Duration span_;
  const util::Duration tstep_;
  util::Duration tlstep_;

  double alpha_;

  mutable std::unique_ptr<Increment_> xdep_;
  mutable std::unique_ptr<Increment_> gradFG_;

  std::unique_ptr<WeightingFct> wfct_;
  std::unique_ptr<WeightingFct> wfcttraj_;
  mutable std::shared_ptr<WeightedDiff<MODEL, Increment_, State_> > filter_;
  mutable std::shared_ptr<WeightedDiff<MODEL, Increment_, State_> > filtertraj_;
  mutable std::unique_ptr<JcCovar_> jccov_;
  std::unique_ptr<eckit::LocalConfiguration> jcovconf_;
};

// =============================================================================

template <typename MODEL>
CostJcDFI<MODEL>::CostJcDFI(const eckit::Configuration &conf,
                            const Geometry_ &resol, const util::DateTime &vt,
                            const util::Duration &span,
                            const util::Duration &tstep)
    : conf_(conf),
      resol_(resol),
      tlres_(),
      ltraj_(false),
      vt_(vt),
      span_(span),
      tstep_(tstep),
      tlstep_(),
      alpha_(0),
      xdep_(),
      gradFG_(),
      wfct_(),
      wfcttraj_(),
      filter_(),
      filtertraj_() {
  alpha_ = conf.getDouble("alpha");
  if (conf.has("ftime")) vt_ = util::DateTime(conf.getString("ftime"));
  if (conf.has("span")) span_ = util::Duration(conf.getString("span"));
  if (conf.has("covariance"))
    jcovconf_.reset(new eckit::LocalConfiguration(conf, "covariance"));
  //  wfct_.reset(WeightFactory::create(config)); YT
  wfct_.reset(new DolphChebyshev(conf_));
  wfcttraj_.reset(new DolphChebyshev(conf_));
  Log::debug() << "CostJcDFI created vt = " << vt_ << ", span = " << span_
               << std::endl;
  Log::trace() << "CostJcDFI created" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBase<State<MODEL> > > CostJcDFI<MODEL>::initialize(
    const CtrlVar_ &x, const eckit::Configuration &) const {
  filter_.reset(new WeightedDiff<MODEL, Increment_, State_>(
      vt_, span_, resol_, conf_, tstep_, *wfct_));
  if (jcovconf_)
    jccov_.reset(
        new JcCovar_(x.state()[0], x.state()[0].geometry(), *jcovconf_));
  return filter_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJcDFI<MODEL>::finalize(const eckit::Configuration &) const {
  double zz = 0.5 * alpha_;
  std::unique_ptr<Increment_> dx(filter_->releaseDiff());
  std::unique_ptr<Increment_> dxnorm(new Increment_(*dx));
  if (jcovconf_) {
    jccov_->inverseMultiply(*dx, *dxnorm);
  }
  zz *= dot_product(*dx, *dxnorm);
  filter_.reset();
  Log::test() << "CostJcDFI: Nonlinear Jc = " << zz << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBase<State<MODEL> > > CostJcDFI<MODEL>::initializeTraj(
    const CtrlVar_ &x, const Geometry_ &tlres,
    const eckit::Configuration &innerConf) {
  ltraj_ = true;
  tlres_.reset(new Geometry_(tlres));
  tlstep_ = util::Duration(innerConf.getString("linearmodel.tstep"));
  filtertraj_.reset(new WeightedDiff<MODEL, Increment_, State_>(
      vt_, span_, *tlres_, conf_, tlstep_, *wfcttraj_));
  if (jcovconf_) {
    if (!jccov_) {
      jccov_.reset(new JcCovar_(x.state()[0], tlres, *jcovconf_));
    }
    jccov_->linearize(x.state()[0], tlres);
  }
  return filtertraj_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJcDFI<MODEL>::finalizeTraj(const eckit::Configuration &) {
  ASSERT(ltraj_ == true);
  xdep_.reset(filtertraj_->releaseDiff());
  bool lresetGradFG = false;
  if (conf_.has("resetGradFG")) lresetGradFG = conf_.getBool("resetGradFG");
  double zz = 0.;
  if (lresetGradFG) {
    gradFG_.reset(new Increment_(*xdep_));
    xdep_->zero();
    gradFG_->zero();
  } else {
    *xdep_ *= -1.0;
    gradFG_.reset(this->multiplyCoInv(*xdep_));
    zz = 0.5 * dot_product(*xdep_, *gradFG_);
  }
  ltraj_ = false;
  filtertraj_.reset();
  Log::info() << "CostJcDFI: Nonlinear Jc (traj) = " << zz << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> *CostJcDFI<MODEL>::newDualVector() const {
  Variables_ vars(conf_);
  Increment_ *dx = new Increment_(*tlres_, vars, vt_);
  return dx;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBaseTL<Increment<MODEL> > > CostJcDFI<MODEL>::setupTL(
    const CtrlInc_ &) const {
  std::shared_ptr<WeightedDiffTL<MODEL, Increment_> > filterTL(
      new WeightedDiffTL<MODEL, Increment_>(vt_, span_, *tlres_, conf_, tlstep_,
                                            *wfcttraj_));
  return filterTL;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBaseAD<Increment<MODEL> > > CostJcDFI<MODEL>::setupAD(
    std::shared_ptr<const GeneralizedDepartures> pv, CtrlInc_ &) const {
  std::shared_ptr<const Increment_> dx =
      std::dynamic_pointer_cast<const Increment_>(pv);
  std::shared_ptr<WeightedDiffAD<Increment_> > filterAD(
      new WeightedDiffAD<Increment_>(vt_, span_, tlstep_, *wfcttraj_, dx));
  return filterAD;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> *CostJcDFI<MODEL>::multiplyCovar(
    const GeneralizedDepartures &dv1) const {
  const Increment_ &dx1 = dynamic_cast<const Increment_ &>(dv1);
  Increment_ *dx2 = new Increment_(dx1);
  if (jcovconf_) jccov_->multiply(dx1, *dx2);
  const double za = 1.0 / alpha_;
  *dx2 *= za;
  return dx2;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> *CostJcDFI<MODEL>::multiplyCoInv(
    const GeneralizedDepartures &dv1) const {
  const Increment_ &dx1 = dynamic_cast<const Increment_ &>(dv1);
  Increment_ *dx2 = new Increment_(dx1);
  if (jcovconf_) jccov_->inverseMultiply(dx1, *dx2);
  *dx2 *= alpha_;
  double zz = dot_product(dx1, *dx2);
  Log::info() << "CostJcDFI: multiplyCoInv norm = " << zz << std::endl;
  return dx2;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJcDFI<MODEL>::resetLinearization() {
  gradFG_.reset();
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJCDFI_H_
