/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJO_H_
#define OOPS_ASSIMILATION_COSTJO_H_

#include <memory>  // for std::dynamic_pointer_cast, std::shared_ptr, std::unique_ptr
#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostTermBase.h"
#include "oops/base/Departures.h"
#include "oops/base/LinearObsOperators.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObsErrors.h"
#include "oops/base/ObsOperators.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/Observations.h"
#include "oops/base/Observer.h"
#include "oops/base/ObserverAD.h"
#include "oops/base/ObserverTL.h"
#include "oops/base/PostBase.h"
#include "oops/base/PostBaseAD.h"
#include "oops/base/PostBaseTL.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief \f$ J_o \f$ term of the cost function

 *  The CostJo class encapsulates the \f$ J_o \f$ term of the cost function. It
 * inherits from a pure abstract base class CostTermBase which defines a common
 * interface to all the cost function terms except for the \f$ J_b \f$ term.

 *  The \f$ J_o \f$ term has the following structure:

 * \f[ J_o = \frac{1}{2} [y - {\cal G}(x)]^T R^{-1} [y - {\cal G}(x)] \f]

 * \f[ J_o = \sum_l J_o^l
           = \frac{1}{2} \sum_l [\bar{y}_{l} -
                                 {\cal \bar{G}}_{l}(\bar{x}_{l})]^T
                 \bar{R}_{l}^{-1}
                                 [\bar{y}_{l} -
                                 {\cal \bar{G}}_{l}(\bar{x}_{l})]
           =  \frac{1}{2} \sum_j \sum_k \sum_l
                  [ \bar{y}_{k,l}^j -
                    {\cal \bar{G}}_{k,l}^j (\bar{x}_{k,l}^j)]^T
             [\bar{R}_{k,l}^j]^{-1}
                  [ \bar{y}_{k,l}^j -
                    {\cal \bar{G}}_{k,l}^j (\bar{x}_{k,l}^j) ]
   \f]

    -> Loop over l - loop over observation types;

    -> Loop over k - loop over the time steps spanning the assimilation
                     window/sub-window;

    -> Loop over j - loop over sub-windows;

 *   Note: the loop over the number of sub-windows is external to CostJo.

 *   Note: that the \f$ J_o \f$ class contains objects, which are themselves
 *  containers to allow exposing to OOPS multiple observation types (look k
 above):

    -> ObservationSpaces  -> a vector of ObservationSpace objects;

    -> ObsOperators       -> a vector of ObsOperator objects;

    -> LinearObsOperators -> a vector of LinearObsOperator objects;

    -> ObsErrors          -> a vector of ObsErrorCovariance objects;

    -> Observations       -> a vector of ControlObsVector objects;

    -> Departures         -> a vector of ControlObsVector objects;

 *  Note:

    -> \f$ {\cal G}(x) \f$ is computed by the Observer post-processor (attached
      to the non-linear forecast);

    -> \f$ y \f$ is implemented by Observations class;

    -> \f$ y - {\cal G}(x) \f$ is stored by Departures class;

    -> \f$ R \f$ is implemented by ObsErrors class;

 *   In more detail, the \f$ J_o \f$ term includes the constrained observation
 *  bias component (constrained VarBC); consider \f$ J_o^l \f$ corresponding to
 *  the l-th observation space (observation type):

 * \f[ J_o^l =  \frac{1}{2} (\bar{y}_l - {\cal \bar{G}}_l (\bar{x}_l))^T
                             [\bar{R}_l]^{-1}
                            (\bar{y}_l - {\cal \bar{G}}_l (\bar{x}_l))
             =  \frac{1}{2}
                    (\begin{bmatrix}
                       \tilde{y}_l \\ \tilde{b}_l
                    \end{bmatrix}
                    -
                    \begin{bmatrix}
                      {\cal \tilde{G}}_l && {\cal \tilde{P}}_l \\
                      0                  && {\cal \tilde{P}}_l
                    \end{bmatrix}
                    \begin{bmatrix}
                       x \\ \tilde{\beta}_l
                    \end{bmatrix})^T
                       \begin{bmatrix}
                          \tilde{R}_l^{-1} && 0                       \\
                          0                && \tilde{R}_{\beta,l}^{-1}
                       \end{bmatrix}
                    (\begin{bmatrix}
                        \tilde{y}_l \\ \tilde{b}_l
                     \end{bmatrix}
                     -
                     \begin{bmatrix}
                        {\cal \tilde{G}}_l && {\cal \tilde{P}}_l \\
                        0                  && {\cal \tilde{P}}_l
                     \end{bmatrix}
                     \begin{bmatrix}
                        x \\ \tilde{\beta}_l
                     \end{bmatrix})^T
   \f]

   ->  the non-linear observation operator associated with l-th ObservationSpace

      \f[ \begin{bmatrix} {\cal \tilde{H}}_l && {\cal \tilde{P}}_l
           \end{bmatrix}
      \f]

       is implemented by obsEquiv method of ObsOperator; note that
       \f$ {\cal \tilde{G}}_l = {\cal \tilde{H}}_l {\cal I} \f$, where
       \f$ {\cal I} \f$ is an interpolation operator

   ->  the non-linear observation bias opeartor \f$ {\cal \tilde{P}}_l \f$
      associated with l-th Observation space is is implemented by obsBiasEquiv
      method of ObsOperator;

   ->  the observation and bias error covariance matrix associated with l-th
      ObservationSpace:

       \f[ \begin{bmatrix} \tilde{R}_l^{-1} && 0 \\
                           0                && \tilde{R}_{\beta,l}^{-1}
            \end{bmatrix}
       \f]

      is implemented by inverseMultiply method of ObsErrorCovariance class;
      \f$ \tilde{R}_l \f$ is observation error covariance matrix and
      \f$ \tilde{R}_{\beta,l}\f$ is observation bias error covariance matrix;

   ->  the control observation vector associated with l-th ObservationSpace:

       \f[ \begin{bmatrix} \tilde{y}_l \\
                           \tilde{b}_l
           \end{bmatrix}
       \f]

       is implemented by ControlObsVector class. Observations class contains a
       vector of ControlObsVector objects;

   ->  The departures associated with l-th ObservationSpace:

       \f[ \begin{bmatrix} \tilde{d}_l \\
                           \tilde{d}_{c,l}
           \end{bmatrix} =
           \begin{bmatrix} \tilde{y}_l \\
                           \tilde{b}_l
           \end{bmatrix}
         - \begin{bmatrix} {\cal \tilde{G}}_l && {\cal \tilde{P}}_; \\
                             0                && {\cal \tilde{P}}_l
           \end{bmatrix}
           \begin{bmatrix} x \\
                           \beta_l
           \end{bmatrix}
       \f]

       are implemented by ControlObsVector class. Departures class contains a
       vector of ControlObsVector objects.

 *   For details on the implementation of VarQC refere to ObsErrorCovariance
 *  documentation.

 *   Consider the quadratic \f$ J_o \f$ cost function term:

      \f[  J_o(dx) = G^T R^{-1} G (dx - d) \f]

     -> \f$ G (dx) \f$ is computed by ObserverTL; ObserverTL is a post processor
       which is attached to the tangent linear model trajectory
       (LinearModel::forecastTL). \f$ G \f$ has the same structure as
       \f$ {\cal G} \f$ and \f$ dx \f$ has the same structure as \f$ x \f$;

     -> \f$ G^T (dy) \f$ is computed by ObserverAD; ObserverAD is a post
       processor which is attached to the adjoint of the tangent linear model
       trajectory (LinearModel::forecastAD).
 */

template <typename MODEL>
class CostJo : public CostTermBase<MODEL>, private eckit::NonCopyable {
  using CtrlVar_ = ControlVariable<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using Departures_ = Departures<MODEL>;
  using Observations_ = Observations<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using State_ = State<MODEL>;
  using Increment_ = Increment<MODEL>;
  using ObsAuxControls_ = ObsAuxControls<MODEL>;
  using ObsErrors_ = ObsErrors<MODEL>;
  using ObsOperators_ = ObsOperators<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;
  using LinearObsOperators_ = LinearObsOperators<MODEL>;

 public:
  const std::string classname() const override { return "Jo"; }

  /// Define which ObservationSpace to use
  enum OS : size_t { NotScreened = 0, Screened = 1, Quadratic = 2 };

  /// Constructor
  CostJo(const eckit::Configuration &, const Geometry_ &geom,
         const util::DateTime &, const util::DateTime &, const util::Duration &,
         const bool subwindows = false);

  /// Destructor
  virtual ~CostJo() {}

  /// Various set ups prior to the integration of the model
  std::shared_ptr<PostBase<State_> > initialize(
      const CtrlVar_ &, const eckit::Configuration &) const override;

  /// Various set ups prior to the linearisation trajectory computation
  std::shared_ptr<PostBase<State_> > initializeTraj(
      const CtrlVar_ &, const Geometry_ &,
      const eckit::Configuration &) override;

  /// Various computations after the integration of the model
  double finalize(const eckit::Configuration &) const override;

  /// Various computations after the linearisation trajectory computation
  double finalizeTraj(const eckit::Configuration &) override;

  /// Screen observations
  void screenObservations(std::unique_ptr<Observations_> &,
                          const eckit::Configuration &) const;

  /// Initialize ObserverTL (computes \f$ G(dx) \f$) before starting the TL run
  std::shared_ptr<PostBaseTL<Increment_> > setupTL(
      const CtrlInc_ &) const override;

  /// Initialize ObserverAD (computes \f$ G^T y\f$) before starting the AD run
  std::shared_ptr<PostBaseAD<Increment_> > setupAD(
      std::shared_ptr<const GeneralizedDepartures>, CtrlInc_ &) const override;

  /// Multiply by \f$ R\f$
  Departures_ *multiplyCovar(const GeneralizedDepartures &) const override;

  /// Multiply by \f$ R^{-1}\f$.
  Departures_ *multiplyCoInv(const GeneralizedDepartures &) const override;

  /// Randomize \f$ R\f$.
  Departures_ *randomizeCovar() const override;

  /// Provide new (zeroed) departures.
  Departures_ *newDualVector() const override;

  /// Return departures ie \f$ y - {\cal G}(x) \f$.
  Departures_ *newDepartures() const override {
    return new Departures_(*ydep_);
  }

  /// Return gradient at first guess ie \f$ R^{-1}(y - {\cal G}(x))\f$.
  Departures_ *newGradientFG() const override {
    return new Departures_(*gradFG_);
  }

  /// Return observations reference
  const Observations_ &observations() const { return *yobs_; }

  /// Reset obs operator trajectory.
  void resetLinearization() override;

  /// Print Jo table
  double printJo(const Departures_ &, const Departures_ &,
                 const size_t &type) const;

  /// Access observation spaces
  const ObsSpaces_ &obspaces() const { return *obsSpaces_; }

  /// Checkpoint Jo
  void checkpoint(eckit::LocalConfiguration &) const;

  /// Restart Jo
  void restart(const eckit::Configuration &);

  /// Set xb
  void setBackground(const CtrlVar_ &xb) const { xb_ = &xb; }

 private:
  const eckit::LocalConfiguration conf_;  ///< Configuration

  const Geometry_ &geom_;  ///< Constant reference to geometry

  const util::DateTime winbgn_;  ///< Beginning of the window
  const util::DateTime winend_;  ///< End of the window
  const util::Duration tslot_;   ///< Time slot duration

  const bool subwindows_;    ///< Logical controlling if sub-windows are enabled
  mutable bool lsetupOnce_;  ///< Logical controlling various set ups
  mutable bool
      lscreenObservations_;  ///< Logical controlling screening of observations

  const CtrlVar_ mutable *xb_;  ///< Pointer to background
  const CtrlVar_ mutable *fg_;  ///< Pointer to first guess

  mutable std::unique_ptr<ObsSpaces_> obsSpaces_;  ///< Observation spaces

  mutable std::unique_ptr<ObsOperators_>
      hop_;  ///< Non-linear observation operators
  std::shared_ptr<LinearObsOperators_>
      hoptlad_;  ///< Linear observation operators
  mutable std::unique_ptr<ObsErrors_>
      R_;  ///< Observation error covariance matrices

  mutable std::unique_ptr<Observations_> yobs_;  ///< Observations
  mutable std::unique_ptr<Departures_>
      ydep_;  ///< Departures : \f$ y_{obs} - {\cal G}(x_{fg}) \f$
  mutable std::unique_ptr<Departures_> gradFG_;  ///< Gradient at first guess :
                                                 /// \f$ R^{-1} (y_{obs} - {\cal
                                                 ///     G}(x_{fg})) \f$

  /// Observer attached by \f$ J_o\f$ to the model integration
  mutable std::shared_ptr<Observer<MODEL, State_> > pobs_;

  ///  Observer attached by \f$ J_o\f$ to the TrajectorySaver for
  /// linearisation trajectory computation
  mutable std::shared_ptr<Observer<MODEL, State_> > pobstraj_;

  /// ObserverTL attached by \f$ J_o\f$ to the linear model integration
  mutable std::shared_ptr<ObserverTL<MODEL, Increment_> > pobstl_;

  /// ObserverAD attached by \f$ J_o\f$ to the adjoint model integration
  mutable std::shared_ptr<ObserverAD<MODEL, Increment_> > pobsad_;
};

// =============================================================================

template <typename MODEL>
CostJo<MODEL>::CostJo(const eckit::Configuration &joConf, const Geometry_ &geom,
                      const util::DateTime &winbgn,
                      const util::DateTime &winend, const util::Duration &ts,
                      const bool subwindows)
    : conf_(joConf),
      geom_(geom),
      winbgn_(winbgn),
      winend_(winend),
      tslot_(ts),
      subwindows_(subwindows),
      lsetupOnce_(true),
      lscreenObservations_(false),
      xb_(),
      fg_(),
      obsSpaces_(),
      hop_(),
      hoptlad_(),
      R_(),
      yobs_(),
      ydep_(),
      gradFG_(),
      pobs_(),
      pobstraj_(),
      pobstl_(),
      pobsad_() {
  /*!  Depending on the configuration set up either not-screened or screened
   *  observation spaces.
   */
  // Check if screening block is present
  if (conf_.has("screening")) {
    // Check if screening is requested
    bool lscreen = conf_.getBool("screening");
    Log::info() << "CostJo::CostJo lscreen = " << lscreen << std::endl;
    if (lscreen) {
      // Load un-screened ObservationSpaces if screening is requested
      obsSpaces_.reset(
          new ObsSpaces_(conf_, geom_, winbgn_, winend_, OS::NotScreened));
    } else {
      // Load screened ObservationSpaces if screening is not requested
      obsSpaces_.reset(
          new ObsSpaces_(conf_, geom_, winbgn_, winend_, OS::Screened));
    }
  } else {
    // Load screened ObservationSpaces if screening block is missing
    obsSpaces_.reset(
        new ObsSpaces_(conf_, geom_, winbgn_, winend_, OS::Screened));
  }

  Log::trace() << "CostJo:CostJo done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBase<State<MODEL> > > CostJo<MODEL>::initialize(
    const CtrlVar_ &fg, const eckit::Configuration &conf) const {
  /*!  This method is called prior to the non-linear model integration. Its
   * purpose is to construct the observation operator, observation error
   * covariance matrix, read observations. It returns an Observer to be
   * attached to the non-linear trajectory; the purpose of the Observer is to
   * compute model equivalents at observation locations.
   */
  ASSERT(lscreenObservations_ == false);
  /// 1. At each outer loop setup a switch to control if screening is on;
  if (conf.has("screening"))
    lscreenObservations_ = conf.getBool("screening.active");

  /// 2. Store a reference to the first guess
  fg_ = &fg;

  /// 3. Reset observation space if required
  if (not lsetupOnce_ && lscreenObservations_) {
    // Close ObservationSpaces
    obsSpaces_.reset();
    // Open new ObservationSpaces
    obsSpaces_.reset(
        new ObsSpaces_(conf_, geom_, winbgn_, winend_, OS::NotScreened));
  }

  /// 4. Construct \f$\cal H\f$, R, y if required
  if (lsetupOnce_ || lscreenObservations_) {
    // Setup once or reset operators if new screening is to be performed
    hop_.reset(new ObsOperators_(*obsSpaces_));
    R_.reset(new ObsErrors_(*obsSpaces_));
    yobs_.reset(new Observations_(*obsSpaces_));
    yobs_->read(conf_);
  }

  /// 5. Set up the Observer
  pobs_.reset(new Observer<MODEL, State_>(*obsSpaces_, *hop_, fg_->obsVar(),
                                          tslot_, subwindows_));
  /// 6. Return the Observer
  return pobs_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJo<MODEL>::finalize(const eckit::Configuration &conf) const {
  /*!  This method is called after the non-linear model integration. Its purpose
   *  is to compute departures, perform screening if required, set up the
   *  observation error covariance matrix, VarQC weights, print Jo table,
   *  perform diagnostics if required. The Jo cost value is returned.
   */
  /// 1. Retrieve observation equivalents \f$ {\cal G}(x) \f$
  std::unique_ptr<Observations_> yeqv(pobs_->release());
  Log::info() << "Jo Observation Equivalent: " << *yeqv << std::endl;

  // 1.5. Reload observations if H operator change the observation vector
  ///   AROME: this is required for radar observations
  if (conf.has("reload_observations")) {
    yobs_->read(conf_);
  }

  /// 2. Calculate departures \f$ d = y - {\cal G}(x) \f$
  ydep_.reset(new Departures_(*yobs_ - *yeqv));
  Log::info() << "Jo Departures: " << *ydep_ << std::endl;

  /// 3. Write the observation space component of the first guess
  ///   IFS: this is required to write out the VarBC cycle file
  if (conf.has("statesObsAux")) {
    fg_->obsVar().write(conf);
  }

  /// 4. Screen observations
  if (lscreenObservations_) this->screenObservations(yeqv, conf);

  /// 5. Linearize the observation error covariance matrix R
  R_->linearize(fg_->obsVar(), *yeqv);

  /// 6. Setup VarQC weights
  if (conf.getBool("varQC.activate", false)) R_->setupWeights(*ydep_);

  /// 7. Compute FG gradient \f$ R^{-1}d \f$
  gradFG_.reset(R_->inverseMultiply(*ydep_));

  /// 8. Print JO table
  double zjo = this->printJo(*ydep_, *gradFG_, OS::Screened);

  /// 9. Diagnostics
  eckit::LocalConfiguration diagnostics;
  if (conf.has("diagnostics")) {
    diagnostics = eckit::LocalConfiguration(conf, "diagnostics");
    if (diagnostics.has("departures")) {
      const eckit::LocalConfiguration depconf(diagnostics, "departures");
      ydep_->save(depconf);
    }
  }

  /// 10. Reset the Observer
  pobs_.reset();

  // Reset the first guess pointer
  fg_ = nullptr;

  return zjo;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJo<MODEL>::screenObservations(std::unique_ptr<Observations_> &yeqv,
                                       const eckit::Configuration &conf) const {
  /*!  This method is called to perform screening of observations. Once the
   * screening is completed, the observation operator, observation error
   * covariance matrix, observations, departures and observation equivalents
   * are reset.
   */
  // ---------------------------------------------------------------------------
  // MC: should be implemented as a post-processor in the Observer at some point
  // ---------------------------------------------------------------------------
  /// 1.  Print unscreened Jo table (set up of observation error covariance
  /// matrix is required)
  R_->linearize(fg_->obsVar(), *yeqv);
  std::unique_ptr<Departures_> gradFG(R_->inverseMultiply(*ydep_));
  double zjo = this->printJo(*ydep_, *gradFG, OS::NotScreened);
  Log::debug() << "CostJo (before screening)" << zjo << std::endl;

  /// 2. Save departures in the current ObservationSpace
  const eckit::LocalConfiguration depconf(conf, "screening.departures");
  ydep_->save(depconf);

  /// 3. Get a vector of pointers to GeoVaLs from the Observer
  const std::vector<std::shared_ptr<GeoVaLs_> > &gvals = pobs_->geovals();

  /// 4. Perform screening - this saves a screened ObservationSpaces
  obsSpaces_->screenObservations(*ydep_, gvals);

  /// 5. Close the current ObservationSpaces and read in the screened
  /// ObservationSpaces
  obsSpaces_.reset();
  obsSpaces_.reset(
      new ObsSpaces_(conf_, geom_, winbgn_, winend_, OS::Screened));

  /// 6. Reset the observation error covariance matrix \f$R\f$
  R_.reset(new ObsErrors_(*obsSpaces_));

  /// 7. Reset the non-linear observation operator \f$\cal G\f$
  hop_.reset(new ObsOperators_(*obsSpaces_));

  /// 8. Reset observations
  yobs_.reset(new Observations_(*obsSpaces_));
  yobs_->read(conf_);

  /// 9. Reset departures
  ydep_.reset(new Departures_(*obsSpaces_));
  ydep_->read(depconf);

  /// 10. Compute observation equivalents as \f$ {\cal G}(x) = y - d \f$
  yeqv.reset(new Observations_(*yobs_));
  *yeqv -= *ydep_;

  /// 11. Reset the observation space component of the first guess
  ObsAuxControls_ &obsvarfg = const_cast<ObsAuxControls_ &>(fg_->obsVar());
  obsvarfg.reset(*obsSpaces_, "fg", conf);

  /// 12. Reset the observation space component of the background
  ObsAuxControls_ &obsvarbg = const_cast<ObsAuxControls_ &>(xb_->obsVar());
  obsvarbg.reset(*obsSpaces_, "xb", conf);
}
// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBase<State<MODEL> > > CostJo<MODEL>::initializeTraj(
    const CtrlVar_ &fg, const Geometry_ &, const eckit::Configuration &conf) {
  /*!  This method is called prior to the forced model linearisation trajectory
   *  computation in the TrajecotrySaver. Its purpose is to construct the linear
   *  observation operator and the linearisation trajectory Observer. The method
   *  returns the Observer which is subsequently enrolled by the
   *  TrajecotrySaver; the purpose of this Observer is to compute the
   *  linearisation trajectory for the linear observation operator.
   */
  util::Duration tltstep = tslot_;
  if (conf.has("linearmodel"))
    tltstep = util::Duration(conf.getString("linearmodel.tstep"));
  if (conf.has("timeslot"))
    tltstep = util::Duration(conf.getString("timeslot"));

  hoptlad_.reset();
  pobstraj_.reset();

  /// 1. Store a reference to the first guess
  fg_ = &fg;

  /// 2. Set up the linear observation operator
  hoptlad_.reset(new LinearObsOperators_(*obsSpaces_));

  /// 4. Set up the linearisation trajectory observer
  pobstraj_.reset(new Observer<MODEL, State_>(*obsSpaces_, *hop_, fg.obsVar(),
                                              tltstep, subwindows_, hoptlad_));

  /// 4. Return the linearisation trajectory observer
  return pobstraj_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJo<MODEL>::finalizeTraj(const eckit::Configuration &conf) {
  /*!  This method is called once the model linearisation trajectory is
   *  computed. Its purpose is to compute departures w.r.t. the linearisation
   *  trajectory and print the corresponding Jo table if required. The computed
   *  Jo cost is returned.
   */

  /// 1. Retrieve observation equivalents
  std::unique_ptr<Observations_> yeqv(pobstraj_->release());
  Log::info() << "Jo Observation Equivalent:" << *yeqv << std::endl;

  /// 2. Compute departures w.r.t. the linearisation trajecotry
  Departures_ ydep(*yobs_ - *yeqv);
  Log::info() << "Jo Departures:" << ydep << std::endl;

  /// 3. Print the linearisation trajectory JO-table if required
  std::unique_ptr<Departures_> grad(R_->inverseMultiply(ydep));

  // Check config if the inner loop trajectory Jo needs to be printed
  bool printInnerLoopTrajJo = false;
  if (conf.has("printInnerLoopTrajJo"))
    printInnerLoopTrajJo = conf.getBool("printInnerLoopTrajJo");

  double zjo = 0.0;
  if (printInnerLoopTrajJo) {
    // Print JO table
    Log::info() << "Print inner loop trajectory JO-table" << std::endl;
    zjo = this->printJo(ydep, *grad, OS::Screened);
  } else {
    // Calculate JO
    for (std::size_t jj = 0; jj < ydep.size(); ++jj) {
      const double zz =
          0.5 * dot_product(ydep[jj].obsvector(), (*grad)[jj].obsvector());
      zjo += zz;
    }
  }

  /// 4. Store departures w.r.t. the linearisation trajectory if required
  if (conf.has("innerloop_departures")) {
    const eckit::LocalConfiguration depconf(conf, "innerloop_departures");
    ydep.save(depconf);
  }

  // Reset the first guess pointer
  fg_ = nullptr;

  pobstraj_.reset();
  return zjo;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBaseTL<Increment<MODEL> > > CostJo<MODEL>::setupTL(
    const CtrlInc_ &dx) const {
  /*!  This method sets up the ObserverTL post-processor-tl. ObserverTL is
   * typically attached to the linear model integration to compute
   * \f$ G(dx) \f$.
   */
  ASSERT(hoptlad_);
  if (pobstl_ == nullptr)
    pobstl_.reset(new ObserverTL<MODEL, Increment_>(
        *obsSpaces_, *hoptlad_, dx.obsVar(), tslot_, subwindows_));
  else {
    pobstl_->setupTL(dx.obsVar());
  }
  return pobstl_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
std::shared_ptr<PostBaseAD<Increment<MODEL> > > CostJo<MODEL>::setupAD(
    std::shared_ptr<const GeneralizedDepartures> pv, CtrlInc_ &dx) const {
  /*!  This method sets up the ObserverAD post-processor-ad. ObserverAD is
   * typically attached to the adjoint model integration to compute
   * \f$ G^T(y) \f$.
   */
  ASSERT(hoptlad_);
  std::shared_ptr<const Departures_> dy =
      std::dynamic_pointer_cast<const Departures_>(pv);
  if (pobsad_ == nullptr)
    pobsad_.reset(new ObserverAD<MODEL, Increment_>(
        *obsSpaces_, *hoptlad_, dy, dx.obsVar(), tslot_, subwindows_));
  else {
    pobsad_->setupAD(dy, dx.obsVar());
  }
  return pobsad_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Departures<MODEL> *CostJo<MODEL>::multiplyCovar(
    const GeneralizedDepartures &v1) const {
  const Departures_ &y1 = dynamic_cast<const Departures_ &>(v1);
  return R_->multiply(y1);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Departures<MODEL> *CostJo<MODEL>::multiplyCoInv(
    const GeneralizedDepartures &v1) const {
  const Departures_ &y1 = dynamic_cast<const Departures_ &>(v1);
  return R_->inverseMultiply(y1);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Departures<MODEL> *CostJo<MODEL>::randomizeCovar() const {
  Departures_ *ydep = this->newDepartures();
  R_->randomize(*ydep);
  return ydep;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Departures<MODEL> *CostJo<MODEL>::newDualVector() const {
  Departures_ *ydep = new Departures_(*obsSpaces_);
  ydep->zero();
  return ydep;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJo<MODEL>::resetLinearization() {
  hoptlad_.reset();
  pobs_.reset();
  pobstraj_.reset();
  pobstl_.reset();
  pobsad_.reset();
  lsetupOnce_ = false;
  lscreenObservations_ = false;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double CostJo<MODEL>::printJo(const Departures_ &dy, const Departures_ &grad,
                              const size_t &type) const {
  obsSpaces_->printJo(dy, grad);

  double zjo = 0.0;
  for (std::size_t jj = 0; jj < dy.size(); ++jj) {
    const double zz =
        0.5 * dot_product(dy[jj].obsvector(), grad[jj].obsvector());
    const unsigned nobs = dy[jj].size();
    if (nobs > 0) {
      switch (type) {
        case OS::Screened:
          Log::test() << "CostJo   : Nonlinear Jo = " << zz
                      << ", nobs = " << nobs << ", Jo/n = " << zz / nobs
                      << ", err = " << (*R_)[jj].getRMSE() << std::endl;
          break;
        case OS::NotScreened:
          Log::info() << "CostJo (not screened)   : Nonlinear Jo = " << zz
                      << ", nobs = " << nobs << ", Jo/n = " << zz / nobs
                      << ", err = " << (*R_)[jj].getRMSE() << std::endl;
          break;
        case OS::Quadratic:
          Log::info() << "  CostJo : Quadratic Jo = " << zz
                      << ", nobs = " << nobs << ", Jo/n = " << zz / nobs
                      << std::endl;
          break;
        default:
          Log::info() << "CostJo : wrong settings" << std::endl;
      }
      if (dy[jj].hasBias()) {
        ASSERT(grad[jj].hasBias());
        double zjc =
            0.5 * dot_product(dy[jj].biasvector(), grad[jj].biasvector());
        switch (type) {
          case OS::Screened:
            Log::test() << "CostJo   : Nonlinear JoCVarBC = " << zjc
                        << std::endl;
            break;
          case OS::NotScreened:
            Log::info() << "CostJo (not screened)  : Nonlinear JoCVarBC = "
                        << zjc << std::endl;
            break;
        }
      }
    } else {
      Log::info() << "CostJo --- No Observations" << std::endl;
    }
    zjo += zz;
  }

  return zjo;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJo<MODEL>::checkpoint(eckit::LocalConfiguration &conf) const {
  if (conf.has("Jo")) {
    Log::info() << "CostJo<MODEL>::checkpoint starting" << std::endl;
    eckit::LocalConfiguration joconf(conf, "Jo");
    if (joconf.has("departures")) {
      Log::info() << "CostJo<MODEL>::checkpoint savedepartures" << std::endl;
      const eckit::LocalConfiguration depconf(joconf, "departures");
      ydep_->save(depconf);
    }
    obsSpaces_->saveObservations();
    Log::info() << "CostJo<MODEL>::checkpoint done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void CostJo<MODEL>::restart(const eckit::Configuration &conf) {
  if (conf.has("Jo")) {
    Log::info() << "CostJo<MODEL>::restart starting" << std::endl;
    const eckit::LocalConfiguration joConf(conf.getSubConfiguration("Jo"));

    /// 1. Initializations (R, H)
    hop_.reset(new ObsOperators_(*obsSpaces_));
    R_.reset(new ObsErrors_(*obsSpaces_));

    /// 2. Read-in observations
    yobs_.reset(new Observations_(*obsSpaces_));
    yobs_->read(conf_);

    /// 3. Read-in departures
    const eckit::LocalConfiguration depconf(joConf, "departures");
    ydep_.reset(new Departures_(*obsSpaces_));
    ydep_->read(depconf);

    /// 4. Compute observation equivalents as \f$ {\cal G}(x) = y - d
    std::unique_ptr<Observations_> yeqv(new Observations_(*yobs_));
    *yeqv -= *ydep_;

    /// 5. Linearize the observation error covariance matrix R
    R_->linearize(*yeqv);

    /// 6. Setup VarQC weights
    if (joConf.getBool("varQC.activate", false)) {
      R_->setupWeights(*ydep_);
    }

    /// 7. Compute FG gradient \f$ R^{-1}d \f$
    gradFG_.reset(R_->inverseMultiply(*ydep_));

    /// 8. Print JO table
    this->printJo(*ydep_, *gradFG_, OS::Screened);

    Log::info() << "CostJo<MODEL>::restart done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJO_H_
