/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTTERMBASE_H_
#define OOPS_ASSIMILATION_COSTTERMBASE_H_

#include <memory>

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/base/PostBase.h"
#include "oops/base/PostBaseAD.h"
#include "oops/base/PostBaseTL.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"

namespace eckit {
class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief Abstract base class for the terms of the cost function (e.g. Jo, Jc).

 *   A vector of CostTermBase objects is stored by the CostFunction allowing for
 *  flexibility of the choice of the terms of the cost function.

 *   Methods initialize (returns a post-processor)/finalize are called
 *  prior/post the non-linear model trajectory integration.

 *   Methods initializeTraj (returns a post-processor)/finalizeTraj are called
 *  prior/post the linearisation trajectory computation by the TrajecotrySaver.

 *   SetupTL/SetupAD set up post-processors-tl/ad attached to the linear/adjoint
 *  model integration.
 */

template <typename MODEL>
class CostTermBase {
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;
  using Increment_ = Increment<MODEL>;
  using PostPtr_ = std::shared_ptr<PostBase<State_> >;
  using PostTLPtr_ = std::shared_ptr<PostBaseTL<Increment_> >;
  using PostADPtr_ = std::shared_ptr<PostBaseAD<Increment_> >;

 public:
  virtual const std::string classname() const = 0;

  /// Destructor
  virtual ~CostTermBase() {}

  /// Return a post-processor to be attached to the non-linear model integration
  virtual PostPtr_ initialize(const ControlVariable<MODEL> &,
                              const eckit::Configuration &) const = 0;

  /// Return a post-processor to be attached to the linearisation trajectory
  /// computation
  virtual PostPtr_ initializeTraj(const ControlVariable<MODEL> &,
                                  const Geometry_ &,
                                  const eckit::Configuration &) = 0;

  /// Perform computations after nonlinear model integration.
  virtual double finalize(const eckit::Configuration &) const = 0;

  /// Perform computation after the linearisation trajectory computation
  virtual double finalizeTraj(const eckit::Configuration &) = 0;

  /// Return a post-processor-tl to be attached to the TL run.
  virtual PostTLPtr_ setupTL(const ControlIncrement<MODEL> &) const = 0;

  /// Return a post-processor-ad to be attached to the AD run.
  virtual PostADPtr_ setupAD(std::shared_ptr<const GeneralizedDepartures>,
                             ControlIncrement<MODEL> &) const = 0;

  /// Multiply by covariance (or weight) matrix and its inverse.
  virtual GeneralizedDepartures *multiplyCovar(
      const GeneralizedDepartures &) const = 0;

  /// Multiply by inverse covariance (or weight) matrix
  virtual GeneralizedDepartures *multiplyCoInv(
      const GeneralizedDepartures &) const = 0;

  /// Randomize the covariance (or weight) matrix.
  virtual GeneralizedDepartures *randomizeCovar() const = 0;

  /// Provide new dual space vector (for example a Departure for Jo).
  virtual GeneralizedDepartures *newDualVector() const = 0;

  /// Provide departures (for example a Departure for Jo).
  virtual GeneralizedDepartures *newDepartures() const = 0;

  /// Return gradient at first guess.
  virtual GeneralizedDepartures *newGradientFG() const = 0;

  /// Reset trajectory.
  virtual void resetLinearization() = 0;

  /// Checkpoint
  virtual void checkpoint(eckit::LocalConfiguration &) const = 0;

  /// Restart
  virtual void restart(const eckit::Configuration &) = 0;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTTERMBASE_H_
