/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_HTRINVHDXFG_H_
#define OOPS_ASSIMILATION_HTRINVHDXFG_H_

#include <memory>  // for std::unique_ptr

#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/DualVector.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/interface/Increment.h"
#include "util/Logger.h"
#include "util/PrintAdjTest.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {

/// Calculate \f$ H^T R^{-1} H dx_FG \f$.

template <typename MODEL>
class HtRinvHdxFG : private eckit::NonCopyable {
  using Increment_ = Increment<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CostFct_ = CostFunction<MODEL>;

 public:
  explicit HtRinvHdxFG(const CostFct_& j);

  void compute(CtrlInc_& dz) const;

 private:
  CostFct_ const& j_;
  mutable CtrlInc_ dx_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
HtRinvHdxFG<MODEL>::HtRinvHdxFG(const CostFct_& j)
    : j_(j), dx_(j.jb().getFirstGuessIncrement(), true) {}

// -----------------------------------------------------------------------------

template <typename MODEL>
void HtRinvHdxFG<MODEL>::compute(CtrlInc_& dz) const {
  // Setup TL terms of cost function
  PostProcessorTL<Increment_> costtl;
  for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
    costtl.enrollProcessor(j_.jterm(jj).setupTL(dx_));
  }

  // Run TLM
  CtrlInc_ mdx(dx_);
  j_.runTL(mdx, costtl);

  // Get TLM outputs, multiply by covariance inverses, and setup ADJ forcing
  // terms
  j_.zeroAD(dz);
  PostProcessorAD<Increment_> costad;

  DualVector<MODEL> ww;
  DualVector<MODEL> zz;

  for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
    ww.append(costtl.releaseOutputFromTL(jj));
    zz.append(j_.jterm(jj).multiplyCoInv(*ww.getv(jj)));
    costad.enrollProcessor(j_.jterm(jj).setupAD(zz.getv(jj), dz));
  }

  // Run ADJ
  j_.runAD(dz, costad);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_HTRINVHDXFG_H_
