/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_
#define OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_

#include <memory>  // for std::shared_ptr, std::unique_ptr
#include <vector>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/Late4DStart.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/StateInfo.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/Restart.h"

namespace oops {

/*! @brief IncrementalAssimilation implements variational algorithms
 *
 *   The variational algorithms are implemented using the incremental
 *  formulation. IncrementalAssimilation orchestrates the execution of outer
 *  loops. Each outer loop consists of three key stages:

 *    1. Set up the quadratic cost function: compute departures, linearization
        trajectory, set up linear operators

 *    2. Minimize the quadratic cost function

 *    3. Update the first guess

 *   IncrementalAssimilation implements the late 4D start mechanism,
 *  orchestrates checkpointing/restart mechanism and warm starts the first
 *  guess/minimizer if requested.

 *   IncrementalAssimilation is adapted for single- and multi-executable modes.

 *   IFS: multi-executable mode is default, late 4D start is active by default
 */

template <typename MODEL>
class IncrementalAssimilation : private eckit::NonCopyable {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVariable_ = ControlVariable<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Late4DStart_ = Late4DStart<MODEL>;
  using Minimizer_ = Minimizer<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() {
    return "oops::IncrementalAssimilation";
  }

  /// Constructor
  IncrementalAssimilation(CostFunction<MODEL> &);

  /// Destructor
  ~IncrementalAssimilation() {}

  /// Perform incremental assimilation
  void assimilate(const eckit::Configuration &);

  /// Return the analysis
  std::shared_ptr<CtrlVariable_> getAnalysis() const {
    ASSERT(xx_);
    return xx_;
  }

  /// Skip the final non-linear trajectory
  const bool &skipFinalNonlinearTrajectory() const { return skipFinalTraj_; }

 private:
  /// Restart mechanism: checkpoint variational assimilation
  void checkpoint(eckit::LocalConfiguration &conf) const;

  /// Restart mechanism: restart variational assimilation
  void restart(const eckit::Configuration &conf) const;

  /// Restart mechanism: connect with the restart mechanism
  void connectToRestart(util::Restart &restarter) const;

  /// Warm start First Guess
  void warmStartFirstGuess(const eckit::Configuration &config);

  // Other utilities
  /// Skip the final non-linear trajectory
  void setSkipFinalNonlinearTrajectory(const eckit::LocalConfiguration &);

  /// Set the outer loop bounds
  void setOuterLoopBounds(const size_t &);

  /// Set the outer loop index
  void setOuterLoopIndex(const size_t &, const eckit::LocalConfiguration &);

  /// Data members
  CostFunction<MODEL> &J_;                       ///< Cost function
  std::shared_ptr<ControlVariable<MODEL> > xx_;  ///< First guess
  std::unique_ptr<Minimizer<MODEL> > minim_;     ///< Minimizer
  std::unique_ptr<util::Restart> restarter_;     ///< Restart

  size_t jouter_;          ///< Outer loop index
  mutable size_t jstart_;  ///< Index of the first outer loop
  mutable size_t jend_;    ///< Index of the last outer loop
  bool skipFinalTraj_;     ///< Logical controlling skipping the final traj
};

// =============================================================================

template <typename MODEL>
IncrementalAssimilation<MODEL>::IncrementalAssimilation(CostFunction<MODEL> &J)
    : J_(J),
      xx_(),
      minim_(),
      restarter_(),
      jouter_(0),
      jstart_(0),
      jend_(0),
      skipFinalTraj_(false) {
  /// Setup the restarter
  restarter_.reset(new util::Restart(
      {"FirstGuess", "Minimizer", "IncrementalAssimilation", "CostFunction"}));
  this->connectToRestart(*restarter_);
  Log::trace() << "IncrementalAssimilation:constructed" << std::endl;
}

// -----------------------------------------------------------------------------
//
template <typename MODEL>
void IncrementalAssimilation<MODEL>::assimilate(
    const eckit::Configuration &config) {
  /// 1. Initialize first guess from background
  xx_.reset(new CtrlVariable_(J_.jb().getBackground()));
  Log::trace() << classname() << "::assimilate first guess has been set up"
               << std::endl;

  /// 2. Setup configurations of outer loops
  std::vector<eckit::LocalConfiguration> iterconfs;
  config.get("variational.iteration", iterconfs);
  const size_t nouterconfSize = iterconfs.size();

  /// 3. Setup the minimizer
  eckit::LocalConfiguration minConf(config, "minimizer");
  minim_.reset(MinFactory<MODEL>::create(minConf, J_));

  /// 4. Warm start and/or re-centre the first guess
  this->warmStartFirstGuess(config);

  /// 5. Restart the first guess and incremental assimilation
  restarter_->restart({"FirstGuess", "IncrementalAssimilation"});

  /// 6. Set bounds of the loop over outer loops
  this->setOuterLoopBounds(nouterconfSize);

  Log::info() << "Running incremental assimilation with " << nouterconfSize
              << " outer iterations." << std::endl;

  /// 7. Perform the outer loop (trajectory + minimization)
  for (size_t jj = jstart_; jj < jend_; ++jj) {
    /// 7.1 Set current outer loop index
    this->setOuterLoopIndex(jj, iterconfs[jj]);

    //  Get configuration for current outer iteration
    Log::info() << classname()
                << "::assimilate Configuration for outer iteration " << jouter_
                << ":\n"
                << iterconfs[jj];

    /// 7.2 Register the outer loop in the logbook
    util::LogbookWriter<size_t> log("OuterLoop", jouter_);

    /// 7.3 Allow skipping the final nonlinear trajectory; default off
    this->setSkipFinalNonlinearTrajectory(iterconfs[jj]);

    /// 7.4 Setup post-processors for the non-linear trajectory run
    PostProcessor<State_> post;

    /// 7.4.1 Setup diagnostic prints
    if (iterconfs[jj].has("prints")) {
      const eckit::LocalConfiguration prtConfig(iterconfs[jj], "prints");
      post.enrollProcessor(new StateInfo<State_>("traj", prtConfig));
    }

    /// 7.5 Setup late 4D start mechanism (enrolls post-processors)
    Late4DStart_ late4DStart(J_, *xx_, post, iterconfs[jj]);

    /// 7.6 Restart the cost function
    restarter_->restart({"CostFunction"});

    /// 7.7 Setup the quadratic cost function
    J_.linearize(*xx_, iterconfs[jj], post);

    /// 7.8 Checkpoint the cost function (to enable restart from
    //      the minimization)
    restarter_->checkpoint(iterconfs[jj], {"CostFunction"});

    /// Skip the minimization
    if (iterconfs[jj].getBool("skipMinimization", false)) break;

    /// 7.9 Warm start the minimizer (if requested)
    minim_->warmStart(config.getSubConfiguration("variational.WarmStart"));

    /// 7.10 Restart the minimizer (if requested)
    if (jj == jstart_) restarter_->restart({"Minimizer"});

    /// 7.11 Perform the minimization
    std::unique_ptr<CtrlInc_> dx(minim_->minimize(iterconfs[jj]));
    if (jj == jend_) minim_.reset();

    /// 7.12 Add the control increment to the first guess
    J_.addIncrement(*xx_, *dx);

    /// 7.13 Write the control increment (if requested)
    if (iterconfs[jj].has("increment"))
      dx->write(iterconfs[jj].getSubConfiguration("increment"));

    /// 7.14 Reset restart logbook and checkpoint the variational assimilation
    restarter_->reset();
    restarter_->checkpoint(
        iterconfs[jj], {"FirstGuess", "Minimizer", "IncrementalAssimilation"});

    /// 7.15 Perform late 4D start
    late4DStart.late4DStart(*dx);

    /// 7.16 Clean-up the quadratic cost function
    if (!late4DStart.isLate4D()) J_.resetLinearization();
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::setOuterLoopBounds(
    const size_t &nouterconfSize) {
  ///  In single executable mode nouterconfSize is set to the total number of
  ///  outer loops
  /// while in multi-executable mode it is set to one.
  jend_ = nouterconfSize;
  if (jstart_ >= jend_) {
    jstart_ = 0;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::setOuterLoopIndex(
    const size_t &jouter, const eckit::LocalConfiguration &conf) {
  ///  Set the current outer loop index; it is set based on the outer loop index
  /// (jouter) or based on the configuration entry (relied upon in multi
  /// executable mode
  jouter_ = jouter;
  if (conf.has("iteration")) jouter_ = conf.getInt("iteration");
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::setSkipFinalNonlinearTrajectory(
    const eckit::LocalConfiguration &conf) {
  skipFinalTraj_ = conf.getBool("skipFinalTraj", false);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::checkpoint(
    eckit::LocalConfiguration &conf) const {
  ///  The checkpoint mechanism stores the current outer loop index to know
  /// where to restart in case of failure
  conf.set("Iteration", jouter_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::restart(
    const eckit::Configuration &conf) const {
  ///  Set the loop over the outer loops to start from the last completed outer
  /// loop recorded in the restart configuration file
  jstart_ = conf.getUnsigned("Iteration");
  ++jstart_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::connectToRestart(
    util::Restart &restarter) const {
  ///  The restart/checkpoint mechanism works by calling callback functions of
  /// various components of the system: CostFunction, Minimizer, FirstGuess and
  /// IncrementalAssimilation; Each callback function takes a configuration as
  /// input

  /// 1. Register checkpoint callback functions
  restarter.connectChkptCallback(
      "CostFunction",
      [this](eckit::LocalConfiguration &conf) { return J_.checkpoint(conf); });
  restarter.connectChkptCallback("Minimizer",
                                 [this](eckit::LocalConfiguration &conf) {
                                   return minim_->checkpoint(conf);
                                 });
  restarter.connectChkptCallback(
      "FirstGuess",
      [this](eckit::LocalConfiguration &conf) { return xx_->write(conf); });
  restarter.connectChkptCallback("IncrementalAssimilation",
                                 [this](eckit::LocalConfiguration &conf) {
                                   return this->checkpoint(conf);
                                 });

  /// 2. Register restart callback functions
  restarter.connectRstrtCallback(
      "CostFunction",
      [this](const eckit::Configuration &conf) { return J_.restart(conf); });
  restarter.connectRstrtCallback("Minimizer",
                                 [this](const eckit::Configuration &conf) {
                                   return minim_->restart(conf);
                                 });
  restarter.connectRstrtCallback(
      "FirstGuess",
      [this](const eckit::Configuration &conf) { return xx_->read(conf); });
  restarter.connectRstrtCallback(
      "IncrementalAssimilation",
      [this](const eckit::Configuration &conf) { return this->restart(conf); });
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrementalAssimilation<MODEL>::warmStartFirstGuess(
    const eckit::Configuration &config) {
  Log::trace() << classname() << "::warmStartFirstGuess start" << std::endl;

  /// 1. Warm start the first guess
  if (config.has("variational.WarmStart.FirstGuess")) {
    Log::info() << classname() << "::warmStartFirstGuess read First Guess"
                << std::endl;
    xx_->read(config.getSubConfiguration("variational.WarmStart.FirstGuess"));
  }

  /// 2. Re-centre the first guess
  if (config.has("variational.WarmStart.RecentreFirstGuess")) {
    Log::info() << classname() << "::warmStartFirstGuess re-centre First Guess"
                << std::endl;

    eckit::LocalConfiguration conf(config,
                                   "variational.WarmStart.RecentreFirstGuess");

    /// Setup re-centring increment
    Geometry_ fgRecResol(conf.getSubConfiguration("resolution"));
    Variables_ fgRecVars(conf);
    util::DateTime fgRecValidTime(xx_->state()[0].validTime());

    Increment_ fgRecIncr(fgRecResol, fgRecVars, fgRecValidTime);
    fgRecIncr.read(conf.getSubConfiguration("increment"));

    /// Re-centre first guess
    xx_->state()[0] += fgRecIncr;
  }

  Log::trace() << classname() << "::warmStartFirstGuess done" << std::endl;
}

// =============================================================================

}  // namespace oops

#endif  // OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_
