/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_JBMATRIX_H_
#define OOPS_ASSIMILATION_JBMATRIX_H_

#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/base/ObsAuxIncrements.h"
#include "util/Logger.h"
#include "util/PrintAdjTest.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {

/// The \f$ Jb \f$ matrix.
/*!
 *  The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply the \f$ Jb \f$ matrix for
 *  a general case where U' is not equal to U, where B = UU^t
 */

template <typename MODEL>
class JbMatrix : private eckit::NonCopyable {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using ObsAuxIncrements_ = ObsAuxIncrements<MODEL>;

 public:
  explicit JbMatrix(const CostFct_& j, bool adjointTest = false)
      : j_(j),
        adjointTest_(adjointTest),
        dxObsAux_(j.jb().getFirstGuessIncrement().obsVar(), false),
        invBdxObsAux_(j.jb().getFirstGuessIncrement().obsVar(), false) {}
  ~JbMatrix() {}

  void multiply(const CtrlVec_& v, CtrlVec_& jbv) const {
    /*! JbMatrix = | I              |
     *             |   U'^T B^-1 U' |
     * where B = UU^T and U' does not have to be equal to U
     */

    // State part
    jbv.state() = v.state();

    // Mod Aux part
    jbv.modVar() = v.modVar();

    // Obs Aux part
    dxObsAux_.zero();
    invBdxObsAux_.zero();

    // U'v
    j_.jb().jbObsBias().multiplySqrt(v.obsVar(), dxObsAux_);
    // B^-1 U' v
    j_.jb().jbObsBias().inverseMultiply(dxObsAux_, invBdxObsAux_);
    // U'^T B^-1 U' v
    j_.jb().jbObsBias().multiplySqrtTrans(invBdxObsAux_, jbv.obsVar());

    if (adjointTest_) {
      // Adjoint Test

      // U'^t U'v
      double fwd = dot_product(dxObsAux_, invBdxObsAux_);
      double bwd = dot_product(v.obsVar(), jbv.obsVar());

      Log::info() << "ObsBiasCovariance Adjoint Test: " << std::endl
                  << util::PrintAdjTest(fwd, bwd, "B_bias") << std::endl;
    }
  }

 private:
  CostFct_ const& j_;
  const bool adjointTest_;

  mutable ObsAuxIncrements_ dxObsAux_;
  mutable ObsAuxIncrements_ invBdxObsAux_;
};

}  // namespace oops

#endif  // OOPS_ASSIMILATION_JBMATRIX_H_
