/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_LATE4DSTART_H_
#define OOPS_ASSIMILATION_LATE4DSTART_H_

#include <memory>  // for std::shared_ptr, std::unique_ptr
#include <vector>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/utils/Tokenizer.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateSaver.h"
#include "oops/base/StateWriter.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/Restart.h"

namespace oops {

/*! @brief Late4DStart implements the late 4D start mechanism
 *
 */

template <typename MODEL>
class Late4DStart : private eckit::NonCopyable {
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVariable_ = ControlVariable<MODEL>;
  using Increment_ = Increment<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::Late4DStart"; }

  /// Constructor
  Late4DStart(CostFunction<MODEL> &, const ControlVariable<MODEL> &,
              PostProcessor<State_> &, const eckit::LocalConfiguration &);

  /// Destructor
  ~Late4DStart() {}

  /// Late 4D start mechanism
  void late4DStart(const CtrlInc_ &);

  /// Check if the late 4D start mechanism is active
  const bool &isLate4D() const { return late4D_; }

 private:
  /// Data members
  CostFunction<MODEL> &J_;            ///< Cost function
  const ControlVariable<MODEL> &xx_;  ///< First guess
  std::shared_ptr<StateSaver<State_> > stateSaver_;
  std::shared_ptr<StateWriter<State_> > stateWriter_;
  std::unique_ptr<eckit::LocalConfiguration> conf_;

  const bool late4D_;  ///< Logical controlling the late 4D start
};

// =============================================================================

template <typename MODEL>
Late4DStart<MODEL>::Late4DStart(CostFunction<MODEL> &J,
                                const ControlVariable<MODEL> &xx,
                                PostProcessor<State_> &post,
                                const eckit::LocalConfiguration &conf)
    : J_(J),
      xx_(xx),
      stateSaver_(),
      stateWriter_(),
      conf_(),
      late4D_(conf.has("late4d")) {
  if (isLate4D()) {
    conf_.reset(new eckit::LocalConfiguration(conf, "late4d"));

    /// 1 Setup diagnostic prints
    if (conf_->has("prints")) {
      const eckit::LocalConfiguration prtConfig(*conf_, "prints");
      post.enrollProcessor(new StateInfo<State_>("traj", prtConfig));
    }
    /// 2 Setup state writer/saver for late 4D start
    const eckit::LocalConfiguration late4DstateConf(*conf_, "state");
    if (late4DstateConf.has("writeToDisk")) {
      const eckit::LocalConfiguration late4DstateWriteConf(late4DstateConf,
                                                           "writeToDisk");
      stateWriter_.reset(new StateWriter<State_>(late4DstateWriteConf));
      post.enrollProcessor(stateWriter_);
    } else {
      stateSaver_.reset(new StateSaver<State_>(late4DstateConf));
      post.enrollProcessor(stateSaver_);
    }
    Log::trace() << "Late4DStart:constructed" << std::endl;
  }
}

// -----------------------------------------------------------------------------

/*! Late 4D start mechanism is applied in the last outer loop.

 *  The mechanims consists of the following steps:
 *   A.) Record/Save \f$x_{fg}^{steps}\f$ during the first guess trajectory;
 *   B.) Propagate the tangent linear model and record \f$dx_{fg}^{steps}\f$;
 *   C.) Perform sanity checks on \f$dx_{fg}^{steps}\f$;
 *   D.) Obtain 'an' fields: \f$x_{an}^{steps} = x_{fg}^{steps} +
                                                 dx_{fg}^{steps}\f$;
 *   E.) Write out 'an' fields and \f$dx_{fg}^{steps}\f$;
 */

template <typename MODEL>
void Late4DStart<MODEL>::late4DStart(const CtrlInc_ &dx) {
  if (isLate4D()) {
    Log::info() << "Late4DStart::late4DStart starting" << std::endl;
    util::LogbookWriter<bool> logStart("late4DStart", true);
    const eckit::LocalConfiguration late4DstateConf(*conf_, "state");

    /// 1. Propagate the increment to the requested times
    ///  This is achieved by setting up a StateSaver templated on Increment and
    /// adding it to the post processor that is attached to the tangent linear
    /// model propagation
    PostProcessorTL<Increment_> cost;
    PostProcessor<Increment_> post;

    std::shared_ptr<StateSaver<Increment_> > incrementSaver;
    incrementSaver.reset(new StateSaver<Increment_>(late4DstateConf));
    post.enrollProcessor(incrementSaver);

    CtrlInc_ dxtmp(dx);
    J_.runTL(dxtmp, cost, post);

    /// 2. Retrieve the recorded increments
    const std::vector<std::shared_ptr<Increment_> > &increments(
        incrementSaver->releaseFields());

    /// 3. Perform sanity checks on the retrieved increments based on
    /// configuration
    bool lSanityCheckPassed = true;
    std::vector<eckit::LocalConfiguration> confs;
    if (late4DstateConf.has("incrementSanityChecks"))
      late4DstateConf.get("incrementSanityChecks", confs);

    for (auto const &config : confs) {
      const std::vector<std::string> varname =
          config.getStringVector("variables");
      const double zmin = config.getDouble("min");
      const double zmax = config.getDouble("max");
      Variables_ var(config);
      for (auto const &incr : increments) {
        const double zincmin = incr->min(var);
        const double zincmax = incr->max(var);
        Log::info() << "Late4DStart::late4DStart " << std::endl
                    << "  Sanity Test: date: " << incr->validTime()
                    << ", variable: " << varname[0] << ", bounds: (" << zmin
                    << ", " << zmax << ")"
                    << ", min = " << zincmin << ", max = " << zincmax
                    << std::endl;
        if (zincmin >= zmin && zincmax <= zmax) {
          Log::info() << "  Sanity Test: passed" << std::endl;
        } else {
          Log::info() << "  Sanity Test: NOT passed" << std::endl;
          lSanityCheckPassed = false;
        }
      }
    }

    /// 4. Write ObsAux
    xx_.obsVar().write(*conf_);

    /// 4. Write ModAux
    xx_.modVar().write(*conf_);

    /// 6. Reset linearization, close ODB
    J_.resetLinearization();

    /// 7. Write 4V fields (+0H)
    if (conf_->getBool("write4Vfields") || not lSanityCheckPassed) {
      xx_.state()[0].write(late4DstateConf);
      util::LogbookWriter<bool> log("4VfieldsWrittenOut", true);
    }

    /// 8. Register the sanity check outcome in the logbook
    util::LogbookWriter<bool> log("Late4D_check", lSanityCheckPassed);

    /// 9. Check if all the sanity tests passed. Skip the final trajecotry
    ///   if this is the case
    if (lSanityCheckPassed) {
      Log::info() << "Late4DStart::late4DStart " << std::endl
                  << "  All increment sanity tests PASSED   " << std::endl
                  << "  Skipping final non-linear trajectory" << std::endl
                  << *conf_ << std::endl;

      if (late4DstateConf.has("writeToDisk")) {
        ///  If this option is enabled, the first guess fields are not
        /// available in memory. Instead, they were written out to disk during
        /// the non-linear trajectory and need to be read in to produce the
        /// required AN fields.
        //
        /// 10-A. Read in saved FG states
        eckit::LocalConfiguration writeToDiskConf(
            late4DstateConf.getSubConfiguration("writeToDisk"));
        ASSERT(writeToDiskConf.has("restart_file"));
        ASSERT(writeToDiskConf.has("steps"));
        util::DateTime refTime(xx_.state()[0].validTime());
        const std::string restartFileBase =
            writeToDiskConf.getString("restart_file");
        const std::string steps = writeToDiskConf.getString("steps");
        eckit::Tokenizer tok("\t\n\v\f\r !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~");
        std::vector<std::string> tokens;
        tok(steps, tokens);
        /// Check if the number of requested AN fields agrees with the number of
        // stored increments.
        ASSERT(tokens.size() == increments.size());
        size_t jj = 0;
        /// Create a state to hold first guess fields.
        State_ state(State_(xx_.state()[0]));
        /// 11-A. Release first guess
        // if (skipFinalNonlinearTrajectory()) xx_.reset();

        for (const std::string &token : tokens) {
          util::DateTime validTime(refTime);
          const util::Duration step(token);
          validTime += step;
          /// Check if the requested valid time of a FG state agrees with the
          // valid time of the increment.
          ASSERT(validTime == (*increments[jj]).validTime());

          /// 11.1-A Read in first guess fields
          std::string restartFile = restartFileBase + validTime.toString();
          Log::info() << "Late4DStart::late4DStart " << std::endl
                      << "  -> read in FG state: " << restartFile << std::endl;
          eckit::LocalConfiguration config(writeToDiskConf);
          config.set("restart_file", restartFile);
          config.set("step", step.toString());
          state.read(config);

          /// 11.1-B Add increment to the first guess state
          state += *increments[jj];

          /// 11.2-C Write AN fields
          state.write(late4DstateConf);

          /// 11.3-D Write AN increments
          increments[jj]->write(late4DstateConf);

          Log::info() << "IncrementalAssimilation::late4Dstart " << std::endl
                      << "  State An valid at: " << state.validTime()
                      << " written out." << std::endl;
          jj++;
        }
      } else {
        /// The first guess fields are available in memory. AN fields are
        /// produced by adding increments valid at appropriate times.
        //
        /// 10-B. Retrieve saved FG states kept in memory
        std::vector<std::shared_ptr<State_> > states;
        if (stateSaver_) states = stateSaver_->releaseFields();

        size_t nAnStates = states.size();
        ASSERT(increments.size() == nAnStates);

        /// 11-B. Release first guess
        // if (skipFinalNonlinearTrajectory()) xx_.reset();

        /// 12-B. Compute and write AN fields
        for (size_t jj = 0; jj < nAnStates; ++jj) {
          /// 13.1-B Add increment to the first guess state
          *states[jj] += *increments[jj];

          /// 13.2-B Write AN fields
          states[jj]->write(late4DstateConf);

          /// 13.3-B Write AN increments
          increments[jj]->write(late4DstateConf);

          Log::info() << "Late4DStart::late4DStart " << std::endl
                      << "  State An valid at: " << states[jj]->validTime()
                      << " written out." << std::endl;

          /// 13.4 Release AN fields
          states[jj].reset();
        }
      }

      Log::info() << "Late4DStart::late4DStart finished" << std::endl;
    }
  }
}

// =============================================================================

}  // namespace oops

#endif  // OOPS_ASSIMILATION_LATE4DSTART_H_
