/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_MINIMIZER_H_
#define OOPS_ASSIMILATION_MINIMIZER_H_

#include <map>
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/BMatrix.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/DualVector.h"
#include "oops/assimilation/HMatrix.h"
#include "oops/assimilation/HtMatrix.h"
#include "oops/assimilation/MMatrix.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/assimilation/MtMatrix.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/PrintAdjTest.h"
#include "util/abor1_cpp.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief Abstract base class for minimizers
 *
 *  Minimizer defines an interface for families of minimizers; it manages the
 * minimization and adjoint/tangent linear model testing. The following
 * families of minimizers are available: DRMinimizer (right B preconditioned
 * minimizers), DualMinimizer (observation space minimizers), PrimalMinimizer
 * (model space minimizers), SQRTMinimizer (square root  minimizers),
 * SaddlePointMinimizer (saddle point formulation minimizers), LBMinimizer
 * (left B precondtioned minimizers)

 *  Concrete minimizers are produced by the MinFactory

 *  IFS: SQRTMinimizer is employed by default
 */

template <typename MODEL>
class Minimizer : private eckit::NonCopyable {
  using B_ = BMatrix<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using Dual_ = DualVector<MODEL>;
  using H_ = HMatrix<MODEL>;
  using Ht_ = HtMatrix<MODEL>;
  using M_ = MMatrix<MODEL>;
  using Mt_ = MtMatrix<MODEL>;
  using State_ = State<MODEL>;

 public:
  virtual const std::string classname() const = 0;

  /// Construct a minimizer
  explicit Minimizer(const CostFct_ &J) : J_(J), outerIteration_(0){};

  /// Destroy a minimizer
  virtual ~Minimizer() {}

  /// Orchestrate the minimization of a quadratic cost function
  ControlIncrement<MODEL> *minimize(const eckit::Configuration &);

  /// Checkpoint the minimizer (virtual method)
  virtual void checkpoint(eckit::LocalConfiguration &) const = 0;

  /// Restart the minimizer (virtual method)
  virtual void restart(const eckit::Configuration &) = 0;

  /// Warm start the minimizer (virtual method)
  virtual void warmStart(const eckit::Configuration &) = 0;

 private:
  /// Minimizer the quadratic cost function (virtual method)
  virtual ControlIncrement<MODEL> *doMinimize(const eckit::Configuration &) = 0;

  /// Perform adjoint tests
  void adjTests(const eckit::Configuration &) const;

  /// Perform model adjoint tests
  void adjModelTest() const;

  /// Perform observation operator adjoint tests
  void adjObsTest() const;

  /// Perfrom B adjoint tests
  void adjBMatrixTest() const;

  /// Perform tangent linear model tests
  void tlmTests(const eckit::Configuration &) const;

  /// Perform tangent linear model approximation test
  void tlmApproxTest() const;

  /// Perform the Tayolor test
  void tlmTaylorTest() const;

  /// Peroform the tangent linear propagation test
  void tlmPropagTest(const eckit::Configuration &config,
                     const CtrlInc_ &) const;

  const CostFct_ &J_;   ///< Handle to the cost function
  int outerIteration_;  ///< Current outer loop index
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlIncrement<MODEL> *Minimizer<MODEL>::minimize(
    const eckit::Configuration &config) {
  /// 1. Perform TLM tests
  this->tlmTests(config);

  /// 2. Perform ADJ tests
  this->adjTests(config);

  /// 3. Minimize the cost function
  ControlIncrement<MODEL> *dx = this->doMinimize(config);

  /// 4. Perform TLM propagation test
  this->tlmPropagTest(config, *dx);

  /// 5. Update outer loop counter
  outerIteration_++;

  return dx;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::tlmTests(const eckit::Configuration &config) const {
  /// Tangent Linear tests are performed if onlineDiagnostics configuration is
  /// found

  if (config.has("onlineDiagnostics")) {
    const eckit::LocalConfiguration onlineDiag(config, "onlineDiagnostics");
    bool runTlmTaylorTest = onlineDiag.getBool("tlmTaylorTest");
    bool runTlmApproxTest = onlineDiag.getBool("tlmApproxTest");

    // Online tests
    /// 1. TLM linear approximation test
    if (outerIteration_ == 0 && runTlmApproxTest) {
      util::LogbookWriter<std::string> log("TlmTests", "TlmApproxTest");
      this->tlmApproxTest();
    }

    /// 2. TLM Taylor test
    if (outerIteration_ == 0 && runTlmTaylorTest) {
      util::LogbookWriter<std::string> log("TlmTests", "TlmTaylorTest");
      this->tlmTaylorTest();
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::adjTests(const eckit::Configuration &config) const {
  /// Adjoint tests are performed if onlineDiagnostics configuration is found

  if (config.has("onlineDiagnostics")) {
    const eckit::LocalConfiguration onlineDiag(config, "onlineDiagnostics");
    bool runAdjTlmTest = onlineDiag.getBool("adjTlmTest", false);
    bool runAdjObsTest = onlineDiag.getBool("adjObsTest", false);
    bool runAdjBMatrixTest = onlineDiag.getBool("adjBMatrixTest", false);

    // Online tests
    /// 1. Model adjoint test
    if (runAdjTlmTest) {
      util::LogbookWriter<std::string> log("AdjTests", "AdjModelTests");
      this->adjModelTest();
    }

    /// 2. Obs adjoint test
    if (runAdjObsTest) {
      util::LogbookWriter<std::string> log("AdjTests", "AdjObsTests");
      this->adjObsTest();
    }

    /// 3. B matrix adjoint test
    if (runAdjBMatrixTest) {
      util::LogbookWriter<std::string> log("AdjTests", "AdjBMatrixTests");
      this->adjBMatrixTest();
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::tlmApproxTest() const {
  /*! TL approx test:

   *  calculate and store:

   *    M(x + dx), M(x), M'(dx)

   *  where dx is initialized using randomization

   *  TO DO:

   *    write to file depending on requirements
   */

  Log::info() << std::endl
              << "TLM Linear Approximation Test - starting: " << outerIteration_
              << std::endl
              << std::endl;

  /// 1. Construct the linear model matrix operator
  const M_ M(J_);

  /// 2. Initialize the control incremet using randomization
  CtrlInc_ dx(J_.jb().getFirstGuessIncrement(), false);
  J_.jb().randomize(dx);

  /// 3. Run the non-linear model starting from unperturbed initial conditions
  /// (x_b)
  PostProcessor<State_> pp;
  ControlVariable<MODEL> mxx(J_.jb().getBackground());
  J_.runNL(mxx, pp);

  /// 4. Run the non-linear model starting from perturbed initial conditions
  /// (x_b + dx)
  ControlVariable<MODEL> mpertxx(J_.jb().getBackground());
  J_.addIncrement(mpertxx, dx);
  J_.runNL(mpertxx, pp);

  /// 5. Propagate the perurbation with the tangent linear model
  CtrlInc_ mdx(dx);
  M.multiply(mdx);

  // print log
  Log::info() << std::endl
              << "TLM Linear Approximation Test: done." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::tlmPropagTest(const eckit::Configuration &config,
                                     const CtrlInc_ &dx) const {
  /*! TL propagation test:

   *  calculate and store:

   *    M'(dx)

   *  where dx comes from minimization stopped at requested iteration

   *  TO DO:

   *    write to file depending on requirements
   */

  if (config.has("onlineDiagnostics")) {
    const eckit::LocalConfiguration onlineDiag(config, "onlineDiagnostics");
    bool runTlmPropagTest = onlineDiag.getBool("tlmPropagTest");

    if (runTlmPropagTest) {
      // print log
      Log::info() << "TLM Propagation Test - starting: " << outerIteration_
                  << std::endl
                  << std::endl;

      /// 1. Construct linear model matrix operator
      const M_ M(J_);

      /// 2. Propagete the control increment with the tangent linear model:
      /// M'(dx)
      CtrlInc_ mdx(dx);
      M.multiply(mdx);

      // print log
      Log::info() << std::endl << "TLM Propagation Test: done." << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::tlmTaylorTest() const {
  /*! TL Taylor test:

   *  ||M(x + p dx) - M(x)|| / ||M'(p dx)|| -> 1

   *  where p is a scalar damping factor and dx is initialized
   *  using randomization
   */

  Log::info() << std::endl
              << "TLM Taylor Test: " << outerIteration_ << std::endl;

  /// 1. Construct linear model matrix
  const M_ M(J_);

  /// 2. Instantiate and initialize the control increment using randomization
  CtrlInc_ dx(J_.jb().getFirstGuessIncrement(), false);
  J_.jb().randomize(dx);

  /// 3. Run the non-linear model starting from unperturbed initial conditions
  /// (x_b)
  PostProcessor<State_> pp;
  ControlVariable<MODEL> mxx(J_.jb().getBackground());
  J_.runNL(mxx, pp);

  /// 5. Propagate the un-damped perturbation with the tangent linear model:
  /// M'(dx)
  CtrlInc_ mdx(dx);
  M.multiply(mdx);

  /// 6. Loop over increments scaled with increasing damping factor
  for (unsigned int i = 0; i < 14; ++i) {
    /// 6.1 Compute ||p M'(dx)||
    CtrlInc_ pmdx(mdx);
    pmdx *= 1. / pow(10.0, i);
    double denom = sqrt(dot_product(pmdx, pmdx));

    /// 6.2 Propagate the perturbed state with the non-linear model NL: M(x+pdx)
    ControlVariable<MODEL> mpertxx(J_.jb().getBackground());
    CtrlInc_ pdx(dx);
    pdx *= 1. / pow(10.0, i);
    J_.addIncrement(mpertxx, pdx);
    J_.runNL(mpertxx, pp);

    /// 6.3 Compute ||M(x+pdx) - M(x)||
    CtrlInc_ diff_nl(mdx, false);
    diff_nl.state()[0].diff(mpertxx.state()[0], mxx.state()[0]);
    double nom = sqrt(dot_product(diff_nl, diff_nl));

    /// 6.4 Print results
    Log::info() << std::endl
                << "TLM Taylor test:  p = " << std::setw(8) << 1. / pow(10.0, i)
                << ", ||M(x) - M(x+p dx)|| / ||p M'(dx)|| = "
                << util::full_precision(1. + std::abs(1. - nom / denom))
                << std::endl
                << std::endl;
  }
  Log::info() << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::adjModelTest() const {
  /*! Perform the adjoint test of the linear model:

   *  <M dx1, dx2> - <dx1, Mt dx2>)/<M dx1, dx2>

   *  where dx1 and dx2 are increments obtained through randomization
   */

  /// 1. Construct the linear model matrix operator and its adjoint
  const M_ M(J_);
  const Mt_ Mt(J_);

  /// 2. Instantiate and initialize increments using randomization
  CtrlInc_ dx1(J_.jb().getFirstGuessIncrement(), false);
  CtrlInc_ dx2(J_.jb().getFirstGuessIncrement(), false);

  J_.jb().randomize(dx1);
  J_.jb().randomize(dx2);

  /// 3. Propagate dx1 with the tangent linear model
  CtrlInc_ mdx1(dx1);
  M.multiply(mdx1);

  /// 4. Propagate backwards dx2 with the adjoint of the linear model
  CtrlInc_ mtdx2(dx2);
  mtdx2.state()[0].updateTime(mdx1.state()[0].validTime() -
                              dx1.state()[0].validTime());
  Mt.multiply(mtdx2);

  /// 5. Calculate < M dx1, dx2 >
  double adj_tst_fwd = dot_product(mdx1, dx2);

  /// 6. Calculate < dx1, Mt dx2 >
  double adj_tst_bwd = dot_product(dx1, mtdx2);

  /// 7. Print results
  Log::info() << "Model Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "M") << std::endl
              << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::adjObsTest() const {
  /*!  Perform the adjoint test of the linear observation operator:

   *  (<H dx, dy> - <dx, Ht dy>)/<H dx, dy>

   *  where dx is an increment obtained through randomization and
   *  dy is a DualVector obtained by first creating an increment
   *  through randomization and then applying a linear observation
   *  operator

   *   Note that the \f$H = [\overline{H} P] [dx d\beta]\f$, where
   \f$\overline{H}\f$
   *  is the linear observation operator and \f$P\f$ is the linear bias
   *  operator
   */

  /// A. Perform  adjoint test of the composite observation and bias operator
  ///  \f$[\overline{H} P] [dx d\beta]\f$

  /// Construct linear obs operator matrix and its adjoint
  const H_ H(J_);
  const Ht_ Ht(J_);

  /// A.1 Instantiate and initialize increments and dual vectors
  CtrlInc_ dx1(J_.jb().getFirstGuessIncrement(), false);
  CtrlInc_ dx2(J_.jb().getFirstGuessIncrement(), false);
  CtrlInc_ hthdx2(J_.jb().getFirstGuessIncrement(), false);
  Dual_ hdx1;
  Dual_ hdx2;

  // randomize increments
  J_.jb().randomize(dx1);
  J_.jb().randomize(dx2);

  /// A.2 Create a new set of increments
  std::unique_ptr<CtrlInc_> dx1wrk(new CtrlInc_(dx1));
  std::unique_ptr<CtrlInc_> dx2wrk(new CtrlInc_(dx2));

  /// A.3 Compute H dx1 and H dx2
  H.multiply(*dx1wrk, hdx1, true);
  H.multiply(*dx2wrk, hdx2, true);

  /// A.4 Compute H^T (H dx2)
  J_.zeroAD(hthdx2);
  Ht.multiply(hdx2, hthdx2, true);

  /// A.5 Calculate FWD < H dx1, H dx2 >
  double adj_tst_fwd = dot_product(hdx1, hdx2);

  /// A.6 Calculate BWD < dx1, H^T (H dx2) >
  double adj_tst_bwd = dot_product(*dx1wrk, hthdx2);

  /// A.7 Print results
  Log::info() << "Obs Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "HP") << std::endl
              << std::endl;
  //
  ///  B. Perform adjoint test of the observation operator only
  ///  \f$\overline{H}\f$. This is achieved by zeroing out the auxiliary
  ///  observation space component of the control increment
  //
  hdx1.zero();
  hdx2.zero();

  /// B.1 Reset increments
  dx1wrk.reset(new CtrlInc_(dx1));
  dx2wrk.reset(new CtrlInc_(dx2));

  /// B.2 Zero out auxiliary obs space component of ControlIncrements
  dx1wrk->obsVar().zero();
  dx2wrk->obsVar().zero();

  /// B.3 Compute H dx1 and H dx2
  H.multiply(*dx1wrk, hdx1, true);
  H.multiply(*dx2wrk, hdx2, true);

  /// B.4 Copmpute H^T (H dx2)
  J_.zeroAD(hthdx2);
  Ht.multiply(hdx2, hthdx2, true);

  /// B.5 Calculate FWD < H dx1, H dx2 >
  adj_tst_fwd = dot_product(hdx1, hdx2);

  /// B.6 Calculate BWD < dx1, H^T (H dx2) >
  adj_tst_bwd = dot_product(*dx1wrk, hthdx2);

  /// B.7 Print results
  Log::info() << "Obs Only Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "H") << std::endl
              << std::endl;

  //
  ///  C. Perform adjoint test of the bias operator only \f$P\f$. This is
  ///  achieved by zeroing out the state increment component of the control
  ///  increment
  //
  hdx1.zero();
  hdx2.zero();

  /// C.1 Reset increments
  dx1wrk.reset(new CtrlInc_(dx1));
  dx2wrk.reset(new CtrlInc_(dx2));

  /// C.2 Reset state increment component of ControlIncrements
  dx1wrk->state().zero();
  dx2wrk->state().zero();

  /// C.3 Compute H dx1 and H dx2
  H.multiply(*dx1wrk, hdx1, true);
  H.multiply(*dx2wrk, hdx2, true);

  /// C.4 Compute H^T (H dx2)
  J_.zeroAD(hthdx2);
  Ht.multiply(hdx2, hthdx2, true);

  /// C.5 Calculate FWD < H dx1, H dx2 >
  adj_tst_fwd = dot_product(hdx1, hdx2);

  /// C.6 Calculate BWD < dx1, Ht (H dx2) >
  adj_tst_bwd = dot_product(*dx1wrk, hthdx2);

  /// C.7 Print results
  Log::info() << "Obs Bias Only Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "P") << std::endl
              << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Minimizer<MODEL>::adjBMatrixTest() const {
  /*!  Perform the adjoint test of the linear observation operator:

   *  (<UU^T dx, dx> - <U^T dx, U^T dx>)/<UU^T dx, dx>

   *  where dx is an increment obtained through randomization

   */

  /// A. Perform  adjoint test of the composite B matrix diag{Bstate Bmod Bobs}

  /// Construct B matrix operator
  const B_ B(J_);

  /// A.1 Instantiate and initialize increments
  CtrlInc_ dx(J_.jb().getFirstGuessIncrement(), false);
  CtrlVec_ utdx(J_.jb());
  CtrlInc_ uutdx(J_.jb());

  /// A.2 Randomize increments
  J_.jb().randomize(dx);

  /// A.3 Compute U^T dx and UU^T dx
  B.multiplySqrtTrans(dx, utdx);
  B.multiplySqrt(utdx, uutdx);

  /// A.4 Calculate FWD < UU^T dx, dx >
  double adj_tst_fwd = dot_product(uutdx, dx);

  /// A.5 Calculate BWD < U^T dx, U^T dx >
  double adj_tst_bwd = dot_product(utdx, utdx);

  /// A.6 Print results
  Log::info() << "B Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "B", "UU^T dx",
                                    "dx", "U^T dx", "U^T dx")
              << std::endl;
  //
  /// B. Perform  adjoint test of the B matrix diag{Bstate 0 0}
  //

  /// B.1 Instantiate and initialize increments
  CtrlInc_ dxstate(dx, true);

  /// B.2 Zero out the observation and model error components
  dxstate.modVar().zero();
  dxstate.obsVar().zero();
  utdx.zero();
  uutdx.zero();

  /// B.3 Compute U^T dx and UU^T dx
  B.multiplySqrtTrans(dxstate, utdx);
  B.multiplySqrt(utdx, uutdx);

  /// B.4 Calculate FWD < UU^T dx, dx >
  adj_tst_fwd = dot_product(uutdx, dxstate);

  /// B.5 Calculate BWD < U^T dx, U^T dx >
  adj_tst_bwd = dot_product(utdx, utdx);

  /// B.6 Print results
  Log::info() << "B-state Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "B-state",
                                    "UU^T dx", "dx", "U^T dx", "U^T dx")
              << std::endl;

  //
  /// C. Perform  adjoint test of the B matrix diag{0 Bmod 0}
  //

  /// C.1 Instantiate and initialize increments
  CtrlInc_ dxmod(dx, true);

  /// C.2 Zero out the state and observation components
  dxmod.state().zero();
  dxmod.obsVar().zero();
  utdx.zero();
  uutdx.zero();

  /// C.3 Compute U^T dx and UU^T dx
  B.multiplySqrtTrans(dxmod, utdx);
  B.multiplySqrt(utdx, uutdx);

  /// C.4 Calculate FWD < UU^T dx, dx >
  adj_tst_fwd = dot_product(uutdx, dxmod);

  /// C.5 Calculate BWD < U^T dx, U^T dx >
  adj_tst_bwd = dot_product(utdx, utdx);

  /// C.6 Print results
  Log::info() << "B-mod Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "B-mod",
                                    "UU^T dx", "dx", "U^T dx", "U^T dx")
              << std::endl;

  //
  /// D. Perform adjoint test of the B matrix diag{0 0 Bobs}
  //

  /// D.1 Instantiate and initialize increments
  CtrlInc_ dxobs(dx, true);

  /// D.2 Zero out the observation and model error components
  dxobs.state().zero();
  dxobs.modVar().zero();
  utdx.zero();
  uutdx.zero();

  /// D.3 Compute U^T dx and UU^T dx
  B.multiplySqrtTrans(dxobs, utdx);
  B.multiplySqrt(utdx, uutdx);

  /// D.4 Calculate FWD < UU^T dx, dx >
  adj_tst_fwd = dot_product(uutdx, dxobs);

  /// D.5 Calculate BWD < U^T dx, U^T dx >
  adj_tst_bwd = dot_product(utdx, utdx);

  /// D.6 Print results
  Log::info() << "B-obs Adjoint Test: " << outerIteration_ << std::endl
              << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "B-obs",
                                    "UU^T dx", "dx", "U^T dx", "U^T dx")
              << std::endl;
}

// -----------------------------------------------------------------------------

/// Minimizer Factory
template <typename MODEL>
class MinFactory {
  using CostFct_ = CostFunction<MODEL>;

 public:
  static Minimizer<MODEL> *create(const eckit::Configuration &,
                                  const CostFct_ &);
  virtual ~MinFactory() { getMakers().clear(); }

 protected:
  explicit MinFactory(const std::string &);

 private:
  virtual Minimizer<MODEL> *make(const eckit::Configuration &,
                                 const CostFct_ &) = 0;
  static std::map<std::string, MinFactory<MODEL> *> &getMakers() {
    static std::map<std::string, MinFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <class MODEL, class FCT>
class MinMaker : public MinFactory<MODEL> {
  using CostFct_ = CostFunction<MODEL>;
  virtual Minimizer<MODEL> *make(const eckit::Configuration &conf,
                                 const CostFct_ &J) {
    return new FCT(conf, J);
  }

 public:
  explicit MinMaker(const std::string &name) : MinFactory<MODEL>(name) {}
};

// =============================================================================

template <typename MODEL>
MinFactory<MODEL>::MinFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in minimizer factory."
                 << std::endl;
    ABORT("Element already registered in MinFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Minimizer<MODEL> *MinFactory<MODEL>::create(const eckit::Configuration &config,
                                            const CostFct_ &J) {
  std::string id = config.getString("algorithm");
  Log::info() << "Minimizer algorithm=" << id << std::endl;
  typename std::map<std::string, MinFactory<MODEL> *>::iterator j =
      getMakers().find(id);
  if (j == getMakers().end()) {
    Log::error() << id << " does not exist in minimizer factory." << std::endl;
    ABORT("Element does not exist in MinFactory.");
  }
  return (*j).second->make(config, J);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_MINIMIZER_H_
