/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SQRTMINIMIZER_H_
#define OOPS_ASSIMILATION_SQRTMINIMIZER_H_

#include <iterator>
#include <map>
#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/BMatrix.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/JbMatrix.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/assimilation/UtHtRinvHUMatrix.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {

/*! @brief Abstract base class for square root minimizers

 *  SQRTMinimizer is the base class for all minimizers that use the square root
 * \f$ U\f$ of the \f$ B = UU^T\f$ matrix to precondition the variational
 * minimisation problem through the introduction of an auxiliary variable
 * \f$ dx = U \hat{v} \f$ or \f$ dx = U \hat{v} + x_b - x_fg \f$.

 *  IFS: SQRTPLanczosMinimizer is employed by default, the change of variable
 * \f$ dx = U \hat{v} + x_b - x_fg \f$ is default.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class SQRTMinimizer : public Minimizer<MODEL> {
  using Bmat_ = BMatrix<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using Jbmat_ = JbMatrix<MODEL>;
  using Minimizer_ = Minimizer<MODEL>;
  using UtHtRinvHU_ = UtHtRinvHUMatrix<MODEL>;

 public:
  virtual const std::string classname() const override = 0;

  /// Construct the minimizer
  SQRTMinimizer(const eckit::Configuration &conf, const CostFct_ &J)
      : Minimizer_(J),
        J_(J),
        lclassicChaVar_(true),
        linfoConv_(false),
        ninner_(30),
        ninnerMin_(0),
        gnreduc_(1.e-14),
        gradJb_(),
        dv0_(),
        dvh_(),
        costJ0Jb_(0),
        costJ0JbJq_(0),
        setObsAuxFormulation_(false),
        obsAuxFormulation_(),
        iter_(0),
        maxiter_(999) {
    /// Over-write default configuration following user input
    this->setupSolver(conf);
  }
  /// Destroy the minimizer
  ~SQRTMinimizer() {}

  /// Checkpoint the minimizer
  void checkpoint(eckit::LocalConfiguration &) const;

  /// Restart the minimizer
  void restart(const eckit::Configuration &);

  /// Warm start the minimizer
  void warmStart(const eckit::Configuration &);

 private:
  /// Solver (virtual method) - implemented by concrete minimizers
  CtrlInc_ *doMinimize(const eckit::Configuration &) override;
  virtual double solve(CtrlVec_ &, CtrlVec_ &, const Jbmat_ &,
                       const UtHtRinvHU_ &, const bool &, const size_t &,
                       const size_t &, const double &, const double &,
                       const double &) = 0;

  /// Checkpoint the limited memory preconditioner (virtual method)
  virtual void checkpointLMP(eckit::LocalConfiguration &) const = 0;

  /// Restart the limited memory preconditioner (virtual method)
  virtual void restartLMP(const eckit::Configuration &) = 0;

  void checkpointGradJb(eckit::LocalConfiguration &) const;
  void checkpointDv0(eckit::LocalConfiguration &) const;
  void checkpointDvh(eckit::LocalConfiguration &) const;
  void checkpointAux(eckit::LocalConfiguration &) const;

  void restartGradJb(const eckit::Configuration &);
  void restartDv0(const eckit::Configuration &);
  void restartDvh(const eckit::Configuration &);
  void restartAux(const eckit::Configuration &);
  void warmStartAux(const eckit::Configuration &);

  /// Quadratic cost function calculation (Jb component)
  void updateJbCostFunction(const CtrlVec_ &dv, const CtrlVec_ &jbdv);

  /// Setup solver controls (general settings)
  void setupSolver(const eckit::Configuration &);

  /// Setup the minimization (settings specific to the current minimization)
  void setupMinimization(const eckit::Configuration &);

  ///  Enforce incremental/non-incremental formulation of observation space
  /// components of the control increment
  void enforceObsAuxFormulation(CtrlVec_ &dv) const;

  // Data members
  const CostFct_ &J_;  ///< Handle to the cost function

  // Controls
  bool lclassicChaVar_;  ///< Logical controlling the choice of the change of
                         /// variables
  bool linfoConv_;       ///<  Logical controlling if the information content
                         /// convergence
  ///< criterion is enabled
  size_t ninner_;     ///< Maximum number of iterations of the solver
  size_t ninnerMin_;  ///< Minimum number of iterations of the solver
  double gnreduc_;    ///< Gradient norm reduction

  // Control vectors
  std::unique_ptr<CtrlVec_>
      gradJb_;                     ///< Jb component of the first guess gradient
  std::unique_ptr<CtrlVec_> dv0_;  ///< Initial control vector
  std::vector<CtrlVec_> dvh_;      ///<

  // Cost function Jb
  double costJ0Jb_;    ///< value of the Jb component of the cost function
  double costJ0JbJq_;  ///< value of the JbJq component of the cost function

  // Controls for ObsAux formulation
  mutable bool
      setObsAuxFormulation_;  ///<  Logical controlling the enforcement of
                              ///< incremental/non-incremental formulation of
  ///< observation space components of the control increment
  mutable std::map<std::string, bool> obsAuxFormulation_;

 protected:
  size_t iter_;             ///< Outer loop index
  mutable size_t maxiter_;  ///< Number of outer loops
};

// =============================================================================

template <typename MODEL>
ControlIncrement<MODEL> *SQRTMinimizer<MODEL>::doMinimize(
    const eckit::Configuration &config) {
  iter_++;
  ASSERT(iter_ <= maxiter_);
  Log::info() << std::endl;
  Log::info() << classname() << ": starting" << std::endl;

  /// A.1 Setup minimization specific controls
  this->setupMinimization(config);

  /// A.2 Setup the operators B, Jb, UtHtRinvHU
  const Jbmat_ Jb(J_);
  const Bmat_ B(J_);
  const UtHtRinvHU_ UtHtRinvHU(J_, config);

  /// A.3 Instantiate a control increment meant to hold \f$ H^T R^{-1} d \f$ or
  ///   \f$ H^T R^{-1} [d + H (x_{fg} - x_b) - dv_0] \f$
  CtrlInc_ fgrhs(J_.jb().getFirstGuessIncrement(), false);

  /// A.4 Instantiate a control vector meant to hold \f$  U^T H^T R^{-1} d \f$
  /// or
  ///  \f$ H^T R^{-1} [d + H (x_{fg} - x_b) - dv_0] \f$
  CtrlVec_ rhs(J_.jb());

  /// A.5 Instantiate a control vector meant to hold the solution
  CtrlVec_ dv(J_.jb());

  double costJ0JoJc(0.0);
  if (lclassicChaVar_) {
    ///  B.1 Setup the first guess gradient for the classic change of variables
    /// \f$ dx = U dv \f$

    ///  B.2 Interpolate Jb component of the first guess gradient \f$ -\sum dv_i
    ///  \f$
    /// to the new inner loop resolution
    if (gradJb_ == 0) {
      gradJb_.reset(new CtrlVec_(J_.jb()));
    } else {
      gradJb_.reset(new CtrlVec_(J_.jb(), *gradJb_));
    }
    // Calculate first guess gradient
    /// B.3 Compute \f$ H^T R^{-1} d \f$
    J_.computeGradientFG(fgrhs);

    /// B.4 Compute \f$ U^T H^T R^{-1} d \f$
    B.multiplySqrtTrans(fgrhs, rhs);

    /// B.5 Compute the total fist guess gradinet
    ///    \f$ -\sum dv_i + U^T H^T R^{-1} d \f$
    J_.jb().addGradientFG(rhs, *gradJb_);

    /// B.6 Set \f$ J[0] = 0.5 d^T R^{-1} d \f$
    costJ0JoJc = J_.getCostJoJc();
  } else {
    ///  C.1 Setup the first guess gradient for the change of variables
    /// \f$ dx = U dv + x_b - x_{fg} \f$

    /// C.2 Interpolate initial guess control vector to the new inner loop
    /// resolution
    if (dv0_ == 0) {
      dv0_.reset(new CtrlVec_(J_.jb()));
      /// C.3.a Calculate first guess gradient \f$ H^T R^{-1} d \f$
      J_.computeGradientFG(fgrhs);
      /// C.3.b Set \f$ J[0] = 0.5 d^T R^{-1} d \f$
      costJ0JoJc = J_.getCostJoJc();
    } else {
      dv0_.reset(new CtrlVec_(J_.jb(), *dv0_));
      /// C.3.b Calculate first guess gradient
      ///     \f$ H^T R^{-1} [d + H ( dx_{fg}  - U dv_0) ] \f$
      /// and set \f$ J[0] = 0.5 (d + H ( dx_fg  - U dv0))^T R^{-1}
      ///                        (d + H ( dx_fg  - U dv0)) \f$
      costJ0JoJc = J_.computeGradientFG(*dv0_, fgrhs);
    }
    ///  C.4 Calculate first guess gradient in control space
    /// \f$ U^T H^T R^{-1} [d + H ( dx_{fg} - U dv_0 ) ] \f$
    B.multiplySqrtTrans(fgrhs, rhs);
    /*! Note that:

     *  \f$ U^T H^T R^{-1} [d + H ( dx_{fg} - U dv_0 ) ] - J_b dv_0 \f$

     *  is equivalent to
     *
     *  \f$ U^T H^T R^{-1} d + U^T H^T R^{-1} H dx_{fg} -
     *      [J_b + U^T H^T R^{-1} H U] dv_0 \f$

     */
    /// C.5 Compute \f$ J_b dv_0 \f$
    CtrlVec_ jbdv0(J_.jb());
    Jb.multiply(*dv0_, jbdv0);
    rhs -= jbdv0;
    /// C.6 Initialise \f$ dv \f$ to \f$ dv_0 \f$
    dv = *dv0_;
  }

  /// D.1 Solve the linear system
  double reduc = this->solve(dv, rhs, Jb, UtHtRinvHU, linfoConv_, ninner_,
                             ninnerMin_, gnreduc_, costJ0Jb_, costJ0JoJc);

  /// D.2 Enforce the formulation of the auxiliary observation space component
  /// of the control increment
  this->enforceObsAuxFormulation(dv);

  /// D.3 Calculate the increment \f$ dx_k = U dv \f$
  CtrlInc_ *dx = new CtrlInc_(J_.jb().getFirstGuessIncrement(), false);
  B.multiplySqrt(dv, *dx);

  // Print
  if (linfoConv_) {
    Log::test() << classname()
                << ": final relative gain in information content = " << reduc
                << std::endl;
  } else {
    Log::test() << classname() << ": reduction in residual norm = " << reduc
                << std::endl;
  }
  Log::info() << classname() << " output increment:" << *dx << std::endl;

  /// D.4 Calculate quadratic cost function
  CtrlVec_ jbdv(dv, false);
  Jb.multiply(dv, jbdv);

  // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
  this->updateJbCostFunction(dv, jbdv);

  /// D.5 Book-keeping for next minimization/(transformation of the increment)
  if (lclassicChaVar_) {
    ///  \f$ dx = U dv \f$:

    /// Update the Jb component of the first guess gradient
    *gradJb_ += dv;
  } else {
    ///  \f$ dx = U dv + x_b - x_{fg} \f$ :

    /// Save inital guess
    *dv0_ = dv;
    /// Tranform \f$ dx_k \f$ to be defined w.r.t. the first guess
    const CtrlInc_ &dxfg = J_.jb().getFirstGuessIncrement();
    *dx -= dxfg;
  }

  Log::info() << classname() << ": done" << std::endl;

  return dx;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::setupSolver(const eckit::Configuration &conf) {
  ///  Setup general solver settings (consistent for all minimizations)
  /// from configuration
  /// 1. Setup the choice of the change of variable
  if (conf.has("classic_change_of_variables")) {
    // true  : dx = U dv
    // false : dx = U dv + x_b - x_fg
    lclassicChaVar_ = conf.getBool("classic_change_of_variables");
  }

  /// 2. Check if the information content based convergence criteria enabled
  if (conf.has("linfoconv")) linfoConv_ = conf.getBool("linfoconv");

  /// 3. Check if LMP preconditioner is enabled
  if (conf.has("preconditioner") || conf.has("obsAuxFormulations")) {
    // Provide number of outer loops in order to skip the update of the
    // preconditioner in the final minimization or/and to properly apply
    // obs aux incremental formulation
    ASSERT(conf.has("minimizations"));
    maxiter_ = conf.getUnsigned("minimizations");
  }

  /// 4. Set formulation (incremental/non-incremental) of the observation space
  ///   components of the control vector
  if (conf.has("obsAuxFormulations")) {
    Log::info()
        << "SQRTMinimizer<MODEL>::setupSolver() set ObsAux* formulations"
        << std::endl;
    setObsAuxFormulation_ = true;
    std::vector<eckit::LocalConfiguration> configs;
    conf.get("obsAuxFormulations", configs);
    for (size_t ii = 0; ii < configs.size(); ++ii) {
      std::string key(configs[ii].getString("component"));
      bool value(configs[ii].getBool("incremental"));
      obsAuxFormulation_.insert(std::make_pair(key, value));
      Log::info() << "   ObsAuxComponent: " << key
                  << ", incremental formulation: " << value << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::setupMinimization(const eckit::Configuration &conf) {
  ///  Setup settings of the solver specific for the current minimization (can
  /// be different across different minimizations (IFS: consistent settings are
  /// used for all minimizations)

  /// 1. Set the minimum number of iterations
  if (conf.has("ninnermin")) ninnerMin_ = conf.getUnsigned("ninnermin");

  /// 2. Set the maximum number of iterations
  if (conf.has("ninner")) ninner_ = conf.getUnsigned("ninner");
  ASSERT(ninner_ >= ninnerMin_);
  util::LogbookWriter<size_t> log_nin("maxiters", ninner_);

  /// 3. Set the solver tolerance
  if (conf.has("gradient_norm_reduction"))
    gnreduc_ = conf.getDouble("gradient_norm_reduction");
  util::LogbookWriter<double> log_tol("tolerance", gnreduc_);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::enforceObsAuxFormulation(CtrlVec_ &dv) const {
  /*!  The formulation (incremental/non-incremental) is enforced for the
   *  observations space components of the control vector except in the last
   *  outer loop.

   *   IFS: VarBC has non-incremental formulation; TOVSCV has incremental
   *  formulation in a standard AN experiment and non-incremental formulation
   *  in continuous DA.
   */
  if (setObsAuxFormulation_ && iter_ < maxiter_) {
    Log::info() << "SQRTMinimizer<MODEL>::enforceObsAuxFormulation(dv)"
                << std::endl;
    std::map<std::string, bool>::const_iterator icst =
        obsAuxFormulation_.begin();
    while (icst != obsAuxFormulation_.end()) {
      if (!icst->second) {
        for (std::size_t jobs = 0; jobs < dv.obsVar().size(); ++jobs) {
          if (dv.obsVar()[jobs].size() > 0) {
            dv.obsVar()[jobs].obsAuxCtlVec(icst->first).zero();
            Log::info() << "  set component of the control vector: "
                        << icst->first << " to zero" << std::endl;
          }
        }
      }
      ++icst;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::updateJbCostFunction(const CtrlVec_ &dv,
                                                const CtrlVec_ &jbdv) {
  ///  Compute the cost of the Jb component of the cost function. In case of
  /// the classic change of variables formulation store Jb.dv
  if (lclassicChaVar_) {
    dvh_.push_back(jbdv);

    // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
    costJ0Jb_ += 0.5 * dot_product(dv, jbdv);
    costJ0JbJq_ += 0.5 * dot_product(dv.modVar(), jbdv.modVar());

    for (size_t jouter = 1; jouter < dvh_.size(); ++jouter) {
      CtrlVec_ dvhtmp(J_.jb(), dvh_[jouter - 1]);
      // MC temp
      costJ0Jb_ += dot_product(dv.state(), dvhtmp.state());
      costJ0JbJq_ += dot_product(dv.modVar(), dvhtmp.modVar());
    }
  } else {
    // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
    costJ0Jb_ = 0.5 * dot_product(dv, jbdv);
    costJ0JbJq_ = 0.5 * dot_product(dv.modVar(), jbdv.modVar());
  }

  // Print out costJ0Jb and its ObsAux components
  Log::info() << "SQRTMinimizer: costJ0Jb      : "
              << util::full_precision(costJ0Jb_) << std::endl;
  Log::info() << "SQRTMinimizer: costJ0JbJq    : "
              << util::full_precision(costJ0JbJq_) << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::checkpoint(eckit::LocalConfiguration &conf) const {
  /// Checkpoint the minimizer based on the configuration
  Log::info() << " " << classname() << "::checkpoint starting" << std::endl;

  /// \f$ dx = U dv \f$ : \n
  /// A.1 Checkpoint GradJb (
  checkpointGradJb(conf);

  /// A.2 Checkpoint dvh
  // checkpointDvh(conf);

  /// \f$ dx = U dv + x_b - x_{fg} \f$ : \n
  /// B.1 Checkpoint dv0
  checkpointDv0(conf);

  /// Both formulations: \n
  /// C.1 Checkpoint costJ0Jb
  checkpointAux(conf);

  /// C.2 Checkpoint LMP
  if (conf.has("preconditioner")) this->checkpointLMP(conf);

  /// C.3 Checkpoint the current outer loop iteration
  conf.set("outerIteration", iter_);

  Log::info() << " " << classname() << "::checkpoint done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::warmStart(const eckit::Configuration &config) {
  if (config.has("Minimizer") && iter_ == 0) {
    /// Warm start the minimizer based on the configuration
    Log::info() << " " << classname() << "::warmStart starting" << std::endl;

    /// Get the configuration
    eckit::LocalConfiguration conf(config.getSubConfiguration("Minimizer"));

    /// \f$ dx = U dv \f$ : \n
    /// A.1 Warm start GradJb
    if (conf.has("gradJb")) restartGradJb(conf);

    /// A.2 Warm start dvh
    // if (conf.has("dvh"))
    //  restartDvh(conf);

    /// \f$ dx = U dv + x_b - x_{fg} \f$ : \n
    /// B.1 Warm start dv0
    if (conf.has("dv0")) restartDv0(conf);

    /// Both formulations: \n
    /// C.1 Warm start costJ0Jb
    if (conf.has("quadraticCost")) {
      restartAux(conf.getSubConfiguration("quadraticCost"));
    } else {
      warmStartAux(conf);
    }

    /// C.2 Warm start LMP
    if (conf.has("preconditioner")) this->restartLMP(conf);

    Log::info() << " " << classname() << "::warmStart done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::restart(const eckit::Configuration &conf) {
  /// Restart the minimizer following the configuration
  Log::info() << " " << classname() << "::restart starting" << std::endl;

  /// \f$ dx = U dv \f$ : \n
  /// A.1 Restart GradJb
  restartGradJb(conf);

  /// A.2 Restart dvh
  // restartDvh(conf);

  /// \f$ dx = U dv + x_b - x_{fg} \f$ : \n
  /// B.1 Restart dv0
  restartDv0(conf);

  /// Both formulations: \n
  /// C.1 Restart costJ0Jb
  restartAux(conf);

  /// C.2 Restart LMP
  if (conf.has("preconditioner")) {
    this->restartLMP(conf);
  }

  /// C.3 Set the current outer loop iteration
  iter_ = conf.getUnsigned("outerIteration");

  Log::info() << " " << classname() << "::restart done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::checkpointGradJb(
    eckit::LocalConfiguration &conf) const {
  /// Checkpoint GradJb (\f$ -\sum dv_i \f$)
  if (lclassicChaVar_) {
    ASSERT(conf.has("gradJb"));
    {
      eckit::LocalConfiguration gradjbconf(conf, "gradJb");
      gradJb_->write(gradjbconf);
      Log::info() << " " << classname() << "::checkpoint gradJb: "
                  << util::full_precision(dot_product(*gradJb_, *gradJb_))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::checkpointDv0(
    eckit::LocalConfiguration &conf) const {
  /// Checkpoint dv0 (initial guess)
  if (not lclassicChaVar_) {
    ASSERT(conf.has("dv0"));
    {
      eckit::LocalConfiguration dv0conf(conf, "dv0");
      ASSERT(dv0_);
      dv0_->write(dv0conf);
      Log::info() << " " << classname() << "::checkpoint dv0: "
                  << util::full_precision(dot_product(*dv0_, *dv0_))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::checkpointDvh(
    eckit::LocalConfiguration &conf) const {
  ///  Checkpoint dvh (Jb.dv) for all minimizations; required for computation
  /// of the quadratic cost function value only)
  if (lclassicChaVar_) {
    ASSERT(conf.has("dvh"));
    {
      std::vector<eckit::LocalConfiguration> dvhconfs;
      conf.get("dvh", dvhconfs);
      ASSERT(dvh_.size() == dvhconfs.size());

      for (size_t ii = 0; ii < dvh_.size(); ++ii) {
        dvh_[ii].write(dvhconfs[ii]);
        Log::info() << " " << classname() << "::checkpoint dvh[" << ii << "]: "
                    << util::full_precision(dot_product(dvh_[ii], dvh_[ii]))
                    << std::endl;
      }
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::checkpointAux(
    eckit::LocalConfiguration &conf) const {
  ///  Checkpoint costJ0Jb (required for computation of the quadratic cost
  ///  function only)
  conf.set("costJ0Jb", util::full_precision(costJ0Jb_));
  Log::info() << " " << classname()
              << "::checkpoint costJ0Jb: " << util::full_precision(costJ0Jb_)
              << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::restartGradJb(const eckit::Configuration &conf) {
  /// Restart GradJb (\f$ -\sum dv_i \f$)
  if (lclassicChaVar_) {
    ASSERT(conf.has("gradJb"));
    eckit::LocalConfiguration gradjbconf(conf, "gradJb");
    if (gradJb_ == 0) {
      if (gradjbconf.getBool("gatherIO", true)) {
        gradJb_.reset(new CtrlVec_(J_.jb()));
      } else {
        gradJb_.reset(new CtrlVec_(J_.jbRestart()));
      }
    }
    gradJb_->read(gradjbconf);
    Log::info() << " " << classname() << "::restart gradJb: "
                << util::full_precision(dot_product(*gradJb_, *gradJb_))
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::restartDv0(const eckit::Configuration &conf) {
  /// Restart dv0 (initial guess)
  if (not lclassicChaVar_) {
    ASSERT(conf.has("dv0"));
    eckit::LocalConfiguration dv0conf(conf, "dv0");
    if (dv0_ == 0) {
      if (dv0conf.getBool("gatherIO", true)) {
        dv0_.reset(new CtrlVec_(J_.jb()));
      } else {
        dv0_.reset(new CtrlVec_(J_.jbRestart()));
      }
    }
    dv0_->read(dv0conf);
    Log::info() << " " << classname() << "::restart dv0: "
                << util::full_precision(dot_product(*dv0_, *dv0_)) << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::restartDvh(const eckit::Configuration &conf) {
  ///  Restart dvh (Jb.dv for all minimizations; required for computation
  /// of the quadratic cost function value only)
  if (lclassicChaVar_) {
    ASSERT(conf.has("dvh"));
    std::vector<eckit::LocalConfiguration> dvhconfs;
    conf.get("dvh", dvhconfs);
    for (size_t ii = 0; ii < dvhconfs.size(); ++ii) {
      std::unique_ptr<CtrlVec_> dvh;
      if (dvhconfs[ii].getBool("gatherIO", true)) {
        dvh.reset(new CtrlVec_(J_.jb()));
      } else {
        dvh.reset(new CtrlVec_(J_.jbRestart()));
      }
      dvh->read(dvhconfs[ii]);
      dvh_.push_back(*dvh);
      Log::info() << " " << classname() << "::restart dvh[" << ii << "]: "
                  << util::full_precision(dot_product(dvh_[ii], dvh_[ii]))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::restartAux(const eckit::Configuration &conf) {
  ///  Restart costJ0Jb (required for computation of the quadratic cost function
  ///  only)
  ASSERT(conf.has("costJ0Jb"));
  costJ0Jb_ = conf.getDouble("costJ0Jb");
  Log::info() << " " << classname()
              << "::restart costJ0Jb: " << util::full_precision(costJ0Jb_)
              << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SQRTMinimizer<MODEL>::warmStartAux(const eckit::Configuration &conf) {
  ///  Warm start costJ0Jb (required for computation of the quadratic cost
  ///  function only)
  costJ0Jb_ = 0.0;
  if (lclassicChaVar_) {
    // not implemented
  } else {
    if (conf.has("dv0")) {
      eckit::LocalConfiguration dv0conf(conf, "dv0");
      if (not dv0conf.getBool("gatherIO", true))
        dv0_.reset(new CtrlVec_(J_.jb(), *dv0_));
      costJ0Jb_ = 0.5 * dot_product(*dv0_, *dv0_);
    }
  }
  Log::info() << " " << classname()
              << "::warmStart costJ0Jb: " << util::full_precision(costJ0Jb_)
              << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_SQRTMINIMIZ_H_
