/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SPECTRALSQRTLMP_H_
#define OOPS_ASSIMILATION_SPECTRALSQRTLMP_H_

#include <algorithm>
#include <cmath>
#include <memory>
#include <numeric>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/maths/Matrix.h"
#include "oops/assimilation/CholDecomp.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/SymMatSpectrum.h"
#include "oops/assimilation/TriDiagSpectrum.h"
#include "util/Logger.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {

/*!  @brief The SPECTRAL preconditioner matrix \f$ P \f$ for the system matrix
 *  \f$ A \f$;
 *
 *   It is defined as \f$ P_(k+1)  = P_k  +  U (D^{-1} - I_l) U' \f$
 *  where  \f$ D \f$ is a l-by-l diagonal matrix whose diagonal entries are
 *  \f$ lambda_i \f$, \f$ U = [u_1, u_2, ..., u_l] \f$ is a column matrix.
 *
 *   The pairs \f$ (lambda_i,u_i) \f$ define the Ritz pairs of the matrix
 *  \f$ PA \f$ with respect to the subspace \f$ K(PA,Pr0) \f$
 *   Note that \f$ r0 \f$ is the initial residual.
 *
 *   The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply the preconditioner matrix
 *  \f$ P \f$.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class SpectralSqrtLMP {
  using CostFct_ = CostFunction<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;

 public:
  const std::string classname() const { return "SpectralSqrtLMP"; }

  /// Constructor, destructor
  explicit SpectralSqrtLMP(const eckit::Configuration &, const CostFct_ &);
  ~SpectralSqrtLMP() {}

  /// Methods
  void update(const std::vector<std::shared_ptr<CtrlVec_> > &,
              const std::vector<double> &, const std::vector<double> &);
  void postprocess();
  void changeResolution();
  void inverseMultiply(const CtrlVec_ &, CtrlVec_ &) const;
  void inverseMultiplySqrt(const CtrlVec_ &, CtrlVec_ &) const;
  void multiplySqrt(const CtrlVec_ &, CtrlVec_ &) const;

  /// Checkpointing, restart of the limited memory preconditioner
  void checkpoint(eckit::LocalConfiguration &) const;
  void restart(const eckit::Configuration &);

 private:
  const CostFct_ &J_;

  const bool active_;
  bool debug_;
  bool orthogonalize_;
  bool useoldpairs_;
  size_t maxpairs_;
  size_t update_;
  double mineigval_;
  double errthreshold_;

  std::vector<size_t> usedpairIndx_;
  std::vector<double> eigvals_;
  std::vector<std::shared_ptr<CtrlVec_> > U_;
};

// =============================================================================

template <typename MODEL>
SpectralSqrtLMP<MODEL>::SpectralSqrtLMP(const eckit::Configuration &conf,
                                        const CostFct_ &J)
    : J_(J),
      active_(conf.has("preconditioner")),
      debug_(false),
      orthogonalize_(true),
      useoldpairs_(false),
      maxpairs_(0),
      update_(0),
      mineigval_(1.0),
      errthreshold_(0.001),
      usedpairIndx_(),
      eigvals_(),
      U_() {
  if (active_) {
    Log::info() << "SpectralSqrtLMP::SpectralSqrtLMP constructor starting"
                << std::endl;

    const eckit::LocalConfiguration precond(conf, "preconditioner");
    if (precond.has("maxpairs")) maxpairs_ = precond.getInt("maxpairs");
    if (precond.has("error_threshold"))
      errthreshold_ = precond.getDouble("error_threshold");
    if (precond.has("debug")) debug_ = precond.getBool("debug");
    if (precond.has("orthogonalize"))
      orthogonalize_ = precond.getBool("orthogonalize");
    if (precond.has("useoldpairs"))
      useoldpairs_ = precond.getBool("useoldpairs");
    if (precond.has("mineigval")) mineigval_ = precond.getDouble("mineigval");

    ASSERT(maxpairs_ > 0);
    ASSERT(errthreshold_ > 0);

    Log::info() << "  debug_ = " << debug_ << std::endl
                << "  orthogonalize_ = " << orthogonalize_ << std::endl
                << "  useoldpairs_ = " << useoldpairs_ << std::endl
                << "  mineigval_ = " << mineigval_ << std::endl
                << "  maxpairs_ = " << maxpairs_ << std::endl
                << "  errthreshold_ = " << errthreshold_ << std::endl;

    Log::info() << "SpectralSqrtLMP::SpectralSqrtLMP constructor done"
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::update(
    const std::vector<std::shared_ptr<CtrlVec_> > &Zhl,
    const std::vector<double> &alphas, const std::vector<double> &betas) {
  if (active_) {
    Log::info() << "SpectralSqrtLMP::update starting" << std::endl;
    //  If useoldpairs = false, use only current information
    if (!useoldpairs_) {
      eigvals_.clear();
      usedpairIndx_.clear();
      U_.clear();
    }

    const size_t nvec = Zhl.size();
    if (nvec > 0) {
      ASSERT(alphas.size() == nvec - 1);
      ASSERT(betas.size() == nvec - 1);

      //    Compute spectrum of tri-diagonal matrix
      std::vector<double> evals;
      std::vector<std::vector<double> > evecs;
      TriDiagSpectrum(alphas, betas, evals, evecs);

      //    Determine the converged eigenvalues
      std::vector<size_t> convIndx;
      std::vector<double> neweigvals;
      for (int ji = nvec - 2; ji >= 0; ji--) {
        double erritz = evecs[ji][nvec - 2] * betas[nvec - 2];
        double lambda = evals[ji];
        //      Add new converged eigenvalues
        if (std::abs(erritz) <= errthreshold_ * lambda && lambda > mineigval_ &&
            eigvals_.size() < maxpairs_) {
          Log::info() << "Eigenvalue accepted: " << lambda
                      << ", error= " << std::abs(erritz)
                      << ", threshold= " << errthreshold_ * lambda << std::endl;
          convIndx.push_back(ji);
          eigvals_.push_back(lambda);
          neweigvals.push_back(lambda);
        } else {
          if (eigvals_.size() >= maxpairs_)
            Log::info() << "Warning: Maximum allowed number of pairs exceeded"
                        << std::endl;

          Log::info() << "Eigenvalue not accepted: " << lambda
                      << ", error= " << std::abs(erritz)
                      << ", threshold= " << errthreshold_ * lambda << std::endl;
        }
      }
      //    Keep information on how many pairs are used at each minimization
      //    loop
      usedpairIndx_.push_back(convIndx.size());
      Log::info() << " Outer loop                    : " << update_ << std::endl
                  << " Number of inner iterations    : " << Zhl.size() - 1
                  << std::endl
                  << " Maximum number of pairs       : " << maxpairs_
                  << std::endl
                  << " Number of new converged pairs : " << convIndx.size()
                  << std::endl
                  << " New converged eigenvalues     : "
                  << util::full_precision(neweigvals) << std::endl;
      size_t count = 0;
      for (unsigned kiter = 0; kiter < usedpairIndx_.size(); ++kiter) {
        count += usedpairIndx_[kiter];
        Log::info() << " Number of used pairs from outer loop " << kiter
                    << " : " << usedpairIndx_[kiter] << std::endl;
      }
      Log::info() << " Total number of used pairs    : " << count << std::endl;
      //    Calculate/Update the matrix U
      for (size_t ji = 0; ji < convIndx.size(); ++ji) {
        std::shared_ptr<CtrlVec_> uu(new CtrlVec_(J_.jb()));
        uu->zero();
        for (size_t ii = 0; ii < nvec - 1; ++ii) {
          uu->axpy(evecs[convIndx[ji]][ii], *Zhl[ii]);
        }
        //    Insert new eigenvectors
        U_.push_back(uu);
      }
    }
    ++update_;
    Log::info() << "SpectralSqrtLMP::update done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::postprocess() {
  if (active_) {
    /*  The Ritz vectors combined from different minimizations are not
     * necessarily orthogonal, while the preconditioner requires a set of
     * orthogonal vectors. We first compute eigenvectors v_i of (R) (R)^T,
     * where R=Q U and U is a rectangular matrix with columns being Ritz
     * vectors and Q is a Hausholder matrix - R is an upper triangular matrix
     * and (R)(R)^T has only a non-zero block in the top left corner of size
     * LxL, where L is the number of Ritz vectors forming U. The required
     * preconditioning vectors are given by w_i = Q^T v_i and the required
     * eigenvalues are given by eigenvalues of the LxL matrix.
     */
    Log::info() << "SpectralSqrtLMP::postprocess starting" << std::endl;

    if (orthogonalize_) {
      Log::info() << "SpectralSqrtLMP::postprocess orthogonalization starting"
                  << std::endl;
      size_t nvecs = U_.size();
      ASSERT(nvecs <= maxpairs_);

      // Copy and transform preconditioning vectors/eigenvectors
      std::vector<std::shared_ptr<CtrlVec_> > Q;
      Q.clear();
      for (size_t jj = 0; jj < nvecs; ++jj) {
        std::shared_ptr<CtrlVec_> qqvec(new CtrlVec_(*U_[jj]));
        U_[jj].reset();
        double scale = std::sqrt(eigvals_[jj] - 1);
        (*qqvec) *= scale;
        Q.push_back(qqvec);
      }

      // mp_R matrix - stored in a column major order
      Matrix<double> mp_R(nvecs, nvecs);
      for (size_t jj = 0; jj < nvecs; ++jj) {
        std::vector<double> qqvec(jj + 1);
        Q[jj]->state()[0].modCtlVec().getSection(qqvec, 1, jj + 1);
        double zz =
            dot_product(*Q[jj], *Q[jj]) -
            std::inner_product(qqvec.begin(), qqvec.end(), qqvec.begin(), 0.0);
        if (qqvec[jj] < 0) {
          mp_R(jj, jj) = -std::sqrt(zz + qqvec[jj] * qqvec[jj]);
        } else {
          mp_R(jj, jj) = std::sqrt(zz + qqvec[jj] * qqvec[jj]);
        }
        qqvec[jj] = qqvec[jj] - mp_R(jj, jj);
        for (size_t jk = 0; jk < jj; ++jk) {
          mp_R(jk, jj) = qqvec[jk];
        }
        zz = zz + qqvec[jj] * qqvec[jj];
        std::vector<double> zzvec(jj + 1);
        zzvec[jj] = qqvec[jj];
        Q[jj]->state()[0].modCtlVec().setSection(zzvec, 1, jj + 1);
        *Q[jj] *= std::sqrt(2.0 / zz);

        // We now have Householder vectors in Q[jj-1] and the non-zero elements
        // of the transformed vector in mp_R. Now apply the Householder
        // transfromations to the remaining vectors.

        for (size_t ji = jj + 1; ji < nvecs; ++ji) {
          double zz = dot_product(*Q[jj], *Q[ji]);
          Q[ji]->axpy(-zz, *Q[jj]);
        }
      }

      if (debug_) {
        Log::info() << "SpectralSqrtLMP::QR check orthonormality of Q:"
                    << std::endl;
        for (size_t jj = 0; jj < nvecs; ++jj) {
          for (size_t jk = 0; jk < nvecs; ++jk) {
            double zz = dot_product(*Q[jj], *Q[jk]);
            Log::info() << "  <Q[" << jj << "," << jk
                        << "]> = " << util::full_precision(zz) << std::endl;
          }
          Log::info() << std::endl;
        }
      }

      // Multiply the upper triangle by its transpose and find eigenvectors and
      // eigenvalues
      for (size_t jj = 0; jj < nvecs; ++jj) {
        for (size_t ji = jj + 1; ji < nvecs; ++ji) {
          mp_R(ji, jj) = 0.0;
        }
      }

      // RR^T matrix - stored in a column major order
      Matrix<double> mp_RRT = mp_R * mp_R.transpose();

      // Find eigenpairs of RR^T = L U L^T
      ColVector<double> cv_U;
      Matrix<double> mp_L;
      SymMatSpectrum(mp_RRT, cv_U, mp_L);

      // Put eigenvalues in a standard vector
      std::vector<double> evals(cv_U.size());
      for (size_t jj = 0; jj < evals.size(); ++jj) evals[jj] = cv_U(jj);

      // Transform eigenvalues
      for (auto &val : evals) {
        val = val + 1.0;
      }

      Log::info() << "Eigenvalues: " << util::full_precision(evals)
                  << std::endl;

      // Order eigenvalues
      std::vector<int> indices(evals.size());
      std::size_t n(0);
      std::generate(std::begin(indices), std::end(indices),
                    [&] { return n++; });

      std::sort(std::begin(indices), std::end(indices),
                [&](int ii1, int ii2) { return evals[ii1] > evals[ii2]; });

      Log::info() << "Sorted eigenvalues (" << indices.size()
                  << "):" << std::endl;
      for (const auto &ind : indices)
        Log::info() << "Index: " << ind << ", value: " << evals[ind]
                    << std::endl;

      // Accept or reject eigenvectors
      indices.erase(std::remove_if(indices.begin(), indices.end(),
                                   [&](int n) { return evals[n] <= 1.0; }),
                    indices.end());

      Log::info() << "Sorted accepted eigenvalues (" << indices.size()
                  << "):" << std::endl;
      for (const auto &ind : indices)
        Log::info() << "Index: " << ind << ", value: " << evals[ind]
                    << std::endl;

      if (indices.size() == 0)
        Log::info()
            << "Warning: cannot form preconditioner - no positive eigenvalues; "
            << "minimisation will not be preconditioned" << std::endl;

      // Transform back u_i = Q^T v_i
      U_.clear();
      eigvals_.clear();
      for (auto const &ind : indices) {
        // Limit eigenvalues
        evals[ind] = std::min(10.0, evals[ind]);
        eigvals_.push_back(evals[ind]);
        std::shared_ptr<CtrlVec_> uu(new CtrlVec_(J_.jb()));
        std::vector<double> evec(nvecs);
        for (size_t ji = 0; ji < nvecs; ++ji) evec[ji] = mp_L(ji, ind);
        uu->zero();
        uu->state()[0].modCtlVec().setSection(evec, 1, nvecs);
        for (int ji = nvecs - 1; ji >= 0; --ji) {
          double zz = dot_product(*Q[ji], *uu);
          uu->axpy(-zz, *Q[ji]);
        }
        U_.push_back(uu);
      }
      // Release Q
      Q.clear();

      if (debug_) {
        Log::info() << "SpectralSqrtLMP::QR check orthonormality of U:"
                    << std::endl;
        for (size_t jj = 0; jj < nvecs; ++jj) {
          for (size_t jk = 0; jk < nvecs; ++jk) {
            double zz = dot_product(*U_[jj], *U_[jk]);
            Log::info() << "  <U[" << jj << "," << jk
                        << "]> = " << util::full_precision(zz) << std::endl;
          }
          Log::info() << std::endl;
        }
      }

      Log::info() << "SpectralSqrtLMP::postprocess orthogonalization done"
                  << std::endl;
    } else {
      // Limit eigenvalues
      for (auto &val : eigvals_) val = std::min(10.0, val);
    }
    Log::info() << "SpectralSqrtLMP::postprocess done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::changeResolution() {
  if (active_) {
    //  Interpolate to the new resolution
    Log::info() << "SpectralSqrtLMP::changeResolution starting" << std::endl;
    for (size_t jj = 0; jj < U_.size(); ++jj) {
      U_[jj].reset(new CtrlVec_(J_.jb(), *U_[jj]));
    }
    Log::info() << "SpectralSqrtLMP::changeResolution done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::inverseMultiply(const CtrlVec_ &a,
                                             CtrlVec_ &b) const {
  b = a;  // P_0 = I

  for (size_t ii = 0; ii < U_.size(); ++ii) {
    double zeval = eigvals_[ii];
    ASSERT(zeval);
    double zz = 1.0 / zeval - 1.0;
    b.axpy(zz * dot_product(a, *U_[ii]), *U_[ii]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::inverseMultiplySqrt(const CtrlVec_ &a,
                                                 CtrlVec_ &b) const {
  b = a;  // P_0 = I

  for (size_t ii = 0; ii < U_.size(); ++ii) {
    double zeval = eigvals_[ii];
    ASSERT(zeval);
    double zz = 1.0 / std::sqrt(zeval) - 1.0;
    b.axpy(zz * dot_product(a, *U_[ii]), *U_[ii]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::multiplySqrt(const CtrlVec_ &a,
                                          CtrlVec_ &b) const {
  b = a;  // P_0 = I

  for (size_t ii = 0; ii < U_.size(); ++ii) {
    double zeval = eigvals_[ii];
    ASSERT(zeval);
    double zz = sqrt(zeval) - 1.0;
    b.axpy(zz * dot_product(a, *U_[ii]), *U_[ii]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::checkpoint(eckit::LocalConfiguration &conf) const {
  if (active_) {
    size_t nvecs = U_.size();
    eckit::LocalConfiguration precConf;
    conf.get("preconditioner", precConf);
    precConf.set("nvecs", nvecs);

    // Construct restart configuration and write control vectors
    for (size_t ii = 0; ii < nvecs; ++ii) {
      eckit::LocalConfiguration precConfVec;

      // Jb - component
      if (precConf.has("ctlvec")) {
        std::vector<eckit::LocalConfiguration> localConfs;
        precConf.get("ctlvec", localConfs);
        for (auto &config : localConfs) {
          std::string filename = config.getString("filename");
          config.set("filename", filename + "." + std::to_string(ii));
          config.set("nvecs", nvecs);
        }
        precConfVec.set("ctlvec", localConfs);
      }

      // Jq - component
      if (precConf.has("ctlvecModAux")) {
        eckit::LocalConfiguration localConf;
        precConf.get("ctlvecModAux", localConf);
        std::string filename = localConf.getString("filename");
        localConf.set("filename", filename + "." + std::to_string(ii));
        localConf.set("nvecs", nvecs);
        precConfVec.set("ctlvecModAux", localConf);
      }

      // ObsAux - component
      if (precConf.has("ctlvecsObsAux")) {
        std::vector<eckit::LocalConfiguration> localConfs;
        precConf.get("ctlvecsObsAux", localConfs);
        for (auto &config : localConfs) {
          std::vector<eckit::LocalConfiguration> localObsAuxConfs;
          config.get("ctlvecObsAux", localObsAuxConfs);
          for (auto &configObsAux : localObsAuxConfs) {
            std::string filename = configObsAux.getString("filename");
            configObsAux.set("filename", filename + "." + std::to_string(ii));
            configObsAux.set("nvecs", nvecs);
          }
          config.set("ctlvecObsAux", localObsAuxConfs);
        }
        precConfVec.set("ctlvecsObsAux", localConfs);
      }

      // Write control vector
      U_[ii]->write(precConfVec);
      Log::info() << " " << classname() << "::checkpoint svd[" << ii << "] "
                  << std::endl;
    }
    conf.set("preconditioner", precConf);

    // Additional LMP checkpointing
    eckit::LocalConfiguration auxconf;

    // Convert to string to retain full precision
    std::vector<std::string> eigvalstr;
    for (auto const &value : eigvals_) {
      eigvalstr.push_back(util::full_precision(value));
    }
    auxconf.set("eigvals", eigvalstr);
    Log::info() << " " << classname() << "::checkpoint eigvals "
                << util::full_precision(eigvals_) << std::endl;

    auxconf.set("usedpairIndx", usedpairIndx_);
    Log::info() << " " << classname() << "::checkpoint usedpairIndx "
                << util::full_precision(usedpairIndx_) << std::endl;

    auxconf.set("update", update_);
    Log::info() << " " << classname() << "::checkpoint update "
                << util::full_precision(update_) << std::endl;

    // Set auxiliary components of the preconditioner restart config
    conf.set("preconditioner_aux", auxconf);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void SpectralSqrtLMP<MODEL>::restart(const eckit::Configuration &conf) {
  if (active_) {
    ASSERT(conf.has("preconditioner"));

    // Restart preconditioning vectors

    eckit::LocalConfiguration precConf;
    conf.get("preconditioner", precConf);
    size_t nvecs = precConf.getInt("nvecs");
    const bool lgatherIO = precConf.getBool("gatherIO", true);

    // Construct restart configuration and write control vectors
    for (size_t ii = 0; ii < nvecs; ++ii) {
      eckit::LocalConfiguration precConfVec;

      // Jb - component
      if (precConf.has("ctlvec")) {
        std::vector<eckit::LocalConfiguration> localConfs;
        precConf.get("ctlvec", localConfs);
        for (auto &config : localConfs) {
          std::string filename = config.getString("filename");
          config.set("filename", filename + "." + std::to_string(ii));
          config.set("nvecs", nvecs);
        }
        precConfVec.set("ctlvec", localConfs);
      }

      // Jq - component
      if (precConf.has("ctlvecModAux")) {
        eckit::LocalConfiguration localConf;
        precConf.get("ctlvecModAux", localConf);
        std::string filename = localConf.getString("filename");
        localConf.set("filename", filename + "." + std::to_string(ii));
        localConf.set("nvecs", nvecs);
        precConfVec.set("ctlvecModAux", localConf);
      }

      // ObsAux - component
      if (precConf.has("ctlvecsObsAux")) {
        std::vector<eckit::LocalConfiguration> localConfs;
        precConf.get("ctlvecsObsAux", localConfs);
        for (auto &config : localConfs) {
          std::vector<eckit::LocalConfiguration> localObsAuxConfs;
          config.get("ctlvecObsAux", localObsAuxConfs);
          for (auto &configObsAux : localObsAuxConfs) {
            std::string filename = configObsAux.getString("filename");
            configObsAux.set("filename", filename + "." + std::to_string(ii));
            configObsAux.set("nvecs", nvecs);
          }
          config.set("ctlvecObsAux", localObsAuxConfs);
        }
        precConfVec.set("ctlvecsObsAux", localConfs);
      }

      // Read control vector
      std::shared_ptr<CtrlVec_> u;
      if (lgatherIO) {
        u.reset(new CtrlVec_(J_.jb()));
      } else {
        u.reset(new CtrlVec_(J_.jbRestart()));
      }
      u->read(precConfVec);
      U_.push_back(u);
      Log::info() << " " << classname() << "::restart lmp[" << ii << "] "
                  << std::endl;
    }

    // Restart auxiliary data
    if (conf.has("preconditioner_aux")) {
      eckit::LocalConfiguration auxconf =
          conf.getSubConfiguration("preconditioner_aux");
      eigvals_ = auxconf.getDoubleVector("eigvals");
      usedpairIndx_ = auxconf.getUnsignedVector("usedpairIndx");
      update_ = auxconf.getUnsigned("update");
    } else {
      // Warm start
      eigvals_.clear();
      usedpairIndx_.clear();
      eigvals_.resize(nvecs, 10.0);
      usedpairIndx_.push_back(nvecs);
    }

    Log::info() << " " << classname()
                << "::restart eigvals: " << util::full_precision(eigvals_)
                << std::endl;
    Log::info() << " " << classname() << "::restart usedpairIndx: "
                << util::full_precision(usedpairIndx_) << std::endl;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_SPECTRALSQRTLMP_H_
