/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_UTHTRINVHUMATRIX_H_
#define OOPS_ASSIMILATION_UTHTRINVHUMATRIX_H_

#include <memory>  // for std::unique_ptr

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/CostJo.h"
#include "oops/assimilation/DualVector.h"
#include "oops/base/Departures.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/interface/Increment.h"
#include "util/Logger.h"
#include "util/PrintAdjTest.h"
#include "util/dot_product.h"
#include "util/formats.h"

namespace oops {

/*! @brief The \f$ U^T H^T R^{-1} H U\f$ matrix.
 *
 *   The solvers represent matrices as objects that implement a "multiply"
 *  method. This class defines objects that apply a generalized
 *  \f$ U^T H^T R^{-1} H U\f$ matrix that also includes the equivalent
 *  operators for the other terms of the cost function.
 */

template <typename MODEL>
class UtHtRinvHUMatrix : private eckit::NonCopyable {
  using Increment_ = Increment<MODEL>;
  using CtrlInc_ = ControlIncrement<MODEL>;
  using CtrlVec_ = ControlVector<MODEL>;
  using CostFct_ = CostFunction<MODEL>;
  using CostJo_ = CostJo<MODEL>;
  using Departures_ = Departures<MODEL>;

 public:
  explicit UtHtRinvHUMatrix(const CostFct_ &j,
                            const eckit::Configuration &config);

  void multiply(const CtrlVec_ &dv, CtrlVec_ &dz) const;
  void computeDiagnostics(std::vector<double> &ss,
                          const bool &convergenceReached) const;

 private:
  const eckit::LocalConfiguration config_;
  CostFct_ const &j_;
  mutable CtrlInc_ dx_;
  mutable CtrlInc_ mdx_;
  mutable CtrlInc_ dz_;
  mutable std::unique_ptr<DualVector<MODEL> > ww_;
  mutable std::unique_ptr<DualVector<MODEL> > zz_;
  mutable std::vector<DualVector<MODEL> > wwhudvvec_;
  bool diags_;
  mutable unsigned iter_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
UtHtRinvHUMatrix<MODEL>::UtHtRinvHUMatrix(const CostFct_ &j,
                                          const eckit::Configuration &config)
    : j_(j),
      dx_(j.jb().getFirstGuessIncrement(), false),
      mdx_(j.jb().getFirstGuessIncrement(), false),
      dz_(j.jb().getFirstGuessIncrement(), false),
      ww_(),
      zz_(),
      wwhudvvec_(),
      diags_(config.has("onlineDiagnostics")),
      config_(config),
      iter_(0) {}

// -----------------------------------------------------------------------------

template <typename MODEL>
void UtHtRinvHUMatrix<MODEL>::multiply(const CtrlVec_ &dvin,
                                       CtrlVec_ &dvout) const {
  // Increment counter
  iter_++;

  // Multiply by the square root of the covariance
  j_.jb().multiplyBsqrt(dvin, dx_);

  // Setup TL terms of cost function
  PostProcessorTL<Increment_> costtl;
  for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
    costtl.enrollProcessor(j_.jterm(jj).setupTL(dx_));
  }

  // Run TLM
  mdx_ = dx_;
  j_.runTL(mdx_, costtl);

  // Get TLM outputs, multiply by covariance inverses, and setup ADJ forcing
  // terms
  j_.zeroAD(dz_);
  PostProcessorAD<Increment_> costad;

  ww_.reset(new DualVector<MODEL>());
  zz_.reset(new DualVector<MODEL>());

  for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
    ww_->append(costtl.releaseOutputFromTL(jj));
    if (diags_ && j_.jterm(jj).classname() == "Jo") {
      // Required for diagnostic quadratic JO-table
      //  save H dx[i]= H U dv[i], where
      //       dx = sum_i=1^N dx[i] = U sum_i=1^N dv[i] ss[i]
      DualVector<MODEL> wwhudvii;
      wwhudvii.append(costtl.releaseOutputFromTL(jj));
      wwhudvvec_.push_back(wwhudvii);
    }
    zz_->append(j_.jterm(jj).multiplyCoInv(*ww_->getv(jj)));
    costad.enrollProcessor(j_.jterm(jj).setupAD(zz_->getv(jj), dz_));
  }

  // Run ADJ
  j_.runAD(dz_, costad);

  // Multiply by the transpose of the square root of the covariance
  j_.jb().multiplyBsqrtTrans(dz_, dvout);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void UtHtRinvHUMatrix<MODEL>::computeDiagnostics(
    std::vector<double> &ss, const bool &convergenceReached) const {
  if (diags_ && (iter_ <= 2 || convergenceReached)) {
    Log::info() << "Online diagnostics:" << std::endl
                << "------------------" << std::endl;
    // Quadratic JO-table

    for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
      if (j_.jterm(jj).classname() == "Jo") {
        const CostJo_ &jo = dynamic_cast<const CostJo_ &>(j_.jterm(jj));

        // We would like to compute H dx = H U dv
        // Note that dv = sum_i=1^N dv[ii] ss[ii]
        DualVector<MODEL> wwhudv(wwhudvvec_[0]);
        DualVector<MODEL> wwhudvii(wwhudvvec_[0]);
        DualVector<MODEL> dd;
        DualVector<MODEL> zz;
        wwhudv.zero();
        wwhudvii.zero();
        for (unsigned ii = 0; ii < iter_; ++ii) {
          // compute H U dv[i] ss[i]
          wwhudvii = wwhudvvec_[ii];
          wwhudvii *= ss[ii];
          // compute H dx = H U dv = sum_i=1^N H U dv[ii] ss[ii]
          wwhudv += wwhudvii;
        }
        // Release d = y - H_NL(x)
        dd.append(j_.jterm(jj).newDepartures());
        // Compute dp = y - H_NL(x) - H(dx) = d - H(dx)
        dd -= wwhudv;
        // Compute R^-1 (dp)
        zz.append(j_.jterm(jj).multiplyCoInv(*dd.getv(jj)));
        // print JO table
        jo.printJo(dynamic_cast<const Departures_ &>(*dd.getv(jj)),
                   dynamic_cast<const Departures_ &>(*zz.getv(jj)),
                   CostJo_::OS::Quadratic);
        if (config_.has("onlineDiagnostics.residue")) {
          dynamic_cast<const Departures_ &>(*dd.getv(jj))
              .save(eckit::LocalConfiguration(config_,
                                              "onlineDiagnostics.residue"));
        }
        Log::info() << "----------" << std::endl;
      }
    }

    // Online adjoint test

    // <G dx, dy>, where dy = Rinv H dx
    double adj_tst_fwd = dot_product(*ww_, *zz_);
    // <dx, Gt dy> , where dy = Rinv H dx
    double adj_tst_bwd = dot_product(dx_, dz_);

    Log::info() << "Online adjoint test, iteration: " << iter_ << std::endl
                << util::PrintAdjTest(adj_tst_fwd, adj_tst_bwd, "G")
                << "----------" << std::endl;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_UTHTRINVHUMATRIX_H_
