/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef ADVECTOR_H__
#define ADVECTOR_H__

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Accumulator.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Interpolator.h"
#include "oops/interface/State.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/// Advection operator

// -----------------------------------------------------------------------------
template <typename MODEL>
class Advector : public Interpolator<MODEL> {
  using State4D_ = std::vector<std::unique_ptr<State<MODEL> > >;
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Increment_ = Increment<MODEL>;
  using State_ = State<MODEL>;
  using Interpolator_ = Interpolator<MODEL>;

 public:
  Advector(const State_ &, const eckit::Configuration &, const bool is_adjoint);
  Advector &operator=(const Advector &);

  void set_vector(const eckit::Configuration &);
  void set_wind(const State4D_ &, const util::DateTime &);
  void set_wind_inv(const State4D_ &, const util::DateTime &);
  void set_wind(const State4D_ &);
  void set_wind_inv(const State4D_ &);

  void advect(const Increment_ &, const eckit::Configuration &, Increment_ &);
  void advect(const State_ &, const eckit::Configuration &, State_ &);

 private:
  const eckit::LocalConfiguration config_;
  const eckit::LocalConfiguration vars_;
  const util::DateTime time_;
  unsigned niter_;
  double wind_coeff_;
};
// =============================================================================

// -----------------------------------------------------------------------------
template <typename MODEL>
Advector<MODEL>::Advector(const State_ &xb, const eckit::Configuration &config,
                          const bool is_adjoint)
    : Interpolator_(xb, config, is_adjoint),
      config_(config),
      vars_(config, "wind"),
      time_(xb.validTime()) {
  niter_ = config.getInt("nb_iter_point_fixe");

  // Change default value for wind coefficient
  wind_coeff_ = 0.75;
  if (config.has("wind_coeff")) {
    wind_coeff_ = config.getDouble("wind_coeff");
  }
  Log::trace() << "Advector:Advector created";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
Advector<MODEL> &Advector<MODEL>::operator=(const Advector<MODEL> &other) {
  Interpolator_ *this_interp = this;
  (*this_interp) = other;
  return *this;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::set_vector(const eckit::Configuration &cfg) {
  Interpolator_::set_vector(cfg);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::advect(const Increment_ &dxin,
                             const eckit::Configuration &cfg,
                             Increment_ &dxout) {
  const Variables_ var(cfg);
  this->interpolate(dxin, var, dxout);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::advect(const State_ &xin, const eckit::Configuration &cfg,
                             State_ &xout) {
  const Variables_ var(cfg);
  this->interpolate(xin, var, xout);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::set_wind(const State4D_ &wind,
                               const util::DateTime &date_start) {
  util::Duration step(wind[1]->validTime() - wind[0]->validTime());
  double deltat = (double)step.toSeconds();
  unsigned kstart = 0;
  unsigned kstop = 0;

  // Initialize with vector (0,0)
  set_vector(config_);

  for (unsigned k = 0; k < wind.size(); k++) {
    if (wind[k]->validTime() == time_) kstop = k;
    if (wind[k]->validTime() == date_start) kstart = k;
  }

  // Advect the grid step by step
  Advector<MODEL> s_star(*this);
  for (unsigned k = kstop; k > kstart; k -= 1) {
    for (unsigned iter = 0; iter < niter_; iter++) {
      Accumulator<MODEL, State_, State_> uv(*wind[k - 1]);
      uv.accumul(wind_coeff_, *wind[k - 1]);
      s_star.advect(uv, vars_, uv);

      Accumulator<MODEL, State_, State_> delta_xy(uv);
      delta_xy.accumul(deltat, uv);

      s_star = (*this);
      s_star -= delta_xy;
    }
    (*this) = s_star;
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::set_wind_inv(const State4D_ &wind,
                                   const util::DateTime &date_start) {
  util::Duration step(wind[1]->validTime() - wind[0]->validTime());
  double deltat = (double)step.toSeconds();
  unsigned kstart = 0;
  unsigned kstop = 0;

  set_vector(config_);

  for (unsigned k = 0; k < wind.size(); k++) {
    if (wind[k]->validTime() == time_) kstop = k;
    if (wind[k]->validTime() == date_start) kstart = k;
  }

  Advector<MODEL> s_star(*this);
  for (unsigned k = 1 + kstart; k <= kstop; ++k) {
    for (unsigned iter = 0; iter < niter_; iter++) {
      Accumulator<MODEL, State_, State_> uv(*wind[k]);
      uv.accumul(wind_coeff_, *wind[k]);
      s_star.advect(uv, vars_, uv);

      Accumulator<MODEL, State_, State_> delta_xy(uv);
      delta_xy.accumul(deltat, uv);

      s_star = (*this);
      s_star += delta_xy;
    }
    (*this) = s_star;
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::set_wind(const State4D_ &wind) {
  set_wind(wind, wind[0]->validTime());
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Advector<MODEL>::set_wind_inv(const State4D_ &wind) {
  set_wind_inv(wind, wind[0]->validTime());
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // oops
