/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_ADVECTORHANDLER_H_
#define OOPS_BASE_ADVECTORHANDLER_H_

#include <map>
#include <memory>
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Advector.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/base/Increment4D.h"
#include "oops/base/PostBase.h"
#include "oops/base/State4D.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Variables.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace util {
class DateTime;
}

namespace oops {

/// Advector handler class
// This handler is a singleton created by a ModelSpaceCovariance4D and used in
// the linearization of the cost_function. This class provides a post_processor
// that registers some States during a forecast
template <typename MODEL>
class AdvectorHandler : public PostBase<State<MODEL> > {
  using State_ = State<MODEL>;
  using State4D_ = std::vector<std::unique_ptr<State<MODEL> > >;
  using Increment4D_ = Increment4D<MODEL>;
  using Advector_ = Advector<MODEL>;
  using Ensemble_ = Ensemble<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL> >;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;

 public:
  AdvectorHandler(const eckit::Configuration &config);

  static void setup(std::shared_ptr<AdvectorHandler> other) {
    singleton_ = other;
  }
  static std::shared_ptr<AdvectorHandler> get() { return singleton_; }

  void mult(const Increment4D_ &dxi, Increment4D_ &dxo) const;
  void mult_ad(const Increment4D_ &dxi, Increment4D_ &dxo) const;

 private:
  const eckit::LocalConfiguration config_;
  std::vector<util::DateTime> time_;
  std::vector<std::unique_ptr<Advector_> > advector_;
  std::vector<std::unique_ptr<Advector_> > advector_ad_;
  std::vector<std::unique_ptr<Advector_> > advector_inv_;
  std::unique_ptr<eckit::LocalConfiguration> param_adv_;
  State4D_ xx_;

  static std::shared_ptr<AdvectorHandler<MODEL> > singleton_;

  void doInitialize(const State_ &, const util::DateTime &,
                    const util::Duration &) override;
  void doProcessing(const State_ &) override;
  void doFinalize(const State_ &) override;

  void inv_mult_ensemble() const;
};

// ====================================================================================

template <typename MODEL>
std::shared_ptr<AdvectorHandler<MODEL> > AdvectorHandler<MODEL>::singleton_;

template <typename MODEL>
AdvectorHandler<MODEL>::AdvectorHandler(const eckit::Configuration &config)
    : PostBase<State_>(config), config_(config) {
  param_adv_.reset(
      new eckit::LocalConfiguration(config_, "fields_to_be_interpolated"));
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void AdvectorHandler<MODEL>::mult(const Increment4D_ &dxi,
                                  Increment4D_ &dxo) const {
  for (unsigned k = 0; k < advector_.size(); ++k) {
    advector_[k]->advect(dxi[k + 1], *param_adv_, dxo[k + 1]);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void AdvectorHandler<MODEL>::mult_ad(const Increment4D_ &dxi,
                                     Increment4D_ &dxo) const {
  for (unsigned k = 0; k < advector_ad_.size(); ++k) {
    advector_ad_[k]->advect(dxi[k + 1], *param_adv_, dxo[k + 1]);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void AdvectorHandler<MODEL>::inv_mult_ensemble() const {
  for (unsigned k = 0; k < advector_inv_.size(); ++k) {
    if (EnsemblesCollection_::getInstance().has(time_[k])) {
      Log::info() << "AdvectorHandler inv_mult_ensemble() called for time = "
                  << time_[k] << std::endl;
      EnsemblePtr_ ens = EnsemblesCollection_::getInstance()[time_[k]];
      for (unsigned l = 0; l < ens->size(); ++l) {
        advector_inv_[k]->advect((*ens)[l], *param_adv_, (*ens)[l]);
      }
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void AdvectorHandler<MODEL>::doInitialize(const State_ &,
                                          const util::DateTime &,
                                          const util::Duration &) {
  advector_.clear();
  advector_ad_.clear();
  advector_inv_.clear();
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void AdvectorHandler<MODEL>::doProcessing(const State_ &xx) {
  Log::info() << "AdvectorHandler: State registered at " << xx.validTime()
              << std::endl;
  xx_.emplace_back(new State_(xx));
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void AdvectorHandler<MODEL>::doFinalize(const State_ &) {
  for (unsigned k = 1; k < xx_.size(); ++k) {
    time_.push_back(xx_[k]->validTime());
    advector_.emplace_back(new Advector_(*(xx_[k]), config_, false));
    advector_inv_.emplace_back(new Advector_(*(xx_[k]), config_, false));
    advector_ad_.emplace_back(new Advector_(*(xx_[k]), config_, true));

    advector_[k - 1]->set_wind(xx_);
    advector_inv_[k - 1]->set_wind_inv(xx_);
    *advector_ad_[k - 1] = *advector_[k - 1];
  }
  xx_.clear();
  inv_mult_ensemble();
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_ADVECTORHANDLER_H_
