/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_DFIPROCESSOR_H_
#define OOPS_BASE_DFIPROCESSOR_H_

#include <cmath>
#include <map>
#include <memory>  // for std::unique_ptr

#include "eckit/exception/Exceptions.h"
#include "oops/base/Accumulator.h"
#include "oops/base/DolphChebyshev.h"
#include "oops/base/PostBase.h"
#include "oops/base/WeightingFct.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Compute time average of states or increments during model run.
/*!
 *  Derived classes will compute different types of averages (plain
 *  mean, various types of digital filters) by overwriting the weights
 *  computation method.
 *  This class comes from WeightedMean and allows the coefficients to be
 * different depending on the variable: In ARPEGE DFI, we need coeff=1 at the
 * middle of the assimilation window and 0 elsewhere for GFLs.
 */

template <typename MODEL>
class DFIProcessor : public PostBase<State<MODEL> > {
  typedef Geometry<MODEL> Geometry_;
  typedef Variables<MODEL> Variables_;
  typedef State<MODEL> State_;
  typedef GenericMatrix<MODEL> GenericMatrix_;
  typedef GenericMatrixFactory<MODEL> MatrixFactory_;

 public:
  DFIProcessor(const util::DateTime &, const util::Duration &,
               const Geometry_ &, const eckit::Configuration &);
  virtual ~DFIProcessor() {}

  State_ *releaseMean();

 private:
  void doInitialize(const State_ &, const util::DateTime &,
                    const util::Duration &) override;
  void doProcessing(const State_ &) override;

  void accumulate(double, double, const State_ &);

  std::unique_ptr<WeightingFct> wfct_;
  std::map<util::DateTime, double> weights_;
  Accumulator<MODEL, State_, State_> *avg_;
  double sum_;
  bool linit_;
  const util::DateTime time_;
  const util::DateTime bgn_;
  const util::DateTime end_;
  util::DateTime endleg_;
};

// =============================================================================

template <typename MODEL>
DFIProcessor<MODEL>::DFIProcessor(const util::DateTime &vt,
                                  const util::Duration &span,
                                  const Geometry_ &resol,
                                  const eckit::Configuration &config)
    : PostBase<State_>(vt - span / 2, vt + span / 2),
      wfct_(),
      weights_(),
      avg_(0),
      sum_(0.0),
      linit_(false),
      time_(vt),
      bgn_(vt - span / 2),
      end_(vt + span / 2),
      endleg_() {
  const Variables_ vars(config);
  // Accumulator construction delegated to the doInitialize method
  wfct_.reset(new DolphChebyshev(config));
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL> *DFIProcessor<MODEL>::releaseMean() {
  ASSERT(linit_);
  ASSERT(std::abs(sum_ - 1.0) < 1.0e-8);
  return avg_;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DFIProcessor<MODEL>::doInitialize(const State_ &xx,
                                       const util::DateTime &end,
                                       const util::Duration &tstep) {
  const util::DateTime bgn(xx.validTime());
  if (!linit_ && bgn <= end_ && end >= bgn_) {
    weights_ = wfct_->setWeights(bgn_, end_, tstep);
    linit_ = true;
  }
  endleg_ = end;
  avg_ = new Accumulator<MODEL, State_, State_>(xx);
  avg_->validTime() = time_;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void DFIProcessor<MODEL>::doProcessing(const State_ &xx) {
  const util::DateTime now(xx.validTime());
  if (now != endleg_ || now == end_) {
    ASSERT(weights_.find(now) != weights_.end());

    const double zz = weights_[now];
    const double zgfl = (now == time_) ? 1.0 : 0.0;
    accumulate(zz, zgfl, xx);

    sum_ += zz;
    Log::debug() << "DFIProcessor: time = " << now << ", weight = " << zz
                 << ", sum = " << sum_ << std::endl;
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void DFIProcessor<MODEL>::accumulate(double zz, double zgfl,
                                     const State_ &rhs) {
  eckit::LocalConfiguration cfg;
  cfg.set("matrix", "variable_dependent_coeff");
  cfg.set("analysed_variables_coefficient", zz);
  cfg.set("other_variables_coefficient", zgfl);
  Log::debug() << "DFI coefficient: " << cfg << std::endl;
  std::unique_ptr<GenericMatrix_> coeff(MatrixFactory_::create(cfg));
  coeff->accumul(*avg_, rhs);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_DFIPROCESSOR_H_
