/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_DEPARTURES_H_
#define OOPS_BASE_DEPARTURES_H_

#include <cstddef>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include "oops/assimilation/ControlObsVector.h"
#include "oops/base/GeneralizedDepartures.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/ObsErrorBase.h"
#include "oops/interface/ObsVector.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/Printable.h"
#include "util/dot_product.h"

namespace oops {

/*! @brief Difference between two observation vectors.

 *  A departure is the difference between two observations.
 * The archetypal example is

   \f[ d = y - {\cal G}(x) =
          \begin{bmatrix}
            \bar{y}_1  \\
            ..         \\
            \bar{y}_L
          \end{bmatrix}
          -
          \begin{bmatrix}
            \bar{\cal G}_1 && .. && ..             \\
            ..             && .. && ..             \\
            ..             && .. && \bar{\cal G}_L
          \end{bmatrix}
          \begin{bmatrix}
            \bar{x}_1 \\
            ..        \\
            \bar{x}_L
          \end{bmatrix}
   \f]

   where

   \f[ \bar{y}_j = \begin{bmatrix}
                     \tilde{y}_j  \\
                     \tilde{b}_j  \\
                   \end{bmatrix}
   \f]

   and
   \f[ \bar{\cal G}_j = \begin{bmatrix}
                          \tilde{\cal G}_j  && \tilde{\cal P}_j \\
                          0                 && \tilde{\cal P}_j
                        \end{bmatrix}
   \f]

   \f[ \bar{x}_j = \begin{bmatrix}
                      x \\
                      \beta_j
                   \end{bmatrix}
   \f]

 *   and \f$ L \f$ is the number of ObservationSpace objects held by
 *  ObservationSpaces, \f$  \bar{y}_j \f$ is implemented by ControlObsVector
 *  and has two components: observation \f$ \tilde{y}_j \f$ and bias
 *  \f$ \tilde{b}_j \f$ vectors (each implemented by ObsVector).

 *  Departures hold a vector of ControlObsVector objects.

 *   Keeping an observation space vector here is necessary for the
 *  implementation of generic observation error covariance matrices.

 *   Refer to the documentation of CostJo to better understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class Departures : public util::Printable, public GeneralizedDepartures {
  using ControlObsVector_ = ControlObsVector<MODEL>;
  using ObsErrorBase_ = ObsErrorBase<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  /// Constructor
  Departures() : dep_(){};
  explicit Departures(const ObsSpaces_ &);
  /// Constructor
  explicit Departures(std::vector<std::shared_ptr<ControlObsVector_> >);
  /// Copy constructor
  explicit Departures(const Departures &, const bool copy = true);
  /// Destructor
  ~Departures();

  /// Return the number of components
  std::size_t size() const { return dep_.size(); }
  /// Return a constant reference to i-th component
  const ControlObsVector_ &operator[](const std::size_t ii) const {
    return *dep_.at(ii);
  }
  /// Return a reference to i-th component
  ControlObsVector_ &operator[](const std::size_t ii) { return *dep_.at(ii); }

  // Linear algebra operators
  /// \f$ d1 = d2 \f$
  Departures &operator=(const Departures &);
  /// \f$ d1 = d1 + d2 \f$
  Departures &operator+=(const Departures &);
  /// \f$ d1 = d1 - d2 \f$
  Departures &operator-=(const Departures &);
  /// \f$ d1 = c*d1 \f$
  Departures &operator*=(const double &);
  /// \f$ d1 = d1*d2 \f$
  Departures &operator*=(const Departures &);
  /// \f$ d1 = d1/d2 \f$
  Departures &operator/=(const Departures &);
  /// \f$ d1 = 0 \f$
  void zero();
  /// \f$ d1 = 1/d1 \f$
  void invert();
  /// \f$ d1 = d1 + c*d2 \f$
  void axpy(const double &, const Departures &);
  /// \f$ <d1, d1> = d1^T d1\f$
  double dot_product_with(const Departures &) const;

  /// Get departue values
  const ControlObsVector_ &depvalues() const { return *dep_; }

  // I/O
  /// Write departures
  void save(const eckit::Configuration &) const;
  /// Read departures
  void read(const eckit::Configuration &);

 private:
  void print(std::ostream &) const;

  /// Vector of shared pointers to ControlObsVector objects
  std::vector<std::shared_ptr<ControlObsVector_> > dep_;
};

// =============================================================================

template <typename MODEL>
Departures<MODEL>::Departures(const ObsSpaces_ &obsgeom) : dep_(0) {
  for (std::size_t jj = 0; jj < obsgeom.size(); ++jj) {
    std::shared_ptr<ControlObsVector_> tmp(new ControlObsVector_(obsgeom[jj]));
    dep_.push_back(tmp);
  }
  Log::trace() << "Departures created" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL>::Departures(
    std::vector<std::shared_ptr<ControlObsVector_> > val)
    : dep_(val) {
  Log::trace() << "Departures created" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL>::Departures(const Departures &other, const bool copy)
    : dep_(0) {
  // We want deep copy here
  for (std::size_t jj = 0; jj < other.dep_.size(); ++jj) {
    std::shared_ptr<ControlObsVector_> tmp(
        new ControlObsVector_(*other.dep_[jj], copy));
    dep_.push_back(tmp);
  }
  Log::trace() << "Departures copy-created" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL>::~Departures() {
  Log::trace() << "Departures destructed" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL> &Departures<MODEL>::operator=(const Departures &rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] = *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL> &Departures<MODEL>::operator+=(const Departures &rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] += *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL> &Departures<MODEL>::operator-=(const Departures &rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] -= *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL> &Departures<MODEL>::operator*=(const double &zz) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] *= zz;
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL> &Departures<MODEL>::operator*=(const Departures &rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] *= *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Departures<MODEL> &Departures<MODEL>::operator/=(const Departures &rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] /= *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::zero() {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->zero();
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::invert() {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->invert();
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::axpy(const double &zz, const Departures &rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->axpy(zz, *rhs.dep_[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double Departures<MODEL>::dot_product_with(const Departures &other) const {
  double zz = 0.0;
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    zz += dot_product(*dep_[jj], *other.dep_[jj]);
  }
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::save(const eckit::Configuration &conf) const {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->save(conf);
  }
  Log::trace() << "Departures:Departures have been saved: " << conf
               << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::read(const eckit::Configuration &conf) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->read(conf);
  }
  Log::trace() << "Departures:Departures have been read: " << conf << std::endl;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
void Departures<MODEL>::print(std::ostream &os) const {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    os << *dep_[jj];
  }
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_DEPARTURES_H_
