/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#include "oops/base/DolphChebyshev.h"

#include <cmath>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

// -----------------------------------------------------------------------------

static WeightMaker<DolphChebyshev> makerDolph_("DolphChebyshev");

// -----------------------------------------------------------------------------

DolphChebyshev::DolphChebyshev(const eckit::Configuration& config)
    : window_(false), tau_(util::Duration(config.getString("cutoff"))) {
  if (config.has("window")) window_ = config.getBool("window");
  Log::info() << "DolphChebyshev window_ = " << window_ << ", tau_ = " << tau_
              << std::endl;
}

// -----------------------------------------------------------------------------

std::map<util::DateTime, double> DolphChebyshev::setWeights(
    const util::DateTime& bgn, const util::DateTime& end,
    const util::Duration& dt) {
  constexpr double pi = 3.14159265358979323846;
  const util::Duration window(end - bgn);
  const int nstep = window.toSeconds() / dt.toSeconds();
  ASSERT(window.toSeconds() == dt.toSeconds() * nstep);

  const int M = nstep / 2;
  const int N = 2 * M + 1;
  Log::info() << "DolphChebyshev M = " << M << ", N = " << N << std::endl;

  const int tt = tau_.toSeconds() / dt.toSeconds();
  ASSERT(tau_.toSeconds() == dt.toSeconds() * tt);
  ASSERT(tt > 1);
  double thetas = 0.;
  if (window_) {
    thetas = 2.0 * pi / M;
  } else {
    thetas = 2.0 * pi / tt;
  }
  const double x0 = 1.0 / std::cos(thetas / 2.0);
  Log::info() << "DolphChebyshev tt = " << tt << " thetas = " << thetas
              << ", x0 = " << x0 << std::endl;

  std::vector<double> w(M + 1);
  double rr = 0.0;
  for (int n = 0; n <= M; ++n) {
    double tn = 2.0 * pi * n / N;
    double sum = 0.0;
    for (int m = 0; m <= M; ++m) {
      double xx = x0 * std::cos(pi * m / N);
      double t2m = 0.0;
      double tnm2 = 1.0;
      double tnm1 = xx;
      for (int kk = 2; kk <= 2 * M; ++kk) {
        t2m = 2.0 * xx * tnm1 - tnm2;
        tnm2 = tnm1;
        tnm1 = t2m;
      }
      if (n == 0 && m == 0) {
        rr = 1. / t2m;
        Log::info() << "DolphChebyshev 2.rr = " << 2.0 * rr << std::endl;
      } else if (m > 0) {
        sum += t2m * std::cos(m * tn);
      }
    }
    w[n] = (1.0 + 2.0 * rr * sum) / N;
  }
  // Check the sum
  double checksum = 0.0;
  for (int jj = -M; jj <= M; ++jj) {
    int n = std::abs(jj);
    checksum += w[n];
  }
  Log::debug() << "DolphChebyshev checksum = " << checksum << std::endl;
  ASSERT(std::abs(checksum - 1.0) < 1.0e-8);
  // Combining with simple filter
  double sum = 0.0;
  std::vector<double> ws(2 * M + 1);
  if (window_) {
    for (int jj = -M; jj <= M; ++jj) {
      int n = std::abs(jj);
      if (jj == 0) {
        ws[jj + M] = w[n] * thetas / pi;
      } else {
        ws[jj + M] = w[n] * std::sin(jj * thetas) / (jj * pi);
      }
      sum += ws[jj + M];
    }
  }
  std::map<util::DateTime, double> weights;
  util::DateTime now(bgn);
  for (int jj = -M; jj <= M; ++jj) {
    int n = std::abs(jj);
    Log::debug() << "DolphChebyshev now = " << now << ", n = " << n
                 << ", w = " << w[n] << std::endl;
    if (window_) {
      weights[now] = ws[jj + M] / sum;
    } else {
      weights[now] = w[n];
    }
    now += dt;
  }
  ASSERT(now == end + dt);

  return weights;
}

// -----------------------------------------------------------------------------

}  // namespace oops
