/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_ENSEMBLE_H_
#define OOPS_BASE_ENSEMBLE_H_

#include <cmath>
#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/io/Buffer.h"
#include "eckit/maths/Functions.h"
#include "eckit/serialisation/ResizableMemoryStream.h"
#include "oops/base/Accumulator.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/MPIWrapper.h"
#include "util/dot_product.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Ensemble

template <typename MODEL>
class Ensemble {
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  /// Constructor
  Ensemble(const util::DateTime &, const eckit::Configuration &,
           bool distributed = false);

  Ensemble(const Ensemble &);

  /// Destructor
  virtual ~Ensemble() {}

  /// Accessors
  unsigned int size() const { return rank_; }
  Increment_ &operator[](const int ii) { return *ensemblePerturbs_[ii]; }
  const Increment_ &operator[](const int ii) const {
    return *ensemblePerturbs_[ii];
  }

  void linearize(const State_ &, const Geometry_ &);
  void build(const State_ &, const Geometry_ &);
  void read();
  void to_perturbations();
  void to_perturbations(bool distributed, unsigned id = 0);
  void to_perturbations(int nb_sub_ens);
  Increment_ variance() const;

  const Variables_ &controlVariables() const { return vars_; }

  eckit::Buffer &pack() {
    ASSERT(is_distributed_);
    return *arch_;
  }

 private:
  const eckit::LocalConfiguration config_;
  const util::DateTime validTime_;
  const Variables_ vars_;
  unsigned int rank_;
  double infl_;
  std::vector<std::unique_ptr<Increment_> > ensemblePerturbs_;

  bool is_distributed_;
  bool is_pert_;
  std::unique_ptr<eckit::Buffer> arch_;
};

// ====================================================================================

template <typename MODEL>
Ensemble<MODEL>::Ensemble(const util::DateTime &validTime,
                          const eckit::Configuration &conf, bool distributed)
    : config_(conf),
      validTime_(validTime),
      vars_(conf),
      rank_(conf.getInt("members")),
      infl_(conf.getDouble("inflation", 1.0)),
      ensemblePerturbs_(),
      is_distributed_(distributed),
      is_pert_(false) {
  Log::trace() << "Ensemble:contructor done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Ensemble<MODEL>::Ensemble(const Ensemble &e)
    : config_(e.config_),
      validTime_(e.validTime_),
      vars_(e.vars_),
      rank_(e.rank_),
      infl_(e.infl_),
      ensemblePerturbs_(),
      is_distributed_(e.is_distributed_),
      is_pert_(e.is_pert_) {
  for (unsigned int m = 0; m < rank_; ++m) {
    Increment_ *dx = new Increment_(*(e.ensemblePerturbs_[m]));
    ensemblePerturbs_.emplace_back(dx);
  }
  Log::trace() << "Ensemble:copy contructor done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Ensemble<MODEL>::linearize(const State_ &xb, const Geometry_ &resol) {
  ASSERT(xb.validTime() == validTime_);
  build(xb, resol);
  read();
  to_perturbations();
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Ensemble<MODEL>::build(const State_ &xb, const Geometry_ &resol) {
  /* Build objects for storing Ensemble datas */
  unsigned int size = rank_;
  if (is_distributed_) size = 2;

  for (unsigned int m = 0; m < size; ++m) {
    Increment_ *dx = new Increment_(resol, vars_, validTime_);
    dx->zero();
    if (is_distributed_ && (m == 0)) {
      dx->accumul(1.0, xb);
    }
    ensemblePerturbs_.emplace_back(dx);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Ensemble<MODEL>::read() {
  if (is_distributed_) return;

  std::vector<eckit::LocalConfiguration> confs;
  config_.get("state", confs);
  ASSERT(confs.size() == rank_);

  for (unsigned int jm = 0; jm < rank_; ++jm) {
    ensemblePerturbs_[jm]->read(confs[jm]);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void Ensemble<MODEL>::to_perturbations() {
  const double r = 1.0 / std::sqrt(static_cast<double>(rank_ - 1)) * infl_;
  const double invrank = 1.0 / static_cast<double>(rank_);

  /* Update datas */
  Increment_ sum(*ensemblePerturbs_[0]);
  sum *= invrank;

  for (unsigned m = 1; m < rank_; ++m) {
    sum.axpy(invrank, *ensemblePerturbs_[m]);
  }

  Log::info() << "Ensemble.to_perturbations with inflation :" << infl_
              << std::endl;
  for (unsigned m = 0; m < rank_; ++m) {
    *ensemblePerturbs_[m] -= sum;
    *ensemblePerturbs_[m] *= r;
  }

  is_pert_ = true;
}
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
template <typename MODEL>
void Ensemble<MODEL>::to_perturbations(bool distributed, unsigned id) {
  const double r = 1.0 / std::sqrt(static_cast<double>(rank_ - 1)) * infl_;
  const double invrank = 1.0 / static_cast<double>(rank_);

  /* Update datas */
  Increment_ sum(*ensemblePerturbs_[0]);
  sum *= invrank;

  ASSERT(distributed == is_distributed_);
  util::MPIWrapper &mpiw = util::MPIWrapper::Instance();
  Increment_ tmp(*ensemblePerturbs_[0]);
  tmp.zero();

  if (mpiw.my_rank() == 0) {
    int tag = 666 + 1;
    eckit::mpi::Status st = mpiw.comm().probe(int(1), tag);
    size_t size = eckit::round(mpiw.comm().getCount<char>(st),
                               8);  //< round to nearest 8 bytes
    eckit::Buffer b(size);

    // Get members
    for (unsigned m = 1; m < rank_; ++m) {
      tag = 666 + m;
      b.zero();
      mpiw.comm().receive(static_cast<char *>(b.data()), b.size(), m, tag);
      eckit::ResizableMemoryStream sr(b);
      sr >> tmp;
      sum.axpy(invrank, tmp);
    }
    // Send the sum
    b.zero();
    eckit::ResizableMemoryStream ss(b);
    ss << sum;
    mpiw.comm().broadcast(static_cast<char *>(b.data()), ss.position(),
                          (size_t)0);
  } else {
    // Send my member
    int tag = 666 + mpiw.my_rank();
    eckit::Buffer b(32);
    eckit::ResizableMemoryStream ss(b);
    ss << (*ensemblePerturbs_[0]);
    eckit::mpi::Request req = mpiw.comm().iSend(
        static_cast<const char *>(b.data()), ss.position(), 0, tag);

    // Get the sum
    b.zero();
    mpiw.comm().broadcast(static_cast<char *>(b.data()), b.size(), (size_t)0);
    eckit::ResizableMemoryStream sr(b);
    sr >> sum;
    mpiw.comm().wait(req);
  }

  Log::info() << "Ensemble.to_perturbations with inflation :" << infl_
              << std::endl;
  int size = 1;
  for (int m = 0; m < size; ++m) {
    (*ensemblePerturbs_[m]) -= sum;
    (*ensemblePerturbs_[m]) *= r;
  }

  // Prepare package for MPI transfers
  arch_.reset(new eckit::Buffer(32));
  eckit::ResizableMemoryStream s(*arch_);
  s << (*ensemblePerturbs_[0]);

  is_pert_ = true;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Ensemble<MODEL>::to_perturbations(int nb_sub_ensemble) {
  const int sub_rank = rank_ / nb_sub_ensemble;
  const double inv_sub_rank = 1.0 / static_cast<double>(sub_rank);
  const double r = 1.0 / std::sqrt(static_cast<double>(rank_ - 1)) * infl_;

  Log::info() << "Ensemble.to_perturbations with inflation :" << infl_
              << std::endl;

  /* Update datas */
  for (int p = 0; p < nb_sub_ensemble; ++p) {
    Log::info() << "Ensemble.to_perturbations, remove ensemble mean for "
                   "sub_ensemble # :"
                << p + 1 << std::endl;

    Increment_ sum(*ensemblePerturbs_[p * sub_rank]);
    sum *= inv_sub_rank;

    for (int m = 1; m < sub_rank; ++m) {
      sum.axpy(inv_sub_rank, *ensemblePerturbs_[p * sub_rank + m]);
    }

    for (int m = 0; m < sub_rank; ++m) {
      *ensemblePerturbs_[p * sub_rank + m] -= sum;
      *ensemblePerturbs_[p * sub_rank + m] *= r;
    }
  }

  is_pert_ = true;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment<MODEL> Ensemble<MODEL>::variance() const {
  const double rInv = std::sqrt(static_cast<double>(rank_ - 1));
  const double rsq = 1.0 / static_cast<double>(rank_ - 1);

  Increment_ mean(*ensemblePerturbs_[0]);
  Increment_ variance(*ensemblePerturbs_[0]);
  mean.zero();
  variance.zero();
  for (size_t jm = 0; jm < rank_; ++jm) {
    // Get non-normalized perturbation
    Increment_ pert(*ensemblePerturbs_[jm]);
    if (is_pert_) {
      pert *= rInv;
    }
    pert -= mean;

    // Update mean
    const double meanFactor = 1.0 / static_cast<double>(jm + 1);
    mean.axpy(meanFactor, pert);

    if (jm > 0) {
      // Square perturbation
      pert.schur_product_with(pert);

      // Update variance
      const double varianceFactor =
          static_cast<double>(jm) / static_cast<double>(jm + 1);
      variance.axpy(varianceFactor, pert);
    }
  }

  // Normalize variance
  variance *= rsq;

  return variance;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLE_H_
