/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_ENSEMBLECOVARIANCE_H_
#define OOPS_BASE_ENSEMBLECOVARIANCE_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "oops/assimilation/GMRESR.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/IdentityMatrix.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Localization.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename MODEL>
class IncrModCtlVec;

/*! @brief Generic ensemble based model space error covariance
 *
 *  This class implements a generic localized sampled model space error
 * covariance matrix.
 */
// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovariance : public ModelSpaceCovarianceBase<MODEL> {
  using Ensemble_ = Ensemble<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL>>;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Localization_ = Localization<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;
  using MatrixFactory_ = GenericMatrixFactory<MODEL>;

 public:
  // Constructors, destructors
  EnsembleCovariance(const Geometry_ &, const Variables_ &,
                     const eckit::Configuration &, const State_ &);
  ~EnsembleCovariance();

  // Methods
  void linearize(const State_ &, const Geometry_ &,
                 const eckit::Configuration &) override;
  void multiply(const Increment_ &, Increment_ &) const override;
  void inverseMultiply(const Increment_ &, Increment_ &) const override;
  void multiplySqrt(const IncrCtlVec_ &, Increment_ &) const override;
  void multiplySqrtTrans(const Increment_ &, IncrCtlVec_ &) const override;
  void randomize(Increment_ &) const override;

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*geom_, *loc_);
  }

 private:
  const eckit::LocalConfiguration config_;  ///< Configuration
  const Variables_ vars_;                   ///< Variables
  const util::DateTime time_;               ///< Valid time
  bool has_change_of_variable_;             ///< Change of variable
  std::unique_ptr<Localization_> loc_;      ///< Localization matrix
  std::unique_ptr<Geometry_> geom_;         ///< Pointer to a geometry object
  std::unique_ptr<GenericMatrix_> change_of_variable_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance<MODEL>::EnsembleCovariance(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State_ &)
    : config_(config),
      vars_(vars),
      time_(config_.getString("date")),
      has_change_of_variable_(false),
      loc_(),
      geom_(),
      change_of_variable_() {
  geom_.reset(new Geometry_(resol));
  has_change_of_variable_ = config.has("change_of_variable");
  Log::trace() << "EnsembleCovariance created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance<MODEL>::~EnsembleCovariance() {
  loc_.reset();
  geom_.reset();
  Log::trace() << "EnsembleCovariance destructed." << std::endl;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance<MODEL>::linearize(const State_ &xb,
                                          const Geometry_ &resol,
                                          const eckit::Configuration &config) {
  const eckit::LocalConfiguration ensConf(config, "ensemble_covariance");
  geom_.reset(new Geometry_(resol));

  // Compute the ensemble of perturbations at time of xb.
  ASSERT(xb.validTime() == time_);
  EnsemblePtr_ ens(new Ensemble_(xb.validTime(), ensConf));
  ens->build(xb, resol);
  EnsemblesCollection_::getInstance().put(xb.validTime(), ens);
  ens->read();
  if (has_change_of_variable_) {
    change_of_variable_.reset(MatrixFactory_::create(
        eckit::LocalConfiguration(ensConf, "change_of_variable")));
    for (unsigned l = 0; l < ens->size(); ++l) {
      change_of_variable_->multiply((*ens)[l]);
    }
  }
  ens->to_perturbations(1);

  // Build Localization
  if (ensConf.has("localization")) {
    const eckit::LocalConfiguration conf(ensConf, "localization");
    loc_.reset(new Localization_(resol, vars_, conf));
  }

  Log::trace() << "EnsembleCovariance linearized." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance<MODEL>::multiply(const Increment_ &dxi,
                                         Increment_ &dxo) const {
  EnsemblePtr_ e_k2 = EnsemblesCollection_::getInstance()[dxi.validTime()];
  EnsemblePtr_ e_k1 = EnsemblesCollection_::getInstance()[dxo.validTime()];
  Increment_ dxi_tmp(dxi);

  // Apply to dxi the block B_k1k2 of the B matrix which contains the
  // covariance of perturbations at time k1 with covariance at time k2
  if (has_change_of_variable_)
    change_of_variable_->transposeInverseMultiply(dxi_tmp);
  dxo.zero();
  for (unsigned int m = 0; m < e_k1->size(); ++m) {
    if (loc_) {
      // Localized covariance matrix
      Increment_ dx(dxi_tmp);
      dx.schur_product_with((*e_k2)[m]);
      loc_->multiply(dx);
      dx.schur_product_with((*e_k1)[m]);

      // We can't use '+=' , dxo and dx aren't at the same time (QG)
      dxo.axpy(1.0, dx, false);
    } else {
      // Raw covariance matrix
      double wgt = dxi_tmp.dot_product_with((*e_k2)[m]);
      dxo.axpy(wgt, (*e_k1)[m], false);
    }
  }
  if (has_change_of_variable_) change_of_variable_->inverseMultiply(dxo);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance<MODEL>::inverseMultiply(const Increment_ &dxi,
                                                Increment_ &dxo) const {
  IdentityMatrix<Increment_> Id;
  dxo.zero();
  GMRESR(dxo, dxi, *this, Id, 10, 1.0e-3);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance<MODEL>::multiplySqrt(const IncrCtlVec_ &dv,
                                             Increment_ &dxo) const {
  EnsemblePtr_ e_k = EnsemblesCollection_::getInstance()[dxo.validTime()];

  if (!loc_) {
    throw eckit::NotImplemented("mutiplySqrt not implemented yet", Here());
  }

  dxo.zero();
  Increment_ dxi(dxo);
  for (unsigned int m = 0; m < e_k->size(); ++m) {
    loc_->multiplySqrt(dv.ensCtlVec(m), dxi);
    dxi.schur_product_with((*e_k)[m]);
    dxo.axpy(1.0, dxi, false);
  }
  if (has_change_of_variable_) change_of_variable_->inverseMultiply(dxo);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance<MODEL>::multiplySqrtTrans(const Increment_ &dxi,
                                                  IncrCtlVec_ &dv) const {
  EnsemblePtr_ e_k = EnsemblesCollection_::getInstance()[dxi.validTime()];
  Increment_ dxi_tmp(dxi);

  if (!loc_) {
    throw eckit::NotImplemented("mutiplySqrtTrans not implemented yet", Here());
  }

  if (has_change_of_variable_)
    change_of_variable_->transposeInverseMultiply(dxi_tmp);
  for (unsigned int m = 0; m < e_k->size(); ++m) {
    dv.ensCtlVec(m).zero();
    Increment_ dx(dxi_tmp);
    dx.schur_product_with((*e_k)[m]);
    loc_->multiplySqrtTrans(dx, dv.ensCtlVec(m));
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance<MODEL>::randomize(Increment_ &dxo) const {
  // Build dv
  EnsemblePtr_ e_k = EnsemblesCollection_::getInstance()[dxo.validTime()];
  std::vector<std::unique_ptr<IncrEnsCtlVec<MODEL>>> dv;
  for (unsigned int jj = 0; jj < e_k->size(); ++jj) {
    dv.emplace_back(this->newIncrEnsCtlVec());
  }

  // Initialize dv
  for (unsigned int m = 0; m < e_k->size(); ++m) {
    dv[m]->random();
  }

  // dxo = B^1/2 dv
  dxo.zero();
  Increment_ dx(dxo);
  for (unsigned int m = 0; m < e_k->size(); ++m) {
    loc_->multiplySqrt(*dv[m], dx);
    dx.schur_product_with((*e_k)[m]);
    dxo.axpy(1.0, dx, false);
  }
  if (has_change_of_variable_) change_of_variable_->inverseMultiply(dxo);
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLECOVARIANCE_H_
