/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_ENSEMBLECOVARIANCE4D_H_
#define OOPS_BASE_ENSEMBLECOVARIANCE4D_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Localization.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"
//#include "util/MPIWrapper.h"

namespace oops {

/// Generic ensemble based model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovariance4D : public ModelSpaceCovariance4DBase<MODEL> {
  using Ensemble_ = Ensemble<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL>>;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Localization_ = Localization<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  EnsembleCovariance4D(const Geometry_ &, const Variables_ &,
                       const eckit::Configuration &, const State4D_ &);
  ~EnsembleCovariance4D();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*geom_, *loc_);
  }

 private:
  void schur_add(const Increment_ &, const Increment_ &dxi, Increment_ &dxo,
                 Increment_ &tmp) const;

  const eckit::LocalConfiguration config_;
  const Variables_ vars_;
  unsigned nsubwin_;
  std::unique_ptr<Localization_> loc_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL>> static_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4D<MODEL>::EnsembleCovariance4D(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      config_(config),
      vars_(vars),
      nsubwin_(0),
      loc_(),
      geom_(),
      static_() {
  Log::trace() << "EnsembleCovariance4D created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4D<MODEL>::~EnsembleCovariance4D() {
  Log::trace() << "EnsembleCovariance4D destructed.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4D<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol, const eckit::Configuration &) {
  /*
   * Build localization
   */
  geom_.reset(new Geometry_(resol));
  if (config_.has("localization")) {
    const eckit::LocalConfiguration conf(config_, "localization");
    loc_.reset(new Localization_(*geom_, vars_, conf));
  }

  /*
   * Build ensemble
   */
  std::vector<eckit::LocalConfiguration> confEnsemble;
  config_.get("ensemble", confEnsemble);
  nsubwin_ = confEnsemble.size();
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens_k(new Ensemble_(xb[jsub].validTime(), confEnsemble[jsub]));
    ens_k->build(xb[jsub], resol);
    EnsemblesCollection_::getInstance().put(xb[jsub].validTime(), ens_k);
  }

  /*
   * Read files
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->read();
  }

  /*
   * Calculate mean
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->to_perturbations();
  }

  Log::trace() << "EnsembleCovariance4D linearized.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4D<MODEL>::advectedMultiply(const Increment4D_ &dxi,
                                                   Increment4D_ &dxo) const {
  Increment_ dx_loc(dxo[0]);
  Increment_ tmp(dxo[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];

  dxo.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    if (loc_) {
      dx_loc.zero();
      for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
        EnsemblePtr_ e_k2 =
            EnsemblesCollection_::getInstance()[dxi[k2].validTime()];
        schur_add((*e_k2)[l], dxi[k2], dx_loc, tmp);
      }

      loc_->multiply(dx_loc);

      for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
        EnsemblePtr_ e_k1 =
            EnsemblesCollection_::getInstance()[dxi[k1].validTime()];
        schur_add((*e_k1)[l], dx_loc, dxo[k1], tmp);
      }
    } else {
      double wgt = 0;
      for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
        EnsemblePtr_ e_k2 =
            EnsemblesCollection_::getInstance()[dxi[k2].validTime()];
        wgt += dxi[k2].dot_product_with((*e_k2)[l]);
      }
      for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
        EnsemblePtr_ e_k1 =
            EnsemblesCollection_::getInstance()[dxi[k1].validTime()];
        dxo[k1].axpy(wgt, (*e_k1)[l], false);
      }
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4D<MODEL>::advectedMultiplySqrt(
    const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
  if (!loc_) {
    throw eckit::NotImplemented("advectedMultiplySqrt not implemented yet", Here());
  }

  Increment_ dx(dxo[0]);
  Increment_ tmp(dxo[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxo[0].validTime()];

  dxo.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    loc_->multiplySqrt(dvi.ensCtlVec(l), dx);
    for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
      EnsemblePtr_ e_k1 =
          EnsemblesCollection_::getInstance()[dxo[k1].validTime()];
      schur_add((*e_k1)[l], dx, dxo[k1], tmp);
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4D<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  if (!loc_) {
    throw eckit::NotImplemented("advectedMultiplySqrtTrans not implemented yet", Here());
  }

  Increment_ dx(dxi[0]);
  Increment_ tmp(dxi[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];

  for (unsigned l = 0; l < e->size(); ++l) {
    dx.zero();
    dvo.ensCtlVec(l).zero();
    for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
      EnsemblePtr_ e_k2 =
          EnsemblesCollection_::getInstance()[dxi[k2].validTime()];
      schur_add((*e_k2)[l], dxi[k2], dx, tmp);
    }
    loc_->multiplySqrtTrans(dx, dvo.ensCtlVec(l));
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4D<MODEL>::schur_add(const Increment_ &pert,
                                            const Increment_ &dxi,
                                            Increment_ &dxo,
                                            Increment_ &tmp_schur) const {
  tmp_schur = pert;
  tmp_schur.schur_product_with(dxi);
  dxo.axpy(1.0, tmp_schur, false);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4D<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLECOVARIANCE4D_H_
