/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_ENSEMBLECOVARIANCE4DDIST_H_
#define OOPS_BASE_ENSEMBLECOVARIANCE4DDIST_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Localization.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/MPIThreadPool.h"
#include "util/MPIWrapper.h"
#include "util/abor1_cpp.h"

namespace oops {

/// Generic ensemble based model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovariance4DDistributed
    : public ModelSpaceCovariance4DBase<MODEL> {
  using Ensemble_ = Ensemble<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL>>;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Localization_ = Localization<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  EnsembleCovariance4DDistributed(const Geometry_ &, const Variables_ &,
                                  const eckit::Configuration &,
                                  const State4D_ &);
  ~EnsembleCovariance4DDistributed();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*geom_, *loc_);
  }

 private:
  unsigned get_xb_prim(unsigned l) const;
  void schur_add(const Increment_ &, const Increment_ &dxi, Increment_ &dxo,
                 Increment_ &tmp) const;

  const eckit::LocalConfiguration config_;
  const Variables_ vars_;
  unsigned nsubwin_;
  std::vector<util::DateTime> time_;
  std::unique_ptr<Localization_> loc_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL>> static_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4DDistributed<MODEL>::EnsembleCovariance4DDistributed(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      config_(config),
      vars_(vars),
      nsubwin_((unsigned int)0) {
  Log::trace() << "EnsembleCovariance4DDistributed created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4DDistributed<MODEL>::~EnsembleCovariance4DDistributed() {
  Log::trace() << "EnsembleCovariance4DDistributed destructed.";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DDistributed<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol, const eckit::Configuration &) {
  /*
   * Build localization
   */
  geom_.reset(new Geometry_(resol));
  const eckit::LocalConfiguration confLoc(config_, "localization");
  loc_.reset(new Localization_(*geom_, vars_, confLoc));

  /*
   * Build ensemble
   */
  std::vector<eckit::LocalConfiguration> confEnsemble;
  config_.get("ensemble", confEnsemble);

  time_.clear();
  nsubwin_ = confEnsemble.size();
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens_k(
        new Ensemble_(xb[jsub].validTime(), confEnsemble[jsub], true));
    ens_k->build(xb[jsub], resol);
    EnsemblesCollection_::getInstance().put(xb[jsub].validTime(), ens_k);
    time_.push_back(xb[jsub].validTime());
  }

  /*
   * Calculate perturbations
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->to_perturbations((bool)true);
  }

  Log::trace() << "EnsembleCovariance4DDistributed linearized.";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DDistributed<MODEL>::advectedMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  Increment_ dx_loc(dxo[0]);
  Increment_ tmp(dxo[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];

  dxo.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    unsigned ix = get_xb_prim(l);
    dx_loc.zero();
    for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
      EnsemblePtr_ e_k2 =
          EnsemblesCollection_::getInstance()[dxi[k2].validTime()];
      schur_add((*e_k2)[ix], dxi[k2], dx_loc, tmp);
    }

    loc_->multiply(dx_loc);

    for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
      EnsemblePtr_ e_k1 =
          EnsemblesCollection_::getInstance()[dxi[k1].validTime()];
      schur_add((*e_k1)[ix], dx_loc, dxo[k1], tmp);
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DDistributed<MODEL>::advectedMultiplySqrt(
    const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
  Increment_ dx(dxo[0]);
  Increment_ tmp(dxo[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxo[0].validTime()];

  dxo.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    loc_->multiplySqrt(dvi.ensCtlVec(l), dx);
    unsigned ix = get_xb_prim(l);
    for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
      EnsemblePtr_ e_k1 =
          EnsemblesCollection_::getInstance()[dxo[k1].validTime()];
      schur_add((*e_k1)[ix], dx, dxo[k1], tmp);
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DDistributed<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  Increment_ dx(dxi[0]);
  Increment_ tmp(dxi[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];

  for (unsigned l = 0; l < e->size(); ++l) {
    unsigned ix = get_xb_prim(l);
    dx.zero();
    dvo.ensCtlVec(l).zero();
    for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
      EnsemblePtr_ e_k2 =
          EnsemblesCollection_::getInstance()[dxi[k2].validTime()];
      schur_add((*e_k2)[ix], dxi[k2], dx, tmp);
    }
    loc_->multiplySqrtTrans(dx, dvo.ensCtlVec(l));
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DDistributed<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
template <typename MODEL>
unsigned EnsembleCovariance4DDistributed<MODEL>::get_xb_prim(unsigned l) const {
  util::MPIWrapper &mpiw = util::MPIWrapper::Instance();

  // Circular permutation that depends on my rank
  unsigned i_xb_prim =
      (mpiw.nb_worlds() - mpiw.my_rank() + l) % mpiw.nb_worlds();

  if (mpiw.my_rank() == i_xb_prim)
    // Nothing to do, since xb_prim is my perturbation
    return 0;

  Log::info() << "EnsembleCovariance4DDistributed.get_xb_prim(): my_rank="
              << mpiw.my_rank() << " get xb_prim=" << i_xb_prim;

  util::MPIThreadPool<Increment_> pool;
  for (unsigned k = 0; k < nsubwin_; k++) {
    EnsemblePtr_ e_k = EnsemblesCollection_::getInstance()[time_[k]];
    pool.add(&(e_k->pack()), &(*e_k)[1]);
  }
  int tag = l;
  pool.send_recv_nb(i_xb_prim, tag);

  // xb_prim is the perturbation I have received and stored in (*e_k)[1])
  return 1;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DDistributed<MODEL>::schur_add(
    const Increment_ &pert, const Increment_ &dxi, Increment_ &dxo,
    Increment_ &tmp_schur) const {
  tmp_schur = pert;
  tmp_schur.schur_product_with(dxi);
  dxo.axpy(1.0, tmp_schur, false);
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLECOVARIANCEDIST4D_H_
