/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_EnsembleCovariance4DScaleDependentSCALEDEP_H_
#define OOPS_BASE_EnsembleCovariance4DScaleDependentSCALEDEP_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/HorizScaleDecomposition.h"
#include "util/DateTime.h"
#include "util/Logger.h"

namespace oops {

/// Wraps the model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovariance4DScaleDependent
    : public ModelSpaceCovariance4DBase<MODEL> {
  using Ensemble_ = Ensemble<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL>>;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Localization_ = Localization<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;
  using HorizScaleDecomposition_ = HorizScaleDecomposition<MODEL>;

 public:
  EnsembleCovariance4DScaleDependent(const Geometry_ &, const Variables_ &,
                                     const eckit::Configuration &,
                                     const State4D_ &);
  ~EnsembleCovariance4DScaleDependent();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*geom_, *loc_[0]);
  }

 private:
  void schur_add(const Increment_ &, const Increment_ &dxi, Increment_ &dxo,
                 Increment_ &tmp) const;

  const eckit::LocalConfiguration config_;
  const Variables_ vars_;
  std::vector<std::unique_ptr<Localization_>> loc_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL>> static_;
  unsigned nsubwin_;
  int nhorizBand_;
  int SDLtype_;
};

// =============================================================================

// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4DScaleDependent<MODEL>::EnsembleCovariance4DScaleDependent(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      config_(config),
      vars_(vars) {
  Log::trace() << "EnsembleCovariance4DScaleDependent created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4DScaleDependent<
    MODEL>::~EnsembleCovariance4DScaleDependent() {
  Log::trace() << "EnsembleCovariance4DScaleDependent destructed";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DScaleDependent<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol, const eckit::Configuration &) {
  const eckit::LocalConfiguration conf(config_, "localization");
  const eckit::LocalConfiguration confScaleDecomposition(config_,
                                                         "scale_decomposition");

  // What type of SDL is demanded
  SDLtype_ = confScaleDecomposition.getInt("formulation");

  // How many horizontal bands do we have?
  nhorizBand_ = confScaleDecomposition.getInt("nb_wave_band_peaks");

  // How many localization lenghts do we have?
  std::vector<eckit::LocalConfiguration> confHorizBand;
  confScaleDecomposition.get("horiz_band", confHorizBand);
  int nband = confHorizBand.size();
  // Make sure that the above two numbers are equal
  ASSERT(nband == nhorizBand_);

  geom_.reset(new Geometry_(resol));
  for (int horizBandIndex = 0; horizBandIndex < nhorizBand_; ++horizBandIndex) {
    loc_.emplace_back(
        new Localization_(*geom_, vars_, confHorizBand[horizBandIndex]));
  }

  /*
   * Build ensemble
   */
  std::vector<eckit::LocalConfiguration> confEnsemble;
  config_.get("ensemble", confEnsemble);
  nsubwin_ = confEnsemble.size();
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens_k(new Ensemble_(xb[jsub].validTime(), confEnsemble[jsub]));
    ens_k->build(xb[jsub], resol);
    EnsemblesCollection_::getInstance().put(xb[jsub].validTime(), ens_k);
  }

  /*
   * Read files
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->read();
  }

  /*
   * Calculate mean
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->to_perturbations();
  }

  /*
   * Performed a horizontal scale decomposition of the perturbations
   */
  Log::info() << "Scale decomposition of the perturbation object";

  HorizScaleDecomposition_ fltbg(confScaleDecomposition);

  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {  // loop on ensemble time

    Log::info() << " > treating ensemble time # " << jsub + 1;
    EnsemblePtr_ e = EnsemblesCollection_::getInstance()[xb[jsub].validTime()];

    for (int horizBandIndex = 0; horizBandIndex < nhorizBand_;
         ++horizBandIndex) {
      EnsemblePtr_ ens(new Ensemble_(*e));
      EnsembleKey k(xb[jsub].validTime(), horizBandIndex);
      EnsemblesCollection_::getInstance().put(k, ens);
    }

    for (unsigned l = 0; l < e->size(); ++l) {  // loop on ensemble member
      Log::info()
          << " >> calling horiz_scale_decomposition for ensemble member # "
          << l + 1;

      for (int horizBandIndex = 0; horizBandIndex < nhorizBand_;
           ++horizBandIndex) {
        EnsembleKey k(xb[jsub].validTime(), horizBandIndex);
        EnsemblePtr_ horizBandEns = EnsemblesCollection_::getInstance()[k];
        fltbg.collect((*horizBandEns)[l], horizBandIndex);
      }
      fltbg.decompose((*e)[l], SDLtype_);
    }
  }

  Log::trace() << "EnsembleCovariance4DScaleDependent linearized.";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void EnsembleCovariance4DScaleDependent<MODEL>::advectedMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  Increment_ ensAmplitude(dxo[0]);
  Increment_ tmp(dxo[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];
  unsigned ksiIndex;

  // Note on the scale-dependent localization (SDL) formulation...
  // We use here a B^1/2 approach to save computational cost, since:
  // With B^1/2, the cost of the SDL approach is proportional to the number of
  // scale-bands
  // With B    , the cost of the SDL approach is proportional to the SQUARE of
  // the number of scale-bands

  //
  // --- Apply B^T/2 ---
  //

  Log::info() << "Aplying B^T/2";

  std::vector<std::unique_ptr<IncrEnsCtlVec_>> aksi;
  if (SDLtype_ == 1) {
    // Create one Ksi per ensemble member
    for (unsigned l = 0; l < e->size(); ++l) {
      IncrEnsCtlVec_ *ksi = new IncrEnsCtlVec_(*geom_, *loc_[0]);
      aksi.emplace_back(ksi);
    }
  } else {
    // Create one Ksi per ensemble member and per horizontal waveband
    for (int horizBandIndex = 0; horizBandIndex < nhorizBand_;
         ++horizBandIndex) {
      for (unsigned l = 0; l < e->size(); ++l) {
        IncrEnsCtlVec_ *ksi = new IncrEnsCtlVec_(*geom_, *loc_[0]);
        aksi.emplace_back(ksi);
      }
    }
  }

  // loop on horizontal scale band
  for (int horizBandIndex = 0; horizBandIndex < nhorizBand_; ++horizBandIndex) {
    // loop on ensemble members
    for (unsigned memberIndex = 0; memberIndex < e->size(); ++memberIndex) {
      ensAmplitude.zero();
      // loop on time
      for (unsigned timeIndex = 0; timeIndex < nsubwin_; ++timeIndex) {
        EnsembleKey k(dxi[timeIndex].validTime(), horizBandIndex);
        EnsemblePtr_ e_t = EnsemblesCollection_::getInstance()[k];
        schur_add((*e_t)[memberIndex], dxi[timeIndex], ensAmplitude, tmp);
      }
      if (SDLtype_ == 1) {
        ksiIndex = memberIndex;
      } else {
        ksiIndex = horizBandIndex * e->size() + memberIndex;
      }
      loc_[horizBandIndex]->multiplySqrtTrans(ensAmplitude, *aksi[ksiIndex]);
    }
  }

  //
  // --- Apply B^1/2 ---
  //

  Log::info() << "Aplying B^1/2";

  dxo.zero();
  // loop on horizontal scale band
  for (int horizBandIndex = 0; horizBandIndex < nhorizBand_; ++horizBandIndex) {
    // loop on ensemble members
    for (unsigned memberIndex = 0; memberIndex < e->size(); ++memberIndex) {
      ensAmplitude.zero();
      if (SDLtype_ == 1) {
        ksiIndex = memberIndex;
      } else {
        ksiIndex = horizBandIndex * e->size() + memberIndex;
      }
      loc_[horizBandIndex]->multiplySqrt(*aksi[ksiIndex], ensAmplitude);

      // loop on time
      for (unsigned timeIndex = 0; timeIndex < nsubwin_; ++timeIndex) {
        EnsembleKey k(dxi[timeIndex].validTime(), horizBandIndex);
        EnsemblePtr_ e_t = EnsemblesCollection_::getInstance()[k];
        schur_add((*e_t)[memberIndex], ensAmplitude, dxo[timeIndex], tmp);
      }
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DScaleDependent<MODEL>::advectedMultiplySqrt(
    const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
  Increment_ ensAmplitude(dxo[0]);
  Increment_ tmp(dxo[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxo[0].validTime()];

  Log::info() << "Aplying B^1/2";

  dxo.zero();
  // loop on horizontal scale band
  for (int horizBandIndex = 0; horizBandIndex < nhorizBand_; ++horizBandIndex) {
    // loop on ensemble members
    for (unsigned memberIndex = 0; memberIndex < e->size(); ++memberIndex) {
      ensAmplitude.zero();
      loc_[horizBandIndex]->multiplySqrt(dvi.ensCtlVec(memberIndex),
                                         ensAmplitude);

      // loop on time
      for (unsigned timeIndex = 0; timeIndex < nsubwin_; ++timeIndex) {
        EnsembleKey k(dxo[timeIndex].validTime(), horizBandIndex);
        EnsemblePtr_ e_t = EnsemblesCollection_::getInstance()[k];
        schur_add((*e_t)[memberIndex], ensAmplitude, dxo[timeIndex], tmp);
      }
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DScaleDependent<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  Increment_ ensAmplitude(dxi[0]);
  Increment_ tmp(dxi[0]);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];

  Log::info() << "Aplying B^T/2";

  // loop on horizontal scale band
  for (int horizBandIndex = 0; horizBandIndex < nhorizBand_; ++horizBandIndex) {
    // loop on ensemble members
    for (unsigned memberIndex = 0; memberIndex < e->size(); ++memberIndex) {
      ensAmplitude.zero();
      // loop on time
      for (unsigned timeIndex = 0; timeIndex < nsubwin_; ++timeIndex) {
        EnsembleKey k(dxi[timeIndex].validTime(), horizBandIndex);
        EnsemblePtr_ e_t = EnsemblesCollection_::getInstance()[k];
        schur_add((*e_t)[memberIndex], dxi[timeIndex], ensAmplitude, tmp);
      }
      loc_[horizBandIndex]->multiplySqrtTrans(ensAmplitude,
                                              dvo.ensCtlVec(memberIndex));
    }
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DScaleDependent<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DScaleDependent<MODEL>::schur_add(
    const Increment_ &pert, const Increment_ &dxi, Increment_ &dxo,
    Increment_ &tmp_schur) const {
  tmp_schur = pert;
  tmp_schur.schur_product_with(dxi);
  dxo.axpy(1.0, tmp_schur, false);
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_EnsembleCovariance4DScaleDependent_H_
