/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_ENSEMBLECOVARIANCE4DWITHCHANGEOFVARIABLE_H_
#define OOPS_BASE_ENSEMBLECOVARIANCE4DWITHCHANGEOFVARIABLE_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Localization.h"
#include "util/DateTime.h"
#include "util/Logger.h"

namespace oops {

// The 4D Ensemble Covariance is connected to the Ensemble identified by id. It
// contains
// a Localization object and a ChangeOfVariable object.

// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovariance4DWithChangeOfVariable
    : public ModelSpaceCovariance4DBase<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Increment_ = Increment<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Ensemble_ = Ensemble<MODEL>;
  using Localization_ = Localization<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL>>;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;
  using MatrixFactory_ = GenericMatrixFactory<MODEL>;

 public:
  EnsembleCovariance4DWithChangeOfVariable(const Geometry_ &,
                                           const Variables_ &,
                                           const eckit::Configuration &,
                                           const State4D_ &);
  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*geom_, *loc_);
  }

 private:
  void schur_add(const Increment_ &, const Increment_ &dxi, Increment_ &dxo,
                 Increment_ &tmp) const;
  void assign_no_date_ctrl(const Increment_ &, Increment_ &) const;
  void toControlVariableTimes(const Increment4D_ &, Increment4D_ &) const;
  void toControlVariableTimes_ad(const Increment4D_ &, Increment4D_ &) const;
  unsigned ntslots() const { return time_.size(); }

  const Variables_ vars_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL>> static_;
  std::unique_ptr<Localization_> loc_;
  std::vector<util::DateTime> time_;
  std::unique_ptr<GenericMatrix_> change_of_variable_;
  int id_;
};

// =============================================================================

// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovariance4DWithChangeOfVariable<MODEL>::
    EnsembleCovariance4DWithChangeOfVariable(const Geometry_ &resol,
                                             const Variables_ &vars,
                                             const eckit::Configuration &config,
                                             const State4D_ &)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      vars_(vars) {
  id_ = 0;
  if (config.has("id")) id_ = config.getInt("id");

  Log::trace() << "EnsembleCovariance4DWithChangeOfVariable created.";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol,
    const eckit::Configuration &config) {
  unsigned nbtslots;
  std::vector<eckit::LocalConfiguration> confEnsemble;
  config.get("ensemble", confEnsemble);
  nbtslots = confEnsemble.size();

  /*
   * Build localization
   */
  geom_.reset(new Geometry_(resol));
  const eckit::LocalConfiguration conf(config, "localization");
  loc_.reset(new Localization_(*geom_, vars_, conf));

  /*
   * Build ensemble
   */
  time_.clear();
  for (unsigned jsub = 0; jsub < nbtslots; ++jsub) {
    util::DateTime dt(confEnsemble[jsub].getString("reference_time"));
    EnsembleKey id(dt, id_);

    EnsemblePtr_ ens_k(new Ensemble_(dt, confEnsemble[jsub]));
    ens_k->build(xb[0], resol);
    EnsemblesCollection_::getInstance().put(id, ens_k);
    time_.push_back(dt);
  }

  /*
   * Read files
   */
  for (unsigned jsub = 0; jsub < ntslots(); ++jsub) {
    EnsembleKey id(time_[jsub], id_);
    EnsemblePtr_ ens = EnsemblesCollection_::getInstance()[id];
    ens->read();
  }

  change_of_variable_.reset(MatrixFactory_::create(
      eckit::LocalConfiguration(config, "change_of_variable")));
  for (unsigned jsub = 0; jsub < ntslots(); ++jsub) {
    EnsembleKey id(time_[jsub], id_);
    EnsemblePtr_ ens = EnsemblesCollection_::getInstance()[id];
    for (unsigned l = 0; l < ens->size(); ++l) {
      change_of_variable_->multiply((*ens)[l]);
    }
  }

  /*
   * Calculate mean
   */
  int nb_sub_ens;
  for (unsigned jsub = 0; jsub < ntslots(); ++jsub) {
    EnsembleKey id(time_[jsub], id_);
    nb_sub_ens = confEnsemble[jsub].getInt("sub_ens");
    EnsemblePtr_ ens = EnsemblesCollection_::getInstance()[id];
    ens->to_perturbations(nb_sub_ens);
  }

  Log::trace() << "EnsembleCovariance4DWithChangeOfVariable linearized.";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::advectedMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
  Increment_ dx_loc(dxo[0]);
  Increment4D_ dxi_tmp(dxi);
  Increment4D_ dxo_tmp(dxo);
  Increment_ tmp(dxo[0]);
  EnsembleKey id0(time_[0], id_);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[id0];

  /*
   * Change variable in dxi
   */
  toControlVariableTimes_ad(dxi, dxi_tmp);
  for (unsigned k = 0; k < ntslots(); ++k) {
    change_of_variable_->transposeInverseMultiply(dxi_tmp[k]);
  }

  for (unsigned l = 0; l < e->size(); ++l) {
    dx_loc.zero();
    for (unsigned k2 = 0; k2 < ntslots(); ++k2) {
      EnsembleKey idk2(time_[k2], id_);
      EnsemblePtr_ e_k2 = EnsemblesCollection_::getInstance()[idk2];
      schur_add((*e_k2)[l], dxi_tmp[k2], dx_loc, tmp);
    }

    loc_->multiply(dx_loc);

    for (unsigned k1 = 0; k1 < ntslots(); ++k1) {
      EnsembleKey idk1(time_[k1], id_);
      EnsemblePtr_ e_k1 = EnsemblesCollection_::getInstance()[idk1];
      schur_add((*e_k1)[l], dx_loc, dxo_tmp[k1], tmp);
    }

    /*
     * Version without any localization ... for testing and debugging
     *
     dxo.zero();
     for (unsigned l = 0; l < e->size(); ++l) {
     for (unsigned k1 = 0; k1 < ntslots(); ++k1) {
     EnsemblePtr_ e_k1 =
     EnsemblesCollection_::getInstance()[dxi[k1].validTime()];
     for (unsigned k2 = 0; k2 < ntslots(); ++k2) {
     EnsemblePtr_ e_k2 =
     EnsemblesCollection_::getInstance()[dxi[k2].validTime()];
     double zz = (*e_k2)[l].dot_product_with(dxi[k2]);
     dxo[k1].axpy(zz, (*e_k1)[l]);
     }
     }*/
  }

  /*
   * Go back to the right variables
   */
  for (unsigned k = 0; k < ntslots(); ++k) {
    change_of_variable_->inverseMultiply(dxo_tmp[k]);
  }
  toControlVariableTimes(dxo_tmp, dxo);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::advectedMultiplySqrt(
    const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
  dxo.zero();
  Increment_ dx_loc(dxo[0]);
  Increment4D_ dxo_tmp(dxo);
  Increment_ tmp(dxo[0]);
  EnsembleKey id0(time_[0], id_);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[id0];

  for (unsigned l = 0; l < e->size(); ++l) {
    loc_->multiplySqrt(dvi.ensCtlVec(l), dx_loc);

    for (unsigned k1 = 0; k1 < ntslots(); ++k1) {
      EnsembleKey idk1(time_[k1], id_);
      EnsemblePtr_ e_k1 = EnsemblesCollection_::getInstance()[idk1];
      schur_add((*e_k1)[l], dx_loc, dxo_tmp[k1], tmp);
    }
  }

  /*
   * Go back to the right variables
   */
  for (unsigned k = 0; k < ntslots(); ++k) {
    change_of_variable_->inverseMultiply(dxo_tmp[k]);
  }
  toControlVariableTimes(dxo_tmp, dxo);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  Increment_ dx_loc(dxi[0]);
  Increment4D_ dxi_tmp(dxi);
  Increment_ tmp(dxi[0]);
  EnsembleKey id0(time_[0], id_);
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[id0];

  /*
   * Change variable in dxi
   */
  toControlVariableTimes_ad(dxi, dxi_tmp);
  for (unsigned k = 0; k < ntslots(); ++k) {
    change_of_variable_->transposeInverseMultiply(dxi_tmp[k]);
  }

  for (unsigned l = 0; l < e->size(); ++l) {
    dx_loc.zero();
    dvo.ensCtlVec(l).zero();
    for (unsigned k2 = 0; k2 < ntslots(); ++k2) {
      EnsembleKey idk2(time_[k2], id_);
      EnsemblePtr_ e_k2 = EnsemblesCollection_::getInstance()[idk2];
      schur_add((*e_k2)[l], dxi_tmp[k2], dx_loc, tmp);
    }

    loc_->multiplySqrtTrans(dx_loc, dvo.ensCtlVec(l));
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::schur_add(
    const Increment_ &pert, const Increment_ &dxi, Increment_ &dxo,
    Increment_ &tmp_schur) const {
  tmp_schur = pert;
  tmp_schur.schur_product_with(dxi);
  dxo.axpy(1.0, tmp_schur, false);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::toControlVariableTimes(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  if (ntslots() != (dxi.last() - dxi.first() + 1)) {
    assign_no_date_ctrl(dxi[0], dxo[0]);  // dxo[0] = dxi[0];

    for (unsigned k = 1; k < ntslots(); ++k) {
      assign_no_date_ctrl(dxi[k - 1],
                          dxo[2 * k - 1]);  // dxo[2 * k - 1] = dxi[k - 1];
      dxo[2 * k - 1] *= 0.5;
      dxo[2 * k - 1].axpy(0.5, dxi[k], false);

      assign_no_date_ctrl(dxi[k], dxo[2 * k]);  // dxo[2 * k] = dxi[k];
    }
  } else {
    dxo = dxi;
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::toControlVariableTimes_ad(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  // Check times
  int envar_times = dxi.last() - dxi.first() + 1;
  unsigned nb = ntslots();

  if (nb != envar_times) {
    ASSERT(envar_times == (2 * nb - 1));

    assign_no_date_ctrl(dxi[0], dxo[0]);  // dxo[0] = dxi[0];
    dxo[0].axpy(0.5, dxi[1], false);

    for (unsigned k = 1; k < (nb - 1); ++k) {
      assign_no_date_ctrl(dxi[2 * k], dxo[k]);  // dxo[k] = dxi[2 * k];
      dxo[k].axpy(0.5, dxi[2 * k - 1], false);
      dxo[k].axpy(0.5, dxi[2 * k + 1], false);
    }

    assign_no_date_ctrl(dxi[2 * nb - 2],
                        dxo[nb - 1]);  // dxo[nb - 1] = dxi[2 * nb - 2];
    dxo[nb - 1].axpy(0.5, dxi[2 * nb - 3], false);

  } else {
    dxo = dxi;
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovariance4DWithChangeOfVariable<MODEL>::assign_no_date_ctrl(
    const Increment_ &dxi, Increment_ &dxo) const {
  dxo.zero();
  dxo.axpy(1.0, dxi, false);
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLECOVARIANCE4DWITHCHANGEOFVARIABLE_H_
