/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_ENSEMBLECOVSINTERPOL_H_
#define OOPS_BASE_ENSEMBLECOVSINTERPOL_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/GMRESR.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/IdentityMatrix.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Interpolator.h"
#include "oops/interface/Localization.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename MODEL>
class IncrModCtlVec;

/// Generic ensemble based model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovsInterpol : public ModelSpaceCovarianceBase<MODEL> {
  typedef Ensemble<MODEL> Ensemble_;
  typedef std::shared_ptr<Ensemble<MODEL> > EnsemblePtr_;
  typedef EnsemblesCollection<MODEL> EnsemblesCollection_;
  typedef Geometry<MODEL> Geometry_;
  typedef Increment<MODEL> Increment_;
  typedef IncrCtlVec<MODEL> IncrCtlVec_;
  typedef IncrEnsCtlVec<MODEL> IncrEnsCtlVec_;
  typedef IncrModCtlVec<MODEL> IncrModCtlVec_;
  typedef Localization<MODEL> Localization_;
  typedef State<MODEL> State_;
  typedef Variables<MODEL> Variables_;
  typedef Interpolator<MODEL> Interpolator_;

 public:
  EnsembleCovsInterpol(const Geometry_ &, const Variables_ &,
                       const eckit::Configuration &, const State_ &);
  ~EnsembleCovsInterpol();

  void linearize(const State_ &, const Geometry_ &,
                 const eckit::Configuration &) override;
  void multiply(const Increment_ &, Increment_ &) const override;
  void inverseMultiply(const Increment_ &, Increment_ &) const override;
  void multiplySqrt(const IncrCtlVec_ &, Increment_ &) const override;
  void multiplySqrtTrans(const Increment_ &, IncrCtlVec_ &) const override;

  void randomize(Increment_ &) const override;

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*resol_ens_, *loc_);
  }

 private:
  void schur_add(const Increment_ &, const Increment_ &dxi, Increment_ &dxo,
                 Increment_ &tmp) const;

  const eckit::LocalConfiguration config_;
  const Variables_ vars_;
  const util::DateTime time_;
  std::unique_ptr<Localization_> loc_;
  std::unique_ptr<Interpolator_> intdir_;
  std::unique_ptr<Interpolator_> intadj_;
  std::unique_ptr<Geometry_> resol_ens_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovsInterpol<MODEL>::EnsembleCovsInterpol(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State_ &)
    : config_(config),
      vars_(vars),
      time_(config_.getString("date")) {
  Log::trace() << "EnsembleCovsInterpol created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovsInterpol<MODEL>::~EnsembleCovsInterpol() {
  loc_.reset();
  Log::trace() << "EnsembleCovsInterpol destructed." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::linearize(
    const State_ &xb, const Geometry_ &resol,
    const eckit::Configuration &config) {
  /*
   * Build Geometry of the Ensemble
   */
  const eckit::LocalConfiguration confNewGeo(config_, "ensemble_resol");
  resol_ens_.reset(new Geometry_(confNewGeo));

  /*
   * Build localization
   */
  const eckit::LocalConfiguration conf(config_, "localization");
  loc_.reset(new Localization_(*resol_ens_, vars_, conf));

  /*
   * Build Interpolation operators
   */
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  intdir_.reset(new Interpolator_(xb, *resol_ens_, confInterpol, false));
  intadj_.reset(new Interpolator_(xb, *resol_ens_, confInterpol, true));

  /*
   * Build ensemble
   */
  std::vector<eckit::LocalConfiguration> confEnsemble;
  config_.get("ensemble", confEnsemble);
  EnsemblePtr_ ens_k(new Ensemble_(xb.validTime(), confEnsemble[0]));
  ens_k->build(xb, *resol_ens_);
  EnsemblesCollection_::getInstance().put(xb.validTime(), ens_k);

  /*
   * Read files, calculate mean
   */
  EnsemblePtr_ ens = EnsemblesCollection_::getInstance()[xb.validTime()];
  ens->read();
  int nb_sub_ens = confEnsemble[0].getInt("sub_ens", 1);
  ens->to_perturbations(nb_sub_ens);

  Log::trace() << "EnsembleCovsInterpol linearized.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::multiply(const Increment_ &dxi,
                                           Increment_ &dxo) const {
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi.validTime()];
  dxo.zero();

  // Adjoint interpolation to the resolution of the perturbations
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  const Variables_ var(confInterpol);

  Increment_ dx_lr(*resol_ens_, var, dxi.validTime(), dxi.validTime());
  dx_lr.zero();

  Log::trace() << "Adjoint Interpolation (HR -> LR) of the increment";
  intadj_->interpolate(dxi, var, dx_lr);

  Increment_ dx_lr_loc(dx_lr);
  Increment_ tmp(dx_lr);
  Increment_ dxo_lr(dx_lr);

  dxo_lr.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    dx_lr_loc.zero();
    schur_add((*e)[l], dx_lr, dx_lr_loc, tmp);
    loc_->multiply(dx_lr_loc);
    schur_add((*e)[l], dx_lr_loc, dxo_lr, tmp);
  }

  // Back to HR
  Log::trace() << "Direct interpolation (LR -> HR) of the increment";
  intdir_->interpolate(dxo_lr, var, dxo);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::inverseMultiply(const Increment_ &dxi,
                                                  Increment_ &dxo) const {
  dxo.zero();
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::multiplySqrt(const IncrCtlVec_ &dv,
                                               Increment_ &dxo) const {
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxo.validTime()];

  // Interpolation to the resolution of the increment
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  const Variables_ var(confInterpol);

  Increment_ dx_lr(*resol_ens_, var, dxo.validTime(), dxo.validTime());
  Increment_ dx_lr_loc(dx_lr);
  Increment_ tmp(dx_lr_loc);
  Increment_ dxo_lr(dx_lr);

  dxo_lr.zero();
  dxo.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    loc_->multiplySqrt(dv.ensCtlVec(l), dx_lr_loc);
    schur_add((*e)[l], dx_lr_loc, dxo_lr, tmp);
  }

  // Back to HR
  Log::trace() << "Direct interpolation (LR -> HR) of the increment";
  intdir_->interpolate(dxo_lr, var, dxo);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::multiplySqrtTrans(const Increment_ &dxi,
                                                    IncrCtlVec_ &dv) const {
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi.validTime()];

  // Adjoint interpolation to the resolution of the increment
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  const Variables_ var(confInterpol);
  Increment_ dx_lr(*resol_ens_, var, dxi.validTime(), dxi.validTime());
  dx_lr.zero();

  Log::trace() << "Adjoint Interpolation (HR -> LR) of the increment";
  intadj_->interpolate(dxi, var, dx_lr);

  Increment_ dx_lr_loc(dx_lr);
  Increment_ tmp(dx_lr);

  for (unsigned l = 0; l < e->size(); ++l) {
    dv.ensCtlVec(l).zero();
    dx_lr_loc.zero();
    schur_add((*e)[l], dx_lr, dx_lr_loc, tmp);
    loc_->multiplySqrtTrans(dx_lr_loc, dv.ensCtlVec(l));
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::randomize(Increment_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol<MODEL>::schur_add(const Increment_ &pert,
                                            const Increment_ &dxi,
                                            Increment_ &dxo,
                                            Increment_ &tmp_schur) const {
  tmp_schur = pert;
  tmp_schur.schur_product_with(dxi);
  dxo.axpy(1.0, tmp_schur, false);
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLECOVSINTERPOL_H_
