/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_ENSEMBLECOVSINTERPOL4D_H_
#define OOPS_BASE_ENSEMBLECOVSINTERPOL4D_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Interpolator.h"
#include "oops/interface/Localization.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"
//#include "util/MPIWrapper.h"

namespace oops {

/// Generic ensemble based model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class EnsembleCovsInterpol4D : public ModelSpaceCovariance4DBase<MODEL> {
  using Ensemble_ = Ensemble<MODEL>;
  using EnsemblePtr_ = std::shared_ptr<Ensemble<MODEL>>;
  using EnsemblesCollection_ = EnsemblesCollection<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Localization_ = Localization<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Interpolator_ = Interpolator<MODEL>;

 public:
  EnsembleCovsInterpol4D(const Geometry_ &, const Variables_ &,
                         const eckit::Configuration &, const State4D_ &);
  ~EnsembleCovsInterpol4D();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return new IncrModCtlVec_();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return new IncrEnsCtlVec_(*resol_ens_, *loc_);
  }

 private:
  void schur_add(const Increment_ &, const Increment_ &dxi, Increment_ &dxo,
                 Increment_ &tmp) const;

  const eckit::LocalConfiguration config_;
  const Variables_ vars_;
  unsigned nsubwin_;
  std::unique_ptr<Localization_> loc_;
  std::unique_ptr<Interpolator_> intdir_;
  std::unique_ptr<Interpolator_> intadj_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL>> static_;
  std::unique_ptr<Geometry_> resol_ens_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovsInterpol4D<MODEL>::EnsembleCovsInterpol4D(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      config_(config),
      vars_(vars),
      nsubwin_(0),
      loc_(),
      static_() {
  Log::trace() << "EnsembleCovsInterpol4D created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
EnsembleCovsInterpol4D<MODEL>::~EnsembleCovsInterpol4D() {
  Log::trace() << "EnsembleCovsInterpol4D destructed.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol4D<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol, const eckit::Configuration &) {
  /*
   * Build Geometry of the Ensemble
   */
  const eckit::LocalConfiguration confNewGeo(config_, "ensemble_resol");
  resol_ens_.reset(new Geometry_(confNewGeo));

  /*
   * Build localization
   */
  const eckit::LocalConfiguration conf(config_, "localization");
  loc_.reset(new Localization_(*resol_ens_, vars_, conf));

  /*
   * Build Interpolation operators
   */
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  intdir_.reset(new Interpolator_(xb[0], *resol_ens_, confInterpol, false));
  intadj_.reset(new Interpolator_(xb[0], *resol_ens_, confInterpol, true));

  /*
   * Build ensemble
   */
  std::vector<eckit::LocalConfiguration> confEnsemble;
  config_.get("ensemble", confEnsemble);
  nsubwin_ = confEnsemble.size();
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens_k(new Ensemble_(xb[jsub].validTime(), confEnsemble[jsub]));
    ens_k->build(xb[jsub], *resol_ens_);
    EnsemblesCollection_::getInstance().put(xb[jsub].validTime(), ens_k);
  }

  /*
   * Read files
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->read();
  }

  /*
   * Calculate mean
   */
  for (unsigned jsub = 0; jsub < nsubwin_; ++jsub) {
    EnsemblePtr_ ens =
        EnsemblesCollection_::getInstance()[xb[jsub].validTime()];
    ens->to_perturbations();
  }

  Log::trace() << "EnsembleCovsInterpol4D linearized.";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol4D<MODEL>::advectedMultiply(const Increment4D_ &dxi,
                                                     Increment4D_ &dxo) const {
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];
  dxo.zero();

  // Adjoint interpolation to the resolution of the perturbations
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  const Variables_ var(confInterpol);

  Increment4D_ dx_lr(*resol_ens_, dxi);
  dx_lr.zero();

  for (unsigned k = 0; k < nsubwin_; ++k) {
    Log::trace()
        << "Adjoint Interpolation (HR -> LR) of the increment for timeslot #"
        << k;
    intadj_->interpolate(dxi[k], var, dx_lr[k]);
  }

  Increment_ dx_lr_loc(dx_lr[0]);
  Increment_ tmp(dx_lr[0]);
  Increment4D_ dxo_lr(dx_lr);

  dxo_lr.zero();

  for (unsigned l = 0; l < e->size(); ++l) {
    dx_lr_loc.zero();
    for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
      EnsemblePtr_ e_k2 =
          EnsemblesCollection_::getInstance()[dx_lr[k2].validTime()];
      schur_add((*e_k2)[l], dx_lr[k2], dx_lr_loc, tmp);
    }

    loc_->multiply(dx_lr_loc);

    for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
      EnsemblePtr_ e_k1 =
          EnsemblesCollection_::getInstance()[dx_lr[k1].validTime()];
      schur_add((*e_k1)[l], dx_lr_loc, dxo_lr[k1], tmp);
    }
  }

  // Back to HR
  for (unsigned k = 0; k < nsubwin_; ++k) {
    Log::trace()
        << "Direct interpolation (LR -> HR) of the increment for timeslot #"
        << k;
    intdir_->interpolate(dxo_lr[k], var, dxo[k]);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol4D<MODEL>::advectedMultiplySqrt(
    const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxo[0].validTime()];

  // Interpolation to the resolution of the increment
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  const Variables_ var(confInterpol);

  Increment4D_ dx_lr(*resol_ens_, dxo);
  Increment_ dx_lr_loc(dx_lr[0]);
  Increment_ tmp(dx_lr_loc);
  Increment4D_ dxo_lr(dx_lr);

  dxo_lr.zero();
  dxo.zero();
  for (unsigned l = 0; l < e->size(); ++l) {
    loc_->multiplySqrt(dvi.ensCtlVec(l), dx_lr_loc);

    for (unsigned k1 = 0; k1 < nsubwin_; ++k1) {
      EnsemblePtr_ e_k1 =
          EnsemblesCollection_::getInstance()[dxo[k1].validTime()];
      schur_add((*e_k1)[l], dx_lr_loc, dxo_lr[k1], tmp);
    }
  }

  // Back to HR
  for (unsigned k = 0; k < nsubwin_; ++k) {
    Log::trace()
        << "Direct interpolation (LR -> HR) of the increment for timeslot #"
        << k;
    intdir_->interpolate(dxo_lr[k], var, dxo[k]);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol4D<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  EnsemblePtr_ e = EnsemblesCollection_::getInstance()[dxi[0].validTime()];

  // Adjoint interpolation to the resolution of the increment
  const eckit::LocalConfiguration confInterpol(config_, "interpol");
  const Variables_ var(confInterpol);
  Increment4D_ dx_lr(*resol_ens_, dxi);
  dx_lr.zero();

  for (unsigned k = 0; k < nsubwin_; ++k) {
    Log::trace()
        << "Adjoint Interpolation (HR -> LR) of the increment for timeslot #"
        << k;
    intadj_->interpolate(dxi[k], var, dx_lr[k]);
  }

  Increment_ dx_lr_loc(dx_lr[0]);
  Increment_ tmp(dx_lr[0]);

  for (unsigned l = 0; l < e->size(); ++l) {
    dx_lr_loc.zero();
    for (unsigned k2 = 0; k2 < nsubwin_; ++k2) {
      EnsemblePtr_ e_k2 =
          EnsemblesCollection_::getInstance()[dx_lr[k2].validTime()];
      schur_add((*e_k2)[l], dx_lr[k2], dx_lr_loc, tmp);
    }

    loc_->multiplySqrtTrans(dx_lr_loc, dvo.ensCtlVec(l));
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol4D<MODEL>::schur_add(const Increment_ &pert,
                                              const Increment_ &dxi,
                                              Increment_ &dxo,
                                              Increment_ &tmp_schur) const {
  tmp_schur = pert;
  tmp_schur.schur_product_with(dxi);
  dxo.axpy(1.0, tmp_schur, false);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void EnsembleCovsInterpol4D<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLECOVSINTERPOL4D_H_
