/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_ENSEMBLESCOLLECTION_H_
#define OOPS_BASE_ENSEMBLESCOLLECTION_H_

#include <functional>
#include <iostream>
#include <memory>
#include <sstream>
#include <unordered_map>

#include "oops/base/Ensemble.h"
#include "util/DateTime.h"

class EnsembleKey {
 public:
  EnsembleKey(const util::DateTime& d, const int id) : d_(d), id_(id) {}
  bool operator==(const EnsembleKey& other) const {
    return (d_ == other.d_) && (id_ == other.id_);
  }
  std::string toString() const {
    std::stringstream ss;
    ss << d_.toString() << " id:" << id_;
    return ss.str();
  }

 private:
  util::DateTime d_;
  int id_;
};

namespace std {
template <>
class hash<EnsembleKey> {
 public:
  size_t operator()(const EnsembleKey& k) const {
    return std::hash<std::string>()(k.toString());
  }
};
}  // namespace std

namespace oops {

// -----------------------------------------------------------------------------

/// This class handles access to the ensembles of perturbation with a
/// DateTime

template <typename MODEL>
class EnsemblesCollection {
  using EnsemblesMap_ =
      std::unordered_map<EnsembleKey, std::shared_ptr<Ensemble<MODEL> > >;

 public:
  /// Constructor

  /// Destructor
  virtual ~EnsemblesCollection() {}

  /// Accessors
  std::shared_ptr<Ensemble<MODEL> > operator[](const EnsembleKey k) const {
    return ensMap_.at(k);
  }

  void put(const EnsembleKey& k, std::shared_ptr<Ensemble<MODEL> > e) {
    ensMap_[k] = e;
  }

  std::shared_ptr<Ensemble<MODEL> > operator[](const util::DateTime d) const {
    EnsembleKey k(d, 0);
    return ensMap_.at(k);
  }

  void put(const util::DateTime d, std::shared_ptr<Ensemble<MODEL> > e) {
    EnsembleKey k(d, 0);
    ensMap_[k] = e;
  }

  unsigned int size(const util::DateTime d) const {
    if (ensMap_.size() != 0) {
      EnsembleKey k(d, 0);
      return ensMap_.at(k)->size();
    } else {
      return 0;
    }
  }

  unsigned int size() const {
    if (ensMap_.size() != 0) {
      return ensMap_.begin()->second->size();
    } else {
      return 0;
    }
  }

  static EnsemblesCollection<MODEL>& getInstance() {
    static std::shared_ptr<EnsemblesCollection<MODEL> > instance_;
    if (instance_.use_count() == 0)
      instance_.reset(new EnsemblesCollection<MODEL>());
    return *instance_;
  }

  bool isEmpty() { return ensMap_.empty(); }

  bool has(const EnsembleKey k) {
    bool have = false;
    if (ensMap_.count(k)) have = true;
    return have;
  }

  bool has(const util::DateTime d) {
    EnsembleKey k(d, 0);
    return has(k);
  }

 private:
  EnsemblesCollection() {}

  EnsemblesMap_ ensMap_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_ENSEMBLESCOLLECTION_H_
