/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_HYBRIDCOVARIANCE4D_H_
#define OOPS_BASE_HYBRIDCOVARIANCE4D_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

/// Generic hybrid static-ensemble model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class HybridCovariance4D : public ModelSpaceCovariance4DBase<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;
  using MatrixFactory_ = GenericMatrixFactory<MODEL>;

 public:
  HybridCovariance4D(const Geometry_ &, const Variables_ &,
                     const eckit::Configuration &, const State4D_ &);
  ~HybridCovariance4D();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const {
    return static_->covar();
  }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return static_->newIncrModCtlVec();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return ensemble_->newIncrEnsCtlVec();
  }

 private:
  std::unique_ptr<GenericMatrix_> ensemble_weight_;
  std::unique_ptr<GenericMatrix_> static_weight_;
  std::unique_ptr<ModelSpaceCovariance4DBase<MODEL> > static_;
  std::unique_ptr<ModelSpaceCovariance4DBase<MODEL> > ensemble_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
HybridCovariance4D<MODEL>::HybridCovariance4D(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &xb)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      ensemble_weight_(MatrixFactory_::create(
          eckit::LocalConfiguration(config, "ensemble_weight"))),
      static_weight_(MatrixFactory_::create(
          eckit::LocalConfiguration(config, "static_weight"))),
      static_(Covariance4DFactory<MODEL>::create(
          eckit::LocalConfiguration(config, "static_covariance"), resol, vars,
          xb)),
      ensemble_(Covariance4DFactory<MODEL>::create(
          eckit::LocalConfiguration(config, "ensemble_covariance"), resol, vars,
          xb)) {
  Log::trace() << "HybridCovariance4D created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
HybridCovariance4D<MODEL>::~HybridCovariance4D() {
  Log::trace() << "HybridCovariance4D destructed";
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovariance4D<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol,
    const eckit::Configuration &config) {
  const eckit::LocalConfiguration jbConf(config, "Jb");
  const eckit::LocalConfiguration covConf(jbConf, "Covariance");
  const eckit::LocalConfiguration ensConf(covConf, "ensemble_covariance");
  const eckit::LocalConfiguration statConf(covConf, "static_covariance");

  if (covConf.has("override_ensemble_weight")) {
    ensemble_weight_.reset(MatrixFactory_::create(
        eckit::LocalConfiguration(covConf, "override_ensemble_weight")));
  }
  if (covConf.has("override_static_weight")) {
    static_weight_.reset(MatrixFactory_::create(
        eckit::LocalConfiguration(covConf, "override_static_weight")));
  }

  static_->linearize(xb, resol, statConf);
  ensemble_->linearize(xb, resol, ensConf);

  Log::trace() << "HybridCovariance4D linearized.";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void HybridCovariance4D<MODEL>::advectedMultiply(const Increment4D_ &dxi,
                                                 Increment4D_ &dxo) const {
  Increment4D_ tmp(dxo);
  Increment4D_ dxistat(dxi);
  static_weight_->multiply(dxistat);
  static_->multiply(dxistat, dxo);
  static_weight_->multiply(dxo);

  Increment4D_ dxiens(dxi);
  ensemble_weight_->multiply(dxiens);
  ensemble_->multiply(dxiens, tmp);
  ensemble_weight_->multiply(tmp);

  dxo += tmp;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovariance4D<MODEL>::advectedMultiplySqrt(const IncrCtlVec_ &dvi,
                                                     Increment4D_ &dxo) const {
  static_->multiplySqrt(dvi, dxo);
  static_weight_->multiply(dxo);
  Increment4D_ tmp(dxo);
  ensemble_->multiplySqrt(dvi, tmp);
  ensemble_weight_->multiply(tmp);
  dxo += tmp;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovariance4D<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  {  // Separate blocks improve memory handling
    Increment4D_ dxistat(dxi);
    static_weight_->multiply(dxistat);
    static_->multiplySqrtTrans(dxistat, dvo);
  }
  {
    Increment4D_ dxiens(dxi);
    ensemble_weight_->multiply(dxiens);
    ensemble_->multiplySqrtTrans(dxiens, dvo);
  }
}
// -----------------------------------------------------------------------------

template <typename MODEL>
void HybridCovariance4D<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_HYBRIDCOVARIANCE4D_H_
