/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_HYBRIDCOVARIANCEGENCOEFF_H_
#define OOPS_BASE_HYBRIDCOVARIANCEGENCOEFF_H_

#include <memory>  // for std::unique_ptr

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "oops/assimilation/GMRESR.h"
#include "oops/base/EnsembleCovariance.h"
#include "oops/base/IdentityMatrix.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

/// Generic hybrid static-ensemble model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class HybridCovarianceGeneralCoeff : public ModelSpaceCovarianceBase<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;
  using MatrixFactory_ = GenericMatrixFactory<MODEL>;

 public:
  HybridCovarianceGeneralCoeff(const Geometry_ &, const Variables_ &,
                               const eckit::Configuration &, const State_ &);
  ~HybridCovarianceGeneralCoeff();

  void linearize(const State_ &, const Geometry_ &,
                 const eckit::Configuration &) override;
  void multiply(const Increment_ &, Increment_ &) const override;
  void inverseMultiply(const Increment_ &, Increment_ &) const override;
  void multiplySqrt(const IncrCtlVec_ &, Increment_ &) const override;
  void multiplySqrtTrans(const Increment_ &, IncrCtlVec_ &) const override;

  void randomize(Increment_ &) const override;

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const override {
    return static_->newIncrModCtlVec();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const override {
    return ensemble_->newIncrEnsCtlVec();
  }

 private:
  std::unique_ptr<GenericMatrix_> ensemble_weight_;
  std::unique_ptr<GenericMatrix_> static_weight_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL> > static_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL> > ensemble_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
HybridCovarianceGeneralCoeff<MODEL>::HybridCovarianceGeneralCoeff(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State_ &xb)
    : ensemble_weight_(MatrixFactory_::create(
          eckit::LocalConfiguration(config, "ensemble_weight"))),
      static_weight_(MatrixFactory_::create(
          eckit::LocalConfiguration(config, "static_weight"))),
      static_(CovarianceFactory<MODEL>::create(
          eckit::LocalConfiguration(config, "static_covariance"), resol, vars,
          xb)),
      ensemble_(CovarianceFactory<MODEL>::create(
          eckit::LocalConfiguration(config, "ensemble_covariance"), resol, vars,
          xb)) {
  Log::trace() << "HybridCovarianceGeneralCoeff created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
HybridCovarianceGeneralCoeff<MODEL>::~HybridCovarianceGeneralCoeff() {
  Log::trace() << "HybridCovarianceGeneralCoeff destructed" << std::endl;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovarianceGeneralCoeff<MODEL>::linearize(
    const State_ &xb, const Geometry_ &resol,
    const eckit::Configuration &config) {
  const eckit::LocalConfiguration jbConf(config, "Jb");
  const eckit::LocalConfiguration covConf(jbConf, "Covariance");
  const eckit::LocalConfiguration ensConf(covConf, "ensemble_covariance");
  const eckit::LocalConfiguration statConf(covConf, "static_covariance");

  static_->linearize(xb, resol, statConf);
  ensemble_->linearize(xb, resol, ensConf);

  Log::trace() << "HybridCovarianceGeneralCoeff linearized.";
}
// -----------------------------------------------------------------------------

template <typename MODEL>
void HybridCovarianceGeneralCoeff<MODEL>::multiply(const Increment_ &dxi,
                                                   Increment_ &dxo) const {
  Increment_ tmp(dxo);
  Increment_ dxistat(dxi);
  static_weight_->multiply(dxistat);
  static_->multiply(dxistat, dxo);
  static_weight_->multiply(dxo);

  Increment_ dxiens(dxi);
  ensemble_weight_->multiply(dxiens);
  ensemble_->multiply(dxiens, tmp);
  ensemble_weight_->multiply(tmp);

  dxo += tmp;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
void HybridCovarianceGeneralCoeff<MODEL>::inverseMultiply(
    const Increment_ &dxi, Increment_ &dxo) const {
  dxo.zero();
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovarianceGeneralCoeff<MODEL>::multiplySqrt(const IncrCtlVec_ &dvi,
                                                       Increment_ &dxo) const {
  Increment_ tmp(dxo);
  static_->multiplySqrt(dvi, dxo);
  static_weight_->multiply(dxo);
  ensemble_->multiplySqrt(dvi, tmp);
  ensemble_weight_->multiply(tmp);
  dxo += tmp;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovarianceGeneralCoeff<MODEL>::multiplySqrtTrans(
    const Increment_ &dxi, IncrCtlVec_ &dvo) const {
  Increment_ dxistat(dxi);
  static_weight_->multiply(dxistat);
  static_->multiplySqrtTrans(dxistat, dvo);
  Increment_ dxiens(dxi);
  ensemble_weight_->multiply(dxiens);
  ensemble_->multiplySqrtTrans(dxiens, dvo);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void HybridCovarianceGeneralCoeff<MODEL>::randomize(Increment_ &) const {
  ABORT("HybridCovarianceGeneralCoeff::randomize: Would it make sense?");
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_HYBRIDCOVARIANCEGENCOEFF_H_
