/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_INCRCTLVEC4D_H_
#define OOPS_BASE_INCRCTLVEC4D_H_

#include <map>
#include <memory>
#include <string>

#include "oops/assimilation/CostJbState.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/interface/Geometry.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief IncrCtlVec4D class

 *   State part of the control vector. It has a 4D structure:

 *   \f[ dv = \begin{bmatrix}
                dv^0 \\
                ..   \\
                dv^K \\
                ..   \\
                dv^M \\
              \end{bmatrix}
     \f]

    where \f$ dv^K \f$ corresponds to the K-th sub-window and is defined as:

 *    \f[ dv^K = \begin{bmatrix}
                   dv^K_{mod}  \\
                   dv^K_{ens}
                 \end{bmatrix}
      \f]

    ->  \f$ dv^K \f$ is implemented by IncrCtlVec

 */

template <typename MODEL>
class IncrCtlVec4D : public util::Printable,
                     private util::ObjectCounter<IncrCtlVec4D<MODEL> > {
  using Geometry_ = Geometry<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using JbState_ = CostJbState<MODEL>;

 public:
  static const std::string classname() { return "oops::IncrCtlVec4D"; }

  /// Constructor
  explicit IncrCtlVec4D(const JbState_ &);
  /// Constructor with copy initialization
  IncrCtlVec4D(const JbState_ &, const IncrCtlVec4D &);
  /// Copy constructor
  IncrCtlVec4D(const IncrCtlVec4D &, const bool copy = true);
  /// Destructor
  virtual ~IncrCtlVec4D();

  // Interfacing
  /// Return reference to the component corresponding to i-th sub-window
  IncrCtlVec_ &operator[](const int ii) { return this->get(ii); }
  /// Return constant reference to the component corresponding to i-th
  /// sub-window
  const IncrCtlVec_ &operator[](const int ii) const { return this->get(ii); }
  /// Return the index of the first component
  int first() const { return first_; }
  /// Return the index of the last component
  int last() const { return last_; }

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  void zero();
  /// \f$ dv_1 = dv_2 \f$
  IncrCtlVec4D &operator=(const IncrCtlVec4D &);
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  IncrCtlVec4D &operator+=(const IncrCtlVec4D &);
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  IncrCtlVec4D &operator-=(const IncrCtlVec4D &);
  /// \f$ dv_1 = dv_1 * c \f$
  IncrCtlVec4D &operator*=(const double &);
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  void axpy(const double &, const IncrCtlVec4D &, const bool check = true);
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  double dot_product_with(const IncrCtlVec4D &) const;
  /// Hadamard product (element-wise)
  void schur_product_with(const IncrCtlVec4D &);
  /// Random control vector
  void random();

  // I/O and diagnostics
  /// Read control vector
  void read(const eckit::Configuration &);
  /// Write control vector
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

 private:
  /// Return reference to the component corresponding to i-th sub-window
  IncrCtlVec_ &get(const int);
  /// Return constant reference to the component corresponding to i-th
  /// sub-window
  const IncrCtlVec_ &get(const int) const;
  void print(std::ostream &) const;

  /// Iterator over a map of IncrCtlVec
  using iter_ = typename std::map<int, std::unique_ptr<IncrCtlVec_> >::iterator;
  /// Constant iterator over a map of IncrCtlVec
  using icst_ =
      typename std::map<int, std::unique_ptr<IncrCtlVec_> >::const_iterator;
  /// Map of IncrCtlVec (key: integer; value: IncrCtlVec pointer)
  std::map<int, std::unique_ptr<IncrCtlVec_> > incrCtlVec4d_;
  /// Index of the first control vector
  int first_;
  /// Index of the last control vector
  int last_;
};

// =============================================================================
/// Constructor, destructor
// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL>::IncrCtlVec4D(const JbState_ &jb)
    : incrCtlVec4d_(), first_(0), last_(jb.nchistates() - 1) {
  for (int jsub = 0; jsub <= last_; ++jsub) {
    std::unique_ptr<IncrCtlVec_> incrCtlVec(jb.newIncrCtlVec(jsub));
    incrCtlVec4d_.emplace(jsub, std::move(incrCtlVec));
  }
  Log::trace() << "IncrCtlVec4D<MODEL>::IncrCtlVec4D created" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL>::IncrCtlVec4D(const JbState_ &jb, const IncrCtlVec4D &other)
    : incrCtlVec4d_(), first_(0), last_(jb.nchistates() - 1) {
  for (int jsub = 0; jsub <= last_; ++jsub) {
    IncrCtlVec_ *incrCtlVec = jb.newIncrCtlVec(jsub);
    incrCtlVec4d_.insert(jsub, incrCtlVec);
    incrCtlVec4d_[jsub] = other.incrCtlVec4d_[jsub];
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::IncrCtlVec4D done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL>::IncrCtlVec4D(const IncrCtlVec4D &other, const bool copy)
    : incrCtlVec4d_(), first_(0), last_(other.last_) {
  for (int jsub = 0; jsub <= last_; ++jsub) {
    std::unique_ptr<IncrCtlVec_> incrCtlVec(new IncrCtlVec_(other[jsub]));
    incrCtlVec4d_.emplace(jsub, std::move(incrCtlVec));
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::IncrCtlVec4D copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL>::~IncrCtlVec4D() {
  Log::trace() << "IncrCtlVec4D<MODEL>::~IncrCtlVec4D starting" << std::endl;
  Log::trace() << "IncrCtlVec4D<MODEL>::~IncrCtlVec4D done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec<MODEL> &IncrCtlVec4D<MODEL>::get(const int ii) {
  iter_ it = incrCtlVec4d_.find(ii);
  ASSERT(it != incrCtlVec4d_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
const IncrCtlVec<MODEL> &IncrCtlVec4D<MODEL>::get(const int ii) const {
  icst_ it = incrCtlVec4d_.find(ii);
  ASSERT(it != incrCtlVec4d_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::zero() {
  Log::trace() << "IncrCtlVec4D<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");

  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    jsub->second->zero();
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL> &IncrCtlVec4D<MODEL>::operator=(const IncrCtlVec4D &rhs) {
  Log::trace() << "IncrCtlVec4D<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");

  for (int jsub = rhs.first(); jsub <= rhs.last(); ++jsub) {
    this->get(jsub) = rhs[jsub];
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL> &IncrCtlVec4D<MODEL>::operator+=(const IncrCtlVec4D &rhs) {
  Log::trace() << "IncrCtlVec4D<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");

  for (int jsub = rhs.first(); jsub <= rhs.last(); ++jsub) {
    this->get(jsub) += rhs[jsub];
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL> &IncrCtlVec4D<MODEL>::operator-=(const IncrCtlVec4D &rhs) {
  Log::trace() << "IncrCtlVec4D<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");

  for (int jsub = rhs.first(); jsub <= rhs.last(); ++jsub) {
    this->get(jsub) -= rhs[jsub];
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrCtlVec4D<MODEL> &IncrCtlVec4D<MODEL>::operator*=(const double &zz) {
  Log::trace() << "IncrCtlVec4D<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");

  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    *jsub->second *= zz;
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::axpy(const double &zz, const IncrCtlVec4D &rhs,
                               const bool check) {
  Log::trace() << "IncrCtlVec4D<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");

  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    jsub->second->axpy(zz, rhs[jsub->first]);
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double IncrCtlVec4D<MODEL>::dot_product_with(const IncrCtlVec4D &dv) const {
  Log::trace() << "IncrCtlVec4D<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");

  double zz = 0.0;
  for (icst_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    zz += dot_product(*jsub->second, dv[jsub->first]);
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::schur_product_with(const IncrCtlVec4D &dv) {
  Log::trace() << "IncrCtlVec4D<MODEL>::schur_product_with starting"
               << std::endl;
  util::Timer timer(classname(), "schur_product_with");

  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    jsub->second->shur_product_with(dv[jsub->first]);
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::random() {
  Log::trace() << "IncrCtlVec4D<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");

  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    jsub->second->random();
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "IncrCtlVec4D<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  std::vector<eckit::LocalConfiguration> confs;
  conf.get("ctlvec", confs);
  ASSERT(incrCtlVec4d_.size() == confs.size());
  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    jsub->second->read(confs[jsub->first]);
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "IncrCtlVec4D<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  Log::info() << "IncrCtlVec4D<MODEL>::write conf" << conf << std::endl;
  std::vector<eckit::LocalConfiguration> confs;
  conf.get("ctlvec", confs);
  ASSERT(incrCtlVec4d_.size() == confs.size());
  for (icst_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    jsub->second->write(confs[jsub->first]);
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double IncrCtlVec4D<MODEL>::norm() const {
  Log::trace() << "IncrCtlVec4D<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");

  double zz = 0.0;

  for (iter_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    zz += jsub->second->norm();
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrCtlVec4D<MODEL>::print(std::ostream &os) const {
  Log::trace() << "IncrCtlVec4D<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");

  for (icst_ jsub = incrCtlVec4d_.begin(); jsub != incrCtlVec4d_.end();
       ++jsub) {
    os << *jsub->second << std::endl;
  }

  Log::trace() << "IncrCtlVec4D<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_INCRCTLVEC4D_H_
