/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_INCREMENT4D_H_
#define OOPS_BASE_INCREMENT4D_H_

#include <cmath>
#include <map>
#include <memory>
#include <ostream>
#include <sstream>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/CostJbState.h"
#include "oops/base/State4D.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/Printable.h"
#include "util/dot_product.h"

namespace oops {

/*! @brief Increment4D class

 *  The state increment contains the increment to the 3D or 4D state part of
 * the control increment.

 *   \f$ dx_{state4D} \f$ as implemented by Increment4D implements the state
 *  component of a ControlIncrement and contains a vector of Increment objects.

 *   \f[ dx_{state4D} = \begin{bmatrix}
                           dx^0 \\
                           ..   \\
                           dx^K \\
                           ..   \\
                           dx^M \\
                        \end{bmatrix}
     \f]

     where \f$ dx^K \f$ corresponds to the Increment valid for K-th sub-window.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class Increment4D : public util::Printable {
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using JbState_ = CostJbState<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "Increment4D"; }

  /// Constructors
  explicit Increment4D(const JbState_ &);
  Increment4D(const Geometry_ &, const Variables_ &,
              const std::vector<util::DateTime> &);
  /// Copy constructor
  Increment4D(const Increment4D &, const bool copy = true);
  /// Copy constructor with additional configuration
  Increment4D(const Increment4D &, const eckit::Configuration &);
  /// Copy constructor with interpolation if required
  Increment4D(const Geometry_ &, const Increment4D &);
  /// Destructor
  ~Increment4D();

  // Linear algebra operators
  /// Compute 4D increment as a difference between two 4D states
  void diff(const State4D_ &, const State4D_ &);
  /// \f$ dx_1 = 0 \f$
  void zero();
  /// \f$ dx_1 = dx_2 \f$
  Increment4D &operator=(const Increment4D &);
  /// \f$ dx_1 = dx_1 + dx_2 \f$
  Increment4D &operator+=(const Increment4D &);
  /// \f$ dx_1 = dx_1 - dx_2 \f$
  Increment4D &operator-=(const Increment4D &);
  /// \f$ dx_1 = dx_1 * c \f$
  Increment4D &operator*=(const double);
  /// \f$ dx_1 = dx_1 * c dx_2 \f$
  void axpy(const double, const Increment4D &);
  /// \f$ <dx_1,dx_1> = dx_1^T dx_1 \f$
  double dot_product_with(const Increment4D &) const;

  // I/O and diagnostics
  /// Read increment
  void read(const eckit::Configuration &);
  /// Write increment
  void write(const eckit::Configuration &) const;

  /// Get geometry
  Geometry_ geometry() const { return this->get(first_).geometry(); }

  /// Get a reference to i-th Increment
  Increment_ &operator[](const int ii) { return this->get(ii); }
  /// Get a constant reference to i-th Increment
  const Increment_ &operator[](const int ii) const { return this->get(ii); }
  /// Return the index of the first Increment
  int first() const { return first_; }
  /// Return the index of the last Increment
  int last() const { return last_; }

  /// To be removed
  void shift_forward();
  void shift_backward();

  /// Accessor to times
  const std::vector<util::DateTime> times() const { return times_; }

 private:
  /// Get a reference to i-th Increment
  Increment_ &get(const int);
  /// Get a constant reference to i-th Increment
  const Increment_ &get(const int) const;
  void print(std::ostream &) const;

  /// Iterator over a map of Increment objects
  using iter_ = typename std::map<int, std::unique_ptr<Increment_>>::iterator;
  /// Constant iterator over a map of Increment objects
  using icst_ =
      typename std::map<int, std::unique_ptr<Increment_>>::const_iterator;
  /// Map of Increment objects (key: integer; value: Increment pointer)
  std::map<int, std::unique_ptr<Increment_>> incr4d_;
  /// Index of the first Increment
  int first_;
  /// Index of the last Increment
  int last_;
  /// Times
  std::vector<util::DateTime> times_;
};

// =============================================================================

template <typename MODEL>
Increment4D<MODEL>::Increment4D(const JbState_ &jb)
    : incr4d_(), first_(0), last_(jb.nstates() - 1) {
  for (int jsub = 0; jsub <= last_; ++jsub) {
    std::unique_ptr<Increment_> incr(jb.newStateIncrement(jsub));
    times_.push_back(incr->validTime());
    incr4d_.emplace(jsub, std::move(incr));
  }
  Log::trace() << "Increment4D:Increment4D created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL>::Increment4D(const Geometry_ &geom, const Variables_ &vars,
                                const std::vector<util::DateTime> &times)
    : incr4d_(), first_(0), last_(times.size() - 1) {
  for (int jsub = 0; jsub <= last_; ++jsub) {
    std::unique_ptr<Increment_> incr(new Increment_(geom, vars, times[jsub]));
    times_.push_back(incr->validTime());
    incr4d_.emplace(jsub, std::move(incr));
  }
  Log::trace() << "Increment4D:Increment4D created." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL>::Increment4D(const Increment4D &other, const bool copy)
    : incr4d_(), first_(other.first_), last_(other.last_) {
  for (icst_ jsub = other.incr4d_.begin(); jsub != other.incr4d_.end();
       ++jsub) {
    int isub = jsub->first;
    std::unique_ptr<Increment_> incr(new Increment_(*jsub->second, copy));
    times_.push_back(incr->validTime());
    incr4d_.emplace(isub, std::move(incr));
  }
  Log::trace() << "Increment4D:Increment4D copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL>::Increment4D(const Increment4D &other,
                                const eckit::Configuration &tlConf)
    : incr4d_(), first_(other.first_), last_(other.last_) {
  for (icst_ jsub = other.incr4d_.begin(); jsub != other.incr4d_.end();
       ++jsub) {
    int isub = jsub->first;
    std::unique_ptr<Increment_> incr(new Increment_(*jsub->second));
    times_.push_back(incr->validTime());
    incr4d_.emplace(isub, std::move(incr));
  }
  Log::trace() << "Increment4D:Increment4D copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL>::Increment4D(const Geometry_ &geom, const Increment4D &other)
    : incr4d_(), first_(other.first_), last_(other.last_) {
  for (icst_ jsub = other.incr4d_.begin(); jsub != other.incr4d_.end();
       ++jsub) {
    int isub = jsub->first;
    std::unique_ptr<Increment_> incr(new Increment_(geom, *jsub->second));
    times_.push_back(incr->validTime());
    incr4d_.emplace(isub, std::move(incr));
  }
  Log::trace() << "Increment4D:Increment4D copied." << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment<MODEL> &Increment4D<MODEL>::get(const int ii) {
  iter_ it = incr4d_.find(ii);
  ASSERT(it != incr4d_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
const Increment<MODEL> &Increment4D<MODEL>::get(const int ii) const {
  icst_ it = incr4d_.find(ii);
  ASSERT(it != incr4d_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL>::~Increment4D() {}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::zero() {
  for (iter_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    jsub->second->zero();
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::diff(const State4D_ &cv1, const State4D_ &cv2) {
  for (iter_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    jsub->second->diff(cv1[jsub->first], cv2[jsub->first]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL> &Increment4D<MODEL>::operator=(const Increment4D &rhs) {
  incr4d_.clear();
  for (icst_ jsub = rhs.incr4d_.begin(); jsub != rhs.incr4d_.end(); ++jsub) {
    int isub = jsub->first;
    std::unique_ptr<Increment_> incr(new Increment_(*jsub->second));
    incr4d_.emplace(isub, std::move(incr));
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL> &Increment4D<MODEL>::operator+=(const Increment4D &rhs) {
  for (int jsub = rhs.first(); jsub <= rhs.last(); ++jsub) {
    this->get(jsub) += rhs[jsub];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL> &Increment4D<MODEL>::operator-=(const Increment4D &rhs) {
  for (int jsub = rhs.first(); jsub <= rhs.last(); ++jsub) {
    this->get(jsub) -= rhs[jsub];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Increment4D<MODEL> &Increment4D<MODEL>::operator*=(const double zz) {
  for (iter_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    *jsub->second *= zz;
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::axpy(const double zz, const Increment4D &rhs) {
  for (iter_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    jsub->second->axpy(zz, rhs[jsub->first]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::read(const eckit::Configuration &config) {
  std::vector<eckit::LocalConfiguration> confs;
  config.get("increment", confs);
  ASSERT(incr4d_.size() == confs.size());
  for (iter_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    Log::debug() << "Increment4D:read" << confs[jsub->first] << std::endl;
    jsub->second->read(confs[jsub->first]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::write(const eckit::Configuration &config) const {
  std::vector<eckit::LocalConfiguration> confs;
  config.get("increment", confs);
  ASSERT(incr4d_.size() == confs.size());
  for (icst_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    Log::debug() << "Increment4D:write" << confs[jsub->first] << std::endl;
    jsub->second->write(confs[jsub->first]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::print(std::ostream &outs) const {
  for (icst_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    outs << *jsub->second << std::endl;
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double Increment4D<MODEL>::dot_product_with(const Increment4D &x2) const {
  double zz = 0.0;
  for (icst_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    zz += dot_product(*jsub->second, x2[jsub->first]);
  }
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::shift_forward() {
  using rit =
      typename std::map<int, std::unique_ptr<Increment_>>::reverse_iterator;
  for (rit jsub = incr4d_.rbegin(); jsub != incr4d_.rend(); ++jsub) {
    const int isub = jsub->first;
    if (isub > first_) this->get(isub) = this->get(isub - 1);
  }
  incr4d_.erase(first_);
  times_.erase(times_.begin());
  Log::info() << "Increment4D::shift_forward erased " << first_ << std::endl;
  first_ += 1;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Increment4D<MODEL>::shift_backward() {
  for (iter_ jsub = incr4d_.begin(); jsub != incr4d_.end(); ++jsub) {
    const int isub = jsub->first;
    if (isub < last_) this->get(isub) = this->get(isub + 1);
  }
  incr4d_.erase(last_);
  times_.pop_back();
  Log::info() << "Increment4D::shift_backward erased " << last_ << std::endl;
  last_ -= 1;
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_INCREMENT4D_H_
