/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_LINEAROBSOPERATORS_H_
#define OOPS_BASE_LINEAROBSOPERATORS_H_

#include <memory>
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/LinearObsOperator.h"
#include "oops/interface/ObsVector.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

/*! @brief A container class storing a vector of LinearObsOpertor objects.

 *   Refer to the documentation of CostJo class to understand the structure of
 *  the \f$ J_o \f$ term.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class LinearObsOperators : public util::Printable, private eckit::NonCopyable {
  using LinearObsOperator_ = LinearObsOperator<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;
  using ObsVector_ = ObsVector<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::LinearObsOperators"; }

  /// Constructor
  explicit LinearObsOperators(const ObsSpaces_ &);
  /// Destructor
  ~LinearObsOperators();

  // Access
  /// Return the number of LinearObsOpertor objects stored
  std::size_t size() const { return ops_.size(); }
  /// Return a reference to i-th LinearObsOpertor
  LinearObsOperator_ &operator[](const std::size_t ii) { return *ops_.at(ii); }
  /// Return a constant reference to i-th LinearObsOpertor
  const LinearObsOperator_ &operator[](const std::size_t ii) const {
    return *ops_.at(ii);
  }
  /// Retrun variables (required input variables from LinearModel)
  Variables_ variables(const std::size_t) const;

 private:
  void print(std::ostream &) const;

  /// Vector of LinearObsOpertor objects
  std::vector<std::shared_ptr<LinearObsOperator_> > ops_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearObsOperators<MODEL>::LinearObsOperators(const ObsSpaces_ &os) : ops_(0) {
  for (std::size_t jobs = 0; jobs < os.size(); ++jobs) {
    std::shared_ptr<LinearObsOperator_> tmp(new LinearObsOperator_(os[jobs]));
    ops_.push_back(tmp);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearObsOperators<MODEL>::~LinearObsOperators() {}

// -----------------------------------------------------------------------------

template <typename MODEL>
Variables<MODEL> LinearObsOperators<MODEL>::variables(
    const std::size_t jobs) const {
  Variables<MODEL> var(ops_.at(jobs)->variables());
  return var;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperators<MODEL>::print(std::ostream &os) const {
  for (std::size_t jobs = 0; jobs < ops_.size(); ++jobs) os << *ops_[jobs];
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_LINEAROBSOPERATORS_H_
