/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_MODELSPACECOVARIANCE4DBASE_H_
#define OOPS_BASE_MODELSPACECOVARIANCE4DBASE_H_

#include <map>
#include <memory>
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/config/LocalConfiguration.h"
#include "oops/base/AdvectorHandler.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/IncrCtlVec.h"
#include "oops/base/Increment4D.h"
#include "oops/base/PostBase.h"
#include "oops/base/State4D.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Variables.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace util {
class DateTime;
}

namespace oops {

// -----------------------------------------------------------------------------

// Should this be one with the ErrorCovariance class in the interface directory?
// YT

/// Abstract base class for model space error covariances.
template <typename MODEL>
class ModelSpaceCovariance4DBase {
  using Geometry_ = Geometry<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Advector_ = AdvectorHandler<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;

 public:
  ModelSpaceCovariance4DBase(const Geometry_ &,
                             const eckit::Configuration &config)
      : isAdvected_(config.has("advection")),
        covarianceModel_(config.getString("covariance", "none")),
        randomizationSize_(config.getInt("randomization size", 0)) {
    if (isAdvected_) {
      eckit::LocalConfiguration cfgAdv(config, "advection");
      adv_.reset(new Advector_(cfgAdv));
      Advector_::setup(adv_);
    }
  }

  virtual ~ModelSpaceCovariance4DBase() {
    if (isAdvected_) {
      std::shared_ptr<Advector_> noadv;
      Advector_::setup(noadv);
    }
  }

  void linearize(const State4D_ &xb, const Geometry_ &resol,
                 const eckit::Configuration &config) {
    advectedLinearize(xb, resol, config);
  }

  virtual void multiply(const Increment4D_ &dxi, Increment4D_ &dxo) const {
    if (isAdvected_) {
      Increment4D_ dxadv(dxi);
      adv_->mult_ad(dxi, dxadv);
      advectedMultiply(dxadv, dxo);
      adv_->mult(dxo, dxo);
    } else {
      advectedMultiply(dxi, dxo);
    }
  }

  virtual void multiplySqrt(const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
    advectedMultiplySqrt(dvi, dxo);
    if (isAdvected_) adv_->mult(dxo, dxo);
  }

  virtual void multiplySqrtTrans(const Increment4D_ &dxi,
                                 IncrCtlVec_ &dvo) const {
    if (isAdvected_) {
      Increment4D_ dxadv(dxi);
      adv_->mult_ad(dxi, dxadv);
      advectedMultiplySqrtTrans(dxadv, dvo);
    } else {
      advectedMultiplySqrtTrans(dxi, dvo);
    }
  }

  void inverseMultiply(const Increment4D_ &dxi, Increment4D_ &dxo) const {
    Increment4D_ dxadv(dxi);
    if (isAdvected_) adv_->mult_ad(dxi, dxadv);
    advectedInverseMultiply(dxadv, dxo);
    if (isAdvected_) adv_->mult(dxo, dxo);
  }

  virtual void randomize(Increment4D_ &) const {}

  virtual void advectedLinearize(const State4D_ &, const Geometry_ &,
                                 const eckit::Configuration &) = 0;
  virtual void advectedMultiply(const Increment4D_ &, Increment4D_ &) const = 0;
  virtual void advectedMultiplySqrt(const IncrCtlVec_ &,
                                    Increment4D_ &) const = 0;
  virtual void advectedMultiplySqrtTrans(const Increment4D_ &,
                                         IncrCtlVec_ &) const = 0;
  virtual void advectedInverseMultiply(const Increment4D_ &,
                                       Increment4D_ &) const = 0;
  virtual const ModelSpaceCovarianceBase<MODEL> &covar() const = 0;

  // Control Vector
  virtual IncrModCtlVec_ *newIncrModCtlVec() const = 0;
  virtual IncrEnsCtlVec_ *newIncrEnsCtlVec() const = 0;

  /// Covariance parameters
  const std::string covarianceModel() const { return covarianceModel_; }
  size_t randomizationSize() const { return randomizationSize_; }

 private:
  std::shared_ptr<Advector_> adv_;
  bool isAdvected_;
  std::string covarianceModel_;
  size_t randomizationSize_;
};

// -----------------------------------------------------------------------------

/// Covariance Factory
template <typename MODEL>
class Covariance4DFactory {
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;
  using State4D_ = State4D<MODEL>;

 public:
  static ModelSpaceCovariance4DBase<MODEL> *create(const eckit::Configuration &,
                                                   const Geometry_ &,
                                                   const Variables_ &,
                                                   const State4D_ &);
  virtual ~Covariance4DFactory() { makers_.clear(); }

 protected:
  explicit Covariance4DFactory(const std::string &);

 private:
  virtual ModelSpaceCovariance4DBase<MODEL> *make(const eckit::Configuration &,
                                                  const Geometry_ &,
                                                  const Variables_ &,
                                                  const State4D_ &) = 0;
  static std::map<std::string, Covariance4DFactory<MODEL> *> makers_;
};

// -----------------------------------------------------------------------------

template <class MODEL, class COVAR>
class Covar4DMaker : public Covariance4DFactory<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;
  using State4D_ = State4D<MODEL>;

  virtual ModelSpaceCovariance4DBase<MODEL> *make(
      const eckit::Configuration &conf, const Geometry_ &resol,
      const Variables_ &vars, const State4D_ &xb) {
    return new COVAR(resol, vars, conf, xb);
  }

 public:
  explicit Covar4DMaker(const std::string &name)
      : Covariance4DFactory<MODEL>(name) {}
};

// =============================================================================

template <typename MODEL>
std::map<std::string, Covariance4DFactory<MODEL> *>
    Covariance4DFactory<MODEL>::makers_;

// -----------------------------------------------------------------------------

template <typename MODEL>
Covariance4DFactory<MODEL>::Covariance4DFactory(const std::string &name) {
  if (makers_.find(name) != makers_.end()) {
    Log::error() << name << " already registered in covariance factory.";
    ABORT("Element already registered in Covariance4DFactory.");
  }
  makers_[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ModelSpaceCovariance4DBase<MODEL> *Covariance4DFactory<MODEL>::create(
    const eckit::Configuration &conf, const Geometry_ &resol,
    const Variables_ &vars, const State4D_ &xb) {
  const std::string id = conf.getString("covariance");
  Log::trace() << "ModelSpaceCovariance4DBase type = " << id;
  typename std::map<std::string, Covariance4DFactory<MODEL> *>::iterator j =
      makers_.find(id);
  if (j == makers_.end()) {
    Log::error() << id << " does not exist in covariance factory.";
    Log::error() << "Covariance factory contains " << makers_.size()
                 << " elements:";
    for (typename std::map<std::string,
                           Covariance4DFactory<MODEL> *>::const_iterator jj =
             makers_.begin();
         jj != makers_.end(); ++jj) {
      Log::error() << "A" << jj->first << "B";
    }
    ABORT("Element does not exist in Covariance4DFactory.");
  }
  return (*j).second->make(conf, resol, vars, xb);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_MODELSPACECOVARIANCE4DBASE_H_
