/*
 * (C) Copyright 2017-2019 UCAR
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 */

#ifndef OOPS_BASE_OBSAUXCONTROLS_H_
#define OOPS_BASE_OBSAUXCONTROLS_H_

#include <cmath>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/ObsAuxControl.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

/*! @brief Encapsulates a vector of ObsAuxControl objects.

 *   An object of ObsAuxControls class is part of ControlVariable. The class
 *  contains a vector of ObsAuxControl objects. Refer to the documentation of
 *  ControlVariable and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxControls : public util::Printable {
  using ObsAuxControl_ = ObsAuxControl<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxControls"; }

  /// Constructor
  ObsAuxControls(const ObsSpaces_ &, const eckit::Configuration &);
  /// Copy constructor
  ObsAuxControls(const ObsAuxControls &, const bool copy = true);
  /// Destructor
  ~ObsAuxControls();

  // Access
  /// Return the size
  std::size_t size() const { return auxs_.size(); }
  /// Return a constant reference to the i-th ObsAuxControl object
  const ObsAuxControl_ &operator[](const std::size_t ii) const {
    return *auxs_.at(ii);
  }
  /// Return a reference to the i-th ObsAuxControl object
  ObsAuxControl_ &operator[](const std::size_t ii) { return *auxs_.at(ii); }
  /// Return a constant reference to ObsSpaces
  const ObsSpaces_ &obspaces() const { return *odb_; }

  /// Reset
  void reset(const ObsSpaces_ &, const std::string &,
             const eckit::Configuration &);

  // I/O and diagnostics
  /// Read control variable
  void read(const eckit::Configuration &);
  /// Write control variable
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

 private:
  ObsAuxControls &operator=(const ObsAuxControls &);
  void print(std::ostream &) const;

  const ObsSpaces_ mutable *odb_;
  /// Vector of shared pointers to ObsAuxControl objects
  std::vector<std::shared_ptr<ObsAuxControl_> > auxs_;
};

// =============================================================================

template <typename MODEL>
ObsAuxControls<MODEL>::ObsAuxControls(const ObsSpaces_ &odb,
                                      const eckit::Configuration &conf)
    : odb_(), auxs_(0) {
  std::vector<eckit::LocalConfiguration> obsconf;
  odb_ = &odb;
  conf.get("ObsTypes", obsconf);
  for (std::size_t jobs = 0; jobs < obsconf.size(); ++jobs) {
    std::shared_ptr<ObsAuxControl_> tmp(
        new ObsAuxControl_(odb[jobs], obsconf[jobs]));
    auxs_.push_back(tmp);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxControls<MODEL>::ObsAuxControls(const ObsAuxControls &other,
                                      const bool copy)
    : odb_(other.odb_), auxs_(other.size()) {
  Log::trace() << "ObsAuxControls<MODEL>::ObsAuxControls copy starting"
               << std::endl;
  for (std::size_t jobs = 0; jobs < other.size(); ++jobs) {
    auxs_[jobs].reset(new ObsAuxControl_(other[jobs], copy));
  }
  Log::trace() << "ObsAuxControls<MODEL>::ObsAuxControls copy done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxControls<MODEL>::~ObsAuxControls() {
  Log::trace() << "ObsAuxControls<MODEL>::~ObsAuxControls starting"
               << std::endl;
  for (std::size_t jobs = 0; jobs < auxs_.size(); ++jobs) auxs_[jobs].reset();
  odb_ = nullptr;
  Log::trace() << "ObsAuxControls<MODEL>::~ObsAuxControls done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxControls<MODEL>::reset(const ObsSpaces_ &os,
                                  const std::string &stateid,
                                  const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxControls<MODEL>::reset starting" << std::endl;
  odb_ = &os;
  std::vector<eckit::LocalConfiguration> obsconfs;
  if (conf.has("screening") && conf.has("statesObsAuxReset")) {
    if (conf.getBool("screening.active")) {
      conf.get("statesObsAuxReset", obsconfs);
    }
  }
  ASSERT(auxs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs) {
    auxs_[jobs]->reset(os[jobs], stateid, obsconfs[jobs]);
  }
  Log::trace() << "ObsAuxControls<MODEL>::reset done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxControls<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxControls<MODEL>::read starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconfs;
  conf.get("statesObsAux", obsconfs);
  ASSERT(auxs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs)
    auxs_[jobs]->read(obsconfs[jobs]);
  Log::trace() << "ObsAuxControls<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxControls<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "ObsAuxControls<MODEL>::write starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconfs;
  conf.get("statesObsAux", obsconfs);
  ASSERT(auxs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs)
    auxs_[jobs]->write(obsconfs[jobs]);
  Log::trace() << "ObsAuxControls<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double ObsAuxControls<MODEL>::norm() const {
  Log::trace() << "ObsAuxControls<MODEL>::norm starting" << std::endl;
  double zz = static_cast<double>(0.0);
  for (std::size_t jobs = 0; jobs < auxs_.size(); ++jobs) {
    double norm = auxs_[jobs]->norm();
    zz += norm * norm;
  }
  Log::trace() << "ObsAuxControls<MODEL>::norm done" << std::endl;
  return std::sqrt(zz);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxControls<MODEL>::print(std::ostream &os) const {
  for (std::size_t jobs = 0; jobs < auxs_.size(); ++jobs) os << *auxs_[jobs];
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSAUXCONTROLS_H_
