/*
 * (C) Copyright 2017-2019 UCAR
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 */

#ifndef OOPS_BASE_OBSAUXCOVARIANCES_H_
#define OOPS_BASE_OBSAUXCOVARIANCES_H_

#include <cmath>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObsAuxCtlVecs.h"
#include "oops/base/ObsAuxIncrements.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/ObsAuxCovariance.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

/*! @brief Encapsulates a vector of ObsAuxCovariance objects.

 *   An object of ObsAuxCovariances class is part of CostJbTotal. The class
 *  contains a vector of ObsAuxCovariance objects. Refer to the documentation of
 *  ControlVariable, CostJbTotal and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCovariances : public util::Printable, private eckit::NonCopyable {
  using ObsAuxCovariance_ = ObsAuxCovariance<MODEL>;
  using ObsAuxControls_ = ObsAuxControls<MODEL>;
  using ObsAuxCtlVecs_ = ObsAuxCtlVecs<MODEL>;
  using ObsAuxIncrements_ = ObsAuxIncrements<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxCovariances"; }

  /// Constructor
  ObsAuxCovariances(const ObsSpaces_ &, const eckit::Configuration &);
  /// Destructor
  ~ObsAuxCovariances();

  // Access
  /// Return the size
  std::size_t size() const { return covs_.size(); }
  /// Return a constant reference to the i-th ObsAuxCovariance object
  const ObsAuxCovariance_ &operator[](const std::size_t ii) const {
    return *covs_.at(ii);
  }
  /// Return a reference to the i-th ObsAuxCovariance object
  ObsAuxCovariance_ &operator[](const std::size_t ii) { return *covs_.at(ii); }

  /// Update
  void update(const ObsAuxControls_ &, const eckit::Configuration &);

  // Operators
  /// Set up the inner loop resolution B matrix
  void linearize(const ObsAuxControls_ &, const ObsSpaces_ &);
  /// \f$ dx_1 = B dx_2 \f$
  void multiply(const ObsAuxIncrements_ &, ObsAuxIncrements_ &) const;
  /// \f$ dx_1 = B^{-1} dx_2 \f$
  void inverseMultiply(const ObsAuxIncrements_ &, ObsAuxIncrements_ &) const;
  /// \f$ dx = B^{1/2} dv \f$
  void multiplySqrt(const ObsAuxCtlVecs_ &, ObsAuxIncrements_ &) const;
  /// \f$ dv = [B^{1/2}]^T dx \f$
  void multiplySqrtTrans(const ObsAuxIncrements_ &, ObsAuxCtlVecs_ &) const;
  /// \f$ dx = B^{1/2} dv_{random} \f$
  void randomize(ObsAuxIncrements_ &) const;

  /// Retrun a constant reference to configuration
  const eckit::LocalConfiguration &config() const { return conf_; }
  /// Retrun a constant reference to ObsSpaces
  const ObsSpaces_ &obspaces() const { return *odb_; }

 private:
  void print(std::ostream &) const;

  /// Vector of shared pointers to ObsAuxCovariance objects
  std::vector<std::shared_ptr<ObsAuxCovariance_> > covs_;
  const ObsSpaces_ mutable *odb_;
  const eckit::LocalConfiguration conf_;
};

// =============================================================================

template <typename MODEL>
ObsAuxCovariances<MODEL>::ObsAuxCovariances(const ObsSpaces_ &odb,
                                            const eckit::Configuration &conf)
    : covs_(0), odb_(), conf_(conf) {
  Log::trace() << "ObsAuxCovariances<MODEL>::ObsAuxCovariances starting"
               << std::endl;
  odb_ = &odb;
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("ObsTypes", obsconf);
  for (std::size_t jobs = 0; jobs < obsconf.size(); ++jobs) {
    std::shared_ptr<ObsAuxCovariance_> tmp(
        new ObsAuxCovariance_(odb[jobs], obsconf[jobs]));
    covs_.push_back(tmp);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::ObsAuxCovariances done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCovariances<MODEL>::~ObsAuxCovariances() {
  Log::trace() << "ObsAuxCovariances<MODEL>::~ObsAuxCovariances starting"
               << std::endl;
  odb_ = nullptr;
  for (std::size_t jobs = 0; jobs < covs_.size(); ++jobs) {
    covs_[jobs].reset();
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::~ObsAuxCovariances done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::update(const ObsAuxControls_ &xx,
                                      const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxCovariances<MODEL>::update starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("covariancesObsAux", obsconf);
  ASSERT(covs_.size() == xx.size());
  for (std::size_t jobs = 0; jobs < xx.size(); ++jobs) {
    covs_[jobs]->update(xx[jobs], obsconf[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::update done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::linearize(const ObsAuxControls_ &xx,
                                         const ObsSpaces_ &odb) {
  Log::trace() << "ObsAuxCovariances<MODEL>::linearize starting" << std::endl;
  odb_ = &odb;
  ASSERT(covs_.size() == xx.size());
  for (std::size_t jobs = 0; jobs < xx.size(); ++jobs) {
    covs_[jobs]->linearize(xx[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::multiply(const ObsAuxIncrements_ &dx1,
                                        ObsAuxIncrements_ &dx2) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::multiply starting" << std::endl;
  ASSERT(dx1.size() == dx2.size() && covs_.size() == dx1.size());
  for (std::size_t jobs = 0; jobs < dx1.size(); ++jobs) {
    covs_[jobs]->multiply(dx1[jobs], dx2[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::inverseMultiply(const ObsAuxIncrements_ &dx1,
                                               ObsAuxIncrements_ &dx2) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::inverseMultiply starting"
               << std::endl;
  ASSERT(dx1.size() == dx2.size() && covs_.size() == dx1.size());
  for (std::size_t jobs = 0; jobs < dx1.size(); ++jobs) {
    covs_[jobs]->inverseMultiply(dx1[jobs], dx2[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::inverseMultiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::multiplySqrt(const ObsAuxCtlVecs_ &dv,
                                            ObsAuxIncrements_ &dx) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrt starting"
               << std::endl;
  ASSERT(dv.size() == dx.size() && covs_.size() == dx.size());
  for (std::size_t jobs = 0; jobs < dx.size(); ++jobs) {
    covs_[jobs]->multiplySqrt(dv[jobs], dx[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrt done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::multiplySqrtTrans(const ObsAuxIncrements_ &dx,
                                                 ObsAuxCtlVecs_ &dv) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrtTrans starting"
               << std::endl;
  ASSERT(dv.size() == dx.size() && covs_.size() == dx.size());
  for (std::size_t jobs = 0; jobs < dx.size(); ++jobs) {
    covs_[jobs]->multiplySqrtTrans(dx[jobs], dv[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::multiplySqrtTrans done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::randomize(ObsAuxIncrements_ &dx) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::randomize starting" << std::endl;
  ASSERT(covs_.size() == dx.size());
  for (std::size_t jobs = 0; jobs < dx.size(); ++jobs) {
    covs_[jobs]->randomize(dx[jobs]);
  }
  Log::trace() << "ObsAuxCovariances<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariances<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObsAuxCovariances<MODEL>::print starting" << std::endl;
  for (std::size_t jobs = 0; jobs < covs_.size(); ++jobs)
    os << *covs_.at(jobs) << " ";
  Log::trace() << "ObsAuxCovariances<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSAUXCOVARIANCES_H_
