/*
 * (C) Copyright 2017-2019 UCAR
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 */

#ifndef OOPS_BASE_OBSAUXCTLVECS_H_
#define OOPS_BASE_OBSAUXCTLVECS_H_

#include <cmath>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include "oops/base/ObsAuxControls.h"
#include "oops/interface/ObsAuxCtlVec.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

template <typename MODEL>
class ObsAuxCovariances;

/*! @brief Encapsulates a vector of ObsAuxCtlVec objects.

 *   An object of ObsAuxCtlVecs class is part of ControlVector. The class
 *  contains a vector of ObsAuxCtlVec objects. Refer to the documentation of
 *  ControlVector and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCtlVecs : public util::Printable {
  using ObsAuxCtlVec_ = ObsAuxCtlVec<MODEL>;
  using ObsAuxCovariances_ = ObsAuxCovariances<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxCtlVecs"; }

  /// Constructor
  explicit ObsAuxCtlVecs(const ObsAuxCovariances_ &);
  /// Constructor with copy initialization
  ObsAuxCtlVecs(const ObsAuxCovariances_ &, const ObsAuxCtlVecs &);
  /// Copy constructor
  ObsAuxCtlVecs(const ObsAuxCtlVecs &, const bool copy = true);
  /// Destructor
  ~ObsAuxCtlVecs();

  // Access
  /// Return a constant reference to the i-th ObsAuxControl object
  const ObsAuxCtlVec_ &operator[](const std::size_t ii) const {
    return *ctlvecs_.at(ii);
  }
  /// Return a reference to the i-th ObsAuxControl object
  ObsAuxCtlVec_ &operator[](const std::size_t ii) { return *ctlvecs_.at(ii); }
  /// Return the size
  std::size_t size() const { return ctlvecs_.size(); }

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  void zero();
  /// \f$ dv_1 = dv_2 \f$
  ObsAuxCtlVecs &operator=(const ObsAuxCtlVecs &);
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  ObsAuxCtlVecs &operator+=(const ObsAuxCtlVecs &);
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  ObsAuxCtlVecs &operator-=(const ObsAuxCtlVecs &);
  /// \f$ dv_1 = dv_1 * c \f$
  ObsAuxCtlVecs &operator*=(const double &);
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  void axpy(const double &, const ObsAuxCtlVecs &);
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  double dot_product_with(const ObsAuxCtlVecs &) const;

  // I/O and diagnostics
  /// Read control vector
  void read(const eckit::Configuration &);
  /// Write control vector
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

 private:
  void print(std::ostream &) const;

  /// Vector of shared pointers to ObsAuxCtlVec objects
  std::vector<std::shared_ptr<ObsAuxCtlVec_> > ctlvecs_;
};

// =============================================================================

template <typename MODEL>
ObsAuxCtlVecs<MODEL>::ObsAuxCtlVecs(const ObsAuxCovariances_ &covs)
    : ctlvecs_(0) {
  for (std::size_t jobs = 0; jobs < covs.size(); ++jobs) {
    std::shared_ptr<ObsAuxCtlVec_> tmp(new ObsAuxCtlVec_(covs[jobs]));
    ctlvecs_.push_back(tmp);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL>::ObsAuxCtlVecs(const ObsAuxCovariances_ &covs,
                                    const ObsAuxCtlVecs &other)
    : ctlvecs_(0) {
  ASSERT(covs.size() == other.size());
  for (std::size_t jobs = 0; jobs < covs.size(); ++jobs) {
    std::shared_ptr<ObsAuxCtlVec_> tmp(
        new ObsAuxCtlVec_(covs[jobs], other[jobs]));
    ctlvecs_.push_back(tmp);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL>::ObsAuxCtlVecs(const ObsAuxCtlVecs &other, const bool copy)
    : ctlvecs_(other.size()) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::ObsAuxCtlVecs copy starting"
               << std::endl;
  ASSERT(size() == other.size());
  for (std::size_t jobs = 0; jobs < other.size(); ++jobs) {
    ctlvecs_[jobs].reset(new ObsAuxCtlVec_(other[jobs], copy));
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::ObsAuxCtlVecs copy done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL>::~ObsAuxCtlVecs() {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::~ObsAuxCtlVecs starting" << std::endl;
  for (std::size_t jobs = 0; jobs < ctlvecs_.size(); ++jobs)
    ctlvecs_[jobs].reset();
  Log::trace() << "ObsAuxCtlVecs<MODEL>::~ObsAuxCtlVecs done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVecs<MODEL>::zero() {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::zero starting" << std::endl;
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    ctlvecs_[jobs]->zero();
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL> &ObsAuxCtlVecs<MODEL>::operator=(
    const ObsAuxCtlVecs &rhs) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator= starting" << std::endl;
  ASSERT(size() == rhs.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *ctlvecs_[jobs] = rhs[jobs];
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL> &ObsAuxCtlVecs<MODEL>::operator+=(
    const ObsAuxCtlVecs &rhs) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator+= starting" << std::endl;
  ASSERT(size() == rhs.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *ctlvecs_[jobs] += rhs[jobs];
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL> &ObsAuxCtlVecs<MODEL>::operator-=(
    const ObsAuxCtlVecs &rhs) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator-= starting" << std::endl;
  ASSERT(size() == rhs.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *ctlvecs_[jobs] -= rhs[jobs];
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecs<MODEL> &ObsAuxCtlVecs<MODEL>::operator*=(const double &zz) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator*= starting" << std::endl;
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *ctlvecs_[jobs] *= zz;
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVecs<MODEL>::axpy(const double &zz, const ObsAuxCtlVecs &dx) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::axpy starting" << std::endl;
  ASSERT(size() == dx.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    ctlvecs_[jobs]->axpy(zz, dx[jobs]);
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxCtlVecs<MODEL>::dot_product_with(const ObsAuxCtlVecs &dx) const {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::dot_product_with starting"
               << std::endl;
  ASSERT(size() == dx.size());
  double zz = static_cast<double>(0.0);
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    zz += ctlvecs_[jobs]->dot_product_with(dx[jobs]);
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::dot_product_with done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVecs<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::read starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconfs;
  conf.get("ctlvecsObsAux", obsconfs);
  ASSERT(ctlvecs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs)
    ctlvecs_[jobs]->read(obsconfs[jobs]);
  Log::trace() << "ObsAuxCtlVecs<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVecs<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::write starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconfs;
  conf.get("ctlvecsObsAux", obsconfs);
  ASSERT(ctlvecs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs)
    ctlvecs_[jobs]->write(obsconfs[jobs]);
  Log::trace() << "ObsAuxCtlVecs<MODEL>::write done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxCtlVecs<MODEL>::norm() const {
  Log::trace() << "ObsAuxCtlVecs<MODEL>::norm starting" << std::endl;
  double zz = static_cast<double>(0.0);
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    double norm = ctlvecs_[jobs]->norm();
    zz += norm * norm;
  }
  Log::trace() << "ObsAuxCtlVecs<MODEL>::norm done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVecs<MODEL>::print(std::ostream &os) const {
  for (std::size_t jobs = 0; jobs < ctlvecs_.size(); ++jobs)
    os << *ctlvecs_[jobs];
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSAUXCTLVECS_H_
