/*
 * (C) Copyright 2017-2019 UCAR
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 */

#ifndef OOPS_BASE_OBSAUXINCREMENTS_H_
#define OOPS_BASE_OBSAUXINCREMENTS_H_

#include <cmath>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

/*! @brief Encapsulates a vector of ObsAuxIncrement objects.

 *   An object of ObsAuxIncrements class is part of ControlIncrement. The class
 *  contains a vector of ObsAuxIncrement bjects. Refer to the documentation of
 *  ControlIncrement and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxIncrements : public util::Printable {
  using ObsAuxIncrement_ = ObsAuxIncrement<MODEL>;
  using ObsAuxControls_ = ObsAuxControls<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxIncrements"; }

  /// Constructor
  ObsAuxIncrements(const ObsSpaces_ &, const eckit::Configuration &);
  /// Copy constructor
  ObsAuxIncrements(const ObsAuxIncrements &, const bool copy = true);
  /// Copy constructor
  ObsAuxIncrements(const ObsAuxIncrements &, const eckit::Configuration &);
  /// Destructor
  ~ObsAuxIncrements();

  // Access
  /// Return the size
  std::size_t size() const { return auxs_.size(); }
  /// Return a constant reference to the i-th ObsAuxIncrement object
  const ObsAuxIncrement_ &operator[](const std::size_t ii) const {
    return *auxs_.at(ii);
  }
  /// Return a reference to the i-th ObsAuxIncrement object
  ObsAuxIncrement_ &operator[](const std::size_t ii) { return *auxs_.at(ii); }

  // Linear algebra operators
  /// \f$ dx = x_1 - x_2  \f$
  void diff(const ObsAuxControls_ &, const ObsAuxControls_ &);
  /// \f$ dx_1 = 0 \f$
  void zero();
  /// \f$ dx_1 = dx_2 \f$
  ObsAuxIncrements &operator=(const ObsAuxIncrements &);
  /// \f$ dx_1 = dx_1 + dx_2 \f$
  ObsAuxIncrements &operator+=(const ObsAuxIncrements &);
  /// \f$ dx_1 = dx_1 - dx_2 \f$
  ObsAuxIncrements &operator-=(const ObsAuxIncrements &);
  /// \f$ dx_1 = dx_1 * c \f$
  ObsAuxIncrements &operator*=(const double &);
  /// \f$ dx_1 = dx_1 * c dx_2 \f$
  void axpy(const double &, const ObsAuxIncrements &);
  /// \f$ <dx_1,dx_1> = dx_1^T dx_1 \f$
  double dot_product_with(const ObsAuxIncrements &) const;

  // I/O and diagnostics
  /// Read control increment
  void read(const eckit::Configuration &);
  /// Write control increment
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

 private:
  void print(std::ostream &) const;

  /// Vector of shared pointers to ObsAuxIncrement objects
  std::vector<std::shared_ptr<ObsAuxIncrement_> > auxs_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxControls<MODEL> &operator+=(ObsAuxControls<MODEL> &xx,
                                  const ObsAuxIncrements<MODEL> &dx) {
  Log::trace() << "operator+=(ObsAuxControls, ObsAuxIncrements) starting"
               << std::endl;
  ASSERT(xx.size() == dx.size());
  for (std::size_t jobs = 0; jobs < xx.size(); ++jobs) {
    xx[jobs] += dx[jobs];
  }
  Log::trace() << "operator+=(ObsAuxControls, ObsAuxIncrements) done"
               << std::endl;
  return xx;
}

// =============================================================================

template <typename MODEL>
ObsAuxIncrements<MODEL>::ObsAuxIncrements(const ObsSpaces_ &odb,
                                          const eckit::Configuration &conf)
    : auxs_(0) {
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("ObsTypes", obsconf);
  for (std::size_t jobs = 0; jobs < obsconf.size(); ++jobs) {
    std::shared_ptr<ObsAuxIncrement_> tmp(
        new ObsAuxIncrement_(odb[jobs], obsconf[jobs]));
    auxs_.push_back(tmp);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL>::ObsAuxIncrements(const ObsAuxIncrements &other,
                                          const bool copy)
    : auxs_(other.size()) {
  Log::trace() << "ObsAuxIncrements<MODEL>::ObsAuxIncrements starting"
               << std::endl;
  ASSERT(size() == other.size());
  for (std::size_t jobs = 0; jobs < other.size(); ++jobs) {
    auxs_[jobs].reset(new ObsAuxIncrement_(other[jobs], copy));
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::ObsAuxIncrements done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL>::ObsAuxIncrements(const ObsAuxIncrements &other,
                                          const eckit::Configuration &conf)
    : auxs_(other.size()) {
  Log::trace() << "ObsAuxIncrements<MODEL>::ObsAuxIncrements starting"
               << std::endl;
  std::vector<eckit::LocalConfiguration> obsconf;
  ASSERT(size() == other.size());
  for (std::size_t jobs = 0; jobs < other.size(); ++jobs) {
    auxs_[jobs].reset(new ObsAuxIncrement_(other[jobs]));
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::ObsAuxIncrements done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL>::~ObsAuxIncrements() {
  Log::trace() << "ObsAuxIncrements<MODEL>::~ObsAuxIncrements starting"
               << std::endl;
  for (std::size_t jobs = 0; jobs < auxs_.size(); ++jobs) auxs_[jobs].reset();
  Log::trace() << "ObsAuxIncrements<MODEL>::~ObsAuxIncrements done"
               << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrements<MODEL>::diff(const ObsAuxControls_ &x1,
                                   const ObsAuxControls_ &x2) {
  Log::trace() << "ObsAuxIncrements<MODEL>::diff starting" << std::endl;
  ASSERT(x1.size() == x2.size() && size() == x2.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    auxs_[jobs]->diff(x1[jobs], x2[jobs]);
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::diff done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrements<MODEL>::zero() {
  Log::trace() << "ObsAuxIncrements<MODEL>::zero starting" << std::endl;
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    auxs_[jobs]->zero();
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL> &ObsAuxIncrements<MODEL>::operator=(
    const ObsAuxIncrements &rhs) {
  Log::trace() << "ObsAuxIncrements<MODEL>::operator= starting" << std::endl;
  ASSERT(size() == rhs.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *auxs_[jobs] = rhs[jobs];
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL> &ObsAuxIncrements<MODEL>::operator+=(
    const ObsAuxIncrements &rhs) {
  Log::trace() << "ObsAuxIncrements<MODEL>::operator+= starting" << std::endl;
  ASSERT(size() == rhs.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *auxs_[jobs] += rhs[jobs];
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL> &ObsAuxIncrements<MODEL>::operator-=(
    const ObsAuxIncrements &rhs) {
  Log::trace() << "ObsAuxIncrements<MODEL>::operator-= starting" << std::endl;
  ASSERT(size() == rhs.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *auxs_[jobs] -= rhs[jobs];
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrements<MODEL> &ObsAuxIncrements<MODEL>::operator*=(const double &zz) {
  Log::trace() << "ObsAuxIncrements<MODEL>::operator*= starting" << std::endl;
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    *auxs_[jobs] *= zz;
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrements<MODEL>::axpy(const double &zz,
                                   const ObsAuxIncrements &dx) {
  Log::trace() << "ObsAuxIncrements<MODEL>::axpy starting" << std::endl;
  ASSERT(size() == dx.size());
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    auxs_[jobs]->axpy(zz, dx[jobs]);
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxIncrements<MODEL>::dot_product_with(
    const ObsAuxIncrements &dx) const {
  Log::trace() << "ObsAuxIncrements<MODEL>::dot_product_with starting"
               << std::endl;
  ASSERT(size() == dx.size());
  double zz = static_cast<double>(0.0);
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    zz += auxs_[jobs]->dot_product_with(dx[jobs]);
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::dot_product_with done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrements<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxIncrements<MODEL>::read starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconfs;
  conf.get("incrementsObsAux", obsconfs);
  ASSERT(auxs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs)
    auxs_[jobs]->read(obsconfs[jobs]);
  Log::trace() << "ObsAuxIncrements<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrements<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "ObsAuxIncrements<MODEL>::write starting" << std::endl;
  std::vector<eckit::LocalConfiguration> obsconfs;
  conf.get("incrementsObsAux", obsconfs);
  ASSERT(auxs_.size() >= obsconfs.size());
  for (std::size_t jobs = 0; jobs < obsconfs.size(); ++jobs)
    auxs_[jobs]->write(obsconfs[jobs]);
  Log::trace() << "ObsAuxIncrements<MODEL>::write done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxIncrements<MODEL>::norm() const {
  Log::trace() << "ObsAuxIncrements<MODEL>::norm starting" << std::endl;
  double zz = static_cast<double>(0.0);
  for (std::size_t jobs = 0; jobs < size(); ++jobs) {
    double norm = auxs_[jobs]->norm();
    zz += norm * norm;
  }
  Log::trace() << "ObsAuxIncrements<MODEL>::norm done" << std::endl;
  return std::sqrt(zz);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrements<MODEL>::print(std::ostream &os) const {
  for (std::size_t jobs = 0; jobs < auxs_.size(); ++jobs) os << *auxs_[jobs];
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSAUXINCREMENTS_H_
