/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERRORS_H_
#define OOPS_BASE_OBSERRORS_H_

#include <memory>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlObsVector.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/base/Departures.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/Observations.h"
#include "oops/interface/ObsErrorCovariance.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

/*! @brief Encapsulates a vector of ObsErrorCovariance objects.

 *   CostJo holds an ObsErrors object. Refer to the documentation of CostJo
 *  to understand the structure. Recall we may have multiple ObservationsSpace
 *  objects each specific to particular observation types. To that end
 *  ObservationSpaces class holds a vector of ObservationSpace objects.
 *  ObsErrors class wraps their corresponding ObsErrorCovariance matrices.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsErrors : public util::Printable, private eckit::NonCopyable {
  using ControlObsVector_ = ControlObsVector<MODEL>;
  using CtrlVar_ = ControlVariable<MODEL>;
  using Departures_ = Departures<MODEL>;
  using ObsAuxCtrls_ = ObsAuxControls<MODEL>;
  using Observations_ = Observations<MODEL>;
  using ObsError_ = ObsErrorCovariance<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsErrors"; }

  /// Constructor
  explicit ObsErrors(const ObsSpaces_ &);
  /// Destructor
  ~ObsErrors();

  /// Return the number of covariances
  std::size_t size() const { return err_.size(); }
  /// Return a constant reference to i-th covariance
  const ObsError_ &operator[](const std::size_t ii) const {
    return *err_.at(ii);
  }

  /// Setup the covariance matrix
  void linearize(const Observations_ &);

  /// Setup the covariance matrix
  void linearize(const ObsAuxCtrls_ &, const Observations_ &);

  /// Setup VarQC weights
  void setupWeights(const Departures_ &);

  /// Multiply a Departure by \f$R\f$
  Departures_ *multiply(const Departures_ &) const;
  /// Multiply a Departure by \f$R^{-1}\f$
  Departures_ *inverseMultiply(const Departures_ &) const;

  /// Generate random perturbation
  void randomize(Departures_ &) const;

 private:
  void print(std::ostream &) const;

  /// Vector of shared pointers to ObsErrorCovariance objects
  std::vector<std::shared_ptr<ObsError_> > err_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsErrors<MODEL>::ObsErrors(const ObsSpaces_ &os) : err_(0) {
  for (std::size_t jj = 0; jj < os.size(); ++jj) {
    eckit::LocalConfiguration conf(os[jj].config(), "Covariance");
    std::shared_ptr<ObsError_> tmp(new ObsError_(os[jj], conf));
    err_.push_back(tmp);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsErrors<MODEL>::~ObsErrors() {}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrors<MODEL>::linearize(const Observations_ &yy) {
  for (std::size_t jj = 0; jj < err_.size(); ++jj) {
    err_[jj]->linearize(yy[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrors<MODEL>::linearize(const ObsAuxCtrls_ &xx,
                                 const Observations_ &yy) {
  for (std::size_t jj = 0; jj < err_.size(); ++jj) {
    err_[jj]->linearize(xx[jj], yy[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrors<MODEL>::setupWeights(const Departures_ &dy) {
  for (std::size_t jj = 0; jj < err_.size(); ++jj) {
    err_[jj]->setupWeights(dy[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Departures<MODEL> *ObsErrors<MODEL>::multiply(const Departures_ &dy) const {
  std::vector<std::shared_ptr<ControlObsVector_> > ovec;
  for (std::size_t jj = 0; jj < err_.size(); ++jj) {
    std::shared_ptr<ControlObsVector_> tmp(err_[jj]->multiply(dy[jj]));
    ovec.push_back(tmp);
  }
  return new Departures_(ovec);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Departures<MODEL> *ObsErrors<MODEL>::inverseMultiply(
    const Departures_ &dy) const {
  std::vector<std::shared_ptr<ControlObsVector_> > ovec;
  for (std::size_t jj = 0; jj < err_.size(); ++jj) {
    std::shared_ptr<ControlObsVector_> tmp(err_[jj]->inverseMultiply(dy[jj]));
    ovec.push_back(tmp);
  }
  return new Departures_(ovec);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrors<MODEL>::randomize(Departures_ &dy) const {
  for (std::size_t jj = 0; jj < err_.size(); ++jj) {
    err_[jj]->randomize(dy[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrors<MODEL>::print(std::ostream &os) const {
  for (std::size_t jj = 0; jj < err_.size(); ++jj) os << *err_[jj];
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERRORS_H_
