/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVATIONSPACES_H_
#define OOPS_BASE_OBSERVATIONSPACES_H_

#include <cstddef>
#include <memory>
#include <ostream>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ObservationSpace.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {
template <typename T>
class Departures;
template <typename T>
class GeoVaLs;

// -----------------------------------------------------------------------------

/*! @brief Wrapper around a vector of observation spaces
 */

template <typename MODEL>
class ObservationSpaces
    : public util::Printable,
      private eckit::NonCopyable,
      private util::ObjectCounter<ObservationSpaces<MODEL>> {
  using Departures_ = Departures<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using ObsSpace_ = ObservationSpace<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using GeoVaLsVec_ = std::vector<std::shared_ptr<GeoVaLs_>>;

 public:
  static const std::string classname() { return "oops::ObservationSpaces"; }

  ObservationSpaces(const eckit::Configuration &, const Geometry_ &,
                    const util::DateTime &, const util::DateTime &,
                    const bool lscreened = false);
  ~ObservationSpaces();

  /// Access
  std::size_t size() const { return spaces_.size(); }
  const ObsSpace_ &operator[](const std::size_t ii) const {
    return *spaces_.at(ii);
  }
  ObsSpace_ &operator[](const std::size_t ii) { return *spaces_.at(ii); }

  /// Assimilation window
  const util::DateTime &windowStart() const { return wbgn_; }
  const util::DateTime &windowEnd() const { return wend_; }

  /// Screen observations
  void screenObservations(const Departures_ &, const GeoVaLsVec_ &) const;

  /// Save observations
  void saveObservations() const;

  /// Other
  void printJo(const Departures_ &, const Departures_ &) const;

 private:
  void print(std::ostream &) const;

  const util::DateTime wbgn_;
  const util::DateTime wend_;
  const bool lscreened_;
  std::vector<std::shared_ptr<ObsSpace_>> spaces_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObservationSpaces<MODEL>::ObservationSpaces(const eckit::Configuration &conf,
                                            const Geometry_ &geom,
                                            const util::DateTime &bgn,
                                            const util::DateTime &end,
                                            const bool lscreened)
    : wbgn_(bgn), wend_(end), lscreened_(lscreened), spaces_(0) {
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("ObsTypes", obsconf);
  for (std::size_t jj = 0; jj < obsconf.size(); ++jj) {
    Log::debug() << "ObservationSpaces::ObservationSpaces : conf "
                 << obsconf[jj] << std::endl;
    std::shared_ptr<ObsSpace_> tmp(
        new ObsSpace_(obsconf[jj], geom, bgn, end, lscreened));
    spaces_.push_back(tmp);
    //  Generate locations etc... if required
    if (obsconf[jj].has("Generate")) {
      const eckit::LocalConfiguration gconf(obsconf[jj], "Generate");
      spaces_[jj]->generateDistribution(gconf);
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObservationSpaces<MODEL>::~ObservationSpaces() {}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpaces<MODEL>::screenObservations(
    const Departures_ &dep, const GeoVaLsVec_ &geovals) const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    spaces_[jj]->screenObservations(dep[jj].obsvector(), *geovals[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpaces<MODEL>::saveObservations() const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    spaces_[jj]->saveObservations();
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpaces<MODEL>::printJo(const Departures_ &dy,
                                       const Departures_ &grad) const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    spaces_[jj]->printJo(dy[jj].obsvector(), grad[jj].obsvector());
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpaces<MODEL>::print(std::ostream &os) const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    os << *spaces_[jj];
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVATIONSPACES_H_
