/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVATIONS_H_
#define OOPS_BASE_OBSERVATIONS_H_

#include <cstddef>
#include <memory>
#include <ostream>
#include <string>
#include <vector>

#include "eckit/config/Configuration.h"
#include "oops/assimilation/ControlObsVector.h"
#include "oops/base/Departures.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/interface/ObsOperator.h"
#include "oops/interface/ObsVector.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/Printable.h"
#include "util/abor1_cpp.h"

namespace oops {

/*! @brief Observations Class.

 *   Observations class contains observed values or their model equivalents.
 *  It contains a vector of ControlObsVector objects and has the same underlying
 *  structure as Departures class.

    \f[ y = \begin{bmatrix}
              \bar{y}_1  \\
              ..         \\
              \bar{y}_L
            \end{bmatrix}
          = \begin{bmatrix}
              \tilde{y}_1  \\
              \tilde{b}_1  \\
              ..           \\
              \tilde{y}_L  \\
              \tilde{b}_L
            \end{bmatrix}
     \f]

 *   where \f$ L \f$ is the number of ObservationSpace objects held by
 *  ObservationSpaces, \f$  \bar{y}_j \f$ is implemented by ControlObsVector and
    has two components: observation \f$ \tilde{y}_j \f$ and bias
    \f$ \tilde{b}_j \f$ vectors (each implemented by ObsVector).

 *   Refer to the documentation of CostJo to better understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class Observations : public util::Printable {
  using ControlObsVector_ = ControlObsVector<MODEL>;
  using Departures_ = Departures<MODEL>;
  using ObsOperator_ = ObsOperator<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;
  using ObsVector_ = ObsVector<MODEL>;

 public:
  /// Constructor
  explicit Observations(const ObsSpaces_ &);
  /// Copy constructor
  explicit Observations(const Observations &);
  /// Destructor
  ~Observations();
  /// Assignment
  Observations &operator=(const Observations &);

  // Access
  /// Return the number of components
  std::size_t size() const { return obs_.size(); }
  /// Return a constant reference to i-th component
  const ControlObsVector_ &operator[](const std::size_t ii) const {
    return *obs_.at(ii);
  }
  /// Return a reference to i-th component
  ControlObsVector_ &operator[](const std::size_t ii) { return *obs_.at(ii); }

  // Interactions with Departures
  std::vector<std::shared_ptr<ControlObsVector_> > operator-(
      const Observations &other) const;
  /// \f$ y = y + d \f$
  Observations &operator+=(const Departures_ &);
  /// \f$ y = y - d \f$
  Observations &operator-=(const Departures_ &);

  /// Save observations
  void save(const eckit::Configuration &) const;
  /// Read observations
  void read(const eckit::Configuration &);

 private:
  void print(std::ostream &) const;

  /// Vector of pointers to ControlObsVector object
  std::vector<std::unique_ptr<ControlObsVector_> > obs_;
};

// =============================================================================

template <typename MODEL>
Observations<MODEL>::Observations(const ObsSpaces_ &obsdb) : obs_(0) {
  for (std::size_t jj = 0; jj < obsdb.size(); ++jj) {
    obs_.emplace_back(new ControlObsVector_(obsdb[jj]));
  }
  Log::trace() << "Observations created" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Observations<MODEL>::Observations(const Observations &other) : obs_(0) {
  for (std::size_t jj = 0; jj < other.obs_.size(); ++jj) {
    obs_.emplace_back(new ControlObsVector_(*other.obs_[jj]));
  }
  Log::trace() << "Observations copy-created" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Observations<MODEL>::~Observations() {
  Log::trace() << "Observations destructed" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Observations<MODEL> &Observations<MODEL>::operator=(const Observations &rhs) {
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) {
    obs_[jj] = rhs.obs_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
std::vector<std::shared_ptr<ControlObsVector<MODEL> > >
Observations<MODEL>::operator-(const Observations &other) const {
  std::vector<std::shared_ptr<ControlObsVector_> > out;
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) {
    std::shared_ptr<ControlObsVector_> ovec(
        new ControlObsVector_(*obs_[jj], true));
    *ovec -= *other.obs_[jj];
    out.push_back(ovec);
  }
  return out;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Observations<MODEL> &Observations<MODEL>::operator+=(const Departures_ &dy) {
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) {
    *obs_[jj] += dy[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
Observations<MODEL> &Observations<MODEL>::operator-=(const Departures_ &dy) {
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) {
    *obs_[jj] -= dy[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Observations<MODEL>::save(const eckit::Configuration &config) const {
  std::vector<eckit::LocalConfiguration> conf;
  config.get("ObsTypes", conf);
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) {
    const eckit::LocalConfiguration obsconf(conf[jj], "ObsData");
    obs_[jj]->save(obsconf);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Observations<MODEL>::read(const eckit::Configuration &config) {
  std::vector<eckit::LocalConfiguration> conf;
  config.get("ObsTypes", conf);
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) {
    const eckit::LocalConfiguration obsconf(conf[jj], "ObsData");
    obs_[jj]->read(obsconf);
  }
  Log::trace() << "Observations:Observations have been read" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Observations<MODEL>::print(std::ostream &os) const {
  for (std::size_t jj = 0; jj < obs_.size(); ++jj) os << *obs_[jj];
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_OBSERVATIONS_H_
