/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVER_H_
#define OOPS_BASE_OBSERVER_H_

#include <memory>
#include <vector>

#include "oops/base/LinearObsOperators.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObsOperators.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/Observations.h"
#include "oops/base/PostBase.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Locations.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/*! @brief Computes \f$ {\cal G}(x) \f$.

 *   Observer is a post-processor that computes \f$ {\cal G}(x) \f$.
 *  Refer to the documentation of CostJo to understand the structure of
 *  \f$ {\cal G} \f$.
 */

// -----------------------------------------------------------------------------

template <typename MODEL, typename STATE>
class Observer : public PostBase<STATE> {
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using LinearObsOperators_ = LinearObsOperators<MODEL>;
  using Locations_ = Locations<MODEL>;
  using ObsAuxCtrls_ = ObsAuxControls<MODEL>;
  using Observations_ = Observations<MODEL>;
  using ObsOperators_ = ObsOperators<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;

 public:
  /// Constructor
  Observer(const ObsSpaces_ &, const ObsOperators_ &, const ObsAuxCtrls_ &,
           const util::Duration &tslot = util::Duration(0),
           const bool subwin = false,
           std::shared_ptr<LinearObsOperators_> htlad =
               std::shared_ptr<LinearObsOperators_>());
  /// Destructor
  ~Observer() {}

  /// Release \f$ y = {\cal G}(x) \f$
  Observations_ *release() { return yobs_.release(); }
  /// Return a reference to the vector of shared pointers to GeoVaLs
  const std::vector<std::shared_ptr<GeoVaLs_> > &geovals() { return gvals_; }

 private:
  // Methods
  void doInitialize(const STATE &, const util::DateTime &,
                    const util::Duration &) override;
  void doProcessing(const STATE &) override;
  void doFinalize(const STATE &) override;

  // Obs operator
  const ObsSpaces_ &obspaces_;  ///< Constant reference to ObservationSpaces
  const ObsOperators_ &hop_;    ///< Constant reference to ObsOperators
  std::shared_ptr<LinearObsOperators_>
      htlad_;  ///< Share pointer to LinearObsOperators

  // Data
  std::unique_ptr<Observations_> yobs_;  ///< Observations vector
  const ObsAuxCtrls_ &ybias_;  ///< Constant reference to ObsAuxControls

  util::DateTime winbgn_;  ///< Begining of assimilation window
  util::DateTime winend_;  ///< End of assimilation window
  util::DateTime bgn_;     ///< Begining of currently active observations
  util::DateTime end_;     ///< End of currently active observations
  util::Duration hslot_;   ///< Half time slot
  util::Duration tstep_;   //!< Time step
  const bool subwindows_;

  /// Vector of shared pointers to GeoVaLs objects
  std::vector<std::shared_ptr<GeoVaLs_> > gvals_;
};

// ====================================================================================

template <typename MODEL, typename STATE>
Observer<MODEL, STATE>::Observer(const ObsSpaces_ &obsdb,
                                 const ObsOperators_ &hop,
                                 const ObsAuxCtrls_ &ybias,
                                 const util::Duration &tslot, const bool swin,
                                 std::shared_ptr<LinearObsOperators_> htlad)
    : PostBase<STATE>(),
      obspaces_(obsdb),
      hop_(hop),
      htlad_(htlad),
      yobs_(new Observations_(obsdb)),
      ybias_(ybias),
      winbgn_(obsdb.windowStart()),
      winend_(obsdb.windowEnd()),
      bgn_(winbgn_),
      end_(winend_),
      hslot_(tslot / 2),
      tstep_(0),
      subwindows_(swin),
      gvals_(0) {
  Log::info() << "winbgn_ " << winbgn_ << " winend_ " << winend_ << " tslot "
              << tslot << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename STATE>
void Observer<MODEL, STATE>::doInitialize(const STATE &xx,
                                          const util::DateTime &end,
                                          const util::Duration &tstep) {
  /// 1. Determine the beginning and end of the window/sub_window;
  tstep_ = tstep;
  const util::DateTime bgn(xx.validTime());
  if (hslot_ == util::Duration(0)) hslot_ = tstep / 2;
  if (subwindows_) {
    if (bgn == end) {
      bgn_ = bgn - hslot_;
      end_ = end + hslot_;
    } else {
      bgn_ = bgn;
      end_ = end;
    }
  }
  if (bgn_ < winbgn_) bgn_ = winbgn_;
  if (end_ > winend_) end_ = winend_;

  /// 2. Allocate GeoVaLs objects for the corresponding ObservationSpaces
  gvals_.clear();
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    std::shared_ptr<GeoVaLs_> tmp(
        new GeoVaLs_(obspaces_[jj], hop_.variables(jj), xx, bgn_, end_));
    gvals_.push_back(tmp);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename STATE>
void Observer<MODEL, STATE>::doProcessing(const STATE &xx) {
  util::DateTime xxtime(bgn_);
  int ts = hslot_.toSeconds() * 2;

  if (ts == tstep_.toSeconds()) {
    /// 0. 3D case
    xxtime = xx.validTime();
  } else {
    /// 1. Check if current time is the closest timestep to an observation
    /// timeslot
    util::Duration term(xx.validTime() - bgn_);
    int vt = term.toSeconds();

    Log::info() << "Observer doProcessing check at : " << xx.validTime()
                << " : " << vt % ts << " ?>? " << tstep_.toSeconds() / 2
                << std::endl;
    if ((vt % ts) > tstep_.toSeconds() / 2) return;

    /// xxtime : validTime of timeslot to use
    util::Duration slot((int)std::round(((double)vt) / ((double)ts)) * ts);
    xxtime += slot;
  }

  /// 1. Determine the beginning and end of the time slot
  util::DateTime t1(xxtime - hslot_);
  util::DateTime t2(xxtime + hslot_);

  if (t1 < bgn_) t1 = bgn_;
  if (t2 > end_) t2 = end_;

  Log::info() << "Observer doProcessing at: " << xx.validTime()
              << " for : t1=" << t1 << " t2=" << t2 << std::endl;

  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    ///  2. Get observation Locations info for interpolation
    Locations_ locs(obspaces_[jj], t1, t2);

    ///  3. Interpolate state variables to observation Locations
    ///    (Compute \f$ {\cal I}(x) \f$)
    xx.interpolate(locs, *gvals_.at(jj));
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename STATE>
void Observer<MODEL, STATE>::doFinalize(const STATE &) {
  // For each observation space:
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    /// 1. Save the linearisation trajectory
    if (htlad_) (*htlad_)[jj].setTrajectory(*gvals_.at(jj), ybias_[jj]);

    /// 2. Compute \f$ {\cal H}[{\cal I}(x)] + {\cal P}(\beta) \f$
    hop_[jj].obsEquiv(*gvals_.at(jj), (*yobs_)[jj].obsvector(), ybias_[jj]);

    /// 3. Compute \f$ {\cal P}(\beta) \f$
    if ((*yobs_)[jj].hasBias())
      hop_[jj].obsBiasEquiv(*gvals_.at(jj), (*yobs_)[jj].biasvector(),
                            ybias_[jj]);
  }
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVER_H_
