/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVERENVARTL_H_
#define OOPS_BASE_OBSERVERENVARTL_H_

#include <memory>
#include <unordered_map>

#include "oops/base/Departures.h"
#include "oops/base/Increment4D.h"
#include "oops/base/PostBaseTL.h"
#include "oops/interface/GenericMatrix.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/// Computes observation equivalent TL and AD to/from increments.

template <typename MODEL, typename INCR>
class ObserverEnVarTL : public PostBaseTL<INCR> {
  using Departures_ = Departures<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;

 public:
  ObserverEnVarTL(Increment4D_ &, std::shared_ptr<GenericMatrix_>);
  ~ObserverEnVarTL() {}
  GeneralizedDepartures *releaseOutputFromTL() override { return nullptr; }

 private:
  // Methods
  void doInitializeTL(const INCR &, const util::DateTime &,
                      const util::Duration &) override;
  void doProcessingTL(const INCR &) override;
  void doFinalizeTL(const INCR &) override;

  // Data
  using Incr4D_ = std::unordered_map<util::DateTime, INCR *>;
  std::shared_ptr<GenericMatrix_> to_low_resol_;
  Incr4D_ dx_;
};

// ====================================================================================

template <typename MODEL, typename INCR>
ObserverEnVarTL<MODEL, INCR>::ObserverEnVarTL(
    Increment4D_ &dx, std::shared_ptr<GenericMatrix_> to_low_resol)
    : PostBaseTL<INCR>(
          dx[dx.first()].validTime(), dx[dx.last()].validTime(),
          dx[dx.first() + 1].validTime() - dx[dx.first()].validTime()),
      to_low_resol_(to_low_resol) {
  for (int k = dx.first(); k <= dx.last(); ++k) {
    dx_.emplace(dx[k].validTime(), &dx[k]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverEnVarTL<MODEL, INCR>::doInitializeTL(const INCR &dx,
                                                  const util::DateTime &end,
                                                  const util::Duration &tstep) {
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverEnVarTL<MODEL, INCR>::doProcessingTL(const INCR &dx) {
  INCR *dxo = dx_[dx.validTime()];
  // Use axpy instead of copy-cstor, since it is able to interoperate
  // fields and fields_base
  if (to_low_resol_.get()) {
    INCR dxout(dx, false);
    dxout.zero();
    dxout.axpy(1.0, dx);
    to_low_resol_->inverseMultiply(dxout);
    dxo->zero();
    dxo->axpy(1.0, dxout);
  } else {
    dxo->zero();
    dxo->axpy(1.0, dx);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverEnVarTL<MODEL, INCR>::doFinalizeTL(const INCR &) {}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVERENVARTL_H_
