/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_OBSERVERIAU_H_
#define OOPS_BASE_OBSERVERIAU_H_

#include <cmath>
#include <memory>
#include <unordered_map>
#include <vector>

#include "oops/base/PostBase.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Compute time average of states or increments during model run.
/*!
 *  Derived classes will compute IAU
 */

template <typename MODEL, typename STATE>
class ObserverIAU : public PostBase<STATE> {
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Variables_ = Variables<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;
  using MatrixFactory_ = GenericMatrixFactory<MODEL>;
  using Incr_ =
      std::unordered_map<util::DateTime, std::shared_ptr<Increment_> >;
  using MapTime_ = std::unordered_map<util::DateTime, util::DateTime>;

 public:
  ObserverIAU(const eckit::Configuration &);
  virtual ~ObserverIAU() {}

 private:
  void doInitialize(STATE &, const util::DateTime &, const util::Duration &);
  void doProcessing(STATE &);

  std::unique_ptr<GenericMatrix_> inflation_;
  const eckit::LocalConfiguration config_;
  Incr_ increments_;
  MapTime_ steps_;
};

// =============================================================================

template <typename MODEL, typename STATE>
ObserverIAU<MODEL, STATE>::ObserverIAU(const eckit::Configuration &config)
    : PostBase<STATE>(config), inflation_(), config_(config) {
  if (config_.has("inflation")) {
    inflation_.reset(
        MatrixFactory_::create(eckit::LocalConfiguration(config, "inflation")));
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL, typename STATE>
void ObserverIAU<MODEL, STATE>::doInitialize(STATE &xx,
                                             const util::DateTime &end,
                                             const util::Duration &tstep) {
  const util::Duration timestep(config_.getString("frequency"));
  const util::Duration step(config_.getString("increment_frequency"));
  const util::Duration halftstep = step / 2;
  util::DateTime now(xx.validTime());

  std::vector<eckit::LocalConfiguration> confs;
  config_.get("increment", confs);

  const Variables_ vars(eckit::LocalConfiguration(config_, "variables"));
  const Geometry_ resol(eckit::LocalConfiguration(config_, "resolution"));
  Increment_ stp(resol, vars, xx.validTime());

  // Read increments
  MapTime_ bgntime, endtime;
  for (unsigned int jm = 0; jm < confs.size(); ++jm) {
    Log::info() << "ObserverIAU: Read increment at: " << now << std::endl;
    std::shared_ptr<Increment_> dx(new Increment_(resol, vars, xx.validTime()));

    const eckit::LocalConfiguration stpConfig(confs[jm], "starting_point");
    const eckit::LocalConfiguration anaConfig(confs[jm], "analysis");

    stp.read(stpConfig);
    dx->read(anaConfig);
    (*dx) -= stp;
    increments_[now] = dx;

    bgntime.emplace(now, now - halftstep);
    endtime.emplace(now, bgntime[now] + step);
    now += step;
  }

  // Initialize IAU steps
  util::DateTime lasttime(now - step);
  const util::DateTime firsttime(xx.validTime() + timestep);

  if (timestep.toSeconds() == 0) {
    steps_.emplace(firsttime, firsttime);
  } else {
    if (confs.size() == 1) {
      lasttime = firsttime + halftstep;
    }
    for (util::DateTime tm(firsttime); tm <= lasttime; tm += timestep) {
      for (util::DateTime t(xx.validTime()); t <= lasttime; t += step) {
        if ((tm > bgntime[t]) && (tm <= endtime[t])) {
          steps_.emplace(tm, t);
        }
      }
      Log::info() << "ObserverIAU: tstep:" << tm << " increment:" << steps_[tm]
                  << std::endl;
    }

    util::Duration wlen(lasttime - xx.validTime());
    int nb = (wlen.toSeconds() / timestep.toSeconds());
    double weight = 1.0 / ((double)nb);
    Log::info() << "ObserverIAU: weight: " << weight << std::endl;
    for (util::DateTime t(xx.validTime()); t <= lasttime; t += step) {
      (*increments_[t]) *= weight;
    }

    if (config_.has("inflation")) {
      Log::info() << "ObserverIAU: apply extra inflation" << std::endl;
      const eckit::LocalConfiguration inflation(config_, "inflation");
      for (util::DateTime t(xx.validTime()); t <= lasttime; t += step) {
        inflation_->multiply((*increments_[t]));
      }
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL, typename STATE>
void ObserverIAU<MODEL, STATE>::doProcessing(STATE &xx) {
  const util::DateTime now(xx.validTime());

  if (steps_.count(now)) {
    Log::info() << "ObserverIAU: Process IAU at: " << now << std::endl;
    std::shared_ptr<Increment_> dx(increments_[steps_[now]]);
    util::Duration offset(xx.validTime() - dx->validTime());
    dx->updateTime(offset);
    xx += (*dx);
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVERIAU_H_
