/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVERTL_H_
#define OOPS_BASE_OBSERVERTL_H_

#include <map>
#include <memory>
#include <vector>

#include "oops/base/Departures.h"
#include "oops/base/LinearObsOperators.h"
#include "oops/base/ObsAuxIncrements.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/PostBaseTL.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Locations.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/*! @brief Computes \f$ G dx \f$.

 *   ObserverTL is a post-processor that computes \f$ G dx \f$.
 *  Refer to the documentation of CostJo to understand the structure of
 *  \f$ G \f$.
 */

// -----------------------------------------------------------------------------

template <typename MODEL, typename INCR>
class ObserverTL : public PostBaseTL<INCR> {
  using Departures_ = Departures<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using LinearObsOperators_ = LinearObsOperators<MODEL>;
  using Locations_ = Locations<MODEL>;
  using ObsAuxIncrs_ = ObsAuxIncrements<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;
  using gvalsMap_ = typename std::map<util::DateTime,
                                      std::vector<std::unique_ptr<GeoVaLs_> > >;
  using gvalsIter_ = typename gvalsMap_::iterator;

 public:
  /// Constructor
  ObserverTL(const ObsSpaces_ &, const LinearObsOperators_ &,
             const ObsAuxIncrs_ &, const util::Duration &,
             const bool subwin = false);
  /// Destructor
  ~ObserverTL() {}

  /// Release \f$ d = G dx \f$
  Departures_ *releaseOutputFromTL() override { return new Departures_(ydep_); }
  /// Set a pointer to ObsAuxIncrs object
  void setupTL(const ObsAuxIncrs_ &dybias) { dybias_ = &dybias; }

 private:
  // Methods
  void doInitializeTL(const INCR &, const util::DateTime &,
                      const util::Duration &) override;
  void doProcessingTL(const INCR &) override;
  void doFinalizeTL(const INCR &) override;

  // Obs operator
  const ObsSpaces_ &obspaces_;  ///< Constant reference to ObservationSpaces
  const LinearObsOperators_
      &hoptlad_;  ///< Constant reference to LinearObsOperators

  // Data
  Departures_ ydep_;                    ///< Departures
  mutable const ObsAuxIncrs_ *dybias_;  ///< Pointer to ObsAuxIncrs

  util::DateTime winbgn_;  ///< Begining of assimilation window
  util::DateTime winend_;  ///< End of assimilation window
  util::DateTime bgn_;     ///< Begining of currently active observations
  util::DateTime end_;     ///< End of currently active observations
  util::Duration hslot_;   ///< Half time slot
  util::Duration tstep_;   //!< Time step
  const bool subwindows_;

  /// Map of unique pointers to GeoVaLs
  gvalsMap_ gvals_;
};

// ====================================================================================

template <typename MODEL, typename INCR>
ObserverTL<MODEL, INCR>::ObserverTL(const ObsSpaces_ &obsdb,
                                    const LinearObsOperators_ &hoptlad,
                                    const ObsAuxIncrs_ &dybias,
                                    const util::Duration &tslot,
                                    const bool subwin)
    : PostBaseTL<INCR>(obsdb.windowStart(), obsdb.windowEnd()),
      obspaces_(obsdb),
      hoptlad_(hoptlad),
      ydep_(obspaces_),
      dybias_(&dybias),
      winbgn_(obsdb.windowStart()),
      winend_(obsdb.windowEnd()),
      bgn_(winbgn_),
      end_(winend_),
      hslot_(tslot / 2),
      tstep_(0),
      subwindows_(subwin) {}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverTL<MODEL, INCR>::doInitializeTL(const INCR &dx,
                                             const util::DateTime &end,
                                             const util::Duration &tstep) {
  /// 1. Determine the beginning and end of the window/sub_window;
  tstep_ = tstep;
  const util::DateTime bgn(dx.validTime());
  if (hslot_ == util::Duration(0)) hslot_ = tstep / 2;
  if (subwindows_) {
    if (bgn == end) {
      bgn_ = bgn - hslot_;
      end_ = end + hslot_;
    } else {
      bgn_ = bgn;
      end_ = end;
    }
  }
  if (bgn_ < winbgn_) bgn_ = winbgn_;
  if (end_ > winend_) end_ = winend_;

  /// 2. Allocate GeoVaLs objects for the corresponding ObservationSpaces
  ///   (if not already allocated)
  gvalsIter_ it = gvals_.find(bgn_);
  if (it == gvals_.end()) {
    std::vector<std::unique_ptr<GeoVaLs_> > gvalsvec;
    for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
      gvalsvec.emplace_back(
          new GeoVaLs_(obspaces_[jj], hoptlad_.variables(jj), dx, bgn_, end_));
    }
    gvals_.emplace(bgn_, std::move(gvalsvec));
    it = gvals_.find(bgn_);
    ASSERT(it != gvals_.end());
  }
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    it->second[jj]->zero();
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverTL<MODEL, INCR>::doProcessingTL(const INCR &dx) {
  util::DateTime xxtime(bgn_);
  int ts = hslot_.toSeconds() * 2;

  if (ts == tstep_.toSeconds()) {
    /// 0. 3d case
    xxtime = dx.validTime();
  } else {
    /// 1. Check if current time is the closest timestep to an observation
    /// timeslot
    util::Duration term(dx.validTime() - bgn_);
    int vt = term.toSeconds();
    if ((vt % ts) > tstep_.toSeconds() / 2) return;

    // xxtime : validTime of timeslot to use
    util::Duration slot((int)std::round(((double)vt) / ((double)ts)) * ts);
    xxtime += slot;
  }

  /// 2. Determine the beginning and end of the time slot
  util::DateTime t1(xxtime - hslot_);
  util::DateTime t2(xxtime + hslot_);
  if (t1 < bgn_) t1 = bgn_;
  if (t2 > end_) t2 = end_;

  // Log::info() << "ObserverTL doProcessing for : t1=" << t1 << " t2=" << t2 <<
  // std::endl;

  // For each observation space:
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    ///  2. Get observation Locations info for interpolation
    Locations_ locs(obspaces_[jj], t1, t2);

    ///  3. Interpolate increment variables to observation Locations
    ///    (Compute \f$ I dx \f$)
    gvalsIter_ it = gvals_.find(bgn_);
    ASSERT(it != gvals_.end());
    dx.interpolateTL(locs, *it->second[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverTL<MODEL, INCR>::doFinalizeTL(const INCR &) {
  // For each observation space:
  gvalsIter_ it = gvals_.find(bgn_);
  ASSERT(it != gvals_.end());
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    /// 1. Compute \f$ H I dx + P dB \f$
    hoptlad_[jj].obsEquivTL(*it->second[jj], ydep_[jj].obsvector(),
                            (*dybias_)[jj]);
    /// 2. Compute \f$ P dB \f$
    if (ydep_[jj].hasBias())
      hoptlad_[jj].obsBiasEquivTL(*it->second[jj], ydep_[jj].biasvector(),
                                  (*dybias_)[jj]);
  }
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVERTL_H_
