/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_POSTBASE_H_
#define OOPS_BASE_POSTBASE_H_

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/base/PostTimer.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief Handles post-processing of model fields.

 *   PostBase is the base class for all state post processors, it
 *  is mostly used so that PostProcessor can hold a vector of such
 *  processors.

 *   By default processing is performed on every call.

 *   PostBase is templated on FLDS; FLDS would typically be either State
 *  or Increment.

 *   Post-processing is organized in three stages/methods: initialize
 *  (called prior to time-stepping), process (called during the forecast
 *  at each time step) and finalize (called once the time-stepping is
 *  finished).

 *   Methods initialize, process and finalize wrap around virtual methods
 *  (implemented by PostBase children classes) doInitialize, doProcessing
 *  and doFinalize.

 *   PostBase class contains a PostTimer object. PostTimer class controls
 *  the frequency of post-processing.

 *   PostBase is the base class of the following classes (refer to their
 *  documentation for more information):

    - Observer

    - StateInfo

    - StateSaver

    - StateWriter

    - TrajectorySaver

    - WeightedDiff

    - WeightedMean

    - JqTerm

 */

template <typename FLDS>
class PostBase : private eckit::NonCopyable {
 public:
  /// Constructors
  PostBase() : timer_() {}
  explicit PostBase(const util::Duration &freq) : timer_(freq) {}
  explicit PostBase(const eckit::Configuration &conf) : timer_(conf) {}
  PostBase(const util::DateTime &start, const eckit::Configuration &conf)
      : timer_(start, conf) {}
  PostBase(const util::DateTime &start, const util::DateTime &finish,
           const util::Duration &freq = util::Duration(0))
      : timer_(start, finish, freq) {}
  /// Destructor
  virtual ~PostBase() {}

  /// Wraps doInitialize method; doProcessing is called if requested
  void initialize(const FLDS &xx, const util::DateTime &end,
                  const util::Duration &tstep) {
    timer_.initialize(xx.validTime(), end, tstep);
    this->doInitialize(xx, end, tstep);
    if (timer_.itIsTime(xx.validTime())) this->doProcessing(xx);
  }

  void initialize(FLDS &xx, const util::DateTime &end,
                  const util::Duration &tstep) {
    timer_.initialize(xx.validTime(), end, tstep);
    this->doInitialize(xx, end, tstep);
    if (timer_.itIsTime(xx.validTime())) this->doProcessing(xx);
  }

  /// Wraps doProcessing method; doProcessing is is called when requested
  void process(const FLDS &xx) {
    if (timer_.itIsTime(xx.validTime())) this->doProcessing(xx);
  }

  void process(FLDS &xx) {
    if (timer_.itIsTime(xx.validTime())) this->doProcessing(xx);
  }

  /// Wraps doFinalize method
  void finalize(const FLDS &xx) { this->doFinalize(xx); }

  /// Actual processing
  virtual void doProcessing(const FLDS &) {}
  virtual void doProcessing(FLDS &fld) { doProcessing((const FLDS &)fld); }
  /// Actual initialization
  virtual void doInitialize(const FLDS &, const util::DateTime &,
                            const util::Duration &) {}
  virtual void doInitialize(FLDS &fld, const util::DateTime &t,
                            const util::Duration &d) {
    doInitialize((const FLDS &)fld, t, d);
  }
  /// Actual finalization
  virtual void doFinalize(const FLDS &) {}

 private:
  PostTimer timer_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_POSTBASE_H_
