/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_POSTBASETL_H_
#define OOPS_BASE_POSTBASETL_H_

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/base/GeneralizedDepartures.h"
#include "oops/base/PostTimer.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief Handles post-processing of model fields related to cost function.

 *   PostBaseTL is the base class for all cost function processors, it
 *  is mostly used so that PostProcessorTL can hold a vector of such
 *  processors.

 *   PostProcessorTL is typically attached to LinearModel::forecastTL
 *  method.

 *   By default processing is performed on every call.

 *   PostBaseTL is templated on INCR; INCR would typically be Increment.

 *   Post-processing is organized in three stages/methods: initializeTL
 *  (called prior to time-stepping), processTL (called during forecastTL
 *  at each time step) and finalizeTL (called once the time-stepping is
 *  finished).

 *   Methods initializeTL, processTL and finalizeTL wrap around virtual
 *  methods (implemented by PostBaseTL children classes) doInitializeTL,
 *  doProcessingTL and doFinalizeTL.

 *   PostBaseTL class contains a PostTimer object. PostTimer class controls
 *  the frequency of post-processing.

 *   PostBaseTL is the base class of the following classes (refer to their
 *  documentation for more information):

    - ObserverTL

    - WeightedDiffTL

    - JqTermTL
 */

template <typename INCR>
class PostBaseTL : private eckit::NonCopyable {
 public:
  /// Constructors
  PostBaseTL() : timer_() {}
  PostBaseTL(const util::DateTime &start, const util::DateTime &finish,
             const util::Duration &freq = util::Duration(0))
      : timer_(start, finish, freq) {}
  /// Destructor
  virtual ~PostBaseTL() {}

  /// Wraps doInitializerTL method; doProcessingTL is is called if requested
  void initializeTL(const INCR &dx, const util::DateTime &end,
                    const util::Duration &step) {
    timer_.initialize(dx.validTime(), end, step);
    this->doInitializeTL(dx, end, step);
    if (timer_.itIsTime(dx.validTime())) this->doProcessingTL(dx);
  }

  /// Wraps doProcessing method; doProcessing is is called when requested
  void processTL(const INCR &dx) {
    if (timer_.itIsTime(dx.validTime())) this->doProcessingTL(dx);
  }

  /// Wraps doFinalizeTL method
  void finalizeTL(const INCR &dx) { this->doFinalizeTL(dx); }

  /// Return TL dual space output
  virtual GeneralizedDepartures *releaseOutputFromTL() = 0;

 private:
  /// Actual processing
  virtual void doInitializeTL(const INCR &, const util::DateTime &,
                              const util::Duration &) = 0;
  /// Actual initialization
  virtual void doProcessingTL(const INCR &) = 0;
  /// Actual finalization
  virtual void doFinalizeTL(const INCR &) = 0;

  PostTimer timer_;
};

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_POSTBASETL_H_
