/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_POSTPROCESSOR_H_
#define OOPS_BASE_POSTPROCESSOR_H_

#include <memory>
#include <vector>

#include "oops/base/PostBase.h"

namespace oops {

/// Control model post processing
/*!  @brief Controls model post processing

 *   This class controls model post processing in the most general sense,
 *  ie all diagnostics computations that do not affect the model integration.
 *  It contains a vector of PostBase objects. It just calls all the individual
 *  processors one by one.

 *   Post-processing is organized in three stages/methods: initialize
 *  (called prior to time-stepping), process (called during the forecast
 *  at each  time step) and finalize (called once the time-stepping is
 *  finished).

 *   Methods initialize, process and finalize simply call the initialize,
 *  process and finalize methods, respectively, of the enrolled individual
 *  processors.

 *   See the documentation of PostBase for the details of implementation of
 *  processors and the list of available processors.
 */

template <typename FLDS>
class PostProcessor {
  using PostBase_ = PostBase<FLDS>;

 public:
  /// Constructor
  PostProcessor() : processors_() {}
  /// Copy constructor
  PostProcessor(const PostProcessor &pp) : processors_(pp.processors_) {}
  /// Destructor
  ~PostProcessor() {}

  /// Adds a post processor to the vector of post-processors
  void enrollProcessor(PostBase_ *pp) {
    if (pp != 0) {
      std::shared_ptr<PostBase_> sp(pp);
      processors_.push_back(sp);
    }
  }

  /// Adds a post processor to the vector of post-processors
  void enrollProcessor(std::shared_ptr<PostBase_> pp) {
    if (pp != 0) processors_.push_back(pp);
  }

  /// Calls initialize method of individual processors
  void initialize(const FLDS &xx, const util::DateTime &end,
                  const util::Duration &step) {
    for (std::shared_ptr<PostBase_> jp : processors_)
      jp->initialize(xx, end, step);
  }

  void initialize(FLDS &xx, const util::DateTime &end,
                  const util::Duration &step) {
    for (std::shared_ptr<PostBase_> jp : processors_)
      jp->initialize(xx, end, step);
  }

  /// Calls process method of individual processors
  void process(const FLDS &xx) {
    for (std::shared_ptr<PostBase_> jp : processors_) jp->process(xx);
  }

  void process(FLDS &xx) {
    for (std::shared_ptr<PostBase_> jp : processors_) jp->process(xx);
  }

  /// Calls finalize method of individual processors
  void finalize(const FLDS &xx) {
    for (std::shared_ptr<PostBase_> jp : processors_) jp->finalize(xx);
  }

 private:
  PostProcessor operator=(const PostProcessor &);

  /// Vector of shared pointers to individual processors
  std::vector<std::shared_ptr<PostBase_> > processors_;
};

}  // namespace oops

#endif  // OOPS_BASE_POSTPROCESSOR_H_
