/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_POSTPROCESSORSTATE_H_
#define OOPS_BASE_POSTPROCESSORSTATE_H_

#include <map>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/PostBase.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"
#include "util/formats.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief The PostProcessorFactory defines and manages a factory to manage
 * PostProcessor for objects.
 */

// -----------------------------------------------------------------------------

template <typename FLDS>
class PostProcessorFactory {
 public:
  static PostBase<FLDS> *create(const eckit::Configuration &);
  virtual ~PostProcessorFactory() { getMakers().clear(); }

 protected:
  explicit PostProcessorFactory(const std::string &);

 private:
  virtual PostBase<FLDS> *make(const eckit::Configuration &) = 0;
  static std::map<std::string, PostProcessorFactory<FLDS> *> &getMakers() {
    static std::map<std::string, PostProcessorFactory<FLDS> *> makers_;
    return makers_;
  }
};

template <class MODEL, class FLDS, class POST_PROC>
class PostProcessorMaker : public PostProcessorFactory<FLDS> {
 private:
  virtual PostBase<FLDS> *make(const eckit::Configuration &config) {
    return new POST_PROC(config);
  }

 public:
  explicit PostProcessorMaker(const std::string &name)
      : PostProcessorFactory<FLDS>(name) {}
};

// =============================================================================

//  Factory
// -----------------------------------------------------------------------------

template <typename FLDS>
PostProcessorFactory<FLDS>::PostProcessorFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in post processor factory."
                 << std::endl;
    ABORT("Element already registered in PostProcessorFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename FLDS>
PostBase<FLDS> *PostProcessorFactory<FLDS>::create(
    const eckit::Configuration &config) {
  std::string id = config.getString("post_processor_type");
  Log::trace() << "Post processor Type=" << id << std::endl;
  typename std::map<std::string, PostProcessorFactory<FLDS> *>::iterator j =
      getMakers().find(id);
  if (j == getMakers().end()) {
    Log::error() << id << " does not exist in post processor factory."
                 << std::endl;
    ABORT("Element does not exist in PostProcessorFactory.");
  }
  return (*j).second->make(config);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_POSTPROCESSORSTATE_H_
