/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_POSTPROCESSORTL_H_
#define OOPS_BASE_POSTPROCESSORTL_H_

#include <memory>
#include <vector>

#include "oops/base/GeneralizedDepartures.h"
#include "oops/base/PostBaseTL.h"

namespace oops {

/// Control linear model post processing
/*!
 *   This class controls linear model post processing in the most general
 *  sense, ie all diagnostics computations that do not affect the linear
 *  model integration. It just calls all the individual processors one by one.

 *   Post-processing is organized in three stages/methods: initializeTL
 *  (called prior to time-stepping), processTL (called during
 *  LinearModel::forecastTL), nd finalize (called once the time-stepping is
 *  finished).

 *   Methods initializeTL, processTL and finalizeTL simply call the
 *  initializeTL, processTL and finalizeTL methods, respectively, of the
 *  enrolled individual processors.

 *   See the documentation of PostBaseTL for the details of implementation of
 *  processors and the list of available processors.

 */

template <typename INCR>
class PostProcessorTL {
  using PostBaseTL_ = PostBaseTL<INCR>;

 public:
  /// Constructor
  PostProcessorTL() : processors_() {}
  /// Copy constructor
  PostProcessorTL(const PostProcessorTL &pp) : processors_(pp.processors_) {}
  /// Destructor
  ~PostProcessorTL() {}

  /// Adds a post processor to the vector of post-processors
  void enrollProcessor(PostBaseTL_ *pp) {
    if (pp != 0) {
      std::shared_ptr<PostBaseTL_> sp(pp);
      processors_.push_back(sp);
    }
  }

  /// Adds a post processor to the vector of post-processors
  void enrollProcessor(std::shared_ptr<PostBaseTL_> pp) {
    if (pp != 0) processors_.push_back(pp);
  }

  /// Release the dual space output of the post-processor
  GeneralizedDepartures *releaseOutputFromTL(unsigned int ii) {
    GeneralizedDepartures *lambda = processors_[ii]->releaseOutputFromTL();
    return lambda;
  }

  /// Calls initializeTL method of individual processors
  void initializeTL(const INCR &dx, const util::DateTime &end,
                    const util::Duration &step) {
    for (std::shared_ptr<PostBaseTL_> jp : processors_) {
      jp->initializeTL(dx, end, step);
    }
  }

  /// Calls processTL method of individual processors
  void processTL(const INCR &dx) {
    for (std::shared_ptr<PostBaseTL_> jp : processors_) jp->processTL(dx);
  }

  /// Calls finalizeTL method of individual processors
  void finalizeTL(const INCR &dx) {
    for (std::shared_ptr<PostBaseTL_> jp : processors_) jp->finalizeTL(dx);
  }

 private:
  PostProcessorTL operator=(const PostProcessorTL &);

  /// Vector of shared pointers to individual processors
  std::vector<std::shared_ptr<PostBaseTL_> > processors_;
};

}  // namespace oops

#endif  // OOPS_BASE_POSTPROCESSORTL_H_
