/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_STATE4D_H_
#define OOPS_BASE_STATE4D_H_

#include <cmath>
#include <memory>
#include <ostream>
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/State.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace oops {

/// Four dimensional state
/*!
 *  The 4D state is mostly used as part of the VDA control variable.
 */

// -----------------------------------------------------------------------------
template <typename MODEL>
class State4D : public util::Printable {
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;
  using Model_ = Model<MODEL>;

 public:
  static const std::string classname() { return "State4D"; }

  /// The arguments define the number of sub-windows and the resolution
  State4D() : state4d_() {}
  State4D(const eckit::Configuration &, const Geometry_ &, const Model_ &);
  explicit State4D(const State4D &);
  State4D(const Geometry_ &, const State4D &);
  State4D(const Geometry_ &, const Model_ &, const State4D &);
  ~State4D();

  /// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

  /// Get model space control variable
  bool checkStatesNumber(const unsigned int nn) const {
    return state4d_.size() == nn;
  }
  unsigned int statesNumber() const { return state4d_.size(); }
  State_ &operator[](const int ii) { return *state4d_[ii]; }
  const State_ &operator[](const int ii) const { return *state4d_[ii]; }

  /// Push back State_
  void push_back(const State_ &);

  /// Accessor to times
  const std::vector<util::DateTime> times() const { return times_; }

 private:
  State4D &operator=(const State4D &);  // No assignment
  void print(std::ostream &) const;

  std::vector<std::unique_ptr<State_> > state4d_;
  std::vector<util::DateTime> times_;
};

// =============================================================================

template <typename MODEL>
State4D<MODEL>::State4D(const eckit::Configuration &config,
                        const Geometry_ &resol, const Model_ &model) {
  std::vector<eckit::LocalConfiguration> files;
  config.get("state", files);

  for (size_t jsub = 0; jsub < files.size(); ++jsub) {
    State_ *js = new State_(resol, model, files[jsub]);
    times_.push_back(js->validTime());
    state4d_.emplace_back(js);
  }
  Log::trace() << "State4D constructed." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State4D<MODEL>::State4D(const State4D &other) {
  for (const auto &js : other.state4d_) {
    times_.push_back(js->validTime());
    state4d_.emplace_back(new State_(*js));
  }
  Log::trace() << "State4D copied." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State4D<MODEL>::State4D(const Geometry_ &resol, const Model_ &model,
                        const State4D &other) {
  for (const auto &js : other.state4d_) {
    times_.push_back(js->validTime());
    state4d_.emplace_back(new State_(resol, model, *js));
  }
  Log::trace() << "State4D copied." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State4D<MODEL>::State4D(const Geometry_ &resol, const State4D &other) {
  for (const auto &js : other.state4d_) {
    times_.push_back(js->validTime());
    state4d_.emplace_back(new State_(resol, *js));
  }
  Log::trace() << "State4D copied." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State4D<MODEL>::~State4D() {
  Log::trace() << "State4D destructed." << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State4D<MODEL>::read(const eckit::Configuration &config) {
  std::vector<eckit::LocalConfiguration> confs;
  if (config.has("state")) {
    config.get("state", confs);
    ASSERT(state4d_.size() == confs.size());
    unsigned jsub = 0;
    for (std::unique_ptr<State_> &js : state4d_) {
      Log::debug() << "State4D::read" << confs[jsub] << std::endl;
      js->read(confs[jsub]);
      ++jsub;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State4D<MODEL>::write(const eckit::Configuration &config) const {
  std::vector<eckit::LocalConfiguration> confs;
  if (config.has("state")) {
    config.get("state", confs);
    ASSERT(state4d_.size() == confs.size());
    unsigned int jsub = 0;
    for (const std::unique_ptr<State_> &js : state4d_) {
      Log::debug() << "State4D::write" << confs[jsub] << std::endl;
      js->write(confs[jsub]);
      ++jsub;
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State4D<MODEL>::print(std::ostream &outs) const {
  for (const std::unique_ptr<State_> &js : state4d_) {
    outs << *js << std::endl;
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double State4D<MODEL>::norm() const {
  double zn = 0.0;
  for (const std::unique_ptr<State_> &js : state4d_) {
    double zz = js->norm();
    zn += zz * zz;
  }
  return sqrt(zn);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State4D<MODEL>::push_back(const State_ &state) {
  std::unique_ptr<State_> copy(new State_(state));
  times_.push_back(copy->validTime());
  state4d_.push_back(std::move(copy));
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_STATE4D_H_
