/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_STATICCOVARIANCE4D_H_
#define OOPS_BASE_STATICCOVARIANCE4D_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

/// Generic ensemble based model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class StaticCovariance4D : public ModelSpaceCovariance4DBase<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  StaticCovariance4D(const Geometry_ &, const Variables_ &,
                     const eckit::Configuration &, const State4D_ &xb);
  ~StaticCovariance4D();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const {
    return static_->newIncrModCtlVec();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const { return new IncrEnsCtlVec_(); }

 private:
  unsigned nsubwin_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL> > static_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
StaticCovariance4D<MODEL>::StaticCovariance4D(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &xb)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      nsubwin_(0),
      static_(CovarianceFactory<MODEL>::create(
          eckit::LocalConfiguration(config, "covariance_3d"), resol, vars,
          xb[0])) {
  Log::trace() << "StaticCovariance4D created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
StaticCovariance4D<MODEL>::~StaticCovariance4D() {
  Log::trace() << "StaticCovariance4D destructed.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticCovariance4D<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol,
    const eckit::Configuration &config) {
  static_->linearize(xb[0], resol, config);
  Log::trace() << "StaticCovariance4D linearized.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticCovariance4D<MODEL>::advectedMultiply(const Increment4D_ &dxi,
                                                 Increment4D_ &dxo) const {
  Increment_ temp(dxi[0]);
  temp.zero();
  for (int k2 = dxi.first(); k2 <= dxi.last(); ++k2) {
    temp.axpy(1.0, dxi[k2], false);
  }
  static_->multiply(temp, dxo[dxi.first()]);
  for (int k1 = dxi.first() + 1; k1 <= dxi.last(); ++k1) {
    dxo[k1].zero();
    dxo[k1].axpy(1.0, dxo[0], false);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticCovariance4D<MODEL>::advectedMultiplySqrt(const IncrCtlVec_ &dvi,
                                                     Increment4D_ &dxo) const {
  static_->multiplySqrt(dvi, dxo[dxo.first()]);
  for (int k1 = dxo.first() + 1; k1 <= dxo.last(); ++k1) {
    dxo[k1].zero();
    dxo[k1].axpy(1.0, dxo[0], false);
  }
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticCovariance4D<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  Increment_ temp(dxi[0]);
  temp.zero();
  for (int k2 = dxi.first(); k2 <= dxi.last(); ++k2) {
    temp.axpy(1.0, dxi[k2], false);
  }
  static_->multiplySqrtTrans(temp, dvo);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticCovariance4D<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_STATICCOVARIANCE4D_H_
