/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_BASE_STATICWITHTLMODELCOVARIANCE4D_H_
#define OOPS_BASE_STATICWITHTLMODELCOVARIANCE4D_H_

#include <memory>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/ModelSpaceCovariance4DBase.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/base/ObserverEnVarAD.h"
#include "oops/base/ObserverEnVarTL.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/TrajectorySaverMF.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/LinearModel.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

/// Generic ensemble based model space error covariance.

// -----------------------------------------------------------------------------
template <typename MODEL>
class StaticWithTLModelCovariance4D : public ModelSpaceCovariance4DBase<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;
  using State_ = State<MODEL>;
  using State4D_ = State4D<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;
  using IncrCtlVec_ = IncrCtlVec<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using IncrModCtlVec_ = IncrModCtlVec<MODEL>;
  using Variables_ = Variables<MODEL>;
  using ModelAuxCtrl_ = ModelAuxControl<MODEL>;
  using ModelAuxIncrement_ = ModelAuxIncrement<MODEL>;
  using LinearModel_ = LinearModel<MODEL>;
  using Model_ = Model<MODEL>;
  using ObserverTL_ = ObserverEnVarTL<MODEL, Increment_>;
  using ObserverAD_ = ObserverEnVarAD<MODEL, Increment_>;
  using PostProcessorTL_ = PostProcessorTL<Increment_>;
  using PostProcessorAD_ = PostProcessorAD<Increment_>;
  using TrajectorySaver_ = TrajectorySaverMF<MODEL>;
  using GenericMatrix_ = GenericMatrix<MODEL>;
  using MatrixFactory_ = GenericMatrixFactory<MODEL>;

 public:
  StaticWithTLModelCovariance4D(const Geometry_ &, const Variables_ &,
                                const eckit::Configuration &,
                                const State4D_ &xb);
  ~StaticWithTLModelCovariance4D();

  void advectedLinearize(const State4D_ &, const Geometry_ &,
                         const eckit::Configuration &);
  void advectedMultiply(const Increment4D_ &, Increment4D_ &) const;
  void advectedMultiplySqrt(const IncrCtlVec_ &, Increment4D_ &) const;
  void advectedMultiplySqrtTrans(const Increment4D_ &, IncrCtlVec_ &) const;
  void advectedInverseMultiply(const Increment4D_ &, Increment4D_ &) const;
  const ModelSpaceCovarianceBase<MODEL> &covar() const { return *static_; }

  // Control Vector
  IncrModCtlVec_ *newIncrModCtlVec() const {
    return static_->newIncrModCtlVec();
  }
  IncrEnsCtlVec_ *newIncrEnsCtlVec() const { return new IncrEnsCtlVec_(); }

 private:
  unsigned nsubwin_;
  const eckit::LocalConfiguration config_;
  const Variables_ variables_;
  std::unique_ptr<ModelSpaceCovarianceBase<MODEL> > static_;
  std::unique_ptr<ModelAuxIncrement_> modaux_;
  std::unique_ptr<Model_> model_;
  std::vector<std::unique_ptr<LinearModel_> > tlm_;
  std::unique_ptr<Geometry_> low_resol_;
  std::shared_ptr<GenericMatrix_> to_low_resol_;
};

// =============================================================================

/// Constructor, destructor
// -----------------------------------------------------------------------------
template <typename MODEL>
StaticWithTLModelCovariance4D<MODEL>::StaticWithTLModelCovariance4D(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &config, const State4D_ &xb)
    : ModelSpaceCovariance4DBase<MODEL>::ModelSpaceCovariance4DBase(resol,
                                                                    config),
      nsubwin_(0),
      config_(config),
      variables_(vars) {
  static_.reset(CovarianceFactory<MODEL>::create(
      eckit::LocalConfiguration(config, "covariance_3d"), resol, variables_,
      xb[0]));

  Log::trace() << "StaticWithTLModelCovariance4D created.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
StaticWithTLModelCovariance4D<MODEL>::~StaticWithTLModelCovariance4D() {
  Log::trace() << "StaticWithTLModelCovariance4D destructed.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticWithTLModelCovariance4D<MODEL>::advectedLinearize(
    const State4D_ &xb, const Geometry_ &resol,
    const eckit::Configuration &config) {
  const util::Duration length(config.getString("window_length"));
  PostProcessor<State_> post;

  const eckit::LocalConfiguration low_resol_config(config, "low_resolution");
  low_resol_.reset(new Geometry_(low_resol_config));
  const eckit::LocalConfiguration modelConf(config, "model");
  model_.reset(new Model_(*low_resol_, modelConf));
  const eckit::LocalConfiguration tlmConf(config, "linearmodel");
  ModelAuxCtrl_ moderr(*low_resol_, *model_, config);
  modaux_.reset(new ModelAuxIncrement_(*low_resol_, config));

  State_ x0(xb[0]);
  if (config.has("matrix")) {
    to_low_resol_.reset(MatrixFactory_::create(config));
    to_low_resol_->multiply(x0);
  }
  static_->linearize(x0, *low_resol_, config);
  tlm_.clear();
  post.enrollProcessor(
      new TrajectorySaver_(x0, moderr, *low_resol_, tlm_, tlmConf));
  model_->forecast(x0, moderr, length, post);

  Log::trace() << "StaticWithTLModelCovariance4D linearized.";
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticWithTLModelCovariance4D<MODEL>::advectedMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  util::Duration length = dxi[dxi.last()].validTime() - dxi[0].validTime();
  PostProcessor<Increment_> no_post;

  dxo.zero();
  std::shared_ptr<ObserverTL_> obstl(new ObserverTL_(dxo, to_low_resol_));
  PostProcessorTL_ costtl;
  costtl.enrollProcessor(obstl);

  std::shared_ptr<ObserverAD_> obsad(new ObserverAD_(dxi, to_low_resol_));
  PostProcessorAD_ costad;
  costad.enrollProcessor(obsad);

  Increment_ dx(*low_resol_, variables_, dxi[dxi.first()].validTime(),
                dxi[dxi.last()].validTime());
  dx.zero();

  tlm_[0]->forecastAD(dx, *modaux_, length, no_post, costad);
  static_->multiply(dx, dx);
  tlm_[0]->forecastTL(dx, *modaux_, length, no_post, costtl);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticWithTLModelCovariance4D<MODEL>::advectedMultiplySqrt(
    const IncrCtlVec_ &dvi, Increment4D_ &dxo) const {
  util::Duration length = dxo[dxo.last()].validTime() - dxo[0].validTime();
  PostProcessor<Increment_> no_post;

  dxo.zero();
  std::shared_ptr<ObserverTL_> obstl(new ObserverTL_(dxo, to_low_resol_));
  PostProcessorTL_ costtl;
  costtl.enrollProcessor(obstl);

  Increment_ dx(*low_resol_, variables_, dxo[dxo.first()].validTime(),
                dxo[dxo.first()].validTime());
  dx.zero();

  static_->multiplySqrt(dvi, dx);
  tlm_[0]->forecastTL(dx, *modaux_, length, no_post, costtl);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticWithTLModelCovariance4D<MODEL>::advectedMultiplySqrtTrans(
    const Increment4D_ &dxi, IncrCtlVec_ &dvo) const {
  util::Duration length = dxi[dxi.last()].validTime() - dxi[0].validTime();
  PostProcessor<Increment_> no_post;

  std::shared_ptr<ObserverAD_> obsad(new ObserverAD_(dxi, to_low_resol_));
  PostProcessorAD_ costad;
  costad.enrollProcessor(obsad);

  Increment_ dx(*low_resol_, variables_, dxi[dxi.first()].validTime(),
                dxi[dxi.last()].validTime());
  dx.zero();

  tlm_[0]->forecastAD(dx, *modaux_, length, no_post, costad);
  static_->multiplySqrtTrans(dx, dvo);
}

// -----------------------------------------------------------------------------
template <typename MODEL>
void StaticWithTLModelCovariance4D<MODEL>::advectedInverseMultiply(
    const Increment4D_ &dxi, Increment4D_ &dxo) const {
  dxo.zero();
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_STATICWITHTLMODELCOVARIANCE4D_H_
