/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGGEOMETRY_H_
#define OOPS_GENERIC_DUALCONFIGGEOMETRY_H_

#include <memory>
#include <ostream>
#include <string>

#include "oops/interface/Geometry.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class DualConfigGeometry
    : public util::Printable,
      private util::ObjectCounter<DualConfigGeometry<MODEL> > {
  using Geometry_ = Geometry<MODEL>;

 public:
  static const std::string classname() { return "oops::DualConfigGeometry"; }

  explicit DualConfigGeometry(const eckit::Configuration &);
  DualConfigGeometry(const DualConfigGeometry &);
  ~DualConfigGeometry();

  /// Interfacing
  const Geometry_ &geometry1() const { return *geom1_; }
  const Geometry_ &geometry2() const { return *geom2_; }

 private:
  void print(std::ostream &) const;

  std::shared_ptr<const Geometry_> geom1_;
  std::shared_ptr<const Geometry_> geom2_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigGeometry<MODEL>::DualConfigGeometry(const eckit::Configuration &conf)
    : geom1_(), geom2_() {
  Log::trace() << "DualConfigGeometry<MODELS>::DualConfigGeometry starting"
               << std::endl;
  util::Timer timer(classname(), "DualConfigGeometry");
  Log::debug() << "Full config is:" << conf << std::endl;
  {
    util::Timer timer(classname(), "DualConfigGeometry_1");
    eckit::LocalConfiguration conf1(conf, "resolution1");
    geom1_.reset(new Geometry_(conf1));
  }
  {
    util::Timer timer(classname(), "DualConfigGeometry_2");
    eckit::LocalConfiguration conf2(conf, "resolution2");
    geom2_.reset(new Geometry_(conf2));
  }
  Log::trace() << "DualConfigGeometry<MODELS>::DualConfigGeometry done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigGeometry<MODEL>::DualConfigGeometry(const DualConfigGeometry &other) {
  Log::trace() << "DualConfigGeometry<MODELS>::DualConfigGeometry copy starting"
               << std::endl;
  util::Timer timer(classname(), "DualConfigGeometry");
  {
    util::Timer timer(classname(), "DualConfigGeometry_1");
    geom1_(other.geom1_);
  }
  {
    util::Timer timer(classname(), "DualConfigGeometry_2");
    geom2_(other.geom2_);
  }
  Log::trace() << "DualConfigGeometry<MODELS>::DualConfigGeometry copy done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigGeometry<MODEL>::~DualConfigGeometry() {
  Log::trace() << "DualConfigGeometry<MODELS>::~DualConfigGeometry starting"
               << std::endl;
  util::Timer timer(classname(), "~DualConfigGeometry");
  {
    util::Timer timer(classname(), "~DualConfigGeometry_1");
    geom1_.reset();
  }
  {
    util::Timer timer(classname(), "~DualConfigGeometry_2");
    geom2_.reset();
  }
  Log::trace() << "DualConfigGeometry<MODELS>::~DualConfigGeometry done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigGeometry<MODEL>::print(std::ostream &os) const {
  Log::trace() << "DualConfigGeometry<MODELS>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  {
    util::Timer timer(classname(), "print_1");
    os << *geom1_;
  }
  {
    util::Timer timer(classname(), "print_2");
    os << *geom2_;
  }
  Log::trace() << "DualConfigGeometry<MODELS>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGGEOMETRY_H_
