/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGMODEL_H_
#define OOPS_GENERIC_DUALCONFIGMODEL_H_

#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/generic/DualConfigGeometry.h"
#include "oops/generic/DualConfigModelAuxControl.h"
#include "oops/generic/DualConfigPostProcessor.h"
#include "oops/generic/DualConfigState.h"
#include "oops/generic/DualConfigVariables.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/Duration.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

/// Encapsulates the nonlinear dual-config forecast model

// -----------------------------------------------------------------------------

template <typename MODEL>
class DualConfigModel : public util::Printable,
                        private eckit::NonCopyable,
                        private util::ObjectCounter<DualConfigModel<MODEL> > {
  using DualConfigGeometry_ = DualConfigGeometry<MODEL>;
  using DualConfigModel_ = DualConfigModel<MODEL>;
  using DualConfigModelAuxControl_ = DualConfigModelAuxControl<MODEL>;
  using DualConfigState_ = DualConfigState<MODEL>;
  using DualConfigVariables_ = DualConfigVariables<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using ModelAux_ = ModelAuxControl<MODEL>;
  using Model_ = Model<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::DualConfigModel"; }

  DualConfigModel(const DualConfigGeometry_ &, const eckit::Configuration &);
  ~DualConfigModel();

  // Run the forecast
  void forecast(DualConfigState_ &, const DualConfigModelAuxControl_ &,
                const util::Duration &,
                DualConfigPostProcessor<DualConfigState_, State_> &) const;

  // Information and diagnostics
  const util::Duration &timeResolution1() const {
    return model1_->timeResolution();
  }
  const util::Duration &timeResolution2() const {
    return model2_->timeResolution();
  }

  // Interfacing
  const Model_ &model1() const { return *model1_; }
  const Model_ &model2() const { return *model2_; }

 private:
  void initialize(DualConfigState_ &, const DualConfigModelAuxControl_ &) const;
  void step(DualConfigState_ &, const DualConfigModelAuxControl_ &) const;
  void finalize(DualConfigState_ &) const;
  void print(std::ostream &) const;

  std::unique_ptr<Model_> model1_;
  std::unique_ptr<Model_> model2_;
  const DualConfigVariables_ varsForced_;
  const DualConfigVariables_ varsForcedLagged_;
};

// =============================================================================

template <typename MODEL>
DualConfigModel<MODEL>::DualConfigModel(const DualConfigGeometry_ &resol,
                                        const eckit::Configuration &conf)
    : model1_(),
      model2_(),
      varsForced_(conf.getSubConfiguration("model1.forcing"),
                  conf.getSubConfiguration("model2.forcing")),
      varsForcedLagged_(conf.getSubConfiguration("model1.forcingLagged"),
                        conf.getSubConfiguration("model2.forcingLagged")) {
  Log::info() << classname() << "::DualConfigModel starting" << std::endl;
  util::Timer timer(classname(), "DualConfigModel");
  Log::info() << "Full config is:" << conf << std::endl;
  {
    util::Timer timer(classname(), "DualConfigModel_1");
    Log::info() << classname() << "::ctor Model_1 starting" << std::endl;
    eckit::LocalConfiguration conf1(conf, "model1");
    model1_.reset(new Model_(resol.geometry1(), conf1));
  }
  {
    util::Timer timer(classname(), "DualConfigModel_2");
    Log::info() << classname() << "::ctor Model_2 starting" << std::endl;
    eckit::LocalConfiguration conf2(conf, "model2");
    model2_.reset(new Model_(resol.geometry2(), conf2));
  }
  Log::info() << classname() << "::DualConfigModel done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigModel<MODEL>::~DualConfigModel() {
  Log::info() << classname() << "::~DualConfigModel starting" << std::endl;
  util::Timer timer(classname(), "~DualConfigModel");
  {
    util::Timer timer(classname(), "~DualConfigModel_1");
    model1_.reset();
  }
  {
    util::Timer timer(classname(), "~DualConfigModel_2");
    model2_.reset();
  }
  Log::info() << classname() << "::~DualConfigModel done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModel<MODEL>::forecast(
    DualConfigState_ &xx, const DualConfigModelAuxControl_ &maux,
    const util::Duration &len,
    DualConfigPostProcessor<DualConfigState_, State_> &post) const {
  Log::info() << classname() << "::forecast starting" << std::endl;
  util::Timer timer(classname(), "forecast");
  util::LogbookWriter<std::string> log("Forecast", "Nonlinear HR trajectory");

  const util::DateTime end(xx.validTime1() + len);
  Log::info() << classname() << "::forecast: starting 1: " << xx.state1()
              << std::endl;
  Log::info() << classname() << "::forecast: starting 2: " << xx.state2()
              << std::endl;

  /// Forecast initialization
  this->initialize(xx, maux);
  post.initialize(xx, timeResolution1(), timeResolution2(), end);

  /// Forecast time-stepping loop
  while (xx.validTime1() < end) {
    Log::info() << classname() << "::step starting" << std::endl
                << "  t1 = " << xx.validTime1() << std::endl
                << "  t2 = " << xx.validTime2() << std::endl;

    /// Two-way forcing
    xx.forceWith(varsForced_);

    /// Time step (includes forcing of lagged variables)
    this->step(xx, maux);

    /// Post-processing
    post.process(xx);

    Log::info() << classname() << "::step done" << std::endl;
  }

  /// Forecast finalization
  post.finalize(xx);
  this->finalize(xx);

  Log::info() << classname() << "::forecast: forecast finished: " << xx
              << std::endl;
  ASSERT(xx.validTime1() == end);
  ASSERT(xx.validTime2() == end);
  Log::info() << classname() << "::forecast done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModel<MODEL>::initialize(
    DualConfigState_ &xx, const DualConfigModelAuxControl_ &maux) const {
  Log::info() << classname() << "initialize starting" << std::endl;
  util::Timer timer(classname(), "initialize");
  {
    util::Timer timer(classname(), "initialize_1");
    Log::info() << classname() << "::initialize Model_1 starting" << std::endl;
    model1_->initialize(xx.state1(), maux.modelauxcontrol1());
  }
  {
    util::Timer timer(classname(), "initialize_2");
    Log::info() << classname() << "::initialize Model_2 starting" << std::endl;
    model2_->initialize(xx.state2(), maux.modelauxcontrol2());
  }
  Log::info() << classname() << "initialize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModel<MODEL>::step(
    DualConfigState_ &xx, const DualConfigModelAuxControl_ &maux) const {
  Log::info() << classname() << "::step starting" << std::endl;
  util::Timer timer(classname(), "step");
  /// Step Model 1
  {
    util::Timer timer(classname(), "step_1");
    Log::info() << classname() << "::step Model_1 starting"
                << " (" << xx.validTime1() << ")" << std::endl;
    model1_->step(xx.state1(), maux.modelauxcontrol1());
  }
  /// One-way forcing of lagged variables
  xx.forceWithLagged(varsForcedLagged_);
  /// Step Model 2
  while (xx.validTime2() < xx.validTime1()) {
    util::Timer timer(classname(), "step_2");
    Log::info() << classname() << "::step Model_2 starting"
                << " (" << xx.validTime2() << ")" << std::endl;
    model2_->step(xx.state2(), maux.modelauxcontrol2());
  }
  Log::info() << classname() << "::step done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModel<MODEL>::finalize(DualConfigState_ &xx) const {
  Log::info() << classname() << "::finalize starting" << std::endl;
  util::Timer timer(classname(), "finalize");
  {
    util::Timer timer(classname(), "finalize_1");
    Log::info() << classname() << "::finalize Model_1 starting" << std::endl;
    model1_->finalize(xx.state1());
  }
  {
    util::Timer timer(classname(), "finalize_2");
    Log::info() << classname() << "::finalize Model_2 starting" << std::endl;
    model2_->finalize(xx.state2());
  }
  Log::info() << classname() << "::finalize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModel<MODEL>::print(std::ostream &os) const {
  Log::info() << classname() << "::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  {
    util::Timer timer(classname(), "print_1");
    os << *model1_;
  }
  {
    util::Timer timer(classname(), "print_2");
    os << *model2_;
  }
  Log::info() << classname() << "::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGMODEL_H_
