/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGMODELAUXCONTROL_H_
#define OOPS_GENERIC_DUALCONFIGMODELAUXCONTROL_H_

#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "oops/generic/DualConfigGeometry.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxControl.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

template <typename MODEL>
class DualConfigModel;

// -----------------------------------------------------------------------------

template <typename MODEL>
class DualConfigModelAuxControl
    : public util::Printable,
      private util::ObjectCounter<DualConfigModelAuxControl<MODEL> > {
  using DualConfigModelAuxControl_ = DualConfigModelAuxControl<MODEL>;
  using DualConfigGeometry_ = DualConfigGeometry<MODEL>;
  using DualConfigModel_ = DualConfigModel<MODEL>;
  using ModelAuxControl_ = ModelAuxControl<MODEL>;
  using Geometry_ = Geometry<MODEL>;

 public:
  static const std::string classname() {
    return "oops::DualConfigModelAuxControl";
  }

  DualConfigModelAuxControl(const DualConfigGeometry_ &,
                            const DualConfigModel_ &,
                            const eckit::Configuration &);
  DualConfigModelAuxControl(const DualConfigGeometry_ &,
                            const DualConfigModelAuxControl &);
  explicit DualConfigModelAuxControl(const DualConfigModelAuxControl &,
                                     const bool copy = true);
  ~DualConfigModelAuxControl();

  /// Interfacing
  const ModelAuxControl_ &modelauxcontrol1() const { return *aux1_; }
  ModelAuxControl_ &modelauxcontrol1() { return *aux1_; }
  const ModelAuxControl_ &modelauxcontrol2() const { return *aux2_; }
  ModelAuxControl_ &modelauxcontrol2() { return *aux2_; }

  /// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

 private:
  DualConfigModelAuxControl &operator=(const DualConfigModelAuxControl &);
  void print(std::ostream &) const;

  std::unique_ptr<ModelAuxControl_> aux1_;
  std::unique_ptr<ModelAuxControl_> aux2_;
};

// =============================================================================

template <typename MODEL>
DualConfigModelAuxControl<MODEL>::DualConfigModelAuxControl(
    const DualConfigGeometry_ &resol, const DualConfigModel_ &model,
    const eckit::Configuration &conf)
    : aux1_(), aux2_() {
  Log::trace() << classname() << "::DualConfigModelAuxControl starting"
               << std::endl;
  util::Timer timer(classname(), "ctor");
  Log::debug() << "Full config is:" << conf << std::endl;
  {
    util::Timer timer(classname(), "ctor_1");
    eckit::LocalConfiguration conf1(conf, "state1");
    aux1_.reset(new ModelAuxControl_(resol.geometry1(), model.model1(), conf1));
  }
  {
    util::Timer timer(classname(), "ctor_2");
    eckit::LocalConfiguration conf2(conf, "state2");
    aux2_.reset(new ModelAuxControl_(resol.geometry2(), model.model2(), conf2));
  }
  Log::trace() << classname() << "::DualConfigModelAuxControl done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigModelAuxControl<MODEL>::DualConfigModelAuxControl(
    const DualConfigGeometry_ &resol, const DualConfigModelAuxControl &other)
    : aux1_(), aux2_() {
  Log::trace() << classname()
               << "::DualConfigModelAuxControl interpolated starting"
               << std::endl;
  util::Timer timer(classname(), "ctor");
  {
    util::Timer timer(classname(), "ctor_1");
    aux1_.reset(new ModelAuxControl_(resol.geometry1(), *other.aux1_));
  }
  {
    util::Timer timer(classname(), "ctor_2");
    aux2_.reset(new ModelAuxControl_(resol.geometry2(), *other.aux2_));
  }
  Log::trace() << classname() << "::DualConfigModelAuxControl interpolated done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigModelAuxControl<MODEL>::DualConfigModelAuxControl(
    const DualConfigModelAuxControl &other, const bool copy)
    : aux1_(), aux2_() {
  Log::trace() << classname() << "::DualConfigModelAuxControl copy starting"
               << std::endl;
  util::Timer timer(classname(), "ctor");
  {
    util::Timer timer(classname(), "ctor_1");
    aux1_.reset(new ModelAuxControl_(*other.aux1_, copy));
  }
  {
    util::Timer timer(classname(), "ctor_2");
    aux2_.reset(new ModelAuxControl_(*other.aux2_, copy));
  }
  Log::trace() << classname() << "::DualConfigModelAuxControl copy done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
DualConfigModelAuxControl<MODEL>::~DualConfigModelAuxControl() {
  Log::trace() << classname() << "::~DualConfigModelAuxControl starting"
               << std::endl;
  util::Timer timer(classname(), "dtor");
  {
    util::Timer timer(classname(), "dtor_1");
    aux1_.reset();
  }
  {
    util::Timer timer(classname(), "dtor_2");
    aux2_.reset();
  }
  Log::trace() << classname() << "::~DualConfigModelAuxControl done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModelAuxControl<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << classname() << "::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  Log::debug() << "Full config is:" << conf << std::endl;
  {
    util::Timer timer(classname(), "read_1");
    eckit::LocalConfiguration conf1(conf, "model1");
    aux1_->read(conf1);
  }
  {
    util::Timer timer(classname(), "read_2");
    eckit::LocalConfiguration conf2(conf, "model2");
    aux1_->read(conf2);
  }
  Log::trace() << classname() << "::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModelAuxControl<MODEL>::write(
    const eckit::Configuration &conf) const {
  Log::trace() << classname() << "::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  Log::debug() << "Full config is:" << conf << std::endl;
  {
    util::Timer timer(classname(), "write_1");
    eckit::LocalConfiguration conf1(conf, "model1");
    aux1_->write(conf1);
  }
  {
    util::Timer timer(classname(), "write_2");
    eckit::LocalConfiguration conf2(conf, "model2");
    aux2_->write(conf2);
  }
  Log::trace() << classname() << "::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double DualConfigModelAuxControl<MODEL>::norm() const {
  Log::trace() << classname() << "::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0.0;
  {
    util::Timer timer(classname(), "norm_1");
    zz += aux1_->norm();
  }
  {
    util::Timer timer(classname(), "norm_2");
    zz += aux2_->norm();
  }
  Log::trace() << classname() << "::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void DualConfigModelAuxControl<MODEL>::print(std::ostream &os) const {
  Log::trace() << classname() << "::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  {
    util::Timer timer(classname(), "print");
    os << *aux1_;
  }
  {
    util::Timer timer(classname(), "print");
    os << *aux2_;
  }
  Log::trace() << classname() << "::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGMODELAUXCONTROL_H_
