/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_DUALCONFIGPOSTPROCESSOR_H_
#define OOPS_GENERIC_DUALCONFIGPOSTPROCESSOR_H_

#include "oops/base/PostProcessor.h"
#include "oops/generic/DualConfigState.h"
#include "util/Timer.h"

namespace oops {

/// Control model post processing
/*!
 *  This class controls model post processing in the most general sense,
 *  ie all diagnostics computations that do not affect the dual config model
 *  integration. It just calls all the individual processors one by one.
 */

template <typename FLDS, typename FLD>
class DualConfigPostProcessor {
  // using DualConfigState_ = DualConfigState<MODEL>;
  using PostProcessor_ = PostProcessor<FLD>;

 public:
  static const std::string classname() {
    return "oops::DualConfigPostProcessor";
  }

  DualConfigPostProcessor() {}
  DualConfigPostProcessor(const DualConfigPostProcessor &pp)
      : post1_(pp.post1_), post2_(pp.post2_) {}
  ~DualConfigPostProcessor() {}

  /// Calls initialize method of individual processors
  void initialize(const FLDS &xx, const util::Duration &step1,
                  const util::Duration &step2, const util::DateTime &end) {
    util::Timer timer(classname(), "initialize");
    {
      util::Timer timer(classname(), "initialize_1");
      this->post1().initialize(xx.state1(), end, step1);
    }
    {
      util::Timer timer(classname(), "initialize_2");
      this->post2().initialize(xx.state2(), end, step2);
    }
  }

  /// Calls process method of individual processors
  void process(const FLDS &xx) {
    util::Timer timer(classname(), "process");
    {
      util::Timer timer(classname(), "process_1");
      this->post1().process(xx.state1());
    }
    {
      util::Timer timer(classname(), "process_2");
      if (xx.validTime2() == xx.validTime1())
        this->post2().process(xx.state2());
    }
  }

  /// Calls finalize method of individual processors
  void finalize(const FLDS &xx) {
    util::Timer timer(classname(), "finalize");
    {
      util::Timer timer(classname(), "finalize_1");
      this->post1().finalize(xx.state1());
    }
    {
      util::Timer timer(classname(), "finalize_1");
      this->post2().finalize(xx.state2());
    }
  }

  /// Access methods
  const PostProcessor_ &post1() const { return post1_; }
  const PostProcessor_ &post2() const { return post2_; }

  PostProcessor_ &post1() { return post1_; }
  PostProcessor_ &post2() { return post2_; }

 private:
  PostProcessor_ post1_;
  PostProcessor_ post2_;

  DualConfigPostProcessor operator=(const DualConfigPostProcessor &);
};

}  // namespace oops

#endif  // OOPS_GENERIC_DUALCONFIGPOSTPROCESSOR_H_
