/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#include "oops/generic/GenericCtlVec.h"

#include <netcdf.h>

#include <memory>
#include <random>
#include <string>

#include "atlas/array.h"
#include "atlas/field.h"
#include "eckit/exception/Exceptions.h"
#include "util/Logger.h"
#include "util/Mpi.h"
#include "util/Timer.h"
#include "util/abor1_cpp.h"

#define ERR(e, msg) {std::string s(nc_strerror(e)); ABORT(s + " : " + msg);}

namespace oops {
// =============================================================================
/// Constructors, destructor
// -----------------------------------------------------------------------------

GenericCtlVec::GenericCtlVec(const size_t &n) : n_(n) {
  Log::trace() << "GenericCtlVec::GenericCtlVec starting" << std::endl;
  util::Timer timer(classname(), "GenericCtlVec");

  // Initialize field
  data_ = atlas::Field("genericCtlVec", atlas::array::make_datatype<double>(),
                       atlas::array::make_shape(n_));
  this->zero();

  // Prepare I/O
  sendcounts_.resize(eckit::mpi::comm().size());
  eckit::mpi::comm().gather(static_cast<int>(n_), sendcounts_, 0);
  displs_.resize(eckit::mpi::comm().size());
  if (eckit::mpi::comm().rank() == 0) {
    int sum = 0;
    for (size_t i = 0; i < eckit::mpi::comm().size(); ++i) {
      displs_[i] = sum;
      sum += sendcounts_[i];
    }
  }
  nGlb_ = 0;
  for (auto & nLoc : sendcounts_)
    nGlb_ += nLoc;

  Log::trace() << "GenericCtlVec::GenericCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec::GenericCtlVec(const GenericCtlVec &other, const bool copy) :
  n_(other.n_), nGlb_(other.nGlb_), sendcounts_(other.sendcounts_), displs_(other.displs_) {
  Log::trace() << "GenericCtlVec::GenericCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "GenericCtlVec");
  data_ = atlas::Field("genericCtlVec", atlas::array::make_datatype<double>(),
                       atlas::array::make_shape(n_));
  if (copy) {
    auto view = atlas::array::make_view<double, 1>(data_);
    auto viewOther = atlas::array::make_view<double, 1>(other.data_);
    for (size_t jnode = 0; jnode < n_; ++jnode) {
      view(jnode) = viewOther(jnode);
    }
  }
  Log::trace() << "GenericCtlVec::GenericCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec::~GenericCtlVec() {
  Log::trace() << "GenericCtlVec::~GenericCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");
  Log::trace() << "GenericCtlVec::~GenericCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::zero() {
  Log::trace() << "GenericCtlVec::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  auto view = atlas::array::make_view<double, 1>(data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) = 0.0;
  }
  Log::trace() << "GenericCtlVec::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator=(const GenericCtlVec &rhs) {
  Log::trace() << "GenericCtlVec::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  ASSERT(rhs.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewRhs = atlas::array::make_view<double, 1>(rhs.data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) = viewRhs(jnode);
  }
  Log::trace() << "GenericCtlVec::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator+=(const GenericCtlVec &rhs) {
  Log::trace() << "GenericCtlVec::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  ASSERT(rhs.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewRhs = atlas::array::make_view<double, 1>(rhs.data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) += viewRhs(jnode);
  }
  Log::trace() << "GenericCtlVec::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator-=(const GenericCtlVec &rhs) {
  Log::trace() << "GenericCtlVec::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  ASSERT(rhs.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewRhs = atlas::array::make_view<double, 1>(rhs.data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) -= viewRhs(jnode);
  }
  Log::trace() << "GenericCtlVec::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

GenericCtlVec &GenericCtlVec::operator*=(const double &zz) {
  Log::trace() << "GenericCtlVec::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  auto view = atlas::array::make_view<double, 1>(data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) *= zz;
  }
  Log::trace() << "GenericCtlVec::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::axpy(const double &zz, const GenericCtlVec &dx,
                         const bool check) {
  Log::trace() << "GenericCtlVec::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  ASSERT(dx.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewDx = atlas::array::make_view<double, 1>(dx.data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) += zz * viewDx(jnode);
  }
  Log::trace() << "GenericCtlVec::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

double GenericCtlVec::local_dot_product_with(const GenericCtlVec &dx) const {
  Log::trace() << "GenericCtlVec::local_dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "local_dot_product_with");
  double zz = 0;
  ASSERT(dx.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewDx = atlas::array::make_view<double, 1>(dx.data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    zz += view(jnode) * viewDx(jnode);
  }
  Log::trace() << "GenericCtlVec::local_dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::schur_product_with(const GenericCtlVec &dx) {
  Log::trace() << "GenericCtlVec::schur_product_with starting" << std::endl;
  util::Timer timer(classname(), "schur_product_with");
  ASSERT(dx.size() == this->size());
  auto view = atlas::array::make_view<double, 1>(data_);
  auto viewDx = atlas::array::make_view<double, 1>(dx.data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) *= viewDx(jnode);
  }
  Log::trace() << "GenericCtlVec::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::random() {
  Log::trace() << "GenericCtlVec::random starting" << std::endl;
  util::Timer timer(classname(), "random");

  // Global random vector
  std::vector<double> dataGlb;

  if (eckit::mpi::comm().rank() == 0) {
    // Prepare random number generator
    static std::default_random_engine generator;
    std::normal_distribution<double> distribution(0.0, 1.0);

    // Generate global random vector
    dataGlb.resize(nGlb_);
    for (size_t jnodeGlb = 0; jnodeGlb < nGlb_; ++jnodeGlb) {
      dataGlb[jnodeGlb] = distribution(generator);
    }
  }

  // Scatter data from root task
  std::vector<double> dataLoc(n_);
  eckit::mpi::comm().scatterv(dataGlb.cbegin(), dataGlb.cend(), sendcounts_, displs_,
    dataLoc.begin(), dataLoc.end(), 0);

  // Copy data to field
  auto view = atlas::array::make_view<double, 1>(data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) = dataLoc[jnode];
  }
  Log::trace() << "GenericCtlVec::random done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::getSection(std::vector<double> &section, const int &iistart,
                               const int &iiend) const {
  Log::trace() << "GenericCtlVec::getSection starting" << std::endl;
  util::Timer timer(classname(), "getSection");
  ASSERT(iiend >= iistart);
  ASSERT(section.size() > static_cast<size_t>(iiend - iistart));
  auto view = atlas::array::make_view<double, 1>(data_);
  size_t index = 0;
  for (atlas::idx_t jnode = iistart; jnode <= iiend; ++jnode) {
    section[index] = view(jnode);
    ++index;
  }
  Log::trace() << "GenericCtlVec::getSection done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::setSection(const std::vector<double> &section,
                               const int &iistart, const int &iiend) {
  Log::trace() << "GenericCtlVec::setSection starting" << std::endl;
  util::Timer timer(classname(), "setSection");
  ASSERT(iiend >= iistart);
  ASSERT(section.size() > static_cast<size_t>(iiend - iistart));
  auto view = atlas::array::make_view<double, 1>(data_);
  size_t index = 0;
  for (atlas::idx_t jnode = iistart; jnode <= iiend; ++jnode) {
    view(jnode) = section[index];
    ++index;
  }
  Log::trace() << "GenericCtlVec::setSection done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::read(const eckit::Configuration &conf) {
  Log::trace() << "GenericCtlVec::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  // Read data on root task
  std::vector<double> dataGlb;
  if (eckit::mpi::comm().rank() == 0) {
    // Resize global data
    dataGlb.resize(nGlb_);

    // Get file name
    std::string ncFilePath = conf.getString("filepath");
    ncFilePath.append("_");
    ncFilePath.append(std::to_string(eckit::mpi::comm().size()));
    ncFilePath.append("pe.");
    ncFilePath.append(conf.getString("netcdf extension", "nc"));

    Log::info() << "Reading file: " << ncFilePath << std::endl;

    // NetCDF IDs
    int ncid, retval, nGlb_id, var_id;

    // Open NetCDF file
    if ((retval = nc_open(ncFilePath.c_str(), NC_NOWRITE, &ncid))) ERR(retval, ncFilePath);

    // Get dimension
    std::string dimname = "nGlb";
    if ((retval = nc_inq_dimid(ncid, dimname.c_str(), &nGlb_id))) ERR(retval, dimname);

    // Get and check dimension length
    size_t nGlbInFile;
    if ((retval = nc_inq_dimlen(ncid, nGlb_id, &nGlbInFile))) ERR(retval, dimname);
    ASSERT(nGlbInFile == nGlb_);

    // Get variable
    std::string varname = "genericCtlVec";
    if ((retval = nc_inq_varid(ncid, varname.c_str(), &var_id))) ERR(retval, varname);

    // Read data
    if ((retval = nc_get_var_double(ncid, var_id, dataGlb.data()))) ERR(retval, varname);

    // Close file
    if ((retval = nc_close(ncid))) ERR(retval, ncFilePath);
  }

  // Scatter data from root task
  std::vector<double> dataLoc(n_);
  eckit::mpi::comm().scatterv(dataGlb.cbegin(), dataGlb.cend(), sendcounts_, displs_,
    dataLoc.begin(), dataLoc.end(), 0);

  // Copy data to field
  auto view = atlas::array::make_view<double, 1>(data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    view(jnode) = dataLoc[jnode];
  }

  Log::trace() << "GenericCtlVec::read done" << std::endl;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::write(const eckit::Configuration &conf) const {
  Log::trace() << "GenericCtlVec::write starting" << std::endl;
  util::Timer timer(classname(), "write");

  // Copy data from field
  std::vector<double> dataLoc(n_);
  auto view = atlas::array::make_view<double, 1>(data_);
  for (size_t jnode = 0; jnode < n_; ++jnode) {
    dataLoc[jnode] = view(jnode);
  }

  // Gather data on root task
  std::vector<double> dataGlb;
  if (eckit::mpi::comm().rank() == 0) {
    dataGlb.resize(nGlb_);
  }
  eckit::mpi::comm().gatherv(dataLoc.cbegin(), dataLoc.cend(), dataGlb.begin(), dataGlb.end(),
    sendcounts_, displs_, 0);

  // Write data on root task
  if (eckit::mpi::comm().rank() == 0) {
    // Get file name
    std::string ncFilePath = conf.getString("filepath");
    ncFilePath.append("_");
    ncFilePath.append(std::to_string(eckit::mpi::comm().size()));
    ncFilePath.append("pe.");
    ncFilePath.append(conf.getString("netcdf extension", "nc"));

    Log::info() << "Writing file: " << ncFilePath << std::endl;

    // NetCDF IDs
    int ncid, retval, nGlb_id, d_id[1], var_id;

    // Create NetCDF file
    if ((retval = nc_create(ncFilePath.c_str(), NC_CLOBBER, &ncid))) ERR(retval, ncFilePath);

    // Create dimension
    std::string dimname = "nGlb";
    if ((retval = nc_def_dim(ncid, dimname.c_str(), nGlb_, &nGlb_id))) ERR(retval, dimname);
    d_id[0] = nGlb_id;

    // Create variable
    std::string varname = "genericCtlVec";
    if ((retval = nc_def_var(ncid, varname.c_str(), NC_DOUBLE, 1, d_id, &var_id))) ERR(retval, varname);

    // End definition mode
    if ((retval = nc_enddef(ncid))) ERR(retval, ncFilePath);

    // Write data
    if ((retval = nc_put_var_double(ncid, var_id, dataGlb.data()))) ERR(retval, varname);

    // Close file
    if ((retval = nc_close(ncid))) ERR(retval, ncFilePath);
  }

  Log::trace() << "GenericCtlVec::write done" << std::endl;
}

// -----------------------------------------------------------------------------

double GenericCtlVec::norm() const {
  Log::trace() << "GenericCtlVec::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = this->local_dot_product_with(*this);
  util::mpi::comm().allReduceInPlace(zz, eckit::mpi::sum());
  zz = std::sqrt(zz);
  Log::trace() << "GenericCtlVec::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

void GenericCtlVec::print(std::ostream &os) const {
  Log::trace() << "GenericCtlVec::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << "Generic control vector: " << std::endl;
  os << "- name: " << data_.name() << std::endl;
  os << "- size: " << data_.size() << std::endl;
  os << "- norm: " << this->norm() << std::endl;
  Log::trace() << "GenericCtlVec::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops
