/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_GENERIC_LOCALIZATIONBASE_H_
#define OOPS_GENERIC_LOCALIZATIONBASE_H_

#include <map>
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/generic/GenericCtlVec.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Variables.h"
#include "util/Logger.h"
#include "util/abor1_cpp.h"

namespace oops {

// Forward declaration
template <typename MODEL>
class IncrEnsCtlVec;

// -----------------------------------------------------------------------------

/*! @brief Abstract base class for generic localization

 *   LocalizationFactory produces the requested localizations.
 */
template <typename MODEL>
class LocalizationBase : public util::Printable {
  using Increment_ = Increment<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  /// Destructor
  virtual ~LocalizationBase() {}

  /// \f$ dx_1 = L dx_2 \f$
  virtual void multiply(Increment_ &) const = 0;
  /// \f$ dx = L^{1/2} dv \f$
  virtual void multiplySqrt(const GenericCtlVec &, Increment_ &) const = 0;
  /// \f$ dv = [L^{1/2}]^T dx \f$
  virtual void multiplySqrtTrans(const Increment_ &, GenericCtlVec &) const = 0;

  /// Control vector size
  virtual size_t ctlVecSize() const {
    ABORT("ctlVecSize not implemented for this localization");
    return 0;
  }

 private:
  virtual void print(std::ostream &) const = 0;

  std::string id_;
};

// -----------------------------------------------------------------------------

/// Localization Factory
template <typename MODEL>
class LocalizationFactory {
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static LocalizationBase<MODEL> *create(const Geometry_ &, const Variables_ &,
                                         const eckit::Configuration &);
  virtual ~LocalizationFactory() { makers_.clear(); }

 protected:
  explicit LocalizationFactory(const std::string &);

 private:
  virtual LocalizationBase<MODEL> *make(const Geometry_ &, const Variables_ &,
                                        const eckit::Configuration &) = 0;
  static std::map<std::string, LocalizationFactory<MODEL> *> makers_;
};

// -----------------------------------------------------------------------------

template <class MODEL, class LOCALIZATION>
class LocalizationMaker : public LocalizationFactory<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;

  virtual LocalizationBase<MODEL> *make(const Geometry_ &resol,
                                        const Variables_ &vars,
                                        const eckit::Configuration &conf) {
    return new LOCALIZATION(resol, vars, conf);
  }

 public:
  explicit LocalizationMaker(const std::string &name)
      : LocalizationFactory<MODEL>(name) {}
};

// =============================================================================

template <typename MODEL>
std::map<std::string, LocalizationFactory<MODEL> *>
    LocalizationFactory<MODEL>::makers_;

// -----------------------------------------------------------------------------

template <typename MODEL>
LocalizationFactory<MODEL>::LocalizationFactory(const std::string &name) {
  if (makers_.find(name) != makers_.end()) {
    Log::error() << name << " already registered in localization factory."
                 << std::endl;
    ABORT("Element already registered in LocalizationFactory.");
  }
  makers_[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
LocalizationBase<MODEL> *LocalizationFactory<MODEL>::create(
    const Geometry_ &resol, const Variables_ &vars,
    const eckit::Configuration &conf) {
  const std::string id = conf.getString("localization");
  Log::trace() << "LocalizationBase type = " << id << std::endl;
  typename std::map<std::string, LocalizationFactory<MODEL> *>::iterator j =
      makers_.find(id);
  if (j == makers_.end()) {
    Log::error() << id << " does not exist in localization factory."
                 << std::endl;
    Log::error() << "Localization factory contains " << makers_.size()
                 << " elements:" << std::endl;
    for (typename std::map<std::string,
                           LocalizationFactory<MODEL> *>::const_iterator jj =
             makers_.begin();
         jj != makers_.end(); ++jj) {
      Log::error() << "A" << jj->first << "B" << std::endl;
    }
    ABORT("Element does not exist in LocalizationFactory.");
  }
  return (*j).second->make(resol, vars, conf);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_GENERIC_LOCALIZATIONBASE_H_
