/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_GENERICMATRIX_H_
#define OOPS_INTERFACE_GENERICMATRIX_H_

#include <map>
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/base/Increment4D.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "util/Logger.h"
#include "util/Printable.h"
#include "util/abor1_cpp.h"

namespace oops {

// -----------------------------------------------------------------------------
/// Base class generic matrices.
/*!
 *
 */

template <typename MODEL>
class GenericMatrix : public util::Printable, private eckit::NonCopyable {
  using State_ = State<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Increment4D_ = Increment4D<MODEL>;

 public:
  GenericMatrix() {}
  virtual ~GenericMatrix() {}

  State_ &multiply(State_ &) const;

  Increment_ &multiply(Increment_ &) const;
  Increment4D_ &multiply(Increment4D_ &) const;
  Increment_ &inverseMultiply(Increment_ &) const;
  Increment_ &transposeInverseMultiply(Increment_ &) const;

  void accumul(State_ &xx, const State_ &rhs) const;

 private:
  virtual void multiply(typename MODEL::State &) const = 0;
  virtual void multiply(typename MODEL::Increment &) const = 0;
  virtual void inverseMultiply(typename MODEL::Increment &) const = 0;
  virtual void transposeInverseMultiply(typename MODEL::Increment &) const = 0;
  virtual void print(std::ostream &) const = 0;

  virtual void accumul(typename MODEL::State &xx,
                       const typename MODEL::State &rhs) const {}
};

// =============================================================================

/// GenericMatrixFactory Factory
template <typename MODEL>
class GenericMatrixFactory {
 public:
  static GenericMatrix<MODEL> *create(const eckit::Configuration &);
  virtual ~GenericMatrixFactory() { getMakers().clear(); }

 protected:
  explicit GenericMatrixFactory(const std::string &);

 private:
  virtual GenericMatrix<MODEL> *make(const eckit::Configuration &) = 0;
  static std::map<std::string, GenericMatrixFactory<MODEL> *> &getMakers() {
    static std::map<std::string, GenericMatrixFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <class MODEL, class T>
class GenericMatrixMaker : public GenericMatrixFactory<MODEL> {
  virtual GenericMatrix<MODEL> *make(const eckit::Configuration &conf) {
    return new T(conf);
  }

 public:
  explicit GenericMatrixMaker(const std::string &name)
      : GenericMatrixFactory<MODEL>(name) {}
};

// =============================================================================

template <typename MODEL>
GenericMatrixFactory<MODEL>::GenericMatrixFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in observation error factory."
                 << std::endl;
    ABORT("Element already registered in GenericMatrixFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
GenericMatrix<MODEL> *GenericMatrixFactory<MODEL>::create(
    const eckit::Configuration &conf) {
  Log::trace() << "GenericMatrix<MODEL>::create starting" << std::endl;
  const std::string id = conf.getString("matrix");
  typename std::map<std::string, GenericMatrixFactory<MODEL> *>::iterator jerr =
      getMakers().find(id);
  if (jerr == getMakers().end()) {
    Log::error() << id << " does not exist in matrix factory." << std::endl;
    ABORT("Element does not exist in GenericMatrixFactory.");
  }
  GenericMatrix<MODEL> *ptr = jerr->second->make(conf);
  Log::trace() << "GenericMatrix<MODEL>::create done" << std::endl;
  return ptr;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL> &GenericMatrix<MODEL>::multiply(State_ &xx) const {
  Log::trace() << "GenericMatrix<MODEL>::multiply starting" << std::endl;
  this->multiply(xx.state());
  Log::trace() << "GenericMatrix<MODEL>::multiply done" << std::endl;
  return xx;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> &GenericMatrix<MODEL>::multiply(Increment_ &dx) const {
  Log::trace() << "GenericMatrix<MODEL>::multiply starting" << std::endl;
  this->multiply(dx.increment());
  Log::trace() << "GenericMatrix<MODEL>::multiply done" << std::endl;
  return dx;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
Increment4D<MODEL> &GenericMatrix<MODEL>::multiply(Increment4D_ &dx) const {
  Log::trace() << "GenericMatrix<MODEL>::multiply starting" << std::endl;
  for (int jsub = dx.first(); jsub <= dx.last(); ++jsub) {
    this->multiply(dx[jsub].increment());
  }
  Log::trace() << "GenericMatrix<MODEL>::multiply done" << std::endl;
  return dx;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> &GenericMatrix<MODEL>::inverseMultiply(Increment_ &dx) const {
  Log::trace() << "GenericMatrix<MODEL>::multiply starting" << std::endl;
  this->inverseMultiply(dx.increment());
  Log::trace() << "GenericMatrix<MODEL>::multiply done" << std::endl;
  return dx;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Increment<MODEL> &GenericMatrix<MODEL>::transposeInverseMultiply(
    Increment_ &dx) const {
  Log::trace() << "GenericMatrix<MODEL>::transposeMultiply starting"
               << std::endl;
  this->transposeInverseMultiply(dx.increment());
  Log::trace() << "GenericMatrix<MODEL>::transposeMultiply done" << std::endl;
  return dx;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
void GenericMatrix<MODEL>::accumul(State_ &xx, const State_ &rhs) const {
  Log::trace() << "GenericMatrix<MODEL>::accumul starting" << std::endl;
  this->accumul(xx.state(), rhs.state());
  Log::trace() << "GenericMatrix<MODEL>::accumul done" << std::endl;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_GENERICMATRIX_H_
