/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_GEOVALS_H_
#define OOPS_INTERFACE_GEOVALS_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

template <typename MODEL>
class Increment;
template <typename MODEL>
class State;

// -----------------------------------------------------------------------------

/*! @brief Wrapper for geophysical values at observation locations

 *   Consider the \f$ J_o \f$ term of the cost function:

 *   \f$ J_o = \sum ({\cal G}_k (x_k) - y_k)^T R_k^{-1} ({\cal G}_k (x_k) - y_k)
     \f$

 *    ->  \f$ {\cal G}_k x_k = {\cal H}_k {\cal I}_k {\cal M}_{k:0} x_0 \f$;
 *    ->  \f$ {\cal I}_k \f$ is the interpolation operator, such that
 *       \f$ g_k = {\cal I}_k x_k \f$;

 *   In practice the model fields are interpolated to the observation locations
 *  during the trajectory and subsequently the observation operator is called:

 * \f{eqnarray*}{
      \sum {\cal G}_k (x_k) &=& {\cal H} \sum {\cal I}_k {\cal M}_{k:0} x_0 \\
                            &=& {\cal H} \sum {\cal I}_k x_k                \\
                            &=& {\cal H} \sum g_k                           \\
                            &=& {\cal H} g
   \f}

 *   GeoVaLs contains interpolated model values at observation locations
 *  \f$ g \f$.
 */

template <typename MODEL>
class GeoVaLs : public util::Printable,
                private eckit::NonCopyable,
                private util::ObjectCounter<GeoVaLs<MODEL> > {
  using GeoVaLs_ = typename MODEL::GeoVaLs;
  using Increment_ = Increment<MODEL>;
  using ObsSpace_ = ObservationSpace<MODEL>;
  using State_ = State<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::GeoVaLs"; }

  /// Constructor (takes an Increment)
  GeoVaLs(const ObsSpace_ &, const Variables_ &, const Increment_ &,
          const util::DateTime &, const util::DateTime &);
  /// Constructor (takes a State)
  GeoVaLs(const ObsSpace_ &, const Variables_ &, const State_ &,
          const util::DateTime &, const util::DateTime &);
  /// Destructor
  ~GeoVaLs();

  // Interfacing
  /// Return model-specific geovals
  const GeoVaLs_ &geovals() const { return *gvals_; }
  /// Return constant model-specific geovals
  GeoVaLs_ &geovals() { return *gvals_; }

  // Linear algebra and utilities, mostly for writing tests
  /// \f$ g = 0 \f$
  void zero();
  /// \f$ g = rand(size(g)) \f$
  void random();
  /// \f$ <g,g>=g^T g \f$
  double dot_product_with(const GeoVaLs &) const;
  /// Read geovals
  void read(const eckit::Configuration &);
  /// Write geovals
  void write(const eckit::Configuration &) const;

 private:
  void print(std::ostream &) const;
  std::unique_ptr<GeoVaLs_> gvals_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
GeoVaLs<MODEL>::GeoVaLs(const ObsSpace_ &os, const Variables_ &var,
                        const Increment_ &dx, const util::DateTime &t1,
                        const util::DateTime &t2)
    : gvals_() {
  Log::trace() << "GeoVaLs<MODEL>::GeoVaLs starting" << std::endl;
  util::Timer timer(classname(), "GeoVaLs");
  gvals_.reset(new GeoVaLs_(os.observationspace(), var.variables(),
                            dx.increment(), t1, t2));
  Log::trace() << "GeoVaLs<MODEL>::GeoVaLs done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
GeoVaLs<MODEL>::GeoVaLs(const ObsSpace_ &os, const Variables_ &var,
                        const State_ &xx, const util::DateTime &t1,
                        const util::DateTime &t2)
    : gvals_() {
  Log::trace() << "GeoVaLs<MODEL>::GeoVaLs starting" << std::endl;
  util::Timer timer(classname(), "GeoVaLs");
  gvals_.reset(
      new GeoVaLs_(os.observationspace(), var.variables(), xx.state(), t1, t2));
  Log::trace() << "GeoVaLs<MODEL>::GeoVaLs done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
GeoVaLs<MODEL>::~GeoVaLs() {
  Log::trace() << "GeoVaLs<MODEL>::~GeoVaLs starting" << std::endl;
  util::Timer timer(classname(), "~GeoVaLs");
  gvals_.reset();
  Log::trace() << "GeoVaLs<MODEL>::~GeoVaLs done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double GeoVaLs<MODEL>::dot_product_with(const GeoVaLs &other) const {
  Log::trace() << "GeoVaLs<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = gvals_->dot_product_with(other.gvals_);
  Log::trace() << "GeoVaLs<MODEL>::dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void GeoVaLs<MODEL>::zero() {
  Log::trace() << "GeoVaLs<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  gvals_->zero();
  Log::trace() << "GeoVaLs<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void GeoVaLs<MODEL>::random() {
  Log::trace() << "GeoVaLs<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");
  gvals_->random();
  Log::trace() << "GeoVaLs<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void GeoVaLs<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "GeoVaLs<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  gvals_->read(conf);
  Log::trace() << "GeoVaLs<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void GeoVaLs<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "GeoVaLs<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  gvals_->write(conf);
  Log::trace() << "GeoVaLs<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void GeoVaLs<MODEL>::print(std::ostream &os) const {
  Log::trace() << "GeoVaLs<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *gvals_;
  Log::trace() << "GeoVaLs<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_GEOVALS_H_
