/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_INTERFACE_HORIZSCALEDECOMPOSE_H_
#define OOPS_INTERFACE_HORIZSCALEDECOMPOSE_H_

#include <memory>
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class HorizScaleDecomposition
    : private eckit::NonCopyable,
      private util::ObjectCounter<HorizScaleDecomposition<MODEL> > {
  using HorizScaleDecomposition_ = typename MODEL::HorizScaleDecomposition;
  using Geometry_ = Geometry<MODEL>;
  using Increment_ = Increment<MODEL>;

 public:
  static const std::string classname() {
    return "oops::HorizScaleDecomposition";
  }

  HorizScaleDecomposition(const eckit::Configuration &);
  explicit HorizScaleDecomposition(const HorizScaleDecomposition &);
  virtual ~HorizScaleDecomposition();

  void collect(const Increment_ &, const int &) const;
  void decompose(const Increment_ &, const int &) const;

  /// Diagnostics
  friend std::ostream &operator<<(std::ostream &os,
                                  const HorizScaleDecomposition &loc) {
    // loc.print(os);
    return os;
  }

 private:
  std::unique_ptr<HorizScaleDecomposition_> horiz_sd_;
};

// =============================================================================

template <typename MODEL>
HorizScaleDecomposition<MODEL>::HorizScaleDecomposition(
    const eckit::Configuration &conf)
    : horiz_sd_() {
  Log::trace()
      << "HorizScaleDecomposition<MODEL>::HorizScaleDecomposition starting";
  util::Timer timer(classname(), "HorizScaleDecomposition");
  horiz_sd_.reset(new HorizScaleDecomposition_(conf));
  Log::trace()
      << "HorizScaleDecomposition<MODEL>::HorizScaleDecomposition done";
}

template <typename MODEL>
HorizScaleDecomposition<MODEL>::HorizScaleDecomposition(
    const HorizScaleDecomposition &other)
    : horiz_sd_() {
  Log::trace()
      << "HorizScaleDecomposition<MODEL>::HorizScaleDecomposition starting";
  util::Timer timer(classname(), "HorizScaleDecomposition");
  Log::trace()
      << "HorizScaleDecomposition<MODEL>::HorizScaleDecomposition done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
HorizScaleDecomposition<MODEL>::~HorizScaleDecomposition() {
  Log::trace()
      << "HorizScaleDecomposition<MODEL>::~HorizScaleDecomposition starting";
  util::Timer timer(classname(), "~HorizScaleDecomposition");
  horiz_sd_.reset();
  Log::trace()
      << "HorizScaleDecomposition<MODEL>::~HorizScaleDecomposition done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void HorizScaleDecomposition<MODEL>::collect(const Increment_ &dx,
                                             const int &band_index) const {
  Log::trace() << "HorizScaleDecomposition<MODEL>::horizScaleCollect starting";
  util::Timer timer(classname(), "horizScaleCollect");
  horiz_sd_->collect(dx.increment(), band_index);
  Log::trace() << "HorizScaleDecomposition<MODEL>::horizScaleCollect done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void HorizScaleDecomposition<MODEL>::decompose(const Increment_ &dx,
                                               const int &SDLtype) const {
  Log::trace() << "HorizScaleDecomposition<MODEL>::decompose starting";
  util::Timer timer(classname(), "decompose");
  horiz_sd_->decompose(dx.increment(), SDLtype);
  Log::trace() << "HorizScaleDecomposition<MODEL>::decompose done";
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_HORIZSCALEDECOMPOSE_H_
