/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_INCRENSCTLVEC_H_
#define OOPS_INTERFACE_INCRENSCTLVEC_H_

#include <memory>  // for std::unique_ptrs
#include <string>

#include "eckit/mpi/Comm.h"
#include "oops/generic/GenericCtlVec.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Localization.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

template <typename MODEL>
class Localization;

// -----------------------------------------------------------------------------

/*! @brief IncrEnsCtlVec class

 *  Part of the control vector (acted upon by the localized ensemble B matrix).
 * Consider a general formulation of the state component of the control vector
 * for a particular sub-window \f$ K \f$, having two components corresponding
 * to the modelled and ensemble based covariance matrices:

 *    \f[ dv^K = \begin{bmatrix}
                   dv^K_{mod}  \\
                   dv^K_{ens}
                 \end{bmatrix}
      \f]

    ->  \f$ dv^K \f$ is implemented by IncrCtlVec

 *  The component corresponding to the localized ensemble B matrix is defined
 * as:

 *    \f[ dv^K_{ens} = \begin{bmatrix}
                         dv^K_{ens_{1}} \\
                         ..             \\
                         dv^K_{ens_{L}}
                       \end{bmatrix}
      \f]

    ->  \f$ dv^K_{ens_m} \f$ is implemented by IncrEnsCtlVec, L is the
 *     number of ensemble perturbations used to construct the localized
 *     sampled covariance matrix;

 */

template <typename MODEL>
class IncrEnsCtlVec : public util::Printable,
                      private util::ObjectCounter<IncrEnsCtlVec<MODEL> > {
  using IncrEnsCtlVec_ = typename MODEL::IncrEnsCtlVec;
  using Localization_ = Localization<MODEL>;
  using Geometry_ = Geometry<MODEL>;

 public:
  static const std::string classname() { return "oops::IncrEnsCtlVec"; }

  /// Constructor
  IncrEnsCtlVec() : genVec_(), modVec_() {}
  /// Constructor
  IncrEnsCtlVec(const Geometry_ &, const Localization_ &);
  /// Constructor with copy initialization
  IncrEnsCtlVec(const Geometry_ &, const Localization_ &,
                const IncrEnsCtlVec &);
  /// Copy constructor
  IncrEnsCtlVec(const IncrEnsCtlVec &, const bool copy = true);
  /// Destructor
  virtual ~IncrEnsCtlVec();

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  void zero();
  /// \f$ dv_1 = dv_2 \f$
  IncrEnsCtlVec &operator=(const IncrEnsCtlVec &);
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  IncrEnsCtlVec &operator+=(const IncrEnsCtlVec &);
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  IncrEnsCtlVec &operator-=(const IncrEnsCtlVec &);
  /// \f$ dv_1 = dv_1 * c \f$
  IncrEnsCtlVec &operator*=(const double &);
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  void axpy(const double &, const IncrEnsCtlVec &, const bool check = true);
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  double local_dot_product_with(const IncrEnsCtlVec &) const;
  /// Hadamard product (element-wise) ( local in this PE )
  void schur_product_with(const IncrEnsCtlVec &);
  /// Random control vector
  void random();

  // I/O and diagnostics
  /// Read control vector
  void read(const eckit::Configuration &);
  /// Write control vector
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

  // Interfacing
  /// Get generic or ensemble control vector
  GenericCtlVec &genCtlVec() { return *genVec_; }
  IncrEnsCtlVec_ &modCtlVec() { return *modVec_; }
  /// Get constant generic or ensemble control vector
  const GenericCtlVec &genCtlVec() const { return *genVec_; }
  const IncrEnsCtlVec_ &modCtlVec() const { return *modVec_; }

 private:
  void print(std::ostream &) const;

  std::unique_ptr<GenericCtlVec> genVec_;   ///< Generic control vector
  std::unique_ptr<IncrEnsCtlVec_> modVec_;  ///< Model-specific implementation
};

// =============================================================================
/// Constructor, destructor
// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL>::IncrEnsCtlVec(const Geometry_ &resol,
                                    const Localization_ &loc)
    : genVec_(), modVec_() {
  /*! Dimensions of the ensemble control vector are determined by the geometry
   * and localization matrix.
   */

  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  if (loc.isGeneric()) {
    genVec_.reset(new GenericCtlVec(loc.genLoc().ctlVecSize()));
  } else {
    modVec_.reset(new IncrEnsCtlVec_(resol.geometry(), loc.modLoc()));
  }
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL>::IncrEnsCtlVec(const Geometry_ &resol,
                                    const Localization_ &loc,
                                    const IncrEnsCtlVec &other)
    : genVec_(), modVec_() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.genVec_) {
    genVec_.reset(new GenericCtlVec(loc.genLoc().ctlVecSize(), *other.genVec_));
  }
  if (other.modVec_) {
    modVec_.reset(
        new IncrEnsCtlVec_(resol.geometry(), loc.modLoc(), *other.modVec_));
  }
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL>::IncrEnsCtlVec(const IncrEnsCtlVec &other, const bool copy)
    : genVec_(), modVec_() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec copy starting"
               << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.genVec_) genVec_.reset(new GenericCtlVec(*other.genVec_, copy));
  if (other.modVec_) modVec_.reset(new IncrEnsCtlVec_(*other.modVec_, copy));
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL>::~IncrEnsCtlVec() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::~IncrEnsCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");
  genVec_.reset();
  modVec_.reset();
  Log::trace() << "IncrEnsCtlVec<MODEL>::~IncrEnsCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::zero() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  if (genVec_) genVec_->zero();
  if (modVec_) modVec_->zero();
  Log::trace() << "IncrEnsCtlVec<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL> &IncrEnsCtlVec<MODEL>::operator=(
    const IncrEnsCtlVec &rhs) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  if (genVec_) *genVec_ = *rhs.genVec_;
  if (modVec_) *modVec_ = *rhs.modVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL> &IncrEnsCtlVec<MODEL>::operator+=(
    const IncrEnsCtlVec &rhs) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  if (genVec_) *genVec_ += *rhs.genVec_;
  if (modVec_) *modVec_ += *rhs.modVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL> &IncrEnsCtlVec<MODEL>::operator-=(
    const IncrEnsCtlVec &rhs) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  if (genVec_) *genVec_ -= *rhs.genVec_;
  if (modVec_) *modVec_ -= *rhs.modVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrEnsCtlVec<MODEL> &IncrEnsCtlVec<MODEL>::operator*=(const double &zz) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  if (genVec_) *genVec_ *= zz;
  if (modVec_) *modVec_ *= zz;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::axpy(const double &zz, const IncrEnsCtlVec &dx,
                                const bool check) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  if (genVec_) genVec_->axpy(zz, *dx.genVec_, check);
  if (modVec_) modVec_->axpy(zz, *dx.modVec_, check);
  Log::trace() << "IncrEnsCtlVec<MODEL>::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double IncrEnsCtlVec<MODEL>::local_dot_product_with(
    const IncrEnsCtlVec &dx) const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::local_dot_product_with starting"
               << std::endl;
  util::Timer timer(classname(), "local_dot_product_with");
  double zz = 0;
  if (genVec_) zz = genVec_->local_dot_product_with(*dx.genVec_);
  if (modVec_) zz = modVec_->local_dot_product_with(*dx.modVec_);
  Log::trace() << "IncrEnsCtlVec<MODEL>::local_dot_product_with done"
               << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::schur_product_with(const IncrEnsCtlVec &dx) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::schur_product_with starting"
               << std::endl;
  util::Timer timer(classname(), "schur_product_with");
  if (genVec_) genVec_->schur_product_with(*dx.genVec_);
  if (modVec_) modVec_->schur_product_with(*dx.modVec_);
  Log::trace() << "IncrEnsCtlVec<MODEL>::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::random() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");
  if (genVec_) genVec_->random();
  if (modVec_) modVec_->random();
  Log::trace() << "IncrEnsCtlVec<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  if (genVec_) genVec_->read(conf);
  if (modVec_) modVec_->read(conf);
  Log::trace() << "IncrEnsCtlVec<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  if (genVec_) genVec_->write(conf);
  if (modVec_) modVec_->write(conf);
  Log::trace() << "IncrEnsCtlVec<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double IncrEnsCtlVec<MODEL>::norm() const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = this->local_dot_product_with(*this);
  eckit::mpi::comm().allReduceInPlace(zz, eckit::mpi::sum());
  zz = std::sqrt(zz);
  Log::trace() << "IncrEnsCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrEnsCtlVec<MODEL>::print(std::ostream &os) const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  if (genVec_) os << *genVec_;
  if (modVec_) os << *modVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_INCRENSCTLVEC_H_
