/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_INCRMODCTLVEC_H_
#define OOPS_INTERFACE_INCRMODCTLVEC_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "oops/generic/GenericCtlVec.h"
#include "oops/interface/ErrorCovariance.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief IncrModCtlVec class

 *  Part of the control vector (acted upon by the modelled background error
 * covariance matrix). Consider a general formulation of the state component
 * of the control vector for a particular sub-window \f$ K \f$, having two
 * components corresponding to the modelled and ensemble based covariance
 * matrices:

 *    \f[ dv^K = \begin{bmatrix}
                   dv^K_{mod}  \\
                   dv^K_{ens}
                 \end{bmatrix}
      \f]

    ->  \f$ dv^K \f$ is implemented by IncrCtlVec

    ->  \f$ dv^K_{mod} \f$ is implemented by IncrModCtlVec

 */

template <typename MODEL>
class IncrModCtlVec : public util::Printable,
                      private util::ObjectCounter<IncrModCtlVec<MODEL> > {
  using IncrModCtlVec_ = typename MODEL::IncrModCtlVec;
  using ErrorCovariance_ = ErrorCovariance<MODEL>;

 public:
  static const std::string classname() { return "oops::IncrModCtlVec"; }

  /// Constructor
  IncrModCtlVec();
  /// Constructor (generic)
  IncrModCtlVec(const size_t &);
  /// Constructor (model specific)
  IncrModCtlVec(const ErrorCovariance_ &);
  /// Constructor with copy initialization
  IncrModCtlVec(const ErrorCovariance_ &, const IncrModCtlVec &);
  /// Copy constructor
  IncrModCtlVec(const IncrModCtlVec &, const bool copy = true);
  /// Destructor
  virtual ~IncrModCtlVec();

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  void zero();
  /// \f$ dv_1 = dv_2 \f$
  IncrModCtlVec &operator=(const IncrModCtlVec &);
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  IncrModCtlVec &operator+=(const IncrModCtlVec &);
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  IncrModCtlVec &operator-=(const IncrModCtlVec &);
  /// \f$ dv_1 = dv_1 * c \f$
  IncrModCtlVec &operator*=(const double &);
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  void axpy(const double &, const IncrModCtlVec &, const bool check = true);
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  double dot_product_with(const IncrModCtlVec &) const;
  /// Hadamard product (element-wise)
  void schur_product_with(const IncrModCtlVec &);
  /// Random control vector
  void random();

  // Section (required for re-orthogonalization in LMP)
  /// Get section of the control vector
  void getSection(std::vector<double> &, const int &, const int &) const;
  /// Set section of the control vector
  void setSection(const std::vector<double> &, const int &, const int &);

  // I/O and diagnostics
  /// Read control vector
  void read(const eckit::Configuration &);
  /// Write control vector
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

  // Interfacing
  /// Get control vector
  GenericCtlVec &genCtlVec() { return *genVec_; }
  IncrModCtlVec_ &modCtlVec() { return *modVec_; }
  /// Get constant control vector
  const GenericCtlVec &genCtlVec() const { return *genVec_; }
  const IncrModCtlVec_ &modCtlVec() const { return *modVec_; }

 private:
  void print(std::ostream &) const;

  std::unique_ptr<GenericCtlVec> genVec_;   ///< Generic control vector
  std::unique_ptr<IncrModCtlVec_> modVec_;  ///< Model-specific implementation
};

// =============================================================================
/// Constructor, destructor
// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec() : genVec_(), modVec_() {}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const size_t &n) : genVec_(), modVec_() {
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec (generic) starting"
               << std::endl;
  util::Timer timer(classname(), "Incr");
  genVec_.reset(new GenericCtlVec(n));
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec (generic) done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const ErrorCovariance_ &cov)
    : genVec_(), modVec_() {
  Log::trace()
      << "IncrModCtlVec<MODEL>::IncrModCtlVec (model-specific) starting"
      << std::endl;
  util::Timer timer(classname(), "Incr");
  modVec_.reset(new IncrModCtlVec_(cov.errorCovariance()));
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec (model-specific) done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const ErrorCovariance_ &cov,
                                    const IncrModCtlVec &other)
    : genVec_(), modVec_() {
  Log::trace()
      << "IncrModCtlVec<MODEL>::IncrModCtlVec (model-specific) starting"
      << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.modVec_)
    modVec_.reset(new IncrModCtlVec_(cov.errorCovariance(), *other.modVec_));
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec (model-specific) done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const IncrModCtlVec &other, const bool copy)
    : genVec_(), modVec_() {
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec copy starting"
               << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.genVec_) genVec_.reset(new GenericCtlVec(*other.genVec_, copy));
  if (other.modVec_) modVec_.reset(new IncrModCtlVec_(*other.modVec_, copy));
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL>::~IncrModCtlVec() {
  Log::trace() << "IncrModCtlVec<MODEL>::~IncrModCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");
  genVec_.reset();
  modVec_.reset();
  Log::trace() << "IncrModCtlVec<MODEL>::~IncrModCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::zero() {
  Log::trace() << "IncrModCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  if (genVec_) genVec_->zero();
  if (modVec_) modVec_->zero();
  Log::trace() << "IncrModCtlVec<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

// template<typename MODEL>
// void IncrModCtlVec<MODEL>::zero(const util::DateTime & tt) {
//  Log::trace() << "IncrModCtlVec<MODEL>::zero starting" << std::endl;
//  util::Timer timer(classname(), "zero");
//  if (genVec_) genVec_->zero(tt);
//  if (modVec_) modVec_->zero(tt);
//  Log::trace() << "IncrModCtlVec<MODEL>::zero done" << std::endl;
//}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL> &IncrModCtlVec<MODEL>::operator=(
    const IncrModCtlVec &rhs) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  if (genVec_) *genVec_ = *rhs.genVec_;
  if (modVec_) *modVec_ = *rhs.modVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL> &IncrModCtlVec<MODEL>::operator+=(
    const IncrModCtlVec &rhs) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  if (genVec_) *genVec_ += *rhs.genVec_;
  if (modVec_) *modVec_ += *rhs.modVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL> &IncrModCtlVec<MODEL>::operator-=(
    const IncrModCtlVec &rhs) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  if (genVec_) *genVec_ -= *rhs.genVec_;
  if (modVec_) *modVec_ -= *rhs.modVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
IncrModCtlVec<MODEL> &IncrModCtlVec<MODEL>::operator*=(const double &zz) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  if (genVec_) *genVec_ *= zz;
  if (modVec_) *modVec_ *= zz;
  Log::trace() << "IncrModCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::axpy(const double &zz, const IncrModCtlVec &dx,
                                const bool check) {
  Log::trace() << "IncrModCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  if (genVec_) genVec_->axpy(zz, *dx.genVec_, check);
  if (modVec_) modVec_->axpy(zz, *dx.modVec_, check);
  Log::trace() << "IncrModCtlVec<MODEL>::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double IncrModCtlVec<MODEL>::dot_product_with(const IncrModCtlVec &dx) const {
  Log::trace() << "IncrModCtlVec<MODEL>::dot_product_with starting"
               << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = 0.0;
  if (genVec_) {
    double zzgen = genVec_->local_dot_product_with(*dx.genVec_);
    eckit::mpi::comm().allReduceInPlace(zzgen, eckit::mpi::sum());
    zz += zzgen;
  }
  if (modVec_) zz += modVec_->dot_product_with(*dx.modVec_);
  Log::trace() << "IncrModCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::schur_product_with(const IncrModCtlVec &dx) {
  Log::trace() << "IncrModCtlVec<MODEL>::schur_product_with starting"
               << std::endl;
  util::Timer timer(classname(), "schur_product_with");
  if (genVec_) genVec_->schur_product_with(*dx.genVec_);
  if (modVec_) modVec_->schur_product_with(*dx.modVec_);
  Log::trace() << "IncrModCtlVec<MODEL>::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::random() {
  Log::trace() << "IncrModCtlVec<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");
  if (genVec_) genVec_->random();
  if (modVec_) modVec_->random();
  Log::trace() << "IncrModCtlVec<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::getSection(std::vector<double> &section,
                                      const int &iistart,
                                      const int &iiend) const {
  Log::trace() << "IncrModCtlVec<MODEL>::getSection starting" << std::endl;
  util::Timer timer(classname(), "getSection");
  if (genVec_) genVec_->getSection(section, iistart, iiend);
  if (modVec_) modVec_->getSection(section, iistart, iiend);
  Log::trace() << "IncrModCtlVec<MODEL>::getSection done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::setSection(const std::vector<double> &section,
                                      const int &iistart, const int &iiend) {
  Log::trace() << "IncrModCtlVec<MODEL>::setSection starting" << std::endl;
  util::Timer timer(classname(), "setSection");
  if (genVec_) genVec_->setSection(section, iistart, iiend);
  if (modVec_) modVec_->setSection(section, iistart, iiend);
  Log::trace() << "IncrModCtlVec<MODEL>::setSection done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "IncrModCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  if (genVec_) genVec_->read(conf);
  if (modVec_) modVec_->read(conf);
  Log::trace() << "IncrModCtlVec<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "IncrModCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  if (genVec_) genVec_->write(conf);
  if (modVec_) modVec_->write(conf);
  Log::trace() << "IncrModCtlVec<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double IncrModCtlVec<MODEL>::norm() const {
  Log::trace() << "IncrModCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = std::sqrt(this->dot_product_with(*this));
  Log::trace() << "IncrModCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void IncrModCtlVec<MODEL>::print(std::ostream &os) const {
  Log::trace() << "IncrModCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  if (genVec_) os << *genVec_;
  if (modVec_) os << *modVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_INCRMODCTLVEC_H_
