/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_INTERFACE_INTERPOLATOR_H_
#define OOPS_INTERFACE_INTERPOLATOR_H_

#include <memory>
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class Interpolator : private eckit::NonCopyable,
                     private util::ObjectCounter<Interpolator<MODEL> > {
  using Interpolator_ = typename MODEL::Interpolator;
  using Geometry_ = Geometry<MODEL>;
  using Variables_ = Variables<MODEL>;
  using Increment_ = Increment<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() { return "oops::Interpolator"; }

  Interpolator(const State_ &, const eckit::Configuration &, const bool);
  Interpolator(const State_ &, const Geometry_ &, const eckit::Configuration &,
               const bool);
  explicit Interpolator(const Interpolator &);
  virtual ~Interpolator();

  Interpolator &operator=(const Interpolator &);
  Interpolator &operator+=(const State_ &);
  Interpolator &operator-=(const State_ &);

  void set(const Geometry_ &, const Geometry_ &);
  virtual void set_vector(const eckit::Configuration &);

  void interpolate(const Increment_ &, const Variables_ &, Increment_ &) const;
  void interpolate(const State_ &, const Variables_ &, State_ &) const;
  void smooth(Increment_ &) const;
  void save(Increment_ &) const;
  void read(const Increment_ &);
  void write(const eckit::Configuration &) const;

  /// Diagnostics
  friend std::ostream &operator<<(std::ostream &os, const Interpolator &loc) {
    loc.print(os);
    return os;
  }

 private:
  void print(std::ostream &) const;
  std::unique_ptr<Interpolator_> interp_;
};

// =============================================================================

template <typename MODEL>
Interpolator<MODEL>::Interpolator(const State_ &xb,
                                  const eckit::Configuration &conf,
                                  const bool is_adjoint)
    : interp_() {
  Log::trace() << "Interpolator<MODEL>::Interpolator starting";
  util::Timer timer(classname(), "Interpolator");
  interp_.reset(new Interpolator_(xb.state(), conf, is_adjoint));
  Log::trace() << "Interpolator<MODEL>::Interpolator done";
}

template <typename MODEL>
Interpolator<MODEL>::Interpolator(const State_ &xb, const Geometry_ &geo,
                                  const eckit::Configuration &conf,
                                  const bool is_adjoint)
    : interp_() {
  Log::trace()
      << "Interpolator<MODEL>::Interpolator with 2 geometries starting";
  util::Timer timer(classname(), "Interpolator");
  interp_.reset(
      new Interpolator_(xb.state(), geo.geometry(), conf, is_adjoint));
  Log::trace() << "Interpolator<MODEL>::Interpolator with 2 geometries done";
}

template <typename MODEL>
Interpolator<MODEL>::Interpolator(const Interpolator &other) : interp_() {
  Log::trace() << "Interpolator<MODEL>::Interpolator starting";
  util::Timer timer(classname(), "Interpolator");
  interp_.reset(new Interpolator_(*other.interp_));
  Log::trace() << "Interpolator<MODEL>::Interpolator done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Interpolator<MODEL>::~Interpolator() {
  Log::trace() << "Interpolator<MODEL>::~Interpolator starting";
  util::Timer timer(classname(), "~Interpolator");
  interp_.reset();
  Log::trace() << "Interpolator<MODEL>::~Interpolator done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Interpolator<MODEL> &Interpolator<MODEL>::operator=(const Interpolator &rhs) {
  Log::trace() << "Interpolator<MODEL>::operator= starting";
  util::Timer timer(classname(), "operator=");
  *interp_ = *rhs.interp_;
  Log::trace() << "Interpolator<MODEL>::operator= done";
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Interpolator<MODEL> &Interpolator<MODEL>::operator+=(const State_ &rhs) {
  Log::trace() << "Interpolator<MODEL>::operator+= starting";
  util::Timer timer(classname(), "operator+=");
  *interp_ += rhs.state();
  Log::trace() << "Interpolator<MODEL>::operator+= done";
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Interpolator<MODEL> &Interpolator<MODEL>::operator-=(const State_ &rhs) {
  Log::trace() << "Interpolator<MODEL>::operator+= starting";
  util::Timer timer(classname(), "operator+=");
  *interp_ -= rhs.state();
  Log::trace() << "Interpolator<MODEL>::operator+= done";
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::set(const Geometry_ &geo_in,
                              const Geometry_ &geo_out) {
  Log::trace() << "Interpolator<MODEL>::set starting";
  util::Timer timer(classname(), "set");
  interp_->set(geo_in.geometry(), geo_out.geometry());
  Log::trace() << "Interpolator<MODEL>::set done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::set_vector(const eckit::Configuration &conf) {
  Log::trace() << "Interpolator<MODEL>::set_vector starting";
  util::Timer timer(classname(), "set_vector");
  interp_->set_vector(conf);
  Log::trace() << "Interpolator<MODEL>::set_vector done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::interpolate(const Increment_ &dxi,
                                      const Variables_ &var,
                                      Increment_ &dxo) const {
  Log::trace() << "Interpolator<MODEL>::interpolate starting";
  util::Timer timer(classname(), "interpolate");
  interp_->interpolate(dxi.increment(), var.variables(), dxo.increment());
  Log::trace() << "Interpolator<MODEL>::interpolate done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::interpolate(const State_ &xxi, const Variables_ &var,
                                      State_ &xxo) const {
  Log::trace() << "Interpolator<MODEL>::interpolate starting";
  util::Timer timer(classname(), "interpolate");
  interp_->interpolate(xxi.state(), var.variables(), xxo.state());
  Log::trace() << "Interpolator<MODEL>::interpolate done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::smooth(Increment_ &dx) const {
  Log::trace() << "Interpolator<MODEL>::smooth starting";
  util::Timer timer(classname(), "smooth");
  interp_->smooth(dx.increment());
  Log::trace() << "Interpolator<MODEL>::smooth done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::save(Increment_ &dx) const {
  Log::trace() << "Interpolator<MODEL>::save starting";
  util::Timer timer(classname(), "save");
  interp_->save(dx.increment());
  Log::trace() << "Interpolator<MODEL>::save done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::read(const Increment_ &dx) {
  Log::trace() << "Interpolator<MODEL>::read starting";
  util::Timer timer(classname(), "read");
  interp_->read(dx.increment());
  Log::trace() << "Interpolator<MODEL>::read done";
}

// -----------------------------------------------------------------------------
template <typename MODEL>

void Interpolator<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "Interpolator<MODEL>::write starting";
  util::Timer timer(classname(), "write");
  interp_->write(conf);
  Log::trace() << "Interpolator<MODEL>::write done";
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Interpolator<MODEL>::print(std::ostream &os) const {
  Log::trace() << "Interpolator<MODEL>::print starting";
  util::Timer timer(classname(), "print");
  os << *interp_;
  Log::trace() << "Interpolator<MODEL>::print done";
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_INTERPOLATOR_H_
