/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_JCCOVAR_H_
#define OOPS_INTERFACE_JCCOVAR_H_

#include <memory>
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/JcCovarBase.h"
#include "oops/interface/State.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

/*! @brief JcCovar class

 *   This class provides the operations associated with the Jc covariance
 *  matrix. It wraps the actual JcCovarBase object which can be a model
 *  specific one or a generic one (achieved through polymorphism). The
 *  interface for the Jc covariance comprises two levels (JcCovar and
 *  JcCovarBase) because we want run time polymorphism.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class JcCovar : public util::Printable,
                private eckit::NonCopyable,
                private util::ObjectCounter<JcCovar<MODEL> > {
  using Increment_ = Increment<MODEL>;
  using JcCovarBase_ = JcCovarBase<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() { return "oops::JcCovar"; }

  /// Constructor
  JcCovar(const State_ &, const Geometry_ &, const eckit::Configuration &);
  /// Destructor
  ~JcCovar();

  // Methods
  /// Set up the covariance matrix at inner loop resolution
  void linearize(const State_ &, const Geometry_ &);
  /// \f$ dx_1 = C dx_2 \f$
  void multiply(const Increment_ &, Increment_ &) const;
  /// \f$ dx_1 = C^{-1} dx_2 \f$
  void inverseMultiply(const Increment_ &, Increment_ &) const;

 private:
  /// Diagnostics
  void print(std::ostream &) const;

  std::unique_ptr<JcCovarBase_>
      jccov_;  ///< Pointer to a JcCovarBase class (abstract)
  std::unique_ptr<Geometry_> geom_;  ///< Pointer to a Geometry object
};

// =============================================================================

template <typename MODEL>
JcCovar<MODEL>::JcCovar(const State_ &state, const Geometry_ &geom,
                        const eckit::Configuration &conf)
    : jccov_(), geom_() {
  Log::trace() << "JcCovar<MODEL>::JcCovar starting" << std::endl;
  util::Timer timer(classname(), "JcCovar");
  Log::debug() << "JcCovar config is:" << conf << std::endl;
  jccov_.reset(JcCovarFactory<MODEL>::create(state, geom, conf));
  geom_.reset(new Geometry_(geom));
  Log::trace() << "JcCovar<MODEL>::JcCovar done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
JcCovar<MODEL>::~JcCovar() {
  Log::trace() << "JcCovar<MODEL>::~JcCovar starting" << std::endl;
  util::Timer timer(classname(), "~JcCovar");
  jccov_.reset();
  Log::trace() << "JcCovar<MODEL>::~JcCovar done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void JcCovar<MODEL>::linearize(const State_ &x, const Geometry_ &geom) {
  Log::trace() << "JcCovar<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  jccov_->linearize(x, geom);
  Log::trace() << "JcCovar<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void JcCovar<MODEL>::multiply(const Increment_ &dx_in,
                              Increment_ &dx_out) const {
  Log::trace() << "JcCovar<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  jccov_->multiply(dx_in, dx_out);
  Log::trace() << "JcCovar<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void JcCovar<MODEL>::inverseMultiply(const Increment_ &dx_in,
                                     Increment_ &dx_out) const {
  Log::trace() << "JcCovar<MODEL>::inverseMultiply starting" << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  jccov_->inverseMultiply(dx_in, dx_out);
  Log::trace() << "JcCovar<MODEL>::inverseMultiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void JcCovar<MODEL>::print(std::ostream &os) const {
  Log::trace() << "JcCovar<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *jccov_;
  Log::trace() << "JcCovar<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_JCCOVAR_H_
