/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_LINEARMODEL_H_
#define OOPS_INTERFACE_LINEARMODEL_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/LinearModelBase.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "oops/interface/State.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class LocalConfiguration;
class Configuration;
}  // namespace eckit

namespace oops {

/*! @brief Encapsulates the linear forecast model.

 *   This class provides the operations associated with the LinearModel. It
 *  wraps the actual linear model which can be a model specific one or a
 *  generic one (identity). The interface for the linear model comprises two
 *  levels (LinearModel and LinearModelBase) because we want run time
 *  polymorphism. LinearModelFactory produces concrete objects of
 *  LinearModelBase children classes.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class LinearModel : public util::Printable,
                    private eckit::NonCopyable,
                    private util::ObjectCounter<LinearModel<MODEL> > {
  using LinearModelBase_ = LinearModelBase<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using ModelAux_ = ModelAuxControl<MODEL>;
  using ModelAuxIncr_ = ModelAuxIncrement<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() { return "oops::LinearModel"; }

  /// Constructor
  LinearModel(const Geometry_ &, const eckit::Configuration &);
  /// Destructor
  ~LinearModel();

  /// Run the tangent linear forecast
  void forecastTL(
      Increment_ &, const ModelAuxIncr_ &, const util::Duration &,
      PostProcessor<Increment_> post = PostProcessor<Increment_>(),
      PostProcessorTL<Increment_> cost = PostProcessorTL<Increment_>(),
      const bool idmodel = false) const;

  /// Run the adjoint forecast
  void forecastAD(
      Increment_ &, ModelAuxIncr_ &, const util::Duration &,
      PostProcessor<Increment_> post = PostProcessor<Increment_>(),
      PostProcessorAD<Increment_> cost = PostProcessorAD<Increment_>(),
      const bool idmodel = false) const;

  /// Set the linearization trajectory
  void setTrajectory(State_ &, const ModelAux_ &);

  /// Access the linearisation trajectory model
  const Model_ &getTrajectoryModel() const { return *lrmodel_; }

  // Information and diagnostics
  /// Return time step
  const util::Duration &timeResolution() const {
    return tlm_->timeResolution();
  }

  // Interfacing
  /// Return a constant reference to the Geometry object
  const Geometry_ &geom() const { return *geom_; }
  /// Return a constant reference to the LinearModelBase object
  const LinearModelBase_ &model() const { return *tlm_; }
  /// Return a constant reference to the Configuration object
  const eckit::LocalConfiguration &conf() const { return conf_; }

 protected:
  // Run the TL forecast
  /// Initialize the TL model
  void initializeTL(Increment_ &, const ModelAuxIncr_ &) const;
  /// Make a time step TL
  void stepTL(Increment_ &, const ModelAuxIncr_ &) const;
  /// Finalize the TL model
  void finalizeTL(Increment_ &) const;

  // Run the AD forecast
  /// Initialize the adjoint model
  void initializeAD(Increment_ &, const ModelAuxIncr_ &) const;
  /// Make a time step AD
  void stepAD(Increment_ &, ModelAuxIncr_ &) const;
  /// Finalize the adjoint model
  void finalizeAD(Increment_ &) const;

 private:
  // diagnostics
  void print(std::ostream &) const;

  const eckit::LocalConfiguration conf_;
  std::unique_ptr<LinearModelBase_>
      tlm_;                          ///< Pointer to a LinearModelBase object
  std::unique_ptr<Geometry_> geom_;  ///< Pointer to a Geometry object
  std::unique_ptr<Model_> lrmodel_;  ///< Pointer to the linearisation model
};

// =============================================================================

template <typename MODEL>
LinearModel<MODEL>::LinearModel(const Geometry_ &resol,
                                const eckit::Configuration &conf)
    : conf_(conf), tlm_(), geom_(), lrmodel_() {
  Log::trace() << "LinearModel<MODEL>::LinearModel starting" << std::endl;
  util::Timer timer(classname(), "LinearModel");
  Log::debug() << "LinearModel config is:" << conf << std::endl;
  tlm_.reset(LinearModelFactory<MODEL>::create(resol, conf));
  geom_.reset(new Geometry_(resol));
  lrmodel_.reset(new Model_(resol, conf.getSubConfiguration("trajectory")));
  Log::trace() << "LinearModel<MODEL>::LinearModel done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearModel<MODEL>::~LinearModel() {
  Log::trace() << "LinearModel<MODEL>::~LinearModel starting" << std::endl;
  util::Timer timer(classname(), "~LinearModel");
  lrmodel_.reset();
  geom_.reset();
  tlm_.reset();
  Log::trace() << "LinearModel<MODEL>::~LinearModel done" << std::endl;
}

// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
/// Run forecast TL and AD
// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::forecastTL(Increment_ &dx, const ModelAuxIncr_ &mctl,
                                    const util::Duration &len,
                                    PostProcessor<Increment_> post,
                                    PostProcessorTL<Increment_> cost,
                                    const bool idmodel) const {
  /*!  The linear model forecast receives PostProcessor and PostProcessorTL;
   *  PostProcessor can be used for post-processing (storing/writing of an
   *  Increment at requested lead times) and PostProcessorTL can be used to
   *  linear operators (e.g. \f$ H \f$ matrix).

   *   The model error term is handled internally by the model.
   */
  Log::trace() << "LinearModel<MODEL>::forecastTL starting" << std::endl;
  util::Timer timer(classname(), "forecastTL");

  const util::DateTime end(dx.validTime() + len);
  const util::Duration tstep(tlm_->timeResolution());
  Log::info() << "LinearModel<MODEL>::forecastTL: Starting " << dx << std::endl;
  this->initializeTL(dx, mctl);
  cost.initializeTL(dx, end, tstep);
  post.initialize(dx, end, tstep);
  if (idmodel) {
    while (dx.validTime() < end) {
      dx.updateTime(tstep);
      cost.processTL(dx);
      post.process(dx);
    }
  } else {
    while (dx.validTime() < end) {
      this->stepTL(dx, mctl);
      cost.processTL(dx);
      post.process(dx);
    }
  }
  cost.finalizeTL(dx);
  post.finalize(dx);
  this->finalizeTL(dx);
  Log::info() << "LinearModel<MODEL>::forecastTL: Finished " << dx << std::endl;
  ASSERT(dx.validTime() == end);

  Log::trace() << "LinearModel<MODEL>::forecastTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::forecastAD(Increment_ &dx, ModelAuxIncr_ &mctl,
                                    const util::Duration &len,
                                    PostProcessor<Increment_> post,
                                    PostProcessorAD<Increment_> cost,
                                    const bool idmodel) const {
  /*!  The adjoint model forecast receives PostProcessor and PostProcessorAD;
   *  PostProcessor can be used for post-processing (storing/writing of an
   *  Increment at requested lead times) and PostProcessorAD can be used to
   *  evaluate the adjoint of linear operators (e.g. \f$ H^T \f$ matrix).

   *   The model error term is handled internally by the model.
   */
  Log::trace() << "LinearModel<MODEL>::forecastAD starting" << std::endl;
  util::Timer timer(classname(), "forecastAD");

  const util::DateTime bgn(dx.validTime() - len);
  const util::Duration tstep(tlm_->timeResolution());
  Log::info() << "LinearModel<MODEL>::forecastAD: Starting " << dx << std::endl;
  this->initializeAD(dx, mctl);
  post.initialize(dx, bgn, tstep);
  cost.initializeAD(dx, bgn, tstep);
  if (idmodel) {
    while (dx.validTime() > bgn) {
      cost.processAD(dx);
      dx.updateTime(-tstep);
      post.process(dx);
    }
  } else {
    while (dx.validTime() > bgn) {
      cost.processAD(dx);
      this->stepAD(dx, mctl);
      post.process(dx);
    }
  }
  cost.finalizeAD(dx);
  post.finalize(dx);
  this->finalizeAD(dx);
  Log::info() << "LinearModel<MODEL>::forecastAD: Finished " << dx << std::endl;
  ASSERT(dx.validTime() == bgn);

  Log::trace() << "LinearModel<MODEL>::forecastAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::setTrajectory(State_ &xlr, const ModelAux_ &maux) {
  /*!  The low resolution linearisation trajectory model is used to construct
   *  the linearisation trajectory.
   */
  Log::trace() << "LinearModel<MODEL>::setTrajectory starting" << std::endl;
  util::Timer timer(classname(), "setTrajectory");
  tlm_->setTrajectory(xlr, *lrmodel_, maux);
  Log::trace() << "LinearModel<MODEL>::setTrajectory done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::initializeTL(Increment_ &dx,
                                      const ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModel<MODEL>::initializeTL starting" << std::endl;
  util::Timer timer(classname(), "initializeTL");
  tlm_->initializeTL(dx, merr);
  Log::trace() << "LinearModel<MODEL>::initializeTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::stepTL(Increment_ &dx,
                                const ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModel<MODEL>::stepTL starting" << std::endl;
  util::Timer timer(classname(), "stepTL");
  tlm_->stepTL(dx, merr);
  Log::trace() << "LinearModel<MODEL>::stepTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::finalizeTL(Increment_ &dx) const {
  Log::trace() << "LinearModel<MODEL>::finalizeTL starting" << std::endl;
  util::Timer timer(classname(), "finalizeTL");
  tlm_->finalizeTL(dx);
  Log::trace() << "LinearModel<MODEL>::finalizeTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::initializeAD(Increment_ &dx,
                                      const ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModel<MODEL>::initializeAD starting" << std::endl;
  util::Timer timer(classname(), "initializeAD");
  tlm_->initializeAD(dx, merr);
  Log::trace() << "LinearModel<MODEL>::initializeAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::stepAD(Increment_ &dx, ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModel<MODEL>::stepAD starting" << std::endl;
  util::Timer timer(classname(), "stepAD");
  tlm_->stepAD(dx, merr);
  Log::trace() << "LinearModel<MODEL>::stepAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::finalizeAD(Increment_ &dx) const {
  Log::trace() << "LinearModel<MODEL>::finalizeAD starting" << std::endl;
  util::Timer timer(classname(), "finalizeAD");
  tlm_->finalizeAD(dx);
  Log::trace() << "LinearModel<MODEL>::finalizeAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModel<MODEL>::print(std::ostream &os) const {
  Log::trace() << "LinearModel<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *tlm_;
  Log::trace() << "LinearModel<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_LINEARMODEL_H_
