/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_LINEARMODELBASE_H_
#define OOPS_INTERFACE_LINEARMODELBASE_H_

#include <map>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "oops/interface/State.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"
#include "util/abor1_cpp.h"

namespace eckit {
class Configuration;
}

namespace oops {

/*!  @brief Base class for encapsulation of the linear forecast model.

 *   Defines the interfaces for the linear model. LinearModelFactory is used
 *  to produce concrete linear models (model specific or generic).

 *   The public methods (setTrajectory, initializeTL, ..) wrap around the
 *  corresponding protected virtual methods (setTrajectory, initializeTL, ..)
 *  which are implemented by child model specific or generic classes.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class LinearModelBase : public util::Printable, private eckit::NonCopyable {
  using Increment_ = Increment<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using ModelAux_ = ModelAuxControl<MODEL>;
  using ModelAuxIncr_ = ModelAuxIncrement<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() { return "oops::LinearModelBase"; }

  /// Constructor
  LinearModelBase() {}
  /// Destructor
  virtual ~LinearModelBase() {}

  /// Set the linearization trajectory
  void setTrajectory(State_ &, const Model_ &, const ModelAux_ &);

  // Run the TL forecast
  /// Initialize the TL model
  void initializeTL(Increment_ &, const ModelAuxIncr_ &) const;
  /// Make a time step TL
  void stepTL(Increment_ &, const ModelAuxIncr_ &) const;
  /// Finalize the TL model
  void finalizeTL(Increment_ &) const;

  // Run the AD forecast
  /// Initialize the adjoint model
  void initializeAD(Increment_ &, const ModelAuxIncr_ &) const;
  /// Make a time step AD
  void stepAD(Increment_ &, ModelAuxIncr_ &) const;
  /// Finalize the adjoint model
  void finalizeAD(Increment_ &) const;

  // Information and diagnostics
  /// Return time step
  virtual const util::Duration &timeResolution() const = 0;

 protected:
  // Set the linearization trajectory
  virtual void setTrajectory(typename MODEL::State &,
                             const typename MODEL::Model &,
                             const typename MODEL::ModelAuxControl &) = 0;

  // Run the TL forecast
  virtual void initializeTL(
      typename MODEL::Increment &,
      const typename MODEL::ModelAuxIncrement &) const = 0;
  virtual void stepTL(typename MODEL::Increment &,
                      const typename MODEL::ModelAuxIncrement &) const = 0;
  virtual void finalizeTL(typename MODEL::Increment &) const = 0;

  // Run the AD forecast
  virtual void initializeAD(
      typename MODEL::Increment &,
      const typename MODEL::ModelAuxIncrement &) const = 0;
  virtual void stepAD(typename MODEL::Increment &,
                      typename MODEL::ModelAuxIncrement &) const = 0;
  virtual void finalizeAD(typename MODEL::Increment &) const = 0;

  // Information and diagnostics
  virtual void print(std::ostream &) const = 0;
};

// =============================================================================

/// LinearModelFactory Factory
template <typename MODEL>
class LinearModelFactory {
  using Geometry_ = Geometry<MODEL>;

 public:
  static LinearModelBase<MODEL> *create(const Geometry_ &,
                                        const eckit::Configuration &);
  virtual ~LinearModelFactory() { getMakers().clear(); }

 protected:
  explicit LinearModelFactory(const std::string &);

 private:
  virtual LinearModelBase<MODEL> *make(const Geometry_ &,
                                       const eckit::Configuration &) = 0;
  static std::map<std::string, LinearModelFactory<MODEL> *> &getMakers() {
    static std::map<std::string, LinearModelFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <class MODEL, class T>
class LinearModelMaker : public LinearModelFactory<MODEL> {
  using Geometry_ = Geometry<MODEL>;
  virtual LinearModelBase<MODEL> *make(const Geometry_ &geom,
                                       const eckit::Configuration &conf) {
    return new T(geom.geometry(), conf);
  }

 public:
  explicit LinearModelMaker(const std::string &name)
      : LinearModelFactory<MODEL>(name) {}
};

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearModelFactory<MODEL>::LinearModelFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name
                 << " already registered in the tangent linear model factory."
                 << std::endl;
    ABORT("Element already registered in LinearModelFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearModelBase<MODEL> *LinearModelFactory<MODEL>::create(
    const Geometry_ &geom, const eckit::Configuration &conf) {
  Log::trace() << "LinearModelBase<MODEL>::create starting" << std::endl;
  const std::string id = conf.getString("version");
  typename std::map<std::string, LinearModelFactory<MODEL> *>::iterator jerr =
      getMakers().find(id);
  if (jerr == getMakers().end()) {
    Log::error() << id << " does not exist in the tangent linear model factory."
                 << std::endl;
    ABORT("Element does not exist in LinearModelFactory.");
  }
  LinearModelBase<MODEL> *ptr = jerr->second->make(geom, conf);
  Log::trace() << "LinearModelBase<MODEL>::create done" << std::endl;
  return ptr;
}

// =============================================================================

template <typename MODEL>
void LinearModelBase<MODEL>::setTrajectory(State_ &xlr, const Model_ &lrmodel,
                                           const ModelAux_ &maux) {
  Log::trace() << "LinearModelBase<MODEL>::setTrajectory starting" << std::endl;
  util::Timer timer(classname(), "setTrajectory");
  this->setTrajectory(xlr.state(), lrmodel.model(), maux.modelauxcontrol());
  Log::trace() << "LinearModelBase<MODEL>::setTrajectory done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModelBase<MODEL>::initializeTL(Increment_ &dx,
                                          const ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModelBase<MODEL>::initializeTL starting" << std::endl;
  util::Timer timer(classname(), "initializeTL");
  this->initializeTL(dx.increment(), merr.modelauxincrement());
  Log::trace() << "LinearModelBase<MODEL>::initializeTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModelBase<MODEL>::stepTL(Increment_ &dx,
                                    const ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModelBase<MODEL>::stepTL starting" << std::endl;
  util::Timer timer(classname(), "stepTL");
  this->stepTL(dx.increment(), merr.modelauxincrement());
  Log::trace() << "LinearModelBase<MODEL>::stepTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModelBase<MODEL>::finalizeTL(Increment_ &dx) const {
  Log::trace() << "LinearModelBase<MODEL>::finalizeTL starting" << std::endl;
  util::Timer timer(classname(), "finalizeTL");
  this->finalizeTL(dx.increment());
  Log::trace() << "LinearModelBase<MODEL>::finalizeTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModelBase<MODEL>::initializeAD(Increment_ &dx,
                                          const ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModelBase<MODEL>::initializeAD starting" << std::endl;
  util::Timer timer(classname(), "initializeAD");
  this->initializeAD(dx.increment(), merr.modelauxincrement());
  Log::trace() << "LinearModelBase<MODEL>::initializeAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModelBase<MODEL>::stepAD(Increment_ &dx, ModelAuxIncr_ &merr) const {
  Log::trace() << "LinearModelBase<MODEL>::stepAD starting" << std::endl;
  util::Timer timer(classname(), "stepAD");
  this->stepAD(dx.increment(), merr.modelauxincrement());
  Log::trace() << "LinearModelBase<MODEL>::stepAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearModelBase<MODEL>::finalizeAD(Increment_ &dx) const {
  Log::trace() << "LinearModelBase<MODEL>::finalizeAD starting" << std::endl;
  util::Timer timer(classname(), "finalizeAD");
  this->finalizeAD(dx.increment());
  Log::trace() << "LinearModelBase<MODEL>::finalizeAD done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_LINEARMODELBASE_H_
