/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_LINEAROBSOPERBASE_H_
#define OOPS_INTERFACE_LINEAROBSOPERBASE_H_

#include <map>
#include <memory>
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/interface/ObsAuxControlBase.h"
#include "oops/interface/ObsAuxIncrementBase.h"
#include "util/Logger.h"
#include "util/Printable.h"
#include "util/abor1_cpp.h"

namespace oops {

/*!  @brief Base class for the linear observation operator.

 *   Defines the interfaces for the linear observation operator.
 *  LinearObsOperFactory is used to produce concrete linear observation
 *  operators.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class LinearObsOperBase : public util::Printable, private eckit::NonCopyable {
  using GeoVaLs_ = typename MODEL::GeoVaLs;
  using ObsVector_ = typename MODEL::ObsVector;
  using Variables_ = typename MODEL::Variables;

 public:
  /// Constructor
  LinearObsOperBase() {}
  /// Destructor
  virtual ~LinearObsOperBase() {}

  // Obs Operators
  /// Set the linearisation trajectory
  virtual void setTrajectory(
      const GeoVaLs_ &,
      const std::map<std::string, std::unique_ptr<ObsAuxControlBase<MODEL>>>
          &) = 0;
  /// Evaluate the linear observation operator
  virtual void obsEquivTL(
      const GeoVaLs_ &, ObsVector_ &,
      const std::map<std::string, std::unique_ptr<ObsAuxIncrementBase<MODEL>>>
          &) const = 0;
  /// Evaluate the adjoint of the linear observation operator
  virtual void obsEquivAD(
      GeoVaLs_ &, const ObsVector_ &,
      std::map<std::string, std::unique_ptr<ObsAuxIncrementBase<MODEL>>> &)
      const = 0;
  /// Evaluate the linear bias operator
  virtual void obsBiasEquivTL(
      const GeoVaLs_ &, ObsVector_ &,
      const std::map<std::string, std::unique_ptr<ObsAuxIncrementBase<MODEL>>>
          &) const = 0;
  /// Evaluate the adjoint of the bias observation operator
  virtual void obsBiasEquivAD(
      const GeoVaLs_ &, const ObsVector_ &,
      std::map<std::string, std::unique_ptr<ObsAuxIncrementBase<MODEL>>> &)
      const = 0;

  // Other
  /// Retrun variables (required input variables from LinearModel)
  virtual std::shared_ptr<const Variables_> variables() const = 0;

 private:
  virtual void print(std::ostream &) const = 0;
};

// =============================================================================

/// Obs Operator Factory
template <typename MODEL>
class LinearObsOperFactory {
  using ObsSpace_ = typename MODEL::ObsSpace;

 public:
  static LinearObsOperBase<MODEL> *create(const ObsSpace_ &,
                                          const eckit::Configuration &);
  virtual ~LinearObsOperFactory() { getMakers().clear(); }

 protected:
  explicit LinearObsOperFactory(const std::string &);

 private:
  virtual LinearObsOperBase<MODEL> *make(const ObsSpace_ &,
                                         const eckit::Configuration &) = 0;
  static std::map<std::string, LinearObsOperFactory<MODEL> *> &getMakers() {
    static std::map<std::string, LinearObsOperFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <class MODEL, class T>
class LinearObsOpMaker : public LinearObsOperFactory<MODEL> {
  using ObsSpace_ = typename MODEL::ObsSpace const;
  virtual LinearObsOperBase<MODEL> *make(const ObsSpace_ &odb,
                                         const eckit::Configuration &conf) {
    return new T(odb, conf);
  }

 public:
  explicit LinearObsOpMaker(const std::string &name)
      : LinearObsOperFactory<MODEL>(name) {}
};

// =============================================================================

template <typename MODEL>
LinearObsOperFactory<MODEL>::LinearObsOperFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in LinearObsOperFactory."
                 << std::endl;
    ABORT("Element already registered in LinearObsOperFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearObsOperBase<MODEL> *LinearObsOperFactory<MODEL>::create(
    const ObsSpace_ &odb, const eckit::Configuration &conf) {
  const std::string id = conf.getString("ObsType");
  typename std::map<std::string, LinearObsOperFactory<MODEL> *>::iterator jloc =
      getMakers().find(id);
  if (jloc == getMakers().end()) {
    Log::error() << id << " does not exist in LinearObsOperFactory."
                 << std::endl;
    ABORT("Element does not exist in LinearObsOperFactory.");
  }
  LinearObsOperBase<MODEL> *ptr = jloc->second->make(odb, conf);
  return ptr;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_LINEAROBSOPERBASE_H_
