/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_LINEAROBSOPERATOR_H_
#define OOPS_INTERFACE_LINEAROBSOPERATOR_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/LinearObsOperBase.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "oops/interface/ObsVector.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

/*! @brief Encapsulates the linear observation operator.

 *   This class provides the operations associated with the linear observation
 *  operator. It wraps the actual linear observation operator which is derived
 *  from the linear observation operator base class LinearObsOperBase. The
 *  interface for the linear observation operator comprises two levels because
 *  we want run time polymorphism. Note that it may be desirable to define a
 *  linear observation operator for each individual observation type rather than
 *  a single composite operator (which is the case the IFS).
 *  LinearObsOperFactory produces concrete objects of LinearObsOperBase children
 *  classes.

 *   LinearObsOperators class contains a vector of LinearObsOperator objects.
 *  Refer to the documentation of CostJo class to understand the structure of
 *  the \f$ J_o \f$ term.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class LinearObsOperator
    : public util::Printable,
      private eckit::NonCopyable,
      private util::ObjectCounter<LinearObsOperator<MODEL> > {
  using LinearObsOperBase_ = LinearObsOperBase<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using ObsAuxControl_ = ObsAuxControl<MODEL>;
  using ObsAuxIncrement_ = ObsAuxIncrement<MODEL>;
  using ObsSpace_ = ObservationSpace<MODEL>;
  using ObsVector_ = ObsVector<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::LinearObsOperator"; }

  /// Constructor
  explicit LinearObsOperator(const ObsSpace_ &);
  /// Destructor
  ~LinearObsOperator();

  /// Interfacing - return a constant reference to the concrete operator object
  const LinearObsOperBase_ &linearobsoperator() const { return *oper_; }

  // Obs Operators
  /// Set the linearisation trajectory
  void setTrajectory(const GeoVaLs_ &, const ObsAuxControl_ &);
  /// Evaluate the linear observation operator \f$ \bar{H} \f$
  void obsEquivTL(const GeoVaLs_ &, ObsVector_ &,
                  const ObsAuxIncrement_ &) const;
  /// Evaluate the adjoint of the linear observation operator \f$ \bar{H} \f$
  void obsEquivAD(GeoVaLs_ &, const ObsVector_ &, ObsAuxIncrement_ &) const;
  /// Evaluate the linear bias operator
  void obsBiasEquivTL(const GeoVaLs_ &, ObsVector_ &,
                      const ObsAuxIncrement_ &) const;
  /// Evaluate the adjoint of the bias observation operator
  void obsBiasEquivAD(const GeoVaLs_ &, const ObsVector_ &,
                      ObsAuxIncrement_ &) const;

  // Other
  /// Retrun variables (required input variables from LinearModel)
  Variables_ variables() const;

 private:
  void print(std::ostream &) const;

  std::unique_ptr<LinearObsOperBase_>
      oper_;  ///< Pointer to a LinearObsOperBase object
};

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearObsOperator<MODEL>::LinearObsOperator(const ObsSpace_ &os) : oper_() {
  Log::trace() << "LinearObsOperator<MODEL>::LinearObsOperator starting"
               << std::endl;
  util::Timer timer(classname(), "LinearObsOperator");
  oper_.reset(
      LinearObsOperFactory<MODEL>::create(os.observationspace(), os.config()));
  Log::trace() << "LinearObsOperator<MODEL>::LinearObsOperator done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
LinearObsOperator<MODEL>::~LinearObsOperator() {
  Log::trace() << "LinearObsOperator<MODEL>::~LinearObsOperator starting"
               << std::endl;
  util::Timer timer(classname(), "~LinearObsOperator");
  oper_.reset();
  Log::trace() << "LinearObsOperator<MODEL>::~LinearObsOperator done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperator<MODEL>::setTrajectory(const GeoVaLs_ &gvals,
                                             const ObsAuxControl_ &aux) {
  Log::trace() << "LinearObsOperator<MODEL>::obsEquiv starting" << std::endl;
  util::Timer timer(classname(), "ObsEquiv");
  oper_->setTrajectory(gvals.geovals(), aux.obsauxcontrol());
  Log::trace() << "LinearObsOperator<MODEL>::obsEquiv done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperator<MODEL>::obsEquivTL(const GeoVaLs_ &gvals, ObsVector_ &yy,
                                          const ObsAuxIncrement_ &aux) const {
  /*! \f[ \begin{bmatrix} \tilde{H} && \tilde{P} \end{bmatrix}
          \begin{bmatrix} I (dx) \\ dx_{obsaux} \end{bmatrix}
      \f]

      Note: -> \f$ I (dx) \f$ is stored by GeoVaLs.
   */
  Log::trace() << "LinearObsOperator<MODEL>::obsEquivTL starting" << std::endl;
  util::Timer timer(classname(), "ObsEquivTL");
  oper_->obsEquivTL(gvals.geovals(), yy.obsvector(), aux.obsauxincrement());
  Log::trace() << "LinearObsOperator<MODEL>::obsEquivTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperator<MODEL>::obsEquivAD(GeoVaLs_ &gvals, const ObsVector_ &yy,
                                          ObsAuxIncrement_ &aux) const {
  /*! \f[ \begin{bmatrix} \tilde{H}^T \\ \tilde{P}^T \end{bmatrix} dy
      \f]
   */
  Log::trace() << "LinearObsOperator<MODEL>::obsEquivAD starting" << std::endl;
  util::Timer timer(classname(), "ObsEquivAD");
  oper_->obsEquivAD(gvals.geovals(), yy.obsvector(), aux.obsauxincrement());
  Log::trace() << "LinearObsOperator<MODEL>::obsEquivAD done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperator<MODEL>::obsBiasEquivTL(
    const GeoVaLs_ &gvals, ObsVector_ &obias,
    const ObsAuxIncrement_ &aux) const {
  /// \f$ \tilde{P} dx_{obsaux} \f$
  Log::trace() << "LinearObsOperator<MODEL>::obsBiasEquivTL starting"
               << std::endl;
  util::Timer timer(classname(), "ObsBiasEquivTL");
  oper_->obsBiasEquivTL(gvals.geovals(), obias.obsvector(),
                        aux.obsauxincrement());
  Log::trace() << "LinearObsOperator<MODEL>::obsBiasEquivTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperator<MODEL>::obsBiasEquivAD(const GeoVaLs_ &gvals,
                                              const ObsVector_ &obias,
                                              ObsAuxIncrement_ &aux) const {
  /// \f$ \tilde{P}^T dy \f$
  Log::trace() << "LinearObsOperator<MODEL>::obsBiasEquivAD starting"
               << std::endl;
  util::Timer timer(classname(), "ObsBiasEquivAD");
  ObsAuxIncrement_ auxbias(aux, false);
  auxbias.zero();
  oper_->obsBiasEquivAD(gvals.geovals(), obias.obsvector(),
                        auxbias.obsauxincrement());
  aux += auxbias;
  Log::trace() << "LinearObsOperator<MODEL>::obsBiasEquivAD done" << std::endl;
}

// -----------------------------------------------------------------------------
template <typename MODEL>
Variables<MODEL> LinearObsOperator<MODEL>::variables() const {
  Log::trace() << "LinearObsOperator<MODEL>::variables starting" << std::endl;
  util::Timer timer(classname(), "variables");
  Variables_ inputs(oper_->variables());
  Log::trace() << "LinearObsOperator<MODEL>::variables done" << std::endl;
  return inputs;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void LinearObsOperator<MODEL>::print(std::ostream &os) const {
  Log::trace() << "LinearObsOperator<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *oper_;
  Log::trace() << "LinearObsOperator<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_LINEAROBSOPERATOR_H_
