/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_LOCALIZATION_H_
#define OOPS_INTERFACE_LOCALIZATION_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/generic/LocalizationBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/IncrEnsCtlVec.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Variables.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

/*! @brief Wrapper for localization

 */

template <typename MODEL>
class Localization : public util::Printable,
                     private util::ObjectCounter<Localization<MODEL> >,
                     private eckit::NonCopyable {
  using Localization_ = typename MODEL::LocalizationMatrix;
  using LocalizationBase_ = LocalizationBase<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using IncrEnsCtlVec_ = IncrEnsCtlVec<MODEL>;
  using Increment_ = Increment<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::Localization"; }

  Localization(const Geometry_ &, const Variables_ &,
               const eckit::Configuration &);
  virtual ~Localization();

  /// Implementations of localization
  void multiply(Increment_ &) const;
  void multiplySqrt(const IncrEnsCtlVec_ &, Increment_ &) const;
  void multiplySqrtTrans(const Increment_ &, IncrEnsCtlVec_ &) const;

  // Interfacing
  /// Return a constant reference to the generic or model-specific localization
  /// matrix
  const LocalizationBase_ &genLoc() const { return *genLoc_; }
  const Localization_ &modLoc() const { return *modLoc_; }
  /// Return a reference to the generic or model-specific localization matrix
  LocalizationBase_ &genLoc() { return *genLoc_; }
  Localization_ &modLoc() { return *modLoc_; }
  /// Is the localization matrix generic?
  bool isGeneric() const { return (genLoc_ != 0); };

 private:
  void print(std::ostream &) const;

  std::unique_ptr<LocalizationBase_>
      genLoc_;  ///< Pointer to a generic localization
  std::unique_ptr<Localization_>
      modLoc_;  ///< Pointer to a model-specific localization
};

// =============================================================================

template <typename MODEL>
Localization<MODEL>::Localization(const Geometry_ &resol,
                                  const Variables_ &vars,
                                  const eckit::Configuration &conf)
    : genLoc_(), modLoc_() {
  Log::trace() << "Localization<MODEL>::Localization starting" << std::endl;
  if (conf.has("localization")) {
    // Generic localization
    genLoc_.reset(LocalizationFactory<MODEL>::create(resol, vars, conf));
  } else {
    // Model specific localization
    modLoc_.reset(new Localization_(resol.geometry(), vars.variables(), conf));
  }
  Log::trace() << "Localization<MODEL>::Localization done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Localization<MODEL>::~Localization() {
  Log::trace() << "Localization<MODEL>::~Localization starting" << std::endl;
  util::Timer timer(classname(), "~Localization");
  genLoc_.reset();
  modLoc_.reset();
  Log::trace() << "Localization<MODEL>::~Localization done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Localization<MODEL>::multiply(Increment_ &dx) const {
  Log::trace() << "Localization<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  if (genLoc_) genLoc_->multiply(dx);
  if (modLoc_) modLoc_->multiply(dx.increment());
  Log::trace() << "Localization<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Localization<MODEL>::multiplySqrt(const IncrEnsCtlVec_ &dv,
                                       Increment_ &dx) const {
  Log::trace() << "Localization<MODEL>::multiplySqrt starting" << std::endl;
  util::Timer timer(classname(), "multiplySqrt");
  if (genLoc_) genLoc_->multiplySqrt(dv.genCtlVec(), dx);
  if (modLoc_) modLoc_->multiplySqrt(dv.modCtlVec(), dx.increment());
  Log::trace() << "Localization<MODEL>::multiplySqrt done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Localization<MODEL>::multiplySqrtTrans(const Increment_ &dx,
                                            IncrEnsCtlVec_ &dv) const {
  Log::trace() << "Localization<MODEL>::multiplySqrtTrans starting"
               << std::endl;
  util::Timer timer(classname(), "multiplySqrtTrans");
  if (genLoc_) genLoc_->multiplySqrtTrans(dx, dv.genCtlVec());
  if (modLoc_) modLoc_->multiplySqrtTrans(dx.increment(), dv.modCtlVec());
  Log::trace() << "Localization<MODEL>::multiplySqrtTrans done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Localization<MODEL>::print(std::ostream &os) const {
  Log::trace() << "Localization<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  if (genLoc_) os << *genLoc_;
  if (modLoc_) os << *modLoc_;
  Log::trace() << "Localization<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_LOCALIZATION_H_
