/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_MODEL_H_
#define OOPS_INTERFACE_MODEL_H_

#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/State.h"
#include "util/Duration.h"
#include "util/LogbookWriter.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

template <typename MODEL>
class DualConfigModel;

/*! @brief Encapsulates the nonlinear forecast model.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class Model : public util::Printable,
              private eckit::NonCopyable,
              private util::ObjectCounter<Model<MODEL> > {
  using Model_ = typename MODEL::Model;
  using Geometry_ = Geometry<MODEL>;
  using ModelAux_ = ModelAuxControl<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() { return "oops::Model"; }

  /// Constructor
  Model(const Geometry_ &, const eckit::Configuration &);
  /// Destructor
  ~Model();

  /// Run the forecast
  void forecast(State_ &, const ModelAux_ &, const util::Duration &,
                PostProcessor<State_> &) const;

  /// Setup linearizatin trajectory
  void setupTrajTL(const State_ &, const util::Duration &,
                   const util::Duration &, PostProcessor<State_> &) const;

  // Information and diagnostics
  /// Return the model time step
  const util::Duration &timeResolution() const {
    return model_->timeResolution();
  }

  // Interfacing
  /// Return a constant reference to the model specific model object.
  const Model_ &model() const { return *model_; }

 private:
  /// Initialize the model (prior to forecast)
  void initialize(State_ &, const ModelAux_ &) const;
  /// Model time step
  void step(State_ &, const ModelAux_ &) const;
  /// Finalize the model (after the forecast)
  void finalize(State_ &) const;
  void print(std::ostream &) const;

  std::unique_ptr<Model_>
      model_;  ///< Pointer to the model specific model object

  // DualConfigModel needs to access private methods of the Model
  friend class DualConfigModel<MODEL>;
};

// =============================================================================

template <typename MODEL>
Model<MODEL>::Model(const Geometry_ &resol, const eckit::Configuration &conf)
    : model_() {
  Log::trace() << "Model<MODEL>::Model starting" << std::endl;
  util::Timer timer(classname(), "Model");
  model_.reset(new Model_(resol.geometry(), conf));
  Log::trace() << "Model<MODEL>::Model done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Model<MODEL>::~Model() {
  Log::trace() << "Model<MODEL>::~Model starting" << std::endl;
  util::Timer timer(classname(), "~Model");
  model_.reset();
  Log::trace() << "Model<MODEL>::~Model done" << std::endl;
}

// -----------------------------------------------------------------------------
//  ****** NOTE: The forecast method below is not from MODEL::Model ******
// -----------------------------------------------------------------------------

template <typename MODEL>
void Model<MODEL>::forecast(State_ &xx, const ModelAux_ &maux,
                            const util::Duration &len,
                            PostProcessor<State_> &post) const {
  /*!  The model forecast receives a PostProcessor, which can be used for
   *  post-processing (storing/writing of the State; constructing the
   linearisation
   *  trajectory for the quadratic cost function)

   *   The model error term is handled internally by the model.
   */
  Log::trace() << "Model<MODEL>::forecast starting" << std::endl;
  util::Timer timer(classname(), "forecast");
  util::LogbookWriter<std::string> log("Forecast", "Nonlinear HR trajectory");
  const util::DateTime end(xx.validTime() + len);
  {
    size_t totsteps = len / model_->timeResolution();
    util::LogbookWriter<int> log_totsteps("traj_totsteps", totsteps);
  }
  Log::info() << "Model::forecast: forecast starting: " << xx << std::endl;
  this->initialize(xx, maux);
  post.initialize(xx, end, model_->timeResolution());
  size_t step = 0;
  while (xx.validTime() < end) {
    this->step(xx, maux);
    post.process(xx);
    ++step;
    util::LogbookWriter<int> log_step("traj_step", step);
  }
  post.finalize(xx);
  this->finalize(xx);
  Log::info() << "Model::forecast: forecast finished: " << xx << std::endl;
  ASSERT(xx.validTime() == end);

  Log::trace() << "Model<MODEL>::forecast done" << std::endl;
}

// -----------------------------------------------------------------------------
//  ****** NOTE: The forecast method above is not from MODEL::Model ******
// -----------------------------------------------------------------------------

template <typename MODEL>
void Model<MODEL>::setupTrajTL(const State_ &xx, const util::Duration &dt,
                               const util::Duration &len,
                               PostProcessor<State_> &post) const {
  /*!
   *   This method is designed to run an identity forecast to correctly
   *  setup structures holding the linearisation trajectory
   */
  Log::trace() << "Model<MODEL>::setupTrajTL starting" << std::endl;
  util::Timer timer(classname(), "setupTrajTL");
  util::DateTime time(xx.validTime());
  const util::DateTime end(xx.validTime() + len);
  Log::info() << "Model:setupTrajTL starting: " << std::endl;
  post.initialize(xx, end, dt);
  while (time < end) {
    time += dt;
    post.process(xx);
  }
  post.finalize(xx);
  Log::info() << "Model:forecast::setupTrajTL finished: " << std::endl;
  Log::trace() << "Model<MODEL>::setupTrajTL done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Model<MODEL>::initialize(State_ &xx, const ModelAux_ &maux) const {
  Log::trace() << "Model<MODEL>::initialize starting" << std::endl;
  util::Timer timer(classname(), "initialize");
  model_->initialize(xx.state(), maux.modelauxcontrol());
  Log::trace() << "Model<MODEL>::initialize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Model<MODEL>::step(State_ &xx, const ModelAux_ &maux) const {
  Log::trace() << "Model<MODEL>::step starting" << std::endl;
  util::Timer timer(classname(), "step");
  model_->step(xx.state(), maux.modelauxcontrol());
  Log::trace() << "Model<MODEL>::step done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Model<MODEL>::finalize(State_ &xx) const {
  Log::trace() << "Model<MODEL>::finalize starting" << std::endl;
  util::Timer timer(classname(), "finalize");
  model_->finalize(xx.state());
  Log::trace() << "Model<MODEL>::finalize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void Model<MODEL>::print(std::ostream &os) const {
  Log::trace() << "Model<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *model_;
  Log::trace() << "Model<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_MODEL_H_
