/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_MODELAUXCONTROL_H_
#define OOPS_INTERFACE_MODELAUXCONTROL_H_

#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "oops/interface/Geometry.h"
#include "oops/interface/State.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

template <typename MODEL>
class Model;

/*! @brief Encapsulates the auxiliary model space component of ControlVariable.

 *   The auxiliary model space component of ControlVariable is designed as
 *  a constant (time step independent) model error forcing term that can be
 *  jointly estimated by the assimilation algorithm.

 *   \f$ x_{k+1} = {\cal M}_{k+1:k} x_k + \eta \f$, where \f$ \eta \f$ is the
 *  model error forcing term implemented by ModelAuxControl.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ModelAuxControl : public util::Printable,
                        private util::ObjectCounter<ModelAuxControl<MODEL> > {
  using ModelAuxControl_ = typename MODEL::ModelAuxControl;
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() { return "oops::ModelAuxControl"; }

  /// Constructor
  ModelAuxControl(const Geometry_ &, const Model_ &,
                  const eckit::Configuration &);
  /// Copy constructor with interpolation
  ModelAuxControl(const Geometry_ &, const ModelAuxControl &);
  /// Copy constructor
  explicit ModelAuxControl(const ModelAuxControl &, const bool copy = true);
  /// Destructor
  ~ModelAuxControl();

  // Interfacing
  /// Return a constant reference to the encapsulated model specific object
  const ModelAuxControl_ &modelauxcontrol() const { return *aux_; }
  /// Return a reference to the encapsulated model specific object
  ModelAuxControl_ &modelauxcontrol() { return *aux_; }

  // I/O and diagnostics
  /// Read
  void read(const eckit::Configuration &);
  /// Write
  void write(const eckit::Configuration &) const;
  /// Compute L-2 norm
  double norm() const;

 private:
  ModelAuxControl &operator=(const ModelAuxControl &);
  void print(std::ostream &) const;

  /// Pointer to the model specific object
  std::unique_ptr<ModelAuxControl_> aux_;
};

// =============================================================================

template <typename MODEL>
ModelAuxControl<MODEL>::ModelAuxControl(const Geometry_ &resol,
                                        const Model_ &model,
                                        const eckit::Configuration &conf)
    : aux_() {
  Log::trace() << "ModelAuxControl<MODEL>::ModelAuxControl starting"
               << std::endl;
  util::Timer timer(classname(), "ModelAuxControl");
  aux_.reset(new ModelAuxControl_(resol.geometry(), model.model(),
                                  conf.getSubConfiguration("stateModAux")));
  Log::trace() << "ModelAuxControl<MODEL>::ModelAuxControl done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ModelAuxControl<MODEL>::ModelAuxControl(const Geometry_ &resol,
                                        const ModelAuxControl &other)
    : aux_() {
  Log::trace()
      << "ModelAuxControl<MODEL>::ModelAuxControl interpolated starting"
      << std::endl;
  util::Timer timer(classname(), "ModelAuxControl");
  aux_.reset(new ModelAuxControl_(resol.geometry(), *other.aux_));
  Log::trace() << "ModelAuxControl<MODEL>::ModelAuxControl interpolated done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ModelAuxControl<MODEL>::ModelAuxControl(const ModelAuxControl &other,
                                        const bool copy)
    : aux_() {
  Log::trace() << "ModelAuxControl<MODEL>::ModelAuxControl copy starting"
               << std::endl;
  util::Timer timer(classname(), "ModelAuxControl");
  aux_.reset(new ModelAuxControl_(*other.aux_, copy));
  Log::trace() << "ModelAuxControl<MODEL>::ModelAuxControl copy done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ModelAuxControl<MODEL>::~ModelAuxControl() {
  Log::trace() << "ModelAuxControl<MODEL>::~ModelAuxControl starting"
               << std::endl;
  util::Timer timer(classname(), "~ModelAuxControl");
  aux_.reset();
  Log::trace() << "ModelAuxControl<MODEL>::~ModelAuxControl done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxControl<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "ModelAuxControl<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  aux_->read(conf.getSubConfiguration("stateModAux"));
  Log::trace() << "ModelAuxControl<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxControl<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "ModelAuxControl<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  aux_->write(conf.getSubConfiguration("stateModAux"));
  Log::trace() << "ModelAuxControl<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double ModelAuxControl<MODEL>::norm() const {
  Log::trace() << "ModelAuxControl<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = aux_->norm();
  Log::trace() << "ModelAuxControl<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxControl<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ModelAuxControl<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *aux_;
  Log::trace() << "ModelAuxControl<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_MODELAUXCONTROL_H_
