/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_MODELAUXCONTROLESTIMATOR_H_
#define OOPS_INTERFACE_MODELAUXCONTROLESTIMATOR_H_

#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/State.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class ModelAuxControlEstimator
    : public util::Printable,
      private eckit::NonCopyable,
      private util::ObjectCounter<ModelAuxControlEstimator<MODEL> > {
  using ModelAuxControlEstimator_ = typename MODEL::ModelAuxControlEstimator;
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using ModelAuxControl_ = ModelAuxControl<MODEL>;
  using State_ = State<MODEL>;

 public:
  static const std::string classname() {
    return "oops::ModelAuxControlEstimator";
  }

  explicit ModelAuxControlEstimator(const Geometry_ &, const Model_ &,
                                    const eckit::Configuration &);
  virtual ~ModelAuxControlEstimator();

  /// Methods
  void estimate(const State_ &, ModelAuxControl_ &) const;
  void estimate(const State_ &, const State_ &, ModelAuxControl_ &) const;

  /// Interfacing
  const ModelAuxControlEstimator_ &modelAuxControlEstimator() const {
    return *aux_;
  }
  ModelAuxControlEstimator_ &modelAuxControlEstimator() { return *aux_; }

 private:
  ModelAuxControlEstimator &operator=(const ModelAuxControlEstimator &);
  void print(std::ostream &) const;
  std::unique_ptr<ModelAuxControlEstimator_> aux_;
};

// =============================================================================

template <typename MODEL>
ModelAuxControlEstimator<MODEL>::ModelAuxControlEstimator(
    const Geometry_ &resol, const Model_ &model,
    const eckit::Configuration &conf)
    : aux_() {
  Log::trace()
      << "ModelAuxControlEstimator<MODEL>::ModelAuxControlEstimator starting"
      << std::endl;
  util::Timer timer(classname(), "ModelAuxControlEstimator");
  aux_.reset(new ModelAuxControlEstimator_(
      resol.geometry(), model.model(),
      conf.getSubConfiguration("stateModAuxEstimator")));
  Log::trace()
      << "ModelAuxControlEstimator<MODEL>::ModelAuxControlEstimator done"
      << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ModelAuxControlEstimator<MODEL>::~ModelAuxControlEstimator() {
  Log::trace()
      << "ModelAuxControlEstimator<MODEL>::~ModelAuxControlEstimator starting"
      << std::endl;
  util::Timer timer(classname(), "~ModelAuxControlEstimator");
  aux_.reset();
  Log::trace()
      << "ModelAuxControlEstimator<MODEL>::~ModelAuxControlEstimator done"
      << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxControlEstimator<MODEL>::estimate(const State_ &xb,
                                               ModelAuxControl_ &xerr) const {
  Log::trace() << "ModelAuxControlEstimator<MODEL>::estimate starting"
               << std::endl;
  util::Timer timer(classname(), "estimate");
  aux_->estimate(xb.state(), xerr.modelauxcontrol());
  Log::trace() << "ModelAuxControlEstimator<MODEL>::estimate done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxControlEstimator<MODEL>::estimate(const State_ &xb,
                                               const State_ &an,
                                               ModelAuxControl_ &xerr) const {
  Log::trace() << "ModelAuxControlEstimator<MODEL>::estimate starting"
               << std::endl;
  util::Timer timer(classname(), "estimate");
  aux_->estimate(xb.state(), an.state(), xerr.modelauxcontrol());
  Log::trace() << "ModelAuxControlEstimator<MODEL>::estimate done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxControlEstimator<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ModelAuxControlEstimator<MODEL>::print starting"
               << std::endl;
  util::Timer timer(classname(), "print");
  os << *aux_;
  Log::trace() << "ModelAuxControlEstimator<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_MODELAUXCONTROLESTIMATOR_H_
