/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_MODELAUXCOVARIANCE_H_
#define OOPS_INTERFACE_MODELAUXCOVARIANCE_H_

#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxCtlVec.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

/*!  @brief Encapsulates covariance for auxiliary model space component of
 *  ControlVariable

 *   This class provides the operations associated with the covariance matrix
 *  of the auxiliary model space component of ControlVariable. It wraps the
 *  actual model specific covariance matrix.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ModelAuxCovariance
    : public util::Printable,
      private eckit::NonCopyable,
      private util::ObjectCounter<ModelAuxCovariance<MODEL> > {
  using ModelAuxCovariance_ = typename MODEL::ModelAuxCovariance;
  using Geometry_ = Geometry<MODEL>;
  using ModelAuxControl_ = ModelAuxControl<MODEL>;
  using ModelAuxCtlVec_ = ModelAuxCtlVec<MODEL>;
  using ModelAuxIncrement_ = ModelAuxIncrement<MODEL>;

 public:
  static const std::string classname() { return "oops::ModelAuxCovariance"; }

  /// Constructor
  ModelAuxCovariance(const eckit::Configuration &, const Geometry_ &);
  /// Destructor
  ~ModelAuxCovariance();

  // Operators
  /// Set up the inner loop resolution covariance matrix
  void linearize(const ModelAuxControl_ &, const Geometry_ &);
  /// \f$ dx_1 = Q dx_2 \f$
  void multiply(const ModelAuxIncrement_ &, ModelAuxIncrement_ &) const;
  /// \f$ dx_1 = Q^{-1} dx_2 \f$
  void inverseMultiply(const ModelAuxIncrement_ &, ModelAuxIncrement_ &) const;
  /// \f$ dx = Q^{1/2} dv \f$
  void multiplySqrt(const ModelAuxCtlVec_ &, ModelAuxIncrement_ &) const;
  /// \f$ dv = [Q^{1/2}]^T dx \f$
  void multiplySqrtTrans(const ModelAuxIncrement_ &, ModelAuxCtlVec_ &) const;
  /// \f$ dx = Q^{1/2} dv_{random} \f$
  void randomize(ModelAuxIncrement_ &) const;

  // Interfacing
  /// Return a constant reference to configuration
  const eckit::Configuration &config() const { return cov_->config(); }
  /// Return a reference to the model specific object
  ModelAuxCovariance_ &errorCovariance() { return *cov_; }
  /// Return a constant reference to the model specific object
  const ModelAuxCovariance_ &errorCovariance() const { return *cov_; }

 private:
  void print(std::ostream &) const;

  std::unique_ptr<ModelAuxCovariance_>
      cov_;  ///< Pointer to the model specific object
};

// =============================================================================

template <typename MODEL>
ModelAuxCovariance<MODEL>::ModelAuxCovariance(const eckit::Configuration &conf,
                                              const Geometry_ &resol)
    : cov_() {
  Log::trace() << "ModelAuxCovariance<MODEL>::ModelAuxCovariance starting"
               << std::endl;
  util::Timer timer(classname(), "ModelAuxCovariance");
  cov_.reset(new ModelAuxCovariance_(conf, resol.geometry()));
  Log::trace() << "ModelAuxCovariance<MODEL>::ModelAuxCovariance done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ModelAuxCovariance<MODEL>::~ModelAuxCovariance() {
  Log::trace() << "ModelAuxCovariance<MODEL>::~ModelAuxCovariance starting"
               << std::endl;
  util::Timer timer(classname(), "~ModelAuxCovariance");
  cov_.reset();
  Log::trace() << "ModelAuxCovariance<MODEL>::~ModelAuxCovariance done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::linearize(const ModelAuxControl_ &xx,
                                          const Geometry_ &resol) {
  Log::trace() << "ModelAuxCovariance<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  cov_->linearize(xx.modelauxcontrol(), resol.geometry());
  Log::trace() << "ModelAuxCovariance<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::multiply(const ModelAuxIncrement_ &dx1,
                                         ModelAuxIncrement_ &dx2) const {
  Log::trace() << "ModelAuxCovariance<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  cov_->multiply(dx1.modelauxincrement(), dx2.modelauxincrement());
  Log::trace() << "ModelAuxCovariance<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::inverseMultiply(const ModelAuxIncrement_ &dx1,
                                                ModelAuxIncrement_ &dx2) const {
  Log::trace() << "ModelAuxCovariance<MODEL>::inverseMultiply starting"
               << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  cov_->inverseMultiply(dx1.modelauxincrement(), dx2.modelauxincrement());
  Log::trace() << "ModelAuxCovariance<MODEL>::inverseMultiply done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::multiplySqrt(const ModelAuxCtlVec_ &dvin,
                                             ModelAuxIncrement_ &dxout) const {
  Log::trace() << "ModelAuxCovariance<MODEL>::multiplySqrt starting"
               << std::endl;
  util::Timer timer(classname(), "multiplySqrt");
  cov_->multiplySqrt(dvin.modelAuxCtlVec(), dxout.modelauxincrement());
  Log::trace() << "ModelAuxCovariance<MODEL>::multiplySqrt done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::multiplySqrtTrans(
    const ModelAuxIncrement_ &dxin, ModelAuxCtlVec_ &dvout) const {
  Log::trace() << "ModelAuxCovariance<MODEL>::multiplySqrtTrans starting"
               << std::endl;
  util::Timer timer(classname(), "multiplySqrtTrans");
  cov_->multiplySqrtTrans(dxin.modelauxincrement(), dvout.modelAuxCtlVec());
  Log::trace() << "ModelAuxCovariance<MODEL>::multiplySqrtTrans done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::randomize(ModelAuxIncrement_ &dx) const {
  Log::trace() << "ModelAuxCovariance<MODEL>::randomize starting" << std::endl;
  util::Timer timer(classname(), "randomize");
  cov_->randomize(dx.modelauxincrement());
  Log::trace() << "ModelAuxCovariance<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ModelAuxCovariance<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ModelAuxCovariance<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *cov_;
  Log::trace() << "ModelAuxCovariance<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_MODELAUXCOVARIANCE_H_
