/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCONTROLBASE_H_
#define OOPS_INTERFACE_OBSAUXCONTROLBASE_H_

#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/interface/ObservationSpace.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename MODEL>
class ObsAuxIncrementBase;

/*! @brief Base class for auxiliary observation space components of
 ControlVariable.

 *   Defines interfaces for the auxiliary observation space components of
 *  ControlVariable (e.g. VarBC, TOVSCV). ObsAuxControlFactory is used to
 *  produce concrete objects.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxControlBase : public util::Printable {
  using ObsAuxIncrementBase_ = ObsAuxIncrementBase<MODEL>;
  using ObservationSpace_ = ObservationSpace<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxControlBase"; }

  /// Constructor
  ObsAuxControlBase() {}
  /// Destructor
  virtual ~ObsAuxControlBase() {}

  /// Cloning
  virtual ObsAuxControlBase *clone(const bool copy = true) const = 0;

  /// \f$ x += dx \f$
  virtual ObsAuxControlBase &operator+=(const ObsAuxIncrementBase_ &) = 0;

  /// Reset
  void reset(const ObservationSpace_ &, const std::string &,
             const eckit::Configuration &);

  // I/O and diagnostics
  /// Read
  virtual void read(const eckit::Configuration &) = 0;
  /// Write
  virtual void write(const eckit::Configuration &) const = 0;
  /// L-2 norm
  virtual double norm() const = 0;

  // Interfacing
  virtual const int &getAccess() const = 0;
  virtual int &getAccess() = 0;
  virtual const int &getAccessToObsSpace() const = 0;

 protected:
  virtual void reset(const typename MODEL::ObsSpace &, const std::string &,
                     const eckit::Configuration &) = 0;

 private:
  virtual void print(std::ostream &) const = 0;
};

// =============================================================================

/// ObsAuxControl Factory
template <typename MODEL>
class ObsAuxControlFactory {
 public:
  static ObsAuxControlBase<MODEL> *create(const ObservationSpace<MODEL> &,
                                          const eckit::Configuration &);
  virtual ~ObsAuxControlFactory() { getMakers().clear(); }

 protected:
  explicit ObsAuxControlFactory(const std::string &);

 private:
  virtual ObsAuxControlBase<MODEL> *make(const ObservationSpace<MODEL> &,
                                         const eckit::Configuration &) = 0;
  static std::map<std::string, ObsAuxControlFactory<MODEL> *> &getMakers() {
    static std::map<std::string, ObsAuxControlFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <typename MODEL, class T>
class ObsAuxControlMaker : public ObsAuxControlFactory<MODEL> {
  virtual ObsAuxControlBase<MODEL> *make(const ObservationSpace<MODEL> &os,
                                         const eckit::Configuration &conf) {
    return new T(os.observationspace(), conf);
  }

 public:
  explicit ObsAuxControlMaker(const std::string &name)
      : ObsAuxControlFactory<MODEL>(name) {}
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxControlFactory<MODEL>::ObsAuxControlFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in ObsAuxControlFactory."
                 << std::endl;
    ABORT("Element already registered in ObsAuxControlFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxControlBase<MODEL> *ObsAuxControlFactory<MODEL>::create(
    const ObservationSpace<MODEL> &os, const eckit::Configuration &conf) {
  const std::string id = conf.getString("type");
  typename std::map<std::string, ObsAuxControlFactory<MODEL> *>::iterator jloc =
      getMakers().find(id);
  if (jloc == getMakers().end()) {
    Log::error() << id << " does not exist in ObsAuxControlFactory."
                 << std::endl;
    ABORT("Element does not exist in ObsAuxControlFactory.");
  }
  ObsAuxControlBase<MODEL> *ptr = jloc->second->make(os, conf);
  return ptr;
}

// =============================================================================

template <typename MODEL>
void ObsAuxControlBase<MODEL>::reset(const ObservationSpace_ &os,
                                     const std::string &stateid,
                                     const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxControlBase<MODEL>::reset starting" << std::endl;
  util::Timer timer(classname(), "reset");
  this->reset(os.observationspace(), stateid, conf);
  Log::trace() << "ObsAuxControlBase<MODEL>::reset done" << std::endl;
}

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCONTROLBASE_H_
