/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCOVARIANCE_H_
#define OOPS_INTERFACE_OBSAUXCOVARIANCE_H_

#include <iostream>
#include <map>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsAuxCovarianceBase.h"
#include "oops/interface/ObsAuxCtlVec.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "oops/interface/ObservationSpace.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

/*!  @brief Encapsulates the covariance matrix of auxiliary observation space
 *  components of ControlVariable.

 *   This class provides operations associated with the covariance matrix of
 *  auxiliary observation space components (e.g. VarBC, TOVSC) of
 *  ControlVariable. It contains a vector of ObsAuxCovarianceBase objects,
 *  where ObsAuxCovarianceBase is a base class for model specific (or
 *  potentially generic) classes.

 *   ObsAuxCovariances class wraps around a vector of ObsAuxCovariance objects
 *  and is part of CostJbTotal. Refer to the documentation of ControlVariable
 *  and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCovariance : public util::Printable,
                         private eckit::NonCopyable,
                         private util::ObjectCounter<ObsAuxCovariance<MODEL>> {
  using ObsAuxCovarianceBase_ = ObsAuxCovarianceBase<MODEL>;
  using ObsAuxControl_ = ObsAuxControl<MODEL>;
  using ObsAuxIncrement_ = ObsAuxIncrement<MODEL>;
  using ObsAuxCtlVec_ = ObsAuxCtlVec<MODEL>;
  using ObservationSpace_ = ObservationSpace<MODEL>;
  using ObsAuxCovariancePtrMap_ =
      std::map<std::string, std::unique_ptr<ObsAuxCovarianceBase_>>;
  using iter_ = typename ObsAuxCovariancePtrMap_::iterator;
  using icst_ = typename ObsAuxCovariancePtrMap_::const_iterator;

 public:
  static const std::string classname() { return "oops::ObsAuxCovariance"; }

  /// Constructor
  ObsAuxCovariance(const ObservationSpace_ &, const eckit::Configuration &);
  /// Destructor
  ~ObsAuxCovariance();

  /// Update
  void update(const ObsAuxControl_ &, const eckit::Configuration &);

  // Operators
  /// Set up the inner loop resolution B matrix
  void linearize(const ObsAuxControl_ &);
  /// \f$ dx_1 = B dx_2 \f$
  void multiply(const ObsAuxIncrement_ &, ObsAuxIncrement_ &) const;
  /// \f$ dx_1 = B^{-1} dx_2 \f$
  void inverseMultiply(const ObsAuxIncrement_ &, ObsAuxIncrement_ &) const;
  /// \f$ dx = B^{1/2} dv \f$
  void multiplySqrt(const ObsAuxCtlVec_ &, ObsAuxIncrement_ &) const;
  /// \f$ dv = [B^{1/2}]^T dx \f$
  void multiplySqrtTrans(const ObsAuxIncrement_ &, ObsAuxCtlVec_ &) const;
  /// \f$ dx = B^{1/2} dv_{random} \f$
  void randomize(ObsAuxIncrement_ &) const;

  // Interfacing
  /// Retrun a constant reference to configuration
  const eckit::Configuration &config() const { return *conf_; }

  /// Return a constant reference to the model specific object
  const ObsAuxCovarianceBase_ &errorCovariance(const std::string &id) const {
    return this->get(id);
  }
  /// Return a reference to the model specific object
  ObsAuxCovarianceBase_ &errorCovariance(const std::string &id) {
    return this->get(id);
  }
  /// Return the number of ObsAuxIncrementBase objects
  const size_t &size() const { return size_; }

 private:
  /// Return a constant reference to the model specific object
  const ObsAuxCovarianceBase_ &get(const std::string &) const;
  /// Return a reference to the model specific object
  ObsAuxCovarianceBase_ &get(const std::string &);
  void print(std::ostream &) const;

  mutable size_t size_;           ///< Number of model specific objects
  ObsAuxCovariancePtrMap_ covs_;  ///< Map of pointers to model specific objects
  std::unique_ptr<eckit::LocalConfiguration> conf_;  ///< Configuration
};

// =============================================================================

template <typename MODEL>
ObsAuxCovariance<MODEL>::ObsAuxCovariance(const ObservationSpace_ &os,
                                          const eckit::Configuration &conf)
    : size_(0), covs_(), conf_(new eckit::LocalConfiguration(conf)) {
  Log::trace() << "ObsAuxCovariance<MODEL>::ObsAuxCovariance starting"
               << std::endl;
  util::Timer timer(classname(), "ObsAuxCovariance");

  std::vector<eckit::LocalConfiguration> covconfs;
  conf.get("covarianceObsAux", covconfs);
  size_ = covconfs.size();

  for (auto const &config : covconfs) {
    std::string key(config.getString("type"));
    std::unique_ptr<ObsAuxCovarianceBase_> auxcov(
        ObsAuxCovarianceFactory<MODEL>::create(os, config));
    covs_.emplace(key, std::move(auxcov));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::ObsAuxCovariance done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCovariance<MODEL>::~ObsAuxCovariance() {
  Log::trace() << "ObsAuxCovariance<MODEL>::~ObsAuxCovariance starting"
               << std::endl;
  util::Timer timer(classname(), "~ObsAuxCovariance");
  Log::trace() << "ObsAuxCovariance<MODEL>::~ObsAuxCovariance done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCovarianceBase<MODEL> &ObsAuxCovariance<MODEL>::get(
    const std::string &id) {
  iter_ it = covs_.find(id);
  ASSERT(it != covs_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
const ObsAuxCovarianceBase<MODEL> &ObsAuxCovariance<MODEL>::get(
    const std::string &id) const {
  icst_ it = covs_.find(id);
  ASSERT(it != covs_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::linearize(const ObsAuxControl_ &xx) {
  Log::trace() << "ObsAuxCovariance<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  for (iter_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->linearize(xx.obsauxcontrol(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::update(const ObsAuxControl_ &xx,
                                     const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxCovariance<MODEL>::update starting" << std::endl;
  util::Timer timer(classname(), "update");

  conf_.reset(new eckit::LocalConfiguration(conf));

  std::vector<eckit::LocalConfiguration> confs;
  conf.get("covarianceObsAux", confs);

  for (iter_ it = covs_.begin(); it != covs_.end(); ++it) {
    for (auto const &config : confs) {
      if (config.getString("type") == it->first && config.getBool("update")) {
        it->second->update(xx.obsauxcontrol(it->first), config);
      }
    }
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::update done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::multiply(const ObsAuxIncrement_ &dx1,
                                       ObsAuxIncrement_ &dx2) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->multiply(dx1.obsauxincrement(it->first),
                         dx2.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::inverseMultiply(const ObsAuxIncrement_ &dx1,
                                              ObsAuxIncrement_ &dx2) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::inverseMultiply starting"
               << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->inverseMultiply(dx1.obsauxincrement(it->first),
                                dx2.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::inverseMultiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::multiplySqrt(const ObsAuxCtlVec_ &dvin,
                                           ObsAuxIncrement_ &dxout) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrt starting" << std::endl;
  util::Timer timer(classname(), "multiplySqrt");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->multiplySqrt(dvin.obsAuxCtlVec(it->first),
                             dxout.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrt done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::multiplySqrtTrans(const ObsAuxIncrement_ &dxin,
                                                ObsAuxCtlVec_ &dvout) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrtTrans starting"
               << std::endl;
  util::Timer timer(classname(), "multiplySqrtTrans");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->multiplySqrtTrans(dxin.obsauxincrement(it->first),
                                  dvout.obsAuxCtlVec(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrtTrans done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::randomize(ObsAuxIncrement_ &dx) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::randomize starting" << std::endl;
  util::Timer timer(classname(), "randomize");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->randomize(dx.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsAuxCovariance<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    os << it->first << ": " << *it->second;
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCOVARIANCE_H_
