/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCTLVEC_H_
#define OOPS_INTERFACE_OBSAUXCTLVEC_H_

#include <iostream>
#include <map>
#include <memory>
#include <string>

#include "oops/interface/ObsAuxCtlVecBase.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

template <typename MODEL>
class ObsAuxCovariance;

/*! @brief Encapsulates auxiliary observation space components of ControlVector.

 *   This class provides operations associated with the auxiliary observation
 *  space components (e.g. VarBC, TOVSC) of ControlVector. It contains a
 *  vector of ObsAuxCtlVecBase objects, where ObsAuxCtlVecBase is a base
 *  class for model specific (or potentially generic) classes.

 *   ObsAuxCtlVecs class wraps around a vector of ObsAuxCtlVec objects
 *  and is part of ControlVector. Refer to the documentation of ControlVector
 *  and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCtlVec : public util::Printable,
                     private util::ObjectCounter<ObsAuxCtlVec<MODEL>> {
  using ObsAuxCtlVecBase_ = ObsAuxCtlVecBase<MODEL>;
  using ObsAuxCovariance_ = ObsAuxCovariance<MODEL>;
  using ObsAuxCtlVecPtrMap_ =
      std::map<std::string, std::unique_ptr<ObsAuxCtlVecBase_>>;
  using iter_ = typename ObsAuxCtlVecPtrMap_::iterator;
  using icst_ = typename ObsAuxCtlVecPtrMap_::const_iterator;

 public:
  static const std::string classname() { return "oops::ObsAuxCtlVec"; }

  /// Constructor
  ObsAuxCtlVec(const ObsAuxCovariance_ &);
  /// Constructor with copy initialization
  ObsAuxCtlVec(const ObsAuxCovariance_ &, const ObsAuxCtlVec &);
  /// Copy constructor
  ObsAuxCtlVec(const ObsAuxCtlVec &, const bool copy = true);
  /// Destructor
  ~ObsAuxCtlVec();

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  void zero();
  /// \f$ dv_1 = dv_2 \f$
  ObsAuxCtlVec &operator=(const ObsAuxCtlVec &);
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  ObsAuxCtlVec &operator+=(const ObsAuxCtlVec &);
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  ObsAuxCtlVec &operator-=(const ObsAuxCtlVec &);
  /// \f$ dv_1 = dv_1 * c \f$
  ObsAuxCtlVec &operator*=(const double &);
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  void axpy(const double &, const ObsAuxCtlVec &);
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  double dot_product_with(const ObsAuxCtlVec &) const;

  // I/O and diagnostics
  /// Read control vector
  void read(const eckit::Configuration &);
  /// Write control vector
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

  // Interfacing
  /// Return a constant reference to the model specific object
  const ObsAuxCtlVecBase_ &obsAuxCtlVec(const std::string &id) const {
    return this->get(id);
  }
  /// Return a reference to the model specific object
  ObsAuxCtlVecBase_ &obsAuxCtlVec(const std::string &id) {
    return this->get(id);
  }

  /// Return a constant reference to the vector of keys
  const std::vector<std::string> &getKeys() const { return keys_; }

  /// Return the number of ObsAuxCtlVecBase objects
  const size_t &size() const { return size_; }

 private:
  /// Return a constant reference to the model specific object
  const ObsAuxCtlVecBase_ &get(const std::string &) const;
  /// Return a reference to the model specific object
  ObsAuxCtlVecBase_ &get(const std::string &);
  void print(std::ostream &) const;

  mutable size_t size_;            ///< Number of model specific objects
  std::vector<std::string> keys_;  ///< Vector of keys to model specific objects
  ObsAuxCtlVecPtrMap_ ctlvecs_;  ///< Map of pointers to model specific objects
};

// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL>::ObsAuxCtlVec(const ObsAuxCovariance_ &cov)
    : size_(0), keys_(), ctlvecs_() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCtlVec");

  std::vector<eckit::LocalConfiguration> covconfs;
  cov.config().get("covarianceObsAux", covconfs);
  size_ = covconfs.size();

  for (auto const &config : covconfs) {
    std::string key(config.getString("type"));
    std::unique_ptr<ObsAuxCtlVecBase<MODEL>> auxctl(
        ObsAuxCtlVecFactory<MODEL>::create(cov.errorCovariance(key)));
    ctlvecs_.emplace(key, std::move(auxctl));
    keys_.push_back(key);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL>::ObsAuxCtlVec(const ObsAuxCovariance_ &cov,
                                  const ObsAuxCtlVec &other)
    : size_(other.size_), keys_(), ctlvecs_() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCtlVec");
  for (icst_ it = other.ctlvecs_.begin(); it != other.ctlvecs_.end(); ++it) {
    std::string incname = it->first;
    std::unique_ptr<ObsAuxCtlVecBase<MODEL>> auxctl(it->second->clone(cov));
    ctlvecs_.emplace(incname, std::move(auxctl));
    keys_.push_back(incname);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL>::ObsAuxCtlVec(const ObsAuxCtlVec &other, const bool copy)
    : size_(other.size_), keys_(), ctlvecs_() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCtlVec");
  for (icst_ it = other.ctlvecs_.begin(); it != other.ctlvecs_.end(); ++it) {
    std::string incname = it->first;
    std::unique_ptr<ObsAuxCtlVecBase<MODEL>> auxctl(it->second->clone(copy));
    ctlvecs_.emplace(incname, std::move(auxctl));
    keys_.push_back(incname);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL>::~ObsAuxCtlVec() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::~ObsAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~ObsAuxCtlVec");
  Log::trace() << "ObsAuxCtlVec<MODEL>::~ObsAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVecBase<MODEL> &ObsAuxCtlVec<MODEL>::get(const std::string &id) {
  iter_ it = ctlvecs_.find(id);
  ASSERT(it != ctlvecs_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
const ObsAuxCtlVecBase<MODEL> &ObsAuxCtlVec<MODEL>::get(
    const std::string &id) const {
  icst_ it = ctlvecs_.find(id);
  ASSERT(it != ctlvecs_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVec<MODEL>::zero() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    it->second->zero();
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL> &ObsAuxCtlVec<MODEL>::operator=(const ObsAuxCtlVec &rhs) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) = rhs.obsAuxCtlVec(it->first);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL> &ObsAuxCtlVec<MODEL>::operator+=(const ObsAuxCtlVec &rhs) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) += rhs.obsAuxCtlVec(it->first);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL> &ObsAuxCtlVec<MODEL>::operator-=(const ObsAuxCtlVec &rhs) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) -= rhs.obsAuxCtlVec(it->first);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxCtlVec<MODEL> &ObsAuxCtlVec<MODEL>::operator*=(const double &zz) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) *= zz;
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVec<MODEL>::axpy(const double &zz, const ObsAuxCtlVec &dx) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    it->second->axpy(zz, dx.obsAuxCtlVec(it->first));
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxCtlVec<MODEL>::dot_product_with(const ObsAuxCtlVec &dx) const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = 0.0;
  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    zz += it->second->dot_product_with(dx.obsAuxCtlVec(it->first));
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVec<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("ctlvecObsAux")) conf.get("ctlvecObsAux", confs);

  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    for (auto const &config : confs) {
      if (config.getString("type") == it->first) it->second->read(config);
    }
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVec<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("ctlvecObsAux")) conf.get("ctlvecObsAux", confs);

  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    for (auto const &config : confs) {
      if (config.getString("type") == it->first) it->second->write(config);
    }
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::write done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxCtlVec<MODEL>::norm() const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0.0;
  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    double norm = it->second->norm();
    zz += norm * norm;
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::norm done" << std::endl;
  return sqrt(zz);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxCtlVec<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    os << it->first << ": " << *it->second;
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::print done" << std::endl;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCTLVEC_H_
