/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCTLVECBASE_H_
#define OOPS_INTERFACE_OBSAUXCTLVECBASE_H_

#include <iostream>
#include <string>

#include "eckit/config/Configuration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename MODEL>
class ObsAuxCovarianceBase;

/*!  @brief Base class for auxiliary observation space components of
 *  ControlVector.

 *   Defines interfaces for the auxiliary observation space components of
 *  ControlVector (e.g. VarBC, TOVSCV). ObsAuxCtlVecFactory is used to
 *  produce concrete objects.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCtlVecBase : public util::Printable {
  using ObsAuxCovarianceBase_ = ObsAuxCovarianceBase<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxCtlVecBase"; }

  /// Constructor
  ObsAuxCtlVecBase() {}
  /// Destructor
  virtual ~ObsAuxCtlVecBase() {}

  /// Cloning
  virtual ObsAuxCtlVecBase *clone(const ObsAuxCovarianceBase_ &) const = 0;
  virtual ObsAuxCtlVecBase *clone(const bool copy = true) const = 0;

  // Linear algebra operators
  /// \f$ dv_1 = 0 \f$
  virtual void zero() = 0;
  /// \f$ dv_1 = dv_2 \f$
  virtual ObsAuxCtlVecBase &operator=(const ObsAuxCtlVecBase &) = 0;
  /// \f$ dv_1 = dv_1 + dv_2 \f$
  virtual ObsAuxCtlVecBase &operator+=(const ObsAuxCtlVecBase &) = 0;
  /// \f$ dv_1 = dv_1 - dv_2 \f$
  virtual ObsAuxCtlVecBase &operator-=(const ObsAuxCtlVecBase &) = 0;
  /// \f$ dv_1 = dv_1 * c \f$
  virtual ObsAuxCtlVecBase &operator*=(const double &) = 0;
  /// \f$ dv_1 = dv_1 * c dv_2 \f$
  virtual void axpy(const double &, const ObsAuxCtlVecBase &) = 0;
  /// \f$ <dv_1,dv_1> = dv_1^T dv_1 \f$
  virtual double dot_product_with(const ObsAuxCtlVecBase &) const = 0;

  // I/O and diagnostics
  /// Read control vector
  virtual void read(const eckit::Configuration &) = 0;
  /// Write control vector
  virtual void write(const eckit::Configuration &) const = 0;
  /// L-2 norm
  virtual double norm() const = 0;

  /// Interfacing
  virtual const int &getAccess() const = 0;
  virtual int &getAccess() = 0;

 private:
  virtual void print(std::ostream &) const = 0;
};

// -----------------------------------------------------------------------------

/// ObsAuxCtlVec Factory
template <typename MODEL>
class ObsAuxCtlVecFactory {
  using ObsAuxCovarianceBase_ = ObsAuxCovarianceBase<MODEL>;

 public:
  static ObsAuxCtlVecBase<MODEL> *create(const ObsAuxCovarianceBase_ &);
  virtual ~ObsAuxCtlVecFactory() { getMakers().clear(); }

 protected:
  explicit ObsAuxCtlVecFactory(const std::string &);

 private:
  virtual ObsAuxCtlVecBase<MODEL> *make(const ObsAuxCovarianceBase_ &) = 0;
  static std::map<std::string, ObsAuxCtlVecFactory<MODEL> *> &getMakers() {
    static std::map<std::string, ObsAuxCtlVecFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <typename MODEL, class T>
class ObsAuxCtlVecMaker : public ObsAuxCtlVecFactory<MODEL> {
  using ObsAuxCovarianceBase_ = ObsAuxCovarianceBase<MODEL>;
  virtual ObsAuxCtlVecBase<MODEL> *make(const ObsAuxCovarianceBase_ &cov) {
    return new T(cov);
  }

 public:
  explicit ObsAuxCtlVecMaker(const std::string &name)
      : ObsAuxCtlVecFactory<MODEL>(name) {}
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCtlVecFactory<MODEL>::ObsAuxCtlVecFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in ObsAuxCtlVecFactory."
                 << std::endl;
    ABORT("Element already registered in ObsAuxCtlVecFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxCtlVecBase<MODEL> *ObsAuxCtlVecFactory<MODEL>::create(
    const ObsAuxCovarianceBase<MODEL> &cov) {
  const std::string id = cov.config().getString("type");
  typename std::map<std::string, ObsAuxCtlVecFactory<MODEL> *>::iterator jloc =
      getMakers().find(id);
  if (jloc == getMakers().end()) {
    Log::error() << id << " does not exist in ObsAuxCtlVecFactory."
                 << std::endl;
    ABORT("Element does not exist in ObsAuxCtlVecFactory.");
  }
  ObsAuxCtlVecBase<MODEL> *ptr = jloc->second->make(cov);
  return ptr;
}

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCTLVECBASE_H_
