/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXINCREMENT_H_
#define OOPS_INTERFACE_OBSAUXINCREMENT_H_

#include <iostream>
#include <map>
#include <memory>
#include <string>

#include "eckit/exception/Exceptions.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsAuxIncrementBase.h"
#include "oops/interface/ObservationSpace.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
class Configuration;
}

namespace oops {

/*!  @brief Encapsulates auxiliary observation space components of
 *  ControlIncrement.

 *   This class provides operations associated with the auxiliary observation
 *  space components (e.g. VarBC, TOVSC) of ControlIncrement. It contains a
 *  vector of ObsAuxIncrementBase objects, where ObsAuxIncrementBase is a base
 *  class for model specific (or potentially generic) classes.

 *   ObsAuxIncrements class wraps around a vector of ObsAuxIncrement objects
 *  and is part of ControlIncrement. Refer to the documentation of
 *  ControlIncrement and CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxIncrement : public util::Printable,
                        private util::ObjectCounter<ObsAuxIncrement<MODEL>> {
  using ObsAuxIncrementBase_ = ObsAuxIncrementBase<MODEL>;
  using ObsAuxControl_ = ObsAuxControl<MODEL>;
  using ObservationSpace_ = ObservationSpace<MODEL>;
  using ObsAuxIncrementPtrMap_ =
      std::map<std::string, std::unique_ptr<ObsAuxIncrementBase_>>;
  using iter_ = typename ObsAuxIncrementPtrMap_::iterator;
  using icst_ = typename ObsAuxIncrementPtrMap_::const_iterator;

 public:
  static const std::string classname() { return "oops::ObsAuxIncrement"; }

  /// Constructor
  ObsAuxIncrement(const ObservationSpace_ &, const eckit::Configuration &);
  /// Copy constructor
  ObsAuxIncrement(const ObsAuxIncrement &, const bool copy = true);
  /// Copy constructor
  ObsAuxIncrement(const ObsAuxIncrement &, const eckit::Configuration &);
  /// Destructor
  ~ObsAuxIncrement();

  // Linear algebra operators
  /// \f$ dx = x_1 - x_2  \f$
  void diff(const ObsAuxControl_ &, const ObsAuxControl_ &);
  /// \f$ dx_1 = 0 \f$
  void zero();
  /// \f$ dx_1 = dx_2 \f$
  ObsAuxIncrement &operator=(const ObsAuxIncrement &);
  /// \f$ dx_1 = dx_1 + dx_2 \f$
  ObsAuxIncrement &operator+=(const ObsAuxIncrement &);
  /// \f$ dx_1 = dx_1 - dx_2 \f$
  ObsAuxIncrement &operator-=(const ObsAuxIncrement &);
  /// \f$ dx_1 = dx_1 * c \f$
  ObsAuxIncrement &operator*=(const double &);
  /// \f$ dx_1 = dx_1 * c dx_2 \f$
  void axpy(const double &, const ObsAuxIncrement &);
  /// \f$ <dx_1,dx_1> = dx_1^T dx_1 \f$
  double dot_product_with(const ObsAuxIncrement &) const;

  // I/O and diagnostics
  /// Read control increment
  void read(const eckit::Configuration &);
  /// Write control increment
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

  // Interfacing
  /// Return a constant reference to the model specific object
  const ObsAuxIncrementBase_ &obsauxincrement(const std::string &id) const {
    return this->get(id);
  }
  /// Return a reference to the model specific object
  ObsAuxIncrementBase_ &obsauxincrement(const std::string &id) {
    return this->get(id);
  }

  /// Return a constant reference to the map of model specific objects
  const ObsAuxIncrementPtrMap_ &obsauxincrement() const { return incs_; }
  /// Return a reference to the map of model specific objects
  ObsAuxIncrementPtrMap_ &obsauxincrement() { return incs_; }

  /// Return a constant reference to the vector of keys
  const std::vector<std::string> &getKeys() const { return keys_; }

  /// Return the number of ObsAuxIncrementBase objects
  const size_t &size() const { return size_; }

 private:
  /// Return a constant reference to the model specific object
  const ObsAuxIncrementBase_ &get(const std::string &) const;
  /// Return a reference to the model specific object
  ObsAuxIncrementBase_ &get(const std::string &);
  void print(std::ostream &) const;

  mutable size_t size_;            ///< Number of model specific objects
  std::vector<std::string> keys_;  ///< Vector of keys to model specific objects
  ObsAuxIncrementPtrMap_ incs_;  ///< Map of pointers to model specific objects
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxControl<MODEL> &operator+=(ObsAuxControl<MODEL> &xx,
                                 const ObsAuxIncrement<MODEL> &dx) {
  Log::trace() << "operator+=(ObsAuxControl, ObsAuxIncrement) starting"
               << std::endl;
  util::Timer timer("oops::ObsAuxIncrement", "operator+=ObsAuxControl");
  using icst = typename std::map<
      std::string, std::unique_ptr<ObsAuxIncrementBase<MODEL>>>::const_iterator;
  for (icst it = dx.obsauxincrement().begin(); it != dx.obsauxincrement().end();
       ++it) {
    xx.obsauxcontrol(it->first) += dx.obsauxincrement(it->first);
  }
  Log::trace() << "operator+=(ObsAuxControl, ObsAuxIncrement) done"
               << std::endl;
  return xx;
}

// =============================================================================

template <typename MODEL>
ObsAuxIncrement<MODEL>::ObsAuxIncrement(const ObservationSpace_ &os,
                                        const eckit::Configuration &conf)
    : size_(0), keys_(), incs_() {
  Log::trace() << "ObsAuxIncrement<MODEL>::ObsAuxIncrement starting"
               << std::endl;
  util::Timer timer(classname(), "ObsAuxIncrement");

  std::vector<eckit::LocalConfiguration> confs;
  conf.get("covarianceObsAux", confs);
  size_ = confs.size();

  for (auto const &config : confs) {
    std::string key(config.getString("type"));
    std::unique_ptr<ObsAuxIncrementBase<MODEL>> auxinc(
        ObsAuxIncrementFactory<MODEL>::create(os, config));
    incs_.emplace(key, std::move(auxinc));
    keys_.push_back(key);
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::ObsAuxIncrement done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL>::ObsAuxIncrement(const ObsAuxIncrement &other,
                                        const bool copy)
    : size_(other.size_), keys_(), incs_() {
  Log::trace() << "ObsAuxIncrement<MODEL>::ObsAuxIncrement starting"
               << std::endl;
  util::Timer timer(classname(), "ObsAuxIncrement");
  for (icst_ it = other.incs_.begin(); it != other.incs_.end(); ++it) {
    std::string incname = it->first;
    std::unique_ptr<ObsAuxIncrementBase<MODEL>> auxinc(it->second->clone(copy));
    incs_.emplace(incname, std::move(auxinc));
    keys_.push_back(incname);
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::ObsAuxIncrement done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL>::ObsAuxIncrement(const ObsAuxIncrement &other,
                                        const eckit::Configuration &conf)
    : size_(other.size_), keys_(), incs_() {
  Log::trace() << "ObsAuxIncrement<MODEL>::ObsAuxIncrement starting"
               << std::endl;
  util::Timer timer(classname(), "ObsAuxIncrement");
  for (icst_ it = other.incs_.begin(); it != other.incs_.end(); ++it) {
    std::string incname = it->first;
    std::unique_ptr<ObsAuxIncrementBase<MODEL>> auxinc(it->second->clone(conf));
    incs_.emplace(incname, std::move(auxinc));
    keys_.push_back(incname);
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::ObsAuxIncrement done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL>::~ObsAuxIncrement() {
  Log::trace() << "ObsAuxIncrement<MODEL>::~ObsAuxIncrement starting"
               << std::endl;
  util::Timer timer(classname(), "~ObsAuxIncrement");
  Log::trace() << "ObsAuxIncrement<MODEL>::~ObsAuxIncrement done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrementBase<MODEL> &ObsAuxIncrement<MODEL>::get(const std::string &id) {
  iter_ it = incs_.find(id);
  ASSERT(it != incs_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
const ObsAuxIncrementBase<MODEL> &ObsAuxIncrement<MODEL>::get(
    const std::string &id) const {
  icst_ it = incs_.find(id);
  ASSERT(it != incs_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrement<MODEL>::diff(const ObsAuxControl_ &x1,
                                  const ObsAuxControl_ &x2) {
  Log::trace() << "ObsAuxIncrement<MODEL>::diff starting" << std::endl;
  util::Timer timer(classname(), "diff");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    it->second->diff(x1.obsauxcontrol(it->first), x2.obsauxcontrol(it->first));
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::diff done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrement<MODEL>::zero() {
  Log::trace() << "ObsAuxIncrement<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    it->second->zero();
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL> &ObsAuxIncrement<MODEL>::operator=(
    const ObsAuxIncrement &rhs) {
  Log::trace() << "ObsAuxIncrement<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    *(it->second) = rhs.obsauxincrement(it->first);
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL> &ObsAuxIncrement<MODEL>::operator+=(
    const ObsAuxIncrement &rhs) {
  Log::trace() << "ObsAuxIncrement<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    *(it->second) += rhs.obsauxincrement(it->first);
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL> &ObsAuxIncrement<MODEL>::operator-=(
    const ObsAuxIncrement &rhs) {
  Log::trace() << "ObsAuxIncrement<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    *(it->second) -= rhs.obsauxincrement(it->first);
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ObsAuxIncrement<MODEL> &ObsAuxIncrement<MODEL>::operator*=(const double &zz) {
  Log::trace() << "ObsAuxIncrement<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    *(it->second) *= zz;
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrement<MODEL>::axpy(const double &zz, const ObsAuxIncrement &dx) {
  Log::trace() << "ObsAuxIncrement<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    it->second->axpy(zz, dx.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxIncrement<MODEL>::dot_product_with(
    const ObsAuxIncrement &dx) const {
  Log::trace() << "ObsAuxIncrement<MODEL>::dot_product_with starting"
               << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = 0.0;
  for (icst_ it = incs_.begin(); it != incs_.end(); ++it) {
    zz += it->second->dot_product_with(dx.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::dot_product_with done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrement<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "ObsAuxIncrement<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  std::vector<eckit::LocalConfiguration> confs;
  conf.get("incrementObsAux", confs);

  for (iter_ it = incs_.begin(); it != incs_.end(); ++it) {
    for (auto const &config : confs) {
      if (config.getString("type") == it->first) it->second->read(config);
    }
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrement<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "ObsAuxIncrement<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("incrementObsAux")) conf.get("incrementObsAux", confs);

  for (icst_ it = incs_.begin(); it != incs_.end(); ++it) {
    for (auto const &config : confs) {
      if (config.getString("type") == it->first) it->second->write(config);
    }
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::write done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ObsAuxIncrement<MODEL>::norm() const {
  Log::trace() << "ObsAuxIncrement<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0.0;
  for (icst_ it = incs_.begin(); it != incs_.end(); ++it) {
    double norm = it->second->norm();
    zz += norm * norm;
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::norm done" << std::endl;
  return sqrt(zz);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ObsAuxIncrement<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObsAuxIncrement<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  for (icst_ it = incs_.begin(); it != incs_.end(); ++it) {
    os << it->first << ": " << *it->second;
  }
  Log::trace() << "ObsAuxIncrement<MODEL>::print done" << std::endl;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXINCREMENT_H_
