/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXINCREMENTBASE_H_
#define OOPS_INTERFACE_OBSAUXINCREMENTBASE_H_

#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/interface/ObservationSpace.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"
#include "util/abor1_cpp.h"

namespace oops {

template <typename MODEL>
class ObsAuxControlBase;

/*!  @brief Base class for auxiliary observation space components of
 *  ControlIncrement.

 *   Defines interfaces for the auxiliary observation space components of
 *  ControlIncrement (e.g. VarBC, TOVSCV). ObsAuxIncrementFactory is used to
 *  produce concrete objects.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxIncrementBase : public util::Printable {
  using ObsAuxControlBase_ = ObsAuxControlBase<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsAuxIncrementBase"; }

  /// Constructor
  ObsAuxIncrementBase() {}
  /// Destructor
  virtual ~ObsAuxIncrementBase() {}

  /// Cloning
  virtual ObsAuxIncrementBase *clone(const bool copy = true) const = 0;
  virtual ObsAuxIncrementBase *clone(const eckit::Configuration &) const = 0;

  // Linear algebra operators
  /// \f$ dx = x_1 - x_2  \f$
  virtual void diff(const ObsAuxControlBase_ &, const ObsAuxControlBase_ &) = 0;
  /// \f$ dx_1 = 0 \f$
  virtual void zero() = 0;
  /// \f$ dx_1 = dx_2 \f$
  virtual ObsAuxIncrementBase &operator=(const ObsAuxIncrementBase &) = 0;
  /// \f$ dx_1 = dx_1 + dx_2 \f$
  virtual ObsAuxIncrementBase &operator+=(const ObsAuxIncrementBase &) = 0;
  /// \f$ dx_1 = dx_1 - dx_2 \f$
  virtual ObsAuxIncrementBase &operator-=(const ObsAuxIncrementBase &) = 0;
  /// \f$ dx_1 = dx_1 * c \f$
  virtual ObsAuxIncrementBase &operator*=(const double &) = 0;
  /// \f$ dx_1 = dx_1 * c dx_2 \f$
  virtual void axpy(const double &, const ObsAuxIncrementBase &) = 0;
  /// \f$ <dx_1,dx_1> = dx_1^T dx_1 \f$
  virtual double dot_product_with(const ObsAuxIncrementBase &) const = 0;

  // virtual / I/O and diagnostics
  /// Read
  virtual void read(const eckit::Configuration &) = 0;
  /// Write
  virtual void write(const eckit::Configuration &) const = 0;
  /// L-2 norm
  virtual double norm() const = 0;

  // Interfacing
  virtual const int &getAccess() const = 0;
  virtual int &getAccess() = 0;

 private:
  virtual void print(std::ostream &) const = 0;
};

// -----------------------------------------------------------------------------

/// ObsAuxIncrement Factory

template <typename MODEL>
class ObsAuxIncrementFactory {
 public:
  static ObsAuxIncrementBase<MODEL> *create(const ObservationSpace<MODEL> &,
                                            const eckit::Configuration &);
  virtual ~ObsAuxIncrementFactory() { getMakers().clear(); }

 protected:
  explicit ObsAuxIncrementFactory(const std::string &);

 private:
  virtual ObsAuxIncrementBase<MODEL> *make(const ObservationSpace<MODEL> &,
                                           const eckit::Configuration &) = 0;
  static std::map<std::string, ObsAuxIncrementFactory<MODEL> *> &getMakers() {
    static std::map<std::string, ObsAuxIncrementFactory<MODEL> *> makers_;
    return makers_;
  }
};

// -----------------------------------------------------------------------------

template <typename MODEL, class T>
class ObsAuxIncrementMaker : public ObsAuxIncrementFactory<MODEL> {
  virtual ObsAuxIncrementBase<MODEL> *make(const ObservationSpace<MODEL> &os,
                                           const eckit::Configuration &conf) {
    return new T(os.observationspace(), conf);
  }

 public:
  explicit ObsAuxIncrementMaker(const std::string &name)
      : ObsAuxIncrementFactory<MODEL>(name) {}
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxIncrementFactory<MODEL>::ObsAuxIncrementFactory(const std::string &name) {
  if (getMakers().find(name) != getMakers().end()) {
    Log::error() << name << " already registered in ObsAuxIncrementFactory."
                 << std::endl;
    ABORT("Element already registered in ObsAuxIncrementFactory.");
  }
  getMakers()[name] = this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsAuxIncrementBase<MODEL> *ObsAuxIncrementFactory<MODEL>::create(
    const ObservationSpace<MODEL> &os, const eckit::Configuration &conf) {
  const std::string id = conf.getString("type");
  typename std::map<std::string, ObsAuxIncrementFactory<MODEL> *>::iterator
      jloc = getMakers().find(id);
  if (jloc == getMakers().end()) {
    Log::error() << id << " does not exist in ObsAuxIncrementFactory."
                 << std::endl;
    ABORT("Element does not exist in ObsAuxIncrementFactory.");
  }
  ObsAuxIncrementBase<MODEL> *ptr = jloc->second->make(os, conf);
  return ptr;
}

// =============================================================================

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXINCREMENTBASE_H_
