/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSERRORCOVARIANCE_H_
#define OOPS_INTERFACE_OBSERRORCOVARIANCE_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/assimilation/ControlObsVector.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsErrorBase.h"
#include "oops/interface/ObsVector.h"
#include "oops/interface/ObservationSpace.h"
#include "util/Printable.h"

namespace eckit {
class Configuration;
}

namespace oops {

/*!  @brief Encapsulates observation and observation bias error covariance
 *  matrices.

 *   This class provides the operations associated with the observation and
 *  observation bias error covariance matrices. It wraps the actual observation
 *  and observation bias error covariance matrices which can be model specific
 *  or generic.

 *   The interface for the observation error comprises two levels
 *  (ObsErrorCovariance and ObsErrorBase) because we want run time polymorphism.

 *   The observation and observation bias error covariance matrix has block
 *  diagonal structure and is defined as:

    \f[
      \bar{R} \bar{y} =
           \begin{bmatrix}
              \tilde{R}  &&  0                  \\
              0          &&  \tilde{R}_{\beta}
           \end{bmatrix}
           \begin{bmatrix}
              \tilde{y} \\
              \tilde{b}
           \end{bmatrix}

    \f]

 *   where \f$ \tilde{R} \f$ is the observation error covariance matrix and
 *  \f$ \tilde{R}_{\beta} \f$ is the observation bias error covariance matrix.
 *  \f$ \bar{y} \f$ is implemented by ControlObsVector, which holds two
 *  ObsVector components corresponding to observations \f$ \tilde{y} \f$ and
 *  biases \f$ \tilde{b} \f$.

 *   If VarQC is active, it is applied in a square root symmetric form:

    \f[
      \bar{W}^{1/2} \bar{R} \bar{W}^{1/2} \bar{y} =
          \begin{bmatrix}
           \tilde{W}^{1/2}  &&  0  \\
           0                &&  I
          \end{bmatrix}
          \begin{bmatrix}
           \tilde{R}  &&  0                 \\
           0          &&  \tilde{R}_{\beta}
          \end{bmatrix}
          \begin{bmatrix}
           \tilde{W}^{1/2}  &&  0  \\
           0                &&  I
          \end{bmatrix}
          \begin{bmatrix}
           \tilde{y} \\
           \tilde{b}
          \end{bmatrix}
    \f]

     where \f$ \tilde{W}^{1/2} \f$ is a diagonal matrix of VarQC weights.

 *    ObsErrors class holds a vector of ObsErrorCovariance objects. Refere to
 *   the documentation of CostJo to understand the structure.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsErrorCovariance
    : public util::Printable,
      private util::ObjectCounter<ObsErrorCovariance<MODEL> >,
      private eckit::NonCopyable {
  using ObsAuxCtrl_ = ObsAuxControl<MODEL>;
  using ObsErrorBase_ = ObsErrorBase<MODEL>;
  using ObsSpace_ = ObservationSpace<MODEL>;
  using ObsVector_ = ObsVector<MODEL>;
  using ControlObsVector_ = ControlObsVector<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsErrorCovariance"; }

  /// Constructor
  ObsErrorCovariance(const ObsSpace_ &, const eckit::Configuration &);
  /// Destructor
  ~ObsErrorCovariance();

  /// Setup the covariance matrix
  void linearize(const ControlObsVector_ &);
  /// Setup the covariance matrix
  void linearize(const ObsAuxCtrl_ &, const ControlObsVector_ &);
  /// Setup the VarQC weights
  void setupWeights(const ControlObsVector_ &);

  /// Multiply a Departure by \f$\bar{R}\f$
  ControlObsVector_ *multiply(const ControlObsVector_ &) const;
  /// Multiply a Departure by \f$\bar{R}^{-1}\f$
  ControlObsVector_ *inverseMultiply(const ControlObsVector_ &) const;
  /// Multiply a Departure by \f$\tilde{R}^{-1/2}\f$
  ControlObsVector_ *inverseMultiplySqrt(
      const ControlObsVector_ &) const;
  /// Multiply a Departure by \f$\tilde{R}^{-T/2}\f$
  ControlObsVector_ *inverseMultiplySqrtTrans(
      const ControlObsVector_ &) const;

  /// Generate random perturbation
  void randomize(ControlObsVector_ &) const;

  /// Get mean error for Jo table
  double getRMSE() const;

 private:
  /// \f$ \bar{y}_1 = \bar{W}^{1/2} \bar{y}_2 \f$
  ControlObsVector_ *multiplyWghtSqrt_(const ControlObsVector_ &) const;

  /// \f$ \bar{y}_1 = \bar{W} \bar{y}_2 \f$
  ControlObsVector_ *multiplyWght_(const ControlObsVector_ &) const;

  /// \f$ \bar{y}_1 = \bar{R}^{-1} \bar{y}_2 \f$
  ControlObsVector_ *inverseMultiply_(const ControlObsVector_ &) const;

  /// \f$ \bar{y}_1 = \bar{R}^{-1/2} \bar{y}_2 \f$
  ControlObsVector_ *inverseMultiplySqrt_(
      const ControlObsVector_ &) const;

  /// \f$ \bar{y}_1 = \bar{R}^{-T/2} \bar{y}_2 \f$
  ControlObsVector_ *inverseMultiplySqrtTrans_(
      const ControlObsVector_ &) const;
  void print(std::ostream &) const;

  bool lvarqc_;  ///< Controls VarQC
  const bool leigen_;
  std::unique_ptr<ObsErrorBase_>
      covar_;  ///< Pointer to observation error covariance
  std::unique_ptr<ObsErrorBase_>
      biascovar_;  ///< Pointer to bias error covariance
};

// ====================================================================================

template <typename MODEL>
ObsErrorCovariance<MODEL>::ObsErrorCovariance(const ObsSpace_ &obsdb,
                                              const eckit::Configuration &conf)
    : lvarqc_(false),
      leigen_(conf.getBool("eigenProjection", false)),
      covar_(),
      biascovar_() {
  Log::trace() << "ObsErrorCovariance<MODEL>::ObsErrorCovariance starting"
               << std::endl;
  util::Timer timer(classname(), "ObsErrorCovariance");
  eckit::LocalConfiguration obscovarconf(conf, "ObsErrorCovariance");
  covar_.reset(ObsErrorFactory<MODEL>::create(obsdb, obscovarconf));
  if (conf.has("BiasErrorCovariance")) {
    eckit::LocalConfiguration biascovarconf(conf, "BiasErrorCovariance");
    biascovar_.reset(ObsErrorFactory<MODEL>::create(obsdb, biascovarconf));
  }
  Log::trace() << "ObsErrorCovariance<MODEL>::ObsErrorCovariance done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsErrorCovariance<MODEL>::~ObsErrorCovariance() {
  Log::trace() << "ObsErrorCovariance<MODEL>::~ObsErrorCovariance starting"
               << std::endl;
  util::Timer timer(classname(), "~ObsErrorCovariance");
  covar_.reset();
  biascovar_.reset();
  Log::trace() << "ObsErrorCovariance<MODEL>::~ObsErrorCovariance done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::linearize(const ControlObsVector_ &yy) {
  Log::trace() << "ObsErrorCovariance<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  lvarqc_ = false;
  covar_->linearize(yy.obsvector().obsvector());
  if (biascovar_) biascovar_->linearize(yy.biasvector().obsvector());
  Log::trace() << "ObsErrorCovariance<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::linearize(const ObsAuxCtrl_ &xx,
                                          const ControlObsVector_ &yy) {
  Log::trace() << "ObsErrorCovariance<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  lvarqc_ = false;
  covar_->linearize(yy.obsvector().obsvector());
  if (biascovar_)
    biascovar_->linearize(xx.obsauxcontrol(), yy.biasvector().obsvector());
  Log::trace() << "ObsErrorCovariance<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::setupWeights(const ControlObsVector_ &yy) {
  Log::trace() << "ObsErrorCovariance<MODEL>::setupWeights starting"
               << std::endl;
  util::Timer timer(classname(), "setupWeights");
  //  std::unique_ptr<ControlObsVector<MODEL> >
  //  rsqrtinvdy(this->inverseMultiplySqrt_(yy));
  //  covar_->setupWeights(rsqrtinvdy->obsvector().obsvector());
  covar_->setupWeights(yy.obsvector().obsvector());
  lvarqc_ = true;
  Log::trace() << "ObsErrorCovariance<MODEL>::setupWeights done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::multiply(
    const ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  ObsVector_ *dyobs =
      new ObsVector_(covar_->multiply(dy.obsvector().obsvector()));
  ObsVector_ *dybias = nullptr;
  if (biascovar_)
    dybias = new ObsVector_(biascovar_->multiply(dy.biasvector().obsvector()));
  ControlObsVector_ *dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::multiply done" << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::multiplyWghtSqrt_(
    const ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::multiplyWghtSqrt_ starting"
               << std::endl;
  util::Timer timer(classname(), "multiplyWghtSqrt_");
  ObsVector_ *dyobs =
      new ObsVector_(covar_->multiplyWghtSqrt(dy.obsvector().obsvector()));
  ObsVector_ *dybias = nullptr;
  if (biascovar_)
    dybias =
        new ObsVector_(covar_->multiplyWghtSqrt(dy.biasvector().obsvector()));
  ControlObsVector_ *dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::multiplyWghtSqrt_ done"
               << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::multiplyWght_(
    const ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::multiplyWght_ starting"
               << std::endl;
  util::Timer timer(classname(), "multiplyWght_");
  ObsVector_ *dyobs =
      new ObsVector_(covar_->multiplyWght(dy.obsvector().obsvector()));
  ObsVector_ *dybias = nullptr;
  if (biascovar_)
    dybias = new ObsVector_(covar_->multiplyWght(dy.biasvector().obsvector()));
  ControlObsVector_ *dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::multiplyWght_ done" << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::inverseMultiply_(
    const ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply_ starting"
               << std::endl;
  util::Timer timer(classname(), "inverseMultiply_");
  ObsVector_ *dyobs =
      new ObsVector_(covar_->inverseMultiply(dy.obsvector().obsvector()));
  ObsVector_ *dybias = nullptr;
  if (biascovar_)
    dybias = new ObsVector_(
        biascovar_->inverseMultiply(dy.biasvector().obsvector()));
  ControlObsVector_ *dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply_ done"
               << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::inverseMultiplySqrtTrans_(
    const ControlObsVector_ &dy) const {
  Log::trace()
      << "ObsErrorCovariance<MODEL>::inverseMultiplySqrtTrans_ starting"
      << std::endl;
  util::Timer timer(classname(), "inverseMultiplySqrtTrans_");
  ObsVector_ *dyobs = new ObsVector_(
      covar_->inverseMultiplySqrtTrans(dy.obsvector().obsvector()));
  ObsVector_ *dybias = nullptr;
  if (biascovar_)
    dybias = new ObsVector_(
        biascovar_->inverseMultiplySqrtTrans(dy.biasvector().obsvector()));
  ControlObsVector_ *dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiplySqrtTrans_ done"
               << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::inverseMultiplySqrt_(
    const ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiplySqrt_ starting"
               << std::endl;
  util::Timer timer(classname(), "inverseMultiplySqrt_");
  ObsVector_ *dyobs =
      new ObsVector_(covar_->inverseMultiplySqrt(dy.obsvector().obsvector()));
  ObsVector_ *dybias = nullptr;
  if (biascovar_)
    dybias = new ObsVector_(
        biascovar_->inverseMultiplySqrt(dy.biasvector().obsvector()));
  ControlObsVector_ *dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiplySqrt_ done"
               << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> *ObsErrorCovariance<MODEL>::inverseMultiply(
    const ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply starting"
               << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  ControlObsVector<MODEL> *rinvwrinvdy = nullptr;
  ControlObsVector<MODEL> *wrinvwdy = nullptr;
  ControlObsVector<MODEL> *dyout = nullptr;
  if (leigen_) {
    if (lvarqc_) {
      std::unique_ptr<ControlObsVector<MODEL> > rinvdy(
          this->inverseMultiplySqrt_(dy));
      std::unique_ptr<ControlObsVector<MODEL> > wrinvdy(
          this->multiplyWght_(*rinvdy));
      rinvwrinvdy = this->inverseMultiplySqrtTrans_(*wrinvdy);
    } else {
      std::unique_ptr<ControlObsVector<MODEL> > rinvdy(
          this->inverseMultiplySqrt_(dy));
      rinvwrinvdy = this->inverseMultiplySqrtTrans_(*rinvdy);
    }
    dyout = rinvwrinvdy;
  } else {
    if (lvarqc_) {
      std::unique_ptr<ControlObsVector<MODEL> > wdy(
          this->multiplyWghtSqrt_(dy));
      std::unique_ptr<ControlObsVector<MODEL> > rinvwdy(
          this->inverseMultiply_(*wdy));
      wrinvwdy = this->multiplyWghtSqrt_(*rinvwdy);
    } else {
      wrinvwdy = this->inverseMultiply_(dy);
    }
    dyout = wrinvwdy;
  }
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply done" << std::endl;
  return dyout;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::randomize(ControlObsVector_ &dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::randomize starting" << std::endl;
  util::Timer timer(classname(), "randomize");
  covar_->randomize(dy.obsvector().obsvector());
  Log::trace() << "ObsErrorCovariance<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double ObsErrorCovariance<MODEL>::getRMSE() const {
  Log::trace() << "ObsErrorCovariance<MODEL>::getRMSE starting" << std::endl;
  util::Timer timer(classname(), "getRMSE");
  double zz = covar_->getRMSE();
  Log::trace() << "ObsErrorCovariance<MODEL>::getRMSE done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *covar_;
  if (biascovar_) os << *biascovar_;
  Log::trace() << "ObsErrorCovariance<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSERRORCOVARIANCE_H_
