/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSOPERATOR_H_
#define OOPS_INTERFACE_OBSOPERATOR_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/memory/NonCopyable.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsOperatorBase.h"
#include "oops/interface/ObsVector.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

/*! @brief Encapsulates the observation operator.

 *   This class provides the operations associated with the observation
 *  operator. It wraps the actual observation operator which is derived
 *  from the operator base class ObsOperatorBase. The interface for the
 *  observation operator comprises two levels because we want run time
 *  polymorphism. Note that it may be desirable to define an observation
 *  operator for each individual observation type rather than a single
 *  composite operator (which is the case the IFS). ObsOperatorFactory
 *  produces concrete objects of ObsOperatorBase children classes.

 *   ObsOperators  class contains a vector of ObsOperator objects. Refer
 *  to the documentation of CostJo class to understand the structure of
 *  the \f$ J_o \f$ term.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsOperator : public util::Printable,
                    private eckit::NonCopyable,
                    private util::ObjectCounter<ObsOperator<MODEL> > {
  using ObsOperatorBase_ = ObsOperatorBase<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using ObsAuxControl_ = ObsAuxControl<MODEL>;
  using ObsVector_ = ObsVector<MODEL>;
  using ObsSpace_ = ObservationSpace<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::ObsOperator"; }

  /// Constructor
  explicit ObsOperator(const ObsSpace_ &);
  /// Destructor
  ~ObsOperator();

  /// Interfacing
  const ObsOperatorBase_ &obsoperator() const { return *oper_; }

  // Obs Operators
  /// Evaluate the non-linear observation operator \f$ {\cal \bar{H}} \f$
  void obsEquiv(const GeoVaLs_ &, ObsVector_ &, const ObsAuxControl_ &) const;
  /// Evaluate the non-linear bias operator \f$ {\cal \tilde{P}} \f$
  void obsBiasEquiv(const GeoVaLs_ &, ObsVector_ &,
                    const ObsAuxControl_ &) const;

  // Other
  /// Retrun variables (required input variables from Model)
  Variables_ variables() const;

 private:
  void print(std::ostream &) const;

  std::unique_ptr<ObsOperatorBase_>
      oper_;  ///< Pointer to a ObsOperatorBase object
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsOperator<MODEL>::ObsOperator(const ObsSpace_ &os) : oper_() {
  Log::trace() << "ObsOperator<MODEL>::ObsOperator starting" << std::endl;
  util::Timer timer(classname(), "ObsOperator");
  oper_.reset(
      ObsOperatorFactory<MODEL>::create(os.observationspace(), os.config()));
  Log::trace() << "ObsOperator<MODEL>::ObsOperator done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsOperator<MODEL>::~ObsOperator() {
  Log::trace() << "ObsOperator<MODEL>::~ObsOperator starting" << std::endl;
  util::Timer timer(classname(), "~ObsOperator");
  oper_.reset();
  Log::trace() << "ObsOperator<MODEL>::~ObsOperator done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsOperator<MODEL>::obsEquiv(const GeoVaLs_ &gvals, ObsVector_ &yy,
                                  const ObsAuxControl_ &aux) const {
  /*! \f[ \begin{bmatrix} {\cal \tilde{H}} && {\cal \tilde{P}} \end{bmatrix}
          \begin{bmatrix} {\cal I} (x) \\ x_{obsaux} \end{bmatrix}
      \f]

      Note: -> \f$ {\cal I} (x) \f$ is stored by GeoVaLs.
   */
  Log::trace() << "ObsOperator<MODEL>::obsEquiv starting" << std::endl;
  util::Timer timer(classname(), "ObsEquiv");
  oper_->obsEquiv(gvals.geovals(), yy.obsvector(), aux.obsauxcontrol());
  Log::trace() << "ObsOperator<MODEL>::obsEquiv done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsOperator<MODEL>::obsBiasEquiv(const GeoVaLs_ &gvals, ObsVector_ &yybias,
                                      const ObsAuxControl_ &aux) const {
  Log::trace() << "ObsOperator<MODEL>::obsEquiv starting" << std::endl;
  util::Timer timer(classname(), "ObsBiasEquiv");
  oper_->obsBiasEquiv(gvals.geovals(), yybias.obsvector(), aux.obsauxcontrol());
  Log::trace() << "ObsOperator<MODEL>::obsBiasEquiv done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Variables<MODEL> ObsOperator<MODEL>::variables() const {
  Log::trace() << "ObsOperator<MODEL>::variables starting" << std::endl;
  util::Timer timer(classname(), "variables");
  Variables<MODEL> var(oper_->variables());
  Log::trace() << "ObsOperator<MODEL>::variables done" << std::endl;
  return var;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsOperator<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObsOperator<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *oper_;
  Log::trace() << "ObsOperator<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSOPERATOR_H_
