/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSERVATIONSPACE_H_
#define OOPS_INTERFACE_OBSERVATIONSPACE_H_

#include <memory>
#include <ostream>
#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "oops/interface/Geometry.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace util {
class DateTime;
}

namespace oops {
template <typename T>
class ObsVector;
template <typename T>
class GeoVaLs;

/*! @brief Encapsulates the observation space.

 *   ObservationSpace is a bucket for observations; ObservationSpace object can
 *  contain observations of specific observation type/types, or it can include
 *  observations of all the observation types. This information is agnostic to
 *  the OOPS control layer.

 *   ObservationSpace is responsible for performing screening of observations.

 *   ObservationSpaces class contains a vector of ObservationSpace objects.
 *  Refer to the documentation of CostJo class to understand the structure of
 *  the \f$ J_o \f$ term.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObservationSpace : public util::Printable,
                         private util::ObjectCounter<ObservationSpace<MODEL> > {
  using ObsSpace_ = typename MODEL::ObsSpace;
  using Geometry_ = Geometry<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using ObsVector_ = ObsVector<MODEL>;

 public:
  static const std::string classname() { return "oops::ObservationSpace"; }

  /// Constructor
  ObservationSpace(const eckit::Configuration &, const Geometry_ &,
                   const util::DateTime &, const util::DateTime &,
                   const bool lscreened = false);
  /// Destructor
  ~ObservationSpace();

  // Interfacing
  /// Return a reference to the model specific object
  ObsSpace_ &observationspace() const { return *obsdb_; }
  /// Return a constant reference to the configuration
  const eckit::Configuration &config() const { return conf_; }

  /// Assimilation window start
  const util::DateTime &windowStart() const { return winbgn_; }
  /// Assimilation window end
  const util::DateTime &windowEnd() const { return winend_; }
  /// Assimilation window length
  const util::Duration &windowLength() const { return winlen_; }

  /// Perform screening
  void screenObservations(const ObsVector_ &, const GeoVaLs_ &) const;

  /// Save observations
  void saveObservations() const;

  // Other
  /// Generate observations
  void generateDistribution(const eckit::Configuration &);
  /// Print the Jo table
  void printJo(const ObsVector_ &, const ObsVector_ &) const;

 private:
  void print(std::ostream &) const;

  const util::DateTime winbgn_;       ///< Window start
  const util::DateTime winend_;       ///< Window end
  const util::Duration winlen_;       ///< Window length
  eckit::LocalConfiguration conf_;    ///< Configuration
  std::unique_ptr<ObsSpace_> obsdb_;  ///< Pointer to the model specific object
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObservationSpace<MODEL>::ObservationSpace(const eckit::Configuration &conf,
                                          const Geometry_ &geom,
                                          const util::DateTime &bgn,
                                          const util::DateTime &end,
                                          const bool lscreened)
    : winbgn_(bgn), winend_(end), winlen_(end - bgn), conf_(conf), obsdb_() {
  Log::trace() << "ObservationSpace<MODEL>::ObservationSpace starting"
               << std::endl;
  util::Timer timer(classname(), "ObservationSpace");
  obsdb_.reset(new ObsSpace_(conf, geom.geometry(), bgn, end, lscreened));
  Log::trace() << "ObservationSpace<MODEL>::ObservationSpace done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObservationSpace<MODEL>::~ObservationSpace() {
  Log::trace() << "ObservationSpace<MODEL>::~ObservationSpace" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::screenObservations(const ObsVector_ &dy,
                                                 const GeoVaLs_ &gvals) const {
  Log::trace() << "ObservationSpace<MODEL>::screenObservations starting"
               << std::endl;
  util::Timer timer(classname(), "screenObservations");
  obsdb_->screenObservations(dy.obsvector(), gvals.geovals());
  Log::trace() << "ObservationSpace<MODEL>::screenObservations done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::saveObservations() const {
  Log::trace() << "ObservationSpace<MODEL>::saveObservations starting"
               << std::endl;
  util::Timer timer(classname(), "saveObservations");
  obsdb_->saveObservations();
  Log::trace() << "ObservationSpace<MODEL>::saveObservations done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::generateDistribution(
    const eckit::Configuration &conf) {
  Log::trace() << "ObservationSpace<MODEL>::generateDistribution starting"
               << std::endl;
  util::Timer timer(classname(), "generateDistribution");
  obsdb_->generateDistribution(conf);
  Log::trace() << "ObservationSpace<MODEL>::generateDistribution done"
               << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::printJo(const ObsVector_ &dy,
                                      const ObsVector_ &grad) const {
  Log::trace() << "ObservationSpace<MODEL>::printJo starting" << std::endl;
  util::Timer timer(classname(), "printJo");
  obsdb_->printJo(dy.obsvector(), grad.obsvector());
  Log::trace() << "ObservationSpace<MODEL>::printJo done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::print(std::ostream &os) const {
  Log::trace() << "ObservationSpace<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *obsdb_;
  Log::trace() << "ObservationSpace<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSERVATIONSPACE_H_
