/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_STATE_H_
#define OOPS_INTERFACE_STATE_H_

#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/Configuration.h"
#include "oops/base/PostProcessor.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Locations.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/LogbookWriter.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

template <typename MODEL>
class Model;
template <typename MODEL>
class LinearModel;

/*! @brief Encapsulates model state

 *   Wrapper around a model specific state class.

 *   \f$ x_{state4D} \f$ implemented by State4D implements the state component
 *  of ControlVariable and contains a vector of State objects.
 */

// -----------------------------------------------------------------------------

template <typename MODEL>
class State : public util::Printable,
              private util::ObjectCounter<State<MODEL> > {
  using State_ = typename MODEL::State;
  using Geometry_ = Geometry<MODEL>;
  using GeoVaLs_ = GeoVaLs<MODEL>;
  using Locations_ = Locations<MODEL>;
  using Model_ = Model<MODEL>;
  using LinearModel_ = LinearModel<MODEL>;
  using Variables_ = Variables<MODEL>;

 public:
  static const std::string classname() { return "oops::State"; }

  /// Constructor (with Model)
  State(const Geometry_ &, const Model_ &, const eckit::Configuration &);
  /// Constructor (with LinearModel)
  State(const Geometry_ &, const LinearModel_ &, const eckit::Configuration &);
  /// Copy constructor (with Model and Configuration)
  State(const Geometry_ &, const Model_ &, const State &,
        const eckit::Configuration &);
  /// Copy constructor (with Model)
  State(const Geometry_ &, const Model_ &, const State &);
  /// Copy constructor
  State(const Geometry_ &, const State &);
  /// Copy constructor (no interpolation)
  State(const State &);
  /// Destructor
  ~State();
  /// Assignment
  State &operator=(const State &);

  // Interfacing
  /// Return a reference to the model specific state
  State_ &state() { return *state_; }
  /// Return a constant reference to the model specific state
  const State_ &state() const { return *state_; }
  /// Get geometry
  Geometry_ geometry() const;

  /// Interpolate to observation locations
  void interpolate(const Locations_ &, GeoVaLs_ &) const;

  /// Interpolate full fields (for forcing linearisation trajectory)
  void forceWith(const State &, const Variables_ &);

  /// Time
  const util::DateTime &validTime() const { return state_->validTime(); }
  util::DateTime &validTime() { return state_->validTime(); }

  // I/O and diagnostics
  /// Read state
  void read(const eckit::Configuration &);
  /// Write state
  void write(const eckit::Configuration &) const;
  /// L-2 norm
  double norm() const;

  /// Zero out
  void zero();

 protected:
  /// Protected methods are for Accumulator. Could we find a better design?
  void accumul(const double &, const State &);

 private:
  void print(std::ostream &) const;

  std::unique_ptr<State_> state_;  ///< Pointer to the model specific object
  const Model_ &model_;            ///< Constant reference to the model
};

// =============================================================================

template <typename MODEL>
State<MODEL>::State(const Geometry_ &resol, const Model_ &model,
                    const eckit::Configuration &conf)
    : state_(), model_(model) {
  Log::trace() << "State<MODEL>::State read starting" << std::endl;
  util::Timer timer(classname(), "State");
  state_.reset(new State_(resol.geometry(), model.model(), conf));
  Log::trace() << "State<MODEL>::State read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL>::State(const Geometry_ &resol, const LinearModel_ &model,
                    const eckit::Configuration &conf)
    : state_(), model_(model) {
  Log::trace() << "State<MODEL>::State read starting" << std::endl;
  util::Timer timer(classname(), "State");
  state_.reset(new State_(resol.geometry(), model.model(), conf));
  Log::trace() << "State<MODEL>::State read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL>::State(const Geometry_ &resol, const Model_ &model,
                    const State &other, const eckit::Configuration &conf)
    : state_(), model_(model) {
  Log::trace() << "State<MODEL>::State interpolated starting" << std::endl;
  util::Timer timer(classname(), "State");
  state_.reset(
      new State_(resol.geometry(), model.model(), *other.state_, conf));
  Log::trace() << "State<MODEL>::State interpolated done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL>::State(const Geometry_ &resol, const Model_ &model,
                    const State &other)
    : state_(), model_(model) {
  Log::trace() << "State<MODEL>::State interpolated starting" << std::endl;
  util::Timer timer(classname(), "State");
  state_.reset(new State_(resol.geometry(), model.model(), *other.state_));
  Log::trace() << "State<MODEL>::State interpolated done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL>::State(const Geometry_ &resol, const State &other)
    : state_(), model_(other.model_) {
  Log::trace() << "State<MODEL>::State starting copy" << std::endl;
  util::Timer timer(classname(), "State");
  state_.reset(new State_(resol.geometry(), model_.model(), *other.state_));
  Log::trace() << "State<MODEL>::State copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL>::State(const State &other) : state_(), model_(other.model_) {
  Log::trace() << "State<MODEL>::State starting copy" << std::endl;
  util::Timer timer(classname(), "State");
  state_.reset(new State_(*other.state_));
  Log::trace() << "State<MODEL>::State copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL>::~State() {
  Log::trace() << "State<MODEL>::~State starting" << std::endl;
  util::Timer timer(classname(), "~State");
  state_.reset();
  Log::trace() << "State<MODEL>::~State done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
State<MODEL> &State<MODEL>::operator=(const State &rhs) {
  Log::trace() << "State<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  *state_ = *rhs.state_;
  Log::trace() << "State<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::interpolate(const Locations_ &locs, GeoVaLs_ &gvals) const {
  Log::trace() << "State<MODEL>::interpolate starting" << std::endl;
  util::Timer timer(classname(), "interpolate");
  state_->interpolate(locs.locations(), gvals.geovals());
  Log::trace() << "State<MODEL>::interpolate done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::forceWith(const State &xx, const Variables_ &vars) {
  Log::trace() << "State<MODEL>::forceWith starting" << std::endl;
  util::Timer timer(classname(), "forceWith");
  state_->forceWith(*xx.state_, vars.variables());
  Log::trace() << "State<MODEL>::forceWith done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::read(const eckit::Configuration &conf) {
  Log::trace() << "State<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  state_->read(conf);
  Log::trace() << "State<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::write(const eckit::Configuration &conf) const {
  Log::trace() << "State<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  util::LogbookWriter<bool> log("StateWriting", true);
  state_->write(conf);
  Log::trace() << "State<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double State<MODEL>::norm() const {
  Log::trace() << "State<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = state_->norm();
  Log::trace() << "State<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
Geometry<MODEL> State<MODEL>::geometry() const {
  Log::trace() << "State<MODEL>::geometry starting" << std::endl;
  util::Timer timer(classname(), "geometry");
  Geometry<MODEL> geom(state_->geometry());
  Log::trace() << "State<MODEL>::geometry done" << std::endl;
  return geom;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::print(std::ostream &os) const {
  Log::trace() << "State<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *state_;
  Log::trace() << "State<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::zero() {
  Log::trace() << "State<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  state_->zero();
  Log::trace() << "State<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void State<MODEL>::accumul(const double &zz, const State &xx) {
  Log::trace() << "State<MODEL>::accumul starting" << std::endl;
  util::Timer timer(classname(), "accumul");
  state_->accumul(zz, *xx.state_);
  Log::trace() << "State<MODEL>::accumul done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_STATE_H_
