/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_RUNS_DUALCONFIGFORECAST_H_
#define OOPS_RUNS_DUALCONFIGFORECAST_H_

#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateWriter.h"
#include "oops/generic/DualConfigGeometry.h"
#include "oops/generic/DualConfigModel.h"
#include "oops/generic/DualConfigModelAuxControl.h"
#include "oops/generic/DualConfigPostProcessor.h"
#include "oops/generic/DualConfigState.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/State.h"
#include "oops/runs/Application.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

template <typename MODEL>
class DualConfigForecast : public Application {
  using DualConfigGeometry_ = DualConfigGeometry<MODEL>;
  using DualConfigModel_ = DualConfigModel<MODEL>;
  using DualConfigModelAux_ = DualConfigModelAuxControl<MODEL>;
  using DualConfigState_ = DualConfigState<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using ModelAux_ = ModelAuxControl<MODEL>;
  using State_ = State<MODEL>;

 public:
  // -----------------------------------------------------------------------------
  DualConfigForecast() {}
  // -----------------------------------------------------------------------------
  virtual ~DualConfigForecast() {}
  // -----------------------------------------------------------------------------
  int execute(const eckit::Configuration& fullConfig) const {
    //  Setup resolution
    const eckit::LocalConfiguration resolConfig(fullConfig, "resolution");
    const DualConfigGeometry_ resol(resolConfig);

    //  Setup Model
    const eckit::LocalConfiguration modelConfig(fullConfig, "model");
    const DualConfigModel_ model(resol, modelConfig);

    //  Setup initial state
    const eckit::LocalConfiguration initialConfig(fullConfig, "initial");
    DualConfigState_ xx(resol, model, initialConfig);
    Log::test() << "Initial state: " << xx.norm() << std::endl;

    //  Setup augmented state
    const DualConfigModelAux_ moderr(resol, model, initialConfig);

    //  Setup times
    const util::Duration fclength(fullConfig.getString("forecast_length"));
    const util::DateTime bgndate(xx.validTime1());
    const util::DateTime enddate(bgndate + fclength);
    Log::info() << "Running dual-config forecast from " << bgndate << " to "
                << enddate << std::endl;

    //  Setup forecast outputs
    DualConfigPostProcessor<DualConfigState_, State_> post;

    eckit::LocalConfiguration prtConfig;
    eckit::LocalConfiguration prtConfig1;
    eckit::LocalConfiguration prtConfig2;
    if (fullConfig.has("prints")) {
      prtConfig = eckit::LocalConfiguration(fullConfig, "prints");
      prtConfig1 = eckit::LocalConfiguration(prtConfig, "prints1");
      prtConfig2 = eckit::LocalConfiguration(prtConfig, "prints2");
    }
    post.post1().enrollProcessor(new StateInfo<State_>("fc", prtConfig1));
    post.post2().enrollProcessor(new StateInfo<State_>("fc", prtConfig2));

    const eckit::LocalConfiguration outConfig(fullConfig, "output");
    const eckit::LocalConfiguration outConfig1(outConfig, "output1");
    const eckit::LocalConfiguration outConfig2(outConfig, "output2");
    post.post1().enrollProcessor(new StateWriter<State_>(bgndate, outConfig1));
    post.post2().enrollProcessor(new StateWriter<State_>(bgndate, outConfig2));

    //  Run forecast
    model.forecast(xx, moderr, fclength, post);

    Log::test() << "Final state: " << xx.norm() << std::endl;

    return 0;
  }
  // -----------------------------------------------------------------------------
 private:
  std::string appname() const {
    return "oops::DualConfigForecast<" + MODEL::name() + ">";
  }
  // -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_DUALCONFIGFORECAST_H_
