/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_RUNS_MAKEOBS_H_
#define OOPS_RUNS_MAKEOBS_H_

#include <memory>  // for std::shared_ptr, std::unique_ptr
#include <string>
#include <vector>

#include "eckit/config/LocalConfiguration.h"
#include "oops/base/Departures.h"
#include "oops/base/ObsAuxControls.h"
#include "oops/base/ObsErrors.h"
#include "oops/base/ObsOperators.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/Observations.h"
#include "oops/base/Observer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/StateInfo.h"
#include "oops/generic/instantiateObsErrorFactory.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/State.h"
#include "oops/runs/Application.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

template <typename MODEL>
class MakeObs : public Application {
  using Departures_ = Departures<MODEL>;
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using ModelAux_ = ModelAuxControl<MODEL>;
  using ObsAuxCtrls_ = ObsAuxControls<MODEL>;
  using Observations_ = Observations<MODEL>;
  using ObsSpaces_ = ObservationSpaces<MODEL>;
  using ObsOperators_ = ObsOperators<MODEL>;
  using State_ = State<MODEL>;

 public:
  // -----------------------------------------------------------------------------
  MakeObs() { instantiateObsErrorFactory<MODEL>(); }
  // -----------------------------------------------------------------------------
  virtual ~MakeObs() {}
  // -----------------------------------------------------------------------------
  int execute(const eckit::Configuration& fullConfig) const {
    //  Setup observation window
    const eckit::LocalConfiguration windowConf(fullConfig,
                                               "Assimilation Window");
    const util::DateTime bgn(windowConf.getString("Begin"));
    const util::DateTime end(windowConf.getString("End"));
    const util::Duration fclen(end - bgn);
    Log::info() << "Observation window is:" << windowConf << std::endl;

    //  Setup resolution
    const eckit::LocalConfiguration geomConfig(fullConfig, "Geometry");
    const Geometry_ geom(geomConfig);

    //  Setup Model
    const eckit::LocalConfiguration modelConfig(fullConfig, "Model");
    const Model_ model(geom, modelConfig);

    //  Setup initial "true" state
    const eckit::LocalConfiguration initialConfig(fullConfig,
                                                  "Initial Condition");
    Log::info() << "Initial configuration is:" << initialConfig << std::endl;
    State_ xx(geom, model, initialConfig);
    Log::test() << "Initial state: " << xx.norm() << std::endl;

    //  Setup augmented state
    ModelAux_ moderr(geom, model, initialConfig);

    //  Setup forecast outputs
    PostProcessor<State_> post;

    eckit::LocalConfiguration prtConf;
    fullConfig.get("prints", prtConf);
    post.enrollProcessor(new StateInfo<State_>("fc", prtConf));

    //  Setup observations
    const eckit::LocalConfiguration obsconf(fullConfig, "Observations");
    Log::info() << "Observation configuration is:" << obsconf << std::endl;
    ObsSpaces_ obspace(obsconf, geom, bgn, end);
    ObsOperators_ hop(obspace);

    //  Setup observation bias
    ObsAuxCtrls_ ybias(obspace, obsconf);

    std::shared_ptr<Observer<MODEL, State_> > pobs(
        new Observer<MODEL, State_>(obspace, hop, ybias));
    post.enrollProcessor(pobs);

    //  Run forecast and generate observations
    model.forecast(xx, moderr, fclen, post);
    Log::test() << "Final state: " << xx.norm() << std::endl;
    Log::info() << "MakeObs: Finished observation generation." << std::endl;

    std::unique_ptr<Observations_> yobs(pobs->release());
    Log::info() << "Generated observation: " << *yobs << std::endl;

    //  Perturb observations
    if (obsconf.has("obspert")) {
      Departures_ ypert(obspace);
      ObsErrors<MODEL> matR(obspace);
      matR.randomize(ypert);
      double opert = obsconf.getDouble("obspert");
      ypert *= opert;
      *yobs += ypert;
      Log::info() << "Perturbed observation: " << *yobs << std::endl;
    }

    //  Save observations
    for (std::size_t jj = 0; jj < yobs->size(); ++jj) {
      Log::test() << "Generated observation: " << (*yobs)[jj] << std::endl;
    }
    yobs->save(obsconf);

    return 0;
  }
  // -----------------------------------------------------------------------------
 private:
  std::string appname() const { return "oops::MakeObs<" + MODEL::name() + ">"; }
  // -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_MAKEOBS_H_
