/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#include "oops/runs/Run.h"

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/config/YAMLConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/log/OStreamTarget.h"
#include "oops/runs/Application.h"
#include "util/LibOOPS.h"
#include "util/LogbookHelper.h"
#include "util/Logger.h"
#include "util/Mpi.h"
#include "util/ObjectCountHelper.h"
#include "util/RestartHelper.h"
#include "util/TimerHelper.h"

namespace oops {

// -----------------------------------------------------------------------------

Run::Run(int argc, char** argv)
    : eckit::Main(argc, argv, "OOPS_HOME"), config_(), timer_() {
  // Get configuration file from command line
  ASSERT(argc >= 2);
  eckit::PathName configfile = argv[argc - 1];

  // Read configuration
  config_.reset(new eckit::YAMLConfiguration(configfile));

  Log::info() << "Configuration input file is: " << configfile << std::endl;
  Log::info() << "Full configuration is:" << *config_ << std::endl;

  // Start recording logbook
  util::LogbookHelper::start();

  // Start restart mechanism
  util::RestartHelper::start();

  // Start measuring performance
  util::ObjectCountHelper::start();
  util::TimerHelper::start();
}

// -----------------------------------------------------------------------------

Run::~Run() {
  // Stop recording logbook
  util::LogbookHelper::stop();

  // Stop restart mechanism
  util::RestartHelper::stop();

  // Performance diagnostics
  util::ObjectCountHelper::stop();
  util::TimerHelper::stop();

  LibOOPS::instance().finalise();
}

// -----------------------------------------------------------------------------

void Run::execute(const Application& app) {
  int status = 1;
  Log::info() << "Run: Starting " << app << std::endl;
  try {
    Log::info() << "Full configuration is:" << *config_ << std::endl;
    status = app.execute(*config_);
  } catch (const eckit::Exception& e) {
    status = 1;
    Log::error() << e.what() << " caught in " << Here() << std::endl;
    Log::error() << "Exception: " << app << " terminating..." << std::endl;
    eckit::Exception::exceptionStack(eckit::Log::error(), true);
  } catch (const std::exception& e) {
    status = 1;
    Log::error() << "Exception: " << e.what() << std::endl;
    Log::error() << "Exception: " << app << " terminating..." << std::endl;
  } catch (...) {
    status = 1;
    Log::error() << "Unknown exception: " << app << " terminating..."
                 << std::endl;
  }
  Log::info() << "Run: Finishing " << app << " with status = " << status
              << std::endl;
  if (status) ::exit(status);
}

eckit::LogTarget* Run::createInfoLogTarget() const {
  const char* envlogfile = getenv("OOPS_LOGFILE");
  if (envlogfile) {
    static std::ofstream logfile;
    char oopslogfile[100];
    int rank = static_cast<int>(eckit::mpi::comm().rank());
    if (rank == 0) {
      sprintf(oopslogfile, "%s", envlogfile);
    } else {
      sprintf(oopslogfile, "/dev/null");
    }
    logfile.open(oopslogfile, std::ofstream::out | std::ofstream::app);
    return new eckit::OStreamTarget(logfile);
  } else {
    return new eckit::OStreamTarget(std::cout);
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops
