/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef OOPS_RUNS_VARIATIONAL_H_
#define OOPS_RUNS_VARIATIONAL_H_

#include <memory>  // for std::shared_ptr, std::unique_ptr
#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/IncrementalAssimilation.h"
#include "oops/assimilation/instantiateCostFactory.h"
#include "oops/assimilation/instantiateMinFactory.h"
#include "oops/base/Observations.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateWriter.h"
#include "oops/base/instantiateCovarFactory.h"
#include "oops/generic/instantiateObsErrorFactory.h"
#include "oops/generic/instantiateTlmFactory.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/runs/Application.h"
#ifdef ECSABER
#include "saber/oops/instantiateCovarFactory.h"
#include "saber/oops/instantiateMinFactory.h"
#endif
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

/*! @brief Variational data assimilation application

 * Variational is the top-level driver of variational data assimilation
 * algorithms: 3D-Var, 4D-Var, 4D-Ens-Var and Weak Constraint 4D-Var.
 */

template <typename MODEL>
class Variational : public Application {
  using Geometry_ = Geometry<MODEL>;
  using Model_ = Model<MODEL>;
  using State_ = State<MODEL>;

 public:
  /// Construct a variational application
  Variational() {
    /// 1. Instantiate various factories
    instantiateCostFactory<MODEL>();
    instantiateCovarFactory<MODEL>();
    instantiateMinFactory<MODEL>();
    instantiateObsErrorFactory<MODEL>();
    instantiateTlmFactory<MODEL>();
#ifdef ECSABER
    saber::instantiateCovarFactory<MODEL>();
    saber::instantiateMinFactory<MODEL>();
#endif
  }

  /// Destroy a variational application
  virtual ~Variational() {}

  /// Execute a variational application
  int execute(const eckit::Configuration& fullConfig) const {
    /// 1. Setup the outer-loop geometry
    const eckit::LocalConfiguration resolConfig(fullConfig, "resolution");
    const Geometry_ resol(resolConfig);

    /// 2. Setup the non-linear model
    const eckit::LocalConfiguration modelConfig(fullConfig, "model");
    const Model_ model(resol, modelConfig);
    Log::trace() << "Variational: model has been set up" << std::endl;

    /*! 3. Setup the cost function; the cost function type (3D-Var, 4D-Var, ..)
     *   is selected at run time; the specific cost function object is
     *   produced by the cost function factory;

     *   The background is constructed inside the cost function because its
     valid
     *   time within the assimilation window can be different (3D-Var vs.
     4D-Var),
     *   it can be 3D or 4D (strong vs weak constraint), etc...
     */

    const eckit::LocalConfiguration cfConf(fullConfig, "cost_function");
    std::unique_ptr<CostFunction<MODEL> > J(
        CostFactory<MODEL>::create(cfConf, resol, model));
    Log::trace() << "Variational: cost function has been set up" << std::endl;

    /// 4. Perform incremental variational assimilation
    IncrementalAssimilation<MODEL> incrementalAssimilation(*J);
    incrementalAssimilation.assimilate(fullConfig);
    Log::trace() << "Variational: incremantal assimilation done" << std::endl;

    /// 5. Skip final trajectory (if the late 4D start mechnism is active)
    if (incrementalAssimilation.skipFinalNonlinearTrajectory()) return 0;

    /// 6. Obtain the analysis
    std::shared_ptr<ControlVariable<MODEL> > xx(
        incrementalAssimilation.getAnalysis());

    /// 7. Save the analysis and compute final diagnostics (evaluate cost
    /// function)
    PostProcessor<State_> post;
    const util::DateTime winbgn(cfConf.getString("window_begin"));
    if (fullConfig.has("output")) {
      const eckit::LocalConfiguration outConfig(fullConfig, "output");
      post.enrollProcessor(new StateWriter<State_>(winbgn, outConfig));
    }

    const eckit::LocalConfiguration finalConfig(fullConfig, "final");
    if (finalConfig.has("prints")) {
      const eckit::LocalConfiguration prtConfig(finalConfig, "prints");
      post.enrollProcessor(new StateInfo<State_>("final", prtConfig));
    }

    J->evaluate(*xx, finalConfig, post);

    return 0;
  }
  // -----------------------------------------------------------------------------
 private:
  std::string appname() const {
    return "oops::Variational<" + MODEL::name() + ">";
  }
  // -----------------------------------------------------------------------------
};

}  // namespace oops
#endif  // OOPS_RUNS_VARIATIONAL_H_
