/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_INCREMENT_H_
#define TEST_INTERFACE_INCREMENT_H_

#include <cmath>
#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "oops/runs/Test.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"
#include "util/dot_product.h"

#include "test/TestEnvironment.h"

namespace test {

// =============================================================================

template <typename MODEL>
class IncrementFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using Model_ = oops::Model<MODEL>;
  using State_ = oops::State<MODEL>;
  using Variables_ = oops::Variables<MODEL>;

 public:
  static const Geometry_& resol() { return *getInstance().resol_; }
  static const Variables_& ctlvars() { return *getInstance().ctlvars_; }
  static const util::DateTime& time() { return *getInstance().time_; }

 private:
  static IncrementFixture<MODEL>& getInstance() {
    static IncrementFixture<MODEL> theIncrementFixture;
    return theIncrementFixture;
  }

  IncrementFixture<MODEL>() {
    //  Setup a geometry
    const eckit::LocalConfiguration resolConfig(TestEnvironment::config(),
                                                "Geometry");
    resol_.reset(new Geometry_(resolConfig));

    const eckit::LocalConfiguration varConfig(TestEnvironment::config(),
                                              "Variables");
    ctlvars_.reset(new Variables_(varConfig));

    //  Setup reference model
    const eckit::LocalConfiguration modconf(TestEnvironment::config(), "Model");
    model_.reset(new Model_(*resol_, modconf));

    //  Setup reference state
    const eckit::LocalConfiguration fgconf(TestEnvironment::config(), "State");
    State_ xx(*resol_, *model_, fgconf);

    time_.reset(new util::DateTime(xx.validTime()));
  }

  ~IncrementFixture<MODEL>() {}

  std::unique_ptr<Geometry_> resol_;
  std::unique_ptr<Model_> model_;
  std::unique_ptr<Variables_> ctlvars_;
  std::unique_ptr<util::DateTime> time_;
};

// =============================================================================

template <typename MODEL>
void testIncrementConstructor() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx(Test_::resol(), Test_::ctlvars(), Test_::time());

  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testIncrementCopyConstructor() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx1(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx1.random();
  EXPECT(dx1.norm() > 0.0);

  Increment_ dx2(dx1);
  EXPECT(dx2.norm() > 0.0);

  // Check that the copy is equal to the original
  dx2 -= dx1;
  EXPECT(dx2.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testIncrementTriangle() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx1(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx1.random();
  Increment_ dx2(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx2.random();

  // test triangle inequality
  double dot1 = dx1.norm();
  EXPECT(dot1 > 0.0);

  double dot2 = dx2.norm();
  EXPECT(dot2 > 0.0);

  dx2 += dx1;
  double dot3 = dx2.norm();
  EXPECT(dot3 > 0.0);

  EXPECT(dot3 <= dot1 + dot2);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testIncrementOpPlusEq() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx1(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx1.random();
  Increment_ dx2(dx1);

  // test *= and +=
  dx2 += dx1;
  dx1 *= 2.0;

  dx2 -= dx1;
  EXPECT(dx2.norm() < 1e-8);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testIncrementDotProduct() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx1(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx1.random();
  Increment_ dx2(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx2.random();

  // test symmetry of dot product
  double zz1 = dot_product(dx1, dx2);
  double zz2 = dot_product(dx2, dx1);

  EXPECT(zz1 == zz2);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testIncrementZero() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx.random();
  EXPECT(dx.norm() > 0.0);

  // test zero
  dx->zero();
  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testIncrementAxpy() {
  using Test_ = IncrementFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;

  Increment_ dx1(Test_::resol(), Test_::ctlvars(), Test_::time());
  dx1.random();

  // test axpy
  Increment_ dx2(dx1);
  dx2.axpy(2.0, dx1);

  dx2 -= dx1;
  dx2 -= dx1;
  dx2 -= dx1;

  EXPECT(dx2.norm() < 1e-8);
}

// =============================================================================

template <typename MODEL>
class Increment : public oops::Test {
 public:
  Increment() {}
  virtual ~Increment() {}

 private:
  std::string testid() const {
    return "test::Increment<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(CASE("interface/Increment/testIncrementConstructor") {
      testIncrementConstructor<MODEL>();
    });
    ts.emplace_back(CASE("interface/Increment/testIncrementCopyConstructor") {
      testIncrementCopyConstructor<MODEL>();
    });
    ts.emplace_back(CASE("interface/Increment/testIncrementTriangle") {
      testIncrementTriangle<MODEL>();
    });
    ts.emplace_back(CASE("interface/Increment/testIncrementOpPlusEq") {
      testIncrementOpPlusEq<MODEL>();
    });
    ts.emplace_back(CASE("interface/Increment/testIncrementDotProduct") {
      testIncrementDotProduct<MODEL>();
    });
    ts.emplace_back(CASE("interface/Increment/testIncrementAxpy") {
      testIncrementAxpy<MODEL>();
    });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_INCREMENT_H_
