/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_LINEARMODEL_H_
#define TEST_INTERFACE_LINEARMODEL_H_

#include <algorithm>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <limits>
#include <memory>  // for std::shared_ptr, std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/TrajectorySaver.h"
#include "oops/base/instantiateCovarFactory.h"
#include "oops/generic/instantiateTlmFactory.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/LinearModel.h"
#include "oops/interface/Model.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/dot_product.h"

#include "test/TestEnvironment.h"

using oops::Log;

namespace test {

// =============================================================================

template <typename MODEL>
class LinearModelFixture : private eckit::NonCopyable {
  using Covariance_ = oops::ModelSpaceCovarianceBase<MODEL>;
  using Geometry_ = oops::Geometry<MODEL>;
  using Increment_ = oops::Increment<MODEL>;
  using LinearModel_ = oops::LinearModel<MODEL>;
  using Model_ = oops::Model<MODEL>;
  using ModelAux_ = oops::ModelAuxControl<MODEL>;
  using ModelAuxIncr_ = oops::ModelAuxIncrement<MODEL>;
  using State_ = oops::State<MODEL>;
  using Variables_ = oops::Variables<MODEL>;

 public:
  static const eckit::Configuration& test() { return *getInstance().test_; }
  static const Geometry_& resol() { return *getInstance().resol_; }
  static const Variables_& ctlvars() { return *getInstance().ctlvars_; }
  static const util::DateTime& time() { return *getInstance().time_; }
  static const Covariance_& covariance() { return *getInstance().B_; }
  static const Model_& model() { return *getInstance().model_; }
  static const State_& xref() { return *getInstance().xref_; }
  static const ModelAux_& bias() { return *getInstance().bias_; }
  static const ModelAuxIncr_& dbias() { return *getInstance().dbias_; }
  static const LinearModel_& tlm() { return *getInstance().tlm_; }

 private:
  static LinearModelFixture<MODEL>& getInstance() {
    static LinearModelFixture<MODEL> theLinearModelFixture;
    return theLinearModelFixture;
  }

  LinearModelFixture<MODEL>() {
    test_.reset(new eckit::LocalConfiguration(TestEnvironment::config(),
                                              "LinearModelTest"));
    const util::Duration len(test_->getString("fclength"));

    const eckit::LocalConfiguration resolConfig(TestEnvironment::config(),
                                                "Geometry");
    resol_.reset(new Geometry_(resolConfig));

    const eckit::LocalConfiguration varConfig(TestEnvironment::config(),
                                              "Variables");
    ctlvars_.reset(new Variables_(varConfig));

    const eckit::LocalConfiguration nlConf(TestEnvironment::config(), "Model");
    model_.reset(new Model_(*resol_, nlConf));

    const eckit::LocalConfiguration iniConf(TestEnvironment::config(), "State");
    xref_.reset(new State_(*resol_, *model_, iniConf));
    time_.reset(new util::DateTime(xref_->validTime()));

    const eckit::LocalConfiguration biasConf(TestEnvironment::config(),
                                             "ModelBias");
    bias_.reset(new ModelAux_(*resol_, *model_, biasConf));
    dbias_.reset(new ModelAuxIncr_(*resol_, biasConf));

    //  Create a covariance matrix
    oops::instantiateCovarFactory<MODEL>();
    const eckit::LocalConfiguration covar(TestEnvironment::config(),
                                          "Covariance");
    Covariance_* Bptr = oops::CovarianceFactory<MODEL>::create(
        covar, *resol_, *ctlvars_, *xref_);
    Bptr->linearize(*xref_, *resol_, covar);
    B_.reset(Bptr);

    //  Linear model configuration
    tlConf_.reset(new eckit::LocalConfiguration(TestEnvironment::config(),
                                                "LinearModel"));

    //  Setup trajectory for TL and AD
    oops::instantiateTlmFactory<MODEL>();
    std::vector<std::unique_ptr<LinearModel_> > tlmvec;
    oops::PostProcessor<State_> post;
    std::shared_ptr<oops::TrajectorySaver<MODEL> > trajectorySaver(
        new oops::TrajectorySaver<MODEL>(*xref_, *bias_, *resol_, tlmvec,
                                         *tlConf_));
    post.enrollProcessor(trajectorySaver);
    State_ xx(*xref_);
    model_->forecast(xx, *bias_, len, post);
    trajectorySaver->recordTrajectory();
    LinearModel_* tlmfirst = tlmvec.begin()->release();
    tlmvec.erase(tlmvec.begin());
    tlm_.reset(tlmfirst);
  }

  ~LinearModelFixture<MODEL>() {}

  std::unique_ptr<const eckit::LocalConfiguration> test_;
  std::unique_ptr<const eckit::LocalConfiguration> tlConf_;
  std::unique_ptr<const Geometry_> resol_;
  std::unique_ptr<const util::DateTime> time_;
  std::unique_ptr<const Variables_> ctlvars_;
  std::unique_ptr<const State_> xref_;
  std::unique_ptr<const Model_> model_;
  std::unique_ptr<const ModelAux_> bias_;
  std::unique_ptr<const ModelAuxIncr_> dbias_;
  std::unique_ptr<const Covariance_> B_;
  std::unique_ptr<const LinearModel_> tlm_;
};

// =============================================================================

template <typename MODEL>
void testLinearModelConstructor() {
  using Test_ = LinearModelFixture<MODEL>;

  const util::Duration zero(0);
  EXPECT(Test_::tlm().timeResolution() > zero);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testLinearModelZeroLength() {
  using Test_ = LinearModelFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;
  using ModelAuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  const util::DateTime vt(Test_::time());
  const util::Duration zero(0);

  Increment_ dxref(Test_::resol(), Test_::ctlvars(), vt);
  Test_::covariance().randomize(dxref);
  ModelAuxIncr_ daux(Test_::dbias());
  const double ininorm = dxref.norm();
  EXPECT(ininorm > 0.0);

  Increment_ dx(Test_::resol(), Test_::ctlvars(), vt);
  Increment_ dxm(Test_::resol(), Test_::ctlvars(), vt);
  dxm = dxref;
  Test_::tlm().forecastTL(dxm, daux, zero);
  dx = dxm;
  EXPECT(dx.validTime() == vt);
  EXPECT(dx.norm() == ininorm);

  dxm.zero();
  dxm = dxref;
  Test_::tlm().forecastAD(dxm, daux, zero);
  dx = dxm;
  EXPECT(dx.validTime() == vt);
  EXPECT(dx.norm() == ininorm);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testLinearModelZeroPert() {
  using Test_ = LinearModelFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;
  using ModelAuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  const util::Duration len(Test_::test().getString("fclength"));
  const util::DateTime t1(Test_::time());
  const util::DateTime t2(t1 + len);
  EXPECT(t2 > t1);

  Increment_ dx(Test_::resol(), Test_::ctlvars(), t1);
  ModelAuxIncr_ daux(Test_::dbias());

  dx.zero();
  daux.zero();
  EXPECT(dx.norm() == 0.0);
  Test_::tlm().forecastTL(dx, daux, len);
  EXPECT(dx.validTime() == t2);
  EXPECT(dx.norm() == 0.0);

  dx.zero();
  daux.zero();
  EXPECT(dx.norm() == 0.0);
  Test_::tlm().forecastAD(dx, daux, len);
  EXPECT(dx.validTime() == t1);
  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testLinearModelLinearity() {
  using Test_ = LinearModelFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;
  using ModelAuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  const util::Duration len(Test_::test().getString("fclength"));
  const util::DateTime t1(Test_::time());
  const util::DateTime t2(t1 + len);
  EXPECT(t2 > t1);
  const double zz = 3.1415;

  Increment_ dx1(Test_::resol(), Test_::ctlvars(), t1);
  Test_::covariance().randomize(dx1);
  ModelAuxIncr_ daux1(Test_::dbias());
  EXPECT(dx1.norm() > 0.0);

  Increment_ dx2(dx1);
  ModelAuxIncr_ daux2(daux1);

  Test_::tlm().forecastTL(dx1, daux1, len);
  EXPECT(dx1.validTime() == t2);
  dx1 *= zz;
  daux1 *= zz;

  dx2 *= zz;
  daux2 *= zz;
  Test_::tlm().forecastTL(dx2, daux2, len);
  EXPECT(dx2.validTime() == t2);

  const double tol = Test_::test().getDouble("toleranceAD");
  EXPECT(oops::is_close(dx1.norm(), dx2.norm(), tol));
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testLinearApproximation() {
  using Test_ = LinearModelFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;
  using State_ = oops::State<MODEL>;

  const util::Duration len(Test_::test().getString("fclength"));
  const util::DateTime t1(Test_::time());
  const util::DateTime t2(t1 + len);
  EXPECT(t2 > t1);

  Increment_ dx0(Test_::resol(), Test_::ctlvars(), t1);
  Test_::covariance().randomize(dx0);
  EXPECT(dx0.norm() > 0.0);

  Increment_ dx(dx0);
  Test_::tlm().forecastTL(dx, Test_::dbias(), len);
  const double dxnorm = dx.norm();

  oops::PostProcessor<State_> post;
  State_ xx0(Test_::xref());
  Test_::model().forecast(xx0, Test_::bias(), len, post);

  const unsigned int ntest = Test_::test().getInt("testiterTL");
  double zz = 1.0;
  std::vector<double> errors;
  for (unsigned int jtest = 0; jtest < ntest; ++jtest) {
    State_ xx(Test_::xref());
    Increment_ pert(dx0);
    pert *= zz;
    xx += pert;
    Test_::model().forecast(xx, Test_::bias(), len, post);

    Increment_ diff(Test_::resol(), Test_::ctlvars(), t2);
    diff.diff(xx, xx0);
    const double difnorm = diff.norm();
    const double err = zz * dxnorm / difnorm;
    Increment_ derr(dx);
    derr *= zz;
    derr -= diff;
    const double errnorm = derr.norm();
    errors.push_back(errnorm / difnorm);
    Log::test() << "TL error = " << std::setprecision(16) << err
                << ", relative error = " << errnorm / difnorm << std::endl;
    zz /= 10.0;
  }

  // Analyze results
  const double approx = *std::min_element(errors.begin(), errors.end());
  Log::test() << "Test TL min error = " << approx << std::endl;
  const double tol = Test_::test().getDouble("toleranceTL");
  EXPECT(approx < tol);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testLinearModelAdjoint() {
  using Test_ = LinearModelFixture<MODEL>;
  using Increment_ = oops::Increment<MODEL>;
  using ModelAuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  const util::Duration len(Test_::test().getString("fclength"));
  const util::DateTime t1(Test_::time());
  const util::DateTime t2(t1 + len);
  EXPECT(t2 > t1);

  Increment_ dx11(Test_::resol(), Test_::ctlvars(), t1);
  Test_::covariance().randomize(dx11);
  ModelAuxIncr_ daux1(Test_::dbias());
  EXPECT(dx11.norm() > 0.0);
  Increment_ dx12(dx11);
  Test_::tlm().forecastTL(dx12, daux1, len);
  EXPECT(dx12.norm() > 0.0);

  Increment_ dx22(Test_::resol(), Test_::ctlvars(), t2);
  Test_::covariance().randomize(dx22);
  ModelAuxIncr_ daux2(Test_::dbias());
  EXPECT(dx22.norm() > 0.0);
  Increment_ dx21(dx22);
  Test_::tlm().forecastAD(dx21, daux2, len);
  EXPECT(dx21.norm() > 0.0);

  EXPECT(dx11.norm() != dx22.norm());
  EXPECT(dx11.validTime() == t1);
  EXPECT(dx21.validTime() == t1);
  EXPECT(dx12.validTime() == t2);
  EXPECT(dx22.validTime() == t2);

  const double dot1 = dot_product(dx11, dx21);
  const double dot2 = dot_product(dx12, dx22);
  const double tol = Test_::test().getDouble("toleranceAD");
  EXPECT(oops::is_close(dot1, dot2, tol));
}

// =============================================================================

template <typename MODEL>
class LinearModel : public oops::Test {
 public:
  LinearModel() {}
  virtual ~LinearModel() {}

 private:
  std::string testid() const {
    return "test::LinearModel<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(
        CASE("interface/GeometryIterator/testLinearModelConstructor") {
          testLinearModelConstructor<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/GeometryIterator/testLinearModelZeroLength") {
          testLinearModelZeroLength<MODEL>();
        });
    ts.emplace_back(CASE("interface/GeometryIterator/testLinearModelZeroPert") {
      testLinearModelZeroPert<MODEL>();
    });
    ts.emplace_back(
        CASE("interface/GeometryIterator/testLinearModelLinearity") {
          testLinearModelLinearity<MODEL>();
        });
    ts.emplace_back(CASE("interface/GeometryIterator/testLinearApproximation") {
      testLinearApproximation<MODEL>();
    });
    ts.emplace_back(CASE("interface/GeometryIterator/testLinearModelAdjoint") {
      testLinearModelAdjoint<MODEL>();
    });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_LINEARMODEL_H_
