/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_LOCATIONS_H_
#define TEST_INTERFACE_LOCATIONS_H_

#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"

#include "oops/interface/Locations.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"

#include "test/TestEnvironment.h"

namespace test {

// -----------------------------------------------------------------------------
template <typename MODEL>
class LocationsFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using ObsSpace_ = oops::ObservationSpace<MODEL>;

 public:
  static const ObsSpace_& getObSpace() { return *getInstance().os_; }
  static const util::DateTime& gett1() { return *getInstance().t1_; }
  static const util::DateTime& gett2() { return *getInstance().t2_; }

 private:
  static LocationsFixture<MODEL>& getInstance() {
    static LocationsFixture<MODEL> theLocationsFixture;
    return theLocationsFixture;
  }

  LocationsFixture() {
    const util::DateTime bgn(
        TestEnvironment::config().getString("window_begin"));
    const util::DateTime end(TestEnvironment::config().getString("window_end"));

    const eckit::LocalConfiguration geomConfig(TestEnvironment::config(),
                                               "Geometry");
    geom_.reset(new Geometry_(geomConfig));

    std::vector<eckit::LocalConfiguration> obsConfs;
    TestEnvironment::config().get("Observations", obsConfs);
    EXPECT(obsConfs.size() > 0);
    const eckit::LocalConfiguration obsConf(obsConfs[0], "Observation");
    os_.reset(new ObsSpace_(obsConf, *geom_, bgn, end));

    t1_.reset(new util::DateTime("2010-01-01T12:00:00Z"));
    t2_.reset(new util::DateTime("2010-01-02T00:00:00Z"));
  }

  ~LocationsFixture() {}

  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ObsSpace_> os_;
  std::unique_ptr<util::DateTime> t1_;
  std::unique_ptr<util::DateTime> t2_;
};
// -----------------------------------------------------------------------------

template <typename MODEL>
void testConstructors() {
  using Locations_ = oops::Locations<MODEL>;

  std::unique_ptr<Locations_> locs(new Locations_(
      LocationsFixture<MODEL>::getObSpace(), LocationsFixture<MODEL>::gett1(),
      LocationsFixture<MODEL>::gett2()));

  EXPECT(locs.get());

  locs.reset();

  EXPECT(!locs.get());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
class Locations : public oops::Test {
 public:
  Locations() {}
  virtual ~Locations() {}

 private:
  std::string testid() const {
    return "test::Locations<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(CASE("interface/Locations/testConstructors") {
      testConstructors<MODEL>();
    });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_LOCATIONS_H_
