/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_MODELAUXINCREMENT_H_
#define TEST_INTERFACE_MODELAUXINCREMENT_H_

#include <cmath>
#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxCovariance.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"
#include "util/dot_product.h"

#include "test/TestEnvironment.h"

namespace test {

// =============================================================================

template <typename MODEL>
class ModelAuxIncrementFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using Covariance_ = oops::ModelAuxCovariance<MODEL>;
  using ModelAux_ = oops::ModelAuxControl<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

 public:
  static const eckit::Configuration& config() { return *getInstance().conf_; }
  static const Covariance_& covariance() { return *getInstance().covar_; }
  static const Geometry_& resol() { return *getInstance().resol_; }

 private:
  static ModelAuxIncrementFixture<MODEL>& getInstance() {
    static ModelAuxIncrementFixture<MODEL> theModelAuxIncrementFixture;
    return theModelAuxIncrementFixture;
  }

  ModelAuxIncrementFixture<MODEL>() {
    //  Setup a geometry
    const eckit::LocalConfiguration resolConfig(TestEnvironment::config(),
                                                "Geometry");
    resol_.reset(new Geometry_(resolConfig));

    //  Setup a covariance matrix
    conf_.reset(new eckit::LocalConfiguration(TestEnvironment::config(),
                                              "ModelBiasCovariance"));
    covar_.reset(new Covariance_(*conf_, *resol_));
  }

  ~ModelAuxIncrementFixture<MODEL>() {}

  std::unique_ptr<const eckit::LocalConfiguration> conf_;
  std::unique_ptr<const Geometry_> resol_;
  std::unique_ptr<const Covariance_> covar_;
};

// =============================================================================

template <typename MODEL>
void testModelAuxIncrementConstructor() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx(Test_::resol(), Test_::config());

  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementCopyConstructor() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx1);

  AuxIncr_ dx2(dx1);
  EXPECT(dx2.norm() > 0.0);
  EXPECT(dx2.norm() == dx1.norm());

  // Check that the copy is equal to the original
  dx2 -= dx1;
  EXPECT(dx2.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementChangeRes() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx1);

  AuxIncr_ dx2(dx1, Test_::config());
  EXPECT(dx2.norm() > 0.0);
  EXPECT(dx2.norm() == dx1.norm());

  // Check that the copy is equal to the original
  dx2 -= dx1;
  EXPECT(dx2.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementTriangle() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx1);
  AuxIncr_ dx2(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx2);

  // test triangle inequality
  double dot1 = dx1.norm();
  EXPECT(dot1 > 0.0);

  double dot2 = dx2.norm();
  EXPECT(dot2 > 0.0);

  dx2 += dx1;
  double dot3 = dx2.norm();
  EXPECT(dot3 > 0.0);

  EXPECT(dot3 <= dot1 + dot2);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementOpPlusEq() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx1);
  AuxIncr_ dx2(dx1);

  // test *= and +=
  dx2 += dx1;
  dx1 *= 2.0;

  dx2 -= dx1;
  EXPECT(dx2.norm() < 1e-8);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementDotProduct() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx1);
  AuxIncr_ dx2(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx2);

  // test symmetry of dot product
  double zz1 = dot_product(dx1, dx2);
  double zz2 = dot_product(dx2, dx1);

  EXPECT(zz1 == zz2);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementZero() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx);
  EXPECT(dx.norm() > 0.0);

  // test zero
  dx->zero();
  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testModelAuxIncrementAxpy() {
  using Test_ = ModelAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ModelAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::resol(), Test_::config());
  ModelAuxIncrementFixture<MODEL>::covariance().randomize(dx1);

  // test axpy
  AuxIncr_ dx2(dx1);
  dx2.axpy(2.0, dx1);

  dx2 -= dx1;
  dx2 -= dx1;
  dx2 -= dx1;

  EXPECT(dx2.norm() < 1e-8);
}

// =============================================================================

template <typename MODEL>
class ModelAuxIncrement : public oops::Test {
 public:
  ModelAuxIncrement() {}
  virtual ~ModelAuxIncrement() {}

 private:
  std::string testid() const {
    return "test::ModelAuxIncrement<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();
    ts.emplace_back(
        CASE("interface/ModelAuxIncrement/testModelAuxIncrementConstructor") {
          testModelAuxIncrementConstructor<MODEL>();
        });
    ts.emplace_back(CASE(
        "interface/ModelAuxIncrement/testModelAuxIncrementCopyConstructor") {
      testModelAuxIncrementCopyConstructor<MODEL>();
    });
    ts.emplace_back(
        CASE("interface/ModelAuxIncrement/testModelAuxIncrementChangeRes") {
          testModelAuxIncrementChangeRes<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ModelAuxIncrement/testModelAuxIncrementTriangle") {
          testModelAuxIncrementTriangle<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ModelAuxIncrement/testModelAuxIncrementOpPlusEq") {
          testModelAuxIncrementOpPlusEq<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ModelAuxIncrement/testModelAuxIncrementDotProduct") {
          testModelAuxIncrementDotProduct<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ModelAuxIncrement/testModelAuxIncrementAxpy") {
          testModelAuxIncrementAxpy<MODEL>();
        });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_MODELAUXINCREMENT_H_
