/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_OBSAUXCOVARIANCE_H_
#define TEST_INTERFACE_OBSAUXCOVARIANCE_H_

#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/Configuration.h"
#include "eckit/memory/NonCopyable.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/ObsAuxCovariance.h"
#include "oops/interface/ObservationSpace.h"

#include "oops/runs/Test.h"

#include "test/TestEnvironment.h"

namespace test {

// -----------------------------------------------------------------------------
template <typename MODEL>
class ObsAuxCovarianceFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using ObsSpace_ = oops::ObservationSpace<MODEL>;

 public:
  static const eckit::Configuration& config() { return *getInstance().conf_; }
  static const ObsSpace_& obspace() { return *getInstance().obspace_; }

 private:
  static ObsAuxCovarianceFixture<MODEL>& getInstance() {
    static ObsAuxCovarianceFixture<MODEL> theObsAuxCovarianceFixture;
    return theObsAuxCovarianceFixture;
  }

  ObsAuxCovarianceFixture() {
    conf_.reset(new eckit::LocalConfiguration(TestEnvironment::config()));
    const eckit::LocalConfiguration geomConfig(TestEnvironment::config(),
                                               "Geometry");
    geom_.reset(new Geometry_(geomConfig));

    const util::DateTime tbgn(
        TestEnvironment::config().getString("window_begin"));
    const util::DateTime tend(
        TestEnvironment::config().getString("window_end"));
    std::vector<eckit::LocalConfiguration> obsConfs;
    TestEnvironment::config().get("Observations", obsConfs);
    EXPECT(obsConfs.size() > 0);
    const eckit::LocalConfiguration obsConf(obsConfs[0], "Observation");
    obspace_.reset(new ObsSpace_(obsConf, *geom_, tbgn, tend));
  }

  ~ObsAuxCovarianceFixture() {}

  std::unique_ptr<const eckit::LocalConfiguration> conf_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ObsSpace_> obspace_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
void testConstructor() {
  using Test_ = ObsAuxCovarianceFixture<MODEL>;
  using Covariance_ = oops::ObsAuxCovariance<MODEL>;

  std::unique_ptr<Covariance_> bias(
      new Covariance_(Test_::obspace(), Test_::config()));
  EXPECT(bias.get());

  bias.reset();
  EXPECT(!bias.get());
}

// -----------------------------------------------------------------------------
//  void linearize(const ObsAuxControl_ &, const Geometry_ &);
//  void multiply(const ObsAuxIncrement_ &, ObsAuxIncrement_ &) const;
//  void invMult(const ObsAuxIncrement_ &, ObsAuxIncrement_ &) const;
//  void randomize(ObsAuxIncrement_ &) const;
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCovariance : public oops::Test {
 public:
  ObsAuxCovariance() {}
  virtual ~ObsAuxCovariance() {}

 private:
  std::string testid() const {
    return "test::ObsAuxCovariance<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(CASE("interface/ObsAuxCovariance/testConstructor") {
      testConstructor<MODEL>();
    });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_OBSAUXCOVARIANCE_H_
