/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_OBSAUXINCREMENT_H_
#define TEST_INTERFACE_OBSAUXINCREMENT_H_

#include <cmath>
#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsAuxCovariance.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"
#include "util/dot_product.h"

#include "test/TestEnvironment.h"

namespace test {

// =============================================================================

template <typename MODEL>
class ObsAuxIncrementFixture : private eckit::NonCopyable {
  using Covariance_ = oops::ObsAuxCovariance<MODEL>;
  using Geometry_ = oops::Geometry<MODEL>;
  using ObsAux_ = oops::ObsAuxControl<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;
  using ObsSpace_ = oops::ObservationSpace<MODEL>;

 public:
  static const eckit::Configuration& config() { return *getInstance().conf_; }
  static const Covariance_& covariance() { return *getInstance().covar_; }
  static const ObsSpace_& obspace() { return *getInstance().obspace_; }

 private:
  static ObsAuxIncrementFixture<MODEL>& getInstance() {
    static ObsAuxIncrementFixture<MODEL> theObsAuxIncrementFixture;
    return theObsAuxIncrementFixture;
  }

  ObsAuxIncrementFixture<MODEL>() {
    conf_.reset(new eckit::LocalConfiguration(TestEnvironment::config()));
    const eckit::LocalConfiguration geomConfig(TestEnvironment::config(),
                                               "Geometry");
    geom_.reset(new Geometry_(geomConfig));

    const util::DateTime tbgn(
        TestEnvironment::config().getString("window_begin"));
    const util::DateTime tend(
        TestEnvironment::config().getString("window_end"));
    std::vector<eckit::LocalConfiguration> obsConfs;
    TestEnvironment::config().get("Observations", obsConfs);
    EXPECT(obsConfs.size() > 0);
    const eckit::LocalConfiguration obsConf(obsConfs[0], "Observation");
    obspace_.reset(new ObsSpace_(obsConf, *geom_, tbgn, tend));
    covar_.reset(new Covariance_(*obspace_, *conf_));
  }

  ~ObsAuxIncrementFixture<MODEL>() {}

  std::unique_ptr<const eckit::LocalConfiguration> conf_;
  std::unique_ptr<const Covariance_> covar_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ObsSpace_> obspace_;
};

// =============================================================================

template <typename MODEL>
void testObsAuxIncrementConstructor() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx(Test_::obspace(), Test_::config());

  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementCopyConstructor() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx1);

  AuxIncr_ dx2(dx1);
  EXPECT(dx2.norm() > 0.0);
  EXPECT(dx2.norm() == dx1.norm());

  // Check that the copy is equal to the original
  dx2 -= dx1;
  EXPECT(dx2.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementChangeRes() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx1);

  AuxIncr_ dx2(dx1, Test_::config());
  EXPECT(dx2.norm() > 0.0);
  EXPECT(dx2.norm() == dx1.norm());

  // Check that the copy is equal to the original
  dx2 -= dx1;
  EXPECT(dx2.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementTriangle() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx1);
  AuxIncr_ dx2(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx2);

  // test triangle inequality
  double dot1 = dx1.norm();
  EXPECT(dot1 > 0.0);

  double dot2 = dx2.norm();
  EXPECT(dot2 > 0.0);

  dx2 += dx1;
  double dot3 = dx2.norm();
  EXPECT(dot3 > 0.0);

  EXPECT(dot3 <= dot1 + dot2);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementOpPlusEq() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx1);
  AuxIncr_ dx2(dx1);

  // test *= and +=
  dx2 += dx1;
  dx1 *= 2.0;

  dx2 -= dx1;
  EXPECT(dx2.norm() < 1e-8);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementDotProduct() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx1);
  AuxIncr_ dx2(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx2);

  // test symmetry of dot product
  double zz1 = dot_product(dx1, dx2);
  double zz2 = dot_product(dx2, dx1);

  EXPECT(zz1 == zz2);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementZero() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx);
  EXPECT(dx.norm() > 0.0);

  // test zero
  dx->zero();
  EXPECT(dx.norm() == 0.0);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testObsAuxIncrementAxpy() {
  using Test_ = ObsAuxIncrementFixture<MODEL>;
  using AuxIncr_ = oops::ObsAuxIncrement<MODEL>;

  AuxIncr_ dx1(Test_::obspace(), Test_::config());
  ObsAuxIncrementFixture<MODEL>::covariance().randomize(dx1);

  // test axpy
  AuxIncr_ dx2(dx1);
  dx2.axpy(2.0, dx1);

  dx2 -= dx1;
  dx2 -= dx1;
  dx2 -= dx1;

  EXPECT(dx2.norm() < 1e-8);
}

// =============================================================================

template <typename MODEL>
class ObsAuxIncrement : public oops::Test {
 public:
  ObsAuxIncrement() {}
  virtual ~ObsAuxIncrement() {}

 private:
  std::string testid() const {
    return "test::ObsAuxIncrement<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(
        CASE("interface/ObsAuxIncrement/testObsAuxIncrementConstructor") {
          testObsAuxIncrementConstructor<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ObsAuxIncrement/testObsAuxIncrementCopyConstructor") {
          testObsAuxIncrementCopyConstructor<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ObsAuxIncrement/testObsAuxIncrementChangeRes") {
          testObsAuxIncrementChangeRes<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ObsAuxIncrement/testObsAuxIncrementTriangle") {
          testObsAuxIncrementTriangle<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ObsAuxIncrement/testObsAuxIncrementOpPlusEq") {
          testObsAuxIncrementOpPlusEq<MODEL>();
        });
    ts.emplace_back(
        CASE("interface/ObsAuxIncrement/testObsAuxIncrementDotProduct") {
          testObsAuxIncrementDotProduct<MODEL>();
        });
    ts.emplace_back(CASE("interface/ObsAuxIncrement/testObsAuxIncrementAxpy") {
      testObsAuxIncrementAxpy<MODEL>();
    });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_OBSAUXINCREMENT_H_
