/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_OBSVECTOR_H_
#define TEST_INTERFACE_OBSVECTOR_H_

#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/ObsVector.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"
#include "util/dot_product.h"

#include "test/TestEnvironment.h"

namespace test {

// -----------------------------------------------------------------------------
template <typename MODEL>
class ObsVectorFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using ObsSpace_ = oops::ObservationSpace<MODEL>;

 public:
  static const ObsSpace_& obspace() { return *getInstance().obspace_; }

 private:
  static ObsVectorFixture<MODEL>& getInstance() {
    static ObsVectorFixture<MODEL> theObsVectorFixture;
    return theObsVectorFixture;
  }

  ObsVectorFixture() {
    const util::DateTime tbgn(
        TestEnvironment::config().getString("window_begin"));
    const util::DateTime tend(
        TestEnvironment::config().getString("window_end"));

    const eckit::LocalConfiguration geomConfig(TestEnvironment::config(),
                                               "Geometry");
    geom_.reset(new Geometry_(geomConfig));

    std::vector<eckit::LocalConfiguration> obsConfs;
    TestEnvironment::config().get("Observations", obsConfs);
    EXPECT(obsConfs.size() > 0);
    const eckit::LocalConfiguration obsConf(obsConfs[0], "Observation");
    obspace_.reset(new ObsSpace_(obsConf, *geom_, tbgn, tend));
  }

  ~ObsVectorFixture() {}

  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ObsSpace_> obspace_;
};
// -----------------------------------------------------------------------------

template <typename MODEL>
void testConstructor() {
  using Test_ = ObsVectorFixture<MODEL>;
  using ObsVector_ = oops::ObsVector<MODEL>;

  std::unique_ptr<ObsVector_> ov(new ObsVector_(Test_::obspace()));
  EXPECT(ov.get());

  ov.reset();
  EXPECT(!ov.get());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testCopyConstructor() {
  using Test_ = ObsVectorFixture<MODEL>;
  using ObsVector_ = oops::ObsVector<MODEL>;

  std::unique_ptr<ObsVector_> ov(new ObsVector_(Test_::obspace()));

  std::unique_ptr<ObsVector_> other(new ObsVector_(*ov));
  EXPECT(other.get());

  other.reset();
  EXPECT(!other.get());

  EXPECT(ov.get());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsVector : public oops::Test {
 public:
  ObsVector() {}
  virtual ~ObsVector() {}

 private:
  std::string testid() const {
    return "test::ObsVector<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(CASE("interface/ObsVector/testConstructor") {
      testConstructor<MODEL>();
    });
    ts.emplace_back(CASE("interface/ObsVector/testCopyConstructor") {
      testCopyConstructor<MODEL>();
    });
  }
};

// =============================================================================

}  // namespace test

#endif  // TEST_INTERFACE_OBSVECTOR_H_
