/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_OBSERVATIONSPACE_H_
#define TEST_INTERFACE_OBSERVATIONSPACE_H_

#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"

#include "test/TestEnvironment.h"

namespace test {

// -----------------------------------------------------------------------------
template <typename MODEL>
class ObservationSpaceFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using ObsSpace_ = oops::ObservationSpace<MODEL>;

 public:
  static const eckit::Configuration& config() { return *getInstance().conf_; }
  static const util::DateTime& bgn() { return *getInstance().tbgn_; }
  static const util::DateTime& end() { return *getInstance().tend_; }
  static const ObsSpace_& obspace() { return *getInstance().obspace_; }

 private:
  static ObservationSpaceFixture<MODEL>& getInstance() {
    static ObservationSpaceFixture<MODEL> theObservationSpaceFixture;
    return theObservationSpaceFixture;
  }

  ObservationSpaceFixture() {
    tbgn_.reset(new util::DateTime(
        TestEnvironment::config().getString("window_begin")));
    tend_.reset(
        new util::DateTime(TestEnvironment::config().getString("window_end")));

    const eckit::LocalConfiguration geomConfig(TestEnvironment::config(),
                                               "Geometry");
    geom_.reset(new Geometry_(geomConfig));

    std::vector<eckit::LocalConfiguration> obsConfs;
    TestEnvironment::config().get("Observations", obsConfs);
    EXPECT(obsConfs.size() > 0);
    const eckit::LocalConfiguration obsConf(obsConfs[0], "Observation");
    obspace_.reset(new ObsSpace_(obsConf, *geom_, *tbgn_, *tend_));
  }

  ~ObservationSpaceFixture() {}

  std::unique_ptr<const util::DateTime> tbgn_;
  std::unique_ptr<const util::DateTime> tend_;
  std::unique_ptr<Geometry_> geom_;
  std::unique_ptr<ObsSpace_> obspace_;
};
// -----------------------------------------------------------------------------

template <typename MODEL>
void testConstructor() {
  using Test_ = ObservationSpaceFixture<MODEL>;

  EXPECT(Test_::obspace().windowStart() == Test_::bgn());
  EXPECT(Test_::obspace().windowEnd() == Test_::end());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObservationSpace : public oops::Test {
 public:
  ObservationSpace() {}
  virtual ~ObservationSpace() {}

 private:
  std::string testid() const {
    return "test::ObservationSpace<" + MODEL::name() + ">";
  }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(CASE("interface/ObservationSpace/testConstructor") {
      testConstructor<MODEL>();
    });
  }
};

// -----------------------------------------------------------------------------

}  // namespace test

#endif  // TEST_INTERFACE_OBSERVATIONSPACE_H_
