/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef TEST_INTERFACE_STATE_H_
#define TEST_INTERFACE_STATE_H_

#include <cmath>
#include <iostream>
#include <memory>  // for std::unique_ptr
#include <string>
#include <vector>

#define ECKIT_TESTING_SELF_REGISTER_CASES 0

#include "eckit/config/LocalConfiguration.h"
#include "eckit/memory/NonCopyable.h"
#include "eckit/testing/Test.h"

#include "oops/interface/Geometry.h"
#include "oops/interface/Model.h"
#include "oops/interface/State.h"
#include "oops/interface/Variables.h"
#include "oops/runs/Test.h"

#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/dot_product.h"

#include "test/TestEnvironment.h"

namespace test {

// -----------------------------------------------------------------------------

template <typename MODEL>
class StateFixture : private eckit::NonCopyable {
  using Geometry_ = oops::Geometry<MODEL>;
  using Model_ = oops::Model<MODEL>;
  using State_ = oops::State<MODEL>;
  using Variables_ = oops::Variables<MODEL>;

 public:
  static const eckit::Configuration& test() { return *getInstance().test_; }
  static const Geometry_& resol() { return *getInstance().resol_; }
  static const Variables_& vars() { return *getInstance().vars_; }
  static const State_& xref() { return *getInstance().xref_; }
  static const double& norm() { return getInstance().refnorm_; }

 private:
  static StateFixture<MODEL>& getInstance() {
    static StateFixture<MODEL> theStateFixture;
    return theStateFixture;
  }

  StateFixture<MODEL>() {
    test_.reset(
        new eckit::LocalConfiguration(TestEnvironment::config(), "StateTest"));

    const eckit::LocalConfiguration resolConfig(TestEnvironment::config(),
                                                "Geometry");
    resol_.reset(new Geometry_(resolConfig));

    const eckit::LocalConfiguration varConfig(TestEnvironment::config(),
                                              "Variables");
    vars_.reset(new Variables_(varConfig));

    const eckit::LocalConfiguration modConfig(TestEnvironment::config(),
                                              "Model");
    model_.reset(new Model_(*resol_, modConfig));

    const eckit::LocalConfiguration conf(TestEnvironment::config(), "State");
    xref_.reset(new State_(*resol_, *model_, conf));
    refnorm_ = xref_->norm();
  }

  ~StateFixture<MODEL>() {}

  std::unique_ptr<const eckit::LocalConfiguration> test_;
  std::unique_ptr<Geometry_> resol_;
  std::unique_ptr<Model_> model_;
  std::unique_ptr<Variables_> vars_;
  std::unique_ptr<State_> xref_;
  double refnorm_;
};

// =============================================================================

template <typename MODEL>
void testStateConstructor() {
  using Test_ = StateFixture<MODEL>;

  EXPECT(Test_::xref().norm() == Test_::norm());

  double norm = Test_::test().getDouble("norm");
  double tol = Test_::test().getDouble("tolerance");
  EXPECT(std::abs(Test_::xref().norm() - norm) < tol);

  const util::DateTime vt(Test_::test().getString("date"));
  EXPECT(Test_::xref().validTime() == vt);
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void testStateCopyConstructor() {
  using Test_ = StateFixture<MODEL>;
  using State_ = oops::State<MODEL>;

  std::unique_ptr<State_> xx(new State_(Test_::xref()));
  EXPECT(xx.get());
  EXPECT(xx->norm() == Test_::norm());

  const util::DateTime vt(Test_::test().getString("date"));
  EXPECT(xx->validTime() == vt);

  xx.reset();
  EXPECT(!xx.get());

  // Recomputing initial norm to make sure nothing bad happened
  EXPECT(Test_::xref().norm() == Test_::norm());
}

// -----------------------------------------------------------------------------

template <typename MODEL>
class State : public oops::Test {
 public:
  State() {}
  virtual ~State() {}

 private:
  std::string testid() const { return "test::State<" + MODEL::name() + ">"; }

  void register_tests() const {
    std::vector<eckit::testing::Test>& ts = eckit::testing::specification();

    ts.emplace_back(CASE("interface/State/testStateConstructor") {
      testStateConstructor<MODEL>();
    });
    ts.emplace_back(CASE("interface/State/testStateCopyConstructor") {
      testStateCopyConstructor<MODEL>();
    });
  }
};

// -----------------------------------------------------------------------------

}  // namespace test

#endif  // TEST_INTERFACE_STATE_H_
