/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_EFSO_H_
#define OOPS_TEST_SUITE_EFSO_H_

#include <memory>
#include <iostream>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(diagnostic,EFSO) {
  TEST_MESSAGE("Diagnostic EFSO");

  /*
   * Setup
   */
  const Config_ advar(getConfig(), "param");
  const Config_ variablesConfig(getConfig(), "variables");
  const Variables_ variables(variablesConfig);

  // Setup (4DEnVar) J
  const Config_ cfConfig(getConfig(), "cost_function");
  const util::Duration winlen(cfConfig.getString("window_length"));
  std::unique_ptr<CostFunction_> J(
      CostFactory_::create(cfConfig, getResolution(), getModel()));
  ControlVariable_ xx(J->jb().getBackground());
  J->linearize(xx, cfConfig);

  /*
   * Read wind fields
   */
  const Config_ windConfig(getConfig(), "wind_states");
  Log::info() << "Wind configuration is:\n" << windConfig;
  std::vector<std::unique_ptr<State_ > > wind;
  // TODO: read wind from files

  const util::DateTime time_t0(wind[0]->validTime());
  const util::DateTime time_t(wind[wind.size() - 1]->validTime());
  const util::DateTime winend(time_t0 + winlen);
  const util::Duration timeslot(wind[1]->validTime() - wind[0]->validTime());

  /*
   * Advectors Setup
   */
  Advectors_ adv, adv_ad;
  for (util::DateTime t0(time_t0); t0 <= winend; t0 += timeslot) {
    pAdvector_ ptr_adv(
        new Advector_(*wind[wind.size() - 1], Config_(getConfig(), "advector"), false));
    pAdvector_ ptr_adv_ad(
        new Advector_(*wind[wind.size() - 1], Config_(getConfig(), "advector"), true));

    adv[t0] = ptr_adv;
    adv_ad[t0] = ptr_adv_ad;

    adv[t0]->set_wind(wind, t0);
    *adv_ad[t0] = *adv[t0];
  }

  /*
   * Localization setup
   */
  Localization_ LocMatrix(getResolution(),
      Config_(getConfig(), "localization"));
  std::unique_ptr<GenericMatrix_> change_of_variable(
      MatrixFactory_::create(
          eckit::LocalConfiguration(getConfig(), "change_of_variable")));
  /*
   * Build ensemble
   */
  std::vector<Config_> ensembleConfig;
  getConfig().get("ensemble", ensembleConfig);
  for (unsigned idate = 0; idate < ensembleConfig.size(); ++idate) {
    const util::DateTime validTime(
        Config_(ensembleConfig[idate]).getString("reference_time"));
    pEnsemble_ ensemble(new Ensemble_(validTime, ensembleConfig[idate]));
    EnsemblesCollection_::getInstance().put(validTime, ensemble);
    ensemble->build(*wind[0], getResolution());
    ensemble->read();
  }
  /*
   * Compute perturbations
   */
  for (unsigned idate = 0; idate < ensembleConfig.size(); ++idate) {
    const util::DateTime t(
        Config_(ensembleConfig[idate]).getString("reference_time"));
    pEnsemble_ ens = EnsemblesCollection_::getInstance()[t];

    for (unsigned l = 0; l < ens->size(); ++l) {
      change_of_variable->multiply((*ens)[l]);
    }
    ens->to_perturbations();
  }

  /*
   * Read forecast errors, apply norm matrix
   */
  Increment_ fcst_err(getResolution(), variables, time_t);
  fcst_err.read(Config_(getConfig(), "background_forecast_error"));
  Increment_ an_fcst_err(getResolution(), variables, time_t);
  an_fcst_err.read(Config_(getConfig(), "analysis_forecast_error"));

  fcst_err += an_fcst_err;
  Increment_ c_fcst_err(fcst_err);
  std::unique_ptr<GenericMatrix_> CMatrix(
      MatrixFactory_::create(
          eckit::LocalConfiguration(getConfig(), "norm_matrix")));
  CMatrix->multiply(c_fcst_err);

  if (getConfig().has("norm_diag")) {
    fcst_err.schur_product_with(c_fcst_err);
    fcst_err.write(Config_(getConfig(), "norm_diag"));
    LocMatrix.multiply(fcst_err);
    fcst_err.write(Config_(getConfig(), "norm_diag_loc"));
  }

  /*
   * Sensitivity calculation
   */
  CtrlInc_ sensitivity(J->jb());
  sensitivity.zero();

  change_of_variable->transposeInverseMultiply(c_fcst_err);

  for (int itslot = sensitivity.state().first();
      itslot <= sensitivity.state().last(); itslot++) {

    Increment_ dx(sensitivity.state()[itslot]);
    const util::DateTime t0(sensitivity.state()[itslot].validTime());
    pEnsemble_ xa_t0 = EnsemblesCollection_::getInstance()[t0];
    pEnsemble_ xa_t = EnsemblesCollection_::getInstance()[time_t];

    for (unsigned int m = 0; m < xa_t->size(); ++m) {
      dx.zero();

      // dx = C(bg_fcst_err + an_fcst_err)
      dx.axpy(1.0, c_fcst_err, false);
      // dx = xa_t o C(bg_fcst_err + an_fcst_err)
      dx.schur_product_with((*xa_t)[m]);

      // dx = L ( xa_t o C(bg_fcst_err + an_fcst_err) )
      adv_ad[t0]->advect(dx, advar, dx);
      LocMatrix.multiply(dx);

      // dx = xa_t0 o ( L ( xa_t o C(bg_fcst_err + an_fcst_err) ) )
      dx.schur_product_with((*xa_t0)[m]);

      sensitivity.state()[itslot] += dx;
    }
    change_of_variable->inverseMultiply(sensitivity.state()[itslot]);
  }

  if (getConfig().has("sensitivity")) {
    sensitivity.write(Config_(getConfig(), "sensitivity"));
  }

  /*
   * Apply R-1 H
   */
  PostProcessorTL_ pp;
  pp.enrollProcessor(J->jterm(0).setupTL(sensitivity));

  J->runTL(sensitivity, pp);
  std::shared_ptr<oops::GeneralizedDepartures> dy(pp.releaseOutputFromTL(0));

  //J->jterm(0).multiplyCoInv(*dy);
  std::shared_ptr<Departures_> ddy = std::dynamic_pointer_cast < Departures_
      > (dy);
  ddy->save(Config_(getConfig(), "oman"));
}

}

#endif
