/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_B_H_
#define OOPS_TEST_SUITE_B_H_

#include <iostream>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(b,diag_b_dirac) {
  TEST_MESSAGE(
      "Apply B matrix ");
  Diag_ xx(state());
  Increment_ dxin(increment());
  Increment_ dxout(increment());

  const Config_ dirac(getConfig(), "dirac");
  dxin.read(dirac);

  B().multiply(dxin,dxout);
  xx += dxout;

  const Config_ diag(getConfig(), "diag");
  xx.write(diag);
}

// -----------------------------------------------------------------------------
TEST_CASE(b,test_b_randomize) {
  TEST_MESSAGE(
      "Test B randomize ");
  std::ostringstream oss;
  State_ xx(state());
  Increment_ dx(increment());

  B().randomize(dx);
  xx += dx;
  const eckit::LocalConfiguration member(getConfig(), "member");
  xx.write(member);

  // Report result
  double dxdx = dot_product(dx, dx);
  oss << std::setprecision(20) << "||dx|| = " << dxdx;
  TEST_MESSAGE(oss.str());
  checkResult(dxdx, "expected_norm");
}

// -----------------------------------------------------------------------------
TEST_CASE(b,test_b_4d) {
  TEST_MESSAGE(
      "Test B4D ");
  std::ostringstream oss;
  const util::Duration zero(0);

  const eckit::LocalConfiguration jbConf(getConfig(), "Jb4D");
  const eckit::LocalConfiguration bgConf(jbConf, "Background");

  State4D_ xx(bgConf, getResolution(), getModel());
  CostJb4D_ jb(jbConf, getResolution(), getLinearVariables(), zero, xx);
  jb.linearize(xx, getResolution(),jbConf);

  Increment4D_ dx(jb);
  Increment4D_ Bdx(dx);

  const eckit::LocalConfiguration diracConf(getConfig(), "dirac_4d");
  dx.read(diracConf);

  jb.Bmult(dx, Bdx);

  const eckit::LocalConfiguration diagConf(getConfig(), "diag_4d");
  Bdx.write(diagConf);
}

TEST_CASE(b,test_sqrtb_4d) {
  TEST_MESSAGE(
      "Test Sqrt B4D ");
  std::ostringstream oss;
  const util::Duration zero(0);

  const eckit::LocalConfiguration jbConf(getConfig(), "Jb4D");
  const eckit::LocalConfiguration bgConf(jbConf, "Background");

  State4D_ xx(bgConf, getResolution(), getModel());
  CostJb4D_ jb(jbConf, getResolution(), getLinearVariables(), zero, xx);
  jb.linearize(xx, getResolution(),jbConf);

  IncrCtlVec4D_ dv(jb);
  Increment4D_ dx(jb);
  Increment4D_ Bdx(dx);

  const eckit::LocalConfiguration diracConf(getConfig(), "dirac_4d");
  dx.read(diracConf);

  jb.BmultSqrtTrans(dx, dv);
  jb.BmultSqrt(dv, Bdx);

  const eckit::LocalConfiguration diagConf(getConfig(), "diag_4d");
  Bdx.write(diagConf);
}

// -----------------------------------------------------------------------------
TEST_CASE(b,test_adjoint) {
  std::ostringstream oss;
  TEST_MESSAGE(
      "Test ensemble covariance adjoint ");

  std::unique_ptr<Increment_> dx1(new Increment_(increment()));
  B().randomize(*dx1);

  std::unique_ptr<IncrCtlVec_> dv2(jb().jbState().newIncrCtlVec(0));
  dv2->random();

  IncrCtlVec_ dv1(*dv2);
  dv1.random();
  jb().jbState().covar().multiplySqrtTrans(*dx1,dv1);

  Increment_ dx2(*dx1);
  B().randomize(dx2);
  jb().jbState().covar().multiplySqrt(*dv2,dx2);

  // Check result
  double dx1dx2 = dot_product(*dx1, dx2);
  double dv1dv2 = dot_product(dv1, *dv2);
  double s = std::abs((dx1dx2 - dv1dv2) / (dx1dx2 + dv1dv2) * 2.0);
  double digits = - log10(s);
  oss << std::setprecision(20) << "dx1.dx2 = " << dx1dx2 << " dy1.dy2 = " << dv1dv2 << " digits = " << digits;
  TEST_MESSAGE(oss.str());

  EXPECT( digits >= 10.0 );
}

// -----------------------------------------------------------------------------
TEST_CASE(b,loc) {
  TEST_MESSAGE(
      "Apply Localization ");
  State_ xx(state());
  Increment_ dx(increment());

  const Config_ dirac(getConfig(), "dirac");
  dx.read(dirac);

  const Config_ conf(getConfig(), "localization");
  Localization_ loc(getResolution(), conf);

  loc.multiply(dx);
  xx += dx;

  const Config_ diag(getConfig(), "diag");
  xx.write(diag);

  double locnorm = dot_product(dx,dx);
  std::ostringstream oss;
  oss << std::setprecision(20) << "localization norm = " << locnorm;
  TEST_MESSAGE(oss.str());

  checkResult(locnorm, "localization_norm");
}


// -----------------------------------------------------------------------------
TEST_CASE(b,test_loc_adjoint) {
  std::ostringstream oss;
  TEST_MESSAGE(
      "Test localization adjoint ");

  const Config_ locConf(getConfig(), "cost_function.Jb.Covariance.localization");
  Localization_ loc(getResolution(), locConf);

  std::unique_ptr<Increment_> dx1(new Increment_(increment()));
  B().randomize(*dx1);

  std::unique_ptr<IncrCtlVec_> dv2(jb().jbState().newIncrCtlVec(0));
  dv2->random();

  IncrCtlVec_ dv1(*dv2);
  loc.multiplySqrtTrans(*dx1, dv1.ensCtlVec(0));

  Increment_ dx2(*dx1);
  loc.multiplySqrt(dv2->ensCtlVec(0), dx2);

  // Check result
  double dx1dx2 = dot_product(*dx1, dx2);
  double dv1dv2 = dot_product(dv1, *dv2);
  double s = std::abs((dx1dx2 - dv1dv2) / (dx1dx2 + dv1dv2) * 2.0);
  double digits = - log10(s);
  oss << std::setprecision(20) << "<dx1,dx2> = " << dx1dx2 << " <dv1,dv2> = " << dv1dv2 << " digits = " << digits;
  TEST_MESSAGE(oss.str());

  EXPECT( digits >= 10.0 );
}

}

#endif
