/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_ENSEMBLE_H_
#define OOPS_TEST_SUITE_ENSEMBLE_H_

#include <iostream>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(ensemble,build) {
  TEST_MESSAGE(
      "Buid Ensemble ");
  Increment_ dx(increment());

  std::vector<Config_> ensConfig;
  getConfig().get("ensemble", ensConfig);
  std::vector<Config_> member;
  ensConfig[0].get("state", member);
  for (int imem=0; imem < member.size(); ++ imem) {
    State_ xx(state());
    B().randomize(dx);
    xx += dx;
    xx.write(member[imem]);
  }

}

// -----------------------------------------------------------------------------
TEST_CASE(ensemble,read) {
  TEST_MESSAGE(
      "Test Ensemble read  ");
  std::vector<std::unique_ptr<Increment_> > ensemble;
  Increment_ sum(increment());
  sum.zero();

  /*
   * Read ensemble
   */
  std::vector<Config_> ensConfig;
  getConfig().get("ensemble", ensConfig);
  std::vector<Config_> member;
  ensConfig[0].get("state", member);
  int rank = ensConfig[0].getInt("members");

  for (int imem=0; imem < rank; ++ imem) {
    ensemble.emplace_back(new Increment_(increment()));
    ensemble[imem]->read(member[imem]);
  }

  /*
   * Calculate perturbations
   */
  double invrank = -1.0 / ((double) rank);
  double r = 1.0 / sqrt(((double) rank - 1.0));

  for (int imem=0; imem < rank; ++ imem) {
    sum.axpy(invrank, *ensemble[imem], false);
  }

  for (int imem=0; imem < rank; ++ imem) {
    ensemble[imem]->axpy(1.0, sum, false);
    *ensemble[imem] *= r;
  }

  /*
   * Calculate variances
   */
  Increment_ dx(increment());
  Increment_ variances(increment());
  variances.zero();

  for (int imem=0; imem < rank; ++ imem) {
    dx = *ensemble[imem];
    dx.schur_product_with(*ensemble[imem]);
    variances += dx;
  }
  double var = dot_product(variances,variances);

  std::ostringstream oss;
  oss << std::setprecision(20) << "variances = " << var;
  TEST_MESSAGE(oss.str());

  checkResult(var, "expected_variances");
}


// -----------------------------------------------------------------------------
TEST_CASE(ensemble,diag_be_bs) {
  TEST_MESSAGE(
      "Diagnostic of Be and Bs  ");
  std::vector<std::unique_ptr<Increment_> > ensemble;
  Increment_ sum(increment());
  sum.zero();

  /*
   * Read ensemble
   */
  std::vector<Config_> ensConfig;
  getConfig().get("ensemble", ensConfig);
  std::vector<Config_> member;
  ensConfig[0].get("state", member);
  int rank = ensConfig[0].getInt("members");

  for (int imem=0; imem < rank; ++ imem) {
    ensemble.emplace_back(new Increment_(increment()));
    ensemble[imem]->read(member[imem]);
  }

  /*
   * Calculate perturbations
   */
  double invrank = -1.0 / ((double) rank);
  double r = 1.0 / sqrt(((double) rank - 1.0));

  for (int imem=0; imem < rank; ++ imem) {
    sum.axpy(invrank, *ensemble[imem], false);
  }

  for (int imem=0; imem < rank; ++ imem) {
    ensemble[imem]->axpy(1.0, sum, false);
    *ensemble[imem] *= r;
  }

  /*
   * Calculate variances
   */
  Increment_ dx(increment());
  Variances_* variances = new Variances_(state());

  for (int imem=0; imem < rank; ++ imem) {
    dx = *ensemble[imem];
    dx.schur_product_with(*ensemble[imem]);
    (*variances) += dx;
  }
  const Config_ diagbe(getConfig(), "variances_b_ens");
  variances->write(diagbe);
  delete variances;

  /*
   * Randomize Bs
   */
  variances = new Variances_(state());
  Increment_ mb(increment());
  for (int imem=0; imem < rank; ++ imem) {
    B().randomize(mb);
    mb *= r;
    dx = mb;
    dx.schur_product_with(mb);
    (*variances) += dx;
  }
  const Config_ diagbs(getConfig(), "variances_b_stat");
  variances->write(diagbs);
  delete variances;
}

}

#endif
