/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_INTERPOLATOR_H_
#define OOPS_TEST_SUITE_INTERPOLATOR_H_

#include <iostream>
#include <memory>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(interpol,advector_direct) {
  TEST_MESSAGE(
      "Advector direct test ");
  const eckit::LocalConfiguration advcfg(getConfig(), "advector");
  const eckit::LocalConfiguration advar(advcfg, "fields_to_be_interpolated");

  // Build the fields to be advected ( correlation of C matrix )
  Increment_ field(increment());
  const Config_ dirac(getConfig(), "dirac");
  field.read(dirac);

  const Config_ conf(getConfig(), "localization");
  Localization_ loc(getResolution(), conf);
  const eckit::LocalConfiguration windConf(getConfig(), "wind");
  std::vector<std::unique_ptr<State_ > > wind;
  // TODO: read wind from files

  const util::DateTime winend(end());

  loc.multiply(field);
  //B().randomize(field);

  std::vector<Config_> cfgAdv;
  getConfig().get("advection_time", cfgAdv);

  unsigned nb_adv = cfgAdv.size();
  for (unsigned ifield = 0; ifield < nb_adv; ifield++) {

    // Build an Advector object
    Advector_ adv(*wind[ifield+1], advcfg, false);
    adv.set_wind(wind);

    // Advect field
    Increment_ adv_field(field);
    adv.advect(adv_field, advar, adv_field);

    const Config_ output(cfgAdv[ifield], "output");
    State_ xx(state());
    xx += adv_field;
    xx.write(output);
  }

}

// -----------------------------------------------------------------------------
TEST_CASE(interpol,advector_adjoint) {
  TEST_MESSAGE(
      "Advector adjoint test ");

  const eckit::LocalConfiguration advcfg(getConfig(), "advector");
  const eckit::LocalConfiguration advar(advcfg, "fields_to_be_interpolated");

  const Config_ conf(getConfig(), "localization");
  Localization_ loc(getResolution(), conf);
  const eckit::LocalConfiguration windConf(getConfig(), "wind");
  std::vector<std::unique_ptr<State_ > > wind;
  // TODO: read wind from files

  const util::DateTime winend(end());

  // Build the fields to be advected ( correlation of C matrix )
  Increment_ field(increment());
  const Config_ dirac(getConfig(), "dirac");
  field.read(dirac);
  //loc.multiply(field);
  B().randomize(field);

  // Build Advector objects
  Advector_ direct(*wind[wind.size()-1], advcfg, false);
  direct.set_wind(wind);
  Advector_ adjoint(*wind[wind.size()-1], advcfg, true);
  adjoint = direct;

  Increment_ ww(field);
  Increment_ zz(ww);
  Increment_ www(zz);
  Increment_ zzz(ww);

  direct.advect(ww, advar, zzz);
  adjoint.advect(zz, advar, www);

  const Config_ output(getConfig(), "output");
  State_ xx(state());
  xx += www;
  xx.write(output);

  double p1 = dot_product(ww, www);
  double p2 = dot_product(zz, zzz);
  double s = std::abs((p1 - p2) / (p1 + p2) * 2.0);
  double digits = -log10(s);

  std::ostringstream oss;
  oss << "ADJOINT TEST (DIRECT): x.Ft(y)= "
  << std::setprecision(20) << ((double) p1) << " y.F(x)= "
  << std::setprecision(20) << ((double) p2) << " digits = " << digits;
  TEST_MESSAGE(oss.str());
  checkDigits (digits);
}

// -----------------------------------------------------------------------------
TEST_CASE(interpol,two_geos_test) {
  TEST_MESSAGE(
      "Interpol 2 geos adjoint test ");

  const eckit::LocalConfiguration confNewGeo(getConfig(), "ensemble_resol");
  const Geometry_ resol_ens(confNewGeo);

  const eckit::LocalConfiguration confInterpol(getConfig(), "interpol");
  const Variables_ var(confInterpol);
  State_ x0(state());
  Interpolator_ intdir(x0,resol_ens,confInterpol, false);
  Interpolator_ intadj(x0,resol_ens,confInterpol, true);

  // Build fields using randomisation
  Increment_ dx(increment());
  B().randomize(dx);

  const eckit::LocalConfiguration jbConf(getConfig(), "Jb");
  const eckit::LocalConfiguration confState(getConfig(), "state_ens");
  State_ x0_lr(resol_ens,getModel(),confState);
  Increment_ dx_lr(resol_ens,increment());
  std::unique_ptr<ModelSpaceCovarianceBase_> B2;
  B2.reset(CovarianceFactory_::create(jbConf, resol_ens, var, x0_lr));
  B2->linearize(x0_lr,resol_ens,jbConf);
  B2->randomize(dx_lr);

  // define increments for test
  Increment_ zz(dx);
  Increment_ zzz(zz);
  zzz.zero();
  Increment_ ww(dx_lr);
  Increment_ www(ww);
  www.zero();

  // Perform direct and adjoint
  intdir.interpolate(ww, var , zzz);
  intadj.interpolate(zz, var, www);

  // Diagnostics
  double p1 = dot_product(ww, www);
  double p2 = dot_product(zz, zzz);
  double s = std::abs((p1 - p2) / (p1 + p2) * 2.0);
  double digits = -log10(s);

  std::ostringstream oss;
  oss << "ADJOINT TEST (DIRECT): x.Ft(y)= "
  << std::setprecision(20) << ((double) p1) << " y.F(x)= "
  << std::setprecision(20) << ((double) p2) << " digits = " << digits;
  TEST_MESSAGE(oss.str());
  checkDigits (digits);
  }
}

#endif
