/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_MIX_H_
#define OOPS_TEST_SUITE_MIX_H_

#include <iostream>
#include <memory>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(mix,change_of_variable_adjoint) {
  TEST_MESSAGE(
      "Change of variable adjoint test ");

  std::unique_ptr<GenericMatrix_> change_of_variable(
      MatrixFactory_::create(Config_(getConfig(), "change_of_variable")));

  Increment_ xx(increment());
  B().randomize(xx);
  Increment_ yy(increment());
  B().randomize(yy);
  change_of_variable->multiply(yy);

  Increment_ Fx(xx);
  change_of_variable->inverseMultiply(Fx);
  Increment_ Fty(yy);
  change_of_variable->transposeInverseMultiply(Fty);

  double p1 = dot_product(xx, Fty);
  double p2 = dot_product(yy, Fx);
  double s = std::abs((p1 - p2) / (p1 + p2) * 2.0);
  double digits = -log10(s);

  xx -= Fx;
  double zz = dot_product(xx, xx);
  std::ostringstream oss1;
  oss1 << "ADJOINT TEST: ||x-F(x)|| = " << std::setprecision(20) << zz;
  TEST_MESSAGE(oss1.str());
  EXPECT( zz > 1.0 );

  std::ostringstream oss2;
  oss2 << "ADJOINT TEST: x.Ft(y)= "
  << std::setprecision(20) << ((double) p1) << " y.F(x)= "
  << std::setprecision(20) << ((double) p2) << " digits = " << digits;
  TEST_MESSAGE(oss2.str());
  checkDigits (digits);
  }

// -----------------------------------------------------------------------------
  TEST_CASE(mix,vertical_profile) {
    TEST_MESSAGE(
        "Vertical profile multiplier test ");

    std::unique_ptr<GenericMatrix_> forward(
        MatrixFactory_::create(
            Config_(getConfig(), "vertical_profile_forward")));
    std::unique_ptr<GenericMatrix_> backward(
        MatrixFactory_::create(
            Config_(getConfig(), "vertical_profile_backward")));

    Increment_ dx(increment());
    B().randomize(dx);
    Increment_ dy(dx);
    forward->multiply(dy);
    backward->multiply(dy);

    dx -= dy;
    double zz = dot_product(dx, dx);
    std::ostringstream oss1;
    oss1 << "Vertical profile multiplier: ||x-F(Finv((x))|| = " << std::setprecision(20) << zz;
    TEST_MESSAGE(oss1.str());
    EXPECT( zz < 0.000001 );
  }

// -----------------------------------------------------------------------------
  TEST_CASE(mix,tapering) {
    TEST_MESSAGE(
        "Tapering test");

    std::unique_ptr<GenericMatrix_> tapering(
        MatrixFactory_::create(Config_(getConfig(), "tapering")));
    std::unique_ptr<GenericMatrix_> tapering_sqrt(
        MatrixFactory_::create(Config_(getConfig(), "tapering_sqrt")));

    Increment_ dx(increment());
    B().randomize(dx);
    Increment_ dy(dx);
    tapering->multiply(dy);

    Increment_ dz(dx);
    tapering_sqrt->multiply(dz);
    tapering_sqrt->multiply(dz);

    dx -= dy;
    double zz = dot_product(dx, dx);
    std::ostringstream oss1;
    oss1 << "Tapering test: ||x-F(x)|| = " << std::setprecision(20) << zz;
    TEST_MESSAGE(oss1.str());
    EXPECT( zz > 1.0 );

    dy -= dz;
    zz = dot_product(dy, dy);
    std::ostringstream oss2;
    oss2 << "Tapering test: ||Fx-sqrtF(sqrtF(x))|| = " << std::setprecision(20) << zz;
    TEST_MESSAGE(oss2.str());
    EXPECT( zz < 0.000001 );

  }

  // -----------------------------------------------------------------------------
  TEST_CASE(mix,norm_matrix) {
    TEST_MESSAGE(
        "Norm Matrix test ");

    std::unique_ptr<GenericMatrix_> norm_matrix(
        MatrixFactory_::create(Config_(getConfig(), "norm_matrix")));

    State_ xx(state());
    State_ Fx(xx);
    norm_matrix->multiply(Fx);

    Increment_ dx(increment());
    dx.diff(Fx,xx);
    double zz = dot_product(dx, dx);
    std::ostringstream oss1;
    oss1 << "NORM MATRIX TEST: ||x-F(x)|| = " << std::setprecision(20) << zz;
    TEST_MESSAGE(oss1.str());
    EXPECT( zz > 1.0 );
  }

  // -----------------------------------------------------------------------------
  TEST_CASE(mix,fields_leak_mem) {
    TEST_MESSAGE(
        "Fields leak memory test ");

    int nalloc = getConfig().getInt("number_of_allocations");
    for (int iter=0; iter < nalloc; iter++) {
      State_ xx(state());
    }
  }

  // -----------------------------------------------------------------------------
  TEST_CASE(mix,fields_base_leak_mem) {
    TEST_MESSAGE(
        "Fields base leak memory test ");

    int nalloc = getConfig().getInt("number_of_allocations");
    for (int iter=0; iter < nalloc; iter++) {
      Increment_ dx(increment());
    }
  }

}

#endif
