/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_MODEL_TLAD_H_
#define OOPS_TEST_SUITE_MODEL_TLAD_H_

#include <iostream>
#include <memory>
#include <cmath>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(mix,test_adjoint) {
  std::ostringstream oss;
  TEST_MESSAGE(
      "Test model adjoint ");

  State_ xx(state());
  propagate(xx, true);

  // Call tl
  Increment_ dx1(increment());
  randomize (dx1);

  Increment_ Mdx1(dx1);
  propagate (Mdx1);

  // Call ad
  Increment_ dx2(increment());
  randomize (dx2);
  Increment_ Mtdx2(dx2);
  propagate_ad (Mtdx2);

  // Check result
  double dx1Mtdx2 = dot_product(dx1, Mtdx2);
  double Mdx1dx2 = dot_product(Mdx1, dx2);
  double s = (dx1Mtdx2 - Mdx1dx2) / (dx1Mtdx2 + Mdx1dx2) * 2.0;

  oss << std::setprecision(20) << "<dx1,Mtdx2> = " << dx1Mtdx2 << " <Mdx1,dx2> = " << Mdx1dx2 << " diff = " << (dx1Mtdx2-Mdx1dx2);
  TEST_MESSAGE(oss.str());

  EXPECT( std::abs(s) < 1.0e-10 );
}

// -----------------------------------------------------------------------------
TEST_CASE(mix,test_tl) {
  TEST_MESSAGE(
      "Test tangent linear model");

  State_ Mx(state());
  State_ Mx_plus_dx(state());
  Increment_ dx(increment());
  randomize (dx);

  Mx_plus_dx += dx;
  propagate(Mx,true);
  Mx.write(Config_(getConfig(), "Mx"));

  propagate (Mx_plus_dx);
  Mx_plus_dx.write(Config_(getConfig(), "Mx_plus_dx"));

  propagate (dx);
  Mx += dx;
  Mx.write(Config_(getConfig(), "Mx_plus_Mdx"));
}

// -----------------------------------------------------------------------------
TEST_CASE(mix,test_model_direct) {
  TEST_MESSAGE(
      "Test direct model");
  std::ostringstream oss;
  std::unique_ptr<Accumulator_> dx(
      new Accumulator_(getResolution(), getLinearVariables(), start()));
  State_ Mx(state());

  propagate (Mx);
  dx->accumul( 1.0, state());
  dx->accumul(-1.0, Mx);
  double dxdx = dot_product(*dx, *dx);

  oss << std::setprecision(20) << "||Mx-x|| = " << dxdx;
  oss << "||Mx|| = " << Mx;
  TEST_MESSAGE(oss.str());

  // Report result
  checkResult(dxdx, "expected_diff");
}

// -----------------------------------------------------------------------------
TEST_CASE(mix,test_external_dfi) {
  TEST_MESSAGE(
      "Test external dfi");
  std::ostringstream oss;
  std::unique_ptr<Accumulator_> dx(
      new Accumulator_(getResolution(), getLinearVariables(), start()));
  State_ Mx(state());
  dx->accumul( 1.0, Mx);

  //  Setup DFI
  PostProcessor_ pp;
  const eckit::LocalConfiguration dfiConf(getConfig(), "dfi");
  const util::Duration dfispan(dfiConf.getString("filter_span"));
  const util::DateTime dfitime(start() + dfispan / 2);
  std::shared_ptr<WeightedMean_> pdfi(
      new WeightedMean_(dfitime, dfispan, getResolution(), dfiConf));
  pp.enrollProcessor(pdfi);

  //  Run DFI forecast
  getModel().forecast(Mx, getModelErr(), dfispan, pp);
  std::unique_ptr<State_> xdfi(pdfi->releaseMean());

  // Report result
  dx->accumul(-1.0, *xdfi);
  double dxdx = dot_product(*dx, *dx);

  oss << std::setprecision(20) << "||Mx-x|| = " << dxdx;
  oss << "||Mx|| = " << Mx;
  TEST_MESSAGE(oss.str());
  checkResult(dxdx, "expected_diff");
}

// -----------------------------------------------------------------------------
TEST_CASE(mix,test_change_resolution_model_adjoint) {
  std::ostringstream oss;
  TEST_MESSAGE(
      "Test model adjoint ");

  // Get configuration
  const Config_ jbConf(getConfig(), "CostJb");
  const Config_ to_low_resol_config(getConfig(), "to_low_resolution");
  const Config_ low_resol_config(to_low_resol_config, "low_resolution");
  const Config_ modelConfig(getConfig(), "model");
  const Config_ tlmConf(getConfig(), "linear_model");

  // Build high resolution objects
  ControlVariable_ xb(Config_(jbConf, "Background"), getResolution(), getModel(), obspace());
  CostJb4D_ jb4d(jbConf, getResolution(), getLinearVariables(), util::Duration(0), xb.state());
  jb4d.linearize(xb.state(), getResolution(),jbConf);

  // Build low resolution objects
  PostProcessor_ post;
  PostProcessorIncr_ no_post;
  const Geometry_ low_resol(low_resol_config);
  Model_ model(low_resol, modelConfig);
  ModelAuxCtrl_ moderr(low_resol, model, getConfig());
  ModelAuxIncrement_ modaux(low_resol, getConfig());
  std::shared_ptr<GenericMatrix_> to_low_resol(MatrixFactory_::create(to_low_resol_config));

  // Build trajectory at low resolution
  State_ x0(xb.state()[0]);
  to_low_resol->multiply(x0);
  TLM().clear();
  post.enrollProcessor(
    new TrajectorySaver_(x0, moderr, low_resol, TLM(), tlmConf));
  model.forecast(x0, moderr, length(), post);

  // Call tl
  Increment4D_ dx1(jb4d);
  jb4d.randomize (dx1);
  Increment_ dx1_3dt0_lr(low_resol, getLinearVariables(), start());
  dx1_3dt0_lr.zero();
  dx1_3dt0_lr.axpy(1.0,dx1[0]);
  Increment4D_ Mdx1_4d_hr(jb4d);

  // Run dx1_3dt0_lr.propagate()
  PostProcessorTL_ post_tl;
  std::shared_ptr < Observer4DTL_
  			> obstl(new Observer4DTL_(Mdx1_4d_hr, to_low_resol));
  post_tl.enrollProcessor(obstl);
  Increment_ tmp(dx1_3dt0_lr);
  getTLM().forecastTL(tmp, modaux, length(), no_post, post_tl);

  // Call ad
  Increment4D_ dx2_4d_hr(jb4d);
  jb4d.randomize (dx2_4d_hr);
  Increment_ Mtdx2_3dt0_lr(low_resol, getLinearVariables(), start(), end());
  Mtdx2_3dt0_lr.zero();

  // Run dx1_3dt0_lr.propagate()
  PostProcessorAD_ post_ad;
  std::shared_ptr < Observer4DAD_
			> obsad(new Observer4DAD_(dx2_4d_hr, to_low_resol));
  post_ad.enrollProcessor(obsad);
  getTLM().forecastAD(Mtdx2_3dt0_lr, modaux, length(), no_post, post_ad);

  // Check result
  double dx1Mtdx2 = dot_product(dx1_3dt0_lr, Mtdx2_3dt0_lr);
  double Mdx1dx2 = dot_product(Mdx1_4d_hr, dx2_4d_hr);
  double s = (dx1Mtdx2 - Mdx1dx2) / (dx1Mtdx2 + Mdx1dx2) * 2.0;

  oss << std::setprecision(20) << "<dx1,Mtdx2> = " << dx1Mtdx2 << " <Mdx1,dx2> = " << Mdx1dx2 << " diff = " << (dx1Mtdx2-Mdx1dx2);
  TEST_MESSAGE(oss.str());

  EXPECT( std::abs(s) < 1.0e-9 );
}

// -----------------------------------------------------------------------------
TEST_CASE(mix,test_surfex) {
  TEST_MESSAGE(
      "Test SURFEX integrity");
  std::ostringstream oss;
  std::unique_ptr<Accumulator_> dx(
      new Accumulator_(getResolution(), getLinearVariables(), start()));

  //  Run forecast
  State_ x0(state());
  dx->zero();

  propagate (x0);
  dx->accumul( 1.0, x0);

  //  Run forecast
  State_ x1(state());
  propagate (x1);
  dx->accumul( -1.0, x1);

  // Report result
  double dxdx = dot_product(*dx, *dx);

  oss << std::setprecision(20) << "||Mx-Mx|| = " << dxdx;
  TEST_MESSAGE(oss.str());
  EXPECT( dxdx < 0.000001 );
}


}

#endif
