/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_OP_OBS_FILE_H_
#define OOPS_TEST_SUITE_OP_OBS_FILE_H_

#include <math.h>
#include <iostream>
#include <memory>

#include "util/Logger.h"
#include "util/Duration.h"
#include "util/dot_product.h"

namespace test {

// -----------------------------------------------------------------------------
TEST_CASE(op_obs_file,test_hop) {
  TEST_MESSAGE(
      "Test observation operator with trajectory files");

  // Setup Trajectory
  std::vector<eckit::LocalConfiguration> tslots;
  const eckit::LocalConfiguration trajConf(getConfig(), "trajectory");
  trajConf.get("state", tslots);
  int nslots = tslots.size();
  State4D_ traj(trajConf, getResolution(), getModel());

  util::Duration tstep(length());
  if ( nslots > 1 ) tstep /= (nslots-1);

  // Call H
  std::shared_ptr<Observer_> pp(
      new Observer_(obspace(), hop(), obias(), tstep));
  oops::PostProcessor<State_> post;
  post.enrollProcessor(pp);

  post.initialize(traj[0], end(), tstep);
  for (int jsub = 1; jsub < nslots; ++jsub) {
    post.process(traj[jsub]);
  }
  post.finalize(traj[0]);

  // Get departure and save
  std::shared_ptr<Observations_> yequ(pp->release());
  Departures_ ydep(yobs() - (*yequ));
  ydep.save(eckit::LocalConfiguration(getConfig(), "fg_depar"));
  // MASTERODB compliance
  ydep.save(eckit::LocalConfiguration(getConfig(), "hires"));
}

// -----------------------------------------------------------------------------
TEST_CASE(op_obs_file,test_hop_with_jo) {
  TEST_MESSAGE(
      "Test observation operator with trajectory files and calculate Jo");

  // Setup Trajectory
  std::vector<eckit::LocalConfiguration> tslots;
  const eckit::LocalConfiguration trajConf(getConfig(), "trajectory");
  trajConf.get("state", tslots);
  int nslots = tslots.size();
  State4D_ traj(trajConf, getResolution(), getModel());

  util::Duration tstep(length());
  if ( nslots > 1 ) tstep /= (nslots-1);

  // Call H
  std::shared_ptr<Observer_> pp(
      new Observer_(obspace(), hop(), obias(), tstep));
  oops::PostProcessor<State_> post;
  post.enrollProcessor(pp);

  post.initialize(traj[0], end(), tstep);
  for (int jsub = 1; jsub < nslots; ++jsub) {
    post.process(traj[jsub]);
  }
  post.finalize(traj[0]);

  // Get departure and save
  std::shared_ptr<Observations_> yequ(pp->release());
  Departures_ ydep(yobs() - (*yequ));
  ydep.save(eckit::LocalConfiguration(getConfig(), "fg_depar"));
  // MASTERODB compliance
  ydep.save(eckit::LocalConfiguration(getConfig(), "hires"));

  // Calculate and print Jo
  ObsErrors_ R(obspace());
  R.linearize(yobs());
  std::unique_ptr<Departures_> ygrad(R.inverseMultiply(ydep));
  const double zjo = 0.5 * dot_product(ydep, *ygrad);
  obspace().printJo(ydep, *ygrad);

  std::ostringstream oss;
  oss << std::setprecision(20) << "Jo = " << zjo;
  TEST_MESSAGE(oss.str());

  // Report result
  checkResult(zjo, "expected_Jo");
}

// -----------------------------------------------------------------------------
TEST_CASE(op_obs_file,test_tl) {
  TEST_MESSAGE(
      "Test Tangent linear observation operator using a trajectory from files");

  State_ xx(state());
  Increment_ dx(increment());
  randomize (dx);

  std::shared_ptr<Observations_> H_x = Hmult3D(xx, true);
  std::shared_ptr<Departures_> H_dx = Hmult3D(dx);

  double eps = getConfig().getDouble("tl_lambda_start");
  for (int lambda = 1; lambda <= getConfig().getInt("tl_nb_tests"); lambda++) {
    eps /= 10.0;

    Departures_ eps_H_dx(*H_dx);
    eps_H_dx *= eps;

    Increment_ eps_dx(dx);
    eps_dx *= eps;

    State_ x_plus_eps_dx(xx);
    x_plus_eps_dx += eps_dx;
    std::shared_ptr<Observations_> H_x_plus_eps_dx = Hmult3D(x_plus_eps_dx);
    Departures_ ratio( (*H_x_plus_eps_dx) - (*H_x));
    ratio /= eps_H_dx;

    //std::stringstream col;
    //TODO: col << "obs_diags_1@update_" << lambda;
    ratio.save(eckit::LocalConfiguration(getConfig(), "obs_diags"));
  }

  std::ostringstream oss;
  oss << "Check in ODB files ( Hnl(x+eps.dx) - Hnl(x) ) / eps.Htl(dx) stored in obs_diags_1@update_*";
  TEST_MESSAGE(oss.str());
}

// -----------------------------------------------------------------------------
TEST_CASE(op_obs_file,test_adjoint) {
  std::ostringstream oss;
  State_ xx(state());
  std::shared_ptr<Observations_> H_x = Hmult3D(xx, true);

  // Call H_tl
  CtrlInc_ dx1(jb());
  randomize (dx1);
  std::shared_ptr<Departures_> dy1 = Hmult3D(dx1);

  // Call H_ad
  CtrlInc_ dx2(dx1);
  dx2.zero();
  std::shared_ptr<Departures_> dy2(new Departures_(*dy1));
  randomize (*dy2);

  Htmult3D(dy2, dx2);

  // Check result
  double dx1dx2 = dot_product(dx1, dx2);
  double dy1dy2 = dot_product(*dy1, *dy2);
  double s = std::abs((dx1dx2 - dy1dy2) / (dx1dx2 + dy1dy2) * 2.0);
  double digits = - log10(s);

  oss << std::setprecision(20) << "dx1.dx2 = " << dx1dx2 << " dy1.dy2 = " << dy1dy2 << " digits = " << digits;
  TEST_MESSAGE(oss.str());
  checkDigits(digits);
}

}

#endif
