/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 */

#ifndef OOPS_TEST_SUITE_VARIATIONAL_FIXTURE_H_
#define OOPS_TEST_SUITE_VARIATIONAL_FIXTURE_H_

#include <sstream>
#include <iostream>
#include <memory>
#include <unordered_map>

#include "eckit/exception/Exceptions.h"
#include "eckit/config/LocalConfiguration.h"
#include "eckit/config/YAMLConfiguration.h"

#include "util/Logger.h"
#include "util/DateTime.h"
#include "util/Duration.h"

#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/JqTerm.h"
#include "oops/assimilation/DualVector.h"
#include "oops/assimilation/HMatrix.h"
#include "oops/assimilation/HtMatrix.h"
#include "oops/assimilation/State4D.h"
#include "oops/assimilation/Increment4D.h"
#include "oops/assimilation/CostJbTotal.h"
#include "oops/assimilation/CostJb4D.h"
#include "oops/base/GeneralizedDepartures.h"
#include "oops/base/Departures.h"
#include "oops/base/Observations.h"
#include "oops/base/Observer.h"
#include "oops/base/ObserverTL.h"
#include "oops/base/ObserverAD.h"
#include "oops/base/StateInfo.h"
#include "oops/base/TrajectorySaverMF.h"
#include "oops/interface/Increment.h"
#include "oops/interface/State.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Variables.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/base/PostProcessorFactory.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/base/GeneralizedDepartures.h"
#include "oops/base/Advector.h"
#include "oops/interface/Interpolator.h"
#include "oops/base/ObsErrors.h"
#include "oops/interface/ModelAuxControl.h"
#include "oops/interface/ModelAuxCovariance.h"
#include "oops/interface/ModelAuxIncrement.h"
#include "oops/interface/LinearModel.h"
#include "oops/interface/Model.h"
#include "oops/interface/ObsAuxCovariance.h"
#include "oops/interface/ObsVector.h"
#include "oops/interface/Localization.h"
#include "oops/interface/GenericMatrix.h"
#include "oops/base/Accumulator.h"
#include "oops/base/WeightedMean.h"
#include "oops/base/Ensemble.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ObsOperators.h"
#include "oops/base/ObsAuxIncrements.h"
#include "oops/base/LinearObsOperators.h"
#include "oops/base/ObservationSpaces.h"
#include "oops/base/ObserverEnVarTL.h"
#include "oops/base/ObserverEnVarAD.h"

#include "oops/assimilation/IncrCtlVec4D.h"
#include "oops/assimilation/Increment4D.h"
#include "oops/assimilation/State4D.h"
#include "oops/assimilation/CostJb4D.h"
#include "oops/assimilation/CostJb3D.h"

#include "oops/assimilation/instantiateCostFactory.h"
#include "oops/assimilation/instantiateMinFactory.h"
#include "oops/base/instantiateCovarFactory.h"
#include "oops/base/instantiatePostProcessorStateFactory.h"

namespace oops {
  template<typename MODEL> class JqTerm;
}

namespace test {

/*!
 *  This class gives access to input datas for the whole test suite. These
 *  datas are calculated once.
 */
template<typename MODEL> class TestSuiteVariationalFixture {
public:
  using Config_ = eckit::LocalConfiguration ;
  using Geometry_ = oops::Geometry<MODEL> ;
  using Variables_ = oops::Variables<MODEL> ;
  using ModelAuxCtrl_ = oops::ModelAuxControl<MODEL> ;
  using ModelAuxCovar_ = oops::ModelAuxCovariance<MODEL> ;
  using ModelAuxIncrement_ = oops::ModelAuxIncrement<MODEL> ;
  using LinearModel_ = oops::LinearModel<MODEL> ;
  using Model_ = oops::Model<MODEL> ;
  using ObsAuxIncr_ = oops::ObsAuxIncrements<MODEL> ;
  using ObsAuxCtrl_ = oops::ObsAuxControls<MODEL> ;
  using ObsAuxCovar_ = oops::ObsAuxCovariance<MODEL> ;
  using Increment_ = oops::Increment<MODEL> ;
  using Increment4D_ = oops::Increment4D<MODEL> ;
  using State_ = oops::State<MODEL> ;
  using State4D_ = oops::State4D<MODEL> ;
  using IncrCtlVec4D_ = oops::IncrCtlVec4D<MODEL> ;
  using IncrCtlVec_ = oops::IncrCtlVec<MODEL> ;
  using JbTotal_ = oops::CostJbTotal<MODEL> ;
  using JbState_ = oops::CostJbState<MODEL> ;
  using CostJb3D_ = oops::CostJb3D<MODEL> ;
  using CostJb4D_ = oops::CostJb4D<MODEL> ;
  using Departures_ = oops::Departures<MODEL> ;
  using Observations_ = oops::Observations<MODEL> ;
  using ObsVector_ = oops::ObsVector<MODEL> ;
  using ObsSpaces_ = oops::ObservationSpaces<MODEL> ;
  using ObsOperators_ = oops::ObsOperators<MODEL> ;
  using LinearObsOperators_ = oops::LinearObsOperators<MODEL> ;
  using ObsErrors_ = oops::ObsErrors<MODEL> ;
  using ModelSpaceCovarianceBase_ = oops::ModelSpaceCovarianceBase<MODEL> ;
  using CtrlInc_ = oops::ControlIncrement<MODEL> ;
  using ControlVariable_ = oops::ControlVariable<MODEL> ;
  using DepartVector_ = oops::DualVector<MODEL> ;
  using HMatrix_ = oops::HMatrix<MODEL> ;
  using HtMatrix_ = oops::HtMatrix<MODEL> ;
  using CostFunction_ = oops::CostFunction<MODEL> ;
  using CostFactory_ = oops::CostFactory<MODEL> ;
  using TrajectorySaver_ = oops::TrajectorySaverMF<MODEL> ;
  using Interpolator_ = oops::Interpolator<MODEL> ;  
  using Advector_ = oops::Advector<MODEL> ;
  using pAdvector_ = std::shared_ptr<Advector_> ;
  using Advectors_ = std::unordered_map<util::DateTime, pAdvector_> ;
  using GenericMatrix_ = oops::GenericMatrix<MODEL> ;
  using MatrixFactory_ = oops::GenericMatrixFactory<MODEL> ;
  using CovarianceFactory_ = oops::CovarianceFactory<MODEL> ;
  using PostProcessorTL_ = oops::PostProcessorTL<Increment_> ;
  using PostProcessorAD_ = oops::PostProcessorAD<Increment_> ;
  using ObserverTL_ = oops::ObserverTL<MODEL, Increment_> ;
  using PtrObserverTL_ = std::shared_ptr<ObserverTL_> ;
  using ObserverAD_ = oops::ObserverAD<MODEL, Increment_> ;
  using PtrObserverAD_ = std::shared_ptr<ObserverAD_> ;
  using PostProcessor_ = oops::PostProcessor<State_> ;
  using PostProcessorIncr_ = oops::PostProcessor<Increment_> ;
  using Observer_ = oops::Observer<MODEL, State_> ;
  using Localization_ = oops::Localization<MODEL> ;
  using Accumulator_ = oops::Accumulator<MODEL, Increment_, State_> ;
  using WeightedMean_ = oops::WeightedMean<MODEL, State_> ;

  using Ensemble_ = oops::Ensemble<MODEL> ;
  using pEnsemble_ = std::shared_ptr<Ensemble_> ;
  using EnsemblesCollection_ = oops::EnsemblesCollection<MODEL> ;
  using Variances_ = oops::Accumulator<MODEL, State_, State_> ;
  using Diag_ = oops::Accumulator<MODEL, State_, State_> ;
  using Observer4DTL_ = oops::ObserverEnVarTL<MODEL, Increment_> ;
  using Observer4DAD_ = oops::ObserverEnVarAD<MODEL, Increment_> ;

  // -----------------------------------------------------------------------------
  TestSuiteVariationalFixture() {
    //TEST_MESSAGE("Ctor TestSuiteVariationalFixture");
  }

  // -----------------------------------------------------------------------------
  ~TestSuiteVariationalFixture() {
    //TEST_MESSAGE("Dtor TestSuiteVariationalFixture");
  }

  // -----------------------------------------------------------------------------
  const util::DateTime & start() {
    return *getInstance().start_;
  }

  // -----------------------------------------------------------------------------
  const util::DateTime & end() {
    return *getInstance().end_;
  }

  // -----------------------------------------------------------------------------
  const util::Duration & length() {
    return *getInstance().fclen_;
  }

  // -----------------------------------------------------------------------------
  const State_ & state() {
    return *getInstance().state_;
  }

  // -----------------------------------------------------------------------------
  const Increment_ & increment() {
    return *getInstance().increment_;
  }

  // -----------------------------------------------------------------------------
  const ModelSpaceCovarianceBase_ & B() {
    return *getInstance().B_;
  }

  // -----------------------------------------------------------------------------
  const JbTotal_ & jb() {
    return *getInstance().jb_;
  }

  // -----------------------------------------------------------------------------
  const Config_ & getConfig() {
    return *getInstance().config_;
  }

  // -----------------------------------------------------------------------------
  const ModelAuxCtrl_ & getModelErr() {
    return *getInstance().moderr_;
  }

  // -----------------------------------------------------------------------------
  const Model_ & getModel() {
    return *getInstance().model_;
  }

  // -----------------------------------------------------------------------------
  std::vector<std::unique_ptr<LinearModel_> > & TLM() {
    return getInstance().tlm_;
  }

  // -----------------------------------------------------------------------------
  LinearModel_ & getTLM(const unsigned i = 0) {
    return *getInstance().tlm_[i];
  }

  // -----------------------------------------------------------------------------
  const Geometry_ & getResolution() {
    return *getInstance().resol_;
  }

  // -----------------------------------------------------------------------------
  const Variables_ & getLinearVariables() {
    return *getInstance().linvar_;
  }

  // -----------------------------------------------------------------------------
  const ObsSpaces_ & obspace() {
    return *getInstance().obspace_;
  }

  // -----------------------------------------------------------------------------
  const Observations_ & yobs() {
    return *getInstance().obs_;
  }

  // -----------------------------------------------------------------------------
  const ObsOperators_ & hop() {
    return *getInstance().hop_;
  }

  // -----------------------------------------------------------------------------
  const std::shared_ptr<LinearObsOperators_> & hoptlad() {
    return getInstance().hoptlad_;
  }

  // -----------------------------------------------------------------------------
  const ObsAuxCtrl_ & obias() {
    return *getInstance().obsbias_;
  }

  // -----------------------------------------------------------------------------
  void checkResult(double result, const char result_key[]) {
    char * expected_result_env = ::getenv("EXPECTED_RESULT");
    char * expected_value_env = ::getenv("EXPECTED_VALUE");
    char * expected_config_env = ((expected_value_env != 0) ? expected_value_env : expected_result_env  );    ASSERT(expected_config_env != 0);
    std::istringstream expected_config(expected_config_env);
    const eckit::YAMLConfiguration config(expected_config);

    double expected_result = config.getDouble(result_key);

    if (config.has("significant_digits")) {
      double expected_significant_digits = config.getDouble(
          "significant_digits");
      double s = std::abs((result - expected_result) / expected_result);
      double significant_digits = -log10(s);

      std::ostringstream oss;
      oss << std::setprecision(20) << "Expected result = " << expected_result
          << " Digits: " << significant_digits;
      TEST_MESSAGE(oss.str());
      EXPECT(significant_digits >= expected_significant_digits);
    } else {
      double tolerance = 1.0e-3;
      if (config.has("tolerance"))
        tolerance = config.getDouble("tolerance");
      EXPECT(std::abs(result - expected_result) < tolerance);
    }
  }

  // -----------------------------------------------------------------------------
  void checkDigits(double significant_digits) {
    char * expected_result_env = ::getenv("EXPECTED_RESULT");
    char * expected_value_env = ::getenv("EXPECTED_VALUE");
    char * expected_config_env = ((expected_value_env != 0) ? expected_value_env : expected_result_env  );    ASSERT(expected_config_env != 0);
    ASSERT(expected_config_env != 0);
    std::istringstream expected_config(expected_config_env);
    const eckit::YAMLConfiguration config(expected_config);

    double expected_significant_digits = config.getDouble("significant_digits");

    EXPECT(significant_digits >= expected_significant_digits);
  }

  // -----------------------------------------------------------------------------
  void randomize(Departures_ & dy) {
    ObsErrors_ R(obspace());
    R.linearize(yobs());
    R.randomize(dy);
  }

  // -----------------------------------------------------------------------------
  void randomize(Increment_ & dx) {
    B().randomize(dx);
  }

  // -----------------------------------------------------------------------------
  void randomize(CtrlInc_ & dx) {
    jb().randomize(dx);
  }

  // -----------------------------------------------------------------------------
  std::shared_ptr<Observations_> Hmult3D(const State_ & xx,
      const bool with_traj = false) {
    util::LogbookWriter<std::string> log("Forecast","Nonlinear HR trajectory");
    oops::PostProcessor<State_> post;
    std::shared_ptr < TrajectorySaver_ > trajectorySaver;

    util::Duration tslot("PT1H");
    std::shared_ptr<Observer_> pp;
    if (with_traj) {
      pp.reset(
          new Observer_(obspace(), hop(), obias(), tslot, false, hoptlad()));
    } else {
      pp.reset(new Observer_(obspace(), hop(), obias(), tslot));
    }
    post.enrollProcessor(pp);

    if (with_traj) {
      TLM().clear();
      const Config_ tlmConf(getConfig(), "linear_model");
      trajectorySaver.reset(
              new TrajectorySaver_(xx, getModelErr(), getResolution(), TLM(),
                  tlmConf));
      post.enrollProcessor(trajectorySaver);
    }

    post.initialize(xx, end(), tslot);
    post.process(xx);
    post.finalize(xx);

    std::shared_ptr<Observations_> yequ(pp->release());

    return yequ;
  }

  // -----------------------------------------------------------------------------
  std::shared_ptr<Observations_> Hmult(const State_ & xx, const bool with_traj =
      false) {
    oops::PostProcessor<State_> post;
    std::shared_ptr < TrajectorySaver_ > trajectorySaver;
    State_ xstart(xx);

    util::Duration tslot(getConfig().getString("timeslot"));
    std::shared_ptr<Observer_> pp;
    if (with_traj) {
      pp.reset(
          new Observer_(obspace(), hop(), obias(), tslot, false, hoptlad()));
    } else {
      pp.reset(new Observer_(obspace(), hop(), obias(), tslot));
    }
    post.enrollProcessor(pp);

    if (with_traj) {
      TLM().clear();
      const Config_ tlmConf(getConfig(), "linear_model");
      trajectorySaver.reset(
              new TrajectorySaver_(xstart, getModelErr(), getResolution(),
                  TLM(), tlmConf));
      post.enrollProcessor(trajectorySaver);
    }

    getModel().forecast(xstart, getModelErr(), length(), post);
    std::shared_ptr<Observations_> yequ(pp->release());

    return yequ;
  }

  // -----------------------------------------------------------------------------
  std::shared_ptr<Departures_> Hmult(const Increment_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorTL_ cost;
    ModelAuxIncrement_ modaux(getResolution(), getConfig());
    const Config_ biasConfig(getConfig(), "ObsBias");
    ObsAuxIncr_ obsaux(obspace(), biasConfig);
    util::Duration tslot(getConfig().getString("timeslot"));
    Increment_ dx0(dx);

    PtrObserverTL_ obstl(
        new ObserverTL_(obspace(), *hoptlad(), obsaux, tslot, false));
    cost.enrollProcessor(obstl);

    getTLM().forecastTL(dx0, modaux, length(), post, cost);

    std::shared_ptr<Departures_> dy(obstl->releaseOutputFromTL());
    return dy;
  }

  // -----------------------------------------------------------------------------
  std::shared_ptr<Departures_> Hmult(const CtrlInc_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorTL_ cost;
    util::Duration tslot(getConfig().getString("timeslot"));
    Increment_ dx0(dx.state()[0]);

    PtrObserverTL_ obstl(
        new ObserverTL_(obspace(), *hoptlad(), dx.obsVar(), tslot, false));
    cost.enrollProcessor(obstl);
    getTLM().forecastTL(dx0, dx.modVar(), length(), post, cost);

    std::shared_ptr<Departures_> dy(obstl->releaseOutputFromTL());
    return dy;
  }

  // -----------------------------------------------------------------------------
  std::shared_ptr<Departures_> Hmult3D(const Increment_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorTL_ cost, no_cost;
    ModelAuxIncrement_ modaux(getResolution(), getConfig());
    const Config_ biasConfig(getConfig(), "ObsBias");
    ObsAuxIncr_ obsaux(obspace(), biasConfig);
    util::Duration tslot("PT1H");
    Increment_ dx0(dx);

    PtrObserverTL_ obstl(
        new ObserverTL_(obspace(), *hoptlad(), obsaux, tslot, false));
    cost.enrollProcessor(obstl);

    getTLM().forecastTL(dx0, modaux, util::Duration(0), post, no_cost, true);
    cost.initializeTL(dx0, dx0.validTime(), util::Duration(0));
    cost.finalizeTL(dx0);

    std::shared_ptr<Departures_> dy(obstl->releaseOutputFromTL());

    return dy;
  }

  // -----------------------------------------------------------------------------
  std::shared_ptr<Departures_> Hmult3D(const CtrlInc_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorTL_ cost, no_cost;
    util::Duration tslot("PT1H");
    Increment_ dx0(dx.state()[0]);

    PtrObserverTL_ obstl(
        new ObserverTL_(obspace(), *hoptlad(), dx.obsVar(), tslot, false));
    cost.enrollProcessor(obstl);

    getTLM().forecastTL(dx0, dx.modVar(), util::Duration(0), post, no_cost, true);
    cost.initializeTL(dx0, dx0.validTime(), util::Duration(0));
    cost.finalizeTL(dx0);

    std::shared_ptr<Departures_> dy(obstl->releaseOutputFromTL());

    return dy;
  }

  // -----------------------------------------------------------------------------
  void Htmult3D(std::shared_ptr<Departures_> dy, Increment_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorAD_ cost, no_cost;
    ModelAuxIncrement_ modaux(getResolution(), getConfig());
    const Config_ biasConfig(getConfig(), "ObsBias");
    ObsAuxIncr_ obsaux(obspace(), biasConfig);
    util::Duration tslot("PT1H");

    PtrObserverAD_ obsad(
        new ObserverAD_(obspace(), *hoptlad(), dy, obsaux, tslot, false));
    cost.enrollProcessor(obsad);

    getTLM().forecastAD(dx, modaux, util::Duration(0), post, no_cost, true);
    cost.initializeAD(dx, dx.validTime(), util::Duration(0));
    cost.finalizeAD(dx);
  }

  // -----------------------------------------------------------------------------
  void Htmult3D(std::shared_ptr<Departures_> dy, CtrlInc_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorAD_ cost, no_cost;
    util::Duration tslot("PT1H");

    PtrObserverAD_ obsad(
        new ObserverAD_(obspace(), *hoptlad(), dy, dx.obsVar(), tslot, false));
    cost.enrollProcessor(obsad);

    getTLM().forecastAD(dx.state()[0], dx.modVar(), util::Duration(0), post, no_cost, true);
    cost.initializeAD(dx.state()[0], dx.state()[0].validTime(), util::Duration(0));
    cost.finalizeAD(dx.state()[0]);
  }

  // -----------------------------------------------------------------------------
  void Htmult(std::shared_ptr<Departures_> dy, CtrlInc_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorAD_ cost;
    util::Duration tslot(getConfig().getString("timeslot"));
    dx.state()[0].updateTime(length());

    PtrObserverAD_ obsad(
        new ObserverAD_(obspace(), *hoptlad(), dy, dx.obsVar(), tslot, false));
    cost.enrollProcessor(obsad);

    getTLM().forecastAD(dx.state()[0], dx.modVar(), length(), post, cost);
  }

  // -----------------------------------------------------------------------------
  void Htmult(std::shared_ptr<Departures_> dy, Increment_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorAD_ cost;
    ModelAuxIncrement_ modaux(getResolution(), getConfig());
    const Config_ biasConfig(getConfig(), "ObsBias");
    ObsAuxIncr_ obsaux(obspace(), biasConfig);
    util::Duration tslot(getConfig().getString("timeslot"));
    dx.updateTime(length());

    PtrObserverAD_ obsad(
        new ObserverAD_(obspace(), *hoptlad(), dy, obsaux, tslot, false));
    cost.enrollProcessor(obsad);

    getTLM().forecastAD(dx, modaux, length(), post, cost);
  }

  // -----------------------------------------------------------------------------
  void propagate(State_ & xx, const bool with_traj = false) {
    oops::PostProcessor<State_> post;
    std::shared_ptr < TrajectorySaver_ > trajectorySaver;

    std::vector<eckit::LocalConfiguration> postConfigs;
    getConfig().get("post_processors", postConfigs);
    for (size_t jj = 0; jj < postConfigs.size(); ++jj) {
      oops::PostBase<State_> * pp = oops::PostProcessorFactory<State_>::create(
          postConfigs[jj]);
      post.enrollProcessor(pp);
    }

    if (with_traj) {
      TLM().clear();
      const Config_ tlmConf(getConfig(), "linear_model");
      trajectorySaver.reset(
           new TrajectorySaver_(xx, getModelErr(), getResolution(),
                TLM(), tlmConf));
      post.enrollProcessor(trajectorySaver);
    }
    getModel().forecast(xx, getModelErr(), length(), post);

    return;
  }

  // -----------------------------------------------------------------------------
  void propagate(Increment_ & dx) {
    oops::PostProcessor<Increment_> post;
    PostProcessorTL_ cost;
    ModelAuxIncrement_ modaux(getResolution(), getConfig());
    getTLM().forecastTL(dx, modaux, length(), post, cost);
    return;
  }

  // -----------------------------------------------------------------------------
  void propagate_ad(Increment_ & dx) {
    dx.updateTime(length());
    oops::PostProcessor<Increment_> post;
    PostProcessorAD_ cost;
    ModelAuxIncrement_ modaux(getResolution(), getConfig());
    getTLM().forecastAD(dx, modaux, length(), post, cost);
    return;
  }

// -----------------------------------------------------------------------------
  void setup(const eckit::Configuration & fullConfig) {
    config_.reset(new eckit::LocalConfiguration(fullConfig));

    oops::instantiateCostFactory<MODEL>();
    oops::instantiateCovarFactory<MODEL>();
    oops::instantiateMinFactory<MODEL>();
    oops::instantiatePostProcessorStateFactory<MODEL>();

    fclen_.reset(new util::Duration(fullConfig.getString("window_length")));
    start_.reset(new util::DateTime(fullConfig.getString("window_begin")));
    end_.reset(new util::DateTime(*start_));
    *end_ += *fclen_;

    //  Setup resolution
    const Config_ resolConfig(fullConfig, "resolution");
    resol_.reset(new Geometry_(resolConfig));

    //  Setup variables
    const Config_ linearVariablesConfig(fullConfig, "linear_variables");
    linvar_.reset(new Variables_(linearVariablesConfig));

    //  Setup model
    const Config_ modelConfig(fullConfig, "model");
    model_.reset(new Model_(*resol_, modelConfig));
    moderr_.reset(new ModelAuxCtrl_(*resol_, *model_, fullConfig));

    //  Setup variables
    const Config_ variablesConfig(fullConfig, "variables");
    const Variables_ var(variablesConfig);

    // Setup State
    const Config_ stateConf(fullConfig, "state");
    state_.reset(new State_(*resol_, *model_, stateConf));

    // Setup Observations
    if (fullConfig.has("cost_function.Jo")) {
      const Config_ obsConfig(fullConfig, "cost_function.Jo");
      obspace_.reset(
          new ObsSpaces_(obsConfig, *resol_, *start_, *end_, obsConfig.has("screened")));
      hop_.reset(new ObsOperators_(*obspace_));
      hoptlad_.reset(new LinearObsOperators_(*obspace_));

      obs_.reset(new Observations_(*obspace_));
      obs_->read(obsConfig);

      if (!fullConfig.has("cost_function.Jb")) {
        const Config_ biasConfig(fullConfig, "ObsBias");
        obsbias_.reset(new ObsAuxCtrl_(*obspace_, biasConfig));      
      }
    }
    
    // Setup standalone Bmat
    if (fullConfig.has("Bmat")) {
      const Config_ jbConf(fullConfig, "Bmat");
      B_.reset(
          oops::CovarianceFactory<MODEL>::create(jbConf, *resol_, var,
              *state_));
      B_->linearize(*state_, *resol_, jbConf);
    }

    // Setup Jb
    if (fullConfig.has("cost_function.Jb")) {
      const Config_ jConf(fullConfig, "cost_function");
      const Config_ jbConf(jConf, "Jb");
      xx_.reset(new ControlVariable_(jConf, *resol_, *model_, *obspace_));
      jb_.reset(new JbTotal_(*xx_, new CostJb3D_(jbConf, *resol_, *linvar_, *fclen_,
          xx_->state()[0]), jConf, *resol_, *obspace_));
      oops::JqTerm<MODEL> * jq = jb_->initializeTraj(*xx_, *resol_, jConf);
      jb_->finalizeTraj(jq, jbConf);
      obsbias_.reset(new ObsAuxCtrl_(*obspace_, jConf));
    }

    increment_.reset(new Increment_(*resol_, *linvar_, *start_));
    increment_->zero();

  }

private:

  /* This singleton must be attribute of the Run object, so that
   * all objects will be destroyed in the right order. Especially before
   * MPI finalizing. */
  TestSuiteVariationalFixture & getInstance();

  std::unique_ptr<util::Duration> fclen_;
  std::unique_ptr<util::DateTime> start_;
  std::unique_ptr<util::DateTime> end_;
  std::unique_ptr<Config_> config_;
  std::unique_ptr<Config_> expectedResults_;
  std::unique_ptr<Geometry_> resol_;
  std::unique_ptr<Variables_> linvar_;
  std::unique_ptr<Model_> model_;

  std::vector<std::unique_ptr<LinearModel_> > tlm_;

  std::unique_ptr<ControlVariable_> xx_;
  std::unique_ptr<State_> state_;
  std::unique_ptr<Increment_> increment_;
  std::unique_ptr<ModelSpaceCovarianceBase_> B_;
  std::unique_ptr<JbTotal_> jb_;
  std::unique_ptr<ModelAuxCtrl_> moderr_;

  std::unique_ptr<ObsSpaces_> obspace_;
  std::unique_ptr<Observations_> obs_;
  std::unique_ptr<ObsOperators_> hop_;
  std::unique_ptr<ObsAuxCtrl_> obsbias_;
  std::shared_ptr<LinearObsOperators_> hoptlad_;

};

template<typename MODEL>
TestSuiteVariationalFixture<MODEL> & TestSuiteVariationalFixture<MODEL>::getInstance() {
  extern TestSuiteVariationalFixture<MODEL>& getGlobalFixture();
  return getGlobalFixture();
}

}
#endif
