/*
 * (C) Copyright 1996- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef TestSuite_h
#define TestSuite_h

#include <fstream>
#include <numeric>
#include <random>
#include <iomanip>
#include <string>
#include <iostream>
#include <unistd.h>
#include <memory>
#include <unordered_map>

#include "eckit/filesystem/LocalPathName.h"
#include "eckit/log/Log.h"
#include "eckit/mpi/Comm.h"
#include "eckit/types/FloatCompare.h"
#include "eckit/types/Types.h"
#include "eckit/testing/Test.h"

namespace testsuite {

//----------------------------------------------------------------------------------------------------------------------
/// A test is defined by a description and a function pointer.

class TestSuite {

  public:  // methods
    TestSuite(const std::string& description, std::function<void()> testFn) :
      description_(description), testFn_(std::move(testFn)) {}

    bool run() {
      bool success = true;
      try {
        testFn_();
      }
      catch (eckit::testing::TestException& e) {
        success = false;
        eckit::Log::error()  << "Test \"" << description() << "\" failed: " << e.what()
                                            << " @ " << e.location()  << std::endl;
      }
      catch (eckit::Exception& e) {
        success = false;
        eckit::Log::error()  << "Test \"" << description()
                                            << "\" failed with unhandled eckit::Exception: " << e.what() << " @ "
                                            << e.location()  << std::endl;
        eckit::Log::error() << "    Stack trace: " << e.callStack() << std::endl;
      }
      catch (std::exception& e) {
        success = false;
        eckit::Log::error()  << "Test \"" << description()
                                            << "\" failed with unhandled exception: " << e.what()
                                            << std::endl;
      }
      catch (...) {
        success = false;
        eckit::Log::error()  << "Test \"" << description()
                                            << "\" failed with unknown unhandled exception."
                                            << std::endl;
      }

      return success;
    }

    std::string description() const {
        return description_;
    }

  private:  // members
    std::string description_;

    std::function<void()> testFn_;
};

///----------------------------------------------------------------------------------------------------------------------
std::unordered_map<std::string, std::unique_ptr<TestSuite> > & get_testsuite() {
  static std::unordered_map<std::string, std::unique_ptr<TestSuite> > tests;
  return tests;
}

/// Tests are registered using a static object, that is initialised before main()
class TestSuiteRegister {
  public:
    TestSuiteRegister(const std::string& family, const std::string& testid, void (*testFn)()) {
      std::ostringstream description;
      description << family << "/" << testid;
      get_testsuite()[description.str()].reset(new TestSuite(description.str(), testFn));
    }
};

/// Launch test method
std::string init(int argc, char* argv[]) {
  int opt;
  std::string testid;

  // Parse arguments
  ::setenv("AUTO_LOAD_PLUGINS", "false", true);
  std::string configfile=argv[1];
  while ( (opt = getopt(argc, argv, "tlh")) != -1 ) {
    switch ( opt ) {
      case 't':
        testid = std::string(argv[optind]);
        break;
      case 'l':
        eckit::Log::info() << "List of available tests : " <<  std::endl;
        for( const auto& n : get_testsuite() ) {
          eckit::Log::info() << "   " << n.first << std::endl;
        }
        exit(0);
        break;
      case 'h':
        eckit::Log::info() << "Usage : -t <test_type> : Execute <test_type> test" <<  std::endl;
        eckit::Log::info() << "        -l             : Liste available tests " <<  std::endl;
        eckit::Log::info() << "        -h             : Usage message " <<  std::endl;
        exit(0);
        break;
    }
  }

  return testid;
}

//----------------------------------------------------------------------------------------------------------------------

}  // namespace testsuite

//Macros
#define FIXTURE_TEST_SUITE_( F , M )            \
  typedef F TestSuiteFixture;                   \
  typedef M MODEL;



#define TEST_CASE(family, testid)                                                          \
    template< typename MODEL >                                                             \
    struct test_##family##testid : public TestSuiteFixture  { void test_method(); };       \
    void fn_##family##testid() {                                                           \
        test_##family##testid< MODEL > t; t.test_method();}                                \
    static testsuite::TestSuiteRegister test_registration_##family##testid(#family,#testid,\
        &fn_##family##testid);                                                             \
    template< typename MODEL >                                                             \
    void test_##family##testid< MODEL >::test_method()

#define TEST_MESSAGE(msg) \
    eckit::Log::info() << "<Message file=\"" << __FILE__ << "\" line=\"" << __LINE__ << "\">"  \
       << "<![CDATA[" << msg << "]]></Message>" << std::endl;



#endif  // TestSuite_h
