/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef UTIL_LOGBOOKHELPER_H_
#define UTIL_LOGBOOKHELPER_H_

#include <functional>
#include <iomanip>
#include <map>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/filesystem/PathName.h"
#include "eckit/memory/NonCopyable.h"
#include "util/Logbook.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace util {

// -----------------------------------------------------------------------------

/*! @brief LogbookHelper

 * The LogbookHelper is a static class that holds an instance of a Logbook;
 * LogbookHelper is used by the LogbookWriter to record logs:
 *  - LogbookHelper::startRecording(..) method is called by the constructor of
 *     the LogbookWriter class;
 *  - LogbookHelper::finishRecording(..) method is called by the destructor of
 *     the LogbookWriter
 * LogbookHelper allows to connect callbacks (e.g. coming from a Model), which
 * are invoked when writing logs in the Logbook.
 */

class LogbookHelper : public util::Printable, private eckit::NonCopyable {
  using cit_ = std::map<std::string, int>::const_iterator;
  using CallbackFunction_ =
      std::function<void(const eckit::Configuration &conf)>;
  using Callbacks_ = std::vector<CallbackFunction_>;

 public:
  enum Execution : bool { Active = true, NotActive = false };

  enum Callback : bool { Call = true, DontCall = false };

  /// Initialization, finalization
  static void start();
  static void stop();

  /// Utilities
  template <typename ValueT>
  static void startRecording(const std::string &, const ValueT &);
  static void finishRecording(const std::string &);

  /// Connect callback - allow models to connect to the LogbookHelper
  static void connectCallback(const CallbackFunction_ &cb);

  /// Accessors
  static bool has(const std::string &s);
  template <typename ValueT>
  static void query(const std::string &s, ValueT &v);
  template <typename ValueT>
  static void record(const std::string &s, const ValueT &v,
                     const bool callbacks = false);

  /// Interfacing
  static const eckit::Configuration &getConstLogbook();
  static eckit::LocalConfiguration &getLogbook();

  ~LogbookHelper();

 private:
  LogbookHelper();
  static LogbookHelper &getInstance();
  const Callbacks_ &getCallbacks() { return callbacks_; }
  void print(std::ostream &os) const;

  bool on_;
  std::map<std::string, bool> status_;
  std::map<std::string, int> counts_;
  Callbacks_ callbacks_;
  std::unique_ptr<Logbook> logbook_;
};

// -----------------------------------------------------------------------------

template <typename ValueT>
void LogbookHelper::query(const std::string &s, ValueT &v) {
  ASSERT(getInstance().logbook_);
  return getInstance().logbook_->get<ValueT>(s, v);
}

// -----------------------------------------------------------------------------

template <typename ValueT>
void LogbookHelper::record(const std::string &s, const ValueT &v,
                           const bool callbacks) {
  ASSERT(getInstance().logbook_);
  getInstance().logbook_->set<ValueT>(s, v);
  // Call callback functions
  if (callbacks)
    for (auto const &cb : getInstance().getCallbacks()) {
      cb(getConstLogbook());
    }
}

// -----------------------------------------------------------------------------

template <typename ValueT>
void LogbookHelper::startRecording(const std::string &name, const ValueT &val) {
  if (getInstance().on_) {
    getInstance().status_[name] = Execution::Active;
    getInstance().record<bool>(name + "_active", Execution::Active,
                               Callback::DontCall);
    getInstance().record<ValueT>(name, val, Callback::Call);
  }
}

// -----------------------------------------------------------------------------

}  // namespace util

#endif  // UTIL_LOGBOOKHELPER_H_
