/*
 * Copyright 2011 ECMWF
 *
 * This software was developed at ECMWF for evaluation
 * and may be used for academic and research purposes only.
 * The software is provided as is without any warranty.
 *
 * This software can be used, copied and modified but not
 * redistributed or sold. This notice must be reproduced
 * on each copy made.
 *
 *
 * This class performs MPI transfers with an serializable object
 * in multi-threading way.
 * It doesn't work with OpenMPI, but works with IntelMPI
 */
#ifndef UTIL_MPI_THREAD_POOL_H_
#define UTIL_MPI_THREAD_POOL_H_

#include <memory>
#include <string>

#include "eckit/io/Buffer.h"
#include "eckit/maths/Functions.h"
#include "eckit/serialisation/ResizableMemoryStream.h"
#include "eckit/thread/ThreadPool.h"
#include "util/MPIWrapper.h"

namespace util {

// -----------------------------------------------------------------------------
template <typename serializable>
class threaded_bcast : public eckit::ThreadPoolTask {
 public:
  threaded_bcast(unsigned ithread, serializable* dx)
      : ithread_(ithread), dx_(dx) {}
  void execute() {
    MPIWrapper& mpiw = MPIWrapper::Instance();
    eckit::Buffer b(32);
    eckit::ResizableMemoryStream s(b);

    if (mpiw.comm(ithread_).rank() == 0) s << *dx_;
    mpiw.comm(ithread_).broadcast(static_cast<char*>(b.data()), s.position(),
                                  (size_t)0);
    if (mpiw.comm(ithread_).rank() != 0) s >> *dx_;
  }

 private:
  const unsigned ithread_;
  serializable* dx_;
};

// -----------------------------------------------------------------------------
template <typename serializable>
class threaded_send_recv : public eckit::ThreadPoolTask {
 public:
  threaded_send_recv(unsigned ithread, eckit::Buffer* xsend,
                     serializable* xrecv, unsigned source, int tag)
      : ithread_(ithread),
        source_(source),
        tag_(tag),
        xsend_(xsend),
        xrecv_(xrecv) {}

  void execute() {
    MPIWrapper& mpiw = MPIWrapper::Instance();
    int n1 = mpiw.nb_worlds();
    int n2 = mpiw.nb_threads();
    int tag_send = (tag_ * n1 + mpiw.my_rank()) * n2 + ithread_;
    int tag_recv = (tag_ * n1 + source_) * n2 + ithread_;

    eckit::ResizableMemoryStream ss(*xsend_);
    eckit::mpi::Request req =
        mpiw.comm(ithread_).iSend(static_cast<const char*>(xsend_->data()),
                                  ss.position(), source_, tag_send);

    eckit::mpi::Status st = mpiw.comm(ithread_).probe(source_, tag_recv);
    size_t size = eckit::round(mpiw.comm(ithread_).template getCount<char>(st),
                               8);  //< round to nearest 8 bytes
    eckit::Buffer b(size);
    b.zero();
    mpiw.comm(ithread_).receive(static_cast<char*>(b.data()), b.size(), source_,
                                tag_recv);
    eckit::ResizableMemoryStream sr(b);
    sr >> *xrecv_;

    mpiw.comm(ithread_).wait(req);
  }

 private:
  const unsigned ithread_, source_, tag_;
  eckit::Buffer* xsend_;
  serializable* xrecv_;
};

// -----------------------------------------------------------------------------
template <typename serializable>
class MPIThreadPool {
 public:
  MPIThreadPool() {}

  void add(serializable* x1) { vec_x_.insert(vec_x_.begin(), x1); }

  void add(eckit::Buffer* x1, serializable* x2) {
    vec_arch_.insert(vec_arch_.begin(), x1);
    vec_x_.insert(vec_x_.begin(), x2);
  }

  void broadcast(unsigned source) {
    MPIWrapper& mpiw = MPIWrapper::Instance();

    if (mpiw.nb_threads() == 1) {
      // Mono-thread version ( for OpenMPI )

      for (unsigned i = 0; i < vec_x_.size(); i++) {
        threaded_bcast<serializable> bcast(0, vec_x_[i], source);
        bcast.execute();
      }
    } else {
      // Multi-threads version ( for IntelMPI )
      eckit::ThreadPool pool("p", mpiw.nb_threads());
      for (unsigned i = 0; i < vec_x_.size(); i++) {
        pool.push(new threaded_bcast<serializable>(i, vec_x_[i], source));
      }
      pool.waitForThreads();
    }
  }

  void send_recv_nb(unsigned source, int tag) {
    MPIWrapper& mpiw = MPIWrapper::Instance();

    if (mpiw.nb_threads() == 1) {
      // Mono-thread version ( for OpenMPI )

      for (unsigned i = 0; i < vec_x_.size(); i++) {
        threaded_send_recv<serializable> sendrecv(0, vec_arch_[i], vec_x_[i],
                                                  source, tag);
        sendrecv.execute();
      }
    } else {
      // Multi-threads version ( for IntelMPI )
      eckit::ThreadPool pool("p", mpiw.nb_threads());
      for (unsigned i = 0; i < vec_x_.size(); i++) {
        pool.push(new threaded_send_recv<serializable>(i, vec_arch_[i],
                                                       vec_x_[i], source, tag));
      }
      pool.waitForThreads();
    }
  }

 private:
  std::vector<serializable*> vec_x_;
  std::vector<eckit::Buffer*> vec_arch_;
};

}  // namespace util

#endif  // UTIL_MPI_THREAD_POOL_H_
