#include "util/MPIWrapper.h"

#include <stdlib.h>
#include <string.h>

#include <sstream>
#include <string>

#include "util/Logger.h"

// -----------------------------------------------------------------------------
namespace util {

// -----------------------------------------------------------------------------
MPIWrapper MPIWrapper::m_instance_ = MPIWrapper();

// -----------------------------------------------------------------------------
void MPIWrapper::setup(const eckit::Configuration& config) {
  int argc = 0;
  char** argv = 0;

  oops::Log::trace() << "Constructing MPIWrapper";

  nb_worlds_ = 1;
  nb_threads_ = 1;
  if (config.has("mpi")) {
    const eckit::LocalConfiguration conffMPI(config, "mpi");

    nb_worlds_ = conffMPI.getInt("number_of_ifs_worlds", 1);
    nb_threads_ = conffMPI.getInt("number_of_threads", 1);
  }

  // Define the global communicator
  eckit::mpi::Comm& all = eckit::mpi::comm();
  ASSERT((all.size() % nb_worlds_) == 0);

  unsigned int my_global_rank;
  int comm_size = all.size() / nb_worlds_;
  my_global_rank = all.rank();

  // Define a communicator for the model layer
  int num_world = my_global_rank / comm_size;
  std::ostringstream ossi;
  ossi << "comm_ifs_" << num_world;
  all.split(num_world, ossi.str().c_str());
  model_comm_ = ossi.str();
  my_model_rank_ = eckit::mpi::comm(model_comm_.c_str()).rank();
  comm_in_ftn_ = eckit::mpi::comm(model_comm_.c_str()).communicator();

  // Define communicators for the OOPS layer, one per threads
  int num_oops_world = nb_threads_ * (my_global_rank % comm_size);
  num_oops_world += nb_worlds_;
  for (unsigned int i = 0; i < nb_threads_; i++) {
    num_world = num_oops_world + i;
    std::ostringstream osso;
    osso << "comm_oops_" << num_world;
    all.split(num_world, osso.str().c_str());
    oops_comm_.push_back(osso.str().c_str());
  }
  my_rank_ = comm(0).rank();

  // TODO: To be moved in IFS
  // Direct profiling properly
  std::ostringstream oss;
  oss << "DR_HOOK_PROFILE=drhook." << my_global_rank << ".prof";
  putenv(strdup(oss.str().c_str()));
}

// -----------------------------------------------------------------------------
double MPIWrapper::finalize_dot_product(const double local_pdt) const {
  double global_pdt = local_pdt;

  if (nb_worlds_ == 1) return global_pdt;

  if (my_model_rank_ == 0) {
    comm().allReduceInPlace(global_pdt, eckit::mpi::sum());
  }

  eckit::mpi::comm(model_comm_.c_str()).broadcast(global_pdt, 0);
  return global_pdt;
}

// -----------------------------------------------------------------------------
MPIWrapper::MPIWrapper()
    : nb_threads_(1), nb_worlds_(1), my_model_rank_(0), my_rank_(0) {}

// -----------------------------------------------------------------------------
MPIWrapper::~MPIWrapper() {}

// -----------------------------------------------------------------------------
MPIWrapper& MPIWrapper::Instance() { return m_instance_; }

// -----------------------------------------------------------------------------
}  // namespace util
