/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#include "util/RestartHelper.h"

#include <fstream>
#include <iostream>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/config/YAMLConfiguration.h"
#include "eckit/filesystem/LocalPathName.h"
#include "eckit/parser/JSON.h"
#include "eckit/parser/JSONParser.h"
#include "util/Mpi.h"

using oops::Log;

namespace util {

// -----------------------------------------------------------------------------

RestartHelper& RestartHelper::getInstance() {
  static RestartHelper theHelper;
  return theHelper;
}

// -----------------------------------------------------------------------------

void RestartHelper::start() {
  Log::info() << "RestartHelper::start() Start Restart Logbook recording"
              << std::endl;

  if (getInstance().path_.exists()) {
    getInstance().isReadIn_ = true;
    Log::info() << "RestartHelper::RestartHelper Path : " << getInstance().path_
                << " exists." << std::endl;
    if (mpi::comm().rank() == getInstance().root_) {
      eckit::YAMLConfiguration yaml(getInstance().path_);
      getInstance().logbook_.reset(new Logbook(yaml));
    } else {
      getInstance().logbook_.reset(new Logbook());
    }
    getInstance().logbook_->broadcast(getInstance().root_);
  } else {
    Log::info() << "RestartHelper::RestartHelper Path : " << getInstance().path_
                << " does not exist." << std::endl;
    getInstance().logbook_.reset(new Logbook());
  }
  getInstance().on_ = true;

  Log::info() << *getInstance().logbook_ << std::endl;
}

// -----------------------------------------------------------------------------

void RestartHelper::stop() {
  Log::info() << getInstance() << std::endl;
  getInstance().logbook_.reset();
  getInstance().on_ = false;
}

// -----------------------------------------------------------------------------

void RestartHelper::reset() {
  ASSERT(getInstance().logbook_);
  getInstance().logbook_.reset(new Logbook());
  Log::info() << "RestartHelper::reset() done" << std::endl;
}

// -----------------------------------------------------------------------------

void RestartHelper::write() {
  mpi::comm().barrier();
  if (mpi::comm().rank() == getInstance().root_) {
    eckit::PathName uniquepath = eckit::PathName::unique(getInstance().path_);
    std::ofstream ofs(uniquepath.localPath());
    if (!ofs) throw eckit::CantOpenFile(uniquepath.localPath(), Here());
    eckit::JSON j(ofs);
    j << *(getInstance().logbook_);
    ofs.close();
    eckit::PathName::rename(uniquepath, getInstance().path_);
  }
}

// -----------------------------------------------------------------------------

bool RestartHelper::has(const std::string& s) {
  ASSERT(getInstance().logbook_);
  return getInstance().logbook_->has(s);
}

// -----------------------------------------------------------------------------

const bool& RestartHelper::isReadIn() {
  ASSERT(getInstance().on_);
  return getInstance().isReadIn_;
}

// -----------------------------------------------------------------------------

RestartHelper::RestartHelper()
    : on_(false), isReadIn_(false), root_(0), path_(), logbook_() {
  eckit::PathName path = eckit::LocalPathName::cwd() + "/logbook.json";
  path_ = path;
}

// -----------------------------------------------------------------------------

RestartHelper::~RestartHelper() {}

// -----------------------------------------------------------------------------

void RestartHelper::print(std::ostream& os) const {
  os << " " << std::endl;
  os << "---------------------------------------------------------------------"
     << std::endl;
  os << "------------------------- Restart Content ---------------------------"
     << std::endl;
  os << "---------------------------------------------------------------------"
     << std::endl;
  os << "Restart Logbook path" << path_ << std::endl;
  os << "---------------------------------------------------------------------"
     << std::endl;
  os << *logbook_ << std::endl;
  os << "------------------------- Restart Content End -----------------------"
     << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace util
