/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef UTIL_RESTARTHELPER_H_
#define UTIL_RESTARTHELPER_H_

#include <map>
#include <memory>  // for std::unique_ptr
#include <string>

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/filesystem/PathName.h"
#include "eckit/memory/NonCopyable.h"
#include "util/Logbook.h"
#include "util/Logger.h"
#include "util/Printable.h"

namespace util {

// -----------------------------------------------------------------------------

/*! @brief RestartHelper class.

 * RestartHelper is a static class that holds an instance of the Logbook class,
 *  It provides:
 *   - interfaces to record in/query the restart logbook
 *   - read/write the restart logbook from/to file
 *
 */

class RestartHelper : public util::Printable, private eckit::NonCopyable {
 public:
  /// Initialization, finalization
  static void start();
  static void stop();

  /// Utilities
  static void write();
  static const bool& isReadIn();

  /// Accessors
  static bool has(const std::string& s);
  template <typename ValueT>
  static void query(const std::string& s, ValueT& v);
  template <typename ValueT>
  static void record(const std::string& s, const ValueT& v,
                     const bool writeToDisc = false);

  /// Reset the logbook
  static void reset();

  ~RestartHelper();

 private:
  RestartHelper();
  static RestartHelper& getInstance();
  void print(std::ostream& os) const;

  bool on_;
  mutable bool isReadIn_;
  const size_t root_;
  mutable eckit::PathName path_;
  std::unique_ptr<Logbook> logbook_;
};

// -----------------------------------------------------------------------------

template <typename ValueT>
void RestartHelper::query(const std::string& s, ValueT& v) {
  ASSERT(getInstance().logbook_);
  return getInstance().logbook_->get<ValueT>(s, v);
}

// -----------------------------------------------------------------------------

template <typename ValueT>
void RestartHelper::record(const std::string& s, const ValueT& v,
                           const bool writeToDisc) {
  ASSERT(getInstance().logbook_);
  getInstance().logbook_->set<ValueT>(s, v);
  if (writeToDisc) getInstance().write();
}

// -----------------------------------------------------------------------------

}  // namespace util

#endif  // UTIL_RESTARTHELPER_H_
